# UPC Parser helper functions

import re

# Initialize all regular expressions
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #

regex_include = re.compile(r"""
                           [#]\s*?include   # Match '#include' directive
                           \s+              # Match whitespace
                           ["<]             # Match < or "
                           \s* (.*?)\.h \s* # Match header.h
                           [">]             # Match > or "
                           """, re.VERBOSE)
regex_identifier = re.compile(r"""
                              ([a-zA-Z_]\w*)    # Match a valid C identifier: alphanumerics, that don't start with a number
                              """, re.VERBOSE)

# Print verbose output.
# --------------------------------------------------------------------------- #
def print_output(state):
    """Prints output for this line from the list stored in the state object."""
    linenum = state.linenumber_out()
    outlist = state.output_get()
    state.output_reset()
    for s in outlist:
        print str(linenum) + ":\t" + s

# Determine the indices of the beginning and end of an enclosed expression, e.g. [ARRAY_LEN]
# --------------------------------------------------------------------------- #
def paired_separators_span(tokens, separators, index = 0, backwards = False):
    """Return the indices of a pair of separator tokens, given the index of one and a direction."""
    indices = None
    if not backwards:
        assert tokens[index] == separators[0]
        i = index+1
        while i < len(tokens) and tokens[i] != separators[1]:
            if tokens[i] in "{[(":
                tpl = any_separator_span(tokens, i)
                i = tpl[1]
            i = i+1
        indices = (index, i)
    else:
        assert tokens[index] == separators[1]
        i = index-1
        while i >= 0 and tokens[i] != separators[0]:
            if tokens[i] in "}])":
                tpl = any_separator_span(tokens, i)
                i = tpl[0]
            i = i-1
        indices = (i, index)
    return indices

# Finds the indices of an associated pair of square brackets.
# --------------------------------------------------------------------------- #
def brackets_span(tokens, index = 0, backwards = False):
    """Detects a pair of square brackets and returns a tuple with their indices in a list of tokens."""
    return paired_separators_span(tokens, ('[', ']'), index, backwards)

# Finds the indices of an associated pair of curly braces.
# --------------------------------------------------------------------------- #
def braces_span(tokens, index = 0, backwards = False):
    """Detects a pair of curly braces and returns a tuple with their indices in the list of tokens."""
    return paired_separators_span(tokens, ('{', '}'), index, backwards)

# Finds the indices of an associated pair of parentheses.
# --------------------------------------------------------------------------- #
def paren_span(tokens, index = 0, backwards = False):
    """Detects a pair of parentheses and returns a tuple with their indices in the list of tokens."""
    return paired_separators_span(tokens, ('(', ')'), index, backwards)

separator_map = {'{':('{','}'), '}':('{','}'), '[':('[',']'), ']':('[',']'), '(':('(',')'), ')':('(',')'),}

# Finds indices for an associated pair of separators, given the initial separator.
# --------------------------------------------------------------------------- #
def any_separator_span(tokens, first_separator_index = 0):
    """Determines separators and directions from starting index, and spans the separators."""
    separator = tokens[first_separator_index]
    backwards = False
    assert separator in "{[()]}"
    if separator in ")]}": backwards = True
    return paired_separators_span(tokens, separator_map[separator], first_separator_index, backwards)

# Find a full statement, given the first token.
# --------------------------------------------------------------------------- #
def statement_span(tokens, first_token = None, first_token_index = None):
    """Finds and returns a tuple containing the token indices spanning a full statement."""
    assert first_token_index != None or first_token != None
    expstart = 0
    if first_token_index != None:
        expstart = first_token_index
    else:
        expstart = tokens.index(first_token)
    return paired_separators_span(tokens, (tokens[expstart], ';'), expstart)

# Prepend a specified prefix to a token.
# --------------------------------------------------------------------------- #
def token_prefix(tokens, prefix, token = None, token_index = None):
    """Replace a token in the list with a prefixed version."""
    assert token_index != None or token != None
    t = token
    ti = token_index
    if ti != None:
        t = tokens[ti]
    elif t != None:
        ti = tokens.index(t)
    tokens[ti] = prefix + t

# Append a a postfix to a token.
# --------------------------------------------------------------------------- #
def token_append(tokens, postfix, token = None, token_index = None):
    """Replace a token in the list with a postfixed version."""
    assert token_index != None or token != None
    t = token
    ti = token_index
    if ti != None:
        t = tokens[ti]
    elif t != None:
        ti = tokens.index(t)
    tokens[ti] = t + postfix

# Construct a list of tokens for a global object.
# --------------------------------------------------------------------------- #
def global_build(statement, name, dtype, ptr = [], arr = [['1']]):
    """Constructs a token list for a global object declaration."""
    # Clean out original statement:
    del statement[0:len(statement)]
    # Check arr for defaults:
    if len(arr) < 1: arr = [['1']]
    elif len(arr[0]) < 1: arr[0] = ['1']
    # Start adding in the needed tokens:
    obj_decl = None
    obj_construct = None
    obj_args = None
    
    if arr[0][0] == '1' and len(arr[0]) == 1:
        # Looks like: upc_global<dtype *> name = new upc_global<dtype *>();
        obj_decl = ['upc_global', '<', dtype] + ptr + ['>', '::', 'type', name]
        obj_construct = ['=', 'new', 'upc_global', '<', dtype] + ptr + ['>', '::', 'type', '(']
        obj_args = [')']
    else:
        # Looks like: upc_global_arr<dtype *, array_dim> name = new upc_global_arr<dtype *, array_dim>();
        obj_decl = ['upc_global_arr', '<', dtype] + ptr + [','] + arr[0] + ['>', '::', 'type', name]
        obj_construct = ['=', 'new', 'upc_global_arr', '<', dtype] + ptr + [','] + arr[0] + ['>', '::', 'type', '(']
        obj_args = [')']
        
    tokens = obj_decl + obj_construct + obj_args
    tokens.append(';')
#    print tokens
    statement.extend(tokens)

# Construct a list of tokens for a upc_shared object.
# --------------------------------------------------------------------------- #
def upc_shared_build(statement, name, dtype, ptr = [], block = ['1'], arr = [['1']], ptr_to_shared = False, block2 = ['1']):
    """Constructs a token list for a upc_shared object declaration."""
    # Clean out original statement:
    del statement[0:len(statement)]
    # Check block and arr for defaults:
    if len(block) < 1: block = ['1']
    if len(block2) < 1: block2 = ['1']
    if len(arr) < 1: arr = [['1']]
    elif len(arr[0]) < 1: arr[0] = ['1']
    # Start adding in the needed tokens:
    
    ptr_type = None
    
    if ptr_to_shared:
        ptr_type = ['shared', '['] + block + [']', dtype] + ptr
    else:
        ptr_type = [dtype] + ptr;
        block2 = block
    
    # The number of blocks assigned 1 per thread in a round-robin fashion: array_dim / block2
    nblocks = []
    for i in xrange(len(arr)):
        nblocks.append('(')
        nblocks.extend(arr[i])
        nblocks.extend([')', '*'])
    nblocks[len(nblocks)-1] = '/'
    nblocks.append('(')
    nblocks.extend(block2)
    nblocks.append(')')
    
    # The number of bytes per block: block2 * sizeof(ptr_type)
    nbytes = ['('] + block2 + [')', '*', 'sizeof', '('] + ptr_type + [')']
    
    # Looks like:
    #   upc_shared::type::ptr sstmac_upc_shared_obj_name = upc_shared::type::construct(nblocks, nbytes);
    #   upc_shared_ptr<ptr_type, block2>::type name = sstmac_upc_shared_obj_name;
    obj_decl = ['upc_shared', '::', 'type', '::', 'ptr', 'sstmac_upc_shared_obj_' + name]
    obj_construct = ['=', 'upc_shared', '::', 'type', '::', 'construct', '(']
    obj_args = nblocks + [','] + nbytes + [')', ';']
    obj_ptr = ['upc_shared_ptr', '<'] + ptr_type + [','] + block2 + ['>', '::', 'type', name]
    obj_ptr_args = ['(', 'sstmac_upc_shared_obj_' + name, ')', ';']
    
    tokens = obj_decl + obj_construct + obj_args + ['\n'] + obj_ptr + obj_ptr_args
#    print tokens
    statement.extend(tokens)

# Construct a list of tokens for a global upc_shared_ptr object.
def upc_global_shared_ptr_build(statement, name, dtype, ptr = [], block = ['1'], arr = [[]]):
    """Constructs a token list for a upc_global<upc_shared_ptr> object declaration."""
    # Clean out original statement:
    del statement[0:len(statement)]
    # Check block and arr for defaults:
    if len(block) < 1: block = ['1']
    if len(arr) < 1: 
    	typename = ['upc_global_shrd_p']
    	doargs = 'false'
    else:
    	typename = ['upc_global_shrd_p_arr']
    	doargs = 'true'
    # Construct the statement:
    # Looks like: upc_global<shared [block] dtype *> name = upc_global<shared [block] dtype *>::construct(arrDim1, arrDim2);
    #obj_decl = ['upc_global_shrd_p', '<', 'shared', '['] + block + [']', dtype] + ptr + ['>', '::', 'type', name]
    obj_decl = typename + ['<'] + [ dtype] + [','] + block + ['>', '::', 'type', name]
    obj_construct = ['=', 'new', 'upc_global', '<', 'shared', '['] + block + [']', dtype] + ptr + ['>', '::', 'type', '(']
    if doargs == 'true':
    	threadtok = ['THREADS']
    	mythreadtok = ['MYTHREAD']
    	obj_args = ['(']
    	for i in xrange(len(arr)):
    		
    		if arr[i] == threadtok:
    			obj_args.append('-2')
    	   	elif arr[i] == mythreadtok:
    	   		obj_args.append('-1')
    	   	else:
        		obj_args.extend(arr[i])
        	obj_args.append(',')
        obj_args[len(obj_args)-1] = ')'
    else: 
		obj_args = []
#    tokens = obj_decl + obj_construct + obj_args
    tokens = obj_decl + obj_args
    tokens.append(';')
#    print tokens
    statement.extend(tokens)

# Construct a list of tokens for a upc_shared_ptr object.
# --------------------------------------------------------------------------- #
def upc_shared_ptr_build(statement, dtype, ptr, block = ['1']):
    """Constructs a token list for a upc_shared_ptr object declaration."""
    # Clean out original statement:
    del statement[0:len(statement)]
    # Check block for defaults:
    if len(block) < 1: block = ['1']
    # Construct tokens list:
    # Looks like: upc_shared_ptr<dtype *, block>
    tokens = ['upc_shared_ptr', '<', dtype] + ptr[1:] + [','] + block + ['>' , '::', 'type']
    statement.extend(tokens)
