# UPC Parser rules

import helper
import keys

# Collect important keywords for use in the parser
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #

# Variable declaration keywords
data_types = keys.cspec_base_types | keys.cspec_more_types
type_extensions = keys.cspec_type_qualifiers | keys.cspec_type_modifiers | keys.cspec_storage_class | keys.upcspec_type_qualifiers

# Statement checking keywords
statements = keys.upcspec_statements
loops      = keys.upcspec_loops

# Identifiers for replacement
identifiers = keys.upcspec_identifiers

# Expressions for replacement
expressions = keys.upcspec_expressions

# Header files that need handling
headers = keys.upclib_headers


# Parsing Rules
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #

# Generic Checks
# =========================================================================== #

# Update line number for each new line.
# --------------------------------------------------------------------------- #
def linenumber(state):
    """Increase line number stored in the parser state by one."""
    state.linenumber_inc()
    state.linenumber_out_inc()
    return state

# Checks for scope changes
# --------------------------------------------------------------------------- #
def scopecheck(state, tokens):
    """Determines if the line results in a change in scope."""
    line_scope = state.scope()
    if '{' in tokens: line_scope = line_scope + 1
    if '}' in tokens: line_scope = line_scope - 1
    state.scope(line_scope)
    return state

# Rules for C constructs
# =========================================================================== #

# Needs rules for: global/static variables

# Check for the main function and replace with 'user_skeleton_main'
# --------------------------------------------------------------------------- #
def maincheck(state, tokens):
    """Replaces the name of the main function for use in SSTmacro."""
    if 'main' in tokens:
        helper.token_prefix(tokens, 'user_skeleton_', 'main')
        state.modified_set()
        state.output_append("replaced 'main' with 'user_skeleton_main' from source line " + str(state.linenumber()) + '.', 2)
    return state

# Check for typedef statements and add new types/aliases to a lists
# --------------------------------------------------------------------------- #
def typedefcheck(state, tokens):
    """Determines if the line contains a typedef statement, and stores any resulting typenames."""
    types = data_types
    new_type = None
    verbose_text = ""
    if 'typedef' in tokens:
        expstart,expend = helper.statement_span(tokens, 'typedef')
        # Now, we have the beginning and ending indices of the statement
        index = expend
        if expend >= len(tokens):
            index = len(tokens)-1
        index = index-1 if tokens[index] in ";{" else index
        # Skip over all array size specifications / struct definitions
        while tokens[index] in "]}":
            tpl = helper.any_separator_span(tokens, index)
            index = tpl[0]-1
        # Now, check new typename is a valid identifier and add to list
        new_type = tokens[index]
        if helper.regex_identifier.match(new_type) != None:
            types.add(new_type)
            state.output_append("new typename '" + new_type + "' detected on source line " + str(state.linenumber()) + '.', 2)
        else:
            state.output_append("WARNING: Failed typedef detection from source line " + str(state.linenumber()) + ".")
    return state

# Check for and handle static scoped variables.
# --------------------------------------------------------------------------- #
def staticvarcheck(pstate, tokens):
    """Checks for and handles non-global static scoped variables."""
    if pstate.scope() != 0 and 'static' in tokens:
        # Because static variables are not currently supported, print a warning.
        pstate.output_append("Warning: unsupported static variable declaration detected on source line " + str(pstate.linenumber()) + '.', 0)
    return pstate

# Check for global variable declarations
# --------------------------------------------------------------------------- #
def globalvarcheck(state, tokens):
    """Checks for global variable declarations and replaces with sstmac global object."""
    types = data_types
    qualifiers = keys.cspec_type_qualifiers | keys.cspec_type_modifiers | keys.cspec_storage_class
    # Check scope first - only continue checks if at global scope.
    if state.scope() == 0 and len(tokens) > 0:
        # Get the start and end indexes of the statement
        expstart, expend = helper.statement_span(tokens, first_token_index=0)
        expend = expend+1 if expend < len(tokens) else expend
        statement = tokens[expstart:expend]
        # Break the statement up on ','
        s_count = 1
        if ',' in statement:
            s_count = s_count + statement.count(',')
        
        # Proceed through statement(s) checking for:
        decl = [dict(quals=[], dtype='', ptr=[], name='', arr=[], exp=[]) for j in xrange(s_count)]
        is_declaration = True
        i = 0
        # Zero or more type qualifiers
        while i < len(statement) and statement[i] in qualifiers:
            for k in xrange(s_count): decl[k]['quals'].append(statement[i])
            i = i+1
        # One data type (may be last qualifier because of 'short' and 'long')
        if i < len(statement) and statement[i] not in types: i = i-1
        if i < len(statement) and statement[i] in types:
            for k in xrange(s_count): decl[k]['dtype'] = statement[i]
            i = i+1
        else:
            is_declaration = False
            i = len(statement)  # skip the remaining checks...
        # Check each of multiple comma-separated declarations
        for k in xrange(s_count):
            # Zero or more pointer indicators
            while i < len(statement) and statement[i] == '*':
                decl[k]['ptr'].append(statement[i])
                i = i+1
            # Exactly one valid name
            if i < len(statement) and helper.regex_identifier.match(statement[i]) != None:
                decl[k]['name'] = statement[i]
                i = i+1
            else:
                is_declaration = False
                i = len(statement)  # skip the remaining checks...
            # Zero or more array size descriptors
            while i < len(statement) and statement[i] == '[':
                tpl = helper.brackets_span(statement, i)
                decl[k]['arr'].append(statement[tpl[0]+1:tpl[1]])
                i = tpl[1]+1
            # Zero or more assignments
            if i < len(statement) and statement[i] == '=':
                tpl = helper.paired_separators_span(statement, ('=', ','), i)
                if tpl[1] >= len(statement): tpl = (tpl[0], tpl[1]-1)
                if tpl[1] < len(statement) and statement[tpl[1]] in ',;': tpl = (tpl[0], tpl[1]-1)
                decl[k]['exp'].extend(statement[tpl[0]:tpl[1]+1])
                i = tpl[1]+1
            # Zero or one commas
            if i < len(statement) and statement[i] == ',':
                i = i+1
            elif k < s_count-1:
                is_declaration = False
                i = len(statement)
        # Ensure that was everything...
        if i < len(statement) and statement[i] != ";":
            is_declaration = False
#        print decl
#        print is_declaration
        if is_declaration:
            new_tokens = []
            for k in xrange(s_count):
                state.output_append("Global variable '" + decl[k]['name'] + "' of type '" + decl[k]['dtype'] + '*'*len(decl[k]['ptr']) + "' detected on source line " + str(state.linenumber()) + '.', 2)
                if len(decl[k]['arr']) > 1:
                    state.output_append("Warning: unsupported multidimensional (>1D) global array '" + decl[k]['name'] + "' on source line " + str(state.linenumber()) + '.', 0)
                if len(decl[k]['exp']) > 0:
                    state.output_append("Warning: unsupported static assignment detected for '" + decl[k]['name'] + "' on source line " + str(state.linenumber()) + '.', 0)
                # Replace statement with a upc_global declaration.
                temp_toks = []
                helper.global_build(temp_toks, decl[k]['name'], decl[k]['dtype'], decl[k]['ptr'], decl[k]['arr'])
                new_tokens.extend(temp_toks + ['\n'])
            new_tokens = tokens[0:expstart] + new_tokens[:len(new_tokens)-1] + tokens[expend:]
            del tokens[0:len(tokens)]
            tokens.extend(new_tokens)
            state.modified_set()
    return state

# Rules for UPC constructs
# =========================================================================== #

# Needs rules for: shared declarations, upc_forall

# Check for static (global) UPC shared variable declarations.
# --------------------------------------------------------------------------- #
def upc_sharedvarcheck(state, tokens):
    """Check for static shared variable declarations and replace with sstmac shared object."""
    types = data_types
    qualifiers = type_extensions
    # Only check at global/file scope, and if the 'shared' keyword is present.
    if state.scope() == 0 and len(tokens) > 0 and 'shared' in tokens:
        # Get the start and end indexes of the statement
        expstart, expend = helper.statement_span(tokens, first_token_index=0)
        expend = expend+1 if expend < len(tokens) else expend
        statement = tokens[expstart:expend]
        # Break the statement up on ','
        s_count = 1
        if ',' in statement:
            s_count = s_count + statement.count(',')
        
        # Proceed through statement(s) checking for:
        decl = [dict(quals=[], ptr_to_shared=False, block=[], dtype='', ptr=[], block2=[], name='', arr=[], exp=[]) for j in xrange(s_count)]
        # need to check if variable is thread-local global ptr to shared or shared pointer
        is_declaration = True
        is_shared = True
        i = 0
        # Zero or more type qualifiers
        while i < len(statement) and statement[i] in qualifiers:
            for j in xrange(s_count): decl[j]['quals'].append(statement[i])
            i = i+1
        # Zero or one blocking factors
        if i < len(statement) and statement[i-1] == 'shared' and statement[i] == '[':
            tpl = helper.brackets_span(statement, i)
            if (tpl[1]-tpl[0]) > 1:
                for j in xrange(s_count): decl[j]['block'].extend(statement[tpl[0]+1:tpl[1]])
            else:
                for j in xrange(s_count): decl[j]['block'].append('0')
            i = tpl[1]+1
        # One data type (may be last qualifier because of 'short' and 'long')
        if i < len(statement) and statement[i] not in types: i = i-1
        if i < len(statement) and statement[i] in types:
            for j in xrange(s_count): decl[j]['dtype'] = statement[i]
            i = i+1
        else:
            is_declaration = False
            i = len(statement)  # skip the remaining checks...
        # Check each of multiple comma-separated declarations
        for k in xrange(s_count):
            # Zero or more pointer indicators
            while i < len(statement) and statement[i] == '*':
                decl[k]['ptr'].append(statement[i])
                i = i+1
            # Zero or one instance of 'shared' after '*'
            if i < len(statement) and len(decl[k]['ptr']) > 0 and statement[i] == 'shared':
                i = i+1
            elif i < len(statement) and len(decl[k]['ptr']) > 0 and statement[i] != 'shared':
                # if this is a pointer declaration (even a pointer to shared) without the 'shared' after the '*', it is not a shared object
                is_shared = False
            if i < len(statement) and statement[i-1] == 'shared' and statement[i] == '[':
                tpl = helper.brackets_span(statement, i)
                if (tpl[1]-tpl[0]) > 1:
                    decl[k]['block2'].extend(statement[tpl[0]+1:tpl[1]])
                else:
                    decl[k]['block2'].append('0')
                i = tpl[1]+1
            # Exactly one valid name
            if i < len(statement) and helper.regex_identifier.match(statement[i]) != None:
                decl[k]['name'] = statement[i]
                i = i+1
            else:
                is_declaration = False
                i = len(statement)  # skip the remaining checks...
            # Zero or more array size descriptors
            while i < len(statement) and statement[i] == '[':
                tpl = helper.brackets_span(statement, i)
                decl[k]['arr'].append(statement[tpl[0]+1:tpl[1]]) # append to a list of token lists...
                i = tpl[1]+1
            # Zero or more assignments
            if i < len(statement) and statement[i] == '=':
                tpl = helper.paired_separators_span(statement, ('=', ','), i)
                if tpl[1] >= len(statement): tpl = (tpl[0], tpl[1]-1)
                if tpl[1] < len(statement) and statement[tpl[1]] in ',;': tpl = (tpl[0], tpl[1]-1)
                decl[k]['exp'].extend(statement[tpl[0]:tpl[1]+1])
                i = tpl[1]+1
            # Zero or one commas
            if i < len(statement) and statement[i] == ',':
                i = i+1
            elif k < s_count-1:
                is_declaration = False
                i = len(statement)
        # Ensure that was everything...
        if i < len(statement) and statement[i] != ";":
            is_declaration = False
#        print decl
#        print is_declaration
        if is_declaration:
            new_tokens = []
            for k in xrange(s_count):
                state.output_append("Shared variable '" + decl[k]['name'] + "' of type '" + decl[k]['dtype'] + '*'*len(decl[k]['ptr']) + "' detected on source line " + str(state.linenumber()) + '.', 2)
                if len(decl[k]['arr']) > 1:
                    state.output_append("Warning: unsupported multidimensional (>1D) shared array '" + decl[k]['name'] + "' on source line " + str(state.linenumber()) + '.', 0)
                if len(decl[k]['exp']) > 0:
                    state.output_append("Warning: unsupported static assignment detected for '" + decl[k]['name'] + "' on source line " + str(state.linenumber()) + '.', 0)
                # Check if this variable is a pointer to shared:
                if 'shared' in decl[k]['quals'] and len(decl[k]['ptr']) > 0:
                    decl[k]['ptr_to_shared'] = True
                # Replace statement with a upc_shared declaration.
                temp_toks = []
                if is_shared:
                    helper.upc_shared_build(temp_toks, decl[k]['name'], decl[k]['dtype'], decl[k]['ptr'], decl[k]['block'], decl[k]['arr'], decl[k]['ptr_to_shared'], decl[k]['block2'])
                elif decl[k]['ptr_to_shared']:
                    helper.upc_global_shared_ptr_build(temp_toks, decl[k]['name'], decl[k]['dtype'], decl[k]['ptr'], decl[k]['block'], decl[k]['arr'])
                new_tokens.extend(temp_toks + ['\n'])
            new_tokens = tokens[0:expstart] + new_tokens[:len(new_tokens)-1] + tokens[expend:]
            del tokens[0:len(tokens)]
            tokens.extend(new_tokens)
            state.modified_set()
    return state

# Check for UPC shared pointers. Must be run after upc_sharedvarcheck().
# --------------------------------------------------------------------------- #
def upc_sharedptrcheck(state, tokens):
    """Check for and handle UPC shared pointers."""
    types = data_types
    qualifiers = type_extensions
    new_change = True
    while new_change:
        new_change = False
        # Check if the 'shared' keyword is present.
        if 'shared' in tokens:
            expstart = tokens.index('shared')
            i = expstart+1
            # build up the expression starting with 'shared':
            decl = dict(block=[], dtype='', ptr=[])
            is_shared_ptr = True
            # Look for blocking indicator:
            if i < len(tokens) and tokens[i] == '[':
                tpl = helper.brackets_span(tokens, i)
                if (tpl[1]-tpl[0]) > 1:
                    decl['block'].extend(tokens[tpl[0]+1:tpl[1]])
                else:
                    decl['block'].append('0')
                i = tpl[1]+1
            # Look for datatype:
            if i < len(tokens) and tokens[i] in types:
                decl['dtype'] = tokens[i]
                i = i+1
            else:
                is_shared_ptr = False
                i = len(tokens)
            # One or more pointer indicators:
            if i < len(tokens) and tokens[i] == '*':
                decl['ptr'].append(tokens[i])
                i = i+1
            else:
                is_shared_ptr = False
                i = len(tokens)
            while i < len(tokens) and tokens[i] == '*':
                decl['ptr'].append(tokens[i])
                i = i+1
            expend = i-1
            # Replace pointer to shared:
            if is_shared_ptr and decl['dtype'] != '' and len(decl['ptr']) > 0:
                state.output_append("Pointer to shared of type 'shared " + decl['dtype'] + '*'*len(decl['ptr']) + "' detected on source line " + str(state.linenumber()) + '.', 2)
                temp_toks = []
                helper.upc_shared_ptr_build(temp_toks, decl['dtype'], decl['ptr'], decl['block'])
                tokens[expstart:expend+1] = temp_toks
                state.modified_set()
                new_change = True
            else:
                state.output_append("Warning: error encountered translating a 'shared' variable on source line " + str(state.linenumber()) + '.', 0)
    return state

# Check for UPC statements and replace with function calls.
# --------------------------------------------------------------------------- #
def upc_statementcheck(state, tokens):
    """Checks for and converts UPC statements into function calls."""
    # First, match a token in the line to a UPC statement
    new_change = True
    while new_change:
        new_change = False
        for s in statements:
            if s in tokens:
                expstart,expend = helper.statement_span(tokens, s)
                # Prepend statement with 'sstmac_' prefix.
                helper.token_prefix(tokens, 'sstmac_', s)
                # Insert ')' before the last token in the statement (the ';')
                tokens.insert(expend, ')')
                # Insert '(' before the token after the first token
                tokens.insert(expstart+1, '(')
                # Update parser state.
                state.modified_set()
                state.output_append("'" + s + "' replaced with 'sstmac_" + s + "' from source line " + str(state.linenumber()) + '.', 2)
                new_change = True
    return state

# Check for and handle upc_forall loops. (Currently unsupported.)
# --------------------------------------------------------------------------- #
def upc_forallcheck(pstate, tokens):
    """Checks for and handles upc_forall loops."""
    if 'upc_forall' in tokens:
        # Because the forall looping construct is not currently supported, simply print a warning.
        pstate.output_append("Warning: unsupported 'upc_forall' looping construct on source line " + str(pstate.linenumber()) + '.', 0)
    return pstate

identifier_functions = {'THREADS':'upc_threads', 'MYTHREAD':'upc_mythread', 'UPC_MAX_BLOCK_SIZE':'upc_maxblocksize'}

# Check for UPC identifiers and replaces with function calls.
# --------------------------------------------------------------------------- #
def upc_identifiercheck(state, tokens):
    """Checks for UPC identifiers and replaces them with function calls."""
    # First, match a token in the line to a UPC identifier
    new_change = True
    while new_change:
        new_change = False
        for i in identifiers:
            if i in tokens:
                index = tokens.index(i)
                # Make changes in the tokens list.
                tokens[index] = identifier_functions[i]
                helper.token_prefix(tokens, 'sstmac_', tokens[index], index)
                tokens.insert(index+1, ')')
                tokens.insert(index+1, '(')
                # Update parser state.
                state.modified_set()
                state.output_append("'" + i + "' replaced with 'sstmac_" + identifier_functions[i] + "' from source line " + str(state.linenumber()) + '.', 2)
                new_change = True
    return state

# Check for UPC sizeof type expressions and replace them with function calls.
# --------------------------------------------------------------------------- #
def upc_expressioncheck(state, tokens):
    """Check for and replace UPC expressions."""
    new_change = True
    while new_change:
        new_change = False
        for e in expressions:
            if e in tokens:
                index = tokens.index(e)
                # Make changes in the tokens list.
                helper.token_prefix(tokens, 'sstmac_', tokens[index], index)
                if tokens[index+1] == '(':  # a typename has been passed - convert to a string
                    helper.token_append(tokens, '_s', tokens[index], index)
                    pstart, pend = helper.paren_span(tokens, index+1)
                    tokens.insert(pend, '"')
                    tokens.insert(pstart+1, '"')
                else:  # an expression - enclose in parentheses
                    expstart, expend = helper.statement_span(tokens, first_token_index=index)
                    tokens.insert(expend, ')')
                    tokens.insert(expstart+1, '(')
                # Update the parser state.
                state.modified_set()
                state.output_append("'" + e + "' replaced with 'sstmac_" + e + "' from source line " + str(state.linenumber()) + '.', 2)
                new_change = True
    return state

# Check for UPC header file includes and replace with sstmac headers.
# --------------------------------------------------------------------------- #
def upc_headerscheck(state, directive_tokens):
    """Find include directives for UPC header files and replace with sstmac versions."""
    for i, tok in enumerate(directive_tokens):
        m = helper.regex_include.match(tok)
        if m != None:
            header = m.group(1)
            span = m.span(1)
            if header in headers:
                new_header = 'sstmac_' + header
                directive_tokens[i] = tok[:span[0]] + new_header + tok[span[1]:]
                state.modified_set()
                state.output_append("'" + header + ".h' replaced with 'sstmac_" + header + ".h' from source line " + str(state.linenumber()) + '.', 2)
    return state
