# UPC tokenizer

import re

# Used to divide a line into tokens.
# --------------------------------------------------------------------------- #
# IMPORTANT: All 2+ character combinations must come before '\W'.
regex_tokenize = re.compile(r"""(
                            ".*?" |              # Divide on string literals
                            ".*?$ |
                            //.*?$ |             # Divide on C++ style comments '//'
                            [#].*?$ |            # Divide on preprocessor directives
                            /\*.*?\*/ |          # Divide on single-line C style comments '/*' and '*/'
                            \d+\.\d+e-?\+?\d+ |  # Divide on floating point numbers
                            \d+e-?\+?\d+
                            \d+\.\d* |       
                            \s+ |                # Divide on sections of extended whitespace
                            /\* | \*/ |          # Divide on C style comments that are not on a single line
                            '..?' |              # Divide on character constants
                            \+= | -= |           # Combination operator/assignment
                            \*= | /= | %= | 
                            <<= | >>= |
                            &= | ^= | \|= |
                            && | \|\| |          # Divide on logical and and or
                            == | != |            # Divide on comparison operators
                            >= | <= |
                            \+\+ | -- |          # Divide on increment/decrement operators
                            << | >> |            # Divide on bitshifting operators
                            -> |                 # Divide on indirect access operator
                            \W                   # Divide on all "special" chars (non-alphanumeric)
                            )""", re.VERBOSE | re.DOTALL)

regex_tokenize2 = re.compile(r"""(
                             \w+ |
                             \s+ 
                             )""", re.VERBOSE)

# Tokenize a line of the UPC source file
# --------------------------------------------------------------------------- #
def tokenize(line):
    """Divides a line of UPC source code into a list of tokens and returns the list."""
    tokens = regex_tokenize.split(line)
    tokens = [tokens[i] for i in range(len(tokens)) if tokens[i] != None]
    tokens = [s.strip() for s in tokens]
    tokens = [tokens[i] for i in range(len(tokens)) if tokens[i] != '']
    return tokens

# Separates the tokens list into a list of code tokens and comment tokens
# --------------------------------------------------------------------------- #
def comments(state, tokens, comment_tokens):
    """Separates comments from code into separate lists."""
    open_comment = state.comment()
    del comment_tokens[0:len(comment_tokens)]
    for token in tokens:
        if token.startswith('/*') and token.endswith('*/'):
            comment_tokens.append(token)
        elif token.startswith('//'):
            comment_tokens.append(token)
        elif token.startswith('/*'):
            comment_tokens.append(token)
            open_comment = True
        elif token.endswith('*/'):
            comment_tokens.append(token)
            open_comment = False
        elif open_comment:
            comment_tokens.append(token)
    for c in comment_tokens:
        tokens.remove(c)
    state.comment(open_comment)
    return state

# Separates preprocessor directives into a separate list from code/comment tokens
# --------------------------------------------------------------------------- #
def directives(state, tokens, directive_tokens):
    """Separates preprocessor directives from code into separate lists."""
    ml_directive = state.directive()
    del directive_tokens[0:len(directive_tokens)]
    for token in tokens:
        if token.startswith('#') and token.endswith('\\'):
            directive_tokens.append(token)
            ml_directive = True
        elif token.startswith('#'):
            directive_tokens.append(token)
        elif ml_directive == True and not token.endswith('\\') and token == tokens[len(tokens)-1]:
            directive_tokens.append(token)
            ml_directive = False
        elif ml_directive == True:
            directive_tokens.append(token)
    for d in directive_tokens:
        tokens.remove(d)
    state.directive(ml_directive)
    return state

