/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_SKELETONS_MINI_ACTORMATMUL_ACTORMATMUL_COMPUTOR_H_INCLUDED
#define SSTMAC_SOFTWARE_SKELETONS_MINI_ACTORMATMUL_ACTORMATMUL_COMPUTOR_H_INCLUDED

#include <actormatmul.h>
#include <actormatmul-messages.h>
#include <sstmac/software/libraries/compute/lib_compute_matrix.h>
#include <sstmac/software/libraries/util/lib_display.h>
#include <sstmac/common/stats/stat_time_window.h>

/** Bring sstmac stuff into namespace */
using sstmac::sw::lib_compute_matrix;
using sstmac::sw::GREEN;
using sstmac::sw::RED;
using sstmac::sw::PURPLE;
using sstmac::sw::BLACK;

namespace mini
{

  /**
   * An actor that performs computations (even actors have to make a living).
   */
  class actormatmul::computor : public simplemboxapp
  {
    typedef actormatmul::localstore localstore;
    typedef actormatmul::remotestore remotestore;
    typedef actormatmul::work work;
    typedef actormatmul::putdata putdata;

    /// The parent id.
    actorid parent_;
    /// The number of threads managed by this computor.
    int threads_;
    /// The performance.
    timestamp time_per_fma_;

    lib_display::ptr disp_lib_;

    lib_compute_matrix::ptr compute_lib_;

    static int active_workers_;
    static stat_time_window_time_avg::ptr stat_;

    static int active_workers_donated_;
    static stat_time_window_time_avg::ptr stat_donated_;

    /// Private callback to pretend to do computation.
    void
    compute(const work::ptr& msg)
    {
      int64_t md = msg->matdim();

      if (msg->store_to() == parent_)
        {
          disp_lib_->change_fill(GREEN);

          active_workers_++;
        }
      else
        {
          disp_lib_->change_fill(PURPLE);

          active_workers_donated_++;
        }
      stat_->record(now(), active_workers_);
      stat_donated_->record(now(), active_workers_donated_);
      //std::cout << "t=" << now() << " -- active_workers: " << active_workers_ << std::endl;
      // std::cout << "t=" << now() << " -- active_donated: " << active_workers_donated_ << std::endl;
      compute_lib_->double_mxm(md, md, md);

      if (msg->store_to() == parent_)
        {
          disp_lib_->change_fill(BLACK);

          active_workers_--;
        }
      else
        {
          disp_lib_->change_fill(RED);

          active_workers_donated_--;
        }

      stat_donated_->record(now(), active_workers_donated_);
      stat_->record(now(), active_workers_);

      // std::cout << "t=" << now() << " -- done - active_workers: " << active_workers_ << std::endl;
      // std::cout << "t=" << now() << " -- done - active_donated: " << active_workers_donated_
      //     << std::endl;

      //int64_t multiplies = md * md * md;
      // timestamp t((time_per_fma_ * multiplies) / threads_);
      // simplemboxapp::compute(t, threads_);
    }

    /// Hello.
    computor(softwareid id, actorid parent, int threads,
        const timestamp &time_per_FMA) :
        simplemboxapp(id), parent_(parent), threads_(threads), time_per_fma_(
            time_per_FMA)
    {
      compute_lib_ = lib_compute_matrix::construct(id_);
      register_lib(compute_lib_);

      disp_lib_ = lib_display::construct(id_);
      register_lib(disp_lib_);

      if (!stat_)
        stat_ = stat_time_window_time_avg::construct("<stats> app | actormatmul",
            "stats_actormatmul_work.csv");

      if (!stat_donated_)
        stat_donated_ = stat_time_window_time_avg::construct("<stats> app | actormatmul",
            "stats_actormatmul_donate.csv");

    }

  public:
    typedef boost::intrusive_ptr<computor> ptr;

    /// Create a new computor.
    static computor::ptr
    construct(softwareid id, actorid parent, int threads,
        const timestamp &per_FMA)
    {
      computor::ptr retval(
          new computor(id, parent, threads, per_FMA));
      //  retval->set_self(retval);
      return retval;
    }

    /// Clone this object.
    app::ptr
    clone(softwareid id)
    {
      return construct(id, parent_, threads_, time_per_fma_);
    }

    void
    consume_params(const sim_parameters::ptr &params)
    {

    }

    /// Go.
    void
    skeleton_main()
    {
      mbox()->init();
      do
        {
          simplembox::recvresult_t res = mbox()->recv(actorid::any(),
              actorpattern::any());
          work::ptr msg = boost::dynamic_pointer_cast<work>(
              res.second);
          if (msg->matdim() > 0)
            {
              this->compute(msg);
              if (msg->store_to() == parent_)
                {
                  // We send back 8 bytes -- we are assuming we can share pointers
                  actormsg::ptr amsg = localstore::construct(
                      msg->iteration());
                  mbox()->send(parent_, amsg, 8);
                }
              else
                {
                  // remote store.  We have to send back a copy of the actual matrix.
                  actormsg::ptr amsg = remotestore::construct(
                      msg->iteration());
                  mbox()->send(msg->store_to(), amsg,
                      msg->matdim() * msg->matdim());
                }
            }
          else
            break;
        }
      while (true);
    }
  };

  int actormatmul::computor::active_workers_ = 0;
  stat_time_window_time_avg::ptr actormatmul::computor::stat_;

  int actormatmul::computor::active_workers_donated_ = 0;
  stat_time_window_time_avg::ptr actormatmul::computor::stat_donated_;

} //end of namespace mini

#endif
