/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2011 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include "lusweep.h"
#include <stdio.h>
#include <math.h>

using namespace sstmac;
using namespace sstmac::sw;

namespace lusweep {

    SSTRegisterApp("lusweep", lumain);

 void
lumain::consume_params(const sim_parameters::ptr &params)
{
  s_nx = params->get_optional_int_param("lusweep_nx", 64);
  s_ny = params->get_optional_int_param("lusweep_ny", 64);
  s_nz = params->get_optional_int_param("lusweep_nz", 64);
  s_itr = params->get_optional_int_param("lusweep_itr", 250);
  s_inorm = params->get_optional_int_param("lusweep_inorm", 250);

  s_wg_rhs1 = params->get_optional_double_param("lusweep_rhs1", 1e-6);
  s_wg_rhs2 = params->get_optional_double_param("lusweep_rhs2", 1e-6);
  s_wg_rhs3 = params->get_optional_double_param("lusweep_rhs3", 1e-6);
  s_wg_rhs4 = params->get_optional_double_param("lusweep_rhs4", 1e-6);

  s_wg_jacld = params->get_optional_double_param("lusweep_jacld", 1e-6);
  s_wg_jacu = params->get_optional_double_param("lusweep_jacu", 1e-6);
  s_wg_blts = params->get_optional_double_param("lusweep_blts", 1e-6);
  s_wg_buts = params->get_optional_double_param("lusweep_buts", 1e-6);
}


// Overrides thread::run virtual function
// The task thread context is started by a call into this function
void
lumain::skeleton_main()
{

  // initialize mpi
  if (mpi() == 0)
    throw sstmac::nullerror("lumain:  mpiapi pointer is null.");

  timestamp start = mpi()->init();

  mpicomm::ptr world = mpi()->comm_world();

  const mpiid size = world->size();


  int pex = (int) sqrt(size.id);
  int pey;

  for (; pex * ((int) (size.id / pex)) != size.id; pex--)
    ;
  pey = size.id / pex;

  mpiid rank(world->rank());

  //SSTMAC_DEBUG << "multmpi[" << rank << "]::run(): mpi()->init() at "
  //          << start << "\n";

  if (rank.id == 0)
    {
      SSTMAC_DEBUG
          << "------------------------------------------------------------\n";
      SSTMAC_DEBUG << "Performing 2D decomposition:\n";
      SSTMAC_DEBUG << "     PE-X:         " << pex << "\n";
      SSTMAC_DEBUG << "     PE-Y:         " << pey << "\n";
      SSTMAC_DEBUG << "     Nx:           " << s_nx << "\n";
      SSTMAC_DEBUG << "     Ny:           " << s_ny << "\n";
      SSTMAC_DEBUG << "     Nz:           " << s_nz << "\n";
      //SSTMAC_DEBUG << "     Wg:           " << s_wg << "\n";
      SSTMAC_DEBUG << "\n";
    }

  //place holder routine mimicing communication
  bcast_inputs();
  ssor(0);
  ssor(s_itr - 1);
  error();
  pintgr();

  // simulate broadcasting 1 double from rank 0 with no real payload
  //const int count=1024;
  //const mpiid root(0);
  //mpi()->bcast(count, mpitype::mpi_double, root, world);

  //SSTMAC_DEBUG << "launching sweep...\n";
  timestamp start_sweep = mpi()->wtime();
  sweep(pex, pey, size.id, rank.id, s_nx, s_ny, s_nz);
  timestamp end_sweep = mpi()->wtime();
  //SSTMAC_DEBUG << "sweep done.\n";

  //mpi()->barrier(mpi()->comm_world());
  // finalize
  timestamp done = mpi()->finalize();

  if (rank.id == 0)
    {
      SSTMAC_DEBUG
          << "------------------------------------------------------------\n";
      SSTMAC_DEBUG << "Simulation Finalized\n";
      SSTMAC_DEBUG << "\n";
      SSTMAC_DEBUG << "End Time:          " << done.sec() << " / "
          << (done.sec() * s_itr) << "\n";
      SSTMAC_DEBUG << "Sweep Time:        "
          << (end_sweep.sec() - start_sweep.sec()) << " / ";
      SSTMAC_DEBUG << (end_sweep.sec() - start_sweep.sec()) * s_itr << "\n";
      SSTMAC_DEBUG
          << "------------------------------------------------------------\n";
    }
}


void
lumain::sweep(int pex, int pey, int size, int rank, int nx, int ny, int nz)
{
  int tilex = ceil(nx / pex);
  int tiley = ceil(ny / pey);

  int north = get_north(rank, pex, pey);
  int south = get_south(rank, pex, pey);
  int east = get_east(rank, pex, pey);
  int west = get_west(rank, pex, pey);

  //SSTMAC_DEBUG << "rank " << rank << " north: " << north << ", south: " << south
  //        << ", east: " << east << ", west: " << west << "\n";

  if (rank == 0)
    {
      SSTMAC_DEBUG << "  -> Nx/PE-X:      " << tilex << "\n";
      SSTMAC_DEBUG << "  -> Ny/PE-Y:      " << tiley << "\n";
      //SSTMAC_DEBUG << "  -> W:            " << ((tilex * tiley) * s_wg) << "\n";
      SSTMAC_DEBUG << "\n";
    }

  for (int itr_count = 0; itr_count < s_itr; itr_count++)
    {
      if (rank == 0)
        {
          SSTMAC_DEBUG << "Iteration " << itr_count << "...\n";
        }

      lower(north, south, east, west, tilex, tiley, s_nz);
      upper(north, south, east, west, tilex, tiley, s_nz);

      //SSTMAC_DEBUG << "rank " << rank << " calling rhs(" << north << "," << south
      //        << "," << east << "," << west << "...)\n";
      rhs(north, south, east, west, tilex, tiley, s_nz);
    }
}

void
lumain::lower(int north, int south, int east, int west, int tilex,
    int tiley, int nz)
{
  mpicomm::ptr commworld(mpi()->comm_world());
  mpitag tag(0);
  const_mpistatus_t status = mpistatus::construct();
  const mpiid northrank(north);
  const mpiid southrank(south);
  const mpiid eastrank(east);
  const mpiid westrank(west);

  timestamp w_jacld(s_wg_jacld * (tilex * tiley));
  timestamp w_blts(s_wg_blts * (tilex * tiley));

  for (int zcounter = 0; zcounter < nz; zcounter++)
    {
      this->compute(w_jacld);

      if (north > -1)
        {
          mpi()->recv(tiley * 5, mpitype::mpi_double, northrank, tag,
              commworld, status);
        }

      if (east > -1)
        {
          mpi()->recv(tilex * 5, mpitype::mpi_double, eastrank, tag,
              commworld, status);
        }

      this->compute(w_blts);

      if (south > -1)
        {
          mpi()->send(tiley, mpitype::mpi_double, southrank, tag,
              commworld);
        }

      if (west > -1)
        {
          mpi()->send(tilex, mpitype::mpi_double, westrank, tag, commworld);
        }
    }
}

void
lumain::upper(int north, int south, int east, int west, int tilex,
    int tiley, int nz)
{
  mpicomm::ptr commworld(mpi()->comm_world());
  mpitag tag(0);
  const_mpistatus_t status = mpistatus::construct();

  const mpiid northrank(north);
  const mpiid southrank(south);
  const mpiid eastrank(east);
  const mpiid westrank(west);

  const timestamp w_jacu(s_wg_jacu * (tilex * tiley));
  const timestamp w_buts(s_wg_buts * (tilex * tiley));

  for (int zcounter = 0; zcounter < nz; zcounter++)
    {
      this->compute(w_jacu);

      if (north > -1)
        {
          mpi()->recv(tiley * 5, mpitype::mpi_double, northrank, tag,
              commworld, status);
        }

      if (east > -1)
        {
          mpi()->recv(tilex * 5, mpitype::mpi_double, eastrank, tag,
              commworld, status);
        }

      this->compute(w_buts);

      if (south > -1)
        {
          mpi()->send(tiley, mpitype::mpi_double, southrank, tag,
              commworld);
        }

      if (west > -1)
        {
          mpi()->send(tilex, mpitype::mpi_double, westrank, tag, commworld);
        }
    }
}

void
lumain::rhs(int north, int south, int east, int west, int tilex, int tiley,
    int nz)
{
  const timestamp w_rhs1(s_wg_rhs1 * (tilex * tiley * nz));
  const timestamp w_rhs2(s_wg_rhs2 * (tilex * tiley * nz));
  const timestamp w_rhs3(s_wg_rhs3 * (tilex * tiley * nz));
  const timestamp w_rhs4(s_wg_rhs4 * (tilex * tiley * nz));

  this->compute(w_rhs1);

  //SSTMAC_DEBUG << "rank: " << mpi()->comm_rank(mpi()->comm_world()) <<
  //       " calling exchange(" << north << "," << south << "...)\n";
  exchange(north, south, 10 * tilex * nz, 10 * tilex * nz);
  exchange(south, north, 10 * tilex * nz, 10 * tilex * nz);

  this->compute(w_rhs2);

  exchange(east, west, 10 * tiley * nz, 10 * tiley * nz);
  exchange(west, east, 10 * tiley * nz, 10 * tiley * nz);

  this->compute(w_rhs3);
  this->compute(w_rhs4);
}

void
lumain::exchange(int recvfrom, int sendto, int recvlen, int sendlen)
{
  mpicomm::ptr commworld(mpi()->comm_world());
  mpitag tag(99);
  const_mpistatus_t status = mpistatus::construct();

  mpirequest_t req;

  const mpiid recvfromrank(recvfrom);
  const mpiid sendtorank(sendto);

  int rank = mpi()->comm_rank(commworld).id;
  //SSTMAC_DEBUG << "rank " << rank << " recv: " << recvfrom << " send: " << sendto << "\n";

  if (recvfrom > -1)
    {
      mpi()->irecv(recvlen, mpitype::mpi_double, recvfromrank, tag, commworld, req);
      mpi()->wait(req, status);
      //mpi()->recv(recvlen, mpitype::mpi_double, recvfromrank, tag, commworld, status);
    }

  if (sendto > -1)
    {
      mpi()->send(sendlen, mpitype::mpi_double, sendtorank, tag, commworld);
    }
}

int
lumain::get_north(int rank, int pex, int pey)
{
  if (rank < pex)
    {
      return -1;
    }
  else
    {
      return rank - pex;
    }
}

int
lumain::get_south(int rank, int pex, int pey)
{
  if (rank >= (pex * (pey - 1)))
    {
      return -1;
    }
  else
    {
      return rank + pex;
    }
}

int
lumain::get_east(int rank, int pex, int pey)
{
  if (((rank + 1) % pex) == 0)
    {
      return -1;
    }
  else
    {
      return rank + 1;
    }
}

int
lumain::get_west(int rank, int pex, int pey)
{
  if ((rank % pex) == 0)
    {
      return -1;
    }
  else
    {
      return rank - 1;
    }
}

void
lumain::bcast_inputs()
{
  //simulating bcast input params
  const mpiid root(0);
  mpicomm::ptr world = mpi()->comm_world();

  mpi()->bcast(1, mpitype::mpi_int, root, world);
  mpi()->bcast(1, mpitype::mpi_int, root, world);
  mpi()->bcast(1, mpitype::mpi_int, root, world);
  mpi()->bcast(1, mpitype::mpi_double, root, world);
  mpi()->bcast(1, mpitype::mpi_double, root, world);
  mpi()->bcast(1, mpitype::mpi_double, root, world);
  mpi()->bcast(1, mpitype::mpi_int, root, world);
  mpi()->bcast(1, mpitype::mpi_int, root, world);
  mpi()->bcast(1, mpitype::mpi_int, root, world);
}

void 
lumain::error()
{
    mpi()->allreduce(5, mpitype::mpi_double, mpiop::sum, mpi()->comm_world());
}

void
lumain::pintgr()
{
    mpi()->allreduce(1, mpitype::mpi_double, mpiop::sum, mpi()->comm_world());
    mpi()->allreduce(1, mpitype::mpi_double, mpiop::sum, mpi()->comm_world());
    mpi()->allreduce(1, mpitype::mpi_double, mpiop::sum, mpi()->comm_world());
}

void
lumain::ssor(int iter)
{
    //compute the L2 norms of newton iteration residuals
    l2norm();
    mpi()->barrier(mpi()->comm_world());

    //compute the max-norms of newton iteration corrections
    if((iter + 1) % s_inorm == 0)
    {
        l2norm();
    }

    //compute the max-norms of newton iteration resituals
    if((iter + 1) % s_inorm == 0 || 
        iter == s_itr)
    {
        l2norm();
    }

    mpi()->allreduce(1, mpitype::mpi_double, mpiop::max, mpi()->comm_world());
}

void
lumain::l2norm()
{
    mpi()->allreduce(5, mpitype::mpi_double, mpiop::sum, mpi()->comm_world());
}

}
