/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include "lib_compute_minimd.h"
#include <sstmac/sstmacro.h>

#include "minimd-atom.h"
#include "minimd-force.h"
#include "minimd-integrate.h"
#include "minimd-neighbor.h"
#include "minimd-in.h"
#include "minimd-output.h"

#include "perf.h"
/* 
WKMODEL: 0: naive/buggy
WKMODEL: 1: based on lib_compute_loops with some care
WKMODEL: 2: eiger model
WKMODEL: 3: something based on snl nearest neighbor interpolation?
*/
#define WKMODEL 1
// for eiger with wkmodel 2
#define MODELDIR "x5550/"
using namespace sstmac;
using namespace std;

namespace mini
{

  lib_compute_minimd::lib_compute_minimd(softwareid id) :
    lib_compute_time(id)

  {
    id_ = id;
    using_loops_ = sstmac::sstmac_env::params->get_bool_param(
        "lib_compute_loops_enable");
    if (!using_loops_)
      {
std::cout << "################ not using_loops_ ###################" <<std::endl;

        funcs_["create_atoms"] = timestamp(1E-7);
        funcs_["create_velocity"] = timestamp(1E-7);
        funcs_["output"] = timestamp(1E-7);
        funcs_["create_atoms"] = timestamp(1E-7);
        funcs_["create_atoms"] = timestamp(1E-7);

        funcs_["Neighbor::build"] = timestamp(1E-7);
        funcs_["Thermo::temperature"] = timestamp(1E-7);
        funcs_["Thermo::energy"] = timestamp(1E-7);
        funcs_["Force::compute"] = timestamp(1E-7);
        funcs_["Thermo::pressure"] = timestamp(1E-7);
        funcs_["Thermo::compute"] = timestamp(1E-7);
        funcs_["Integrate::run"] = timestamp(1E-7);

      }
std::cout << "################ using_loops_ ###################" <<std::endl;
    libname_ = "minimdcomputelib" + id.to_string();

  }

  lib_compute_minimd::~lib_compute_minimd()
  {

  }

  //
  // Allocate an object of this type.
  //
  lib_compute_minimd::ptr
  lib_compute_minimd::construct(softwareid it)
  {
    lib_compute_minimd::ptr retval(new lib_compute_minimd(it));

    return retval;
  }

  void
  lib_compute_minimd::compute(std::string func,
      const boost::intrusive_ptr<ptr_type> &a1,
      const boost::intrusive_ptr<ptr_type> &a2, int index)
  {
    if (using_loops_)
      {
        if (!libloop_)
          {
            libloop_ = sstmac::sw::lib_compute_loops::construct(id_);
            register_lib(libloop_);
          }

	// std::cerr << "lib_compute_minimd::compute " << func <<std::endl;
        if (func == "create_atoms")
          {
            minimd::atom::ptr at =
                boost::dynamic_pointer_cast<minimd::atom>(a1);
            if (!at)
              {
                sst_throw(ssterror, "lib_compute_minimd::compute: create_atoms problem casting atom");
              }

            minimd::in::ptr in = boost::dynamic_pointer_cast<minimd::in>(a2);
            if (!in)
              {
                sst_throw(ssterror, "lib_compute_minimd::compute: create_atoms problem casting 'in'");
              }

            double alat = pow((4.0 / in->rho), (1.0 / 3.0));
            int ilo = static_cast<int> (at->box.xlo / (0.5 * alat) - 1);
            int ihi = static_cast<int> (at->box.xhi / (0.5 * alat) + 1);
            int jlo = static_cast<int> (at->box.ylo / (0.5 * alat) - 1);
            int jhi = static_cast<int> (at->box.yhi / (0.5 * alat) + 1);
            int klo = static_cast<int> (at->box.zlo / (0.5 * alat) - 1);
            int khi = static_cast<int> (at->box.zhi / (0.5 * alat) + 1);

            ilo = std::max(ilo, 0);
            ihi = std::min(ihi, 2 * in->nx - 1);
            jlo = std::max(jlo, 0);
            jhi = std::min(jhi, 2 * in->ny - 1);
            klo = std::max(klo, 0);
            khi = std::min(khi, 2 * in->nz - 1);

            libloop_->compute_loop3(klo, khi, jlo, jhi, ilo, ihi, 12);
          }
        else if (func == "create_velocity")
          {
            minimd::atom::ptr at =
                boost::dynamic_pointer_cast<minimd::atom>(a1);
            if (!at)
              {
                sst_throw(ssterror, "lib_compute_minimd::compute: create_velocity problem casting atom");
              }

            if (index == 0)
              {
		std::cerr << "lib_compute_minimd::compute [create_velocity] compute_loop(0," << at->nlocal<<",3)"<<std::endl;
                libloop_->compute_loop(0, at->nlocal, 3);
              }
            else
              {
		std::cerr << "lib_compute_minimd::compute [create_velocity] compute_loop(0," << at->nlocal<<",12)"<<std::endl;
                libloop_->compute_loop(0, at->nlocal, 12);
              }
          }
        else if (func == "Force::compute")
          {
            minimd::atom::ptr at =
                boost::dynamic_pointer_cast<minimd::atom>(a1);
            if (!at)
              {
                sst_throw(ssterror, "lib_compute_minimd::Force::compute: problem casting atom");
              }

            minimd::neighbor::ptr ne = boost::dynamic_pointer_cast<
                minimd::neighbor>(a2);
            if (!ne)
              {
                sst_throw(ssterror, "lib_compute_minimd::Force::compute: problem casting neighbor");
              }
#if (WKMODEL == 0)
            libloop_->compute_loop(0, at->nlocal + at->nghost, 3);
            libloop_->compute_loop(0, at->nlocal, 5);
            libloop_->compute_loop2(0, at->nlocal, 0, 10, 8);
#elif (WKMODEL == 1)
            PERFLOG(Fclear,  IP(nall));
            libloop_->compute_loop(0, at->nlocal + at->nghost, at->Fclear_nl);
            PERFSTOP(Fclear,  IP(at->nlocal + at->nghost));

            PERFLOG(Fscatter,  IP(nlocal), IP(nscat), IP(nneigh));

            libloop_->compute_loop(0, at->nlocal, 3);
// 5 and 9 (now 2 and 6) could come from parameter file
            const int avgneigh = at->ratio_neighbors; 
	    const double frac_scat = at->frac_scat; // avg scats per neighbor
            double nlocal_lines = at->Fscatter_offset + at->Fscatter_coef*frac_scat;
            libloop_->compute_loop2(0, at->nlocal, 0, avgneigh, nlocal_lines);

            PERFSTOP(Fscatter,  IP(at->nlocal), IP(int(floor(frac_scat*avgneigh*at->nlocal))), IP(avgneigh*at->nlocal));
#elif (WKMODEL == 2)
            const int avgneigh = at->ratio_neighbors; 
	    const double frac_scat = at->frac_scat; // avg scats per neighbor
	    eiger_metrics m("Fclear");
	    eiger_metrics m2("Fscatter");
	    m.add("nall", at->nlocal + at->nghost);
	    m2.add("nlocal", at->nlocal);
	    m2.add("nneigh", avgneigh*at->nlocal);
	    m2.add("nscat", int(floor(frac_scat*avgneigh*at->nlocal)));

            PERFLOG(Fclear,  IP(nall));
	    SSTMAC_compute_eiger(m.m, MODELDIR "Fclear.model");
            PERFSTOP(Fclear,  IP(at->nlocal + at->nghost));

            PERFLOG(Fscatter,  IP(nlocal), IP(nscat), IP(nneigh));
	    SSTMAC_compute_eiger(m2.m, MODELDIR "Fscatter.model");
            PERFSTOP(Fscatter,  IP(at->nlocal), IP(int(floor(frac_scat*avgneigh*at->nlocal))), IP(avgneigh*at->nlocal));

#else
#error "lib_compute_minimd::Force::compute: unsupported WKMODEL == " ## #WKMODEL
#endif
          }
        else if (func == "Neighbor::build")
          {
            minimd::atom::ptr at =
                boost::dynamic_pointer_cast<minimd::atom>(a1);
            if (!at)
              {
                sst_throw(ssterror, "lib_compute_minimd::Neighbor::build: problem casting atom");
              }
            minimd::neighbor::ptr ne = boost::dynamic_pointer_cast<
                minimd::neighbor>(a2);
            if (!ne)
              {
                sst_throw(ssterror, "lib_compute_minimd::Neighbor::build problem casting neighbor");
              }

#if (WKMODEL == 0)
            libloop_->compute_loop(0, at->nlocal, 25);
//binatoms nall*8; nall =~ 2.2*nlocal so say 16 for binatoms
// neighboor search: need better instrumentation to get average loop trip counts. eyeball number of at least 40 read equivalents in main loop.
	    // const int nlines_local = 16 + 40;
            // libloop_->compute_loop(0, at->nlocal, nlines_local);
#elif (WKMODEL == 1)
// this is a mess. mbins must be input since neighbor skips all the math in the skeleton.
            int nall = at->nlocal+at->nghost;
            int nlocal = at->nlocal;
            int mbins = ne->mbins;
            PERFLOG(Nbuild, IP(nlocal), IP(nall), IP(mbins),IP(nstencil), IP(avgtripj1), IP(avgtripj2), IP(avgtripj3), IP(avgallocs) );
            libloop_->compute_loop(0, nlocal, 12); // approximate for i loop, assuming no realloc hits


            PERFLOG(Nbinatoms, IP(nlocal), IP(nall), IP(mbins));
            libloop_->compute_loop(0, ne->mbins, 1); // binhead init to -1
            libloop_->compute_loop(0, nall, at->Nbinatoms_nl); // update bins loops including coord2bin
            PERFSTOP(Nbinatoms, IP(nlocal), IP(nall), IP(mbins));

	    // hundred is the assumed trips per bin through entire while j loop
            libloop_->compute_loop2(0, nlocal, 0, ne->avgtripsj1, at->Nbuild_nl1);
            libloop_->compute_loop2(0, nlocal, 0, ne->avgtripsj2, at->Nbuild_nl2);
            libloop_->compute_loop3(0, nlocal, 0, ne->nstencil, 0, ne->avgtripsj3, at->Nbuild_nl3);

	    PERFSTOP(Nbuild, IP(nlocal), IP(nall), IP(mbins), IP(ne->nstencil), IP(ne->avgtripsj1), IP(ne->avgtripsj2), IP(ne->avgtripsj3), IP(ne->avgallocs));

#elif (WKMODEL == 2)

            int nall = at->nlocal+at->nghost;
            int nlocal = at->nlocal;
            int mbins = ne->mbins;
	    eiger_metrics m("Nbuild");
	    m.add("nlocal", at->nlocal);
	    m.add("nall", nall);
	    m.add("mbins", mbins);
	    m.add("nstencil", ne->nstencil);
	    m.add("avgtripj1", ne->avgtripsj1);
	    m.add("avgtripj2", ne->avgtripsj2);
	    m.add("avgtripj3", ne->avgtripsj3);
	    m.add("avgallocs", ne->avgallocs);

            PERFLOG(Nbuild, IP(nlocal), IP(nall), IP(mbins),IP(nstencil), IP(avgtripj1), IP(avgtripj2), IP(avgtripj3), IP(avgallocs) );

            PERFLOG(Nbinatoms, IP(nlocal), IP(nall), IP(mbins));
#if 0
 // cannot nest models in eiger like in real codes or in decomposable libloops blocks, so collect 0-delta-time placeholder data to keep postprocessors happy.
            libloop_->compute_loop(0, ne->mbins, 1); // binhead init to -1
            libloop_->compute_loop(0, nall, 4); // update bins loops including coord2bin
#endif
            PERFSTOP(Nbinatoms, IP(nlocal), IP(nall), IP(mbins));

	    SSTMAC_compute_eiger(m.m, MODELDIR "Nbuild.model");

	    PERFSTOP(Nbuild, IP(nlocal), IP(nall), IP(mbins), IP(ne->nstencil), IP(ne->avgtripsj1), IP(ne->avgtripsj2), IP(ne->avgtripsj3), IP(ne->avgallocs));
#else
#error "lib_compute_minimd::Neighbor::build: unsupported WKMODEL == " ## #WKMODEL
#endif
          }
        else if (func == "Neighbor::setup")
          {
            libloop_->compute_loop(0, 1, 54);
          }
        else if (func == "Thermo::temperature")
          {
            minimd::atom::ptr at =
                boost::dynamic_pointer_cast<minimd::atom>(a1);
            if (!at)
              {
                sst_throw(ssterror, "lib_compute_minimd::Thermo::temperature: problem casting atom");
              }

            PERFLOG(Ttemperature,  IP(nlocal));
#if (WKMODEL == 0)
            libloop_->compute_loop(0, at->nlocal, 4);
#elif (WKMODEL == 1)
            libloop_->compute_loop(0, at->nlocal, at->Ttemperature_nl);
#elif (WKMODEL == 2)
	    eiger_metrics m("Ttemperature");
	    m.add("nlocal", at->nlocal);
	    SSTMAC_compute_eiger(m.m, MODELDIR "Ttemperature.model");
#else
#error "lib_compute_minimd::Thermo::temperature: unsupported WKMODEL == " ## #WKMODEL
#endif
            PERFSTOP(Ttemperature,  IP(at->nlocal));
          }
        else if (func == "Thermo::energy")
          {
            minimd::atom::ptr at =
                boost::dynamic_pointer_cast<minimd::atom>(a1);
            if (!at)
              {
                sst_throw(ssterror, "lib_compute_minimd::Thermo::energy: problem casting atom");
              }
            const int avgneighbors = at->ratio_neighbors; 
            PERFLOG(Tenergy, IP(nlocal),IP(nneigh));
#if (WKMODEL == 0)
            libloop_->compute_loop(0, at->nlocal, 16);
#elif (WKMODEL == 1)
            double nlines_local = at->Tenergy_offset + avgneighbors* at->Tenergy_coef;
            libloop_->compute_loop(0, at->nlocal, nlines_local);
#elif (WKMODEL == 2)
	    eiger_metrics m("Tenergy");
	    m.add("nlocal", at->nlocal);
	    m.add("nneigh", at->nlocal*at->ratio_neighbors);
	    SSTMAC_compute_eiger(m.m, MODELDIR "Tenergy.model");
#else
#error "lib_compute_minimd::Thermo::energy unsupported WKMODEL == " ## #WKMODEL
#endif
            PERFSTOP(Tenergy, IP(at->nlocal),IP(avgneighbors*at->nlocal));
          }
        else if (func == "Thermo::pressure")
          {
            minimd::atom::ptr at =
                boost::dynamic_pointer_cast<minimd::atom>(a1);
            if (!at)
              {
                sst_throw(ssterror, "lib_compute_minimd::Thermo::pressure: problem casting atom");
              }
            PERFLOG(Tpressure, IP(nall));
#if (WKMODEL == 0)
            libloop_->compute_loop(0, at->nlocal, 2); // buggy loop counter model
#elif (WKMODEL == 1)
            libloop_->compute_loop(0, at->nlocal+at->nghost, at->Tpressure_nl);
#elif (WKMODEL == 2)
	    eiger_metrics m("Tpressure");
	    m.add("nall", at->nlocal+at->nghost);
	    SSTMAC_compute_eiger(m.m, MODELDIR "Tpressure.model");
#else
#error "lib_compute_minimd::Thermo::pressure unsupported WKMODEL == " ## #WKMODEL
#endif
            PERFSTOP(Tpressure, IP(at->nlocal+ at->nghost));
          }
        else if (func == "Thermo::compute")
          {

            libloop_->compute_loop(0, 1, 14); // trivial amount of work 
          }
        else if (func == "Integrate::run")
          {
            minimd::atom::ptr at =
                boost::dynamic_pointer_cast<minimd::atom>(a1);
            if (!at)
              {
                sst_throw(ssterror, "lib_compute_minimd::Force::compute: problem casting atom");
              }

            if (index == 1)
              {
                PERFLOG(Imove, IP(nlocal));
#if (WKMODEL == 0)
                libloop_->compute_loop(0, at->nlocal, 3);
#elif (WKMODEL == 1)
		libloop_->compute_loop(0, at->nlocal, at->Imove_nl);
#elif (WKMODEL == 2)
		eiger_metrics m("Imove");
		m.add("nlocal", at->nlocal);
		SSTMAC_compute_eiger(m.m, MODELDIR "Imove.model");
#else
#error "lib_compute_minimd::Integrate::run unsupported WKMODEL == " ## #WKMODEL
#endif
                PERFSTOP(Imove, IP(at->nlocal));
              }
            else
              {
                PERFLOG(Iapply, IP(nlocal));

#if (WKMODEL == 0)
                libloop_->compute_loop(0, at->nlocal, 6);
#elif (WKMODEL == 1)
                libloop_->compute_loop(0, at->nlocal, at->Iapply_nl);
#elif (WKMODEL == 2)
		eiger_metrics m("Iapply");
		m.add("nlocal", at->nlocal);
		SSTMAC_compute_eiger(m.m, MODELDIR "Iapply.model");
#else
#error "lib_compute_minimd::Integrate::run unsupported WKMODEL == " ## #WKMODEL
#endif
                PERFSTOP(Iapply, IP(at->nlocal));
              }

          }
      }
    else
      {
        if (funcs_.find(func) == funcs_.end())
          {
            std::cout << "function: " << func << endl;
            throw sstmac::libraryerror(
                "lib_compute_minimd: cannot find function");
          }

        lib_compute_time::compute(funcs_[func]);
      }
  }

  int
  lib_compute_minimd::get_int(const std::string &func, int index)
  {
    return 1;
  }

}
