/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <minimd.h>
#include <minimd-in.h>
#include <minimd-atom.h>
#include <minimd-force.h>
#include <minimd-neighbor.h>
#include <minimd-integrate.h>
#include <minimd-thermo.h>
#include <minimd-comm.h>
#include <minimd-timer.h>
#include <lib_compute_minimd.h>
#include <sstmac/software/api/mpi/mpiapi.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>
#include <sstmac/common/basicstringtokenizer.h>
#include <boost/lexical_cast.hpp>
#include <fstream>
#include <math.h>

#include "minimd-output.h"
#include "perf.h"
#include "mysql++/mysql++.h"

int force_link_msqlpp(int i) {
  if (i) {
    mysqlpp::Connection* conn = new mysqlpp::Connection(true);
  }
  return 0;
}

double get_time()
{
    double t = sstmac::eventmanager::global->now().sec();
    return t;
}

using namespace sstmac;
using namespace sstmac::sw;

namespace mini {
    SSTRegisterApp("minimd", minimd, "miniapp for molecular dynamics");

    // Pointer types.

    // Utility function.
    inline void
    gettok(std::ifstream &infile, std::string &line,
        std::deque<std::string> &tok, size_t min_tokens, const char *message)
    {
      getline(infile, line);
      if (!infile.good())
        throw valueerror(std::string(message) + ":  Insufficient input lines.");
      tok.clear();
      pst::BasicStringTokenizer::tokenize(line, tok);
      if (tok.size() < min_tokens)
        throw valueerror(std::string(message) + "  Insufficient input fields.");
    }

    // It takes really long to type lexical_cast<> -- blc is boost::lexical_cast
    template<typename T>
      T
      blc(const std::string &value)
      {
        return lexical_cast<T>(value);
      }

    //
    // Simulate the parameter broadcasting 
    //
    void
    minimd::input()
    {
        mpicomm::ptr world = mpi()->comm_world();
        mpi()->comm_rank(world);
        mpi()->bcast(1, mpitype::mpi_int, mpiid(0), world);

        mpi()->barrier(world);

        mpi()->bcast(1, mpitype::mpi_double, mpiid(0), world);
        mpi()->bcast(1, mpitype::mpi_double, mpiid(0), world);
        mpi()->bcast(1, mpitype::mpi_double, mpiid(0), world);
        mpi()->bcast(1, mpitype::mpi_double, mpiid(0), world);
        mpi()->bcast(1, mpitype::mpi_double, mpiid(0), world);
        mpi()->bcast(1, mpitype::mpi_double, mpiid(0), world);
        mpi()->bcast(1, mpitype::mpi_int, mpiid(0), world);
        
        mpi()->bcast(1, mpitype::mpi_int, mpiid(0), world);
        mpi()->bcast(1, mpitype::mpi_int, mpiid(0), world);
        mpi()->bcast(1, mpitype::mpi_int, mpiid(0), world);
        mpi()->bcast(1, mpitype::mpi_int, mpiid(0), world);
        mpi()->bcast(1, mpitype::mpi_int, mpiid(0), world);
        mpi()->bcast(1, mpitype::mpi_int, mpiid(0), world);
        mpi()->bcast(1, mpitype::mpi_int, mpiid(0), world);
    }

    //
    // Set up the box.
    //
    void
    minimd::create_box()
    {
      double latt = pow((4 / in_->rho), (1.0 / 3.0));
      atom_->box.set_prd(in_->nx * latt, in_->ny * latt, in_->nz * latt);
    }

    //
    // Pretend to compute all atom positions etc.
    //
    void
    minimd::create_atoms()
    {
      mpicomm::ptr world = mpi()->comm_world();
      int atoms = 4 * in_->nx * in_->ny * in_->nz;
      assert(atoms>0 || 0 == "atoms overflow bad");
      int size = world->size().id;
      int rank = world->rank().id;
      atom_->natoms = atoms;
      atom_->nghost = in_->nghostavg;
      atom_->ratio_neighbors = in_->ratio_neighbors;
      atom_->frac_scat = in_->frac_scat;
      // We're faking the nlocal to be evenly divided (best-case scenario).
      int eachatoms = atoms / size;
      int sum = eachatoms * size;
      atom_->neighatoms.clear();
      atom_->neighatoms.resize(size, eachatoms);
      for (int i = 0; sum + i < atoms; ++i)
        ++atom_->neighatoms.at(i);
      atom_->nlocal = atom_->neighatoms.at(rank);
      // Now pretend to create the actual atoms like miniMD does.
      execution_->compute("create_atoms", atom_, in_);

      // Simulated mpi call
      mpi()->comm_rank(world);
      mpi()->allreduce(1, mpitype::mpi_int, mpiop::max, world);
      mpi()->allreduce(1, mpitype::mpi_int, mpiop::sum, world);
    }

    //
    // Pretend to assign velocities to all atoms.
    //
    void
    minimd::create_velocity()
    {
      mpicomm::ptr world = mpi()->comm_world();
      std::cerr << "minimd::create_velocity compute1" << std::endl;
      execution_->compute("create_velocity", atom_, ptr_type::ptr(), 0);
      std::cerr << "minimd::create_velocity compute1 done" << std::endl;
      thermo_->temperature(atom_);
      mpi()->allreduce(1, mpitype::mpi_double, mpiop::sum, world);
      mpi()->allreduce(1, mpitype::mpi_double, mpiop::sum, world);
      mpi()->allreduce(1, mpitype::mpi_double, mpiop::sum, world);
      std::cerr << "minimd::create_velocity compute2" << std::endl;
      execution_->compute("create_velocity", atom_, ptr_type::ptr(), 1);
      std::cerr << "minimd::create_velocity compute2 done" << std::endl;
    }

    //
    // We're sort of ignoring output for now.
    //
    void
    minimd::output()
    {
      // get the sum of all profiled timings for this method.
      execution_->compute("output", ptr_type::ptr(), ptr_type::ptr());
      
      mpicomm::ptr world = mpi()->comm_world();
      output_->output(in_ , atom_, force_, neighbor_, comm_, thermo_, integrate_, timer_);

    }

    //
    // Hi.
    //
    minimd::minimd(softwareid id) :
        mpiapp(id, to_string()), execution_(lib_compute_minimd::construct(id)), in_(new in), atom_(
            new atom), force_(new force), neighbor_(new neighbor), integrate_(
            new integrate), thermo_(new thermo), comm_(new comm), timer_(
            new timer), output_(new Output)
    {
      register_lib(execution_);
    }

    //
    // Build a new minimd object.
    //
    minimd::ptr
    minimd::construct(softwareid id)
    {
      minimd::ptr retval(new minimd(id));
      //retval->set_self(retval);
      return retval;
    }

    //
    // Goodbye.
    //
    minimd::~minimd() throw ()
    {
    }

    //
    // Get a copy of this object.
    //
    app::ptr
    minimd::clone(softwareid newid)
    {
      return construct(newid);
    }

    //
    // This object gets access to the mpi api.
    //
    void
    minimd::consume_params(const sim_parameters::ptr &params)
    {
      std::cerr << "minimd::consume_params"<<std::endl;
      in_->init(execution_, mpi());
      atom_->init(execution_, mpi());
      force_->init(execution_, mpi());
      neighbor_->init(execution_, mpi());
      integrate_->init(execution_, mpi());
      thermo_->init(execution_, mpi());
      comm_->init(execution_, mpi());
      timer_->init(execution_, mpi());
      output_->init(execution_, mpi());

      in_->nx = params->get_int_param("minimd_nx");
      in_->ny = params->get_int_param("minimd_ny");
      in_->nz = params->get_int_param("minimd_nz");
      assert(in_->nx>0 || 0 == "minimd_nx parameter bad");
      assert(in_->ny>0 || 0 == "minimd_ny parameter bad");
      assert(in_->nz>0 || 0 == "minimd_nz parameter bad");

      integrate_->ntimes = params->get_int_param("minimd_ntimes");
      std::cerr << "minimd_ntimes " << integrate_->ntimes  << std::endl;


      neighbor_->set_bins(params->get_int_param("minimd_bins_x"),
          params->get_int_param("minimd_bins_y"),
          params->get_int_param("minimd_bins_z"));
      std::cerr << "minimd_bins_x,y,z " << neighbor_->nbinx 
	<< " " << neighbor_->nbiny 
	<< " " << neighbor_->nbinz  << std::endl;

      integrate_->dt = params->get_double_param("minimd_timestep");
      std::cerr << "minimd_timestep " << integrate_->dt  << std::endl;

      in_->t_request = params->get_double_param("minimd_temp");
      std::cerr << "minimd_temp " << in_->t_request << std::endl;

      in_->rho = params->get_double_param("minimd_density");
      std::cerr << "minimd_density " << in_->rho << std::endl;

      neighbor_->every = params->get_int_param("minimd_reneighbor");
      std::cerr << "minimd_reneighbor " << neighbor_->every << std::endl;

      force_->cutforce = params->get_double_param( "minimd_cutoff_inner");
      std::cerr << "minimd_cutoff_inner " << force_->cutforce << std::endl;
      neighbor_->cutneigh = params->get_double_param("minimd_cutoff_outer");
      std::cerr << "minimd_cutoff_outer " << neighbor_->cutneigh << std::endl;

      thermo_->nstat = params->get_int_param("minimd_thermo");
      std::cerr << "minimd_thermo "  << thermo_->nstat << std::endl;

      neighbor_->mbins = params->get_int_param("minimd_neighbor_mbins");
      std::cerr << "minimd_neighbor_mbins " << neighbor_->mbins << std::endl;

      in_->nghostavg = params->get_int_param("minimd_atom_nghostavg");
      std::cerr << "minimd_atom_nghostavg " << in_->nghostavg << std::endl;

      in_->ratio_neighbors = params->get_int_param("minimd_atom_ratio_neighbors");
      std::cerr << "minimd_atom_ratio_neighbors " << in_->ratio_neighbors << std::endl;

      in_->frac_scat = params->get_double_param("minimd_atom_frac_scat");
      std::cerr << "minimd_atom_frac_scat " << in_->frac_scat << std::endl;

      neighbor_->avgtripsj1 = params->get_int_param("minimd_neighbor_avgtripsj1");
      std::cerr << "minimd_neighbor_avgtripsj1 " << neighbor_->avgtripsj1 << std::endl;
      neighbor_->avgtripsj2 = params->get_int_param("minimd_neighbor_avgtripsj2");
      std::cerr << "minimd_neighbor_avgtripsj2 " << neighbor_->avgtripsj2 << std::endl;
      neighbor_->avgtripsj3 = params->get_int_param("minimd_neighbor_avgtripsj3");
      std::cerr << "minimd_neighbor_avgtripsj3 " << neighbor_->avgtripsj3 << std::endl;
      neighbor_->nstencil = params->get_int_param("minimd_neighbor_nstencil");
      std::cerr << "minimd_neighbor_nstencil " << neighbor_->nstencil << std::endl;
      neighbor_->avgallocs = params->get_int_param("minimd_neighbor_avgallocs");
      std::cerr << "minimd_neighbor_avgallocs " << neighbor_->avgallocs << std::endl;
// pattern
#define GETDOUBLE(s,x) s->x = params->get_double_param("minimd_" #x); std::cerr << "minimd_" #x " " << s->x  << std::endl
      atom_->Fclear_nl = params->get_double_param("minimd_Fclear_nl"); std::cerr << "minimd_Fclear_nl " << atom_->Fclear_nl << std::endl;
      atom_->Nbinatoms_nl = params->get_double_param("minimd_Nbinatoms_nl"); std::cerr << "minimd_Nbinatoms_nl " << atom_->Nbinatoms_nl << std::endl;
      atom_->Nbuild_nl1 = params->get_double_param("minimd_Nbuild_nl1"); std::cerr << "minimd_Nbuild_nl1 " << atom_->Nbuild_nl1 << std::endl;
      atom_->Nbuild_nl2 = params->get_double_param("minimd_Nbuild_nl2"); std::cerr << "minimd_Nbuild_nl2 " << atom_->Nbuild_nl2 << std::endl;
      atom_->Nbuild_nl3 = params->get_double_param("minimd_Nbuild_nl3"); std::cerr << "minimd_Nbuild_nl3 " << atom_->Nbuild_nl3 << std::endl;
      atom_->Ttemperature_nl = params->get_double_param("minimd_Ttemperature_nl"); std::cerr << "minimd_Ttemperature_nl " << atom_->Ttemperature_nl << std::endl;
      atom_->Tpressure_nl = params->get_double_param("minimd_Tpressure_nl"); std::cerr << "minimd_Tpressure_nl " << atom_->Tpressure_nl << std::endl;
      atom_->Imove_nl = params->get_double_param("minimd_Imove_nl"); std::cerr << "minimd_Imove_nl " << atom_->Imove_nl << std::endl;
      atom_->Iapply_nl = params->get_double_param("minimd_Iapply_nl"); std::cerr << "minimd_Iapply_nl " << atom_->Iapply_nl << std::endl;
      atom_->Fscatter_offset = params->get_double_param("minimd_Fscatter_offset"); std::cerr << "minimd_Fscatter_offset " << atom_->Fscatter_offset << std::endl;
      atom_->Fscatter_coef = params->get_double_param("minimd_Fscatter_coef"); std::cerr << "minimd_Fscatter_coef " << atom_->Fscatter_coef << std::endl;
      atom_->Tenergy_offset = params->get_double_param("minimd_Tenergy_offset"); std::cerr << "minimd_Tenergy_offset " << atom_->Tenergy_offset << std::endl;
      atom_->Tenergy_coef = params->get_double_param("minimd_Tenergy_coef"); std::cerr << "minimd_Tenergy_coef " << atom_->Tenergy_coef << std::endl;

#undef GETDOUBLE
    }

    //
    // Go.
    //
    void
    minimd::skeleton_main()
    {
      force_link_msqlpp(0);
#define SSTMAC_DB(command) std::cerr << rank << " " << #command << "\n"; command
//#define SSTMAC_DB(command) command
      timestamp startat = mpi()->init();
      mpiid rank = mpi()->comm_rank(mpi()->comm_world());
      int msz = mpi()->comm_world()->size().id;
      //int size = world->size().id;
      PERFDECL(PERF::init();)
      PERFDECL(PERF::mpiArgs(rank.id,msz);) // fixme: serial hardwired for now
      if (msz==1) {
        // PERFDECL(PERF::stringOptions("sst-x5550","gcc","minimd.cpp","minimdEcpu","tesla.",".log");)
        PERFDECL(PERF::stringOptions("sst-x5550","gcc","minimd.cpp","minimdcpu","tesla.",".log");)
      } else { // expecting 8 for now
        // PERFDECL(PERF::stringOptions("sst-x5550","gcc","minimd.cpp","minimd8Ecpu","tesla.",".log");)
        PERFDECL(PERF::stringOptions("sst-x5550","gcc","minimd.cpp","minimd8cpu","tesla.",".log");)
      }

      SSTMAC_DB( this->input());
      SSTMAC_DB( this->create_box());

      // This is, conveniently, pretty much a fixed overhead.
      // unsigned int seedp = rank.id * 12 + 123;
      // double randnoise = (double(rand_r(&seedp)) / RAND_MAX - 0.5) * 0.05;
      timestamp time((0.0003481));
      app::compute(time);
      //SSTMAC_DB( mpi()->compute(execution_->get("main", 0)) );
      SSTMAC_DEBUG << "minimd[" << rank.id << "]: setting up \n";
      SSTMAC_DB( comm_->setup(neighbor_->cutneigh, atom_));
      SSTMAC_DB( this->create_atoms());
      SSTMAC_DB( this->create_velocity());

      SSTMAC_DEBUG << "minimd[" << rank.id << "]: exchanging atoms \n";
      SSTMAC_DB( comm_->exchange(atom_));
      SSTMAC_DB( comm_->borders(atom_));
      SSTMAC_DB( neighbor_->build(atom_));
      SSTMAC_DB( thermo_->compute(0, atom_, neighbor_, force_));

      // Starting dynamics.
      SSTMAC_DEBUG << "minimd[" << rank.id << "]: starting dynamics \n";
      SSTMAC_DB( timer_->barrier_start(timer::TIME_TOTAL));
      SSTMAC_DB(
          integrate_->run(atom_, force_, neighbor_, comm_, thermo_, timer_));
      SSTMAC_DB( timer_->barrier_stop(timer::TIME_TOTAL));

      SSTMAC_DB( thermo_->compute(-1, atom_, neighbor_, force_));
      SSTMAC_DB( this->output());


      // This compute time is also not really dependent on problem size,
      // although it is a little noisier than the earlier value.
      // randnoise = (double(rand_r(&seedp)) / RAND_MAX - 0.5) * 0.10;
      time = timestamp(0.001495);
      app::compute(time);
      //SSTMAC_DB( mpi()->compute(execution_->get("main", 1)) );
      /* timestamp endat = */SSTMAC_DB( mpi()->finalize() );
      //timestamp endat = SSTMAC_DB( mpi()->finalize() );
      PERFDECL(PERF::finalize();)

#undef SSTMAC_DB
    }

} //end namespace
