/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <gtc/gtc_point.h>

#include <sstmac/software/api/hpx/runtime/hpx_thread_manager.h>
#include <sstmac/software/api/hpx/lcos/future_wait.h>
#include <iostream>
#include <fstream>
#include <gtc/gtc_point_stubs.h>
#include <math.h>


namespace hpx_gtc
{

// char INIT_NAME[] = "gtc_point::init";
// char LOAD_NAME[] = "gtc_point::load";
// char CHARGEI_NAME[] = "gtc_point::chargei";

  using namespace hpx;
  using namespace sstmac;
  using namespace sstmac::sw;

/// Initialize the accumulator
  int
  gtc_app::gtc_point::init(std::size_t objectid, config_parameters::ptr par,
       std::vector<stub_ptr> points)
  {

    idx_ = objectid;
    SSTMAC_DEBUG << "gtc_point(" << idx_ << ")::init() \n";
    part_decomp_ = particle_decomp::construct(par);
    part_decomp_->setup(par, idx_);
    point_components_ = points;
    params_ = par;
    // --- Define poloidal grid ---

    // grid shift associated with fieldline following coordinates

    field_array_ = field_array::construct(par->mpsi_);
    field_array_->setup(par, par->gyroradius_, part_decomp_, idx_);

    // initial data
    SSTMAC_DEBUG << "computing loop to " << par->mpsi_ << "\n";
    comp_->compute_loop(0, par->mpsi_ + 1, 15);

    // When doing mode diagnostics, we need to switch from the field-line following
    // coordinates alpha-zeta to a normal geometric grid in theta-zeta. This
    // translates to a greater number of grid points in the zeta direction, which
    // is mtdiag. Precisely, mtdiag should be mtheta/q but since mtheta changes
    // from one flux surface to another, we use a formula that gives us enough
    // grid points for all the flux surfaces considered.

    // starting point for a poloidal grid

    comp_->compute_loop(0, par->mpsi_ + 1, 1);

    // number of grids on a poloidal plane

    comp_->compute_loop(0, par->mpsi_ + 1, 1);

    //mimax_ = mi_ + 100*(std::size_t)(std::ceil(sqrt(mi_))); // ions array upper bound
    //std::size_t memax = static_cast<std::size_t>(me_ + 100*std::ceil(sqrt(tmp14))); // electrons array upper bound

    // initialize arrays
    // temperature and density on the grid, T_i=n_0=1 at mid-radius
    comp_->compute_array_init<double>(par->mpsi_ + 1);
    comp_->compute_array_init<double>(par->mpsi_ + 1);
    comp_->compute_array_init<double>(par->mpsi_ + 1);
    comp_->compute_array_init<double>(par->mpsi_ + 1);
    comp_->compute_array_init<double>(par->mpsi_ + 1);
    comp_->compute_array_init<double>(par->mpsi_ + 1);
    comp_->compute_array_init<double>(par->mpsi_ + 1);
    comp_->compute_array_init<double>(par->mpsi_ + 1);
    comp_->compute_array_init<double>(par->mpsi_ + 1);
    comp_->compute_array_init<double>(par->mpsi_ + 1);

    // # of marker per grid, Jacobian=(1.0+r*cos(theta+r*sin(theta)))*(1.0+r*cos(theta))
    comp_->compute_array_init<double>(par->mpsi_ + 1);

    comp_->compute_loop2(0, field_array_->mgrid_, 1, part_decomp_->mzeta_, 4);
    comp_->compute_loop(0, field_array_->mgrid_, 5);
    comp_->compute_loop2(0, field_array_->mgrid_, 1, part_decomp_->mzeta_, 3);

    // 4-point gyro-averaging for sqrt(mu)=gyroradius on grid of magnetic coordinates
    // rho=gyroradius*sqrt(2/(b/b_0))*sqrt(mu/mu_0), mu_0*b_0=m*v_th^2
    // dtheta/delta_x=1/(r*(1+r*cos(theta))), delta_x=poloidal length increase
    comp_->compute_loop(0, field_array_->mgrid_, 13);

    // initiate radial interpolation for grid
    comp_->compute_loop2(1, part_decomp_->mzeta_, 1, field_array_->mgrid_ * 2,
        15);

    return HPX_SUCCESS;
  }

  int
  gtc_app::gtc_point::load()
  {
    SSTMAC_DEBUG << "gtc_point(" << idx_ << ")::load() \n";
    // initialize random number generator
    comp_->compute_loop(1, field_array_->mi_, 2);

    // Set zion(2:6,1:mi) to uniformly distributed random values between 0 and 1
    // set random zion
    comp_->compute_loop2(2, 6, 1, field_array_->mi_, 1);

    // poloidal: uniform in alpha=theta_0+r*sin(alpha_0), theta_0=theta+r*sin(theta)
    comp_->compute_loop(1, field_array_->mi_, 2);

    comp_->compute_loop2(1, 10, 1, field_array_->mi_, 2);

    // Maxwellian distribution in v_para, <v_para^2>=1.0, use zion0(4,:) as temporary storage

    comp_->compute_loop(1, field_array_->mi_, 8);

    comp_->compute_loop(1, field_array_->mi_, 4);

    // transform zion(1,:) to psi, zion(4,:) to rho_para, zion(6,:) to sqrt(mu)

    comp_->compute_loop(1, field_array_->mi_, 4);

    if (params_->iload_ == false)
      {
        comp_->compute_loop(1, field_array_->mi_, 1);
      }

    return HPX_SUCCESS;
  }

  int
  gtc_app::gtc_point::chargei(std::size_t istep)
  {
    SSTMAC_DEBUG << "gtc_point(" << idx_ << ")::chargei(" << istep << ") \n";

    comp_->compute_loop(0, params_->mpsi_ + 1, 1);

    comp_->compute_array_init<double>(
        part_decomp_->mzeta_ * field_array_->mgrid_);

    comp_->compute_loop(1, field_array_->mi_, 16);
    comp_->compute_loop2(1, field_array_->mi_, 1, 4, 17);

    if (istep == 0)
      return HPX_SUCCESS;

    comp_->compute_loop2(1, field_array_->mi_, 1, 4, 18);

    // poloidal end cell, discard ghost cell j=0
    comp_->compute_loop2(0, params_->mpsi_, 0, part_decomp_->mzeta_, 1);

    // send densityi to the left; receive from the right
      {
        typedef std::vector<hpx::lcos::promise<fake_array<double> >::ptr  > lazy_results_type;
        lazy_results_type lazy_results;
        lazy_results.push_back(
            point_components_[part_decomp_->right_pe_]->get_densityi_async());
        hpx::lcos::wait<fake_array<double> >(lazy_results);
      }

    if (part_decomp_->myrank_toroidal_ == params_->ntoroidal_ - 1)
      {
        // B.C. at zeta=2*pi is shifted
        size_t strides[] =
          { 1, 1 };
        std::size_t lengths[2];
        comp_->compute_loop(0, params_->mpsi_, 8);
        comp_->compute_loop(0, field_array_->mgrid_, 1);

      }
    else
      {
        // B.C. at zeta<2*pi is continuous
        comp_->compute_loop(0, field_array_->mgrid_, 1);

      }

    // zero out charge in radial boundary cell
    comp_->compute_loop2(0, params_->nbound_, 0, field_array_->mgrid_, 2);

    // flux surface average and normalization
    comp_->compute_array_init<double>(params_->mpsi_ + 1);

    comp_->compute_loop2(0, field_array_->mgrid_, 1, part_decomp_->mzeta_, 3);

    // global sum of phi00 broadcast to every toroidal PE
      {
        comp_->compute_array_init<double>(params_->mpsi_ + 1);

        typedef std::vector<hpx::lcos::promise<fake_array<double> >::ptr  > lazy_results_type;
        lazy_results_type lazy_results;
        for (size_t i = 0; i < point_components_.size(); i++)
          {
            lazy_results.push_back(point_components_[i]->get_zonali_async());
          }
        hpx::components::hpx_component::ptr self = ptr(this);
        hpx::lcos::wait<fake_array<double>, gtc_app::gtc_point>(lazy_results,
            &gtc_app::gtc_point::chargei_zonali_callback, self);
      }

    comp_->compute_loop(0, params_->mpsi_ + 1, 1);

    comp_->compute_loop2(0, field_array_->mgrid_, 1, part_decomp_->mzeta_, 2);
    comp_->compute_loop2(0, params_->mpsi_, 1, part_decomp_->mzeta_, 1);

    // enforce charge conservation for zonal flow mode

    comp_->compute_loop(0, params_->mpsi_ - 1, 4);

    return HPX_SUCCESS;
  }

  bool
  gtc_app::gtc_point::chargei_zonali_callback(std::size_t i,
      fake_array<double> const& zonali)
  {
    SSTMAC_DEBUG << "gtc_point(" << idx_ << ")::chargei_zonali_callback() \n";
    comp_->compute_loop(0, params_->mpsi_ + 1, 1);
    return HPX_SUCCESS;
  }

  int
  gtc_app::gtc_point::smooth(std::size_t iflag, std::size_t idiag)
  {
    SSTMAC_DEBUG << "gtc_point(" << idx_ << ")::smooth(" << iflag << ", "
          << idiag << ") \n";
    //short names
    field_array::ptr fa = field_array_;
    particle_decomp::ptr pd = part_decomp_;
    config_parameters::ptr p = params_;

    comp_->compute_loop2(0, pd->mzeta_, 0, fa->mgrid_, 1);

    if (iflag == 0)
      {
        comp_->compute_loop2(1, pd->mzeta_, 1, fa->mgrid_, 1);

      }
    else if (iflag == 1)
      {
        comp_->compute_loop2(1, pd->mzeta_, 1, fa->mgrid_, 1);
      }
    else
      {
        comp_->compute_loop2(1, pd->mzeta_, 1, fa->mgrid_, 1);
      }

    std::size_t ismooth;
    ismooth = 1;
    if (p->nonlinear_ < 0.5)
      ismooth = 0;

    for (std::size_t ip = 1; ip <= ismooth; ip++)
      {
        // radial smoothing
        comp_->compute_loop2(1, p->mpsi_ - 1, 0, pd->mzeta_, 1);

        comp_->compute_loop2(1, pd->mzeta_, 1, fa->mgrid_, 5);

        // poloidal smoothing (-0.0625 0.25 0.625 0.25 -0.0625)
        comp_->compute_loop2(1, fa->mgrid_, 1, pd->mzeta_, 3);
        comp_->compute_loop2(1, p->mpsi_, 1, pd->mzeta_, 4);

        // parallel smoothing
        // get phi from the left
          {
            typedef std::vector<hpx::lcos::promise<fake_array<double> >::ptr  > lazy_results_type;
            lazy_results_type lazy_results;
            lazy_results.push_back(
                point_components_[pd->left_pe_]->get_phi_async(pd->mzeta_));
            hpx::lcos::wait(lazy_results);
          }

        // get phi from the right
          {
            typedef std::vector<hpx::lcos::promise<fake_array<double> >::ptr  > lazy_results_type;
            lazy_results_type lazy_results;
            lazy_results.push_back(
                point_components_[pd->right_pe_]->get_phi_async(1));
            hpx::lcos::wait(lazy_results);
          }

        comp_->compute_loop(1, fa->mgrid_, 3);

        if (pd->mzeta_ == 1)
          {
            comp_->compute_loop(1, fa->mgrid_, 1);
          }
        else if (pd->mzeta_ == 2)
          {
            comp_->compute_loop(1, fa->mgrid_, 2);
          }
        else
          {
            comp_->compute_loop(1, fa->mgrid_, 2);
            comp_->compute_loop2(1, fa->mgrid_, 2, pd->mzeta_ - 1, 1);
          }

      }

    // toroidal BC: send phi to right and receive from left
      {
        typedef std::vector<hpx::lcos::promise<fake_array<double> >::ptr  > lazy_results_type;
        lazy_results_type lazy_results;
        lazy_results.push_back(
            point_components_[pd->left_pe_]->get_phi_async(pd->mzeta_));
        hpx::lcos::wait(lazy_results);
      }

    if (pd->myrank_toroidal_ == 0)
      {
        comp_->compute_loop(1, fa->mgrid_, 2);
      }
    else
      {
        comp_->compute_loop(1, pd->mzeta_, 1);

      }

    // poloidal BC
    comp_->compute_loop2(1, p->mpsi_, 0, pd->mzeta_, 1);

    // radial boundary
    comp_->compute_array_init<double>(p->mpsi_ * pd->mzeta_);
    comp_->compute_array_init<double>(p->mpsi_ * pd->mzeta_);

    if (iflag == 0)
      {
        comp_->compute_loop2(1, fa->mgrid_, 0, pd->mzeta_, 1);

      }
    else if (iflag == 1)
      {
        comp_->compute_loop2(1, fa->mgrid_, 0, pd->mzeta_, 1);

      }
    else
      {
        comp_->compute_loop2(1, fa->mgrid_, 0, pd->mzeta_, 1);

      }

    // solve zonal flow: phi00=r*E_r, E_r(a0)=0. Trapezoid rule

    if (iflag == 3)
      {
        if (p->nhybrid_ == 0)
          {
            comp_->compute_loop(0, p->mpsi_, 1);
          }
        if (p->nhybrid_ > 0)
          {
            comp_->compute_loop(0, p->mpsi_, 1);
          }

        for (std::size_t ismooth = 1; ismooth <= 1; ismooth++)
          {
            comp_->compute_loop(2, p->mpsi_ - 2, 1);
            comp_->compute_loop(1, p->mpsi_ - 1, 2);

          }

        comp_->compute_loop(1, p->mpsi_ - 1, 2);

        // d phi/dr, in equilibrium unit
        comp_->compute_loop(0, p->mpsi_, 2);

        // add FLR contribution using Pade approximation: b*<phi>=(1+b)*<n>

        comp_->compute_loop(0, p->mpsi_, 1);

        comp_->compute_loop(1, p->mpsi_ - 1, 2);

        // (0,0) mode potential store in phi00
        comp_->compute_array_init<double>(p->mpsi_);
        comp_->compute_loop(1, p->mpsi_, 2);

        if (p->mode00_ == 0)
          {
            comp_->compute_array_init<double>(p->mpsi_);
          }
      }

    // Interpolate on a flux surface from fieldline coordinates to magnetic
    // coordinates. Use mtdiag for both poloidal and toroidal grid points.

    if (iflag > 1)
      {
        if (p->nonlinear_ < 0.5 || (iflag == 3 && idiag == 0))
          {
            comp_->compute_array_init<double>(p->mtdiag_);
            comp_->compute_array_init<long double>(p->mtdiag_ / 2 + 1 + 1);

            // ESSL would normally go here
            int one = 1;
            std::size_t mzbig = (std::max)(one, p->mtdiag_ / p->mzetamax_);
            std::size_t mzmax = p->mzetamax_ * mzbig;
            std::size_t mz = pd->mzeta_ * mzbig;
            std::size_t meachtheta = p->mtdiag_ / p->ntoroidal_;
            std::size_t icount = meachtheta * mz
                * (p->idiag2_ - p->idiag1_ + 1);

            comp_->compute_array_init<double>(p->mtdiag_ / 2 + 1 + 1);
            comp_->compute_loop(0, p->nmode_.size(), 1);

            comp_->compute_array_init<double>(
                (p->idiag2_ - p->idiag1_ + 1) * p->mtdiag_ * p->mtdiag_
                    / p->ntoroidal_ + 1);

            comp_->compute_loop4(1, pd->mzeta_, 1, mzbig, 1,
                p->idiag2_ - p->idiag1_, 1, p->mtdiag_, 6);

            // transpose 2-d matrix from (ntoroidal,mzeta*mzbig) to (1,mzetamax*mzbig)

            for (std::size_t jpe = 0; jpe <= p->ntoroidal_ - 1; jpe++)
              {
                comp_->compute_loop3(1, meachtheta, 1, p->idiag2_ - p->idiag1_,
                    1, mz, 2);

                // Gather
                  {
                    typedef std::vector<hpx::lcos::promise<fake_array<double> >::ptr  > lazy_results_type;
                    lazy_results_type lazy_results;
                    for (size_t i = 0; i < point_components_.size(); i++)
                      {
                        lazy_results.push_back(
                            point_components_[i]->get_eachzeta_async());
                      }

                    std::size_t length = (p->idiag2_ - p->idiag1_ + 1)
                        * p->mtdiag_ * p->mtdiag_ / p->ntoroidal_
                        / p->ntoroidal_;
                    ptr self(this);
                    hpx::lcos::wait<fake_array<double>, gtc_app::gtc_point,
                        size_t>(lazy_results,
                        &gtc_app::gtc_point::eachzeta_callback, self, length);

                  }

              }

            // transform to k space
            comp_->compute_loop4(1, meachtheta, 1, p->idiag2_ - p->idiag1_, 0,
                p->ntoroidal_, 1, mz, 2);

            for (std::size_t j = 1; j <= meachtheta; j++)
              {

                for (std::size_t i = p->idiag1_; i <= p->idiag2_; i++)
                  {

                    // call fftr1d(1,mtdiag,scale,xz,yz,2)
                    comp_->compute_fft();

                    // record mode information for diagnostic
                    if (i == p->mpsi_ / 2)
                      {
                        comp_->compute_loop(1, p->num_mode_, 1);

                      }

                    if (p->nonlinear_ < 0.5)
                      {
                        // linear run only keep a few modes
                        comp_->compute_loop(0, p->mtdiag_ / 2 + 1 + 1, 1);

                        // transform back to real space

                        // call fftr1d(-1,mtdiag,scale,xz,yz,2)
                        comp_->compute_fft();

                        // transpose back to (ntoroidal,mz)
                        comp_->compute_loop2(0, p->ntoroidal_, 1, mz, 2);

                      }
                  }
              }

            if (p->nonlinear_ < 0.5)
              {

              }

          }
      }

    if (iflag == 2)
      {

      }

    if (iflag == 3 && idiag == 0)
      {
        // Diagnostic not implemented yet
      }

    return HPX_SUCCESS;
  }

  bool
  gtc_app::gtc_point::eachzeta_callback(std::size_t i,
      fake_array<double> const& eachzeta, std::size_t length)
  {
    SSTMAC_DEBUG << "gtc_point(" << idx_ << ")::eachzeta_callback() \n";
    comp_->compute_loop(0, length, 1);
    return HPX_SUCCESS;
  }

  fake_array<double>
  gtc_app::gtc_point::get_densityi()
  {
    SSTMAC_DEBUG << "gtc_point(" << idx_ << ")::get_densityi() \n";
    fake_array<double> ret(1);
    return ret;
  }

  fake_array<double>
  gtc_app::gtc_point::get_zonali()
  {
    SSTMAC_DEBUG << "gtc_point(" << idx_ << ")::get_zonali() \n";
    fake_array<double> ret(params_->mpsi_ + 1);
    return ret;
  }

  fake_array<double>
  gtc_app::gtc_point::get_phi(std::size_t depth)
  {
    SSTMAC_DEBUG << "gtc_point(" << idx_ << ")::get_phi(" << depth << ") \n";
    fake_array<double> ret(depth);
    return ret;
  }

  fake_array<double>
  gtc_app::gtc_point::get_eachzeta()
  {
    SSTMAC_DEBUG << "gtc_point(" << idx_ << ")::get_eachzeta() \n";
    fake_array<double> ret(
        (params_->idiag2_ - params_->idiag1_ + 1) * params_->mtdiag_
            * params_->mtdiag_ / params_->ntoroidal_ / params_->ntoroidal_ + 1);
    return ret;
  }

  int
  gtc_app::gtc_point::field()
  {
    SSTMAC_DEBUG << "gtc_point(" << idx_ << ")::smooth() \n";
    //short names
    field_array::ptr fa = field_array_;
    particle_decomp::ptr pd = part_decomp_;
    config_parameters::ptr p = params_;

    // finite difference for e-field in equilibrium unit

    comp_->compute_loop(0, p->mpsi_, 1);

    comp_->compute_array_init<double>(fa->mgrid_ * pd->mzeta_ * 3);

    comp_->compute_loop2(1, pd->mzeta_, 1, fa->mgrid_, 6);

    // get phi from the left
      {
        typedef std::vector<hpx::lcos::promise<fake_array<double> >::ptr  > lazy_results_type;
        lazy_results_type lazy_results;
        lazy_results.push_back(
            point_components_[pd->left_pe_]->get_phi_async(pd->mzeta_));
        hpx::lcos::wait(lazy_results);
      }

    // get phi from the right
      {
        typedef std::vector<hpx::lcos::promise<fake_array<double> >::ptr  > lazy_results_type;
        lazy_results_type lazy_results;
        lazy_results.push_back(
            point_components_[pd->right_pe_]->get_phi_async(1));
        hpx::lcos::wait(lazy_results);
      }

    // unpack phi_boundary and calculate E_zeta at boundaries, mzeta=1
    int add =
        (pd->mzeta_ == 1) ? 1 : ((pd->mzeta_ == 2) ? 2 : (2 + pd->mzeta_ - 3));
    comp_->compute_loop(1, fa->mgrid_, 3 + add);

    // adjust the difference between safety factor q and qtinv for fieldline coordinate
    comp_->compute_loop2(1, fa->mgrid_, 1, fa->mgrid_, 1);

    // add (0,0) mode, d phi/d psi
    if (p->mode00_ == 1)
      {
        comp_->compute_loop2(1, p->mpsi_, 1, pd->mzeta_, 1);

      }

    // get evector from the left
      {
        typedef std::vector<hpx::lcos::promise<fake_array<double> >::ptr  > lazy_results_type;
        lazy_results_type lazy_results;
        lazy_results.push_back(
            point_components_[pd->left_pe_]->get_evector_async(1, pd->mzeta_));
        lazy_results.push_back(
            point_components_[pd->left_pe_]->get_evector_async(2, pd->mzeta_));
        lazy_results.push_back(
            point_components_[pd->left_pe_]->get_evector_async(3, pd->mzeta_));
        ptr self(this);
        hpx::lcos::wait<fake_array<double>, gtc_point>(lazy_results,
            &gtc_point::evector_callback, self);
      }

    // unpack end point data for k=0
    if (pd->myrank_toroidal_ == 0)
      { // down-shift for zeta=0
        std::valarray<double> trecvlA, trecvlB, trecvlC;
        comp_->compute_loop(1, fa->mgrid_, 6);

      }
    else
      {
        comp_->compute_loop(1, fa->mgrid_, 3);

      }

    // poloidal end point
    comp_->compute_loop2(1, p->mpsi_, 1, pd->mzeta_, 3);

    return HPX_SUCCESS;
  }

  bool
  gtc_app::gtc_point::evector_callback(std::size_t i,
      fake_array<double> const& evector)
  {
    comp_->compute_loop(1, field_array_->mgrid_, 1);
    return HPX_SUCCESS;
  }

  fake_array<double>
  gtc_app::gtc_point::get_evector(std::size_t depth, std::size_t extent)
  {
    fake_array<double> ret(depth * extent);
    return ret;
  }

  int
  gtc_app::gtc_point::pushi(size_t irk, std::size_t istep, size_t idiag)
  {
    SSTMAC_DEBUG << "gtc_point(" << idx_ << ")::pushi(" << irk << ", " << istep
          << ", " << idiag << ") \n";
    //short names
    field_array::ptr fa = field_array_;
    particle_decomp::ptr pd = part_decomp_;
    config_parameters::ptr p = params_;

    std::size_t limit_vpara = 0; // limit_vpara=1 :  parallel velocity kept <= abs(umax)
    std::size_t conserve_particles = 0;

    if (irk == 1)
      {
        // 1st step of Runge-Kutta method

        comp_->compute_loop2(1, fa->mi_, 1, 5, 1);

        comp_->compute_array_init<double>(p->mflux_ + 1);

        // 2nd step of Runge-Kutta method
      }
    else
      {

        if (p->nonlinear_ < 0.5)
          comp_->compute_array_init<double>(p->mflux_ + 1);

      }

    // gather e_field using 4-point gyro-averaging, sorting in poloidal angle
    comp_->compute_loop2(1, fa->mi_, 1, 4, 22);

    // primary ion marker temperature and parallel flow velocity

    comp_->compute_array_init<double>(p->mpsi_ + 1);
    comp_->compute_array_init<double>(p->mpsi_ + 1);
    comp_->compute_loop(0, p->mpsi_, 1);

    //!********** test gradual kappan **********
    //!     kappati=min(6.9_wp,(kappan+(6.9_wp-kappan)*real(istep,wp)/4000._wp))
    //!     if(mype==0.and.irk==2)write(36,*)istep,kappan,kappati
    //! ****************************************

    //! update GC position
    comp_->compute_loop(1, fa->mi_, 73);

    if (irk == 2)
      {
        if (limit_vpara == 1)
          {
            std::cerr << " Not a default parameter; not implemented "
                << std::endl;
          }
        //! out of boundary particle
        comp_->compute_loop(1, fa->mi_, 5);

        if (conserve_particles == 1)
          {
            std::cerr
                << " conserve_particles not a parameter activated by default.  Not implemented "
                << std::endl;
          }

        // Restore temperature profile when running a nonlinear calculation
        // (nonlinear=1.0) and parameter fixed_Tprofile > 0.

        if (p->nonlinear_ > 0.5 && p->fixed_Tprofile_ > 0)
          {
            if (istep % p->ndiag_ == 0)
              {
                comp_->compute_array_init<double>(p->mflux_ + 1);
                comp_->compute_array_init<double>(p->mflux_ + 1);
                comp_->compute_loop(1, fa->mi_, 6);

                comp_->compute_loop(1, fa->mi_, 4);

                // global sum of dtem broadcast to every toroidal PE
                  {
                    comp_->compute_array_init<double>(p->mpsi_ + 1);

                    typedef std::vector<hpx::lcos::promise<fake_array<double> >::ptr  > lazy_results_type;
                    lazy_results_type lazy_results;
                    for (size_t i = 0; i < point_components_.size(); i++)
                      {
                        lazy_results.push_back(
                            point_components_[i]->get_dtem_async());
                      }
                    ptr self(this);
                    hpx::lcos::wait<fake_array<double>, gtc_point>(lazy_results,
                        &gtc_point::dtem_callback, self);
                  }

                // global sum of dden broadcast to every toroidal PE
                  {
                    comp_->compute_array_init<double>(p->mpsi_ + 1);
                    typedef std::vector<hpx::lcos::promise<fake_array<double> >::ptr  > lazy_results_type;
                    lazy_results_type lazy_results;
                    for (size_t i = 0; i < point_components_.size(); i++)
                      {
                        lazy_results.push_back(
                            point_components_[i]->get_dden_async());
                      }
                    ptr self(this);
                    hpx::lcos::wait<fake_array<double>, gtc_point>(lazy_results,
                        &gtc_point::dden_callback, self);
                  }

                comp_->compute_loop(1, p->mpsi_, 1);
                comp_->compute_loop(1, p->mpsi_, 1);
                comp_->compute_loop(1, fa->mi_, 3);

              }
          }
      }

    if (idiag == 0)
      {
        // flux diagnosis at irk=1
        // Not implemented at this time
      }

    return HPX_SUCCESS;
  }

  bool
  gtc_app::gtc_point::dtem_callback(std::size_t i,
      fake_array<double> const& dtem)
  {
    comp_->compute_loop(1, dtem.size(), 1);
    return HPX_SUCCESS;
  }

  bool
  gtc_app::gtc_point::dden_callback(std::size_t i,
      fake_array<double> const& dden)
  {
    comp_->compute_loop(1, dden.size(), 1);
    return HPX_SUCCESS;
  }

  fake_array<double>
  gtc_app::gtc_point::get_dden()
  {
    fake_array<double> ret(params_->mflux_ + 1);
    return ret;
  }

  fake_array<double>
  gtc_app::gtc_point::get_dtem()
  {
    fake_array<double> ret(params_->mflux_ + 1);
    return ret;
  }

  int
  gtc_app::gtc_point::shifti()
  {
    SSTMAC_DEBUG << "gtc_point(" << idx_ << ")::shifti() \n";
    //short names
    field_array::ptr fa = field_array_;
    particle_decomp::ptr pd = part_decomp_;
    config_parameters::ptr p = params_;

    std::size_t m0 = 1;

    if (p->numberpe_ == 1)
      return HPX_SUCCESS;

    for (size_t iteration; iteration < p->numberpe_; iteration++)
      {

        comp_->compute_array_init<size_t>(3 * 2);

        if (m0 <= fa->mi_)
          {
            comp_->compute_array_init<double>(fa->mimax_);
          }
        comp_->compute_loop(m0, fa->mi_, 6);

        if (iteration > 1)
          {
            // global sum of msend broadcast to every toroidal PE
              {

                typedef std::vector<hpx::lcos::promise<std::size_t>::ptr> lazy_results_type;
                lazy_results_type lazy_results;
                for (size_t i = 0; i < point_components_.size(); i++)
                  {
                    lazy_results.push_back(
                        point_components_[i]->get_msend_async());
                  }
                hpx::lcos::wait<size_t>(lazy_results);
              }

          }

        // an extra space to prevent zero size when msendright(1)=msendleft(1)=0

        long msendleft = 0.1 * fa->mi_;
        long msendright = 0.1 * fa->mi_;
        long msend = 0.8 * fa->mi_;
        // pack particle to move right
        comp_->compute_loop(1, msendright, 2);

        // pack particle to move left
        comp_->compute_loop(1, msendleft, 2);

        std::size_t mtop = fa->mi_;
        // # of particles remain on local PE
        fa->mi_ = fa->mi_ - msendleft - msendright;
        // fill the hole
        comp_->compute_loop(1, msend, 2);

        // send # of particle to move right to neighboring PEs of same particle
          {
            typedef std::vector<hpx::lcos::promise<fake_array<std::size_t> >::ptr  > lazy_results_type;
            lazy_results_type lazy_results;
            lazy_results.push_back(
                point_components_[pd->left_pe_]->get_msendright_async());
            hpx::lcos::wait<fake_array<size_t> >(lazy_results);
          }

        // send particle to right and receive from left
          {

            typedef std::vector<hpx::lcos::promise<fake_array<double> >::ptr  > lazy_results_type;
            lazy_results_type lazy_results;
            lazy_results.push_back(
                point_components_[pd->left_pe_]->get_sendright_async());
            hpx::lcos::wait(lazy_results);
          }

        // send # of particle to move left
          {
            typedef std::vector<hpx::lcos::promise<fake_array<std::size_t> >::ptr  > lazy_results_type;
            lazy_results_type lazy_results;
            lazy_results.push_back(
                point_components_[pd->right_pe_]->get_msendleft_async());
            hpx::lcos::wait(lazy_results);
          }

        // send particle to left and receive from right
          {
            typedef std::vector<hpx::lcos::promise<fake_array<double> >::ptr  > lazy_results_type;
            lazy_results_type lazy_results;
            lazy_results.push_back(
                point_components_[pd->right_pe_]->get_sendleft_async());
            hpx::lcos::wait(lazy_results);
          }

        // tracer particle -- not implemented yet

        // need extra particle array
        if (fa->mi_ + msendleft + msendright > fa->mimax_)
          {
            std::cerr << " Need bigger particle array " << std::endl;
          }

        // unpack particle, particle moved from left
        comp_->compute_loop2(1, msendleft, 1, p->nparam_, 2);

        // particle moved from right
        comp_->compute_loop2(1, msendright, 1, p->nparam_, 2);

      }

    return HPX_SUCCESS;
  }

  std::size_t
  gtc_app::gtc_point::get_msend()
  {
    return 0.1 * field_array_->mi_;
  }

  fake_array<std::size_t>
  gtc_app::gtc_point::get_msendright()
  {
    fake_array<size_t> ret(3);
    return ret;
  }

  fake_array<double>
  gtc_app::gtc_point::get_sendright()
  {
    size_t nzion = 2 * params_->nparam_;
    fake_array<double> ret((nzion + 1) * (0.1 * field_array_->mi_));
    return ret;
  }

  fake_array<std::size_t>
  gtc_app::gtc_point::get_msendleft()
  {
    fake_array<size_t> ret(3);
    return ret;
  }

  fake_array<double>
  gtc_app::gtc_point::get_sendleft()
  {
    size_t nzion = 2 * params_->nparam_;
    fake_array<double> ret((nzion + 1) * (0.1 * field_array_->mi_));
    return ret;
  }

  int
  gtc_app::gtc_point::poisson(std::size_t iflag, std::size_t istep, size_t irk)
  {
    SSTMAC_DEBUG << "gtc_point(" << idx_ << ")::poisson(" << iflag << ", "
          << irk << ", " << istep << ") \n";
    //short names
    field_array::ptr fa = field_array_;
    particle_decomp::ptr pd = part_decomp_;
    config_parameters::ptr p = params_;

    // number of gyro-ring
    std::size_t mring = 2;

    // number of summation: maximum is 32*mring+1
    std::size_t mindex = 32 * mring + 1;

    // gamma=0.75: max. resolution for k=0.577
    double gamma = 0.75;

    std::size_t iteration = 5;

    // initialize poisson solver
    if (istep == 1 && irk == 1 && iflag == 0)
      {
        poisson_initial(mring, mindex);
      }

    std::size_t ipartd, izeta1, izeta2;
    if (p->npartdom_ > 1 && pd->mzeta_ % p->npartdom_ == 0)
      {
        std::cerr
            << " This version of GTC does not support npartdom > 1 at present: "
            << p->npartdom_ << std::endl;
        ipartd = 0;
        izeta1 = 1;
        izeta2 = pd->mzeta_;
      }
    else
      {
        ipartd = 0;
        izeta1 = 1;
        izeta2 = pd->mzeta_;
      }

    comp_->compute_loop2(izeta1, izeta2, 1, fa->mgrid_, 5);
    comp_->compute_loop2(izeta1, izeta2, 2, iteration, 1);
    comp_->compute_loop2(izeta1, izeta2, 0, p->mthetamax_, 2);

    if (ipartd == 1)
      {
        std::cerr << " Unsupported at this time. " << std::endl;
      }

    // in equilibrium unit
    comp_->compute_loop2(0, fa->mgrid_, 1, pd->mzeta_, 2);
    comp_->compute_loop2(0, p->mpsi_, 1, pd->mzeta_, 1);
    return HPX_SUCCESS;
  }

  int
  gtc_app::gtc_point::poisson_initial(std::size_t mring, std::size_t mindex)
  {
    SSTMAC_DEBUG << "gtc_point(" << idx_ << ")::poisson_initial(" << mring
          << ", " << mindex << ") \n";
    //short names
    field_array::ptr fa = field_array_;
    particle_decomp::ptr pd = part_decomp_;
    config_parameters::ptr p = params_;


    comp_->compute_loop(0, p->mpsi_, 1);

    // initialize
    comp_->compute_array_init<int>(
        (pd->mzeta_ + 1) * (fa->mgrid_ + 1) * (mindex + 1));
    comp_->compute_array_init<int>((pd->mzeta_ + 1) * (fa->mgrid_ + 1));
    comp_->compute_array_init<double>(
        (pd->mzeta_ + 1) * (fa->mgrid_ + 1) * (mindex + 1));

    comp_->compute_loop2(1, pd->mzeta_, 0, fa->mgrid_, 10);
    comp_->compute_loop4(1, pd->mzeta_, 0, fa->mgrid_, 1, mring, 1, 8,
        22 + 4 * 5);
    return HPX_SUCCESS;
  }

  std::string
  gtc_app::gtc_point::get_action_name(ACTION a)
  {
    switch (a)
      {
    case INIT:
      return "gtc_point::init";
      break;

    case LOAD:
      return "gtc_point::load";
      break;

    case CHARGEI:
      return "gtc_point::chargei";
      break;

    case GET_DENSITYI:
      return "gtc_point::get_densityi";
      break;

    case GET_ZONALI:
      return "gtc_point::get_zonali";
      break;

    case SMOOTH:
      return "gtc_point::smooth";
      break;

    case GET_PHI:
      return "gtc_point::get_phi";
      break;

    case GET_EACHZETA:
      return "gtc_point::get_eachzeta";
      break;

    case FIELD:
      return "gtc_point::field";
      break;

    case GET_EVECTOR:
      return "gtc_point::get_evector";
      break;

    case PUSHI:
      return "gtc_point::pushi";
      break;

    case GET_DDEN:
      return "gtc_point::get_dden";
      break;

    case GET_DTEM:
      return "gtc_point::get_dtem";
      break;

    case SHIFTI:
      return "gtc_point::shifti";
      break;

    case GET_MSEND:
      return "gtc_point::get_msend";
      break;

    case GET_MSENDRIGHT:
      return "gtc_point::get_msendright";
      break;

    case GET_SENDRIGHT:
      return "gtc_point::get_sendright";
      break;

    case GET_MSENDLEFT:
      return "gtc_point::get_msendleft";
      break;

    case GET_SENDLEFT:
      return "gtc_point::get_sendleft";
      break;

    case POISSON:
      return "gtc_point::poisson";
      break;

    default:
      throw sstmac::ssterror("gtc_point::get_action_name: unknown action");
      }
  }

  void
  gtc_app::gtc_point::register_actions(
      boost::intrusive_ptr<hpx::hpx_thread_manager> threadman)
  {

    threadman->register_action<init_action>(get_action_name(INIT),
        init_action::construct());

    threadman->register_action<load_action>(get_action_name(LOAD),
        load_action::construct());

    threadman->register_action<chargei_action>(get_action_name(CHARGEI),
        chargei_action::construct());

    threadman->register_action<get_densityi_action>(
        get_action_name(GET_DENSITYI), get_densityi_action::construct());

    threadman->register_action<get_zonali_action>(get_action_name(GET_ZONALI),
        get_zonali_action::construct());

    threadman->register_action<smooth_action>(get_action_name(SMOOTH),
        smooth_action::construct());

    threadman->register_action<get_phi_action>(get_action_name(GET_PHI),
        get_phi_action::construct());

    threadman->register_action<get_eachzeta_action>(
        get_action_name(GET_EACHZETA), get_eachzeta_action::construct());

    threadman->register_action<field_action>(get_action_name(FIELD),
        field_action::construct());

    threadman->register_action<get_evector_action>(get_action_name(GET_EVECTOR),
        get_evector_action::construct());

    threadman->register_action<pushi_action>(get_action_name(PUSHI),
        pushi_action::construct());

    threadman->register_action<get_dden_action>(get_action_name(GET_DDEN),
        get_dden_action::construct());

    threadman->register_action<get_dtem_action>(get_action_name(GET_DTEM),
        get_dtem_action::construct());

    threadman->register_action<shifti_action>(get_action_name(SHIFTI),
        shifti_action::construct());

    threadman->register_action<get_msend_action>(get_action_name(GET_MSEND),
        get_msend_action::construct());

    threadman->register_action<get_msendright_action>(
        get_action_name(GET_MSENDRIGHT), get_msendright_action::construct());

    threadman->register_action<get_sendright_action>(
        get_action_name(GET_SENDRIGHT), get_sendright_action::construct());

    threadman->register_action<get_msendleft_action>(
        get_action_name(GET_MSENDLEFT), get_msendleft_action::construct());

    threadman->register_action<get_sendleft_action>(
        get_action_name(GET_SENDLEFT), get_sendleft_action::construct());

    threadman->register_action<poisson_action>(get_action_name(POISSON),
        poisson_action::construct());

  }

}

#define DeclareHPXPayload(ty) \
    typedef sstmac::vector1payload<ty, hpx_gtc::fake_array<ty> > hpx_##ty##_vec_payload; \
    DeclareSerializable(hpx_##ty##_vec_payload)

DeclareHPXPayload(double);
DeclareHPXPayload(long);
DeclareHPXPayload(size_t);


