/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SKELETONS_PARALLEX_GTC_GTC_POINT_H_INCLUDED
#define SKELETONS_PARALLEX_GTC_GTC_POINT_H_INCLUDED

#include <gtc/gtc_app.h>
#include <gtc/gtc_particle_stubs.h>
#include <sstmac/software/api/hpx/components/hpx_component.h>
#include <sstmac/software/api/hpx/lcos/local_mutex.h>
#include <sstmac/software/api/hpx/lcos/hpx_callback.h>
#include <gtc/gtc_params.h>
#include <valarray>
#include "array1d.h"
#include "fake_array.h"
#include <sstmac/software/libraries/compute/lib_compute_loops.h>

//#include <sstmac/software/skeletons/mini/gtc_charge/lib_compute_gtc.h>

namespace hpx_gtc
{

  // this is just to optimize performance a little -- this number can be exceeded
#define MAX_NUM_NEIGHBORS 20

  class gtc_app::gtc_point : public hpx::components::hpx_component
  {
  public:
    class gtc_point_stubs;

    typedef boost::intrusive_ptr<gtc_point> ptr;
    typedef boost::intrusive_ptr<gtc_point_stubs> stub_ptr;
    typedef gtc_point_stubs stubs;

    /*class point_search_callback : public hpx::lcos::hpx_callback<bool>
     {

     public:
     virtual bool
     func(hpx::hpx_action_params::ptr params)
     {

     hpx::hpx_action_params3<size_t, double,
     hpx::components::hpx_component::ptr>::ptr pcast =
     boost::dynamic_pointer_cast<
     hpx::hpx_action_params3<size_t, double,
     hpx::components::hpx_component::ptr> >(params);

     if (!pcast)
     {
     throw sstmac::ssterror(
     "gtc_point::search_callback -- can't cast params");
     }

     gtc_app::gtc_point::ptr point = boost::dynamic_pointer_cast<
     gtc_app::gtc_point>(pcast->get_arg2());
     return point->search_callback(pcast->get_arg0(), pcast->get_arg1());

     }

     };*/

  protected:

    friend class search_callback;

    std::size_t idx_;
    particle_decomp::ptr part_decomp_;
    field_array::ptr field_array_;
    hpx::lcos::local_mutex::ptr mtx_;

    sstmac::sw::lib_compute_loops::ptr comp_;
    std::vector<stub_ptr> point_components_;
    config_parameters::ptr params_;


    int mgrid_;

    // -------------------------------- //
    gtc_point(hpx::globalid gid) :
        hpx_component(gid)
    {
      mtx_ = hpx::lcos::local_mutex::construct();

      std::stringstream ss;
      ss << "gtc_point-" << gid.gid_;

      comp_ = sstmac::sw::lib_compute_loops::construct(ss.str());
      register_lib(comp_);
    }

    // ---- these are the functions that actually do something,
    // ------- they are protected so that only actions can access them
    // ---------------- these are the same function signatures from point--server --------
    int
    init(std::size_t objectid, config_parameters::ptr par,  std::vector<stub_ptr> points);

    int
    load();

    int
    chargei(std::size_t istep);

    int
    smooth(std::size_t iflag, std::size_t idiag);


    bool
    chargei_zonali_callback(std::size_t i, fake_array<double> const& zonali);


    bool
    eachzeta_callback(std::size_t i, fake_array<double> const& eachzeta,
        std::size_t length);

    fake_array<double>
    get_densityi();

    fake_array<double>
    get_zonali();

    fake_array<double>
    get_phi(std::size_t depth);

    fake_array<double>
    get_eachzeta();

    int
    field();

    bool
    evector_callback(std::size_t i, fake_array<double> const& evector);

    fake_array<double>
    get_evector(std::size_t depth, std::size_t extent);

    int
    pushi(size_t irk, std::size_t istep, size_t idiag);

    bool
    dtem_callback(std::size_t i, fake_array<double> const& dtem);

    bool
    dden_callback(std::size_t i, fake_array<double> const& dden);

    fake_array<double>
    get_dden();

    fake_array<double>
    get_dtem();

    int
    shifti();


    std::size_t
    get_msend();



    fake_array<std::size_t>
    get_msendright();



    fake_array<double>
    get_sendright();

    fake_array<std::size_t>
    get_msendleft();



    fake_array<double>
    get_sendleft();

    int
    poisson(std::size_t iflag, std::size_t istep, size_t irk);

    int
    poisson_initial(std::size_t mring, std::size_t mindex);

  public:

    static ptr
    construct(hpx::globalid gid)
    {
      return ptr(new gtc_point(gid));
    }

    virtual
    ~gtc_point()
    {
    }

    static std::string
    component_name()
    {
      return "gtc_app::gtc_point";
    }

    virtual void
    register_actions(boost::intrusive_ptr<hpx::hpx_thread_manager> threadman);

    virtual hpx::components::hpx_component::ptr
    clone(hpx::globalid gid)
    {
      return construct(gid);
    }

    enum ACTION
    {
      INIT,
      LOAD,
      CHARGEI,
      GET_DENSITYI,
      GET_ZONALI,
      SMOOTH,
      GET_PHI,
      GET_EACHZETA,
      FIELD,
      GET_EVECTOR,
      PUSHI,
      GET_DDEN,
      GET_DTEM,
      SHIFTI,
      GET_MSEND,
      GET_MSENDRIGHT,
      GET_SENDRIGHT,
      GET_MSENDLEFT,
      GET_SENDLEFT,
      POISSON
    };

    static std::string
    get_action_name(ACTION a);

    typedef hpx::actions::component_result_action3<int, gtc_app::gtc_point,
        std::size_t, config_parameters::ptr, std::vector<gtc_point::stub_ptr>, // argument
        &gtc_app::gtc_point::init // function
    > init_action;
    typedef hpx::hpx_action_params3<size_t, config_parameters::ptr, std::vector<gtc_point::stub_ptr> > init_paramtype;

    typedef hpx::actions::component_result_action0<int, gtc_app::gtc_point,
        &gtc_app::gtc_point::load // function
    > load_action;
    typedef hpx::hpx_action_params load_paramtype;

    typedef hpx::actions::component_result_action1<int, gtc_app::gtc_point,
        std::size_t, // argument
        &gtc_app::gtc_point::chargei // function
    > chargei_action;
    typedef hpx::hpx_action_params1<size_t> chargei_paramtype;

    typedef hpx::actions::component_result_action2<int, gtc_app::gtc_point,
        std::size_t,  std::size_t, // argument
        &gtc_app::gtc_point::smooth // function
    > smooth_action;
    typedef hpx::hpx_action_params2<size_t,
        size_t> smooth_paramtype;

    typedef hpx::actions::component_result_action0<fake_array<double>,
        gtc_app::gtc_point, &gtc_app::gtc_point::get_densityi, // function
        hpx::parcels::vector1_payloadizer<double, fake_array<double> > > get_densityi_action;
    typedef hpx::hpx_action_params get_densityi_paramtype;

    typedef hpx::actions::component_result_action0<fake_array<double>,
        gtc_app::gtc_point, &gtc_app::gtc_point::get_zonali, // function
        hpx::parcels::vector1_payloadizer<double, fake_array<double> > > get_zonali_action;
    typedef hpx::hpx_action_params get_zonali_paramtype;

    typedef hpx::actions::component_result_action1<fake_array<double>,
        gtc_app::gtc_point, std::size_t, &gtc_app::gtc_point::get_phi, // function
        hpx::parcels::vector1_payloadizer<double, fake_array<double> > > get_phi_action;
    typedef hpx::hpx_action_params1<size_t> get_phi_paramtype;

    typedef hpx::actions::component_result_action0<fake_array<double>,
        gtc_app::gtc_point, &gtc_app::gtc_point::get_eachzeta, // function
        hpx::parcels::vector1_payloadizer<double, fake_array<double> > > get_eachzeta_action;
    typedef hpx::hpx_action_params get_eachzeta_paramtype;

    typedef hpx::actions::component_result_action0<int, gtc_app::gtc_point,
        &gtc_app::gtc_point::field // function
    > field_action;
    typedef hpx::hpx_action_params field_paramtype;

    typedef hpx::actions::component_result_action2<fake_array<double>,
        gtc_app::gtc_point, std::size_t, std::size_t,
        &gtc_app::gtc_point::get_evector, // function
        hpx::parcels::vector1_payloadizer<double, fake_array<double> > > get_evector_action;
    typedef hpx::hpx_action_params2<size_t, size_t> get_evector_paramtype;

    typedef hpx::actions::component_result_action3<int, gtc_app::gtc_point,
        size_t, size_t, size_t, // argument
        &gtc_app::gtc_point::pushi // function
    > pushi_action;
    typedef hpx::hpx_action_params3<size_t, size_t, size_t> pushi_paramtype;

    typedef hpx::actions::component_result_action0<fake_array<double>,
        gtc_app::gtc_point, &gtc_app::gtc_point::get_dden, // function
        hpx::parcels::vector1_payloadizer<double, fake_array<double> > > get_dden_action;
    typedef hpx::hpx_action_params get_dden_paramtype;

    typedef hpx::actions::component_result_action0<fake_array<double>,
        gtc_app::gtc_point, &gtc_app::gtc_point::get_dtem, // function
        hpx::parcels::vector1_payloadizer<double, fake_array<double> > > get_dtem_action;
    typedef hpx::hpx_action_params get_dtem_paramtype;

    typedef hpx::actions::component_result_action0<int, gtc_app::gtc_point,
        &gtc_app::gtc_point::shifti // function
    > shifti_action;
    typedef hpx::hpx_action_params shifti_paramtype;

    typedef hpx::actions::component_result_action0<size_t, gtc_app::gtc_point,
        &gtc_app::gtc_point::get_msend // function
    > get_msend_action;
    typedef hpx::hpx_action_params get_msend_paramtype;

    typedef hpx::actions::component_result_action0<fake_array<size_t>,
        gtc_app::gtc_point, &gtc_app::gtc_point::get_msendright, // function
        hpx::parcels::vector1_payloadizer<size_t, fake_array<size_t> > > get_msendright_action;
    typedef hpx::hpx_action_params get_msendright_paramtype;

    typedef hpx::actions::component_result_action0<fake_array<double>,
        gtc_app::gtc_point, &gtc_app::gtc_point::get_sendright, // function
        hpx::parcels::vector1_payloadizer<double, fake_array<double> > > get_sendright_action;
    typedef hpx::hpx_action_params get_sendright_paramtype;

    typedef hpx::actions::component_result_action0<fake_array<size_t>,
        gtc_app::gtc_point, &gtc_app::gtc_point::get_msendleft, // function
        hpx::parcels::vector1_payloadizer<size_t, fake_array<size_t> > > get_msendleft_action;
    typedef hpx::hpx_action_params get_msendleft_paramtype;

    typedef hpx::actions::component_result_action0<fake_array<double>,
        gtc_app::gtc_point, &gtc_app::gtc_point::get_sendleft, // function
        hpx::parcels::vector1_payloadizer<double, fake_array<double> > > get_sendleft_action;
    typedef hpx::hpx_action_params get_sendleft_paramtype;

    typedef hpx::actions::component_result_action3<int, gtc_app::gtc_point,
        size_t, size_t, size_t, // argument
        &gtc_app::gtc_point::poisson // function
    > poisson_action;
    typedef hpx::hpx_action_params3<size_t, size_t, size_t> poisson_paramtype;

    typedef hpx::actions::component_result_action2<int, gtc_app::gtc_point,
        size_t, size_t, // argument
        &gtc_app::gtc_point::poisson_initial // function
    > poisson_initial_action;
    typedef hpx::hpx_action_params3<size_t, size_t, config_parameters::ptr> poisson_initial_paramtype;

  };

}
#endif /* HPX_RMA_ELEMENT_H_ */
