/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <rma/hpx_random_mem_access.h>
#include <rma/hpx_rma_element.h>
#include <rma/hpx_rma_element_stubs.h>
#include <sstmac/software/api/hpx/components/hpx_distributing_factory.h>
#include <sstmac/software/api/hpx/lcos/promise.h>
#include <sstmac/software/api/hpx/lcos/future_wait.h>

using namespace sstmac;
using namespace sstmac::sw;

namespace hpx
{
    namespace rma 
    {

    SSTRegisterApp("hpx_rma", hpx_random_mem_access);

    void
    hpx_random_mem_access::consume_params(const sstmac::sim_parameters::ptr &params)
    {
      array_size_ = params->get_optional_int_param("hpx_rma_array_size",
          4);
      iterations_ = params->get_optional_int_param("hpx_rma_iterations",
          3);
      seed_ = params->get_optional_int_param("hpx_rma_seed", 0);

      // If the specified seed is 0, then we pick a random seed.
      if (!seed_)
        seed_ = std::size_t(std::time(0));

      rng_ = RNG::SHR3::construct(seed_);

    }

    void
    hpx_random_mem_access::register_actions()
    {

    }

    void
    hpx_random_mem_access::register_components()
    {
      hpx::globalid dumm;
      threadman_->register_component<hpx_rma_element>(hpx_rma_element::component_name(), hpx_rma_element::construct(dumm));
      threadman_->register_component<hpx::components::hpx_distributing_factory>(hpx::components::hpx_distributing_factory::component_name(),
          hpx::components::hpx_distributing_factory::construct(dumm));
    }

    void
    hpx_random_mem_access::hpx_main()
    {

      SSTMAC_DEBUG << " Parameters: " << "\n";
      SSTMAC_DEBUG << " --------------: " << "\n";
      SSTMAC_DEBUG << " array_size  : " << array_size_ << "\n";
      SSTMAC_DEBUG << " iterations  : " << iterations_ << "\n";
      SSTMAC_DEBUG << " seed        : " << seed_ << "\n";
      SSTMAC_DEBUG << " --------------: " << "\n";

      // Create a distributing factory. This object can be used to create
      // a distributed array of components.
      hpx::components::hpx_distributing_factory::ptr factory =
          hpx::components::hpx_distributing_factory::construct(
              agas_->new_gid(hpx::lva::COMPONENT));

      // Get the global component type of our element component.
      // hpx::components::component_type type =
      //     hpx::components::get_component_type<random_mem_access::server::element>();

      // Create a global array of element components via distributing factory.
      // The elements will be divided up equally among the localities that
      // supports our element component will
      std::vector<hpx_rma_element::stub_ptr> array = factory->create_components<hpx_rma_element::stubs>(
          hpx_rma_element::component_name(), array_size_);

      SSTMAC_DEBUG << "app: starting initialization phase \n";

      // Initialize the value of each element with its index in the array.
      for (std::size_t i = 0; i < array_size_; ++i)
        {
          array[i]->init(i);
        }

      ///////////////////////////////////////////////////////////////////////
      // Access phase - increment elements in the array at random.
      std::vector<hpx::lcos::promise<int>::ptr > access_phase;

      SSTMAC_DEBUG << "app: starting access phase \n";

      // Spawn a future for each remote
      for (std::size_t i = 0; i < iterations_; ++i)
        {
          SSTMAC_DEBUG << "app: starting access " << i << "\n";
          // Compute the element to access.
          std::size_t const rn = rng_->value() % array_size_;
          access_phase.push_back(array[rn]->add_async());
        }

      SSTMAC_DEBUG << "app: waiting for accesses to complete \n";
      // Wait for the access operations to complete.
      hpx::lcos::wait(access_phase);

      ///////////////////////////////////////////////////////////////////////
      // Print phase - print each element in the array.
      std::vector<hpx::lcos::promise<int>::ptr > print_phase;

      SSTMAC_DEBUG << "app: starting printing phase \n";

      for (std::size_t i = 0; i < array_size_; ++i)
        {
          // Start an asynchronous print action, which will be executed on
          // the locality where each element lives. The I/O, however, will
          // all go to standard output on the console locality.
          print_phase.push_back(array[i]->print_async());
        }

      // Wait for all print operations to finish.
      hpx::lcos::wait(print_phase);


      SSTMAC_DEBUG << "app: finished ! \n";
    }

} }
