/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/backends/native/eventheap.h>

#include <sstmac/common/debug.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/eventmodule.h>

namespace sstmac
{
  namespace native
  {

    static uint64_t num_events = 0;

    stat_loghistogram<int64_t>::ptr eventheap::event_hist_ = 0;

    // Pointer types.

    //
    // Hello.
    //
    eventheap::eventheap(const timestamp &tv) :
      running_(false)
    {
      set_now(tv);
      last_update_ = time(NULL);
      update_interval_ = 2;
      sim_start_ = last_update_;
      last_update_sim_ = timestamp(tv);
      complete_ = false;

      if (!event_hist_)
      {
        event_hist_ = stat_loghistogram<int64_t>::construct("<trace> events",
            "eventhistogram.dat");
        logger::register_logger(event_hist_.get());
        register_stat(event_hist_);
      }

    }

    //
    // Hello.
    //
    eventheap::ptr
    eventheap::construct(const timestamp &tv)
    {
      eventheap::ptr retval(new eventheap(tv));
      return retval;
    }

    //
    // Goodbye.
    //
    eventheap::~eventheap() throw ()
    {
      if (running_)
      {
        SSTMAC_DEBUG
              << "Warning: eventheap (eventmanager) exiting while running\n.";
      }
    }

    //
    // Clear all events and set time back to a zero of your choice.
    //
    void
    eventheap::clear(const timestamp &zero_time)
    {
      if (running_)
      {
        throw illformederror("eventheap::clear:  Event manager is running.");
      }
      while(!queue_.empty())
        queue_.pop();
      set_now(zero_time);
    }

    //
    // Run the eventmanager.
    //
    void
    eventheap::run()
    {
      if (running_)
      {
        throw illformederror("eventheap::run:  Event manager already running.");
      }
      running_ = true;
      stopped_ = false;
      while (!queue_.empty() && !stopped_)
      {
        do_one_event();
      }

      running_ = false;

      if (queue_.empty() || finish_on_stop_)
      {
        complete_ = true;
        SSTMAC_DEBUG << "eventheap[" << this << "]::run:  All out of events.\n";

        finish();
      }
    }

    void
    eventheap::cancel_all_messages(const eventmodule::ptr &module)
    {
      SSTMAC_DEBUG << "eventheap: cancelling all messages to "
            << module->to_string() << "\n";
      queue_t temp;
      //eventhandler::ptr hand = handler;

      while (!queue_.empty())
      {
        event::ptr t = queue_.top();
        if (t->get_module() == module)
        {

          SSTMAC_DEBUG << "eventheap: cancelling message " << t->to_string()
                << "\n";
          t->cancel();

        }
        queue_.pop();
        temp.push(t);
      }

      while (!temp.empty())
      {
        queue_.push(temp.top());
        temp.pop();
      }

    }

    void
    eventheap::step()
    {
      if (complete_)
      {
        std::cout
            << "manager: step: eventmanager is complete, need to restart. \n";
        return;
      }
      else
      {
        do_one_event();
      }

      if (queue_.empty())
      {
        complete_ = true;
        SSTMAC_DEBUG << "eventheap[" << this << "]::run:  All out of events.\n";

        finish();

      }

    }

    void
    eventheap::do_one_event()
    {

      event::ptr ev = queue_.top();
      if (!ev->is_cancelled())
        set_now(ev->time());
      queue_.pop();
      SSTMAC_DEBUG << "eventheap[" << this << "]::run:  Processing event " << ev
            << " of type " << ev->to_string() << " at time " << now() << "\n";
      ev->execute();
      SSTMAC_DEBUG << "eventheap[" << this << "]::run:  Event " << ev
            << " complete, " << queue_.size() << " events pending\n";



      if (updates_.is_active())
      {
        time_t t = time(NULL);
        if (t >= last_update_ + update_interval_)
        {
          double simsecpersec = (now() - last_update_sim_).sec() / (t
              - last_update_);
          updates_ << "elapsed time: " << (t - sim_start_)
              << ", simulated time: " << now() << ", simsec/sec: "
              << simsecpersec << "\n";
          last_update_ = t;
          last_update_sim_ = now();
        }
      }
      ++num_events;
    }

    void
    eventheap::finish()
    {
      SSTMAC_DEBUG << "reporting for " << finishers_.size() << " modules \n";
      std::list<ptr_type::ptr>::iterator it, end = finishers_.end();
      for (it = finishers_.begin(); it != end; it++)
      {
        eventmodule::ptr mod = boost::dynamic_pointer_cast<eventmodule>(*it);
        mod->finish();
      }
      finish_stats();
      finishers_.clear();
      //std::cout << sstprintf("SST/macro ran a total of %lu events\n", num_events);
    }

    //
    // Set off the given eventhandler at the given time.
    //
    void
    eventheap::schedule(const timestamp& start_time, const event::ptr& ev)
    {
      //timestamp delta_t = start_time - now();
      //if (delta_t.sec() > 1.3){
      //    abort();
      //}

      if (event_hist_->is_active())
      {
        timestamp t = start_time - now_;
        event_hist_->add_one(t.ticks_int64());
      }

      ev->set_time(start_time);

      queue_.push( ev);
      SSTMAC_DEBUG << "eventheap[" << this << "]::schedule:  Scheduling event "
            << ev << " which is " << ev->to_string() << start_time << ", "
            << queue_.size() << " events pending.\n";
      ++nevents_;
    }

    void
    eventheap::register_finish(ptr_type::ptr p)
    {
      finishers_.push_back(p);
    }

  }
} // end of namespace sstmac
