/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_BACKENDS_NATIVE_MANAGER_H_INCLUDED
#define SSTMAC_BACKENDS_NATIVE_MANAGER_H_INCLUDED

#include <sstmac/common/timestamp.h>
#include <sstmac/software/process/appid.h>
#include <sstmac/backends/native/nodeid.h>
#include <sstmac/common/nocopy.h>
#include <sstmac/common/debug.h>
#include <sstmac/common/eventhandler.h>
#include <sstmac/common/ptr_type.h>

#include <sstmac/software/services/launch/allocation/allocationstrategy.h>
#include <sstmac/software/services/launch/indexing/indexstrategy.h>
#include <sstmac/hardware/network/interconnect/interconnect.h>
//#include <sstmac/hardware/network/interconnect/dist_traininterconnect.h>
#include <sstmac/common/eventmanager.h>
#include <sstmac/software/services/launch/launcher.h>
#include <sstmac/hardware/node/node.h>


namespace sstmac  { namespace native
  {

  // Forward declaration of relevant types.

  class networkinterface;

  /**
   * The owner of a discrete-event network simulation.
   *
   * All time progression is handled by this object
   * and messages between nodes are managed here as well.
   */
  class manager: virtual public nocopy, public ptr_type
  {
  public:
    virtual std::string to_string() const{
      return "manager";
    }

    typedef boost::intrusive_ptr<manager> ptr;
  protected:
    /// The nested specialization of the environment type.
    /// One of these gets handed to each launcher started under this manager.
    // class env;

    /// The event manager.
    eventmanager::ptr eventmanager_;

    /// The network simulator.
    sstmac::hw::interconnect::ptr interconnect_;

    /// Monitor whether the simulator is currently running.
    bool running_;

    /// Next parallel process id.
    sstmac::sw::appid next_ppid_;

    debug dbg_;

    /// The processes to be fired up at startup.
    typedef eventhandler::ptr event_t;
    typedef std::vector<std::pair<timestamp, event_t> > startup_vec_t;
    startup_vec_t startup_;

  protected:
    /// Default constructor.
    /// By default the manager has no nodes or network; it can only
    /// be run after one or more nodes have been added.
    manager(const eventmanager::ptr& ev_mgr);

  public:
    static manager::ptr global;

    /// Construct a new manager.
    static manager::ptr
    construct(const eventmanager::ptr& ev_mgr);

    /// Goodbye.
    virtual
    ~manager() throw ();

    eventmanager::ptr
    get_eventmanager(){
      return eventmanager_;
    }

    virtual void finish();

    /// Define a interconnect for this manager.
    void
    build_interconnect(const sim_parameters::ptr& params);

    void
    init_failure_model(const sim_parameters::ptr& params);

    /// Access the simulator used by this manager.
    sstmac::hw::interconnect::ptr
    get_interconnect() const;

    /// Set a pre-built interconnect for this manager. This is used for parallel simulations.
    void
    set_interconnect(const sstmac::hw::interconnect::ptr& fab);

    /// Access the number of nodes added to date.
    /// This can be any value <= get_simulator()->size().
    int64_t
    nodes() const;

    /// Get the node corresponding to nodeid.
    /// \param id is the nodeid of the desired node.
    /// \return the node object correspondint to nodeid.
    sstmac::hw::node::ptr get_node(const nodeid::ptr& id) const;

    // NOTES:
    //     What should we do about launchers added after the simulation
    //     has started with a start-time in the future (when more or fewer
    //     nodes may be available)?  How much of a "queueing system"
    //     do we want to work in here?

    // TODO:
    // 1)  Add a way to specify which nodes get used for a given job
    //     (to facilitate simulating non-uniform node/network topologies).
    // 2)  Make it easier to name (or label) nodes, either individually by
    //     class (same reasoning as above).
    // 3)  Decide on a more final strategy for handling tiered topologies
    //     such as systems with GPGPUs or accelerators (e.g. RoadRunner).

    /// Setup the simulation.
    /// Initializes eventmanager, interconnect, nodes, and scheduler.
    void setup();

    /// Run the simulation.
    /// \return the time at the end of the simulation.
    /// \throw illformederror if the simulation cannot be run
    /// (e.g. if no nodes have been added).
    timestamp run(timestamp until = timestamp(-1));

    /// End the simulation.
    /// Does error checking and prints statistics.
    void done();

    class stophandler : public eventhandler {
     public:
        typedef boost::intrusive_ptr<stophandler> ptr;

      protected:
        stophandler()
        {
        }

      public:
        virtual std::string to_string() const{
          return "stophandler";
        }

        virtual ~stophandler()
        {
        }

        virtual void handle(const sst_message::ptr &msg)
        {
          std::cout << "--- STOP event -----" << std::endl;
          eventmanager::stopped_ = true;
        }

        static ptr
        construct(){
          return ptr(new stophandler());
        }
     };
  };
  } } // end of namespace sstmac.

#endif
