/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_BACKENDS_NATIVE_NODEID_H_INCLUDED
#define SSTMAC_BACKENDS_NATIVE_NODEID_H_INCLUDED

#include <sstmac/common/ptr_type.h>
#include <sstmac/common/nodeaddress.h>
#include <sstmac/common/logger.h>
#include <iostream>
#include <stdint.h>

namespace sstmac { namespace native
  {

  /**
   * A ligthweight wrapper for a node index.
   * Done to avoid annoying mistakes where index values are mixed up.
   *
   * I can't wait for C++ to add opaque typedefs.
   */
  class nodeid:
    public nodeaddress,
    public serializable_type<nodeid>
  {
  ImplementSerializable(nodeid);

  private:
    long id_;

  protected:
    nodeid() :
    	id_(-1)
    {
        id_to_node_map_[id_] = nodeaddress::ptr(this);
    }

    nodeid(long i) :
    	id_(i)
    {
        id_to_node_map_[id_] = nodeaddress::ptr(this);
    }



    friend class native_address_converter;

    class native_address_converter : 
        public nodeaddress::address_converter
    {

      native_address_converter()
                    {
                    }

    public:
      virtual nodeaddress::ptr
      convert(long uid)
      {
       return nodeid::construct(uid);
      }

      virtual
      ~native_address_converter()
      {
      }

      typedef boost::intrusive_ptr<native_address_converter> ptr;

      static native_address_converter::ptr
      construct()
      {
        return ptr(new native_address_converter());
      }
    };


    public:

    typedef boost::intrusive_ptr<nodeid> ptr;

    static void init_statics()
    {
        converter_ = native_address_converter::construct();
    }

    static void clear_statics()
    {
        converter_ = 0;
    }

    static nodeid::ptr
    construct(long i){
      nodeid::ptr ret(new nodeid(i));
      return ret;
    }

    virtual long unique_id() const;



    virtual void
    serialize_order(sst_serializer* ser);



    virtual
    ~nodeid();

    virtual std::string to_string() const;

    virtual bool equals(const nodeaddress::ptr &p) const;

    long id() const;

    void id(long i);

  };

  inline bool operator==(const nodeid::ptr &a, const nodeid::ptr &b)
        {
    return (a->id() == b->id());
        }
  inline bool operator!=(const nodeid::ptr &a, const nodeid::ptr &b)
        {
    return (a->id() != b->id());
        }
  inline bool operator<(const nodeid::ptr &a, const nodeid::ptr &b)
  {
    return (a->id() < b->id());
  }
  inline bool operator<=(const nodeid::ptr &a, const nodeid::ptr &b)
        {
    return (a->id() <= b->id());
    }
  inline bool operator>(const nodeid::ptr &a, const nodeid::ptr &b)
  {
    return (a->id() > b->id());
  }
  inline bool operator>=(const nodeid::ptr &a, const nodeid::ptr &b)
    {
    return (a->id() >= b->id());
    }
  inline std::ostream&
  operator<<(std::ostream &os, const nodeid::ptr &n)
  {
    return (os << n->to_string());
  }

  inline logger&
  operator<<(logger &os, const nodeid::ptr &n)
  {
    return (os << n->to_string());
  }

  } } // end of namespace sstmac



#endif
