#ifndef sstmac_common_allocators_h
#define sstmac_common_allocators_h

#include <sstmac/common/sstnew.h>
#include <memory>
#include <limits>

namespace sstmac {

template <class T>
class basic_allocator {
  public:
    // type definitions
    typedef T        value_type;
    typedef T*       pointer;
    typedef const T* const_pointer;
    typedef T&       reference;
    typedef const T& const_reference;
    typedef std::size_t    size_type;
    typedef std::ptrdiff_t difference_type;

    // rebind allocator to type U
    template <class U>
    struct rebind {
        typedef basic_allocator<U> other;
    };

    // return address of values
    pointer address (reference value) const {
        return &value;
    }
    const_pointer address (const_reference value) const {
        return &value;
    }

    /* constructors and destructor
     * - nothing to do because the allocator has no state
     */
    basic_allocator() throw() {
    }

    basic_allocator(const basic_allocator&) throw() {
    }
    template <class U>
      basic_allocator (const basic_allocator<U>&) throw() {
    }
    ~basic_allocator() throw() {
    }

    // return maximum number of elements that can be allocated
    size_type max_size () const throw() {
        return std::numeric_limits<std::size_t>::max() / sizeof(T);
    }

    // allocate but don't initialize num elements of type T
    pointer allocate (size_type num, const void* = 0) {
        // print message and allocate memory with global new
        pointer ret = (pointer)(::operator new(num*sizeof(T)));
        return ret;
    }

    // initialize elements of allocated storage p with value value
    void construct (pointer p, const T& value) {
        // initialize memory with placement new
        new((void*)p)T(value);
    }

    // destroy elements of initialized storage p
    void destroy (pointer p) {
        // destroy objects by calling their destructor
        p->~T();
    }

    // deallocate storage p of deleted elements
    void deallocate (pointer p, size_type num) {
        ::operator delete((void*)p);
    }
};

// return that all specializations of this allocator are interchangeable
template <class T1, class T2>
bool operator== (const basic_allocator<T1>&,
                 const basic_allocator<T2>&) throw() {
    return true;
}
template <class T1, class T2>
bool operator!= (const basic_allocator<T1>&,
                 const basic_allocator<T2>&) throw() {
    return false;
}


}

#endif
