/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_COMMON_EVENTCALLBACK_H_INCLUDED
#define SSTMAC_COMMON_EVENTCALLBACK_H_INCLUDED

#include <sstmac/common/eventhandler.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/sstevent.h>

namespace sstmac
{

  /**
   * The main interface for something that can respond to an event (sst_message).
   */
  class eventcallback : public eventhandler
  {
  public:
    typedef boost::intrusive_ptr<eventcallback> ptr;

    timestamp time_;

  protected:
    eventcallback();

  public:
    virtual ~eventcallback();

    virtual void
    handle(const sst_message::ptr& msg);

    virtual void
    callback(const sst_message::ptr& msg) = 0;

    virtual std::string to_string() const;

    timestamp get_time() const;

    void set_time(const timestamp& t);

  };

  class donothingcallback : public eventcallback
  {
  public:
    typedef boost::intrusive_ptr<donothingcallback> ptr;

  public:
    static ptr construct();

    void callback(const sst_message::ptr& msg);

    virtual std::string to_string() const;
  };

  /** This throws an exception because messages with callbacks should never be handled */
  class callbackhandler : public eventhandler
  {
    protected:
     callbackhandler();

     static debug dbg_;

    public:
      typedef boost::intrusive_ptr<callbackhandler> ptr;

      static ptr construct();

      std::string to_string() const;

      void handle(const sst_message::ptr &msg);
  };

  template<class Cls, typename Fxn>
    class eventcallback_msg : public eventcallback
    {
    private:
      boost::intrusive_ptr<Cls> obj_;

      Fxn fxn_;

    public:
      eventcallback_msg(const boost::intrusive_ptr<Cls>& obj, Fxn fxn) :
        obj_(obj), fxn_(fxn)
      {
      }

      void
      callback(const sst_message::ptr& msg)
      {
        (obj_.get()->*fxn_)(msg);
      }
    };

    template<class Cls, typename Fxn, typename Arg1>
        class eventcallback_msg_1_args : public eventcallback
        {
        private:
          boost::intrusive_ptr<Cls> obj_;

          Fxn fxn_;

          Arg1 arg1_;


        public:
          eventcallback_msg_1_args(const boost::intrusive_ptr<Cls>& obj, Fxn fxn,
              const Arg1& arg1) :
            obj_(obj), fxn_(fxn), arg1_(arg1)
          {
          }

          void
          callback(const sst_message::ptr& msg)
          {
            Cls& obj = *obj_;
            (obj_.get()->*fxn_)(msg, arg1_);
          }
        };

  template<class Cls, typename Fxn, typename Arg1, typename Arg2>
    class eventcallback_msg_2_args : public eventcallback
    {
    private:
      boost::intrusive_ptr<Cls> obj_;

      Fxn fxn_;

      Arg1 arg1_;

      Arg2 arg2_;

    public:
      eventcallback_msg_2_args(const boost::intrusive_ptr<Cls>& obj, Fxn fxn,
          const Arg1& arg1, const Arg2& arg2) :
        obj_(obj), fxn_(fxn), arg1_(arg1), arg2_(arg2)
      {
      }

      void
      callback(const sst_message::ptr& msg)
      {
        Cls& obj = *obj_;
        (obj_.get()->*fxn_)(msg, arg1_, arg2_);
      }
    };

  template<class Cls, typename Fxn>
    eventcallback::ptr
    msg_callback(const boost::intrusive_ptr<Cls>& cls, Fxn fxn)
    {
      eventcallback::ptr callback = new eventcallback_msg<Cls, Fxn> (cls, fxn);
      return callback;
    }

  template<class Cls, typename Fxn, typename Arg1>
     eventcallback::ptr
     msg_callback(const boost::intrusive_ptr<Cls>& cls, Fxn fxn,
         const Arg1& arg1)
     {
       eventcallback::ptr callback = new eventcallback_msg_1_args<Cls, Fxn,
           Arg1> (cls, fxn, arg1);
       return callback;
     }

  template<class Cls, typename Fxn, typename Arg1, typename Arg2>
    eventcallback::ptr
    msg_callback(const boost::intrusive_ptr<Cls>& cls, Fxn fxn,
        const Arg1& arg1, const Arg2& arg2)
    {
      eventcallback::ptr callback = new eventcallback_msg_2_args<Cls, Fxn,
          Arg1, Arg2> (cls, fxn, arg1, arg2);
      return callback;
    }


    template <class Cls, class Fxn>
    class generic_event_0_args :
         public generic_event
    {
     private:
        boost::intrusive_ptr<Cls> obj_;

        Fxn fxn_;

    public:
        generic_event_0_args(const boost::intrusive_ptr<Cls>& obj, Fxn fxn) :
            obj_(obj), fxn_(fxn)
        {
        }

        void cancel(){
            obj_ = 0;
            fxn_ = 0;
        }

        virtual bool
      is_cancelled() const
      {
        return !obj_ && !fxn_;
      }

        void execute(){
            if (obj_)
                (obj_.get()->*fxn_)();
        }

        std::string
        to_string() const
        {
            return "generic event with 0 args";
        }
    };

    template <class Cls, class Fxn, class Arg1>
    class generic_event_1_args :
         public generic_event
    {
     private:
        boost::intrusive_ptr<Cls> obj_;

        Fxn fxn_;

        Arg1 arg1_;

    public:
        generic_event_1_args(const boost::intrusive_ptr<Cls>& obj, Fxn fxn, const Arg1& arg1) :
            obj_(obj), fxn_(fxn), arg1_(arg1)
        {
        }

        void cancel(){
            obj_ = 0;
            fxn_ = 0;
        }

        virtual bool
      is_cancelled() const
      {
        return !obj_ && !fxn_;
      }

        void execute(){
            if (obj_)
                (obj_.get()->*fxn_)(arg1_);
        }

        std::string
        to_string() const
        {
            return "generic event with 1 arg";
        }

    };

    template <class Cls, class Fxn, class Arg1, class Arg2>
    class generic_event_2_args :
         public generic_event
    {
     private:
        boost::intrusive_ptr<Cls> obj_;

        Fxn fxn_;

        Arg1 arg1_;

        Arg2 arg2_;

    public:
        generic_event_2_args(const boost::intrusive_ptr<Cls>& obj, Fxn fxn,
                             const Arg1& arg1, const Arg2& arg2) :
            obj_(obj), fxn_(fxn), arg1_(arg1), arg2_(arg2)
        {
        }

        std::string
        to_string() const
        {
            return "generic event with 2 args";
        }

        void cancel(){
            obj_ = 0;
            fxn_ = 0;
        }

        virtual bool
      is_cancelled() const
      {
        return !obj_ && !fxn_;
      }

        void execute(){
            if (obj_)
                (obj_.get()->*fxn_)(arg1_, arg2_);
        }


    };

    template<class Cls, typename Fxn>
    event::ptr
    new_event(const boost::intrusive_ptr<Cls>& cls, Fxn fxn)
    {
    event::ptr callback = new generic_event_0_args<Cls, Fxn> (cls, fxn);
    return callback;
    }

    template<class Cls, typename Fxn, class Arg1>
    event::ptr
    new_event(const boost::intrusive_ptr<Cls>& cls, Fxn fxn, const Arg1& arg1)
    {
      event::ptr callback = new generic_event_1_args<Cls, Fxn, Arg1> (cls, fxn, arg1);
      return callback;
    }

    template<class Cls, typename Fxn, class Arg1, class Arg2>
    event::ptr
    new_event(const boost::intrusive_ptr<Cls>& cls, Fxn fxn, const Arg1& arg1, const Arg2& arg2)
    {
        event::ptr callback = new generic_event_2_args<Cls, Fxn, Arg1, Arg2> (cls, fxn, arg1, arg2);
        return callback;
    }


  template<class Cls, typename Fxn>
    class eventcallback_tmpl_0_args : public eventcallback
    {
    private:
      boost::intrusive_ptr<Cls> obj_;

      Fxn fxn_;

    public:
      eventcallback_tmpl_0_args(const boost::intrusive_ptr<Cls>& obj, Fxn fxn) :
        obj_(obj), fxn_(fxn)
      {
      }

      void
      callback(const sst_message::ptr& msg)
      {
        Cls& obj = *obj_;
        (obj_.get()->*fxn_)();
      }

    };

  template<class Cls, typename Fxn, typename Arg1>
    class eventcallback_tmpl_1_args : public eventcallback
    {
    private:
      boost::intrusive_ptr<Cls> obj_;

      Fxn fxn_;

      Arg1 arg1_;

    public:
      eventcallback_tmpl_1_args(const boost::intrusive_ptr<Cls>& obj, Fxn fxn,
          const Arg1& arg1) :
        obj_(obj), fxn_(fxn), arg1_(arg1)
      {
      }

      void
      callback(const sst_message::ptr& msg)
      {
        (obj_.get()->*fxn_)(arg1_);
      }

    };

  template<class Cls, typename Fxn, typename Arg1, typename Arg2>
    class eventcallback_tmpl_2_args : public eventcallback
    {
    private:
      boost::intrusive_ptr<Cls> obj_;

      Fxn fxn_;

      Arg1 arg1_;

      Arg2 arg2_;

    public:
      eventcallback_tmpl_2_args(const boost::intrusive_ptr<Cls>& obj, Fxn fxn,
          const Arg1& arg1, const Arg2& arg2) :
        obj_(obj), fxn_(fxn), arg1_(arg1), arg2_(arg2)
      {
      }

      void
      callback(const sst_message::ptr& msg)
      {
        (obj_.get()->*fxn_)(arg1_, arg2_);
      }

    };

  template<class Cls, typename Fxn, typename Arg1, typename Arg2, typename Arg3>
    class eventcallback_tmpl_3_args : public eventcallback
    {
    private:
      boost::intrusive_ptr<Cls> obj_;

      Fxn fxn_;

      Arg1 arg1_;

      Arg2 arg2_;

      Arg3 arg3_;

    public:
      eventcallback_tmpl_3_args(const boost::intrusive_ptr<Cls>& obj, Fxn fxn,
          const Arg1& arg1, const Arg2& arg2, const Arg3& arg3) :
        obj_(obj), fxn_(fxn), arg1_(arg1), arg2_(arg2), arg3_(arg3)
      {
      }

      void
      callback(const sst_message::ptr& msg)
      {
        (obj_.get()->*fxn_)(arg1_, arg2_, arg3_);
      }

    };

    class event_step_callback : public eventcallback
    {

        protected:
            bool executed_;

            event_step_callback();

        public:
            typedef boost::intrusive_ptr<event_step_callback> ptr;

            virtual void callback(long nbytes, long byte_offset) = 0;

            void callback(const sst_message::ptr &msg);

            std::string to_string() const;
    };

    template<class Cls, typename Fxn>
      class event_step_callback_tmpl : public event_step_callback
      {
      private:
        boost::intrusive_ptr<Cls> obj_;

        Fxn fxn_;

        sst_message::ptr msg_;

      public:
        event_step_callback_tmpl(const boost::intrusive_ptr<Cls>& obj, Fxn fxn, const sst_message::ptr& msg) :
            obj_(obj), fxn_(fxn), msg_(msg)
        {
        }

        void
        callback(long nbytes, long byte_offset)
        {
            executed_ = true;
            (obj_.get()->*fxn_)(nbytes, byte_offset, msg_);
        }

      };

  template<class Cls, typename Fxn>
    event_step_callback::ptr
    new_event_step_callback(const boost::intrusive_ptr<Cls>& cls, Fxn fxn, const sst_message::ptr& msg)
    {
      event_step_callback::ptr callback = new event_step_callback_tmpl<Cls, Fxn> (cls, fxn, msg);
      return callback;
    }



  template<class Cls, typename Fxn>
    eventcallback::ptr
    new_callback(const boost::intrusive_ptr<Cls>& cls, Fxn fxn)
    {
      eventcallback::ptr callback = new eventcallback_tmpl_0_args<Cls, Fxn> (
          cls, fxn);
      return callback;
    }

  template<class Cls, typename Fxn, typename Arg1>
    eventcallback::ptr
    new_callback(const boost::intrusive_ptr<Cls>& cls, Fxn fxn,
        const Arg1& arg1)
    {
      eventcallback::ptr callback = new eventcallback_tmpl_1_args<Cls, Fxn,
          Arg1> (cls, fxn, arg1);
      return callback;
    }

  template<class Cls, typename Fxn, typename Arg1, typename Arg2>
    eventcallback::ptr
    new_callback(const boost::intrusive_ptr<Cls>& cls, Fxn fxn,
        const Arg1& arg1, const Arg2& arg2)
    {
      eventcallback::ptr callback = new eventcallback_tmpl_2_args<Cls, Fxn,
          Arg1, Arg2> (cls, fxn, arg1, arg2);
      return callback;
    }

  template<class Cls, typename Fxn, typename Arg1, typename Arg2, typename Arg3>
    eventcallback::ptr
    new_callback(const boost::intrusive_ptr<Cls>& cls, Fxn fxn,
        const Arg1& arg1, const Arg2& arg2, const Arg3& arg3)
    {
      eventcallback::ptr callback = new eventcallback_tmpl_3_args<Cls, Fxn,
          Arg1, Arg2, Arg3> (cls, fxn, arg1, arg2, arg3);
      return callback;
    }

  template<class Cls, typename Fxn, typename Arg1>
    sst_message::ptr
    new_message_callback(const boost::intrusive_ptr<Cls>& cls, Fxn fxn,
        const Arg1& arg1)
    {
      eventcallback::ptr callback = new eventcallback_tmpl_1_args<Cls, Fxn,
          Arg1> (cls, fxn, arg1);
      sst_message::ptr msg = sst_message::construct();
      msg->add_callback(callback);
      return msg;
    }

  template<class Cls, typename Fxn, typename Arg1, typename Arg2>
     sst_message::ptr
     new_message_callback(const boost::intrusive_ptr<Cls>& cls, Fxn fxn,
         const Arg1& arg1, const Arg2& arg2)
     {
       eventcallback::ptr callback = new eventcallback_tmpl_2_args<Cls, Fxn,
           Arg1, Arg2> (cls, fxn, arg1, arg2);
       sst_message::ptr msg = sst_message::construct();
       msg->add_callback(callback);
       return msg;
     }

  template<class Cls, typename Fxn, typename Arg1, typename Arg2, typename Arg3>
    sst_message::ptr
    new_message_callback(const boost::intrusive_ptr<Cls>& cls, Fxn fxn,
        const Arg1& arg1, const Arg2& arg2, const Arg3& arg3)
    {
      eventcallback::ptr callback = new eventcallback_tmpl_3_args<Cls, Fxn,
          Arg1, Arg2, Arg3> (cls, fxn, arg1, arg2, arg3);
      sst_message::ptr msg = sst_message::construct();
      msg->add_callback(callback);
      return msg;
    }

} // end of namespace sstmac
#endif
