/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_COMMON_EVENTMODULE_H_INCLUDED
#define SSTMAC_COMMON_EVENTMODULE_H_INCLUDED

#include <sstmac/common/eventmanager.h>
#include <sstmac/common/eventhandler.h>
#include <sstmac/common/timestamp.h>
#include <sstmac/common/ptr_type.h>
#include <sstmac/common/messages/sst_message.h>

namespace sstmac
{

  /**
   * The interface for something that can both handle and schedule messages,
   * such as a network model.
   */
  class eventmodule :
    virtual public eventhandler
  {
  public:
    typedef boost::intrusive_ptr<eventmodule> ptr;

  protected:
    bool finished_;

    eventmodule() :
        finished_(false)
    {
    }

  public:

    virtual
    ~eventmodule()
    {
    }

    virtual void finish()
    {
        finished_ = true;
    }

    virtual void handle(const sst_message::ptr& msg)
    {
        //do nothing by default
    }

    /**
     * Set the eventmanager for this module.  Unfortunately, this always has to be called
     * after an eventmodule is constructed.
     * @param m the simulation eventmanager
     */
    virtual void
    set_eventmanager(const eventmanager::ptr &m);

    eventmanager::ptr
    get_eventmanager() const;

    /**
     * get the current time
     * @return a timestamp
     */
    timestamp
    now() const;

    virtual std::string
    to_string() const
    {
      return "eventmodule";
    }

    /**
     * Register a stat_logger that should be notified when the simulation is done
     * @param stat a smart pointer to a stat_logger
     */
    void
    register_stat(stat_logger::ptr stat);

    void
    cancel_all_messages();

    /**
     * Add an event to the event queue, where msg will get delivered to handler at time t.
     * @param t Time at which the event should happen
     * @param handler The handler for the event
     * @param msg The message to deliver to the handler
     */
    void
    schedule(const timestamp &t,
        const eventhandler::ptr &handler,
        const sst_message::ptr &msg = sst_message::ptr());

    void schedule(const timestamp& t, const event::ptr& ev);

    void schedule_now(const event::ptr& ev);

    void schedule_delay(
        const timestamp &delay,
        const eventhandler::ptr& handler,
        const sst_message::ptr& msg);

    void schedule_delay(const timestamp& delay, const event::ptr& ev);

  private:
    eventmanager::ptr eventman_;
    std::list<stat_logger::ptr> need_reg_;

  };

} // end of namespace sstmac
#endif

