/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_COMMON_MESSAGES_MESSAGEID_H_INCLUDED
#define SSTMAC_COMMON_MESSAGES_MESSAGEID_H_INCLUDED

#include <iostream>
#include <sstream>
#include <sstmac/common/types.h>

namespace sstmac
{

  /**
   * A ligthweight wrapper for a message index.
   * Done to avoid annoying mistakes where index values are mixed up.
   * 
   */
  struct messageid
  {
    // These get pretty large, and we need to avoid collisions
    // (otherwise, pending ack requests can start clobbering each other)
    long long int id;

    /**
     *
     * @param v the value to set this id to
     */
    explicit
    messageid(long long int v = -1) :
      id(v)
    {
    }

    /**
     * Increment operator
     * @return the same object with it's value incremented
     */
    inline messageid&
    operator++()
    {
      ++id;
      return *this;
    }

    /**
     * Increment operator
     * @param not used
     * @return a new object with a value of one more
     */
    inline messageid
    operator++(int)
    {
      messageid other(*this);
      ++id;
      return other;
    }

    /**
     * std::stringifier
     * @return a std::string describing this object
     */
    std::string
    to_string() const
    {
      std::stringstream ss(std::stringstream::in | std::stringstream::out);
      ss << "messageid(id=" << id << ")";
      return ss.str();
    }


  };

  inline bool
  operator==(const messageid &a, const messageid &b)
  {
    return (a.id == b.id);
  }
  inline bool
  operator!=(const messageid &a, const messageid &b)
  {
    return (a.id != b.id);
  }
  inline bool
  operator<(const messageid &a, const messageid &b)
  {
    return (a.id < b.id);
  }
  inline bool
  operator<=(const messageid &a, const messageid &b)
  {
    return (a.id <= b.id);
  }
  inline bool
  operator>(const messageid &a, const messageid &b)
  {
    return (a.id > b.id);
  }
  inline bool
  operator>=(const messageid &a, const messageid &b)
  {
    return (a.id >= b.id);
  }
  inline std::ostream&
  operator<<(std::ostream &os, const messageid &n)
  {
    return (os << "messageid(" << n.id << ")");
  }

} // end of namespace sstmac

/*namespace boost
 {
 size_t
 hash_value(const sstmac::messageid &t);
 }*/

#endif
