/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_COMMON_MESSAGES_PAIRPAYLOAD_H_INCLUDED
#define SSTMAC_COMMON_MESSAGES_PAIRPAYLOAD_H_INCLUDED

#include <sstream>

#include <sstmac/common/messages/payload.h>

namespace sstmac
{

  /**
   * Network payload consisting of a value with sensible copy semantics.
   *
   * Don't use this for pointer types, since you will break assumptions
   * about const-ness or protection of underlying data.
   */
  template<typename T1, typename T2>
    class pairpayload : public payload
    {
      /// The data we keep.
      mutable T1 data1_;
      mutable T2 data2_;
      size_t size_;

      /// Construction time.
      pairpayload(const T1 &inval1, const T2 &inval2,
          size_t size = sizeof(T1) + sizeof(T2)) :
        data1_(inval1), data2_(inval2), size_(size)
      {
      }

      /// Private copy constructor.
      pairpayload(const pairpayload<T1, T2> &other,
          size_t size = sizeof(T1) + sizeof(T2)) :
        data1_(other.data1_), data2_(other.data2_), size_(size)
      {
      }

    public:
      typedef boost::intrusive_ptr<pairpayload<T1, T2> > ptr;
      typedef boost::intrusive_ptr<const pairpayload<T1, T2> > const_ptr;

      virtual
      ~pairpayload()
      {
      }

      /// Construct a new valuepayload.
      static typename pairpayload<T1, T2>::const_ptr
      construct(const T1 &inval1, const T2 &inval2)
      {
        // Results in ICE in gcc-4.0.
        //return typename boost::shared_ptr< valuepayload<T> >
        //  (new valuepayload(inval));
        return typename pairpayload::const_ptr(
            new pairpayload<T1, T2> (inval1, inval2));
      }

      /// Clone this object.
      payload::const_ptr
      clone() const
      {
        // Results in ICE in gcc-4.0
        //return typename boost::shared_ptr< valuepayload<T> >
        //  (new valuepayload(*this));
        return typename pairpayload::const_ptr(new pairpayload<T1, T2> (*this));
      }

      /// Access the underlying data in a non-const context.
      std::pair<T1, T2>
      data_unconst()
      {
        return std::pair<T1, T2>(data1_, data2_);
      }

      /// Access the underlying data in a const context.
      const std::pair<T1, T2>
      data() const
      {
        return std::pair<T1, T2>(data1_, data2_);
      }

      virtual long
      get_byte_length() const
      {
        return size_;
      }

      /**
       * Add operator
       * @param other
       * @return a smart pointer to a new valuepayload object
       */
      virtual const payload::const_ptr
      add(const payload::const_ptr &other) const
      {
        const_ptr casted = boost::dynamic_pointer_cast<
            const pairpayload<T1, T2> >(other);
        T1 newval1 = casted->data().first + data().first;
        T2 newval2 = casted->data().second + data().second;
        payload::const_ptr ret = construct(newval1, newval2);
        return ret;
      }

      /**
       * Less than comparator, to only be used for the minloc mpi operation
       * @param other
       * @return
       */
      virtual const payload::const_ptr
      min(const payload::const_ptr &other) const
      {
        const_ptr casted = boost::dynamic_pointer_cast<
            const pairpayload<T1, T2> >(other);
        if (data1_ < casted->data().first)
        {
          return pairpayload<T1, T2>::construct(data1_, data2_);
        }
        else
        {
          return pairpayload<T1, T2>::construct(casted->data().first,
              casted->data().second);
        }

      }

      /**
       * Greater than comparator, to only be used for the maxloc operation
       * @param other
       * @return
       */
      virtual const payload::const_ptr
      max(const payload::const_ptr &other)  const
      {
        const_ptr casted = boost::dynamic_pointer_cast<
            const pairpayload<T1, T2> >(other);
        if (data1_ > casted->data().first)
        {
          return pairpayload<T1, T2>::construct(data1_, data2_);
        }
        else
        {
          return pairpayload<T1, T2>::construct(casted->data().first,
              casted->data().second);
        }

      }

      /**
       * Equals comparator
       * @param other
       * @return
       */
      virtual bool
      equals(const payload::const_ptr &other) const
      {
        const_ptr casted = boost::dynamic_pointer_cast<
            const pairpayload<T1, T2> >(other);
        return data() == casted->data();

      }

      /**
       * Logical or comparator
       * @param other
       * @return
       */
      const payload::const_ptr
      logical_or(const payload::const_ptr &other) const
      {
        sst_throw(unimplementederror,
            "pairpayload: logical_or doesn't make sense");
      }

      /**
       * Logical xor comparator
       * @param other
       * @return
       */
      const payload::const_ptr
      logical_xor(const payload::const_ptr &other) const
      {
        sst_throw(unimplementederror,
            "pairpayload: logical_xor doesn't make sense");
      }

      /**
       * Logical and comparator
       * @param other
       * @return
       */
      const payload::const_ptr
      logical_and(const payload::const_ptr &other) const
      {
        sst_throw(unimplementederror,
            "pairpayload: logical_and doesn't make sense");
      }

      /**
       * Serialize this object into the sst_serializer, for
       * running in parallel simulation
       * @param ser the serializer to use
       */
      virtual void
      serialize(sst_serializer* ser) const
      {
        //  SSTMAC_DEBUG << "valuepayload<" << typeid(T).name()
        //        << ">: serializing \n";
        //ser->pack(get_deser_name());
        //ser->pack(data_);
        sst_throw(unimplementederror, "pairpayload: serialize unimplemented");
      }

      /**
       * Deserialize a valuepayload out of the serializer
       * @param ser the serializer that contains a valuepayload
       * @return a smart pointer to a new valuepayload object
       */
      static payload::const_ptr
      deserialize(sst_serializer* ser)
      {
        //SSTMAC_DEBUG << "valuepayload<" << typeid(T).name()
        //      << ">: deserializing \n";

        //T v;
        // ser->unpack(v);

        //return construct(v);
        sst_throw(unimplementederror, "pairpayload: deserialize unimplemented");
      }

      /**
       * Strinfier
       * @return a std::string description
       */
      virtual std::string
      to_string() const
      {
        std::stringstream ss;
        ss << "pairpayload(" << data1_ << ", " << data2_ << ")";
        return ss.str();
      }

      /**
       * The deserialization name, so sst_serializer can reconstruct this
       * @return a std::string name for this class
       */
      static std::string
      get_deser_name()
      {
        std::stringstream ss;
        ss << "pairpayload<" << typeid(T1).name() << ", " << typeid(T2).name()
            << ">";
        return ss.str();
      }
    };

  /**
   * Network payload consisting of a value with sensible copy semantics.
   *
   * Don't use this for pointer types, since you will break assumptions
   * about const-ness or protection of underlying data.
   */
  template<typename T1, typename T2>
    class pairpayloadvector : public payload
    {

      struct structtype
      {
        T1 a;
        T2 b;
      };

      /// The data we keep.
      void* buf_;
      size_t size_;

      /// Construction time.
      pairpayloadvector(void* start, size_t size, bool copy) :
        size_(size)
      {
        if (copy)
        {
          size_t s = (sizeof(T1) + sizeof(T2)) * size;
          buf_ = malloc(s);

          memcpy(buf_, start, s);
        }
        else
        {
          buf_ = start;
        }
      }

      /// Private copy constructor.
      pairpayloadvector(const pairpayloadvector<T1, T2> &other, size_t size) :
        size_(size)
      {
        size_t s = (sizeof(T1) + sizeof(T2)) * size;
        buf_ = malloc(s);

        memcpy(buf_, other.buf_, s);
      }

    public:
      typedef boost::intrusive_ptr<pairpayloadvector<T1, T2> > ptr;
      typedef boost::intrusive_ptr<const pairpayloadvector<T1, T2> > const_ptr;

      virtual
      ~pairpayloadvector()
      {
        free(buf_);
      }

      /// Construct a new valuepayload.
      static typename pairpayloadvector<T1, T2>::const_ptr
      construct(void* start, size_t size, bool copy = true)
      {
        // Results in ICE in gcc-4.0.
        //return typename boost::shared_ptr< valuepayload<T> >
        //  (new valuepayload(inval));
        return typename pairpayloadvector::const_ptr(
            new pairpayloadvector<T1, T2> (start, size, copy));
      }

      /// Clone this object.
      payload::const_ptr
      clone() const
      {
        // Results in ICE in gcc-4.0
        //return typename boost::shared_ptr< valuepayload<T> >
        //  (new valuepayload(*this));
        return typename pairpayloadvector::const_ptr(
            new pairpayloadvector<T1, T2> (*this));
      }

      /// Access the underlying data in a non-const context.
      void*
      data_unconst()
      {
        return buf_;
      }

      /// Access the underlying data in a const context.
      const void*
      data() const
      {
        return buf_;
      }

      virtual long
      get_byte_length() const
      {
        return size_ * (sizeof(T1) + sizeof(T2));
      }

      /**
       * Add operator
       * @param other
       * @return a smart pointer to a new valuepayload object
       */
      virtual const payload::const_ptr
      add(const payload::const_ptr &other) const
      {
        const_ptr casted = boost::dynamic_pointer_cast<const pairpayloadvector<
            T1, T2> >(other);
        size_t s = (sizeof(T1) + sizeof(T2)) * size_;
        structtype* newdata = (structtype*) malloc(s);
        structtype* thisdata = (structtype*) data();
        structtype* thatdata = (structtype*) casted->data();
        for (int i = 0; i < size_; i++)
        {
          newdata->a = thisdata->a + thatdata->a;
          newdata->b = thisdata->b + thatdata->b;
          newdata++;
          thisdata++;
          thatdata++;
        }
        payload::const_ptr ret = construct(newdata, size_, false);
        return ret;
      }

      /**
       * Less than comparator, to only be used for the minloc mpi operation
       * @param other
       * @return
       */
      virtual const payload::const_ptr
      min(const payload::const_ptr &other) const
      {
        sst_throw(unimplementederror, "pairpayloadvector::less_than - doesn't make sense");

      }

      /**
       * Greater than comparator, to only be used for the maxloc operation
       * @param other
       * @return
       */
      virtual const payload::const_ptr
      max(const payload::const_ptr &other)  const
      {
        sst_throw(unimplementederror, "pairpayloadvector::greater_than - doesn't make sense");
      }

      /**
       * Equals comparator
       * @param other
       * @return
       */
      virtual bool
      equals(const payload::const_ptr &other) const
      {
        const_ptr casted = boost::dynamic_pointer_cast<const pairpayloadvector<
            T1, T2> >(other);
        bool ret = true;
        structtype* thisdata = (structtype*) data();
        structtype* thatdata = (structtype*) casted->data();
        for (int i = 0; i < size_ && ret; i++)
        {
          ret = ret && (thisdata->a == thatdata->a);
          ret = ret && (thisdata->b == thatdata->b);
        }

        return ret;

      }

      /**
       * Logical or comparator
       * @param other
       * @return
       */
      const payload::const_ptr
      logical_or(const payload::const_ptr &other) const
      {
        sst_throw(unimplementederror,
            "pairpayloadvector: logical_or doesn't make sense");
      }

      /**
       * Logical xor comparator
       * @param other
       * @return
       */
      const payload::const_ptr
      logical_xor(const payload::const_ptr &other) const
      {
        sst_throw(unimplementederror,
            "pairpayloadvector: logical_xor doesn't make sense");
      }

      /**
       * Logical and comparator
       * @param other
       * @return
       */
      const payload::const_ptr
      logical_and(const payload::const_ptr &other) const
      {
        sst_throw(unimplementederror,
            "pairpayloadvector: logical_and doesn't make sense");
      }

      /**
       * Serialize this object into the sst_serializer, for
       * running in parallel simulation
       * @param ser the serializer to use
       */
      virtual void
      serialize(sst_serializer* ser) const
      {
        //  SSTMAC_DEBUG << "valuepayload<" << typeid(T).name()
        //        << ">: serializing \n";
        //ser->pack(get_deser_name());
        //ser->pack(data_);
        sst_throw(unimplementederror, "pairpayloadvector: serialize unimplemented");
      }

      /**
       * Deserialize a valuepayload out of the serializer
       * @param ser the serializer that contains a valuepayload
       * @return a smart pointer to a new valuepayload object
       */
      static payload::const_ptr
      deserialize(sst_serializer* ser)
      {
        //SSTMAC_DEBUG << "valuepayload<" << typeid(T).name()
        //      << ">: deserializing \n";

        //T v;
        // ser->unpack(v);

        //return construct(v);
        sst_throw(unimplementederror, "pairpayloadvector: deserialize unimplemented");
      }

      /**
       * Strinfier
       * @return a std::string description
       */
      virtual std::string
      to_string() const
      {
        std::stringstream ss;
        ss << "pairpayloadvector(" << size_ << ")";
        return ss.str();
      }

      /**
       * The deserialization name, so sst_serializer can reconstruct this
       * @return a std::string name for this class
       */
      static std::string
      get_deser_name()
      {
        std::stringstream ss;
        ss << "pairpayloadvector<" << typeid(T1).name() << ", "
            << typeid(T2).name() << ">";
        return ss.str();
      }
    };

} // end of namespace sstmac
#endif
