/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_COMMON_MESSAGES_SERIALIZABLE_H_INCLUDED
#define SSTMAC_COMMON_MESSAGES_SERIALIZABLE_H_INCLUDED

#include <sstmac/common/ptr_type.h>
#include <sstmac/common/factories/factory.h>
#include <sstmac/common/messages/sst_serializer.h>
#include <sstmac/common/factories/factory.h>
#include <boost/unordered_map.hpp>

namespace sstmac
{

  class serializable :
    virtual public ptr_type
  {
    protected:
    typedef enum { ConstructorFlag } cxn_flag_t;

    public:
    typedef boost::intrusive_ptr<serializable> ptr;

    typedef boost::intrusive_ptr<const serializable> const_ptr;

   /* virtual void
    serialize(sst_serializer* ser) const = 0;

    virtual void
    deserialize(sst_serializer* ser) = 0;

    virtual long
    data_serialization_size() const = 0;*/


    virtual void
    serialize_order(sst_serializer* ser) = 0;


    virtual long cls_id() const = 0;

    virtual ~serializable() { }


    virtual serializable::ptr serialization_clone() const {
      sst_throw(ssterror, "serializable::serialization_clone - sorry, you're trying to get a non-const clone of %s and it doesn't implement this function", to_string().c_str());
    }


  };

  template <class T>
  class serializable_type
  {
  protected:
    static long cls_id_;


    virtual
    T* you_forgot_to_add_ImplementSerialzable_to_this_class() = 0;

  public:
    virtual ~serializable_type(){}

    long cls_id() const {
        return cls_id_;
    }

  };

  #define ImplementSerializable(obj) \
    protected: \
        obj(cxn_flag_t flag){} \
    public: \
        virtual long cls_id() const { \
            return ::sstmac::serializable_type<obj>::cls_id(); \
        } \
        static obj* construct_deserialize_stub() { \
            return new obj(ConstructorFlag); \
        } \
        virtual std::string \
        get_serialization_name() const { \
            return #obj; \
        } \
        virtual obj* you_forgot_to_add_ImplementSerialzable_to_this_class() { return 0; }

  class serializable_builder
  {
  public:
    virtual serializable::ptr
    build() const = 0;

    virtual const char*
    name() const = 0;

    virtual bool
    sanity(serializable* ser) = 0;
  };

  template<class T>
    class serializable_builder_impl : public serializable_builder
    {
    protected:
      static const char* name_;

    public:
      serializable::ptr
      build() const
      {
        return T::construct_deserialize_stub();
      }

      const char*
      name() const
      {
        return name_;
      }

      virtual bool
      sanity(serializable* ser)
      {
#if SST_SANITY_CHECK
        return (typeid(T) == typeid(*ser));
#else
        return true;
#endif
      }
    };


  class serializable_factory
  {
    protected:
        typedef boost::unordered_map<long, serializable_builder*> builder_map;
        static builder_map* builders_;

    public:
        static serializable::ptr get_serializable(long cls_id);

        /**
            @return The cls id for the given builder
        */
        static long add_builder(serializable_builder* builder);

        static bool sanity(serializable* ser, long cls_id){
          return (*builders_)[cls_id]->sanity(ser);
        }
  };



  /*template <class PTR>
  void
  sst_serializer::unpack_object(PTR& ptr)
  {
    typedef typename PTR::element_type T;
    serializable::ptr obj = unpack_next_object();
    ptr = boost::dynamic_pointer_cast<T>(obj);
  }

  template <class constPTR>
  void
  sst_serializer::unpack_const_object(constPTR& constptr)
  {
    typedef typename constPTR::element_type constT;
    typedef typename Deconstifier<constT>::unconst_type T;
    typedef typename boost::intrusive_ptr<T> PTR;
    PTR ptr; unpack_object(ptr);
    constptr = ptr;
  }*/

}

#define DeclareBaseClassSerializable(obj) \
    namespace sstmac { \
    template<> boost::intrusive_ptr<obj > \
       sstmac::obj_deserializer<obj >::deserialize_object(long cid, sst_serializer* ser) { \
             sstmac::serializable::ptr o = \
                    sstmac::serializable_factory::get_serializable(cid); \
                o->serialize_order(ser); \
                return boost::dynamic_pointer_cast<obj >(o); \
          } \
           template<> bool \
           sstmac::obj_deserializer<obj >::sanity_check(serializable* ser, long cls_id){ \
             return serializable_factory::sanity(ser, cls_id); \
           }\
         }

#define DeclareSerializable(obj) \
        namespace sstmac { \
    template<> long serializable_type<obj >::cls_id_ \
        = serializable_factory::add_builder(new serializable_builder_impl<obj >); \
    template<> const char* serializable_builder_impl<obj >::name_ = #obj; \
    template<> boost::intrusive_ptr<obj > \
    sstmac::obj_deserializer<obj >::deserialize_object(long cid, sst_serializer* ser) { \
          sstmac::serializable::ptr o = \
                 sstmac::serializable_factory::get_serializable(cid); \
             o->serialize_order(ser); \
             return boost::dynamic_pointer_cast<obj >(o); \
        } \
        template<> bool \
        sstmac::obj_deserializer<obj >::sanity_check(serializable* ser, long cls_id){ \
          return serializable_factory::sanity(ser, cls_id); \
        }\
      }

#endif
