/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_COMMON_MESSAGES_VECTORPAYLOAD_H_INCLUDED
#define SSTMAC_COMMON_MESSAGES_VECTORPAYLOAD_H_INCLUDED

#include <sstream>

#include <sstmac/common/config.h>
#include <sstmac/common/messages/payload.h>
#include <boost/multi_array.hpp>

namespace sstmac
{

  /**
   * Network payload consisting of a boost multiarray.
   *
   */
  template<typename Type, typename VectorType = std::vector<Type> >
    class vector1payload :
        public payload,
        public serializable_type<vector1payload<Type,VectorType> >
    {
    ImplementSerializable(vector1payload)

      /// The data we keep.
      typedef VectorType arraytype;

      // arraytype data_;

      std::vector<int> dims_;

      Type* underneath_;

      /// Construction time.
      vector1payload(const arraytype &inval, int dim1, ptrdiff_t stride)
      {
        dims_.push_back(dim1);

        //do a deep copy
        underneath_ = (Type*) malloc(sizeof(Type) * dim1);
        for (int i = 0; i < dim1; i++)
        {
          underneath_[i] = inval[i * stride];
        }
      }

      //Used by clone
      vector1payload(Type* inval, int dim1, bool cloning, ptrdiff_t stride)
      {
        dims_.push_back(dim1);

        //do a deep copy
        underneath_ = (Type*) malloc(sizeof(Type) * dim1);
        for (int i = 0; i < dim1; i++)
        {
          underneath_[i] = inval[i * stride];
        }
      }

      vector1payload() {

      }

    public:
      typedef boost::intrusive_ptr<vector1payload<Type, VectorType> > ptr;
      typedef boost::intrusive_ptr<const vector1payload<Type, VectorType> >
          const_ptr;

      virtual
      ~vector1payload()
      {
        free(underneath_);
      }

      void
      assign(void *data)
      {
        Type* casted = (Type*) data;
        int count = dims_[0];
        for (int i = 0; i < count; ++i)
        {
          underneath_[i] = casted[i];
        }
      }

      static ptr
      construct(){
        return ptr(new vector1payload<Type, VectorType>());
      }

      static const_ptr
      construct(const arraytype &inval, int dim1, ptrdiff_t stride = 1)
      {
        return const_ptr(
            new vector1payload<Type, VectorType> (inval, dim1, stride));
      }

      /// Clone this object.
      payload::const_ptr
      clone() const
      {
        return const_ptr(
            new vector1payload<Type, VectorType> (underneath_, dims_[0], true,
                1));
      }

      /// Access the underlying data in a const context.
      const Type*
      data() const
      {
        return underneath_;
      }

      int
      get_dim() const
      {
        return dims_[0];
      }

      virtual long
      get_byte_length() const
      {
        int dimsize = 1;
        for (uint i = 0; i < dims_.size(); i++)
          dimsize *= dims_[i];
        return sizeof(Type) * dimsize;
      }

      /**
       * Add operator
       * @param other
       * @return a smart pointer to a new valuepayload object
       */
      virtual const payload::const_ptr
      add(const payload::const_ptr &other) const
      {
        const_ptr casted = boost::dynamic_pointer_cast<const vector1payload<
            Type, VectorType> >(other);

        Type* newval = (Type*) malloc(sizeof(Type) * dims_[0]);

        for (int i = 0; i < dims_[0]; i++)
        {

          newval[i] = underneath_[i] + casted->data()[i];

        }

        payload::const_ptr ret = const_ptr(
            new vector1payload<Type, VectorType> (newval, dims_[0], true, 1));

        free(newval);

        return ret;
      }

      virtual const payload::const_ptr
      prod(const payload::const_ptr &other) const
      {
        const_ptr casted = boost::dynamic_pointer_cast<const vector1payload<
            Type, VectorType> >(other);

        Type* newval = (Type*) malloc(sizeof(Type) * dims_[0]);

        for (int i = 0; i < dims_[0]; i++)
        {

          newval[i] = underneath_[i] * casted->data()[i];

        }

        payload::const_ptr ret = const_ptr(
            new vector1payload<Type, VectorType> (newval, dims_[0], true, 1));

        free(newval);

        return ret;
      }

      /**
       * Less than comparator
       * @param other
       * @return
       */
      virtual const payload::const_ptr
      min(const payload::const_ptr &other) const
      {
        const_ptr casted = boost::dynamic_pointer_cast<const vector1payload<
            Type, VectorType> >(other);

        Type* newval = (Type*) malloc(sizeof(Type) * dims_[0]);

        for (int i = 0; i < dims_[0]; i++)
        {

          newval[i] = (underneath_[i] < casted->data()[i]) ? underneath_[i]
              : casted->data()[i];

        }

        payload::const_ptr ret = const_ptr(
            new vector1payload<Type, VectorType> (newval, dims_[0], true, 1));

        free(newval);

        return ret;
      }

      /**
       * Greater than comparator
       * @param other
       * @return
       */
      virtual const payload::const_ptr
      max(const payload::const_ptr &other) const
      {
        const_ptr casted = boost::dynamic_pointer_cast<const vector1payload<
            Type, VectorType> >(other);

        Type* newval = (Type*) malloc(sizeof(Type) * dims_[0]);

        for (int i = 0; i < dims_[0]; i++)
        {

          newval[i] = (underneath_[i] > casted->data()[i]) ? underneath_[i]
              : casted->data()[i];

        }

        payload::const_ptr ret = const_ptr(
            new vector1payload<Type, VectorType> (newval, dims_[0], true, 1));

        free(newval);

        return ret;

      }

      /**
       * Equals comparator
       * @param other
       * @return
       */
      virtual bool
      equals(const payload::const_ptr &other) const
      {
        const_ptr casted = boost::dynamic_pointer_cast<const vector1payload<
            Type, VectorType> >(other);

        for (int i = 0; i < dims_[0]; i++)
        {
          if (underneath_[i] != casted->data()[i])
            return false;
        }

        return true;
      }

      /**
       * Logical or comparator
       * @param other
       * @return
       */
      virtual const payload::const_ptr
      logical_or(const payload::const_ptr &other) const
      {
        const_ptr casted = boost::dynamic_pointer_cast<const vector1payload<
            Type, VectorType> >(other);

        Type* newval = (Type*) malloc(sizeof(Type) * dims_[0]);

        for (int i = 0; i < dims_[0]; i++)
        {
          newval[i] = underneath_[i] || casted->data()[i];
        }

        payload::const_ptr ret = const_ptr(
            new vector1payload<Type, VectorType> (newval, dims_[0], true, 1));

        free(newval);

        return ret;
      }

      /**
       * Logical xor comparator
       * @param other
       * @return
       */
      virtual const payload::const_ptr
      logical_xor(const payload::const_ptr &other) const
      {
        const_ptr casted = boost::dynamic_pointer_cast<const vector1payload<
            Type, VectorType> >(other);

        Type* newval = (Type*) malloc(sizeof(Type) * dims_[0]);

        for (int i = 0; i < dims_[0]; i++)
        {
          newval[i] = ((underneath_[i] || casted->data()[i])
              && !(underneath_[i] && casted->data()[i]));
        }

        payload::const_ptr ret = const_ptr(
            new vector1payload<Type, VectorType> (newval, dims_[0], true, 1));

        free(newval);

        return ret;
      }

      /**
       * Logical and comparator
       * @param other
       * @return
       */
      virtual const payload::const_ptr
      logical_and(const payload::const_ptr &other) const
      {
        const_ptr casted = boost::dynamic_pointer_cast<const vector1payload<
            Type, VectorType> >(other);

        Type* newval = (Type*) malloc(sizeof(Type) * dims_[0]);

        for (int i = 0; i < dims_[0]; i++)
        {
          newval[i] = underneath_[i] && casted->data()[i];
        }

        payload::const_ptr ret = const_ptr(
            new vector1payload<Type, VectorType> (newval, dims_[0], true, 1));

        free(newval);

        return ret;
      }

      /**
       * Logical or comparator
       * @param other
       * @return
       */
      virtual const payload::const_ptr
      bitwise_or(const payload::const_ptr &other) const
      {
        sst_throw(ssterror, "vectorpayoad::bitwise| not implemented");
      }

      /**
       * Logical xor comparator
       * @param other
       * @return
       */
      virtual const payload::const_ptr
      bitwise_xor(const payload::const_ptr &other) const
      {
        sst_throw(ssterror, "vectorpayoad::bitwise^ not implemented");
      }

      /**
       * Logical and comparator
       * @param other
       * @return
       */
      virtual const payload::const_ptr
      bitwise_and(const payload::const_ptr &other) const
      {
        sst_throw(ssterror, "vectorpayoad::bitwise& not implemented");
      }



      virtual void
      serialize_order(sst_serializer* ser)
      {
        sst_throw(unimplementederror, "vector1payload::serialize");
      }



      /**
       * Strinfier
       * @return a std::string description
       */
      virtual std::string
      to_string() const
      {
        std::stringstream ss;
        ss << "vector1payload(" << typeid(Type).name() << ", " << dims_[0]
            << ")";
        return ss.str();
      }

    };

} // end of namespace sstmac
#endif
