/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_COMMON_MESSAGES_WRAPPERPAYLOAD_H_INCLUDED
#define SSTMAC_COMMON_MESSAGES_WRAPPERPAYLOAD_H_INCLUDED

#include <sstream>

#include <sstmac/common/config.h>
#include <sstmac/common/messages/payload.h>

namespace sstmac
{

  /**
   * Network payload that wraps a raw buffer.  Must be
   * careful with this, or else false shared-memory things
   * might happen.
   *
   */
  class wrapperpayload :
    public payload,
    public serializable_type<wrapperpayload>
  {
    ImplementSerializable(wrapperpayload)

    mutable void* base_;
    size_t size_; //in bytes

    /// Construction time.
    wrapperpayload(void *inval, size_t s) :
      base_(inval), size_(s)
    {

    }

    wrapperpayload() {

    }

  public:
    typedef boost::intrusive_ptr<wrapperpayload> ptr;
    typedef boost::intrusive_ptr<const wrapperpayload> const_ptr;

    virtual
    ~wrapperpayload()
    {

    }

    static const_ptr
    construct(void *inval, size_t s)
    {
      return const_ptr(new wrapperpayload(inval, s));
    }
    static ptr
    construct(){
      return ptr(new wrapperpayload());
    }

    /// Clone this object.
    payload::const_ptr
    clone() const
    {
      return const_ptr(new wrapperpayload(base_, size_));
    }

    /// Access the underlying data in a const context.
    const void*
    data() const
    {
      return base_;
    }

    void*
    data_unconst() const
    {
      return base_;
    }

    virtual long
    get_byte_length() const
    {
      return size_;
    }

    /**
     * Add operator
     * @param other
     * @return a smart pointer to a new valuepayload object
     */
    virtual const payload::const_ptr
    add(const payload::const_ptr &other) const
    {

      return other;
    }

    virtual const payload::const_ptr
    prod(const payload::const_ptr &other) const
    {

      return other;
    }

    /**
     * Less than comparator
     * @param other
     * @return
     */
    virtual const payload::const_ptr
    min(const payload::const_ptr &other) const
    {

      return other;
    }

    /**
     * Greater than comparator
     * @param other
     * @return
     */
    virtual const payload::const_ptr
    max(const payload::const_ptr &other) const
    {
      return other;

    }

    /**
     * Equals comparator
     * @param other
     * @return
     */
    virtual bool
    equals(const payload::const_ptr &other) const
    {
      const_ptr casted = boost::dynamic_pointer_cast<const wrapperpayload>(
          other);
      return base_ == casted->base_ && size_ == casted->size_;
    }

    /**
     * Logical or comparator
     * @param other
     * @return
     */
    virtual const payload::const_ptr
    logical_or(const payload::const_ptr &other) const
    {
      return other;
    }

    /**
     * Logical xor comparator
     * @param other
     * @return
     */
    virtual const payload::const_ptr
    logical_xor(const payload::const_ptr &other) const
    {
      return other;
    }

    /**
     * Logical and comparator
     * @param other
     * @return
     */
    virtual const payload::const_ptr
    logical_and(const payload::const_ptr &other) const
    {
      return other;
    }

    virtual const payload::const_ptr
    bitwise_or(const payload::const_ptr &other) const
    {
      return other;
    }

    /**
     * Logical xor comparator
     * @param other
     * @return
     */
    virtual const payload::const_ptr
    bitwise_xor(const payload::const_ptr &other) const
    {
      return other;
    }

    /**
     * Logical and comparator
     * @param other
     * @return
     */
    virtual const payload::const_ptr
    bitwise_and(const payload::const_ptr &other) const
    {
      return other;
    }



    virtual void
    serialize_order(sst_serializer* ser);



    /**
     * Strinfier
     * @return a std::string description
     */
    virtual std::string
    to_string() const
    {
      std::stringstream ss;
      ss << "wrapperpayload(" << (size_t) base_ << ", " << size_ << ")";
      return ss.str();
    }

  };

} // end of namespace sstmac
#endif
