/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/common/parameters/file_parameters.h>
#include <sstmac/keywords/keyword_registration.h>
#include <sstmac/common/fileio.h>

namespace sstmac
{

  void
  file_parameters::try_to_parse(const std::string& fname, bool fail_on_existing)
  {
    std::string inc_file = fname;
    std::string dir = "";
    std::string f_firstchar = inc_file.substr(0, 1);
    if (f_firstchar == "/")
    {
      //do nothing - this is an absolute path
    }
    else
    {
      size_t pos = fname.find_last_of('/');
      if (pos != std::string::npos)
      {
        dir = fname.substr(0, pos + 1);
      }
    }
    inc_file = trim_str(inc_file);
    try
    {
      parse_file(dir + inc_file, fail_on_existing);
    }
    catch (ioerror& e)
    {
      parse_file(inc_file, fail_on_existing);
    }
  }

  void
  file_parameters::parse_file(const std::string& input_fname,
      bool fail_on_existing)
  {
    std::string fname = trim_str(input_fname);

    std::ifstream in;
    SSTFileIO::open_file(in, fname);

    if (in.is_open())
    {

      std::string line;

      while (in.good())
      {
        std::getline(in, line);
        line = trim_str(line);

        if (line[0] == '#')
        { //a comment
          continue;
        }
        else if (line.find("=") != std::string::npos)
        { //an assignment
          std::pair<std::string, std::string> keyval;
          sim_parameters::ptr scope = sim_parameters::parse_param(line, keyval);
          if (fail_on_existing)
          {
            scope->add_param(keyval.first, keyval.second);
          }
          else
          {
            scope->add_param_override(keyval.first, keyval.second);
          }
        }
        else if (line.find("include") != std::string::npos)
        { //an include line
          std::string included_file = trim_str(line.substr(7));
          if (included_file.find("{") != std::string::npos)
          {
            multisim_includes_.push_back(included_file);
          }
          else
          {
            try_to_parse(included_file, fail_on_existing);
          }
        }
        else if (line.find("unset") != std::string::npos)
        {
          std::string unparam = trim_str(line.substr(5));
          std::pair<std::string, std::string> keyval;

          //trick the parser a little
          sim_parameters::ptr scope = sim_parameters::parse_param(
              unparam + "= 0", keyval);
          scope->remove_param(keyval.first);
        }
      }

    }
    else
    {
      SSTFileIO::not_found(fname);
    }
  }

  file_parameters::file_parameters()
  {
  }

  file_parameters::file_parameters(const maptype& p) :
    params_(p)
  {

  }

  sim_parameters::ptr
  file_parameters::clone()
  {
    return ptr(new file_parameters(params_));
  }

  void
  file_parameters::remove_param(const std::string & key)
  {
    params_.erase(key);
  }

  file_parameters::ptr
  file_parameters::construct(const std::string& filename)
  {
    ptr params = new file_parameters;
    params->parse_file(filename);
    return params;
  }

  std::string
  file_parameters::get_param(const std::string& key) const
  {
    SSTMAC_DEBUG << "file_parameters: getting param " << key << "\n";
    maptype::const_iterator it = params_.find(key);
    if (it == params_.end())
    {
      sst_throw(sstmac::valueerror, "file_parameters: could not find parameter %s", key.c_str());
    }
    return it->second;
  }

  void
  file_parameters::append_params(const std::string& filename)
  {
    parse_file(filename);
  }

  bool
  file_parameters::has_param(const std::string& key) const
  {
    maptype::const_iterator it = params_.find(key);
    return (it != params_.end());
  }

  void
  file_parameters::add_param(const std::string& key, const std::string& val)
  {
    SSTMAC_DEBUG << "setting key " << key << " to value " << val << "\n";

    if (params_.find(key) != params_.end())
    {
      sst_throw( sstmac::valueerror,
          "file_parameters::add_param - key already in params: %s", key.c_str());
    }
    else
    {
      if (val.find("{") != std::string::npos)
      {
        multisim_params_[key] = val;
      }

      KeywordRegistration::validate_keyword(key, val);
      params_[key] = val;
    }
  }

  void
  file_parameters::add_param_override(const std::string& key,
      const std::string& val)
  {
    SSTMAC_DEBUG << "setting key " << key << " to value " << val << "...";

    if (params_.find(key) != params_.end())
    {
      SSTMAC_DEBUG << "overriding \n";
    }
    else
    {
      SSTMAC_DEBUG << "adding new param \n";
    }

    if (val.find("{") != std::string::npos)
    {
      multisim_params_[key] = val;
      SSTMAC_DEBUG << "adding multi param! \n";
    }

    KeywordRegistration::validate_keyword(key, val);
    params_[key] = val;

  }

  void
  file_parameters::combine_into(sim_parameters::ptr sp)
  {
    BOOST_FOREACH(maptype::value_type it, params_)
          {
            sp->add_param_override(it.first, it.second);
          }
  }

  std::string
  file_parameters::to_string() const
  {
    return "file_parameters";
  }

  file_parameters::~file_parameters()
  {
    //there seems to be a memory leak in the boost implementation
    //but explicit clearing seems to fix it
    params_.clear();
  }
}
