/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_COMMON_SIM_PARAMETERS_H_INCLUDED
#define SSTMAC_COMMON_SIM_PARAMETERS_H_INCLUDED

#include <sstmac/common/ptr_type.h>
#include <sstmac/common/units.h>
#include <sstmac/common/debug.h>

#include <sstmac/common/basicstringtokenizer.h>
#include <boost/unordered_map.hpp>
#include <boost/foreach.hpp>

namespace sstmac
{
  class sim_parameters : public ptr_type
  {
  public:
    typedef boost::intrusive_ptr<sim_parameters> ptr;

  protected:
    typedef std::set<std::string> value_set;
    typedef std::map<std::string, value_set> params_read_map;
    params_read_map optional_params_read_;

    std::map<std::string, sim_parameters::ptr> subspaces_;

    sim_parameters()
    {
      if (!dbg_.is_real())
      {
        dbg_ = debug("<debug> parameters | params");
        logger::register_logger(&dbg_);
      }
    }

    virtual sim_parameters::ptr
    subspace_clone() = 0;

    typedef boost::unordered_map<std::string, std::string> maptype;

    void
    register_optional_param(const std::string& key, const std::string& val) const;

    void
    print_params(const maptype& pmap, std::ostream& os) const;

    sim_parameters::ptr
    get_param_scope(const std::string& ns);

  public:

    typedef boost::unordered_map<std::string, std::string> multisim_maptype;
    multisim_maptype multisim_params_;

    virtual void
    remove_param(const std::string &key) = 0;

    virtual std::string
    get_param(const std::string& key) const = 0;

    std::string
    reread_param(const std::string& key) const
    {
      return get_param(key);
    }


    /**
     * This is used for parallel simulation debugging.
     */
    void
    set_part_id(int p){
      dbg_.set_partition_id(p);
    }

    std::string
    reread_optional_param(const std::string& key, const std::string& def) const
    {
      return get_optional_param(key, def);
    }

    /// Return the value of the keyword if it exists. Otherwise return
    /// a default value.
    /// @param key gives the keyword
    /// @param def gives the default value (used if has_param(key) is false)
    /// @return the value if it exists, otherwise the default
    std::string
    get_optional_param(const std::string &key, const std::string &def) const;

    std::string
    deprecated_optional_param(const std::string &key, const std::string &def) const;

    std::string
    deprecated_param(const std::string& key) const;

    virtual sim_parameters::ptr
    clone() = 0;

    /**
     * Clone this object, but retain the multisim_params.
     * This function is needed for doing multisim includes, which requires
     * the files to be flattened first
     */
    virtual sim_parameters::ptr
    clone_multi(){
      ptr ret = clone();
      ret->multisim_params_ = multisim_params_;
      return ret;
    }

    virtual void
    add_param(const std::string& key, const std::string& val) = 0;

    virtual void
    add_param_override(const std::string& key, const std::string& val) = 0;

    virtual void
    combine_into(sim_parameters::ptr sp) = 0;

    /**
     @return The parameter set for the scope specific by p
     */
    sim_parameters::ptr
    parse_param(const std::string& p,
        std::pair<std::string, std::string>& keyval);

    virtual void
    print_params(std::ostream& os = std::cerr) const = 0;

    virtual bool
    is_multi_sim() const
    {
      return multisim_params_.size() > 0;
    }

    multisim_maptype
    get_multi_params()
    {
      return multisim_params_;
    }

    virtual bool
    has_param(const std::string& key) const = 0;

    int
    get_int_param(const std::string& key) const;

    int
    deprecated_int_param(const std::string& key) const;

    int
    deprecated_optional_int_param(const std::string &key, int def) const;

    int
    reread_int_param(const std::string& key) const;

    int
    reread_optional_int_param(const std::string& key, int def) const;

    /// Return the value of the keyword if it exists. Otherwise return
    /// a default value.
    /// @param key gives the keyword
    /// @param def gives the default value (used if has_param(key) is false)
    /// @return the value if it exists, otherwise the default
    int
    get_optional_int_param(const std::string &key, int def) const;

    /// Returns the value of the key as a boolean.
    bool
    get_bool_param(const std::string &key) const;

    bool
    reread_bool_param(const std::string& key) const;

    bool
    deprecated_bool_param(const std::string& key) const;

    bool
    deprecated_optional_bool_param(const std::string& key, bool def) const;

    bool
    reread_optional_bool_param(const std::string& key, bool def) const;

    /// Return the value of the keyword if it exists. Otherwise return
    /// a default value.
    /// @param key gives the keyword
    /// @param def gives the default value (used if has_param(key) is false)
    /// @return the value if it exists, otherwise the default
    bool
    get_optional_bool_param(const std::string &key, int def) const;

    double
    get_bandwidth_param(const std::string& key) const;

    double
    deprecated_bandwidth_param(const std::string& key) const;

    double
    reread_bandwidth_param(const std::string& key) const;

    double
    reread_optional_bandwidth_param(const std::string& key, double def) const;

    double
    deprecated_optional_bandwidth_param(const std::string& key, double def) const;

    double
    get_optional_bandwidth_param(const std::string &key, double def) const;

    double
    get_optional_bandwidth_param(
        const std::string& key,
        const std::string& def) const;

    long
    get_byte_length_param(const std::string& key) const;

    long
    get_optional_byte_length_param(const std::string& key, long def) const;

    long
    deprecated_byte_length_param(const std::string& key) const;

    long
    deprecated_optional_byte_length_param(const std::string& key, long def) const;

    double
    get_optional_freq_param(const std::string& key, double def) const;

    double
    get_freq_param(const std::string& key) const;

    double
    deprecated_freq_param(const std::string& key) const;

    double
    deprecated_optional_freq_param(const std::string& key, double def) const;

    /// Return the value of the keyword if it exists. Otherwise return
    /// a default value.
    /// @param key gives the keyword
    /// @param def gives the default value (used if has_param(key) is false)
    /// @return the value if it exists, otherwise the default
    long
    get_optional_long_param(const std::string &key, long def) const;

    long
    get_long_param(const std::string& key) const;

    long
    deprecated_long_param(const std::string& key) const;

    long
    deprecated_optional_long_param(const std::string& key, long def) const;

    long
    reread_long_param(const std::string& key) const;

    void
    reread_optional_long_param(const std::string& key) const;

    double
    get_double_param(const std::string& key) const;

    /// Return the value of the keyword if it exists. Otherwise return
    /// a default value.
    /// @param key gives the keyword
    /// @param def gives the default value (used if has_param(key) is false)
    /// @return the value if it exists, otherwise the default
    double
    get_optional_double_param(const std::string &key, double def) const;

    double
    reread_double_param(const std::string& key) const;

    double
    reread_optional_double_param(const std::string &key, double def) const;

    double
    deprecated_double_param(const std::string& key) const;

    double
    deprecated_optional_double_param(const std::string& key, double def) const;

    timestamp
    get_time_param(const std::string& key) const;

    /// Return the value of the keyword if it exists. Otherwise return
    /// a default value.
    /// @param key gives the keyword
    /// @param def gives the default value (used if has_param(key) is false)
    /// @return the value if it exists, otherwise the default
    timestamp
    get_optional_time_param(const std::string &key, timestamp def) const;

    timestamp
    deprecated_optional_time_param(const std::string &key, timestamp def) const;

    timestamp
    deprecated_time_param(const std::string& key) const;

    void
    get_vector_param(const std::string& key, std::vector<int>& vals) const;

    void
    get_vector_param(const std::string& key, std::vector<std::string>& vals) const;

  protected:
    static sstmac::debug dbg_;

  };

  class empty_parameters : public sstmac::sim_parameters
  {
  protected:

    empty_parameters()
    {
    }

    empty_parameters(maptype p) :
      params_(p)
    {
    }

  public:
    typedef boost::intrusive_ptr<empty_parameters> ptr;

    static ptr
    construct()
    {
      return ptr(new empty_parameters());
    }

    sim_parameters::ptr
    subspace_clone()
    {
      return new empty_parameters;
    }

    virtual sim_parameters::ptr
    clone()
    {
      return ptr(new empty_parameters(params_));
    }

    virtual void
    remove_param(const std::string & key)
    {

    }
    virtual std::string
    get_param(const std::string& key) const
    {
      SSTMAC_DEBUG << "file_parameters: getting param " << key << "\n";
      maptype::const_iterator it = params_.find(key);
      if (it == params_.end())
      {
        sst_throw(sstmac::valueerror, "file_parameters: could not find parameter %s", key.c_str());
      }
      return it->second;
    }

    virtual bool
    has_param(const std::string& key) const
    {
      maptype::const_iterator it = params_.find(key);
      return (it != params_.end());
    }

    virtual void
    print_params(std::ostream& os = std::cerr) const;

    virtual void
    add_param(const std::string& key, const std::string& val)
    {
      SSTMAC_DEBUG << "setting key " << key << " to value " << val << "\n";

      if (params_.find(key) != params_.end())
      {
        throw sstmac::valueerror(
            "file_parameters::add_param - key already in params");
      }
      else
      {
        params_[key] = val;
      }
    }

    virtual void
    add_param_override(const std::string& key, const std::string& val)
    {
      SSTMAC_DEBUG << "setting key " << key << " to value " << val << "...";

      if (params_.find(key) != params_.end())
      {
        SSTMAC_DEBUG << "overriding \n";
      }
      else
      {
        SSTMAC_DEBUG << "adding new param \n";
      }

      params_[key] = val;

    }

    virtual void
    combine_into(sim_parameters::ptr sp)
    {
      BOOST_FOREACH(maptype::value_type it, params_)
            {
              sp->add_param_override(it.first, it.second);
            }
    }

    virtual std::string
    to_string() const
    {
      return "empty_parameters";
    }

    virtual
    ~empty_parameters()
    {
    }

  protected:
    maptype params_;
  };



}

#endif
