/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_BACKENDS_NATIVE_SSTEVENT_H_INCLUDED
#define SSTMAC_BACKENDS_NATIVE_SSTEVENT_H_INCLUDED

#include <sstmac/common/messages/sst_message.h>
#include <sstmac/common/eventhandler.h>
#include <sstmac/common/malloc.h>
#include <sstmac/common/messages/serializable.h>

namespace sstmac
{
  class eventmodule;
  typedef boost::intrusive_ptr<eventmodule> eventmodulePtr;

  class event : virtual public ptr_type,
      public serializable
  {

  protected:
    event();

  public:
    typedef boost::intrusive_ptr<event> ptr;

    virtual
    ~event();

    const timestamp&
    time() const;

    void
    set_time(const timestamp& t);

    event::ptr next;

    virtual void
    cancel() = 0;

    virtual bool
    is_cancelled() const = 0;

    virtual void
    execute() = 0;

    eventmodulePtr
    get_module() const;

    void
    set_module(const eventmodulePtr& mod);


  protected:
    eventmodulePtr module_;

    timestamp time_;

  };

  inline bool
  operator >(event::ptr e1, event::ptr e2){
    return e1->time() > e2->time();
  }

  class simple_event : public event, public serializable_type<simple_event>
  {
  ImplementSerializable(simple_event)
  protected:
    simple_event(const sst_message::ptr &msg, const eventhandler::ptr &hand);

    simple_event()
    {
      canceled_ = false;
    }

    sst_message::ptr msg_to_deliver_;

    eventhandler::ptr handler_;

    bool canceled_;
  public:
    typedef boost::intrusive_ptr<simple_event> ptr;

    ~simple_event();

    static ptr
    construct(const sst_message::ptr &msg, const eventhandler::ptr &handler);

    virtual std::string
    to_string() const;

    sst_message::ptr
    get_message();
    eventhandler::ptr
    get_handler() const
    {
      return handler_;
    }

    void
    execute();

    void
    cancel();

    virtual bool
    is_cancelled() const
    {
      return canceled_;
    }

    virtual void
    serialize_order(sst_serializer* ser)
    {
      sst_throw(unimplementederror, "simple_event::serialize_order - not implemented");
    }

  };

  class generic_event : public event,
  public serializable_type<generic_event> {
    ImplementSerializable(generic_event)

  protected:
    generic_event() { }

  public:
    typedef boost::intrusive_ptr<generic_event> ptr;

    virtual
    ~generic_event(){ }



    virtual void
    cancel() {
      sst_throw(unimplementederror, "generic_event::cancel - not implemented");
    }

    virtual bool
    is_cancelled() const {
      sst_throw(unimplementederror, "generic_event::is_cancelled - not implemented");
    }

    virtual void
    execute() {
      sst_throw(unimplementederror, "generic_event::execute - not implemented");
    }

    virtual void
    serialize_order(sst_serializer* ser)
    {
      sst_throw(unimplementederror, "generic_event::serialize_order - not implemented");
    }

    virtual std::string
    to_string() const {
      return "generic_event";
    }

  };

} // end of namespace sstmac

#endif
