/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_COMMON_STATS_STAT_LOGGER_H_INCLUDED
#define SSTMAC_COMMON_STATS_STAT_LOGGER_H_INCLUDED

#include <sstmac/common/errors.h>
#include <sstmac/common/logger.h>
#include <sstmac/common/nocopy.h>
#include <sstmac/common/ptr_type.h>
#include <iostream>
#include <fstream>

namespace sstmac {


  /**
   * A type of logger that collects some kind of statistic
   * and outputs to a file during or at the end of a simulation.
   * Usually, a static instance of this class should be used,
   * because no merging of stat objects takes place, which means
   * you'll get one file for each stat object.
   */
class stat_logger : public logger
{
public:
  typedef boost::intrusive_ptr<stat_logger> ptr;

  static std::string stats_dir_;
protected:

  /**
   * Public constructor
   * @param namesig The name signature of this stat, presumably of the form "<stats> [the_name] | ... "
   * @param filename Name of the file to output results to
   */
  stat_logger(const std::string& namesig, const std::string& filename) :
      logger(namesig, NULL, true), fname_(filename)
  {

  }

  stat_logger() : logger() {

  }

public:

  virtual
  ~stat_logger();

  virtual void simulation_finished(timestamp end) = 0;

  virtual std::string
  to_string() const{
    return "stat_logger";
  }

  static void
  set_multisim(int n){
    multirun_ = n;
  }

protected:
  std::ofstream myfile;
  std::string fname_;

  static int multirun_;

  /**
   * Check to see if the file is open.  If not, try and open it and set
   * the current output stream to it, and throw an ssterror if something goes
   * wrong.
   * @return
   */
  bool
  check_open()
  {
    if (!myfile.is_open())
      {
        std::stringstream ss;
        ss << stats_dir_ << fname_;
        if(multirun_ >= 0)
           ss << multirun_;
        myfile.open(ss.str().c_str(), std::ios::out);

        if (!myfile.is_open())
          {
            sst_throw(ioerror, "statlogger: cannot open file %s for writing", ss.str().c_str());
          }

        outstream_ = &myfile;
      }

    return true;
  }
};
} // end of namespace sstmac
#endif
