/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/disk/vesper/disk_controller.h>

#include <math.h>

namespace sstmac
{
  namespace hw
  {

    debug disk_controller::dbg_;
    stat_time_window_sum::ptr disk_controller::stat_readbytes_;
    stat_time_window_sum::ptr disk_controller::stat_writebytes_;

    disk_controller::disk_controller(long nodeid, int num, const char *profile,
        const char *name, eventhandler::ptr callWhenDone, eventmanager::ptr man) :
      numdisk_(num), handler_(callWhenDone), lock_(false), nodeid_(nodeid)
    {
      if (!dbg_.is_real())
        {
          dbg_ = debug("<debug> disk_controller");
          logger::register_logger(&dbg_);
        }

      disk_.resize(num);
      for (int i = 0; i < numdisk_; i++)
        {
          disk_[i] = Disk::construct(profile, name, ptr(this));

          disk_[i]->set_eventmanager(man);
        }

      set_eventmanager(man);

      if (!stat_readbytes_)
        {
          const std::string fname = "stats_disk_readbytes.csv";
          stat_readbytes_ = stat_time_window_sum::construct("<stats> disk",
              fname);
          register_stat(stat_readbytes_);
        }
      if (!stat_writebytes_)
        {
          const std::string fname = "stats_disk_writebytes.csv";
          stat_writebytes_ = stat_time_window_sum::construct("<stats> disk",
              fname);
          register_stat(stat_writebytes_);
        }
    }

    void
    disk_controller::handle(const sst_message::ptr &msg)
    {
      lock_--;

      if (lock_ == 0)
        {
          handler_->handle(current_);
          if (!q_.empty())
            {
              sstmac::sw::disk_message::ptr da = q_.front();
              q_.pop();
              do_access(da);
            }
        }
    }

    void
    disk_controller::request_access(sst_message::ptr req)
    {
      sstmac::sw::disk_message::ptr cast = boost::dynamic_pointer_cast<
          sstmac::sw::disk_message>(req);

      SSTMAC_DEBUG << "disk_controller: requesting access type "
            << cast->get_access_type() << " at sector " << cast->get_sector()
            << " for " << cast->get_numsectors() << " sectors \n";

      if (lock_ == 0)
        {
          do_access(cast);
        }
      else
        {
          SSTMAC_DEBUG << "disk_controller: queuing access \n";
          q_.push(cast);
        }
    }

    void
    disk_controller::do_access(sstmac::sw::disk_message::ptr da)
    {
      use_me();

      current_ = da;
      SSTMAC_DEBUG << "disk_controller: doing access \n";
      switch (da->get_access_type())
        {
      case sstmac::sw::disk_message::OPEN:
        char* data;
        lock_ = numdisk_;

        for (int i = 0; i < numdisk_; i++)
          {
            disk_[i]->ReadRequest(da->get_sector(), 1, data);
          }

        break;

      case sstmac::sw::disk_message::CLOSE:
        lock_ = 1;
        schedule(now(), ptr(this), sst_message::ptr());
        break;

      case sstmac::sw::disk_message::READ:
        {
          char* data;
          lock_ = numdisk_;
          int numsec = ceil((double) da->get_numsectors() / (double) numdisk_);
          for (int i = 0; i < numdisk_; i++)
            {
              disk_[i]->ReadRequest(da->get_sector(), numsec, data);

            }
          stat_readbytes_->record(now(), da->get_numsectors() * Disk::blockSize);

          break;
        }

      case sstmac::sw::disk_message::WRITE:
        {
          char* data;
          lock_ = numdisk_;
          int numsec = ceil((double) da->get_numsectors() / (double) numdisk_);
          for (int i = 0; i < numdisk_; i++)
            {
              disk_[i]->WriteRequest(da->get_sector(), numsec, data);
            }
          stat_writebytes_->record(now(), da->get_numsectors() * Disk::blockSize);
          break;
        }
        }

    }

  }
} /* namespace sstmac */
