/*
 * mlparser.h
 *
 * Reads and parses an XML file into tags with attributes and contents (which may include more tags)
 * Supports multiple top-level tags
 *
 * Jonathan Pearson
 * July 20, 2004 -- Original version written in Java
 * February 18, 2005 -- Translated from Java to C#
 * November 2, 2005 -- Translated from C# to C++
 *
 */

#ifndef MLPARSER_H
#define MLPARSER_H


#include <cstring>
#include <sstmac/common/types.h>

namespace sstmac { namespace hw {

class MLTag;
class MLParser;



//represents either a std::string or an MLTag
//if str is true, use st.str, otherwose use st.tag
typedef struct {
	bool str;
	union {
        std::string *str;
		MLTag *tag;
	} st;
} element;

//taken directly from the SGI STL documentation
struct ltstr {
	bool operator() (const char *s1, const char *s2) const {
		return strcmp(s1, s2) < 0;
	}
};

typedef std::vector<element *> evec_t;
typedef evec_t::const_iterator evec_iter_t;

typedef std::map<const char *, std::string, ltstr, std::allocator<std::pair<const char*, std::string> > > attmap_t;
typedef attmap_t::const_iterator attmap_iter_t;

typedef std::vector<MLTag *> tagvec_t;
typedef tagvec_t::const_iterator tagvec_iter_t;

class MLTag {
public:
	MLTag();
	virtual ~MLTag();
	
	virtual std::string getName() const;
	virtual void setName(const std::string &nm);
	
	virtual void setAttribute(std::string attname, std::string attval);
	
	virtual void addContent(std::string *elem);
	virtual void addContent(MLTag *elem);
	
	virtual evec_iter_t getContents() const;
	virtual evec_iter_t getContentsEnd() const;
	
	virtual int countContents() const;
	
    virtual std::string stringAttribute(const std::string &name);
	virtual int intAttribute(const std::string &name);
	virtual long long longAttribute(const std::string &name);
	virtual float floatAttribute(const std::string &name);
	virtual bool booleanAttribute(const std::string &name);
	
	virtual std::string textContents() const;
	virtual std::string allTextContents() const;
	
	virtual int intContents() const;
	virtual int intAllContents() const;
	
	virtual long long longContents() const;
	virtual long long longAllContents() const;
	
	virtual float floatContents() const;
	virtual float floatAllContents() const;
	
	virtual bool booleanContents() const;
	virtual bool booleanAllContents() const;
	
	virtual const MLTag *findChild(const std::string &name) const;
	virtual const MLTag *findChild(const std::string &name, const MLTag *def) const;
	
private:
    std::string name;
	attmap_t attributes;
	evec_t contents;
};

class ParserData {
private:
	int inp;
	
public:
    std::string token;
    std::string attName;
	MLTag *curTag;
    std::stack<MLTag *> openTags;
	
	int line;
	int column;
	char c;
	int ch;
	int peek;
	
	ParserData(int input);
	virtual ~ParserData();
	
	virtual int Read();
};
	

class MLParser {
private:
	tagvec_t tags;
	
public:
	MLParser(int input);
	virtual ~MLParser();
	
	virtual tagvec_iter_t getTags() const;
	virtual tagvec_iter_t getTagsEnd() const;
	virtual int tagCount() const;
	virtual const MLTag *getTag(int index) const;
	virtual const MLTag *findChild(const std::string &name) const;
	
private:
	virtual void parse(int input);
	virtual void start(ParserData *pd);
	virtual void text(ParserData *pd);
	virtual void closeTag(ParserData *pd);
	virtual void attributes(ParserData *pd);
	virtual void quote1(ParserData *pd);
	virtual void quote2(ParserData *pd);
	
	virtual std::string chars(ParserData *pd);
	virtual std::string spaces(ParserData *pd);
	virtual std::string spacesOpt(ParserData *pd);
    virtual std::string stringx(ParserData *pd);
    virtual std::string stringOpt(ParserData *pd);
	virtual std::string charsWithoutQuote1(ParserData *pd);
	virtual std::string charsWithoutQuote2(ParserData *pd);
	
	virtual void match(ParserData *pd, char c);
	virtual bool consume(ParserData *pd);
};

} }
#endif
