/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_FASTFLOWER_FASTFLOW_H_
#define SSTMAC_HARDWARE_FASTFLOWER_FASTFLOW_H_

#include <sstmac/common/ptr_type.h>
#include <sstmac/common/messages/sst_message.h>
#include <sstmac/hardware/network/interconnect/interconnect_message.h>
namespace sstmac
{
  namespace hw
  {

    class fastflow : public ptr_type
    {

    public:
      class saved_route
      {
      public:
        //map switch ID to dim and dir
        boost::unordered_map<long, long> ports_;

        std::vector<boost::intrusive_ptr<fastflow> > split_;
        long splitid_;

        saved_route(){
          splitid_ = -1;
        }


        bool
        switch_in_route(long swid, long& port)
        {
            boost::unordered_map<long, long>::const_iterator
                it = ports_.find(swid);
            if (it != ports_.end()){
                port = it->second;
                return true;
            }
            else {
                return false;
            }
        }

        void
        add_to_route(long swid, long port)
        {
            ports_[swid] = port;
        }

        void
        split_flow(long swid, const std::vector<boost::intrusive_ptr<fastflow> > &flows){
          splitid_ = swid;
          split_ = flows;
        }

      };
    protected:

      long flowid_;
      static long FLOWIDCNT;

      // number of bytes this flow represents
      long flowbytes_;

      //minimum bandwidth constrained by components (maximum theoretical bw through system)
      double unit_bw_;

      //minimum bandwidth constrained by congestion that the head has seen
      double real_head_bw_;

      //the currently-scheduled head and tail
      interconnect_message::ptr current_tail_;
      interconnect_message::ptr current_head_;

      //tracks when the head got to different components
      std::queue<timestamp> headtimes_;

      //tracks the bandwidth that the head was, so we can subtract it
      std::queue<double> headbws_;

      double timetomakeup_;

      saved_route* route_;

      fastflow(const sst_message::ptr &orig, long bytes);
    public:
      typedef boost::intrusive_ptr<fastflow> ptr;

      static ptr
      construct(const sst_message::ptr &orig)
      {
        long bytes = orig->get_byte_length();
        return ptr(new fastflow(orig, bytes));
      }

      static ptr
      construct(const sst_message::ptr &orig, long bytes)
      {
        return ptr(new fastflow(orig, bytes));
      }

      virtual std::string
      to_string() const
      {
        return "fastflow";
      }

      virtual
      ~fastflow()
      {
        delete route_;
      }

      virtual long
      get_byte_length() const
      {
        return flowbytes_;
      }

      virtual void
      add_to_makeup(double t)
      {
        timetomakeup_ += t;
      }

      virtual void
      makeup_madeup(double t)
      {
        timetomakeup_ -= t;
      }

      virtual double
      get_makeup() const
      {
        return timetomakeup_;
      }

      virtual void
      done()
      {
        current_tail_ = 0;
        current_head_ = 0;
      }

      virtual double
      get_unit_bw() const
      {
        return unit_bw_;
      }

      static bool
      is_singleton(const sst_message::ptr &msg)
      {
        return msg->get_byte_length() <= 128;
      }

      virtual saved_route*
      get_saved_route()
      {
        return route_;
      }

      /**
       * The head has hit a unit bandwidth smaller than the current unit bandwidth.
       * Update the unit bandwidth, which will cause the current tail to be rescheduled
       * to a later time.
       * @param bw
       */
      virtual void
      update_unit_bw(double bw)
      {
        //std::cout << "flow(" << flowid_ << "): updating bw to " << bw << "\n";
        bw_valid_ = false;

        unit_bw_ = bw;

        if (real_head_bw_ == 0 || real_head_bw_ > unit_bw_)
        {
          real_head_bw_ = unit_bw_;
        }
      }

      virtual double
      get_real_head_bw() const
      {
        return real_head_bw_;
      }

      virtual void
      update_real_head_bw(double bw)
      {
        real_head_bw_ = bw;
      }

      virtual void
      add_head(timestamp t, double bw)
      {
        headtimes_.push(t);
        headbws_.push(bw);
      }

      virtual timestamp
      get_current_headtime() const
      {
        if (headtimes_.size() == 0)
        {
          sst_throw(ssterror, "fastflow(%d)::get_current_headtime - don't have a headtime", flowid_);
        }
        return headtimes_.front();
      }

      virtual double
      get_current_headbw() const
      {
        if (headbws_.size() == 0)
        {
          sst_throw(ssterror, "fastflow(%d)::get_current_headbw - don't have a headbw", flowid_);
        }
        return headbws_.front();
      }

      virtual void
      pop_head()
      {
        headtimes_.pop();
        headbws_.pop();
      }

      interconnect_message::ptr
      get_head()
      {
        return current_head_;
      }

      interconnect_message::ptr
      get_tail()
      {
        return current_tail_;
      }

      long
      get_flow_id() const
      {
        return flowid_;
      }

      //stuff for delaying the tail based on changes in unit bandwidth
      bool first_tail_;
      bool bw_valid_;
      timestamp init_tail_time_;
    };

    class fastflow_multipath : public fastflow
    {

    protected:
      fastflow_multipath(const sst_message::ptr &orig, long bytes) :
        fastflow(orig, bytes)
      {

      }

    public:
      typedef boost::intrusive_ptr<fastflow_multipath> ptr;

      static ptr
      construct(const sst_message::ptr &orig, long bytes)
      {
        return ptr(new fastflow_multipath(orig, bytes));
      }

      virtual
      ~fastflow_multipath()
      {
      }

      virtual std::string
      to_string() const
      {
        return fastflow::to_string() + ":multi";
      }
    };

  }
}
#endif
