/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/fastflower/fastflower_component.h>
#include <sstmac/hardware/fastflower/fastflow_message.h>

namespace sstmac
{

  namespace hw
  {

#define BW_MARGIN 0.0001

    debug fastflower_component::flow_dbg_;

    timestamp
    fastflower_component::model(const fastflow_message::ptr &fmsg,
        bool &reschedule)
    {

      if (fmsg->is_head()) //head
      {
        if (fmsg->get_headinfo()->is_flow_)
        {
          fastflow::ptr flow = fmsg->get_flow();
          double ubw = flow->get_unit_bw();

          if (flow_dbg_.is_active())
          {
            flow_dbg_ << "fastflower(" << name_ << ", " << flow->get_flow_id()
                << "): head passing with unit bw " << ubw << " and comp unit bw is " << unit_bw_ << "\n";
          }

          if (ubw < 0 || ubw > unit_bw_)
          {
            flow->update_unit_bw(unit_bw_);
          }

          //fbw = flow->get_unit_bw();
          double fbw = flow->get_real_head_bw();

          if(fbw < 0){
            sst_throw(ssterror, "fastflower_component::model - head - bw uh oh at get real head bw: %f", fbw);
          }

          current_bw_ += fbw;
          numflows_++;
          flow->add_head(now(), fbw);

          flowevent* f = new flowevent;
          f->time_ = now();
          f->numflows_ = numflows_;
          double over = std::max(0.0, current_bw_ - capacity_bw_);
          f->overcommit_ = over;
          history_.push_front(f);

          if (current_bw_ > capacity_bw_)
          {
            if (flow_dbg_.is_active())
            {
              flow_dbg_ << "fastflower(" << name_ << ", "
                  << flow->get_flow_id() << "): overcommitted by "
                  << (current_bw_ - capacity_bw_) << "\n";
            }
            double newf = ubw - over / numflows_;

            if(newf < 0){
              newf = current_bw_ / numflows_;
            }
            //  if(fbw > newf)
            flow->update_real_head_bw(newf);
          }
        }
        reschedule = false;

        // forward(fmsg, unit_delay_);
        return unit_delay_;
      }
      else //tail
      {
        fastflow::ptr flow = fmsg->get_flow();
        double ubw = flow->get_unit_bw();
        // double fbw = flow->get_real_head_bw();
        //figure out tail delay
        sst_message::ptr orig = fmsg->orig();
        long bytes = flow->get_byte_length();

        if (flow->first_tail_) //first time tail is being scheduled
        {
          timestamp delay(bytes / ubw);
          if (flow_dbg_.is_active())
          {
            flow_dbg_ << "fastflower(" << name_ << ", " << flow->get_flow_id()
                << ") rescheduling tail for the first time delayed by "
                << delay << "\n";
          }
          reschedule = true;
          flow->init_tail_time_ = now();
          flow->first_tail_ = false;
          flow->bw_valid_ = true;
          return delay;
        }
        else if (!flow->bw_valid_)
        { //unit bw has been updated

          timestamp timepassed = now() - flow->init_tail_time_;
          long bytespassed = ubw * timepassed.sec();
          long bytesleft = bytes - bytespassed;
          if (bytesleft > 0) //it's possible because of latency that it's fixed already
          {
            reschedule = true;
            double makeup = (bytesleft / ubw);
            if (flow_dbg_.is_active())
            {
              flow_dbg_ << "fastflower(" << name_ << ", "
                  << flow->get_flow_id() << ") rescheduling with delay "
                  << makeup << " for " << bytesleft << " bytes after "
                  << timepassed << " has passed with " << ubw << " bw \n";
            }
            flow->bw_valid_ = true;

            /*if (flow->get_makeup() > 0)
             {
             if (flow->get_makeup() > makeup)
             {
             flow->makeup_madeup(makeup); //cancellation of error, fall through
             makeup = 0;
             }
             else
             {
             double make = flow->get_makeup();
             makeup -= make;
             flow->makeup_madeup(make);

             return timestamp(makeup);
             }

             }
             else
             {*/
            return timestamp(makeup);
            //}
          }

        }

        double fbw = flow->get_unit_bw();

        reschedule = false;

        timestamp last = now();
        bool foundit = false;

        timestamp headtime = flow->get_current_headtime();

        if (flow_dbg_.is_active(2))
        {
          flow_dbg_ << "fastflower(" << name_
              << "): flow tail with total bytes " << bytes << " and headtime "
              << headtime << "\n";

        }

        // Based on the history at this component, delay the tail some if needed
        std::list<flowevent*>::iterator it, end = history_.end();
        for (it = history_.begin(); it != end && !foundit; it++)
        {

          flowevent* f = *it;

          if (flow_dbg_.is_active(2))
          {
            flow_dbg_ << "fastflower(" << name_
                << "): looking at event at time " << f->time_ << ", over: "
                << f->overcommit_ << ", flows: " << f->numflows_ << "...";

          }

          if (f->time_ < headtime) //this is where we stop

          {
            double bw = fbw;
            if (f->numflows_ > 0)
              bw -= (f->overcommit_ / f->numflows_);
            if (bw < 0)
            {
              bw = capacity_bw_ / f->numflows_;
            }

            if (flow_dbg_.is_active(2))
            {
              flow_dbg_ << "bw is " << bw << "\n";
            }
            bytes -= (last - headtime).sec() * (bw);
            foundit = true;

          }
          else
          {
            double bw = fbw;
            if (f->numflows_ > 0)
              bw -= (f->overcommit_ / f->numflows_);
            if (bw < 0)
            {
              bw = capacity_bw_ / f->numflows_;
            }

            if (flow_dbg_.is_active(2))
            {
              flow_dbg_ << "bw is " << bw << "\n";
            }
            bytes -= (last - f->time_).sec() * (bw);
          }
          last = f->time_;
        }

        //if we didn't go past the headtime in the history,
        ///  we have deleted it.  so make up for that
        if (!foundit)
        {
          bytes -= (last - flow->get_current_headtime()).sec() * (fbw);
        }

        // if (bytes < 0)
        // {
        //   std::cout << "i am: " << to_string() << "\n";
        //   sst_throw(ssterror, "fastflower_component:: bytes has gone under 0 in tail delay calculation");
        // }
        double bw = std::min(capacity_bw_ / numflows_, unit_bw_);
        //  if (current_bw_ > capacity_bw_)
        //    bw -= (current_bw_ - capacity_bw_) / numflows_;

        //  if (bw <= 0)
        //  {
        //     bw = fbw / numflows_;
        //  }

        double delay = std::max(0.0, (bytes / bw));

        /* if (delay < 0)
         {
         flow->add_to_makeup(-delay);
         delay = 0;
         }
         else if (flow->get_makeup() > 0)
         {
         if (flow->get_makeup() > delay)
         {
         flow->makeup_madeup(delay);
         delay = 0;
         }
         else
         {
         double make = flow->get_makeup();
         delay -= make;
         flow->makeup_madeup(make);
         }
         }*/

        if (flow_dbg_.is_active())
        {
          flow_dbg_ << "fastflower(" << name_ << ", " << flow->get_flow_id()
              << "): tail delaying for " << delay << ", making up " << bytes
              << " bytes \n";

        }
        //clean up things
        numflows_--;

        if (numflows_ < 0)
        {
          sst_throw(ssterror, "fastflower_component::model - tail - numflows is less than zero");
        }
        current_bw_ -= flow->get_current_headbw();

        if (current_bw_ < 0)
        {
          if(current_bw_ < BW_MARGIN && current_bw_ > -BW_MARGIN){  //correct for round-off error
            current_bw_ = 0;
          }else
            sst_throw(ssterror, "fastflower_component::model - bw uh oh at update current_bw for flow %s, current head is %f, ended up with %f",
                flow->to_string().c_str(), flow->get_current_headbw(), current_bw_);
        }
        flow->pop_head();

        flowevent* f = new flowevent;
        f->time_ = now();
        f->numflows_ = numflows_;
        double over = std::max(0.0, current_bw_ - capacity_bw_);
        f->overcommit_ = over;
        history_.push_front(f);

        //manage history
        while (history_.size() > HISTORY_SIZE)
        {
          flowevent* f = history_.back();
          delete f;
          history_.pop_back();
        }

        // forward(fmsg, delay + unit_delay_);
        return timestamp(delay) + unit_delay_;
      }
    }

  }
}
