/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_FASTFLOWER_JECTOR_H_
#define SSTMAC_HARDWARE_FASTFLOWER_JECTOR_H_

#include <sstmac/hardware/network/switches/entryexit.h>
#include <sstmac/hardware/fastflower/fastflower_component.h>
#include <sstmac/hardware/fastflower/fastflower_switch.h>

namespace sstmac
{
  namespace hw
  {

    class fastflower_injector : virtual public injector,
        virtual public fastflower_component
    {

    protected:
      fastflower_injector(sim_parameters::ptr params,
          fastflower_switch::ptr par) :
        fastflower_component("injector")
      {
        double bw = params->get_bandwidth_param("network_injector_unit_bw");
        double cbw =
            params->get_bandwidth_param("network_injector_capacity_bw");
        timestamp t = params->get_time_param("network_injector_latency");

        init_params(bw, cbw, t);

        parent_ = par;

      }

      fastflower_switch::ptr parent_;

    public:
      typedef boost::intrusive_ptr<fastflower_injector> ptr;

      virtual std::string
      to_string() const
      {
        return "fastflower_injector";
      }

      virtual
      ~fastflower_injector()
      {
      }

      static ptr
      construct(sim_parameters::ptr params, fastflower_switch::ptr par)
      {
        return ptr(new fastflower_injector(params, par));
      }

      virtual void
      handle(const sst_message::ptr &msg)
      {
        send(0, 0, msg);
      }

      void
      send(long nbytes, long bytes_offset, const sst_message::ptr &msg)
      {
        fastflow_message::ptr fmsg = boost::dynamic_pointer_cast<
            fastflow_message>(msg);

        if (!fmsg)
        {
          sst_throw(ssterror, "fastflower_switch::handle - can't cast %s to a fastflow_message", msg->to_string().c_str());
        }

        bool re;
        timestamp d = model(fmsg, re);
        if (re)
          schedule(now() + d, ptr(this), fmsg);
        else
          schedule(now() + d, parent_, fmsg);
      }

      virtual void
      finish()
      {
        parent_ = 0;
        injector::finish();
      }

    };

    class fastflower_ejector : public ejector, public fastflower_component
    {

    protected:
      fastflower_ejector(sim_parameters::ptr params,
          fastflower_switch::ptr par, nodeaddress::ptr addr) :
        fastflower_component("ejector")
      {
        double bw = params->get_bandwidth_param("network_ejector_unit_bw");
        double cbw = params->get_bandwidth_param("network_ejector_capacity_bw");
        timestamp t = params->get_time_param("network_ejector_latency");

        init_params(bw, cbw, t);

        parent_ = par;

        addr_ = addr;
      }

      fastflower_switch::ptr parent_;

      nodeaddress::ptr addr_;

    public:
      typedef boost::intrusive_ptr<fastflower_ejector> ptr;

      virtual std::string
      to_string() const
      {
        return "fastflower_ejector";
      }

      virtual
      ~fastflower_ejector()
      {
      }

      static ptr
      construct(sim_parameters::ptr params, fastflower_switch::ptr par,
          nodeaddress::ptr addr)
      {
        return ptr(new fastflower_ejector(params, par, addr));
      }

      virtual void
      handle(const sst_message::ptr &msg)
      {

        fastflow_message::ptr fmsg = boost::dynamic_pointer_cast<
            fastflow_message>(msg);

        if (!fmsg)
        {
          sst_throw(ssterror, "fastflower_component::handle - can't cast %s to a fastflow_message", msg->to_string().c_str());
        }

        bool re;
        timestamp d = model(fmsg, re);
        if (re)
          schedule(now() + d, ptr(this), msg);
        else
          schedule(now() + d, recvhandles_[fmsg->orig()->toaddr()], fmsg);
      }

      virtual void
      finish()
      {
        parent_ = 0;
        ejector::finish();
      }

    };

  }
}

#endif
