/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/fastflower/fastflower_nic.h>
#include <sstmac/hardware/fastflower/fastflower_memory.h>
#include <sstmac/hardware/network/switches/entryexit.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("fastflower | fastflower_nic", networkinterface, fastflower_nic,
        "implements a nic that models messages as a flow");

    void
    fastflower_nic::init(const node::ptr &parent,
        const interconnect::ptr& interconn)
    {
      if (!parent->get_id())
      {
        return; //just a template nic
      }
      mem_ = safe_cast(parent->get_mem(), fastflower_memory);
      tailcb_ = msg_callback(ptr(this), &fastflower_nic::tail_going_out);
      networkinterface::init(parent, interconn);
    }

    void
    fastflower_nic::connect(connection_type_t ty, const connectable::ptr& mod)
    {
      switch (ty)
      {
      case injection:
      {
        inj_ = safe_cast(mod, fastflower_injector);

        injcb_ = msg_callback(inj_, &injector::handle);
        break;
      }
      case ejection:
      {

        break;
      }
      default:
        connectable_type_invalid(ty);
      }
    }

    void
    fastflower_nic::finish_recv_ack(const sst_message::ptr& netmsg)
    {
        fastflow_message::ptr fmsg = safe_cast(netmsg, fastflow_message);
        fmsg->get_flow()->done();
    }

    void
    fastflower_nic::finish_recv_req(const sst_message::ptr& netmsg)
    {
        fastflow_message::ptr fmsg = safe_cast(netmsg, fastflow_message);
        fmsg->get_flow()->done();
    }

#if 0
    void
    fastflower_nic::handle(const sst_message::ptr& netmsg)
    {

      fastflow_message::ptr fmsg = safe_cast(netmsg, fastflow_message);

      SSTMAC_DEBUG << "fastflower_nic::handle: " << fmsg->to_string() << "...";

      // here we check for messages that are really just signals
      // that should be forwarded to the node and not modeled,
      // or RDMA control messages
      sst_message::ptr orig_msg = fmsg->get_parent_msg();
      switch (orig_msg->get_type())
      {
      case sst_message::RDMA_GET_REQ:
      {
        SSTMAC_DEBUG << "RDMA_GET_REQ, turning around \n";
        orig_msg->mutate(sst_message::RDMA_GET_REQ_TO_RSP);
        send(orig_msg);
        fmsg->get_flow()->done();
        break;
      }

      case sst_message::RDMA_PUT_RSP:
        SSTMAC_DEBUG << "RDMA_PUT_RSP, forwarding to node \n";
        parent_->handle(orig_msg);
        fmsg->get_flow()->done();
        break;
      case sst_message::SEND_ACK:
        SSTMAC_DEBUG << "SEND_ACK, forwarding to node \n";
        parent_->handle(orig_msg);
        fmsg->get_flow()->done();
        break;
      default:
        SSTMAC_DEBUG << "something arriving, sending to memory \n";
        recv(netmsg);
        break;
      }
    }
#endif

    void
    fastflower_nic::recv(const sst_message::ptr& netmsg)
    {
        mem_->do_access(netmsg, memdonecb_); //this is going to come back to me at mem_done
    }

    void
    fastflower_nic::mem_done(const sst_message::ptr &msg)
    {

      fastflow_message::ptr fmsg = safe_cast(msg, fastflow_message);

      SSTMAC_DEBUG << "fastflower_nic(" << parent_->get_id()->to_string()
            << "): mem_done...";
      //check if this thing is done
      if (fmsg->is_head() && !fmsg->get_headinfo()->is_flow_)
      {
        SSTMAC_DEBUG << "small packet arrived \n";
        parent_->handle(fmsg->orig());
      }
      else if (fmsg->is_tail())
      {

        sst_message::ptr orig = fmsg->orig();
        long bytes = orig->get_byte_length();
        bytes_recv_[orig] += fmsg->get_flow()->get_byte_length();

        SSTMAC_DEBUG << "tail arrived with " << fmsg->get_flow()->get_byte_length() << " for a total of " <<
            bytes_recv_[orig] << " and need " << bytes << " to complete \n";

        if (bytes_recv_[orig] >= bytes)
        {
          parent_->handle(orig);
          bytes_recv_.erase(orig);
        }

        fmsg->get_flow()->done();
      }
      else
      {
        SSTMAC_DEBUG << "head arrived \n";
      }

    }

    void
    fastflower_nic::tail_going_out(const sst_message::ptr &msg)
    {
      SSTMAC_DEBUG << "fastflower_nic(" << parent_->get_id()->to_string()
            << "): tail going out, acking if necessary \n";
      fastflow_message::ptr fmsg = safe_cast(msg, fastflow_message);

      if (fmsg->orig()->get_needs_ack())
      {
        sst_message::ptr sendack = fmsg->orig()->clone(
            sst_message::SEND_ACK);
        parent_->schedule(now(), parent_, sendack);
      }
      inj_->handle(msg);
    }

    void
    fastflower_nic::send(const sst_message::ptr &payload)
    {
      SSTMAC_DEBUG << "fastflower_nic(" << parent_->get_id()->to_string()
            << "): send ...";
      fastflow::ptr flow = fastflow::construct(payload);
      if (flow->get_tail())
      {
        SSTMAC_DEBUG << "full flow \n";
        mem_->do_access(flow->get_head(), injcb_);
        mem_->do_access(flow->get_tail(), tailcb_);
      }
      else //it's just a head
      {
        SSTMAC_DEBUG << "small message \n";
        mem_->do_access(flow->get_head(), tailcb_);
      }

    }
  }
}
