/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/fastflower/fastflower_switch.h>
#include <sstmac/hardware/fastflower/fastflower_jector.h>
#include <sstmac/hardware/fastflower/fastflower_link.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("fastflower", networkswitch, fastflower_switch);

    int fastflower_switch::max_paths_;

    void
    fastflower_switch::init_factory_params(const sim_parameters::ptr& params)
    {
      discrete_network_switch::init_factory_params(params);

      double ubw = params->get_bandwidth_param("network_switch_unit_bw");
      double cbw = params->get_bandwidth_param("network_switch_capacity_bw");
      timestamp t = params->get_time_param("network_switch_latency");
      init_params(ubw, cbw, t);

      link_latency_ = params->get_time_param("network_link_latency");

      max_paths_ = params->get_optional_int_param("network_switch_max_paths", 10);
    }

    networkswitch::ptr
    fastflower_switch::clone(router::ptr route, nodeaddress::ptr addr) const
    {

      ptr ret = new fastflower_switch;
      clone_into(ret, addr, route);
      ret->init_members(const_ptr(this));
      return ret;
    }

    void
    fastflower_switch::init_members(const fastflower_switch::const_ptr &sw)
    {
      discrete_network_switch::init_members(sw);
      init_params(sw->unit_bw_, sw->capacity_bw_, sw->unit_delay_);
    }

    injector::ptr
    fastflower_switch::get_new_injector(sim_parameters::ptr params)
    {
      ptr self(this);
      return fastflower_injector::construct(params, self);
    }

    ejector::ptr
    fastflower_switch::get_new_ejector(sim_parameters::ptr params)
    {
      ptr self(this);
      return fastflower_ejector::construct(params, self, my_addr_);
    }

    void
    fastflower_switch::connect(int dim, int port, connection_type_t ty,
        const connectable::ptr &comp, double weight, int red)
    {
      SSTMAC_DEBUG << "fastflower_switch::connect(dim:" << dim << ", port:"
            << port << ", " << port << ", " << ty << ", " << comp->to_string()
            << ", " << weight << ", " << red << "\n";
      switch (ty)
      {
      case network_link:
      {
        fastflower_component::ptr sw = safe_cast(comp, fastflower_component);
        fastflower_link::ptr link = fastflower_link::construct(ptr(this), sw,
            link_bw_ * weight, link_bw_ * red * weight, link_latency_);
        discrete_network_switch::connect(dim, port, ty, link, weight, red);
        break;
      }
      default:
        discrete_network_switch::connect(dim, port, ty, comp, weight, red);
      }

    }

    int
    fastflower_switch::get_queue_length(int port) const
    {
      fastflower_component::const_ptr comp = boost::dynamic_pointer_cast<
          const fastflower_component>(outports_.at(port));

      if (!comp)
      {
        sst_throw(ssterror, "fastflower_switch::get_queue_length - can't cast outport to fastflower_component");
      }

      int ret = comp->get_current_bw() * 1e-6
          / networkswitch::packet_length_bytes_;

      SSTMAC_DEBUG << "fastflower_switch::get_queue_length - bw is "
            << comp->get_current_bw() << ", returning " << ret << "\n";

      if (ret < 0)
      {
        sst_throw(ssterror, "fastflower_switch::get_queue_length - something wrong, negative bandwidth");
      }

      return ret;
    }

    void
    fastflower_switch::handle(const sst_message::ptr &msg)
    {
      fastflow_message::ptr fmsg =
          boost::dynamic_pointer_cast<fastflow_message>(msg);

      if (!fmsg)
      {
        sst_throw(ssterror, "fastflower_component::handle - can't cast %s to a fastflow_message", msg->to_string().c_str());
      }

      SSTMAC_DEBUG << "fastflower_switch(" << my_addr_->to_string()
            << "): handling " << msg->to_string() << "...";

      bool re;
      timestamp d = model(fmsg, re);
      if (re)
        schedule(now() + d, ptr(this), msg);
      else
      {

        if (router_->needs_congestion_info())
        {
          sst_throw(unimplementederror, "fastflower_switch::forward - congestion info not implemented yet");
        }

        fastflow::ptr flow = fmsg->get_flow();
        SSTMAC_DEBUG << "flow(" << flow->get_flow_id() << ")...";
        if (fmsg->is_head())
        {
          SSTMAC_DEBUG << "head...";
          int multi = std::min(max_paths_, router_->get_number_paths(fmsg));

          long bytes = flow->get_byte_length();
          multi = std::min(multi,
              (int) (((double) bytes / MIN_FLOW_SIZE) + 0.5));

          if (multi > 1 && fmsg->get_headinfo()->is_flow_)
          {

            flow->get_saved_route()->splitid_ = my_addr_->unique_id();

            SSTMAC_DEBUG << "spawning " << multi << " flows \n";
            for (int i = 0; i < multi; i++)
            {

              int dim;
              int dir;


              //router_->route(fmsg->rinfo(), dim, dir);

              long newbytes = std::ceil((double) bytes / (double) multi);
              fastflow_multipath::ptr f = fastflow_multipath::construct(
                  fmsg->orig(), newbytes);


              f->update_unit_bw(flow->get_unit_bw() / multi);

              interconnect_message::ptr tosend;
              if (fmsg->is_head())
              {
                tosend = f->get_head();
              }
              else
              {
                tosend = f->get_tail();
              }
              tosend->init_routing(fmsg->rinfo()->clone());
              long port = router_->route(tosend);
              f->get_saved_route()->add_to_route(my_addr_->unique_id(), port);
              flow->get_saved_route()->split_.push_back(f);

              schedule(now() + d, outports_[port], tosend);
            }
          }
          else
          {

            SSTMAC_DEBUG << "just forwarding \n";
            int dim;
            int dir;

            long port = router_->route(fmsg);

            flow->get_saved_route()->add_to_route(my_addr_->unique_id(), port);

            schedule(now() + d, outports_[port], fmsg);
          }
        }
        else
        {
          SSTMAC_DEBUG << "tail...";
          fastflow::saved_route* route = flow->get_saved_route();

          if (route->splitid_ == my_addr_->unique_id())
          {
            std::vector<fastflow::ptr> flows = route->split_;

            SSTMAC_DEBUG << "splitting into " << flows.size() << " \n";

            for (int i = 0; i < flows.size(); i++)
            {
              //int dim, dir;
              long port;
              fastflow::saved_route* r = flows[i]->get_saved_route();
              if (!r->switch_in_route(my_addr_->unique_id(), port))
              {
                sst_throw(ssterror, "fastflower_switch::handle - can't find saved route for tail on multi-flow");
              }

              schedule(now() + d, outports_[port], flows[i]->get_tail());
            }

            flow->done();
          }
          else
          {
            SSTMAC_DEBUG << "forwarding \n";
            long port;
            if (!route->switch_in_route(my_addr_->unique_id(), port))
            {
              sst_throw(ssterror, "fastflower_switch::handle - can't find saved route for tail");
            }

            schedule(now() + d, outports_[port], fmsg);
          }
        }
      }
    }

  }
}

