#include <sstmac/hardware/memory/mc_memorymodel.h>
#include <sstmac/software/libraries/compute/mem_access.h>

namespace sstmac
{
  namespace hw
  {

    using namespace sstmac::sw;

    mc_memorymodel::mc_memorymodel(
        double mem_bw, timestamp mem_lat,
        double proclink_bw, timestamp proclink_lat,
        int num_mc, int num_socket) :
        num_mc_(num_mc), num_socket_(num_socket)
    {

      for (int i = 0; i < num_socket_ * num_mc_; i++)
        {
          mc_cnt_[i] = 0;
          memlinks_[i] = link::construct(mem_bw, mem_lat);
        }

      //assumes point-to-point between sockets (common)
      for (int i = 0; i < num_socket_; i++)
        {

          for (int j = 0; j < num_socket_; j++)
            {
              if (i != j)
                proclinks_[i + j * num_socket_] = link::construct(proclink_bw,
                    proclink_lat);
            }
        }
    }

    void
    mc_memorymodel::handle(const sst_message::ptr &msg)
    {
      if (msg->get_type() == sst_message::FAILURE)
        {
          failure_occurred(msg, ptr(this));
          return;
        }
      mem_access::ptr acc = boost::dynamic_pointer_cast<mem_access>(msg);

      if (!acc)
        {
          throw ssterror(
              "mc_memorymodel:: retiring an access with a non mem_access message");
        }

      int sock, mc;
      uli dstthread = msg->dest_thread(parent_node_->get_os());
      uli srcthread = msg->source_thread(parent_node_->get_os());
      get_location(dstthread, sock, mc);
      memlinks_[sock * num_mc_ + mc]->access_done();
      int src = sched_->get_socket(srcthread);
      if (src != sock)
        {
          proclinks_[src + sock * num_socket_]->access_done();
        }

      done_->handle(msg);
    }

    void
    mc_memorymodel::get_location(long threadid, int &sock, int &mc)
    {
      if (threadid >= 0)
        {
          sock = sched_->get_socket(threadid);

          if (threadmcs_.find(threadid) == threadmcs_.end())
            {
              threadmcs_[threadid] = mc_cnt_[sock];
              mc_cnt_[sock] = (mc_cnt_[sock] + 1) % num_mc_;
            }
          mc = threadmcs_[threadid];
        }
      else
        {
          sock = 0;
          mc = 0;
        }
    }

    void
    mc_memorymodel::access(const sst_message::ptr &msg)
    {
        use_me();

        mem_access::ptr acc = boost::dynamic_pointer_cast<mem_access>(msg);

        if (!acc)
        {
            throw ssterror("mc_memorymodel:: trying to do an access with a non mem_access message");
        }

        mem_bytes_->record(msg->get_byte_length());

        int sourcesock, destsock;
        int mc, temp;

        get_location(acc->srcthread_, sourcesock, temp);
        get_location(acc->dstthread_, destsock, mc);

        timestamp delta_t = get_delta_t(msg, sourcesock, destsock, mc);

        schedule(now() + delta_t, ptr(this), msg);
    }

    timestamp
    mc_memorymodel::get_delta_t(const sst_message::ptr& msg,
                                int sourcesock, int destsock,
                                int mc)
    {
        timestamp delta_t(0);
        if (destsock != sourcesock)
        {
            SSTMAC_DEBUG << "mc_memorymodel : proc link " << (sourcesock
                  + destsock * num_socket_) << "\n";
            delta_t += proclinks_[sourcesock + destsock * num_socket_]->new_access(now(),
                msg->get_byte_length());
        }

        SSTMAC_DEBUG << "mc_memorymodel: doing access of "
              << msg->get_byte_length() << " at socket " << destsock << " at mc "
              << mc << "\n";

        delta_t += memlinks_[destsock * num_mc_ + mc]->new_access(now(),
            msg->get_byte_length());
        return delta_t;
    }
  }
} /* namespace sstmac */
