#ifndef sstmac_hardware_memory_TRAIN_MEMORYMODEL_CC
#define sstmac_hardware_memory_TRAIN_MEMORYMODEL_CC

#include <sstmac/hardware/memory/train_memorymodel.h>

namespace sstmac {
    namespace hw {

debug memoryrouter::dbg_;

SSTRegister("train",memorymodel,train_memorymodel,
            "Implements a train memory model that models memory congestion as a packet-flow link");

void
memoryrouter::handle_network(const packettrain::ptr& train)
{
    if (train->toaddr()->equals(my_addr_)){
        // this has reached its destination!
        // pass it on up to the train station method
        // and give it to the os whenever it is done
        trainstation::handle(train);
    }
    else {
        injector_->send(train);
    }
}

memoryrouter::memoryrouter(const nodeaddress::ptr &addr, const eventhandler::ptr &handler)
    : trainstation(handler), my_addr_(addr)
{
    if (!dbg_.is_real()){
        dbg_ = debug("<debug> memory");
    }
}

memoryrouter::ptr
memoryrouter::construct(const nodeaddress::ptr& addr, const eventhandler::ptr& handler)
{
    return new memoryrouter(addr, handler);
}

void
memoryrouter::handle(const sst_message::ptr &msg)
{
    packettrain::ptr train = safe_cast(msg, packettrain);
    SSTMAC_DEBUG << "memoryrouter::handle: "
                 << train->orig()->to_string() << "\n";
    if (train->orig()->is_network_message()){
        handle_network(train);
    }
    else { //this is local - just send to os
        trainstation::handle(train);
    }
}

void
memoryrouter::finish()
{
    if (finished_)
        return;

    trainstation::finish();

    injector_->finish();
    injector_ = 0;
}

void
memoryrouter::set_injector(const trainpath::ptr& link)
{
    injector_ = link;
}

train_memorymodel::train_memorymodel(
    double freq_hz,
    uint64_t latency_num_cycles,
    double bytes_per_cycle,
    bool enable_multipath,
    bool ignore_congestion,
    const output_bw_arbitrator::ptr& arb)
     :   router_(0), to_router_(0),
       freq_hz_(freq_hz),
       cas_latency_(latency_num_cycles / freq_hz),
       bytes_per_cycle_(bytes_per_cycle),
       multipath_(enable_multipath),
       congestion_(ignore_congestion),
       arb_(arb)
{
}

void
train_memorymodel::init(const nodePtr &parent_node, const computescheduler::ptr &sched)
{
    double payload_fraction = 1.0; //assume all bytes are payload - ignore ECC
    memorymodel::init(parent_node, sched);
    std::string link_name = sstprintf("node(%6ld)", parent_node->get_id()->unique_id());
    router_ = memoryrouter::construct(parent_node->get_id(), parent_node->get_msg_handler());
    to_router_ = trainlink::construct(router_, cas_latency_,
                                      bytes_per_cycle_, freq_hz_, payload_fraction,
                                      multipath_, congestion_,
                                      arb_, "memory", link_name);
}

train_memorymodel::ptr
train_memorymodel::construct(const sim_parameters::ptr &params)
{
    /** sstkeyword { gui=8; docstring=$bytes_per_cycle } */
    double bytes_per_cycle = params->get_double_param("memory_train_bytes_per_cycle");
    /** sstkeyword { gui=10; docstring=$train_latency; } */
    uint64_t latency_cycles = params->get_long_param("memory_train_latency_cycles");
    /** sstkeyword { gui=1000MHz; docstring=$train_freq; } */
    double freq_hz = params->get_freq_param("memory_train_freq");
    /** sstkeyword { docstring=$ignore_congestion; } */
    bool ignore_congestion = params->get_optional_bool_param("memory_train_ignore_congestion", false);
    /** sstkeyword { docstring=$enable_multipath; } */
    bool enable_multipath = params->get_optional_bool_param("memory_train_enable_multipath", false);
    /** sstkeyword { docstring=$train_arbitrator; } */
    output_bw_arbitrator::ptr arb = output_bw_arbitrator_factory
            ::get_optional_param("memory_train_arbitrator", "infinite_buffer", params);

    return new train_memorymodel(freq_hz,
                                 latency_cycles,
                                 bytes_per_cycle,
                                 enable_multipath,
                                 ignore_congestion,
                                 arb);
}

train_memorymodel::~train_memorymodel()
{
}

void
train_memorymodel::set_eventmanager(const eventmanager::ptr &m)
{
    memorymodel::set_eventmanager(m);
    router_->set_eventmanager(m);
    to_router_->set_eventmanager(m);
}

void
train_memorymodel::set_injector(const trainpath::ptr &sw)
{
    router_->set_injector(sw);
}

void
train_memorymodel::handle(const sst_message::ptr &msg)
{
    sst_throw(illformederror, "train_memorymodel::handle: should not be called");
}

void
train_memorymodel::do_access(const sst_message::ptr& msg)
{

    packettrain::ptr train = safe_cast(msg, packettrain);
    SSTMAC_DEBUG << "train_memorymodel::do_access: "
                 << train->to_string() << "\n";
    if (train->is_single_packet()){
        timestamp t = now() + to_router_->link_latency();
        double bw = bytes_per_cycle_ * freq_hz_;
        // set the bandwidth maximum to my bandwidth
        train->throttle_bw(bw);
        schedule(t, router_, msg);
    }
    else {
        //do a full model of the access
        to_router_->send(train);
    }
}

void
train_memorymodel::access(const sst_message::ptr &msg)
{
    packettrain::ptr train = packettrain::construct(msg);
    do_access(train);
}


void
train_memorymodel::finish()
{
    if (finished_)
        return;

    memorymodel::finish();

    router_->finish();
    to_router_->finish();
    router_ = 0;
    to_router_ = 0;


}

    }
}




#endif // TRAIN_MEMORYMODEL_CC
