/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/interconnect/switchinterconnect.h>
#include <sstmac/hardware/network/switches/packet/packetswitch.h>
#include <sstmac/hardware/network/switches/cycleaccurate/cycleaccurateswitch.h>
#include <sstmac/hardware/network/switches/circuit/circuitswitch.h>
#include <sstmac/hardware/network/topology/topology.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("switch", interconnect, switchinterconnect,
        "Models the network interconnect as a topology of connected switches");

    switchinterconnect::switchinterconnect() :
      vis_display("interconnect")
    {
    }

    void
    switchinterconnect::init_factory_params(const sim_parameters::ptr &params)
    {

      /** This builds the nodes */
      interconnect::init_factory_params(params);
      vis_display::init_factory_params(params);
    }

    switchinterconnect::ptr
    switchinterconnect::construct(const sim_parameters::ptr& params)
    {
      return new switchinterconnect();
    }

    networkswitch::ptr
    switchinterconnect::get_template_switch(const sim_parameters::ptr& params, const topology::ptr& top)
    {

      /**
        sstkeyword {
            gui=train;
            docstring=The type of congestion model to use for modeling switch arbitration;
        }
      */
      networkswitch::ptr switchtemp = networkswitch_factory::get_param(
          "network_switchtype", params, top);

      return switchtemp;
    }

    void
    switchinterconnect::set_eventmanager(const eventmanager::ptr &m)
    {
      SSTMAC_DEBUG << "switchinterconnect::set_eventmanager() \n";
      //this needs to happen before everything else, because it might
      // need to ask the simulator about parallel partitions
      topology_->set_evman(m);

      networkswitch::ptr switchtemp = get_template_switch(m->params_, topology_);

      std::vector<long>* swtoadd = topology_->switches_to_add();

      SSTMAC_DEBUG << "switchinterconnect: creating " << swtoadd->size()
            << " switches \n";
      // create all the switches


      std::vector<long>::iterator swit, swend = swtoadd->end();
      for (swit = swtoadd->begin(); swit != swend; swit++)
      {
        nodeaddress::ptr sw_addr = nodeaddress::convert_from_uniqueid(*swit);
        SSTMAC_DEBUG << "creating switch " << sw_addr->to_string() << "\n";
        networkswitch::ptr sw = switchtemp->clone(
            topology_->create_router(sw_addr, m->params_), sw_addr);
        // sw->init_factory_params(params);
        sw->init(topology_);
        switches_[sw_addr] = sw;
      }

      add_nodes(m->params_);
      init_failure_model(m->params_);


      std::vector<node::ptr>::iterator nodeit, nodeend = nodes_.end();
      for (nodeit = nodes_.begin(); nodeit != nodeend; nodeit++)
      {

        node::ptr the_node = *nodeit;
        nodeaddress::ptr nodeaddr = the_node->get_id();
        connectable::ptr nic = the_node->get_nic();

        int inj_port;
        nodeaddress::ptr injaddr = topology_->node_to_injector_addr(nodeaddr,
            inj_port);
        networkswitch::ptr injsw = switches_[injaddr];
        injsw->connect(nodeaddr, inj_port, connectable::injection, nic);

        int ej_port;
        nodeaddress::ptr ejaddr = topology_->node_to_ejector_addr(nodeaddr,
            ej_port);
        networkswitch::ptr ejsw = switches_[ejaddr];
        ejsw->connect(nodeaddr, ej_port, connectable::ejection, nic);
      }

      SSTMAC_DEBUG << "switchinterconnect: created " << swtoadd->size()
            << " gateways switches for " << nodes_.size() << " nodes \n";

      interconnect::set_eventmanager(m);

      topology_->connect_switches(switches_);

      // Now they're connected up, initialize them.
      // Note that connect_switches could have added some switches

      topology::maptype::iterator it, end = switches_.end();
      for (it = switches_.begin(); it != end; it++)
      {
        it->second->initialize();
        it->second->set_eventmanager(m);
      }
    }

    void
    switchinterconnect::finish()
    {
      if (finished_)
        return;

      interconnect::finish();

      topology::maptype::iterator it, end = switches_.end();
      for (it = switches_.begin(); it != end; it++)
      {
        it->second->finish();
      }
    }

    vis::vis_topology::vismaptype
    switchinterconnect::vis_get_nodes()
    {
      vis::vis_topology::vismaptype vismap;
      topology::maptype::iterator it, end = switches_.end();
      for (it = switches_.begin(); it != end; it++)
      {
        vismap[it->first] = it->second;
      }
      return vismap;
    }

    vis::vis_topology::ptr
    switchinterconnect::vis_get_topology()
    {
      return topology_;
    }

  }
}
