/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/interconnect/traininterconnect.h>
#include <sstmac/hardware/packettrain/train_nic.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("train", interconnect, traininterconnect,
        "Models an interconnect that sends messages as a packet train");

    void
    traininterconnect::init_factory_params(const sim_parameters::ptr& params)
    {
      // fix some parameters
      if (params->has_param("network_train_injection_arbitrator"))
      {
        std::string arbtype = params->reread_param(
            "network_train_injection_arbitrator");
        params->add_param_override("network_train_rdma_arbitrator", arbtype);
        params->add_param_override("network_train_udp_arbitrator", arbtype);
      }
      if (params->has_param("network_train_injection_latency"))
      {
        std::string lat = params->reread_param(
            "network_train_injection_latency");
        params->add_param_override("network_train_rdma_injection_latency", lat);
        params->add_param_override("network_train_udp_injection_latency", lat);
      }
      if (params->has_param("network_train_injection_bw"))
      {
        std::string bw = params->reread_param("network_train_injection_bw");
        params->add_param_override("network_train_rdma_injection_bw", bw);
        params->add_param_override("network_train_udp_injection_bw", bw);
      }

      switchinterconnect::init_factory_params(params);
    }

    traininterconnect::ptr
    traininterconnect::construct(const sim_parameters::ptr& params)
    {
      traininterconnect::ptr ret = new traininterconnect();
      return ret;
    }

    std::string
    traininterconnect::label(long node_id) const
    {
      int sw_id;
      long net_id = topology_->node_to_injector_addr(
          nodeaddress::convert_from_uniqueid(node_id), sw_id)->unique_id();

      std::string net_label = topology_->label(net_id);
      std::string node_label = sstprintf("node(%6ld) %s[%2ld]", node_id,
          net_label.c_str(), sw_id);
      return node_label;
    }

    traininterconnect::~traininterconnect()
    {
    }

    void
    traininterconnect::finish()
    {
      if (finished_)
        return;

      interconnect::finish();

      topology::maptype::iterator it, end = switches_.end();
      for (it = switches_.begin(); it != end; it++)
      {
        it->second->finish();
      }
      switches_.clear();

    }

  }
}

