/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_NETWORK_SWITCHES_CYCLEACCURATESWITCH_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_SWITCHES_CYCLEACCURATESWITCH_H_INCLUDED

#include <sstmac/hardware/network/switches/networkswitch.h>

#include <sstmac/hardware/network/switches/orion/ORION_Array.h>
#include <sstmac/hardware/network/switches/orion/ORION_Crossbar.h>
#include <sstmac/hardware/network/switches/orion/ORION_Arbiter.h>
#include <sstmac/hardware/network/switches/orion/ORION_Link.h>

#include <sstmac/common/stats/stat_time_window.h>
#include <sstmac/common/stats/stats_common.h>

#include <sstmac/hardware/network/switches/cycleaccurate/cycleaccuratemessage.h>

namespace sstmac
{
  namespace hw
  {

#define ROUTER_DELAY 0.0000001
#define MAX_GRANTS 8
    //#define ROUTER_DELAY 0

    class cycleaccurateswitch : public discrete_network_switch
    {
    public:
      typedef boost::intrusive_ptr<cycleaccurateswitch> ptr;
      typedef boost::intrusive_ptr<const cycleaccurateswitch> const_ptr;

    protected:

      cycleaccurateswitch();

      class outbuffer : public eventhandler
      {

      protected:
        outbuffer(int pid, cycleaccurateswitch::ptr p, timestamp link_lat,
            double link_bw) :
          pid_(pid), link_lat_(link_lat), link_bw_(link_bw), parent_(p)
        {
          finish_ = timestamp(0);
          outbuff_occ_ = 0;

          total_bytes_ = 0;

          vis_update_ = sst_message::construct();
          first_time_ = true;
        }

        void
        clear_outbuff();

        int pid_;
        timestamp link_lat_;

        double link_bw_;
        long total_bytes_;

        sst_message::ptr vis_update_;

        bool first_time_;

        timestamp finish_;
        cycleaccurateswitch::ptr parent_;

        std::map<int, long> vc_bytes_;

        long outbuff_occ_;
        std::map<timestamp, long> outbuff_occ_clear_;

      public:
        typedef boost::intrusive_ptr<outbuffer> ptr;

        static ptr
        construct(int id, cycleaccurateswitch::ptr p, timestamp link_lat,
            double link_bw)
        {
          return ptr(new outbuffer(id, p, link_lat, link_bw));
        }

        virtual
        ~outbuffer()
        {
        }
        virtual std::string
        to_string() const
        {
          return "cycleaccurateswitch::outbuffer";
        }

        virtual void
        handle(const sst_message::ptr &msg);

        long
        get_occupancy();

      };

      class packetinjector : public injector
      {

      protected:
        packetinjector(sim_parameters::ptr params, cycleaccurateswitch::ptr par)

        {
          MTU_ = params->get_long_param("cycleaccurateswitch_MTU");

          n2r_ = params->get_bandwidth_param(
              "cycleaccurateswitch_bandwidth_n2r");

          parent_ = par;

          next_vc_ = (0);
          dbg_ = parent_->dbg_;
        }

        long MTU_;
        long n2r_;
        long next_vc_;
        debug dbg_;

        static long long MSG_COUNT;

        cycleaccurateswitch::ptr parent_;

      public:
        typedef boost::intrusive_ptr<packetinjector> ptr;

        virtual std::string
        to_string() const
        {
          return "packetinjector";
        }

        virtual
        ~packetinjector()
        {
        }

        static ptr
        construct(sim_parameters::ptr params, cycleaccurateswitch::ptr par)
        {
          return ptr(new packetinjector(params, par));
        }

        void
        send(long nbytes, long byte_offset, const sst_message::ptr &msg);

      };

      class packetejector : public ejector
      {

      protected:
        packetejector(sim_parameters::ptr params, cycleaccurateswitch::ptr par,
            nodeaddress::ptr addr)
        {
          parent_ = par;
          dbg_ = parent_->dbg_;
          addr_ = addr;
        }

        cycleaccurateswitch::ptr parent_;
        debug dbg_;
        nodeaddress::ptr addr_;

        boost::unordered_map<long long, long> recv_bytes_;
        boost::unordered_map<long long, cycleaccuratepayload::ptr> tails_;

      public:
        typedef boost::intrusive_ptr<packetejector> ptr;

        virtual std::string
        to_string() const
        {
          return "packetejector";
        }

        virtual
        ~packetejector()
        {
        }

        static ptr
        construct(sim_parameters::ptr params, cycleaccurateswitch::ptr par,
            nodeaddress::ptr addr)
        {
          return ptr(new packetejector(params, par, addr));
        }

        virtual void
        handle(const sst_message::ptr &msg);

      };

      // ----- parameters ------
      int numVC_;
      int inBuffSize_;
      int outBuffSize_;

      double clockFreq_;

      int mtu;

      timestamp arbLatency_;
      timestamp switchLatency_;
      double switchBandwidth_;

      timestamp link_lat_;

      timestamp packet_timeout_;

      // -----------------------

      sst_message::ptr clockmsg_;
      timestamp clockper_;

      void
      init_buffers();
      void
      init_orion();
      virtual void
      init_params(sim_parameters::ptr params)
      {
      }

      // -------- buffer management --------
      std::map<int, std::queue<cycleaccuratepayload::ptr> > inbuffs_;
      std::map<int, int> inbuffcnt_;
      std::map<int, outbuffer::ptr> outbuffs_;

      std::map<int, int> credits_;

      std::map<nodeaddress::ptr, int, nodeaddress::nodeaddresscomp,
          std::allocator<std::pair<nodeaddress::ptr, int> > > nodebuffers_;

      int nodebufcnt_;

      long totalinbufspace_;
      long totalinbufocc_;

      // -----------------------


      // ---- arbitration ------ //
      void
      arbitrate();

      virtual int
      check_inport(int inport, bool &clock);

      std::map<int, timestamp> inports_free_;
      std::map<int, timestamp> outports_free_; //of the xbar, not the whole switch

      std::map<int, int>::iterator arb_inport_;
      std::map<int, int> arb_vc_;

      // --------- orion stuff -----
      int line_width_;

      ORION_Array* power_buffer_in_;
      ORION_Array* power_buffer_out_;
      ORION_Crossbar* power_xbar_;
      ORION_Arbiter* power_arbiter_;
      bool using_orion_;
      double link_e_per_bm_;
      double link_length_;
      // -------------------------

      double inj_bw_;

      // -------- stats --------
      static stat_time_window_sum::ptr power_stat_;
      static double static_power_;

      static stat_sum::ptr stat_bytes_;

      std::string stat_name_;
      // -------------------------


      virtual injector::ptr
      get_new_injector(sim_parameters::ptr params);

      virtual ejector::ptr
      get_new_ejector(sim_parameters::ptr params);

    public:

      static ptr
      construct(const sim_parameters::ptr& params)
      {
        return new cycleaccurateswitch;
      }

      virtual std::string
      to_string() const
      {
        return "cycleaccurateswitch(" + (my_addr_ ? my_addr_->to_string()
            : "don't know where i am") + ")";
      }

      virtual
      ~cycleaccurateswitch()
      {
      }

      virtual void
      set_inport(const cycleaccurateswitch::ptr& sw, int dim, int dir);

      virtual void
      finish();

      int get_queue_length(int port) const;

      virtual void
      init_factory_params(const sim_parameters::ptr& params);

      /**
       * This gets called after connect
       */
      virtual void
      initialize();

      void
      init_members(const cycleaccurateswitch::const_ptr &sw);

      virtual networkswitch::ptr
      clone(router::ptr route, nodeaddress::ptr addr) const;

      virtual void
      handle(const sst_message::ptr &msg);

      virtual std::string
      get_interprocess_link_name() const
      {
        return "cycleaccurateswitch_latency_r2r";
      }

    };

  }
}
#endif
