/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_NETWORK_SWITCHES_ENTRYEXIT_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_SWITCHES_ENTRYEXIT_H_INCLUDED

#include <sstmac/hardware/common/connection.h>
#include <boost/unordered_map.hpp>

namespace sstmac
{
  namespace hw
  {

    class injector :
        public connectable
    {

    protected:

      typedef boost::unordered_map<nodeaddress::ptr, eventhandler::ptr> handler_map;
      handler_map sendhandle_;


    public:
      typedef boost::intrusive_ptr<injector> ptr;

      virtual std::string
      to_string() const
      {
        return "injector";
      }

      virtual
      ~injector()
      {
      }

      virtual void finish() {
          sendhandle_.clear();
      }

      eventhandler::ptr
      get_sendhandler(const nodeaddress::ptr& addr) const
      {
          handler_map::const_iterator it = sendhandle_.find(addr);
          if (it == sendhandle_.end()){
              sst_throw(valueerror, "caannot find send handler for node address %s",
                        addr->to_string().c_str());
          }
          return it->second;
      }

      void connect(const nodeaddress::ptr& addr, connection_type_t ty, const connectable::ptr &mod)
      {
        switch(ty)
        {
            case injection:
                set_sendhandler(mod, addr);
                break;
            default:
                connectable_type_invalid(ty);
        }
      }

      void
      set_sendhandler(const eventhandler::ptr &hand, nodeaddress::ptr addr)
      {
        sendhandle_[addr] = hand;
      }


      void handle(const sst_message::ptr &msg){
          //fire off the whole message
          send(msg->get_byte_length(), 0, msg);
      }

      /**
        @param nbytes The number of bytes of the given message to send
        @param byte_offset The byte offset of the given message at beginning of send
        @param msg The complete message being sent
      */
      virtual void send(long nbytes, long byte_offset, const sst_message::ptr& msg) = 0;

    };

    class ejector :
        public connectable
    {
    protected:
      boost::unordered_map<nodeaddress::ptr, eventhandler::ptr> recvhandles_;

    public:
      typedef boost::intrusive_ptr<ejector> ptr;

      virtual std::string
      to_string() const
      {
        return "ejector";
      }

      virtual
      ~ejector()
      {
      }

      virtual void finish() {
          recvhandles_.clear();
      }

      void connect(const nodeaddress::ptr& addr, connection_type_t ty, const connectable::ptr &mod)
      {
        switch(ty)
        {
            case ejection:
                set_recvhandler(mod, addr);
                break;
            default:
                connectable_type_invalid(ty);
        }
      }

      void
      set_recvhandler(const eventhandler::ptr &hand, nodeaddress::ptr addr)
      {
        recvhandles_[addr] = hand;
      }

    };

  }
}

#endif
