/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/switches/networkswitch.h>
#include <sstmac/hardware/network/interconnect/switchinterconnect.h>
#include <sstmac/hardware/network/switches/cycleaccurate/cycleaccurateswitch.h>

ImplementFactory(sstmac::hw::networkswitch)
;

namespace sstmac
{
  namespace hw
  {

    debug discrete_network_switch::dbg_;
    int networkswitch::packet_length_bytes_ = 0;

    coordinates
    networkswitch::get_coords() const
    {
      return top_->get_switch_coords(my_addr_->unique_id());
    }

    void
    networkswitch::init_factory_params(const sim_parameters::ptr &params)
    {
      /**
       sstkeyword {
       docstring=The actual physical packet size.ENDL
       This is used primarily by adaptive routing schemesENDL
       in determining queue length.;
       }
       */
      packet_length_bytes_ = params->get_optional_byte_length_param(
          "packet_length", 100);
    }

    void
    discrete_network_switch::connect(const nodeaddress::ptr& addr,
        int port_offset, connection_type_t ty, const connectable::ptr &comp)
    {
      SSTMAC_DEBUG << "discrete_network_switch::connect(" << addr->unique_id()
            << ", " << port_offset << ", " << ty << ", " << comp->to_string()
            << "\n";
      switch (ty)
      {
      case injection:
      {
        injector::ptr inj = get_new_injector(sstmac_env::params);
        inj->connect(addr, ty, comp);
        comp->connect(ty, inj);
        break;
      }
      case ejection:
      {
        ejector::ptr ej = get_new_ejector(sstmac_env::params);
        long port = router_->convert_to_outport(topology::EJECT, port_offset);
        outports_[port] = ej;
        ej->connect(addr, ty, comp);
        comp->connect(ty, ej);
        break;
      }
      default:
        connectable_type_invalid(ty);
      }
    }

    void
    discrete_network_switch::connect(int dim, int port, connection_type_t ty,
        const connectable::ptr &comp, double weight, int red)
    {
      SSTMAC_DEBUG << "discrete_network_switch::connect(dim:" << dim
            << ", port:" << port << ", " << port << ", " << ty << ", "
            << comp->to_string() << ", " << weight << ", " << red << "\n";
      switch (ty)
      {
      case network_link:
      {
        // discrete_network_switch::ptr sw = safe_cast(comp, discrete_network_switch);
        connect(dim, port, comp, weight, red);
        break;
      }
      default:
        connectable_type_invalid(ty);
      }
    }

    void
    discrete_network_switch::connect(int dim, int dir, eventhandler::ptr sw,
        double weight, int redund)
    {
      SSTMAC_DEBUG << "discrete_network_switch::connect(dim:" << dim
            << ", dir:" << dir << ", " << sw->to_string() << ", " << weight
            << ", " << redund << "\n";

      if (link_bw_ < 1)
        abort();

      long port = router_->convert_to_outport(dim, dir);

      outports_[port] = sw;
      portbw_[port] = link_bw_ * weight;
      redund_[port] = redund;

      cycleaccurateswitch::ptr cyc_sw = boost::dynamic_pointer_cast<
          cycleaccurateswitch>(sw);

      if (cyc_sw)
      {
        //tell the switch at my output port to connect the input port
        cyc_sw->set_inport(cyc_sw, dim, dir);
      }

    }

    void
    discrete_network_switch::init_factory_params(
        const sim_parameters::ptr &params)
    {
      link_bw_ = params->get_bandwidth_param("network_bandwidth_link");
      networkswitch::init_factory_params(params);
    }

    std::vector<nodeaddress::ptr>
    discrete_network_switch::get_connected_switches() const
    {
      std::vector<nodeaddress::ptr> ret;

      outport_map::const_iterator it, end = outports_.end();
      for (it = outports_.begin(); it != end; it++)
      {
        networkswitch::ptr net = boost::dynamic_pointer_cast<networkswitch>(
            it->second);
        if (net)
        {
          ret.push_back(net->get_addr());
        }
      }
      return ret;
    }

    std::vector<int>
    discrete_network_switch::get_connected_redund() const
    {
      std::vector<int> ret;

      outport_map::const_iterator it, end = outports_.end();

      for (it = outports_.begin(); it != end; it++)
      {
        networkswitch::ptr net = boost::dynamic_pointer_cast<networkswitch>(
            it->second);
        if (net)
        {
          ret.push_back(redund_.at(it->first));

        }

      }
      return ret;
    }
  }
}
