/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

// butterfly.h: Interface for butterfly networks.
//
// Based on butterfly.c by Jim Schutt
// Adapted by Curtis Janssen <cljanss@ca.sandia.gov>
#ifndef SSTMAC_HARDWARE_NETWORK_TOPOLOGY_butterfly_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_TOPOLOGY_butterfly_H_INCLUDED

#include <sstmac/hardware/network/topology/regular_topology.h>

namespace sstmac
{
  namespace hw
  {

    class abstract_butterfly :
        public regular_topology
    {
        protected:
            int kary_;
            int nfly_;
            long nswitches_per_col_;

            abstract_butterfly(int kary, int nfly, int nps);

            void
            compute_switch_coords(long uid, coordinates& coords) const;

            long
            get_switch_id(const coordinates &coords) const;

        public:
          virtual long
          num_stages() const
          {
            return nfly_;
          }

          int
          kary() const {
            return kary_;
          }

          int
          nfly() const {
            return nfly_;
          }

          long
          num_switches_per_col() const {
            return nswitches_per_col_;
          }

          virtual long
          num_leaf_switches() const
          {
            return nswitches_per_col_;
          }

          int64_t
          diameter() const
          {
            return nfly_ + 1;
          }

    };


    class butterfly :
        public abstract_butterfly
    {
    private:
      long last_col_index_start_;

    protected:
      butterfly(int kary, int nfly, int nps);

      virtual void
      compute_switch_coords(long uid, coordinates& coords) const;

    public:
      static const int up_dimension;

      typedef boost::intrusive_ptr<butterfly> ptr;

      static butterfly::ptr
      construct(sim_parameters::ptr params);

      virtual
      ~butterfly()
      {
      }

      long
      last_col_index_start() const {
        return last_col_index_start_;
      }

      virtual int
      radix() const
      {
        return kary_;
      }

      virtual long
      num_switches() const
      {
        return nswitches_per_col_ * nfly_;
      }

      int
      get_productive_dir(
        int& dim,
        const coordinates &src,
        const coordinates &dst) const;

      void
      minimal_route_to_coords(
        const coordinates &current_coords,
        const coordinates &dest_coords,
        int &dim, int &dir) const;

      int
      minimal_distance(
        const coordinates& current_coords,
        const coordinates& dest_coords
      ) const;

      long
      get_switch_id(const coordinates &coords) const;

      virtual void
      connect_switches(topology::maptype &switches);

      nodeaddress::ptr
      node_to_ejector_addr(
        const nodeaddress::ptr &nodeaddr,
        int &switch_port) const;

      virtual int
      convert_to_outport(int dim, int dir);

      virtual void
      convert_to_dimdir(int outport, int &dim, int&dir);

      virtual std::string
      to_string() const
      {
        return "butterfly";
      }

      std::string
      get_default_router() const {
        return "ordered";
      }

      /**
        Each level in a kary-nfly counts
        as a dimension.
      */
      int ndimensions() const {
        return nfly_;
      }

    };

  }
} //end of namespace sstmac

#endif
