/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */
// crossbar.cc: Implementation of crossbar networks.
//
// Author: Curtis Janssen <cljanss@sandia.gov>
#include <sstream>
#include <sstmac/hardware/network/topology/crossbar.h>
#include <sstmac/common/driver_util.h>
#include <sstmac/hardware/network/switches/networkswitch.h>
#include <sstmac/hardware/packettrain/trainswitch.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("crossbar | xbar", topology, crossbar);

    crossbar::ptr
    crossbar::construct(sim_parameters::ptr params)
    {
      bool errorflag = false;
      std::vector<int> args = get_intvec(
          params->get_param("topology_geometry").c_str(), errorflag);
      int nps = params->get_optional_int_param("network_nodes_per_switch", 1);
      return crossbar::ptr(new crossbar(args[0], nps));
    }

    crossbar::crossbar(long nnode, int nps) :
    regular_topology(nps), nps_(nps), size_(nnode)
    {
    }

    void
    crossbar::compute_switch_coords(long uid, coordinates &coords) const
    {
        coords[0] = uid;
    }

    int
    crossbar::get_productive_dir(
        int &dim,
        const coordinates &src,
        const coordinates &dst) const
    {
        if (dim != 0){
            sst_throw(valueerror,
                "crossbar received non-zero dimension %d", dim);
        }
        return dst[0];
    }

    int
    crossbar::convert_to_outport(int dim, int dir)
    {
      if(dim == topology::EJECT){
          return -(1+dir);
      }
      return dir;
    }

    void
    crossbar::convert_to_dimdir(int outport, int &dim, int&dir)
    {
      if (outport < 0)
        {
          dim = topology::EJECT;
          dir = (-outport) - 1;
        }
      else
        {
          dim = 0;
          dir = outport;
        }
    }

    std::string
    crossbar::name() const
    {
      std::ostringstream ostr;
      ostr << "crossbar(" << num_switches() << ")";
      return ostr.str();
    }

    void
    crossbar::minimal_route_to_coords(
        const coordinates &current_coords,
        const coordinates &dest_coords,
        int &dim, int &dir) const
    {
          dim = 0;
          dir = dest_coords[0];
    }

    void
    crossbar::minimal_route_to_switch(
        const nodeaddress::ptr &current_sw_addr,
        const nodeaddress::ptr &dest_sw_addr,
        int &dim, int &dir) const
    {
        //current switch actually doesn't matter
        dim = 0;
        dir = dest_sw_addr->unique_id();
    }

    int
    crossbar::minimal_distance(const coordinates &current_coords, const coordinates &dest_coords) const
    {
        return 1;
    }

    void
    crossbar::connect_switches(topology::maptype &switches)
    {
      double edge_weight = 1.0;
      int red = 1;
      for (uli i = 0; i < switches.size(); i++)
        {
          nodeaddress::ptr me = nodeaddress::convert_from_uniqueid(i);

          for (uli j = 0; j < switches.size(); j++)
            {
              nodeaddress::ptr them = nodeaddress::convert_from_uniqueid(j);
              if (i != j)
                {
                  switches[me]->connect(0, j, connectable::network_link,
                                        switches[them], edge_weight, red);
                }

            }

        }
    }

  }
} //end of namespace sstmac
