/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

// crossbar.h: Interface for crossbar networks.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>
#ifndef SSTMAC_HARDWARE_NETWORK_TOPOLOGY_CROSSBAR_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_TOPOLOGY_CROSSBAR_H_INCLUDED

#include <sstmac/hardware/network/topology/regular_topology.h>

namespace sstmac
{
  namespace hw
  {

    /** The crossbar network generates a network which connects
     all nodes with only two hops: those to and from the crossbar.
     The inputs for this network can specify the following attributes:

     <table border="1">

     <tr><td>Keyword<td>Type<td>Default<td>Description

     <tr><td><tt>n</tt><td>int<td>256<td>The radix of the crossbar.

     </table>
     */

    class crossbar : public regular_topology
    {
    private:
      int nps_;
      long size_;

    protected:
      crossbar(long nnode, int nps);

      virtual void
      compute_switch_coords(long uid, coordinates& coords) const;

    public:
      typedef boost::intrusive_ptr<crossbar> ptr;

      static crossbar::ptr
      construct(sim_parameters::ptr params);

      virtual
      ~crossbar()
      {
      }

      virtual int
      radix() const
      {
        return num_switches();
      }

      int64_t
      diameter() const {
        return 1;
      }

      std::string
      name() const;

      int
      ndimensions() const {
          return 1;
      }

      int
      ncoords() const {
        return 1;
      }

      long
      num_leaf_switches() const {
          return size_;
      }

      virtual void
      connect_switches(topology::maptype &switches);

      virtual void
      minimal_route_to_switch(
        const nodeaddress::ptr &current_sw_addr,
        const nodeaddress::ptr &dest_sw_addr,
        int &dim, int &dir) const;

      virtual void
      minimal_route_to_coords(
        const coordinates &current_coords,
        const coordinates &dest_coords,
        int &dim, int &dir) const;

      int
      minimal_distance(
        const coordinates& current_coords,
        const coordinates& dest_coords
      ) const;

      int
      get_productive_dir(
        int &dim,
        const coordinates &src,
        const coordinates &dst) const;

      virtual int
      convert_to_outport(int dim, int dir);

      virtual void
      convert_to_dimdir(int outport, int &dim, int&dir);

      long
      get_switch_id(const coordinates& coords) const {
        return coords[0];
      }

      virtual std::string
      to_string() const
      {
        return "crossbartopology";
      }

      virtual long
      num_switches() const
      {
        return size_;
      }

    };

  }
} //end of namespace sstmac

#endif
