/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

// fattree.h: Interface for fat tree networks.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>

#ifndef SSTMAC_HARDWARE_NETWORK_TOPOLOGY_FATTREE_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_TOPOLOGY_FATTREE_H_INCLUDED

#include <sstmac/hardware/network/topology/regular_topology.h>


namespace sstmac
{
  namespace hw
  {

    /** The fattree network generates a k-ary fat tree with l tiers.  In the fat
     tree, not all vertices are nodes. The inputs for this network
     can specify the following attributes:

     <table border="1">

     <tr><td>Keyword<td>Type<td>Default<td>Description

     <tr><td><tt>l</tt><td>int<td>2<td>The number of levels in the fat tree.

     <tr><td><tt>k</tt><td>int<td>2<td>Half the number of links a switch
     element has.

     <tr><td><tt>nndpersw</tt><td>int<td>0<td>The number of nodes per switch.

     <tr><td>width<td>int<td>0<td>The width of edges connecting two switches.

     </table>
     */

    class fattree :
        public regular_topology
    {

    protected:
      int width;

      int l_, k_, numleafswitches_;

      int toplevel_;

      long toplevel_id_start_;

      void
      init_fattree(int l, int k, int nndpersw, int widtharg);

    protected:
      fattree(int l, int k, int nndpersw, int widtharg);

        void
        connect_group(
            maptype& switches,
            int group_stride,
            int down_group_size,
            long down_group_offset,
            int up_group_size,
            long up_group_offset);

        void
        connect_section(
            maptype& switches,
            int group_stride,
            int num_groups_per_section,
            int down_group_size,
            long down_section_offset,
            int up_group_size,
            long up_section_offset);

      virtual void
      compute_switch_coords(long uid, coordinates& coords) const;

    public:

      static const int down_dimension;
      static const int up_dimension;

      typedef boost::intrusive_ptr<fattree> ptr;

      static fattree::ptr
      construct(sim_parameters::ptr params);

      static ptr
      construct(int l, int k, int nps = 1, int wid = 1)
      {
        return ptr(new fattree(l, k, nps, wid));
      }

      virtual
      ~fattree()
      {
      }

      int
      l() const {
        return l_;
      }

      int
      k() const {
        return k_;
      }

      int64_t
      diameter() const {
        return (l_ + 1) * 2;
      }

      std::string
      name() const;

      virtual int
      ndimensions() const {
          //l branches plus one extra
          //that denotes which "redundant" router
          //it is
          return l_ + 1;
      }

      virtual void
      connect_switches(topology::maptype &switches);

      virtual std::string label(long id) const;

      virtual int
      radix() const
      {
        return 2 * k_;
      }

      virtual int
      convert_to_outport(int dim, int dir);

      virtual void
      convert_to_dimdir(int outport, int &dim, int&dir);

      std::string
      get_default_router() const {
        return "fattree";
      }

      virtual long
      num_leaf_switches() const {
        return numleafswitches_;
      }

      virtual long
      num_switches() const
      {
        //there are l-1 levels with numleaf
        //the top level has numleaf / 2
        return numleafswitches_ * (l_ - 1) + numleafswitches_ / 2;
      }

      virtual long
      num_nodes() const
      {
        return numleafswitches_ * nps_;
      }

      long
      get_switch_id(const coordinates &coords) const;

      int
      get_productive_dir(
        int &dim,
        const coordinates &src,
        const coordinates &dst) const;

      void
      minimal_route_to_switch(
        const nodeaddress::ptr &current_sw_addr,
        const nodeaddress::ptr &dest_sw_addr,
        int &dim, int &dir) const;

      void
      minimal_route_to_coords(
        const coordinates &current_coords,
        const coordinates &dest_coords,
        int &dim, int &dir) const;

      virtual std::string
      to_string() const
      {
        return "fattreetopology";
      }

      int
      minimal_distance(
        const coordinates& current_coords,
        const coordinates& dest_coords
      ) const;

    };


  }
} //end of namespace sstmac

#endif
