#include <sstmac/hardware/network/topology/regular_topology.h>
#include <sstmac/hardware/network/switches/networkswitch.h>

namespace sstmac {
    namespace hw {



std::string
regular_topology_descr::to_string() const
{
    return coords_.to_string();
}

coordinates
regular_topology::get_switch_coords(long uid) const
{
    coordinates coords(ndimensions());
    compute_switch_coords(uid, coords);
    return coords;
}

coordinates
regular_topology::get_node_coords(long uid) const
{
    if (nps_ == 1){
        coordinates coords(ndimensions());
        compute_switch_coords(uid, coords);
        return coords;
    }
    else {
        long swid = uid / nps_;
        int lastidx = uid % nps_;
        //extra one for switch coordinates
        int ndim = ndimensions();
        coordinates coords(ndim + 1);
        compute_switch_coords(swid, coords);
        coords[ndim] = lastidx;
        return coords;
    }
}

long
regular_topology::get_node_id(const coordinates &sw_coords, int port) const
{
    long net_id = get_switch_id(sw_coords);
    if (net_id >= num_leaf_switches()){
        sst_throw(valueerror,
            "regular_topology::get_node_id: invalid switch id %ld. check coordinates",
            net_id);
    }
    long nid = nps_ * net_id + port;
    return nid;
}

long
regular_topology::get_node_id(const coordinates& coords) const
{

    if (coords.size() == ndimensions()){
        //there is no "extra" switch index
        return get_switch_id(coords);
    }

    int lastidx = coords.size() - 1;
    return get_node_id(coords, coords[lastidx]);
}

nodeaddress::ptr
regular_topology::node_to_injector_addr(const nodeaddress::ptr& nodeaddr, int& switch_port) const
{
    long node_id = nodeaddr->unique_id();
    long net_id = node_id / nps_;
    switch_port = node_id % nps_;
    return nodeaddress::convert_from_uniqueid(net_id);
}

nodeaddress::ptr
regular_topology::node_to_ejector_addr(const nodeaddress::ptr& nodeaddr, int& switch_port) const
{
    long node_id = nodeaddr->unique_id();
    long net_id = node_id / nps_;
    switch_port = node_id % nps_;
    return nodeaddress::convert_from_uniqueid(net_id);
}

std::vector<long>
regular_topology::get_nodes_connected_to_switch(const nodeaddress::ptr &swaddr) const
{
    std::vector<long> ret;
    for (int i = 0; i < nps_; i++)
      ret.push_back(swaddr->unique_id() * nps_ + i);
    //ret.push_back(i);
    return ret;
}

void
regular_topology::minimal_route_to_switch(
    const nodeaddress::ptr &current_sw_addr,
    const nodeaddress::ptr &dest_sw_addr,
    int &dim, int &dir) const
{
    coordinates current_coords = get_switch_coords(current_sw_addr);
    coordinates dest_coords = get_switch_coords(dest_sw_addr);
    minimal_route_to_coords(current_coords, dest_coords, dim, dir);
}

void
regular_topology::init_factory_params(const sim_parameters::ptr &params)
{
    topology::init_factory_params(params);
    /**
    sstkeyword { gui=1 1 1;
        docstring=The number of redundant physical links that make up
                  each link[edge] in the topology. These vector dimensions
                  should exactly match those in topology_geometry.
                  The total available bandwidth in each topology link is:ENDL
                  BW=Redundant*Link_BW;
     }
    */
    if (params->has_param("topology_redundant")){
      params->get_vector_param("topology_redundant", red_);
      if (red_.size() != ndimensions()){
        sst_throw(inputerror,
                  "topology::init: wrong number of dimensions in topology_redundant."
                  "should be %d, got %d\n",
                  ndimensions(),
                  red_.size());
      }
    }
    else {
        int ndim = ndimensions();
        for (int i = 0; i < ndim; i++)
        {
          red_.push_back(1);
        }
    }
}

networkswitch::ptr
regular_topology::get_switch(const coordinates &coords, const maptype &switches)
{
    long nid = get_switch_id(coords);
    nodeaddress::ptr addr = nodeaddress::convert_from_uniqueid(nid);
    return find_switch(switches, addr);
}

void
regular_topology::get_bit_complement_partners(
    const coordinates &src_sw_coords,
    int port,
    std::vector<nodeaddress::ptr>& partners) const
{
    std::string topname = to_string();
    sst_throw(unimplementederror,
        "%s::get_bit_complement_partners: "
        "don't know how for %s",
        topname.c_str(), topname.c_str());
}

void
regular_topology::get_tornado_send_partners(
    const coordinates &src_sw_coords,
    int port,
    std::vector<nodeaddress::ptr>& partners) const
{
    std::string topname = to_string();
    sst_throw(unimplementederror,
        "%s::get_tornado_partners: "
        "don't know how for %s",
        topname.c_str(), topname.c_str());
}

void
regular_topology::get_tornado_recv_partners(
    const coordinates &src_sw_coords,
    int port,
    std::vector<nodeaddress::ptr>& partners) const
{
    std::string topname = to_string();
    sst_throw(unimplementederror,
        "%s::get_tornado_partners: "
        "don't know how for %s",
        topname.c_str(), topname.c_str());
}

void
regular_topology::get_nearest_neighbor_partners(
    const coordinates &src_sw_coords,
    int port,
    std::vector<nodeaddress::ptr>& partners) const
{
    std::string topname = to_string();
    sst_throw(unimplementederror,
        "%s::get_nearest_neighbor_partners: "
        "don't know how for %s",
        topname.c_str(), topname.c_str());
}

void
regular_topology::get_send_partners(
    traffic_pattern::type_t ty,
    const nodeaddress::ptr &src,
    std::vector<nodeaddress::ptr>& partners) const
{
    bool is_send = true;
    get_partners(is_send, ty, src, partners);
}

void
regular_topology::get_recv_partners(
    traffic_pattern::type_t ty,
    const nodeaddress::ptr &src,
    std::vector<nodeaddress::ptr>& partners) const
{
    bool is_send = false;
    get_partners(is_send, ty, src, partners);
}

void
regular_topology::get_partners(
    bool get_send_partner,
    traffic_pattern::type_t ty,
    const nodeaddress::ptr &src,
    std::vector<nodeaddress::ptr>& partners) const
{
    // convert to switch coordinates
    int port;
    nodeaddress::ptr src_sw = node_to_injector_addr(src, port);
    coordinates src_sw_coords = get_switch_coords(src_sw);
    switch(ty)
    {
        case traffic_pattern::nearest_neighbor:
            get_nearest_neighbor_partners(src_sw_coords, port, partners);
            break;
        case traffic_pattern::tornado:
            if (get_send_partner)
                get_tornado_send_partners(src_sw_coords, port, partners);
            else
                get_tornado_recv_partners(src_sw_coords, port, partners);
            break;
        case traffic_pattern::bit_complement:
            get_bit_complement_partners(src_sw_coords, port, partners);
            break;
        default:
            sst_throw(valueerror,
                "unknown traffic pattern enum %d",
                ty);
    }
}

topology_descr::ptr
regular_topology::get_descr(const nodeaddress::ptr &addr) const
{
    coordinates coords = get_switch_coords(addr);
    return regular_topology_descr::construct(coords);
}

    }
}

