/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_NETWORK_SWITCHES_REGULAR_TOPOLOGY_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_SWITCHES_REGULAR_TOPOLOGY_H_INCLUDED

#include <sstmac/hardware/network/topology/topology.h>

namespace sstmac
{
  namespace hw
  {

    class regular_topology_descr :
        public topology_descr
    {
        protected:
            coordinates coords_;

            regular_topology_descr(const coordinates& coords)
                : coords_(coords)
            {
            }

        public:
            typedef boost::intrusive_ptr<regular_topology_descr> ptr;

            static ptr
            construct(const coordinates& coords){
                return new regular_topology_descr(coords);
            }

            virtual std::string
            to_string() const;
    };

    class regular_topology : public topology
    {

    protected:
      int nps_;

      regular_topology(int nps) :
        topology(), nps_(nps)
      {

      }

      virtual void
      compute_switch_coords(long swid, coordinates& coords) const = 0;

      void
      get_partners(
        bool get_send_partner,
        traffic_pattern::type_t ty,
        const nodeaddress::ptr& src,
        std::vector<nodeaddress::ptr>& partners) const;

      virtual void
      get_nearest_neighbor_partners(
        const coordinates& src_sw_coords,
        int port,
        std::vector<nodeaddress::ptr>& partners) const;

      virtual void
      get_tornado_recv_partners(
        const coordinates& src_sw_coords,
        int port,
        std::vector<nodeaddress::ptr>& partners) const;

      virtual void
      get_tornado_send_partners(
        const coordinates& src_sw_coords,
        int port,
        std::vector<nodeaddress::ptr>& partners) const;

      virtual void
      get_bit_complement_partners(
        const coordinates& src_sw_coords,
        int port,
        std::vector<nodeaddress::ptr>& partners) const;

    public:
      typedef boost::intrusive_ptr<regular_topology> ptr;

      void init_factory_params(const sim_parameters::ptr &params);

      /**
       * Given a switch address, return number of nodes connected to it
       */
      virtual int
      nps(nodeaddress::ptr addr) const
      {
        return nps_;
      }

      virtual long
      num_nodes() const
      {
        return nps_ * num_leaf_switches();
      }

      virtual long
      num_leaf_switches() const = 0;

      virtual std::string
      to_string() const
      {
        return "regular_topology";
      }

      virtual
      ~regular_topology()
      {

      }

      coordinates
      get_switch_coords(long uid) const;

      coordinates
      get_switch_coords(const nodeaddress::ptr& sw_addr) const {
        return get_switch_coords(sw_addr->unique_id());
      }

      coordinates
      get_node_coords(long nid) const;

      coordinates
      get_node_coords(const nodeaddress::ptr& node_addr) const {
        return get_node_coords(node_addr->unique_id());
      }

      networkswitchPtr
      get_switch(const coordinates& coords, const maptype& switches);


      /**
        @param dim [inout] This can be modified if moving in a given dimension
                    requires first moving in another dimension
        @param src The coordinates of the source switch
        @param dst The coordinates of the dest switch
        @return The direction in dimension dim that is productive
      */
      virtual int
      get_productive_dir(
        int& dim,
        const coordinates& src,
        const coordinates& dst) const = 0;

      virtual nodeaddress::ptr
      node_to_injector_addr(const nodeaddress::ptr& nodeaddr, int& switch_port) const;

      int node_to_ejector_port(const nodeaddress::ptr& addr) const {
        return addr->unique_id() % nps_;
      }

      virtual nodeaddress::ptr
      node_to_ejector_addr(const nodeaddress::ptr& nodeaddr, int& switch_port) const;

      virtual int
      node_to_switch_port(long nid) const
      {
        return nid % nps_;
      }

      /**
        @return The number of hops to final destination
      */
      virtual void
      minimal_route_to_switch(
        const nodeaddress::ptr& current_sw_addr,
        const nodeaddress::ptr& dest_sw_addr,
        int& dim, int& dir
      ) const;

      /**
        @return The number of hops to final destination
      */
      virtual void
      minimal_route_to_coords(
        const coordinates& current_coords,
        const coordinates& dest_coords,
        int& dim, int& dir
      ) const = 0;

      virtual int
      minimal_distance(
        const coordinates& current_coords,
        const coordinates& dest_coords
      ) const = 0;

      virtual std::vector<long>
      get_nodes_connected_to_switch(const nodeaddress::ptr &swaddr) const;

      virtual long
      get_node_id(const coordinates& coords) const;

      virtual long
      get_node_id(const coordinates& sw_coords, int port) const;

      virtual long
      get_switch_id(const coordinates& coords) const = 0;

      virtual std::string
      get_default_router() const {
        return "regular";
      }

      virtual void
      get_send_partners(
        traffic_pattern::type_t ty,
        const nodeaddress::ptr &src_node,
        std::vector<nodeaddress::ptr>& partners) const;

      virtual void
      get_recv_partners(
        traffic_pattern::type_t ty,
        const nodeaddress::ptr &src_node,
        std::vector<nodeaddress::ptr>& partners) const;

      /**
          The number of distinct 'dimensions'
          in the topology.  This can correspond directly to standard
          X,Y,Z dimensions or the the number of levels in a fat tree.
          The keyword topology_redundant vector should have this many
          entries.
      */
      virtual int
      ndimensions() const = 0;

      topology_descr::ptr
      get_descr(const nodeaddress::ptr &addr) const;

    };
  }
}

#endif
