/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/topology/routing/router.h>
#include <sstmac/hardware/network/switches/networkswitch.h>
#include <sstmac/hardware/network/topology/routing/routing_info.h>

ImplementFactory(sstmac::hw::router)
;

namespace sstmac
{
  namespace hw
  {

    debug router::dbg_;
    bool router::template_init_ = false;

    int //assume all switches have same input/output port mappings
    router::get_inport(const nodeaddress::ptr &output_buf_sw,
        const nodeaddress::ptr &input_but_bw, int dim, int dir)
    {
      return convert_to_outport(dim, dir);
    }

    router::~router()
    {
    }

    void
    router::set_switch(const networkswitch::ptr& sw)
    {
      sw_ = sw;
      algo_map::const_iterator it, end = algos_.end();
      for (it = algos_.begin(); it != end; ++it)
      {
        it->second->set_switch(sw);
      }
    }

    router::router() :
      default_algo_(routing_info::system_default)
    {
      if (!dbg_.is_real())
      {
        dbg_ = debug("<debug> router | network");
        logger::register_logger(&dbg_);
      }
    }

    bool
    router::get_productive_ports_to_node(const nodeaddress::ptr &dst,
        outports &ports)
    {
      int dir;
      nodeaddress::ptr ej_addr = sw_->get_topology()->node_to_ejector_addr(dst,
          dir);
      if (ej_addr->equals(sw_->get_addr()))
      {
        ports.resize(1);
        ports[0] = convert_to_outport(topology::EJECT, dir);
        return true;
      }
      else
      {
        get_productive_ports_to_switch(ej_addr, ports);
        return false;
      }
    }

    void
    router::convert_to_dimdir(int outport, int &dim, int &dir)
    {
      sw_->get_topology()->convert_to_dimdir(outport, dim, dir);
    }

    int
    router::convert_to_outport(int dim, int dir)
    {
      return sw_->get_topology()->convert_to_outport(dim, dir);
    }

    int
    router::route(const interconnect_message::ptr &msg)
    {
      if (!msg->routing_initialized())
      {
        init_routing(msg);
      }
      return do_route(msg);
    }

    int
    router::do_route(const interconnect_message::ptr &msg)
    {
      routing_algorithm::ptr algo = algos_[msg->rinfo()->route_algo()];
      if (!algo)
      {
        sst_throw(valueerror,
            "routing algorithm %s not supported",
            routing_info::tostr(msg->rinfo()->route_algo()));
      }
      return algo->route(msg);
    }

    int
    router::get_number_paths(const interconnect_message::ptr& msg) const
    {
      if (!msg->routing_initialized())
      {
        init_routing(msg);
      }
      if (!msg->rinfo())
        return 1;
      routing_algorithm::const_ptr algo = algos_.at(msg->rinfo()->route_algo());
      if (!algo)
      {
        sst_throw(valueerror,
            "routing algorithm %s not supported",
            routing_info::tostr(msg->rinfo()->route_algo()));
      }
      return algo->get_number_paths(msg);
    }

    void
    router::init_routing(const interconnect_message::ptr &msg) const
    {
      routing_info::routing_algorithm_t algo_enum =
          msg->get_network_msg()->route_algo();
      if (algo_enum == routing_info::system_default)
        algo_enum = default_algo_;

      routing_algorithm::const_ptr algo = algos_.at(algo_enum);
      if (!algo)
      {
        sst_throw(valueerror,
            "routing algorithm %s not supported",
            routing_info::tostr(algo_enum));
      }
      algo->init_routing(msg);
    }

    routing_info::routing_algorithm_t
    router::str_to_algo(const std::string &str)
    {
      if (str == "minimal")
      {
        return routing_info::minimal;
      }
      else if (str == "valiant")
      {
        return routing_info::valiant;
      }
      else if (str == "min_ad")
      {
        return routing_info::minimal_adaptive;
      }
      else if (str == "ugal")
      {
        return routing_info::ugal;
      }
      else
      {
        sst_throw(inputerror,
            "invalid routing algorithm %s",
            str.c_str());
      }
    }

    void
    router::init_factory_params(const sim_parameters::ptr &params)
    {
      std::string str =
          params->get_optional_param("default_routing", "minimal");
      default_algo_ = str_to_algo(str);

      std::vector<std::string> valid_algos;
      if (params->has_param("valid_routing"))
      {
        params->get_vector_param("valid_routing", valid_algos);
      }
      else
      {
        valid_algos.resize(4);
        valid_algos[0] = "minimal";
        valid_algos[1] = "valiant";
        valid_algos[2] = "min_ad";
        valid_algos[3] = "ugal";
      }

      for (int i = 0; i < valid_algos.size(); ++i)
      {
        routing_info::routing_algorithm_t algo_enum = str_to_algo(
            valid_algos[i]);
        routing_algorithm::ptr algo = routing_algorithm_factory::get_value(
            valid_algos[i], params);
        algos_[algo_enum] = algo;
      }
    }

    void
    router::init(const nodeaddress::ptr& me, const topology::ptr& top)
    {
      me_ = me;
    }

  }
}

