/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */
#ifndef SSTMAC_HARDWARE_NETWORK_TOPOLOGY_TOP_FROM_FILE_H
#define SSTMAC_HARDWARE_NETWORK_TOPOLOGY_TOP_FROM_FILE_H

#include <sstmac/hardware/network/topology/topology.h>

namespace sstmac
{
  namespace hw
  {

    /** Implements a topology that is specified from file.  Each connection
     * must be specified on a separate line.  Use keywords "computer" and "router"
     * to specify a node or a switch.  For example:
     *
     * computer1 router1
     * router1 router2
     * computer2 router2
     * computer3 router3
     * router1 router3
     *
     *
     * is a ring topology.  you get the idea.
     */

    struct connection
    {
      long id_;
      int red_;

      connection(long i)
      {
        id_ = i;
        red_ = 1;
      }
    };

    class top_from_file : public sstmac::hw::topology
    {
    protected:
      std::string fname_;
      long num_switches_;
      mutable bool parsed_node_conn_;
      mutable bool parsed_switch_conn_;

      typedef boost::unordered_map<long, long> node_connection_map;
      mutable node_connection_map node_connections_;

      typedef boost::unordered_map<long, std::vector<connection> >
          rconnection_map;
      mutable rconnection_map rconnections_;

      typedef boost::unordered_map<long, std::vector<long> > nconnection_map;
      mutable boost::unordered_map<long, std::vector<long> > nconnections_;

    protected:
      top_from_file(const std::string fname);

    public:
      typedef boost::intrusive_ptr<top_from_file> ptr;

      ptr
      static
      construct(sstmac::sim_parameters::ptr params)
      {
        std::string fname = params->get_param("top_from_file_name");
        return ptr(new top_from_file(fname));
      }

      /**
       * GH: This is required for flowers, but I don't feel like implementing it.
       * @return NOT the diameter, throws;
       */
      virtual int64_t
      diameter() const
      {
        sst_throw(sstmac::ssterror,
            "top_from_file::diameter() - not implemented");
      }

      virtual std::string
      get_default_router() const {
        return "djikstra";
      }

      void
      minimal_route_to_switch(
        const nodeaddress::ptr &current_sw_addr,
        const nodeaddress::ptr &dest_sw_addr,
        int &dim, int &dir) const;

      /**
       * GH: This is required, but not really used and I don't feel like implementing it.
       * @return NOT the radix, throws
       */
      virtual int
      radix() const
      {
        sst_throw(sstmac::ssterror, "top_from_file::radix() - not implemented");
      }

      virtual void
      connect_switches(topology::maptype &switches);

      virtual long
      num_switches() const
      {
        find_num_nodes();
        return nconnections_.size();
      }

      virtual int
      nps(nodeaddress::ptr addr) const;

      virtual long
      num_nodes() const
      {
        find_num_nodes();
        return node_connections_.size();
      }

      virtual void
      find_num_switches() const;

      virtual void
      find_num_nodes() const;

      nodeaddress::ptr
          node_to_injector_addr(const nodeaddress::ptr &nodeaddr,
              int& switch_port) const;

      nodeaddress::ptr
          node_to_ejector_addr(const nodeaddress::ptr &nodeaddr,
              int& switch_port) const;

      virtual int
      node_to_switch_port(long nid) const;

      virtual std::vector<long>
      get_nodes_connected_to_switch(const nodeaddress::ptr &swaddr) const
      {
        return nconnections_[swaddr->unique_id()];
      }

      long
      get_switch_injecting_from_node(sstmac::nodeaddress::ptr addr) const;

      long
      get_switch_ejecting_to_node(sstmac::nodeaddress::ptr addr) const;

      virtual int
      convert_to_outport(int dim, int dir);

      virtual void
      convert_to_dimdir(int outport, int &dim, int&dir);

      virtual
      ~top_from_file()
      {
      }

      virtual std::string
      to_string() const
      {
        return "top_from_file";
      }

      virtual void
      get_send_partners(
        traffic_pattern::type_t ty,
        const nodeaddress::ptr& src_node,
        std::vector<nodeaddress::ptr>& partners) const;

      virtual void
      get_recv_partners(
        traffic_pattern::type_t ty,
        const nodeaddress::ptr& src_node,
        std::vector<nodeaddress::ptr>& partners) const;

      topology_descr::ptr
      get_descr(const nodeaddress::ptr &addr) const;

    };

  }
}
#endif
