#include <sstmac/hardware/network/topology/topology.h>
#include <sstmac/hardware/network/topology/routing/regular_router.h>
#include <sstmac/hardware/network/switches/networkswitch.h>

#include <sstmac/hardware/network/switches/entryexit.h>

//#include <sstmac/backends/native/par_eventmap.h>
#include <sstmac/backends/native/sim_partition.h>
#include <sstmac/hardware/network/switches/dist_dummyswitch.h>

ImplementFactory(sstmac::hw::topology)
;

namespace sstmac
{
  namespace hw
  {

    using namespace native;

    debug topology::dbg_;
    RNG::Combo::ptr topology::rng_;

    topology::topology()
    {
      if (!dbg_.is_real())
      {
        dbg_ = debug("<debug> topology | network");
        logger::register_logger(&dbg_);
      }

    }

    topology::~topology()
    {
    }

    void
    topology::init_factory_params(const sim_parameters::ptr &params)
    {
        outputgraph_ = params->get_optional_bool_param("topology_output_graph",
          false);

        if (!rng_)
        {
          long seed = time(NULL);
          /**
            sstkeyword {
                gui = 42;
                docstring = In some cases, walking the topology requires a random decision.ENDL
                The seed for all the random decisions can be explicitly set for debugging.;
                extra = true;
            }
          */
          if (params->has_param("topology_seed"))
            seed = params->get_long_param("topology_seed");

          std::vector<RNG::rngint_t> seeds;
          seeds.push_back(seed);
          rng_ = RNG::Combo::construct(seeds);

        }
    }

    topology::direction_t
    topology::get_opp_dir(int dir)
    {
        switch (dir)
          {
        case POS:
          return NEG;

        case NEG:
          return POS;

        case UP:
          return DOWN;

        case DOWN:
          return UP;

        case EAST:
          return WEST;

        case WEST:
          return EAST;

        case NORTH:
          return SOUTH;

        case SOUTH:
          return NORTH;
          }

        sst_throw(valueerror, "No router opposite direction found for %d\n", dir);
        return POS;
    }

    void
    topology::minimal_route_to_node(
        const nodeaddress::ptr &current_sw_addr,
        const nodeaddress::ptr &dest_node_addr,
        int &dim, int &dir) const
    {
        nodeaddress::ptr ej_addr = node_to_ejector_addr(dest_node_addr, dir);
        if (ej_addr->equals(current_sw_addr)){
            dim = EJECT;
        }
        else {
            minimal_route_to_switch(current_sw_addr, ej_addr, dim, dir);
        }
    }

    void
    topology::clear_rng()
    {
        rng_ = 0;
    }

    nodeaddress::ptr
    topology::get_random_intermediate_switch(const nodeaddress::ptr &current, const sst_message::ptr &msg)
    {
        long curid = current->unique_id();
        long nid = curid;
        while (curid == nid) {
            nid = rng_->value_in_range(num_switches());
        }
        return nodeaddress::convert_from_uniqueid(nid);
    }

    coordinates
    topology::get_switch_coords(long uid) const
    {
        sst_throw(unimplementederror,
            "topology::get_switch_coords for %s",
            to_string().c_str());
    }

    std::string
    topology::label(long id) const
    {
      return "vertex " + to_string() + sstprintf("2%d", id);
    }

    router::ptr
    topology::create_router(const nodeaddress::ptr &addr,
        const sim_parameters::ptr &params)
    {
      router::ptr ret = router_factory::get_optional_param("router",
          get_default_router(), params, addr, ptr(this));
      return ret;
    }

    networkswitch::ptr
    topology::find_switch(const maptype &switches, nodeaddress::ptr addr)
    {
      if (sstmac_env::run_mode_ == sstmac_env::SHARED_MEMORY)
      {
        if (switches.find(addr) == switches.end())
        {
          sst_throw(valueerror, "topology::find_switch - couldn't find address %s", addr->to_string().c_str());
        }
        return switches.at(addr);
      }
      else
      {
        if (switches.find(addr) == switches.end())
        {
          if (dummies_.find(addr) != dummies_.end())
          {
            return dummies_[addr];
          }

         // par_eventmap::ptr ev_mgr =
         //     boost::dynamic_pointer_cast<par_eventmap>(evman_);
         // if (!ev_mgr)
         //   sst_throw(ssterror, "topology::find_switch - error getting par_eventmap");
         // ip_queue::ptr lp_queue = ev_mgr->get_ip_queue(addr);

          timestamp lat = sstmac_env::params->get_time_param("ip_latency");
          dist_dummyswitch::ptr dum = dist_dummyswitch::construct(addr, lat);
          dum->init(ptr(this));
          dum->init_addr(addr, regular_router::construct(sstmac_env::params));

          dum->set_eventmanager(evman_);
          dummies_[addr] = dum;
          return dum;
        }
        else
        {
          return switches.at(addr);
        }
      }
    }

    void
    topology::set_evman(const eventmanager::ptr & ev)
    {
      evman_ = ev;
      if (sstmac_env::run_mode_ != sstmac_env::SHARED_MEMORY)
      {
     //   native::par_eventmap::ptr ev_mgr = boost::dynamic_pointer_cast<
      //      native::par_eventmap>(ev);
      //  if (!ev_mgr)
      //    sst_throw(ssterror, "topology::set_eventmanager - error getting dist_eventmap");
        dbg_.set_partition_id(ev->get_lpid());
      }
    }

    std::vector<long>*
    topology::nodes_to_add()
    {
      std::vector<long>* ret = new std::vector<long>();

      if (sstmac_env::run_mode_ == sstmac_env::SHARED_MEMORY)
      {
        long count = num_nodes();
        ret->resize(count);
        for (long i = 0; i < count; i++)
        {
          (*ret)[i] = i;
        }
      }
      else
      {
       // par_eventmap::ptr ev_mgr = boost::dynamic_pointer_cast<par_eventmap>(
       //     evman_);
       // if (!ev_mgr)
       //   sst_throw(ssterror, "topology::switches_to_add - error getting dist_eventmap");

        partition::ptr part = evman_->get_partition();

        int lpid = evman_->get_lpid();
        std::list<int>::const_iterator iterator, end = part->get_partition(lpid)->end();
        int nodeindex = 0;
        //ret->resize(count);
        for (iterator = part->get_partition(lpid)->begin(); iterator != end; ++iterator)
        {
          int swid = *iterator;
          std::vector<long> v = get_nodes_connected_to_switch(
              nodeaddress::convert_from_uniqueid(swid));
          for (int j = 0; j < v.size(); j++)
          {
            ret->push_back(v[j]);
            nodeindex++;
          }
        }

      }

      return ret;
    }

    std::vector<long>*
    topology::switches_to_add()
    {
      std::vector<long>* ret = new std::vector<long>();

      if (sstmac_env::run_mode_ == sstmac_env::SHARED_MEMORY)
      {
        long count = num_switches();
        ret->resize(count);
        for (long i = 0; i < count; i++)
        {
          (*ret)[i] = i;
        }
      }
      else
      {
       // par_eventmap::ptr ev_mgr = boost::dynamic_pointer_cast<par_eventmap>(
       //     evman_);
       // if (!ev_mgr)
       //   sst_throw(ssterror, "topology::switches_to_add - error getting dist_eventmap");

        partition::ptr part = evman_->get_partition();
        int lpid = evman_->get_lpid();
        std::list<int>::const_iterator iterator, end = part->get_partition(lpid)->end();
        int nodeindex = 0;
        ret->resize(part->get_partition(lpid)->size());
        for (iterator = part->get_partition(lpid)->begin(); iterator != end; ++iterator)
        {
          (*ret)[nodeindex] = *iterator;
          nodeindex++;
        }
      }

      return ret;
    }

    coordinates::coordinates()
    {
    }

    coordinates::coordinates(int nelem)
      : std::vector<int>(nelem, 0)
    {
    }

    coordinates::coordinates(const std::vector<int> &vec)
        : std::vector<int>(vec)
    {
    }

    std::string
    coordinates::to_string() const
    {
        std::stringstream sstr;
        sstr << "[";
        for (int i=0; i < size(); ++i)
        {
            sstr << " " << at(i);
        }
        sstr << " ]";
        return sstr.str();
    }

  }
}
