/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_NETWORK_SWITCHES_SWITCHTOPOLOGY_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_SWITCHES_SWITCHTOPOLOGY_H_INCLUDED

#include <sstmac/common/ptr_type.h>
#include <sstmac/common/sstmac-config.h>
#include <sstmac/common/debug.h>
#include <sstmac/common/sim_parameters.h>
#include <sstmac/common/vis/vis.h>
#include <sstmac/common/factories/factory.h>
#include <sstmac/common/eventmodule.h>
#include <sstmac/common/rng.h>
#include <sstmac/hardware/network/topology/traffic/traffic.h>

namespace sstmac
{
  namespace hw
  {

      class coordinates :
        public std::vector<int>
      {
        public:
            coordinates(const std::vector<int>& vec);

            coordinates();

            coordinates(int nelem);

            std::string to_string() const;

      };

    /**
     * This class is an interface for topologies which support
     * the switch network model
     */

    class networkswitch;
    typedef boost::intrusive_ptr<networkswitch> networkswitchPtr;
    class router;
    typedef boost::intrusive_ptr<router> routerPtr;

    class topology_descr :
        public ptr_type
    {
        public:
            typedef boost::intrusive_ptr<topology_descr> ptr;

            virtual std::string
            to_string() const = 0;
    };

    class topology : public virtual vis::vis_topology, public factory_type
    {

    protected:
      bool outputgraph_;

      static debug dbg_;

      eventmanager::ptr evman_;

      std::vector<int> red_;

      std::string name_;

      static RNG::Combo::ptr rng_;

      boost::unordered_map<nodeaddress::ptr, networkswitchPtr> dummies_;

      topology();

    public:

      typedef enum
      {
        POS=0,
        NEG=1,
        UP=2,
        DOWN=3,
        EAST=4,
        WEST=5,
        NORTH=6,
        SOUTH=7,
        EJECT = -1
      } direction_t;

      static direction_t
      get_opp_dir(int dir);

      typedef boost::intrusive_ptr<topology> ptr;

      typedef boost::unordered_map<nodeaddress::ptr, networkswitchPtr> maptype;

      virtual void
      connect_switches(maptype &switches) = 0;

      static void clear_rng();

      virtual routerPtr
      create_router(const nodeaddress::ptr& addr, const sim_parameters::ptr& params);

      virtual int
      radix() const = 0;

      virtual long
      num_switches() const = 0;

      void
      set_evman(const eventmanager::ptr & ev);

      /**
       * nodes_to_add and switches_to_add allocates memory with new, make sure to delete it!!
       */
      virtual std::vector<long>*
      nodes_to_add();

      virtual std::vector<long>*
      switches_to_add();

      virtual networkswitchPtr
      find_switch(const maptype &switches, nodeaddress::ptr addr);

      virtual nodeaddress::ptr
      get_random_intermediate_switch(const nodeaddress::ptr& current_sw, const sst_message::ptr& msg);

      /**
       Given an index input, return a sensible label
       (e.g. hdtorus returns x,y,z coordinates) to
       identify the vertx
       */
      virtual std::string
      label(long id) const;

      virtual std::string
      get_default_router() const  = 0;

      /**
       * Given a switch address, return number of nodes connected to it
       */
      virtual int
      nps(nodeaddress::ptr addr) const = 0;

      virtual long
      num_nodes() const = 0;

      virtual void
      init_factory_params(const sim_parameters::ptr &params);

      virtual std::vector<long>
      get_nodes_connected_to_switch(const nodeaddress::ptr &nodeaddr) const = 0;

      nodeaddress::ptr
      node_to_injector_addr(const nodeaddress::ptr &nodeaddr) const
      {
        int ignore;
        return node_to_injector_addr(nodeaddr, ignore);
      }

      nodeaddress::ptr
      node_to_ejector_addr(const nodeaddress::ptr &nodeaddr) const
      {
        int ignore;
        return node_to_ejector_addr(nodeaddr, ignore);
      }

      virtual nodeaddress::ptr
      node_to_injector_addr(const nodeaddress::ptr& nodeaddr, int& switch_port) const = 0;

      virtual nodeaddress::ptr
      node_to_ejector_addr(const nodeaddress::ptr& nodeaddr, int& switch_port) const = 0;

      virtual int
      node_to_switch_port(long nid) const = 0;

      virtual void
      display_nodes(const vismaptype &switches,
          const vis::vis_engine::ptr& eng, std::list<vis::vis_obj::ptr> &objs)
      {

      }

      /**
        @param The dimension (e.g. x, y, or z) that should be routed on next.
                    Integer value has different meaning depending on type of
                    router (torus/dragonfly/etc)
        @param The direction (e.g. +/-) that should be routed on.
                    Integer value has different physical meaning depending on type
                    of router (torus/dragonly/etc)
        @return The port number on a switch that the message should be routed through
                for a given dim/dir
      */
      virtual int
      convert_to_outport(int dim, int dir) = 0;

        /**
        @param The port number on a switch that the message should be routed through
                for a given dim/dir
        @param[out] The dimension (e.g. x, y, or z) that should be routed on next.
                    Integer value has different meaning depending on type of
                    router (torus/dragonfly/etc)
        @param[out] The direction (e.g. +/-) that should be routed on.
                    Integer value has different physical meaning depending on type
                    of router (torus/dragonly/etc)
        */
      virtual void
      convert_to_dimdir(int outport, int &dim, int&dir) = 0;

      /**
        @return The number of hops to final destination
      */
      virtual void
      minimal_route_to_switch(
        const nodeaddress::ptr& current_sw_addr,
        const nodeaddress::ptr& dest_sw_addr,
        int& dim, int& dir
      ) const = 0;

      /**
        @return The number of hops to final destination
      */
      virtual void
      minimal_route_to_node(
        const nodeaddress::ptr& current_sw_addr,
        const nodeaddress::ptr& dest_node_addr,
        int& dim, int& dir
      ) const;

      virtual coordinates
      get_switch_coords(long uid) const;

      virtual
      ~topology();

      virtual std::string
      to_string() const
      {
        return "topology";
      }

      virtual void
      get_send_partners(
        traffic_pattern::type_t ty,
        const nodeaddress::ptr& src_node,
        std::vector<nodeaddress::ptr>& partners) const = 0;

      virtual void
      get_recv_partners(
        traffic_pattern::type_t ty,
        const nodeaddress::ptr& src_node,
        std::vector<nodeaddress::ptr>& partners) const = 0;

      virtual topology_descr::ptr
      get_descr(const nodeaddress::ptr& addr) const = 0;

    };

    DeclareFactory(topology);
  }
}

#endif
