/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/nic/networkinterface.h>
#include <sstmac/software/process/operatingsystem.h>
#include <sstmac/hardware/network/interconnect/interconnect.h>

ImplementFactory(sstmac::hw::networkinterface)
;

namespace sstmac
{
  namespace hw
  {

    debug networkinterface::dbg_;
    stat_spyplot::ptr networkinterface::spy_;
    stat_spyplot::ptr networkinterface::spy_bytes_;
    stat_spyplot_png::ptr networkinterface::spy_bytes_alloc_png_;
    stat_spyplot_png::ptr networkinterface::spy_bytes_machine_png_;
    stat_histogram<long>::ptr networkinterface::hist_msg_size_;
    stat_sum::ptr networkinterface::total_lat_;

    networkinterface::networkinterface()
    {
      if (!dbg_.is_real())
      {
        dbg_ = debug("<debug> nic");
        logger::register_logger(&dbg_);
      }

    }

    void
    networkinterface::finish()
    {
      if (finished_)
        return;

      component::finish();
      if (lib_mem_)
      {
        lib_mem_->get_os()->unregister_lib(lib_mem_);
        lib_mem_->unregister_all_libs();
        lib_mem_ = 0;
      }
      if (parent_)
      {
        parent_->finish();
        parent_ = 0;
      }
      injector_ = 0;
    }

    void
    networkinterface::clear_statics()
    {
      spy_ = 0;
      spy_bytes_ = 0;
      spy_bytes_alloc_png_ = 0;
      spy_bytes_machine_png_ = 0;
      total_lat_ = 0;
      hist_msg_size_ = 0;
    }

    void
    networkinterface::finish_recv_ack(const sst_message::ptr& msg)
    {
    }

    void
    networkinterface::finish_recv_req(const sst_message::ptr& msg)
    {
    }

    void
    networkinterface::handle(const sst_message::ptr& netmsg)
    {
      sst_message::ptr payload = netmsg->get_parent_msg();
      switch (payload->get_type())
      {
      case sst_message::RDMA_GET_REQ:
        payload->mutate(sst_message::RDMA_GET_REQ_TO_RSP);
        send(payload);
        finish_recv_req(netmsg);
        break;
      case sst_message::RDMA_PUT_RSP:
      case sst_message::SEND_ACK:
        parent_->handle(payload);
        finish_recv_ack(netmsg);
        break;
      case sst_message::RDMA_GET_RSP:
      case sst_message::RDMA_PUT_REQ:
      default:
        recv(netmsg);
        break;
      }
    }

    void
    networkinterface::set_eventmanager(const eventmanager::ptr &m)
    {

      component::set_eventmanager(m);

      if (!spy_)
      {
        spy_ = stat_spyplot::construct("<stats> spyplot | spyplot_csv",
            "network_spyplot.csv");
        register_stat(spy_);

        spy_bytes_ = stat_spyplot::construct("<stats> spyplot | spyplot_csv",
            "network_spyplot_bytes.csv");
        register_stat(spy_bytes_);

        spy_bytes_machine_png_ = stat_spyplot_png::construct(
            "<stats> spyplot | spyplot_png", "machine_spyplot.png", true);

        long switches = interconn_->get_topology()->num_switches();

        if (spy_bytes_machine_png_->is_active() && switches > 1024)
        {
          std::cout << "WARNING: disabling machine spyplot png, since it would produce a large cumbersome png\n";
          spy_bytes_machine_png_->disable();
        }

        spy_bytes_alloc_png_ = stat_spyplot_png::construct(
            "<stats> spyplot | spyplot_png", "allocation_spyplot.png", false);

        register_stat(spy_bytes_alloc_png_);
        register_stat(spy_bytes_machine_png_);

        total_lat_ = stat_sum::construct("<stats> latency", "nic_latency.csv");
        register_stat(total_lat_);

        hist_msg_size_ = stat_histogram<long>::construct("<stats> histograms",
            "network_histogram_msgsize.csv");

        register_stat(hist_msg_size_);
      }

      spy_bytes_machine_png_->add(parent_->get_id()->unique_id(), 0, 0);

    }

    void
    networkinterface::connect(connection_type_t ty, const connectable::ptr &mod)
    {
      switch (ty)
      {
      case injection:
        injector_ = mod;
        mod->connect(parent_->get_id(), injection, ptr(this));
        break;
      case ejection:
        // I don't need a link to the ejector
        mod->connect(parent_->get_id(), ejection, ptr(this));
        break;
      default:
        connectable_type_invalid(ty);
      }
    }

    void
    networkinterface::init(const node::ptr &parent,
        const sstmac::hw::interconnect::ptr &interconn)
    {
      //this is just a template nic
      if (!parent->get_id())
        return;
      {
      }

      interconn_ = interconn;
      parent_ = parent;

      sw::operatingsystem::ptr os = parent->get_os();
      sw::library::ptr memlib = os->get_lib(
          sw::lib_compute_memmove::get_name("nic-lib-mem"));
      sw::lib_compute_memmove::ptr thelib = boost::dynamic_pointer_cast<
          sw::lib_compute_memmove>(memlib);

      if (!thelib)
      {
        lib_mem_ = sw::lib_compute_memmove::construct("nic-lib-mem");
        lib_mem_->init_as_callback_lib();
        parent->get_os()->register_lib(lib_mem_);
      }
      else
      {
        lib_mem_ = thelib;
      }

      //interconn_->set_endpoint(parent->get_id(), ptr(this));
    }

    long
    networkinterface::failure_id() const
    {
      return parent_->get_id()->unique_id();
    }

    networkinterface::~networkinterface()
    {
    }

    void
    networkinterface::validate_msg_supported(const sst_message::ptr &msg) const
    {
      return;
      /*switch (msg->get_type())
       {
       case sst_message::MPI:
       case sst_message::HPX:
       case sst_message::SEND_ACK:
       case sst_message::ACTOR:
       case sst_message::NETWORK:
       case sst_message::FILEMETA:
       case sst_message::FILEDATA:
       return; //valid
       }
       std::string errormsg = to_string()
       + ": network interface card received invalid message type";
       errormsg += sst_message::str(msg->get_type());
       throw unimplementederror(errormsg);*/
    }

  }
} // end of namespace sstmac.
