/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/node/degradednode.h>

#include <sstmac/hardware/processor/simpleprocessor-config.h>
#include <sstmac/hardware/processor/instructionprocessor-config.h>
#include <sstmac/hardware/node/simplenode.h>
#include <sstmac/hardware/memory/simple_memorymodel.h>
#include <sstmac/hardware/processor/migratingcomputescheduler.h>
#include <sstmac/common/errors.h>
#include <sstmac/software/services/launch/machine_descriptor.h>
#include <sstmac/software/process/operatingsystem.h>
#include <sstmac/software/services/launch/launcher.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("degraded | degradednode", node, degradednode);

    void
    degradednode::initialize(eventmanager::ptr man, const topology::ptr& top)
    {
      node::initialize(man, top);

      double tcycle = 1.0 / (freq_ / factor_);

      timestamp memory_latency = params_->get_time_param("node_mem_latency");
      double memory_bandwidth = params_->get_bandwidth_param("node_mem_bandwidth");

      double tflop = tcycle; // 10 clock cycles for a floating point
      double tmemop = memory_latency.sec() + 1.0 / (memory_bandwidth / factor_);

      sstmac::hw::simpleprocessor_config::ptr cfg =
          sstmac::hw::simpleprocessor_config::construct(ncores_, freq_, 10, 30);

      compute_sched_ = sstmac::hw::migratingcomputescheduler::construct(
          nodeid_->unique_id(), ncores_, nsockets_, preempt_, os_);
      compute_sched_->set_eventmanager(man);
      proc_model_ = sstmac::hw::simpleprocessor::construct(cfg, msg_handler_,
          compute_sched_);

      sstmac::hw::instructionprocessor_config::ptr pcfg =
          sstmac::hw::instructionprocessor_config::construct(ncores_,
              eventdata::INST_FLOP, tflop, eventdata::INST_MEMOP, tmemop,
              eventdata::INST_MEMOP_LOCAL, tmemop, eventdata::INST_MEMOP_RAND, tmemop);

      proc_inst_ = sstmac::hw::instructionprocessor::construct(pcfg,
          msg_handler_, compute_sched_);

      mem_model_ = sstmac::hw::simple_memorymodel::construct(params_);
      mem_model_->init(ptr(this), compute_sched_);
      mem_model_->set_eventmanager(man);

      if (params_ && params_->has_param("disk_type"))
        {
          if (params_->get_param("disk_type") == "vesper")
            {
              disk_ = sstmac::hw::disk_controller::construct(
                  nodeid_->unique_id(), params_, msg_handler_, man);
            }
        }

      os_->start_apps();

      launcher_->start();

    }

    node::ptr
    degradednode::clone(nodeaddress::ptr my_id) const
    {
      node::ptr ret = new   degradednode(factor_, params_, my_id);
      return ret;
    }

  }
}
