/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/processor/simpleprocessor-config.h>
#include <sstmac/hardware/processor/instructionprocessor-config.h>
#include <sstmac/hardware/memory/simple_memorymodel.h>
#include <sstmac/hardware/node/detailednode.h>
#include <sstmac/hardware/processor/pinnedcomputescheduler.h>
#include <sstmac/hardware/memory/mc_memorymodel.h>
#include <sstmac/software/process/operatingsystem.h>
#include <sstmac/hardware/nic/networkinterface.h>

#include <sstmac/software/libraries/filesystem/disk_message.h>
#include <sstmac/software/libraries/filesystem/nppfs_server.h>

#include <sstmac/common/errors.h>
#include <sstmac/software/services/launch/machine_descriptor.h>
#include <iostream>
#include <sstmac/software/services/launch/launcher.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("detailed | detailednode", node, detailednode);

    using namespace sstmac::sw;

    detailednode::detailednode(const sim_parameters::ptr &params) :
      simplenode(params)
    {

    }

    detailednode::detailednode(const sim_parameters::ptr &params, const nodeaddress::ptr &my_id) :
      simplenode(params, my_id)
    {
      os_->setAMI(sstmac::sw::AMI::AMI_COMM, this);
      os_->setAMI(sstmac::sw::AMI::AMI_COMPUTE, this);
      os_->setAMI(sstmac::sw::AMI::AMI_QUERY, this);
      os_->setAMI(sstmac::sw::AMI::AMI_HW, this);

      ncores_ = params->get_int_param("node_cores");
      memory_latency_ = params->get_time_param("node_mem_latency");
      memory_bandwidth_ = params->get_bandwidth_param("node_mem_bandwidth");
      freq_ = params->get_freq_param("node_frequency");

      proc_link_bw_ = params->get_bandwidth_param("proc_link_bandwidth");
      proc_link_lat_ = params->get_time_param("proc_link_latency");

      num_mc_ = params->get_int_param("proc_num_mc");
      num_sock_ = params->get_int_param("node_num_procs");

      timestamp pre(-1);
      if (params->has_param("node_preemption"))
        {
          pre = params->get_time_param("node_preemption");
        }
      preempt_ = pre;

      SSTMAC_DEBUG << "constructed a detailednode with parameters mem_lat = "
            << memory_latency_ << ", memory_bandwidth = " << memory_bandwidth_
            << ", freq = " << freq_ << ", cores = " << ncores_ << "\n";
    }

    void
    detailednode::initialize(nodeaddress::ptr my_id,
                             eventmanager::ptr man,
                             const topology::ptr& top)
    {
      node::initialize(man, top);

      SSTMAC_DEBUG << "starting detailednode initialization \n";

      double tcycle = 1.0 / (freq_);

      double tflop = tcycle;
      double tmemop = tflop;

      sstmac::hw::simpleprocessor_config::ptr cfg =
          sstmac::hw::simpleprocessor_config::construct(ncores_, freq_, 10, 30);

      compute_sched_ = sstmac::hw::pinnedcomputescheduler::construct(my_id->unique_id(), ncores_, preempt_, num_sock_, os_);
      compute_sched_->set_eventmanager(man);
      proc_model_ = sstmac::hw::simpleprocessor::construct(cfg, msg_handler_, compute_sched_);

      sstmac::hw::instructionprocessor_config::ptr pcfg =
          sstmac::hw::instructionprocessor_config::construct(ncores_,
              eventdata::INST_FLOP, tflop, eventdata::INST_MEMOP, tmemop,
              eventdata::INST_MEMOP_LOCAL, tmemop, eventdata::INST_MEMOP_RAND, tmemop);

      proc_inst_ = sstmac::hw::instructionprocessor::construct(pcfg,
          msg_handler_, compute_sched_);

      mem_model_ = sstmac::hw::mc_memorymodel::construct(memory_bandwidth_, memory_latency_,
                                                         proc_link_bw_, proc_link_lat_,
                                                         num_mc_, num_sock_);
      mem_model_->init(ptr(this), compute_sched_);
      mem_model_->set_eventmanager(man);

      if (params_ && params_->has_param("disk_type"))
        {
          if (params_->get_param("disk_type") == "vesper")
            {
              disk_ = sstmac::hw::disk_controller::construct(
                  my_id->unique_id(), params_, msg_handler_, man);
            }
        }

      if (params_ && params_->has_param("node_nppfs_server"))
        {
          long nid = params_->get_long_param("node_nppfs_server");
          nodeaddress::ptr addr = nodeaddress::convert_from_uniqueid(nid);
          nppfs_server_ = addr;

          std::string fname = params_->get_param("node_nppfs_diskfile");
          sstmac::sw::nppfs_server::ptr serv =
              sstmac::sw::nppfs_server::construct(fname, addr->equals(nodeid_));
          os_->register_ser(serv);
          serv->start();

        }

      SSTMAC_DEBUG << "detailednode initialization done, starting things \n";
      os_->start_apps();

      launcher_->start();
    }

    node::ptr
    detailednode::clone(nodeaddress::ptr my_id) const
    {
      node::ptr ret = new detailednode(params_, my_id);

      return ret;
    }



  }
} // end of namespace sstmac.
