/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/node/node.h>
#include <sstmac/hardware/nic/networkinterface.h>
#include <sstmac/software/process/operatingsystem.h>
#include <sstmac/software/services/launch/launcher.h>
#include <sstmac/hardware/memory/memorymodel.h>

ImplementFactory(sstmac::hw::node);

namespace sstmac
{
  namespace hw
  {

    using namespace sstmac::sw;


    debug node::dbg_;
    sstmac::sw::allocationstrategy::ptr node::allocation_;

    bool
    node::has_failed()
    {
      return nic_->has_failed();
    }
    void
    node::reset_failed()
    {
      nic_->reset_failed();
    }

    node::node(const sim_parameters::ptr &params) :
        msg_handler_(0), nodeid_(0), params_(params), nsockets_(1)
    {

    }

    node::node(const sim_parameters::ptr &params, const nodeaddress::ptr& my_id) :
        msg_handler_(0), nodeid_(my_id), params_(params), nsockets_(1)
    {
      /** sstkeyword { gui=4; } */
      int ncores = params->get_int_param("node_cores");

      sstmac::sw::hardware_descriptor::ptr desc =
          sstmac::sw::hardware_descriptor::construct(ncores);

      os_ = sstmac::sw::operatingsystem::construct(desc);
      msg_handler_ = message_handler::construct(ptr(this));

      if (!dbg_.is_real())
        {
          dbg_ = debug("<debug> node");
          logger::register_logger(&dbg_);
        }
    }

    void
    node::init_factory_params(const sim_parameters::ptr &params)
    {
      /**
        sstkeyword {
            gui=4;
            docstring=The total number of cores on the node.ENDL
            The number of cores per NUMA block is node_cores/node_sockets.;
        }
      */
      ncores_ = params->reread_int_param("node_cores");
      /**
        sstkeyword {
            gui=4;
            docstring=The total number of distinct sockets on the node.ENDL
            Taken to be synonymous with the number of distinct NUMA blocks.;
        }
      */
      nsockets_ = params->get_optional_int_param("node_sockets", 1);
    }

    node::~node()
    {
    }

    void
    node::set_nic(const networkinterface::ptr &n)
    {
      nic_ = n;
    }

    boost::intrusive_ptr<networkinterface>
    node::get_nic() const
    {
      return nic_;
    }

    sstmac::sw::operatingsystem::ptr
    node::get_os() const
    {
      return os_;
    }

    memorymodel::ptr
    node::get_mem() const
    {
        return mem_model_;
    }

    eventhandler::ptr
    node::get_msg_handler() const
    {
        return msg_handler_;
    }

    std::string
    node::to_string() const
    {
        return this->get_id()->to_string();
    }

    void
    node::initialize(eventmanager::ptr man, const topology::ptr& top)
    {
      SSTMAC_DEBUG << "starting node initialization \n";
      eventmodule::set_eventmanager(man);
      os_->set_eventmanager(man);
      man->register_finish(ptr(this));

      if (!nic_){
          throw valueerror("node::initialize: nic has not been set yet");
      }

      /**
        sstkeyword {
            gui=instant;
            docstring=How the application is actually launched on the nodes.ENDL
            Simplest method is to assume an instant launch at time=0, butENDL
            more complicated schemes might simulate a varying launch delay.;
        }
      */
      launcher_ = launcher_factory::get_param("launch_name", sstmac_env::params, top);
      service::ptr temp = launcher_;
      os_->register_ser(temp);

      ptr self(this);
      service_factory::start_services(self, sstmac_env::params);

      SSTMAC_DEBUG << "done with node initialize \n";

    }

    void
    node::finish()
    {
        if (finished_)
            return;

        eventmodule::finish();
        if (nic_){
            nic_->finish();
            nic_ = 0;
        }
        if (os_){
           os_->finish();
           if (launcher_){
                launcher_->finish();
                os_->unregister_lib(launcher_);
                launcher_->unregister_all_libs();
           }
        }
        os_ = 0;
        launcher_ = 0;
        msg_handler_ = 0;
    }

    void
    node::handle(const sst_message::ptr &msg)
    {
      msg_handler_->handle(msg);
    }

    void
    node::add_service(sstmac::sw::service::ptr serv)
    {
      SSTMAC_DEBUG << "adding service called " << serv->to_string()
            << " at node " << nodeid_->to_string() << "\n";
      os_->register_ser(serv);
      serv->start();
    }

    void
    node::message_handler::handle(const sst_message::ptr &msg)
    {
      if (parent_ == 0)
        {
          throw ssterror("node::message_handler - my parent is null!");
        }

      if (parent_->os_ == 0)
        {
          throw ssterror("node::message_handler - my parent's os is null!");
        }
      SSTMAC_DEBUG << "node[" << parent_->nodeid_->to_string()
            << "]: handling message \n";

      if (msg->get_type() == sst_message::FAILURE)
        {
          parent_->failure_occurred(msg, parent_);
        }
      else
        {
          parent_->os_->handle_message(msg);
        }

      if (parent_->has_failed())
        {
          //reset hardware to be not-failed (self-healing)
          parent_->cancel_all_messages();
          parent_->reset_failed();
          parent_->launcher_->apps_failed();
        }

      parent_->os_->cleanup();
    }

  }
} // end of namespace sstmac

