/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/packetflow/packetflow_nic.h>
#include <sstmac/common/errors.h>
#include <sstmac/software/process/operatingsystem.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("packetflow", networkinterface, packetflow_nic,
        "implements a nic that models messages as a packet flow");

    // Pointer types.
    packetflow_nic::packetflow_nic()
    {
    }

    packetflow_nic::ptr
    packetflow_nic::construct(const sim_parameters::ptr& params)
    {
        return new packetflow_nic;
    }

    //
    // Goodbye.
    //
    packetflow_nic::~packetflow_nic() throw ()
    {
    }

    networkinterface::ptr
    packetflow_nic::clone(const node::ptr &parent,
        const sstmac::hw::interconnect::ptr &interconn,
        const sim_parameters::ptr& params) const
    {
      packetflow_nic::ptr ret = packetflow_nic::construct(params);
      ret->init(parent, interconn);
      return ret;
    }

    void
    packetflow_nic::init(const node::ptr &parent, const interconnect::ptr& interconn)
    {
      if (!parent->get_id())
      {
        return; //just a template nic
      }

      endpoint_ = packetflow_endpoint::construct();
      endpoint_->set_exit(parent->get_msg_handler());
      startpoint_ = packetflow_startpoint::construct();
      //mem_ = parent->get_mem();
      networkinterface::init(parent, interconn);
    }

    void
    packetflow_nic::connect(connection_type_t ty, const connectable::ptr &mod)
    {
        switch(ty)
        {
            case connectable::injection:
                set_inj_switch(safe_cast(mod,packetflow_switch));
                break;
            case connectable::ejection:
                set_ej_switch(safe_cast(mod,packetflow_switch));
                break;
            default:
                networkinterface::connect(ty, mod);
                break;
        }
    }

    void
    packetflow_nic::recv(const sst_message::ptr& netmsg)
    {
      endpoint_->handle(netmsg);
    }

    void
    packetflow_nic::set_inj_switch(const packetflow_switch::ptr &sw)
    {
        startpoint_->set_acker(parent_->get_msg_handler());
        //create an input buffer for injection
        packetflow_crossbar_buffer::ptr inj_buffer
            = sw->get_input_buffer(packetflow_switch::injection_port);
        startpoint_->set_entry(inj_buffer);
    }

    void
    packetflow_nic::set_ej_switch(const packetflow_switch::ptr &sw)
    {
      sw->add_ejector(parent_->get_id(), ptr(this));
    }

    void
    packetflow_nic::send(const sst_message::ptr &payload)
    {
        startpoint_->send(payload);
    }

    void
    packetflow_nic::set_eventmanager(const eventmanager::ptr &m)
    {
      networkinterface::set_eventmanager(m);
      startpoint_->set_eventmanager(m);
      endpoint_->set_eventmanager(m);
    }

    void
    packetflow_nic::validate_msg_supported(const sst_message::ptr &msg) const
    {
      switch (msg->get_type())
      {
      case sst_message::MPI:
      case sst_message::HPX:
      case sst_message::SEND_ACK:
      case sst_message::ACTOR:
      case sst_message::NETWORK:
      case sst_message::RDMA_GET_REQ:
      case sst_message::RDMA_GET_RSP:
        return; //valid
      default:
        sst_throw(unimplementederror, "%s: network interface card received invalid message type %s",
            to_string().c_str(), sst_message::tostr(msg->get_type()));
      }
    }

    void
    packetflow_nic::finish()
    {
      if (finished_)
        return;

      networkinterface::finish();
    }


  }
} // end of namespace sstmac.

