#include <sstmac/hardware/packettrain/arbitrator.h>
#include <sstmac/common/eventcallback.h>
#include <sstmac/hardware/packettrain/trainlink.h>
#include <sstmac/common/factories/factory.h>


ImplementFactory(sstmac::hw::output_bw_arbitrator);

namespace sstmac {
    namespace hw {

    debug output_bw_arbitrator::dbg_;
    debug input_vc_allocator::dbg_;
    static fair_bw_arbitrator::input_train input_train_inserter;
    const double output_bw_arbitrator::free_buffer = -1;
    const double input_vc_allocator::free_vc = -1;
    const double input_vc_allocator::queued_vc = -2;

SSTRegister("infinite_buffer",
            output_bw_arbitrator,
            infinite_buffer_arbitrator,
            "assumes an infinite input buffer that packets are arbitrated from");
SSTRegister("double_buffer",
            output_bw_arbitrator,
            double_buffer_arbitrator,
            "assumes an infinite input AND output buffer that packets are arbitrated from");
SSTRegister("fair",
            output_bw_arbitrator,
            fair_bw_arbitrator,
            "assigns a fair bw based solely on the number of packets waiting on the link");
SSTRegister("throttle",
            output_bw_arbitrator,
            throttle_bw_arbitrator,
            "assumes no congestion, but sets a maximum obtainable bandwidth");
SSTRegister("packet",
            output_bw_arbitrator,
            full_packet_arbitrator,
            "simulates a packet model within the packet-flow-train framework");
SSTRegister("optimistic_packet",
            output_bw_arbitrator,
            optimistic_packet_arbitrator,
            "simulates a packet model with more optimistic forwarding within the packet-flow-train framework");


output_bw_arbitrator::output_bw_arbitrator()
    : bw_bytes_(0),
      mod_(0)
{
    if (!dbg_.is_real())
    {
      dbg_ = debug("<debug> train");
    }
}

void
output_bw_arbitrator::uncongested_bw_input(const packettrain::ptr &train)
{
    train->set_initial_bw(bw_bytes_);
    double exit_bw = std::min(bw_bytes_, train->get_input_bw());
    train->set_input_bw(exit_bw);
}

double
output_bw_arbitrator::allocate_output_buffer(const packettrain::ptr &train, double t_now)
{
    return free_buffer;
}

double
output_bw_arbitrator::min_delay(const packettrain::ptr &train) const
{
    double max_bw = std::min(train->get_input_bw(), bw_bytes_);
    return train->num_bytes() / max_bw;
}

void
output_bw_arbitrator::finish()
{
    mod_ = 0;
}

throttle_bw_arbitrator::throttle_bw_arbitrator()
{

}

throttle_bw_arbitrator::ptr
throttle_bw_arbitrator::construct(const sim_parameters::ptr& params)
{
    return new throttle_bw_arbitrator;
}

void
throttle_bw_arbitrator::arbitrate(const packettrain::ptr &new_train, double t_now,
                                  double &t_start_ser, double &t_stop_ser)
{
    new_train->throttle_bw(bw_bytes_);
}

void
throttle_bw_arbitrator::uncongested_input(const packettrain::ptr &train)
{
    uncongested_bw_input(train);
}


fair_bw_arbitrator::fair_bw_arbitrator()
{
}

fair_bw_arbitrator::ptr
fair_bw_arbitrator::construct(const sim_parameters::ptr& params)
{
    return new fair_bw_arbitrator;
}

void
fair_bw_arbitrator::uncongested_input(const packettrain::ptr &train)
{
    uncongested_bw_input(train);
}

int
fair_bw_arbitrator::get_queue_length(const timestamp &t_now, int packet_length_bytes) const
{
    return inputs_.size();
}

int
fair_bw_arbitrator::input_train::num_vc() const
{
    std::map<uint64_t, int>::const_iterator it = vcs_allocated.begin(), end = vcs_allocated.end();
    int num_vc = 0;
    for ( ; it != end; ++it){
        num_vc += it->second;
    }
    return num_vc;
}

std::string
fair_bw_arbitrator::input_train::to_string() const
{
    std::map<uint64_t, int>::const_iterator it = vcs_allocated.begin(), end = vcs_allocated.end();
    std::stringstream sstr; sstr << " [";
    for ( ; it != end; ++it){
        if (it->second)
            sstr << it->first << " ";
    }
    sstr << "]";
    return sstprintf("(%p,%d,%8.4e,%8.4e,%8.4e:%s)",
                     this, num_trains,
                     input_bw, start_time,
                     start_time + time_length,
                     sstr.str().c_str());
}

void
fair_bw_arbitrator::arbitrate(const packettrain::ptr& new_train, double t_now,
                      double& t_start_ser, double& t_stop_ser)
{
    //


    double train_bw = new_train->get_input_bw();
    double max_bw = std::min(train_bw, bw_bytes_);
    double train_bytes = new_train->num_bytes();
    int msg_id = new_train->msg_id();

    t_start_ser = -1;

    SSTMAC_DEBUG << sstprintf("On %s, new input %s\n",
                              to_string().c_str(),
                              new_train->to_string().c_str());

    { std::list<input_train>::iterator it = inputs_.begin(), end = inputs_.end();
    while (it != end){
        std::list<input_train>::iterator tmp = it++;
        input_train& tr = *tmp;
        SSTMAC_DEBUG << sstprintf("On %s, for %s, current interval %s\n",
                                  to_string().c_str(),
                                  new_train->to_string().c_str(),
                                  tr.to_string().c_str());
    } }

    //clear all trains that are done
    std::list<input_train>::iterator it = inputs_.begin(), end = inputs_.end();
    while (it != end){
        std::list<input_train>::iterator tmp = it++;
        input_train& tr = *tmp;
        double tr_stop_time = tr.start_time + tr.time_length;
        double delta_now = tr_stop_time - t_now;
        SSTMAC_DEBUG << sstprintf("On %s, for %s, checking interval %s at %8.4e for delta_t=%8.4e\n",
                                  to_string().c_str(),
                                  new_train->to_string().c_str(),
                                  tr.to_string().c_str(),
                                  t_now, delta_now);
        if (delta_now < 1e-9){ //if less than a nanosecond left, throw it out
            SSTMAC_DEBUG << sstprintf("On %s, deleting interval %s\n",
                                      to_string().c_str(), tr.to_string().c_str());
            inputs_.erase(tmp);
        }
        else {
            break;
        }
    }

    if (inputs_.empty()){
        double total_time = train_bytes / max_bw;
        t_start_ser = t_now;
        t_stop_ser = t_now + total_time;

        input_train& new_tr = *(inputs_.insert(inputs_.end(), input_train_inserter));
        new_tr.num_trains = 1;
        new_tr.input_bw = train_bw;
        new_tr.start_time = t_now;
        new_tr.time_length = total_time;
        new_tr.vcs_allocated[new_train->msg_id()] = 1;
        SSTMAC_DEBUG << sstprintf("On %s, new interval %s\n",
                                  to_string().c_str(), new_tr.to_string().c_str());

        return;
    }


    it = inputs_.begin();
    end = inputs_.end();
    input_train& first = inputs_.front();
    //bring the first one up to the current time
    first.time_length -= (t_now - first.start_time);
    first.start_time = t_now;

    double bytes_left = train_bytes;
    double total_time_to_send = 0;
    double accumulated_delay = 0;
    while (it != end){
        std::list<input_train>::iterator tmp = it++;
        input_train& tr = *tmp;
        if (t_start_ser < 0){
            //check to see if this interval has already allocated vcs to this flow
            bool vc_allocated = tr.vcs_allocated.find(msg_id) != tr.vcs_allocated.end();
            if (vc_allocated){
                SSTMAC_DEBUG << sstprintf("On %s, for %s, vcs already allocated for interval %s\n",
                                          to_string().c_str(), new_train->to_string().c_str(),
                                          tr.to_string().c_str());
                //we cannot yet send!
                continue;
            }
            SSTMAC_DEBUG << sstprintf("On %s, for %s, setting start send at %8.4e\n",
                                      to_string().c_str(), new_train->to_string().c_str(),
                                      tr.start_time);
            t_start_ser = tr.start_time;
        }


        tr.start_time += accumulated_delay;
        tr.num_trains++;

        double fair_bw = bw_bytes_ / tr.num_trains;
        double available_bw = bw_bytes_ - tr.input_bw;
        double allowed_bw = std::max(fair_bw, available_bw);
        double my_bw = std::min(train_bw, allowed_bw);
        //figure out how much the other flows get slowed down by
        double other_flows_new_bw = std::min(tr.input_bw, bw_bytes_ - my_bw);
        double new_interval_length = (tr.input_bw / other_flows_new_bw) * tr.time_length;
        SSTMAC_DEBUG << sstprintf("On %s, for %s, for interval %s:"
                                  "  ntrains=%d interval=%8.4e old=%8.4e new=%8.4e avail=%8.4e fair=%8.4e bw=%8.4e\n",
                                  to_string().c_str(), new_train->to_string().c_str(), tr.to_string().c_str(),
                                  tr.num_trains, new_interval_length,
                                  tr.input_bw, other_flows_new_bw, available_bw, fair_bw, my_bw);
        SSTMAC_DEBUG << sstprintf("On %s, for %s, lengthen interval %s to %8.4e\n",
                                  to_string().c_str(), new_train->to_string().c_str(),
                                  tr.to_string().c_str(), new_interval_length);
        //figure out how many bytes I could send in this interval
        double bytes_sent = my_bw * new_interval_length;
        if (bytes_sent >= bytes_left){
            //the train finishes in this interval
            double time_needed = bytes_left / my_bw;
            //now figure out how much extra time in the remainder interval
            //the other flows will need to finish
            new_interval_length = tr.time_length - other_flows_new_bw * time_needed / tr.input_bw;
            accumulated_delay += new_interval_length + time_needed - tr.time_length;

            //add the accumulated delay to all the other intervals
            while (it != end){
                input_train& tr = *it;
                SSTMAC_DEBUG << sstprintf("On %s, for %s, interval %s delayed by %8.4e\n",
                                          to_string().c_str(), new_train->to_string().c_str(),
                                          tr.to_string().c_str(), accumulated_delay);
                tr.start_time += accumulated_delay;
                ++it;
            }

            t_stop_ser = t_start_ser + total_time_to_send + time_needed;
            if (time_needed < 1e-9){
                tr.num_trains--;
                //no need to make a new interval
                return;
            }

            //make the new train
            ++tmp;
            input_train& new_tr = *inputs_.insert(tmp, input_train_inserter);

            new_tr.num_trains = tr.num_trains - 1;
            new_tr.input_bw = tr.input_bw;
            new_tr.start_time = tr.start_time + time_needed;
            new_tr.time_length = new_interval_length;
            new_tr.vcs_allocated = tr.vcs_allocated;
            tr.vcs_allocated[msg_id] = 1;
            //decrement the time length of the old interval
            tr.time_length = time_needed;
            tr.input_bw = std::min(bw_bytes_, tr.input_bw + my_bw);

            SSTMAC_DEBUG << sstprintf("On %s, for %s, new interval %s and old interval %s\n",
                                      to_string().c_str(), new_train->to_string().c_str(),
                                      new_tr.to_string().c_str(), tr.to_string().c_str());
            return;
        }
        SSTMAC_DEBUG << sstprintf("On %s, for %s, sent %8.4e bytes at bw %8.4e"
                                  " in interval %s\n",
                                  to_string().c_str(), new_train->to_string().c_str(),
                                  bytes_sent, my_bw, tr.to_string().c_str());
        //add this train's bw to the bw for this interval
        tr.vcs_allocated[msg_id] = 1;
        if (tr.num_trains != (int) tr.vcs_allocated.size()){
            std::cerr << "too many trains: " << tr.to_string() << std::endl;
            abort();
        }
        accumulated_delay += new_interval_length - tr.time_length;
        tr.time_length = new_interval_length;
        tr.input_bw = std::min(bw_bytes_, tr.input_bw + my_bw);
        total_time_to_send += tr.time_length;
        bytes_left -= bytes_sent;
    }

    //we take longer than all current intervals
    double total_time_left = bytes_left / max_bw;
    if (t_start_ser < 0){
        const input_train& last_train = inputs_.back();
        t_start_ser = last_train.start_time + last_train.time_length;
    }
    t_stop_ser = t_start_ser + total_time_to_send + total_time_left;
    if (total_time_left < 1e-9){
        return; //no need for a new interval
    }

    input_train& new_tr = *(inputs_.insert(inputs_.end(), input_train_inserter));
    new_tr.num_trains = 1;
    new_tr.input_bw = max_bw;
    new_tr.start_time = t_start_ser + total_time_to_send;
    new_tr.time_length = total_time_left;
    new_tr.vcs_allocated[msg_id] = 1;

    {double t_last_stop = 0;
    std::list<input_train>::iterator it = inputs_.begin(), end = inputs_.end();
    while (it != end){
        std::list<input_train>::iterator tmp = it++;
        input_train& tr = *tmp;
        SSTMAC_DEBUG << sstprintf("On %s, for %s, current interval %s\n",
                                  to_string().c_str(),
                                  new_train->to_string().c_str(),
                                  tr.to_string().c_str());
        if ((t_last_stop - tr.start_time) > 1e-9){
            std::cerr << "input trains are not ordered correctly" << std::endl;
            abort();
        }
        t_last_stop = tr.start_time + tr.time_length;
    } }


    SSTMAC_DEBUG << sstprintf("On %s, for %s, new interval %s\n",
                              to_string().c_str(), new_train->to_string().c_str(),
                              new_tr.to_string().c_str());

}


void
infinite_buffer_arbitrator::arbitrate(const packettrain::ptr &new_train, double t_now,
                                      double &t_start_ser, double &t_stop_ser)
{
      ++num_trains_;
      double delta_t = t_now - last_input_update_;
      SSTMAC_DEBUG << sstprintf(
            "On %s, %s, sending new train %d after delta_t=%8.4e\n",
            to_string().c_str(), stats().c_str(), new_train->msg_id(), delta_t);
      update_input_bytes(delta_t);

      SSTMAC_DEBUG << sstprintf("On %s, %s, sending %s as \n",
            to_string().c_str(), stats().c_str(),
            new_train->to_string().c_str(), num_trains_);

      input_buffer_fill_bw_ += new_train->get_input_bw();
      pending_bytes_ += new_train->num_bytes();

      update_bw();

      double time_to_arrive = new_train->ser_delay();
      double t_last_packet_arrival = t_now + time_to_arrive;
      //when my last packet arrives, how many bytes will be waiting in the buffers
      double bytes_accumulated_in_buffer = 0;
      if (t_last_packet_arrival >= last_arrival_)
      {
        // my packets finish arriving after all the other packets
        bytes_accumulated_in_buffer = pending_bytes_ - bw_bytes_*time_to_arrive;
        last_arrival_ = t_last_packet_arrival;
      }
      else
      {
          // my packets finish arriving before some of the other packets
          // figure out how many bytes of other messages will arrive when I finish arriving
          double agg_input_bw = pending_bytes_ / (last_arrival_ - t_now);
          double avg_input_bw = 0.5 * (agg_input_bw + input_buffer_fill_bw_);
          bytes_accumulated_in_buffer = time_to_arrive * (avg_input_bw - bw_bytes_);
          // add in my bytes
          bytes_accumulated_in_buffer += new_train->num_bytes();
          SSTMAC_DEBUG << sstprintf("On %s, while sending %s, agg input bw is %8.4e\n",
                                    to_string().c_str(), new_train->to_string().c_str(),
                                    agg_input_bw);
      }

      SSTMAC_DEBUG << sstprintf(
            "On %s, while sending %s, accumulate %8.4e extra bytes\n",
            to_string().c_str(), new_train->to_string().c_str(),
            bytes_accumulated_in_buffer);
      double buffered_bytes_at_end = std::max((double) 0, input_buffer_bytes_ + bytes_accumulated_in_buffer);
      SSTMAC_DEBUG << sstprintf(
                "On %s, when %s finishes arriving after %8.4e, there will be %8.4e buffered bytes\n",
                to_string().c_str(), new_train->to_string().c_str(),
                time_to_arrive, buffered_bytes_at_end);

      // when the new train is done arriving, figure out how long it takes to
      // clear the buffers - that will be the time I finish sending
      double extra_time_to_clear_buffers = buffered_bytes_at_end / bw_bytes_;
      double total_time_to_send = time_to_arrive + extra_time_to_clear_buffers;

      schedule_stop_input(t_now + new_train->ser_delay(),
                          new_train->get_input_bw());

      SSTMAC_DEBUG << sstprintf("On %s, for %s, scheduling stop output at %8.4e\n",
                                to_string().c_str(), new_train->to_string().c_str(), t_now + new_train->ser_delay());


      t_start_ser = t_now;
      t_stop_ser = t_now + total_time_to_send;

      //std::cout << sstprintf("%8.4e: arbitrating for %d trains with total bandwidth of %8.4e\n",
      //                       t_now, num_trains_, bw_bytes_);

      last_input_update_ = t_now;
}

void
infinite_buffer_arbitrator::uncongested_input(const packettrain::ptr &train)
{
    uncongested_bw_input(train);
}

double
infinite_buffer_arbitrator::allocate_output_buffer(const packettrain::ptr &train, double t_now)
{
    return free_buffer;
}

void
infinite_buffer_arbitrator::stop_input(double t_now, double input_bw)
{
  SSTMAC_DEBUG << sstprintf("On %s, stopping input of %8.4e B/s\n",
                            to_string().c_str(), input_bw);

  double delta_t = t_now - last_input_update_;
  update_input_bytes(delta_t);

  input_buffer_fill_bw_ -= input_bw;

  update_bw();

  --num_trains_;

  last_input_update_ = t_now;
}

std::string
infinite_buffer_arbitrator::to_string() const
{
    return sstprintf("infinite buffer arbitrator %p", this);
}

int
infinite_buffer_arbitrator::get_queue_length(const timestamp &t_now, int packet_length_bytes) const
{
    double delta_t = t_now.sec() - last_input_update_;
    double bytes_left = input_buffer_bytes_ - delta_t * bw_bytes_;
    int packets_left = bytes_left / packet_length_bytes;
    return std::max(0, packets_left);
}

void
infinite_buffer_arbitrator::schedule_stop_input(double t_stop, double input_bw)
{
    timestamp stop_time(t_stop);
    event::ptr ev = new_event(ptr(this), &infinite_buffer_arbitrator::stop_input, t_stop, input_bw);
    mod_->schedule(stop_time, ev);
}

void
infinite_buffer_arbitrator::update_bw()
{
  input_buffer_fill_bw_ = input_buffer_fill_bw_ < 0.0001 ? 0 : input_buffer_fill_bw_;
  utilization_bw_ = input_buffer_bytes_ > 0 ? bw_bytes_ : std::min(bw_bytes_, input_buffer_fill_bw_);
}

void
infinite_buffer_arbitrator::update_input_bytes(double delta_t)
{
  //update stuff from previous flow
  input_buffer_bytes_ += (input_buffer_fill_bw_ - utilization_bw_) * delta_t;
  input_buffer_bytes_ = input_buffer_bytes_ < 0.1 ? 0 : input_buffer_bytes_;

  pending_bytes_ -= input_buffer_fill_bw_ * delta_t;
  pending_bytes_ = pending_bytes_ < 0.1 ? 0 : pending_bytes_;
}

infinite_buffer_arbitrator::ptr
infinite_buffer_arbitrator::construct(const sim_parameters::ptr& params)
{
    return new infinite_buffer_arbitrator;
}

infinite_buffer_arbitrator::infinite_buffer_arbitrator()
    : num_trains_(0),
      input_buffer_bytes_(0),
      input_buffer_fill_bw_(0),
      utilization_bw_(0),
      pending_bytes_(0),
      last_arrival_(0),
      last_input_update_(0)
{
}

packet_arbitrator::packet_arbitrator()
    : next_free_(0)
{
}

void
full_packet_arbitrator::arbitrate(
    const packettrain::ptr& new_train,
    double t_now,
    double& t_start_ser,
    double& t_stop_ser)
{
    double ser_latency = new_train->num_bytes() / bw_bytes_;
    double t_next_free = t_now > next_free_ ? t_now : next_free_;
    next_free_ = t_start_ser = t_stop_ser = t_next_free + ser_latency;
    new_train->set_input_bw(bw_bytes_);
}

double
packet_arbitrator::min_delay(const packettrain::ptr &train) const
{
    return train->num_bytes() / bw_bytes_;
}

int
packet_arbitrator::get_queue_length(const timestamp &t_now, int packet_length_bytes) const
{
    double t_now_s = t_now.sec();
    if (next_free_ < t_now_s){
        return 0;
    }
    double bytes_left = (next_free_ - t_now_s) * bw_bytes_;
    int packets_left = bytes_left / packet_length_bytes;
    return packets_left;
}

void
packet_arbitrator::uncongested_input(const packettrain::ptr &train)
{
    train->set_input_bw(bw_bytes_);
}

double
packet_arbitrator::get_util_bw() const
{
    return 0;
}

full_packet_arbitrator::ptr
full_packet_arbitrator::construct(const sim_parameters::ptr& params)
{
    return new full_packet_arbitrator;
}

double
full_packet_arbitrator::allocate_output_buffer(const packettrain::ptr& train, double t_now)
{
    return free_buffer;
}


optimistic_packet_arbitrator::ptr
optimistic_packet_arbitrator::construct(const sim_parameters::ptr& params)
{
    return new optimistic_packet_arbitrator;
}

double
optimistic_packet_arbitrator::allocate_output_buffer(const packettrain::ptr& train, double t_now)
{
    if (train->get_input_bw() < 0){ //unitialized - assume immediately ready
        return free_buffer;
    }
    else if ((train->get_input_bw() - bw_bytes_) > -1e-6){
        return free_buffer; //this will fill the output buffer faster than we can empty it
    }
    else {
        //delay just long enough that so that input buffering can keep up with the send
        double time_needed_to_buffer = train->ser_delay();
        double time_to_send = train->num_bytes() / bw_bytes_;
        double alignment_delay = time_needed_to_buffer - time_to_send;
        return alignment_delay;
    }
}

void
optimistic_packet_arbitrator::arbitrate(
    const packettrain::ptr& new_train,
    double t_now,
    double& t_start_ser,
    double& t_stop_ser)
{
    double ser_latency = new_train->num_bytes() / bw_bytes_;
    double t_next_free = t_now > next_free_ ? t_now : next_free_;
    t_start_ser = t_now;
    next_free_ = t_stop_ser = t_next_free + ser_latency;
    new_train->set_input_bw(bw_bytes_);
}

double_buffer_arbitrator::double_buffer_arbitrator()
    : output_buffer_bytes_(0),
      output_buffer_fill_bw_(0),
      last_output_update_(0)
{
}

double_buffer_arbitrator::ptr
double_buffer_arbitrator::construct(const sim_parameters::ptr& params)
{
    return new double_buffer_arbitrator;
}

void
double_buffer_arbitrator::arbitrate(const packettrain::ptr &new_train, double t_now, double &t_start_ser, double &t_stop_ser)
{
    // whatever bw the train is leaving with, assume an "output" buffer
    // is being filled at exit bw rate
    double output_buf_bw = std::min(new_train->get_input_bw(), bw_bytes_);
    double output_buf_time = new_train->num_bytes() / output_buf_bw;
    double delta_output_update = t_now - last_output_update_;
    update_output_bytes(delta_output_update);
    last_output_update_ = t_now;
    output_buffer_fill_bw_ += output_buf_bw;
    schedule_stop_output(t_now + output_buf_time, output_buf_bw);

    infinite_buffer_arbitrator::arbitrate(new_train, t_now, t_start_ser, t_stop_ser);
}

void
double_buffer_arbitrator::update_output_bytes(double delta_t)
{
    double util_bw = output_buffer_bytes_ > 0 ? bw_bytes_ : std::min(bw_bytes_, output_buffer_fill_bw_);
    output_buffer_bytes_ += (output_buffer_fill_bw_ - util_bw) * delta_t;
    output_buffer_bytes_ = std::max(output_buffer_bytes_, (double) 0);
}

double
double_buffer_arbitrator::allocate_output_buffer(const packettrain::ptr &train, double t_now)
{
    if (train->is_head()){
        return free_buffer;
    }

    double delta_t = t_now - last_output_update_;
    update_output_bytes(delta_t);

    //this cannot go until the output buffer is empty
    //use a loose cutoff of 64 for now
    if (output_buffer_bytes_ > 64){
        double time_to_finish = output_buffer_bytes_ / bw_bytes_;
        SSTMAC_DEBUG << sstprintf("On %s, for %s, waiting %8.4e seconds to clear output buffer of occupancy %8.4e\n",
                                  to_string().c_str(), train->to_string().c_str(),
                                  time_to_finish, output_buffer_bytes_);
        return time_to_finish;
    }
    else {
        return free_buffer;
    }
}

void
double_buffer_arbitrator::stop_output(double t_now, double output_bw)
{
    double delta_t = t_now - last_output_update_;

    update_output_bytes(delta_t);

    SSTMAC_DEBUG << sstprintf("On %s, stopping output at %8.4e after delta_t=%8.4e from last update %8.4e.  New buffer occupancy is %8.4e from fill bw %8.4e\n",
                              to_string().c_str(), t_now, delta_t, last_output_update_, output_buffer_bytes_, output_buffer_fill_bw_);

    output_buffer_fill_bw_ -= output_bw;

    last_output_update_ = t_now;
}

void
double_buffer_arbitrator::schedule_stop_output(double t_stop, double output_bw)
{

    timestamp stop_time(t_stop);
    event::ptr ev = new_event(ptr(this), &double_buffer_arbitrator::stop_output, t_stop, output_bw);
    mod_->schedule(stop_time, ev);
}

std::string
double_buffer_arbitrator::to_string() const
{
    return sstprintf("double buffer arbitrator %p", this);
}

input_vc_allocator::input_vc_allocator()
{
    if (!dbg_.is_real()){
        dbg_ = debug("<debug> train");
    }
}

double
uncongested_vc_allocator::allocate_vc(double t_now, const packettrain::ptr &train)
{
    return free_vc;
}

packettrain::ptr
uncongested_vc_allocator::deallocate_vc(double t_stop_send, const packettrain::ptr &train)
{
    return 0;
}

void
inorder_vc_allocator::finish()
{
    input_vc_allocator::finish();
}

double
inorder_vc_allocator::allocate_vc(double t_now, const packettrain::ptr &train)
{
    size_t prev_size = vc_trains_.size();
    packettrain::ptr& existing = vc_trains_[train->msg_id()];
    if (prev_size != vc_trains_.size()){
        return free_vc;
        // I am the first
    }
    else if (existing){
        // not the first in the queue
        existing->add_next(train);
        return queued_vc;
    }
    else {
        // I am not the first, but I am the first to queue up
        existing = train;
        return queued_vc;
    }
}

packettrain::ptr
inorder_vc_allocator::deallocate_vc(double free_time, const packettrain::ptr& train)
{
    vc_train_map::iterator it = vc_trains_.find(train->msg_id());
    if (it == vc_trains_.end()){
        sst_throw(illformederror, "inorder_vc_allocator::deallocate_vc no trains");
    }

    packettrain::ptr next = it->second;
    if (next){
        //update the map
        it->second = next->pop_next();
        return next;
    }
    else {
        //nothing else doing
        vc_trains_.erase(it);
        return 0;
    }
}

    }
}

