#ifndef ARBITRATOR_H
#define ARBITRATOR_H

#include <sstmac/hardware/packettrain/packettrain.h>
#include <sstmac/common/factories/factory.h>

namespace sstmac
{

  namespace hw
  {

    /**
      @class output_bw_arbitrator
      Class that arbitrates access to the output buffer and allocates
      bw to the output link.  This applies to network links, switches, etc.
    */
    class output_bw_arbitrator :
            public factory_type
    {
    protected:
      static debug dbg_;

      double bw_bytes_;

      eventmodule::ptr mod_;

      output_bw_arbitrator();

      /**
        Compute and assign bandwidth train would achieve
        in the absence of any congestion. This is used by
        several subclasses to implement #uncongested_input.
        This implements the default behavior of
        bw = min(train_bw, my_bw)
        @param train The train being assigned bw
      */
      void
      uncongested_bw_input(const packettrain::ptr& train);

    public:
      typedef boost::intrusive_ptr<output_bw_arbitrator> ptr;

      virtual
      ~output_bw_arbitrator()
      {
      }

      virtual output_bw_arbitrator::ptr clone() const = 0;

      /**
        Flag used to indicate the output buffer is free immediately
        in place of a positive value that indicates a future time
        that output buffer will be free.
      */
      const static double free_buffer;

      /**
        Arbitrate based on current congestion and assign a train
        a start/stop time based on available bandwidth.
        @param train The train arbitrating for bandwidth
        @param t_now The current simulation time in seconds
        @param[inout] t_start_ser The time the train will start serializing/sending
                      through the link/switch
        @param[inout] t_stop_ser  The time the train will stop serializing/sending
                      through the link/switch
      */
      virtual void
      arbitrate(const packettrain::ptr& train, double t_now,
          double& t_start_ser, double& t_stop_ser) = 0;

      /**
        Arbitrate for access to the output buffer and figure out
        when it will next be free
        @param train The train trying to gain access to the output buffer
        @param t_now The current simulation time in seconds
        @return The time at which the output buffer will be free. A value of #free_buffer
                indicates the buffer is free immediately
      */
      virtual double
      allocate_output_buffer(const packettrain::ptr& train, double t_now);

      /**
        Compute and assign bandwidth train would achieve
        in the absence of any congestion.
        @param train The train being assigned bw
      */
      virtual void
      uncongested_input(const packettrain::ptr& train) = 0;

      /**
        Set the bandwidth for this link - usually called following a clone
        @param bw_bytes The bandwidth in bytes/second
      */
      virtual void
      set_bw(double bw_bytes)
      {
        bw_bytes_ = bw_bytes;
      }

      /**
        Set the eventmodule this will use for scheduling events
        @param mod The eventmodule this will use for scheduling events
      */
      virtual void
      set_eventmodule(const eventmodule::ptr& mod)
      {
        mod_ = mod;
      }

      virtual void
      finish();

      virtual int
      get_queue_length(const timestamp& t_now, int packet_length_bytes) const = 0;

      /**
        Compute the minimum delay from arrival to departure a train would
        experience in the absence of congestion
        @param train The train being sent
        @return The minimum delay in seconds
      */
      virtual double
      min_delay(const packettrain::ptr& train) const;

      /**
        Compute the next time the arbitrator is ''free.'' The exact physical meaning
        will vary depending on the arbitrator.  However, higher valules indicate more
        congestion.
        @return The next uncongested time returned as a delay (relative from now) in seconds
      */
      virtual double
      next_free() const = 0;

      /**
       * needed for vis
       * @return
       */
      virtual double
      get_util_bw() const {
        return 0;
      }

    };

    DeclareFactory(output_bw_arbitrator);

    /**
      @class fair_bw_arbitrator
      Output bw arbitrator that assigns bandwidth by simply dividing the max bw
      by the total number of trains
    */
    class fair_bw_arbitrator :
            public output_bw_arbitrator
    {
    public:
      struct input_train
      {
        std::map<uint64_t, int> vcs_allocated;
        int num_trains;
        double start_time;
        double time_length;
        double input_bw;

        std::string
        to_string() const;

        int
        num_vc() const;

        /**
         * GH: not sure if this one is correct.
         * @return
         */

      };

    private:

      std::list<input_train> inputs_;

    public:
      typedef boost::intrusive_ptr<fair_bw_arbitrator> ptr;

      output_bw_arbitrator::ptr
      clone() const {
          return new fair_bw_arbitrator;
      }

      fair_bw_arbitrator();

      virtual int
      get_queue_length(const timestamp& t_now, int packet_length_bytes) const;

      virtual void
      arbitrate(const packettrain::ptr& new_train, double t_now,
          double& t_start_ser, double& t_stop_ser);

      static ptr
      construct(const sim_parameters::ptr& params);

      std::string
      to_string() const
      {
        return "fair bw arbitrator";
      }

      virtual void
      uncongested_input(const packettrain::ptr& train);

      /**
        @throws There is no notion of next_free or congestion with fair_bw arbitrator
      */
      virtual double
      next_free() const
      {
        sst_throw(ssterror, "fair_bw_arbitrator::next_free: I don't know how to estimate congestion");
      }

      virtual double
      get_util_bw() const
      {
        double sum = 0;
        std::list<input_train>::const_iterator it, end = inputs_.end();
        for(it = inputs_.begin(); it != end; it++){
          sum += it->input_bw;
        }

        return std::min(sum, bw_bytes_);
      }

    };

    class packet_arbitrator :
        public output_bw_arbitrator
    {
        protected:
            double next_free_;

        public:
          typedef boost::intrusive_ptr<packet_arbitrator> ptr;

          packet_arbitrator();

          virtual double
          next_free() const
          {
            return next_free_;
          }

          virtual double
          min_delay(const packettrain::ptr& train) const;

          virtual void
          uncongested_input(const packettrain::ptr& train);

          virtual int
          get_queue_length(const timestamp& t_now, int packet_length_bytes) const;

          virtual double
          get_util_bw() const;

    };

    /**
      @class full_packet_arbitrator
      Output bw arbitrator that arbitrates complete, discrete packets without using
      any flow-like characteristics.  Only one packet is serialized on a link
      at a given time.  The entire packet must be in the output buffer
      before beginning serialization.
    */
    class full_packet_arbitrator :
        public packet_arbitrator
    {
    protected:


    public:
      typedef boost::intrusive_ptr<full_packet_arbitrator> ptr;

      output_bw_arbitrator::ptr clone() const {
          return new full_packet_arbitrator;
      }

      static ptr
      construct(const sim_parameters::ptr& params);

      virtual double
      allocate_output_buffer(const packettrain::ptr& train, double t_now);

      std::string
      to_string() const {
        return "full packet arbitrator";
      }

      virtual void
      arbitrate(const packettrain::ptr &new_train, double t_now,
          double &t_start_ser, double &t_stop_ser);

    };

    /**
      @class full_packet_arbitrator
      Output bw arbitrator that arbitrates complete, discrete packets without using
      any flow-like characteristics.  Only one packet is serialized on a link
      at a given time.  Once the packet begins filling the output buffer,
      it can begin serializing on the output link to simulate pipelining
      of packets.  This is in contrast to the #full_packet_arbitrator.
    */
    class optimistic_packet_arbitrator :
        public packet_arbitrator
    {

    public:
      typedef boost::intrusive_ptr<optimistic_packet_arbitrator> ptr;

      output_bw_arbitrator::ptr clone() const {
          return new optimistic_packet_arbitrator;
      }

      static ptr
      construct(const sim_parameters::ptr& params);

      virtual double
      allocate_output_buffer(const packettrain::ptr& train, double t_now);

      std::string
      to_string() const {
        return "optimistic packet arbitrator";
      }

      virtual void
      arbitrate(const packettrain::ptr &new_train, double t_now,
          double &t_start_ser, double &t_stop_ser);

    };

    /**
      @class infinite_buffer_arbitrator
      This assumes the arbitrator can receive all incoming trains and store them in an infinite buffer.
      Packets are arbitrated from the infinite buffer to simulate first-in,first-out order.
    */
    class infinite_buffer_arbitrator :
        public output_bw_arbitrator
    {
    protected:
      int num_trains_;

      double input_buffer_bytes_;

      double input_buffer_fill_bw_;

      double utilization_bw_;

      double pending_bytes_;

      double last_arrival_;

      double last_input_update_;

      bool multipath_;

      std::string
      stats() const
      {
        return sstprintf(
            "inp-buff=%8.4e pend=%8.4e inp-buff-bw=%8.4e util=%8.4e",
            input_buffer_bytes_, pending_bytes_, input_buffer_fill_bw_,
            utilization_bw_);
      }

      /**
        Update the number of bytes in the infinite buffer
        @param delta_t Time since last update in seconds
      */
      void
      update_input_bytes(double delta_t);

      /**
        Update the bandwidth terms based on buffer occupancies
        and the aggregate input bw from trains
      */
      void
      update_bw();

      int
      get_queue_length(const timestamp &t_now, int packet_length_bytes);

      /**
        Notification that incoming train has finished arriving
        and is now completely buffered, i.e. stopped input.
        Update buffer occupancies and bandwidth terms.
        @param stop_time The time the input is being stopped
        @param input_bw The incoming bw that will stop when
                        buffereing is completed
      */
      void
      stop_input(double stop_time, double input_bw);

      /**
        This schedules a #stop_input event at some point
        in the future
        @param stop_time The time the input will be stopped
        @param input_bw The bw (bytes/sec) that has finished buffered
      */
      void
      schedule_stop_input(double stop_time, double input_bw);

      infinite_buffer_arbitrator();

    public:
      typedef boost::intrusive_ptr<infinite_buffer_arbitrator> ptr;

      virtual
      ~infinite_buffer_arbitrator()
      {
      }

      output_bw_arbitrator::ptr clone() const {
          return new infinite_buffer_arbitrator;
      }

      static ptr
      construct(const sim_parameters::ptr& params);

      virtual void
      arbitrate(const packettrain::ptr &new_train, double t_now,
          double &t_start_ser, double &t_stop_ser);

      virtual double
      allocate_output_buffer(const packettrain::ptr& train, double t_now);

      std::string
      to_string() const;

      virtual void
      uncongested_input(const packettrain::ptr& train);

      virtual int
      get_queue_length(const timestamp& t_now, int packet_length_bytes) const;

      virtual double
      next_free() const
      {
        if (utilization_bw_ == 0)
        {
          return 0;
        }
        else
          return input_buffer_bytes_ / utilization_bw_;
      }

      virtual double
      get_util_bw() const
      {
        return utilization_bw_;
      }

    };

    /**
      @class double_buffer_arbitrator
      This assumes the arbitrator can receive all incoming trains and store them in an infinite buffer.
      Packets are arbitrated from the infinite buffer to simulate first-in,first-out order.  This
      adds functionality to #infinite_buffer_arbitrator.  Before being allowed to compete for infinite
      buffer to send, packets must arbitrate for access to an output buffer.
    */
    class double_buffer_arbitrator :
        public infinite_buffer_arbitrator
    {
    protected:
      double output_buffer_bytes_;

      double output_buffer_fill_bw_;

      double last_output_update_;

      /**
        Update the number of bytes in the output buffer
        @param delta_t Time since last update in seconds
      */
      void
      update_output_bytes(double delta_t);

      /**
        Notification that incoming train has finished arriving
        in the output buffer is now completely buffered.
        Update buffer occupancies and bandwidth terms.
        @param stop_time The time the train has finished filling the output buffer
        @param input_bw The incoming bw that will stop when
                        buffereing is completed
      */
      void
      stop_output(double stop_time, double output_bw);

      /**
        This schedules a #stop_output event at some point
        in the future
        @param stop_time The time the output buffer will stop being filled
        @param input_bw The bw (bytes/sec) that has finished buffered
      */
      void
      schedule_stop_output(double stop_time, double output_bw);

    public:
      typedef boost::intrusive_ptr<double_buffer_arbitrator> ptr;

      double_buffer_arbitrator();

      output_bw_arbitrator::ptr clone() const {
          return new double_buffer_arbitrator;
      }

      static ptr
      construct(const sim_parameters::ptr& params);

      virtual void
      arbitrate(const packettrain::ptr &new_train, double t_now,
          double &t_start_ser, double &t_stop_ser);

      virtual double
      allocate_output_buffer(const packettrain::ptr& train, double t_now);

      std::string
      to_string() const;



    };

    /**
      @class throttle_bw_arbitrator
      Implements arbitration by setting a train's bw less than or equal to some maximum.
      BW allocation doesn't monitor congestion - it just sets some predefined maximum.
    */
    class throttle_bw_arbitrator : public output_bw_arbitrator
    {

    public:
      typedef boost::intrusive_ptr<throttle_bw_arbitrator> ptr;

      throttle_bw_arbitrator();

      output_bw_arbitrator::ptr clone() const {
          return new throttle_bw_arbitrator;
      }

      static ptr
      construct(const sim_parameters::ptr& params);

      virtual void
      arbitrate(const packettrain::ptr& new_train, double t_now,
          double& t_start_ser, double& t_stop_ser);

      std::string
      to_string() const
      {
        return "throttle bw arbitrator";
      }

      virtual void
      uncongested_input(const packettrain::ptr& train);

      /**
        @throws There is no notion of next_free or congestion with throttle_bw arbitrator
      */
      virtual double
      next_free() const
      {
          sst_throw(ssterror, "throttle_bw_arbitrator::next_free: I don't know how to estimate congestion");
      }

      virtual int
      get_queue_length(const timestamp& t_now, int packet_length_bytes) const {
        return 0;
      }

      virtual double
      get_util_bw() const
      {
        return bw_bytes_;
      }

    };

    /**
      @class input_vc_allocator
      When packet trains first arrive at a link/switch, they must be allocate
      a virtual channel before progressing.
    */
    class input_vc_allocator : public ptr_type
    {
    protected:
      static debug dbg_;

      input_vc_allocator();

    public:
      virtual
      ~input_vc_allocator()
      {
      }

      typedef boost::intrusive_ptr<input_vc_allocator> ptr;

      /**
        Flag indicating that a virtual channel is immediately free for use
      */
      static const double free_vc;

      /**
        Flag indicating that no virtual channel is free. The incoming train needs
        to queue up behind some other train on its vc.
      */
      static const double queued_vc;

      /**
        Allocate an input virtual channel for the incoming train
        @param t_now The current simulation time in seconds
        @param train The incoming train
        @return The next (absolute) time in seconds that a vc is free.
                Might return a special flag indicated free/queue state.
                See #free_vc and #queued_vc.
      */
      virtual double
      allocate_vc(double t_now, const packettrain::ptr& train) = 0;

      /**
        Deallocate the input virtual channel for other trains.
        @param t_now The current simulation time in seconds
        @param train The train currently using the vc
        @return Any train that was queued up waiting and now needs
                to try allocating again
      */
      virtual packettrain::ptr
      deallocate_vc(double t_now, const packettrain::ptr& train) = 0;

      virtual void
      finish()
      {
      }

    };

    /**
      @class inorder_vc_allocator
      Virtual channel allocator that orders incoming packet trains.  Only one packet train
      from a message is allowed access at any given time.  All distinct messages are
      allowed access.  There is no upper limit to the number of simultaneously active
      virtual channels - only that only one train per message is active at any given time.
    */
    class inorder_vc_allocator : public input_vc_allocator
    {
    protected:

      typedef boost::unordered_map<uint64_t, packettrain::ptr> vc_train_map;
      vc_train_map vc_trains_;

    public:
      typedef boost::intrusive_ptr<inorder_vc_allocator> ptr;

      virtual double
      allocate_vc(double t_now, const packettrain::ptr& train);

      virtual packettrain::ptr
      deallocate_vc(double t_now, const packettrain::ptr& train);

      std::string
      to_string() const
      {
        return sstprintf("inorder_vc_allocator %p", this);
      }

      virtual void
      finish();

    };

    /**
      @class uncongested_vc_allocator
      Virtual channel allocator that always allocates an input vc
      to any packet train regardless of how many other trains are
      active.  Congestion is essentially ignored.  In contrast to
      #inorder_vc_allocator, many trains from the same message
      can be active at the same time.
    */
    class uncongested_vc_allocator : public input_vc_allocator
    {
    public:
      typedef boost::intrusive_ptr<uncongested_vc_allocator> ptr;

      virtual double
      allocate_vc(double t_now, const packettrain::ptr& train);

      virtual packettrain::ptr
      deallocate_vc(double t_now, const packettrain::ptr& train);

      std::string
      to_string() const
      {
        return sstprintf("uncongested_vc_allocator %p", this);
      }

    };

  }

}

#endif // ARBITRATOR_H
