#ifndef CONGESTION_H
#define CONGESTION_H

#include <sstmac/hardware/packettrain/packettrain.h>
#include <sstmac/hardware/packettrain/arbitrator.h>
#include <sstmac/hardware/common/component.h>
#include <sstmac/common/ptr_type.h>
#include <sstmac/common/eventmodule.h>

namespace sstmac {
    namespace hw {


    class traincongestionpoint;
    typedef boost::intrusive_ptr<traincongestionpoint> traincongestionpointPtr;

    /**
      @class stats_congestion
      Stats logger for congestion occurring on network/memory links.  There is
      a separate stat_logger for each type of component - memory,injection,network
    */
    class stats_congestion
        : public stat_logger
    {
        private:
            /**
              @class link_stats
              Private class that tracks statistics for a single component
            */
            struct link_stats {
                double kbytes_sent;
                double ms_delayed;
                link_stats();
            };

            boost::unordered_map<traincongestionpointPtr, link_stats> link_stats_;

        public:
            typedef boost::intrusive_ptr<stats_congestion> ptr;

            /**
              Constructor
              @param stat_str The string used for activating congestion stats for a
                              particular component.  If stat_str == network_link, e.g.
                              then stats are activated by:
                              <stats> congestion | congestion_network_link
              @param fname    The filename where congestion stats will be dumped
            */
            stats_congestion(const std::string& stat_str, const std::string& fname);

            virtual ~stats_congestion(){}

            /**
              Construct a stat_logger for congestion.
              @param devtype The name of the device, e..g network_link.  This will
                             be used in making the file name and the flag for activating.
                             The above would be activated by:
                            <stats> congestion | congestion_network_link
            */
            static ptr construct(const std::string& devtype);

            /**
              Increment the congestion counter for a given link/switch
              @param dev The link or switch device congestion occurs on

            */
            void add_congestion(const traincongestionpointPtr& dev,
                                double bytes_sent, double delay);

            void simulation_finished(timestamp end);

            virtual std::string to_string() const { return "stats_link_congestion"; }
    };


    class trainconfig :
        public ptr_type
    {
        public:
          typedef boost::intrusive_ptr<trainconfig> ptr;

          std::string to_string() const {
            return "train config";
          }

          static ptr
          construct();

          double payload_fraction;

          double link_bytes_per_cycle;
          double link_freq_hz;
          timestamp link_latency;
          bool ignore_link_congestion;
          bool enable_link_multipath;
          output_bw_arbitrator::ptr link_arb_template;

          bool ignore_switch_congestion;
          bool enable_switch_multipath;
          double switch_bytes_per_cycle;
          double switch_freq_hz;
          uint64_t switch_latency_cycles;

          bool vis;
    };

    /**
      @class trainpath
      Generic class for sending message between two points. This
      is usally just a trainlink, but might be a path with
      multiple possible links that might be travsersed
    */
    class trainpath :
        public eventmodule
    {

        public:
            typedef boost::intrusive_ptr<trainpath> ptr;

            virtual ~trainpath(){}

            /**
              Send the train from one point to another
              @param train The train to be setn
            */
            virtual void send(const packettrain::ptr& train) = 0;
    };

    /**
      @class traincongestionpoint
      Abstract class representing any point in the system - memory link, switch, network link -
      where congestion might occur between trains.
    */
    class traincongestionpoint :
        public trainpath
    {
        public:
            typedef boost::intrusive_ptr<traincongestionpoint> ptr;

        protected:
            static std::map<std::string, stats_congestion::ptr> loggers_;

            output_bw_arbitrator::ptr arbiter_;

            input_vc_allocator::ptr vc_alloc_;

            stats_congestion::ptr logger_;

            bool enable_multipath_;

            bool ignore_congestion_;

            static debug dbg_;

            double freq_hz_;

            double bw_bytes_;

            /** number between 0 and 1 that represents what
                fraction of the bytes being sent are actually
                payload bytes */
            double payload_fraction_;

            double num_bytes_per_send_;

            std::string devtype_;

            std::string id_;

            traincongestionpoint(double num_bytes_per_send,
                                 double freq_hz,
                                 double payload_fraction,
                                 const std::string& devtype,
                                 const std::string& id);

            /**
              Compute the bandwidth of the train that would be
              allocated in the absence of congestion
              @param train The train to arbitrate bandwidth for
            */
            void uncongested_input(const packettrain::ptr& train);

            /**
              After received its virtual channel, output buffer,
              and arbitrating for bandwidth, set the event for
              freeing the virtual channel
              @param train The train to free the vc for
            */
            void deallocate_vc_step(const packettrain::ptr& train);

            /**
              After arriving at the congestion point, allocate a virtual channel
              that the train can start to send along
              @param train The train to allocate the vc for
            */
            void allocate_vc_step(const packettrain::ptr& train);

            /**
              After allocating a virtual channel, allocate
              the output buffer for the train
              @param train The train to allocate an ouput buffer for
            */
            void allocate_output_buffer_step(const packettrain::ptr& train);

            /**
              After allocating a virtual channel and output buffer,
              allocate the output bw for the train
              @param train The train to allocate bw for
            */
            void allocate_bw_step(const packettrain::ptr& train);

            void new_input(const packettrain::ptr& train, double& t_start_ser, double& t_stop_ser);

            /** Flag indicating that the send should be indefinitely delayed */
            static const double send_delayed;

            /**
              @param t_send The simulation time a send will occur in seconds
              @return If the send time is an actual time or just a flag indicating delay
            */
            bool delayed_send(double t_send) const {
                return t_send == send_delayed;
            }

            virtual void ack(const timestamp& t, const packettrain::ptr& train) = 0;

            /**
              Once vc, output buffer, and output bw have been assigned, actually do the send.
              This event will be particular to the component (switch, link, etc)
              @param t_start The time the send can begin
              @param train The train being sent
            */
            virtual void do_send(const timestamp& t_start, const packettrain::ptr& train) = 0;

      public:
            /**
              Begin the process of sending.  Currently, this starts the allocate_vc step.
              @param train The train being sent
            */
            void send(const packettrain::ptr& train);

            virtual void finish();

            std::string name() const;

            virtual ~traincongestionpoint(){}

            virtual void set_eventmanager(const eventmanager::ptr &m);
    };

    }

}



#endif // CONGESTION_H
