#include <sstmac/hardware/packettrain/packettrain.h>
#include <sstmac/hardware/network/topology/routing/router.h>

namespace sstmac
{
  namespace hw
  {

    debug packettrain::dbg_;
    uint64_t packettrain::min_num_bytes_per_train_ = 0;
    uint64_t packettrain::max_num_trains_ = 0;
    uint64_t packettrain::id_count_ = 0;
    uint64_t packettrain::negligible_size_ = 0;
    const double packettrain::uninitialized_bw = -1;

    packettrain::packettrain(const sst_message::ptr &parent) :
      interconnect_message(parent),
      input_bw_(uninitialized_bw),
          byte_offset_(0), next_pending_(0), t_arrive_(0),
          vc_allocated_(false), congestion_delay_us_(0)
    {
      if (min_num_bytes_per_train_ == 0){
          sst_throw(illformederror, "packettrain statics not yet initialized");
      }

      msg_id_ = id_count_++;
      if (!dbg_.is_real())
        {
          dbg_ = debug("<debug> train");
        }

      // now determine if we need to split this up into multiple trains
      uint64_t num_trains =
          divide_and_round_up(num_bytes_, min_num_bytes_per_train_);

      num_trains = std::min(num_trains, max_num_trains_);
      uint64_t num_bytes_per_train = num_bytes_ / num_trains;
      uint64_t num_heavy_trains = num_bytes_ % num_trains;
      uint64_t num_light_trains = num_trains - num_heavy_trains;

      num_bytes_ = num_bytes_per_train;
      uint64_t byte_offset = num_bytes_;
      packettrain* prev = this;
      for (uint64_t i = 1; i < num_light_trains; ++i, byte_offset
          += num_bytes_per_train)
        {
          packettrain* next = new packettrain(orig(),
              msg_id_, num_bytes_per_train,
              byte_offset, input_bw_);
          prev->next_pending_ = next;
          prev = next;
        }

      num_bytes_per_train++;
      for (uint64_t i = 0; i < num_heavy_trains; ++i, byte_offset
          += num_bytes_per_train)
        {
          packettrain* next = new packettrain(orig(),
              msg_id_, num_bytes_per_train,
              byte_offset, input_bw_);
          prev->next_pending_ = next;
          prev = next;
        }

      packettrain* tail = prev;
    }

    bool
    packettrain::get_needs_ack() const {
        bool needs_ack = is_tail()
            && orig()->is_network_message()
            && orig()->get_needs_ack();
        return needs_ack;
    }

    void
    packettrain::throttle_bw(double bw)
    {
      set_initial_bw(bw);
      input_bw_ = std::min(bw, input_bw_);
      if (next_pending_)
        next_pending_->throttle_bw(bw);
    }

    void
    packettrain::add_next(const packettrain::ptr &train)
    {
      if (train.get() == this)
        {
          std::cerr << to_string() << std::endl;
          sst_throw(illformederror, "packettrain::add_next: adding self");
        }
      if (next_pending_)
        next_pending_->add_next(train);
      else
        next_pending_ = train;
    }

    void
    packettrain::set_input_bw(double bw)
    {
      //if (input_bw_ > 0 && bw > input_bw_){
      //    sst_throw(valueerror, "cannot reconfigure bw=%8.4e to a higher bandwidth %8.4e",
      //              input_bw_, bw);
      //}
      input_bw_ = bw;
    }

    void
    packettrain::init_statics(int min_bytes, int max_trains, int negligible_size)
    {
      min_num_bytes_per_train_ = min_bytes;
      max_num_trains_ = max_trains;
      negligible_size_ = negligible_size;
      negligible_size_ = std::min(negligible_size_, min_num_bytes_per_train_);
    }

    packettrain::packettrain(const sst_message::ptr& orig,
        int msg_id, uint64_t num_bytes,
        uint64_t offset, double bw) :
      interconnect_message(orig, num_bytes, orig->get_type()),
          input_bw_(bw),
          byte_offset_(offset),
          msg_id_(msg_id),
          next_pending_(0),
          vc_allocated_(false),
          congestion_delay_us_(0)
    {
    }

    void
    packettrain::finalize_congestion_delay()
    {
      orig()->add_delay(timestamp(1e-6 * congestion_delay_us_));
    }

    packettrain::ptr
    packettrain::construct(const sst_message::ptr &parent)
    {
      return new packettrain(parent);
    }

    packettrain::ptr
    packettrain::pop_next()
    {
      packettrain::ptr ret = next_pending_;
      next_pending_ = 0;
      return ret;
    }

    packettrain::~packettrain()
    {
    }

    std::string
    packettrain::to_string() const
    {
      return sstprintf("train %d, %lu->%lu:%lu at bw %8.4e next=%p", msg_id(),
          byte_offset_, byte_offset_ + num_bytes_, num_bytes_total(),
          input_bw_, next_pending_.get());
    }

  }
}
