/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/packettrain/train_nic.h>
#include <sstmac/common/errors.h>
#include <sstmac/software/process/operatingsystem.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("train | train_nic", networkinterface, train_nic,
        "implements a nic that models messages as a packet train");

    // Pointer types.

    using namespace sstmac::hw;

    multipath_nic_injector::multipath_nic_injector(
        const trainlink::ptr &rdma_link, const trainlink::ptr &udp_link) :
      rdma_link_(rdma_link), udp_link_(udp_link)
    {
    }

    multipath_nic_injector::ptr
    multipath_nic_injector::construct(const trainlink::ptr &rdma_link,
        const trainlink::ptr &udp_link)
    {
      return new multipath_nic_injector(rdma_link, udp_link);
    }

    void
    multipath_nic_injector::set_acker(const eventhandler::ptr &ack)
    {
      rdma_link_->set_acker(ack);
      udp_link_->set_acker(ack);
    }

    void
    multipath_nic_injector::send(const packettrain::ptr& train)
    {
      switch (train->orig()->get_type())
      {
      case sst_message::RDMA_GET_RSP:
      case sst_message::RDMA_PUT_RSP:
        rdma_link_->send(train);
        break;
      default:
        udp_link_->send(train);
        break;
      }
    }

    void
    multipath_nic_injector::finish()
    {
      rdma_link_->finish();
      udp_link_->finish();
    }

    void
    multipath_nic_injector::set_eventmanager(const eventmanager::ptr &m)
    {
      rdma_link_->set_eventmanager(m);
      udp_link_->set_eventmanager(m);
      eventmodule::set_eventmanager(m);
    }

    void
    multipath_nic_injector::handle(const sst_message::ptr &msg)
    {
      send(safe_cast(msg, packettrain));
    }

    train_nic::train_nic(const output_bw_arbitrator::ptr& rdma_arb,
        const timestamp &rdma_inj_lat, double rdma_inj_bw,
        const output_bw_arbitrator::ptr& udp_arb, const timestamp &udp_inj_lat,
        double udp_inj_bw, bool inj_multipath, bool inj_congestion,
        bool ej_multi, bool ej_cong) :
      networkinterface(), rdma_arb_template_(rdma_arb),
          rdma_inj_latency_(rdma_inj_lat), rdma_inj_bw_(rdma_inj_bw),
          udp_arb_template_(udp_arb), udp_inj_latency_(udp_inj_lat),
          udp_inj_bw_(udp_inj_bw), inj_enable_multipath_(inj_multipath),
          inj_ignore_congestion_(inj_congestion),
          ej_enable_multipath_(ej_multi), ej_ignore_congestion_(ej_cong)
    {
    }

    train_nic::ptr
    train_nic::construct(const sim_parameters::ptr& params)
    {
      /**
       sstkeyword {
       gui=7GB/s;
       docstring=The bandwidth available to RDMA get/put operations;
       }
       */
      double rdma_inj_bw = params->get_bandwidth_param(
          "network_train_rdma_injection_bw");
      /**
       sstkeyword {
       gui=0.5us;
       docstring=The injection latency of the RDMA pathway on the NIC. This is
       usually a higher latency operation than SMSG or UDP pathway;
       }
       */
      timestamp rdma_inj_lat = params->get_time_param(
          "network_train_rdma_injection_latency");
      /**
       sstkeyword {
       gui=8GB/s;
       docstring=The bandwidth available to simple datagram posts and short message
       mailboxes. This pathway is designed for low-latency on many, small
       messages;
       }
       */
      double udp_inj_bw = params->get_bandwidth_param(
          "network_train_udp_injection_bw");
      /**
       sstkeyword {
       gui=0.2us;
       docstring=The injection latency for datagram posts and short message
       mailboxes. This pathway is designed for low-latency on many, small
       messages;
       }
       */
      timestamp udp_inj_lat = params->get_time_param(
          "network_train_udp_injection_latency");
      /**
       sstkeyword {
       docstring=$train_arbitrator;
       }
       */
      output_bw_arbitrator::ptr rdma_arb =
          output_bw_arbitrator_factory::get_optional_param(
              "network_train_rdma_arbitrator", "infinite_buffer", params);
      /**
       sstkeyword {
       docstring=$train_arbitrator;
       }
       */
      output_bw_arbitrator::ptr udp_arb =
          output_bw_arbitrator_factory::get_optional_param(
              "network_train_udp_arbitrator", "infinite_buffer", params);
      /**
       sstkeyword {
       docstring=$enable_multipath;
       }
       */
      bool inj_multipath = params->get_optional_bool_param(
          "network_train_injection_enable_multipath", false);
      /**
       sstkeyword {
       docstring=$ignore_congestion;
       }
       */
      bool inj_congestion = params->get_optional_bool_param(
          "network_train_injection_ignore_congestion", false);
      /**
       sstkeyword {
       docstring=$enable_multipath;
       }
       */
      bool ej_multipath = params->get_optional_bool_param(
          "network_train_ejection_enable_multipath", inj_multipath);
      /**
       sstkeyword {
       docstring=$ignore_congestion;
       }
       */
      bool ej_congestion = params->get_optional_bool_param(
          "network_train_ejection_ignore_congestion", inj_congestion);

      return new train_nic(rdma_arb, rdma_inj_lat, rdma_inj_bw, udp_arb,
          udp_inj_lat, udp_inj_bw, inj_multipath, inj_congestion, ej_multipath,
          ej_congestion);
    }

    //
    // Goodbye.
    //
    train_nic::~train_nic() throw ()
    {
    }

    networkinterface::ptr
    train_nic::clone(const node::ptr &parent,
        const sstmac::hw::interconnect::ptr &interconn,
        const sim_parameters::ptr& params) const
    {
      train_nic::ptr ret = train_nic::construct(params);
      ret->init(parent, interconn);
      return ret;
    }

    void
    train_nic::init(const node::ptr &parent, const interconnect::ptr& interconn)
    {
      if (!parent->get_id())
      {
        return; //just a template nic
      }
      node_ = trainstation::construct(parent->get_msg_handler());
      mem_ = safe_cast(parent->get_mem(), train_memorymodel);
      networkinterface::init(parent, interconn);
    }

    void
    train_nic::connect(connection_type_t ty, const connectable::ptr &mod)
    {
        switch(ty)
        {
            case connectable::injection:
                set_inj_switch(safe_cast(mod,trainswitch));
                break;
            case connectable::ejection:
                set_ej_switch(safe_cast(mod,trainswitch));
                break;
            default:
                networkinterface::connect(ty, mod);
                break;
        }
    }

/**
    void
    train_nic::handle(const sst_message::ptr &msg)
    {

      packettrain::ptr train = boost::dynamic_pointer_cast<packettrain>(msg);
      if (!train)
      {
        sst_throw(illformederror, "train nic received non-train message");
      }
      SSTMAC_DEBUG << "train_nic::handle: " << train->to_string() << "\n";
      sst_message::ptr orig_msg = msg->get_parent_msg();
      switch (orig_msg->get_type())
      {
      case sst_message::RDMA_GET_REQ:
        send_rdma_rsp(train);
        break;
      case sst_message::RDMA_GET_RSP:
        recv_rdma_rsp(train);
        break;
      case sst_message::RDMA_PUT_REQ:
        mem_->rdma_put(train);
        break;
      case sst_message::RDMA_PUT_RSP:
        parent_->handle(train->parent());
        break;
      case sst_message::SEND_ACK:
        parent_->handle(train->parent());
        break;
      default:
        recv(train);
        break;
      }
    }
*/

    void
    train_nic::recv(const sst_message::ptr& netmsg)
    {
      packettrain::ptr train = safe_cast(netmsg, packettrain);

      /** If the total size of the message is really small, don't bother
       explicitly modeling the memory access */
      if (train->is_single_packet())
      {
        node_->handle(train);
      }
      else
      {
        mem_->do_access(train);
      }
    }

    void
    train_nic::set_inj_switch(const trainswitch::ptr &sw)
    {

      double payload_fraction = 1.0;
      if (sstmac_env::params->has_param("injection_train_payload_fraction"))
      {
        /**
         sstkeyword = {
         gui=1.0;
         docstring=$payload_fraction;
         }
         */
        payload_fraction = sstmac_env::params->get_double_param(
            "injection_train_payload_fraction");
      }
      else if (sstmac_env::params->has_param("network_train_payload_fraction"))
      {
        payload_fraction = sstmac_env::params->get_double_param(
            "network_train_payload_fraction");
      }

      //build a node label for this
      long node_id = parent_->get_id()->unique_id();
      std::string node_label = interconn_->label(node_id);

      trainlink::ptr rdma_in = trainlink::construct(sw, inj_enable_multipath_,
          inj_ignore_congestion_, rdma_inj_latency_, rdma_inj_bw_,
          payload_fraction, rdma_arb_template_->clone(), "rdma_injection",
          node_label);

      trainlink::ptr udp_in = trainlink::construct(sw, inj_enable_multipath_,
          inj_ignore_congestion_, udp_inj_latency_, udp_inj_bw_,
          payload_fraction, udp_arb_template_->clone(), "udp_injection",
          node_label);

      injector_ = multipath_nic_injector::construct(rdma_in, udp_in);
      injector_->set_acker(parent_->get_msg_handler());

      mem_->set_injector(injector_);

    }

    void
    train_nic::set_ej_switch(const trainswitch::ptr &sw)
    {

      double payload_fraction = 1.0;
      if (sstmac_env::params->has_param("injection_train_payload_fraction"))
      {
        /**
         sstkeyword = {
         gui=1.0;
         docstring=$payload_fraction;
         }
         */
        payload_fraction = sstmac_env::params->reread_double_param(
            "injection_train_payload_fraction");
      }
      else if (sstmac_env::params->has_param("network_train_payload_fraction"))
      {
        payload_fraction = sstmac_env::params->get_double_param(
            "network_train_payload_fraction");
      }

      //build a node label for this
      long node_id = parent_->get_id()->unique_id();
      std::string node_label = interconn_->label(node_id);

      trainlink::ptr rdma_out = trainlink::construct(ptr(this),
          ej_enable_multipath_, ej_ignore_congestion_, rdma_inj_latency_,
          rdma_inj_bw_, payload_fraction, rdma_arb_template_->clone(),
          "rdma_ejection", node_label);

      trainlink::ptr udp_out = trainlink::construct(ptr(this),
          ej_enable_multipath_, ej_ignore_congestion_, udp_inj_latency_,
          udp_inj_bw_, payload_fraction, udp_arb_template_->clone(),
          "udp_ejection", node_label);

      ejector_ = multipath_nic_injector::construct(rdma_out, udp_out);

      sw->add_ejector(parent_->get_id(), ejector_);
    }

    void
    train_nic::send(const sst_message::ptr &payload)
    {
      use_me();

      SSTMAC_DEBUG << "train_nic::send("
            << payload->toaddr()->to_string() << ", "
            << payload->to_string() << ")\n";

      if (payload == 0)
      {
        std::cerr << "I'm here throwing this payload error" << std::endl;
        throw nullerror("simplenic::send: null payload pointer.");
      }

      if (spy_)
      {
        spy_->add_one(payload->fromaddr()->unique_id(),
            payload->toaddr()->unique_id());
        spy_bytes_->add(payload->fromaddr()->unique_id(),
            payload->toaddr()->unique_id(), payload->get_byte_length());
        spy_bytes_alloc_png_->add(payload->fromaddr()->unique_id(),
            payload->toaddr()->unique_id(), payload->get_byte_length());
        spy_bytes_machine_png_->add(payload->fromaddr()->unique_id(),
            payload->toaddr()->unique_id(), payload->get_byte_length());
      }

      packettrain::ptr train = packettrain::construct(payload);
      if (train->negligible_size())
      { //don't go to memory - just send it off
        injector_->send(train);
      }
      else if (payload->get_type() == sst_message::RDMA_GET_REQ)
      {
        injector_->send(train);
      }
      else
      {
        mem_->do_access(train);
      }
    }

    void
    train_nic::set_eventmanager(const eventmanager::ptr &m)
    {
      networkinterface::set_eventmanager(m);
      node_->set_eventmanager(m);
      injector_->set_eventmanager(m);
      ejector_->set_eventmanager(m);
    }

    void
    train_nic::validate_msg_supported(const sst_message::ptr &msg) const
    {
      switch (msg->get_type())
      {
      case sst_message::MPI:
      case sst_message::HPX:
      case sst_message::SEND_ACK:
      case sst_message::ACTOR:
      case sst_message::NETWORK:
      case sst_message::RDMA_GET_REQ:
      case sst_message::RDMA_GET_RSP:
        return; //valid
      default:
        sst_throw(unimplementederror, "%s: network interface card received invalid message type %s",
            to_string().c_str(), sst_message::tostr(msg->get_type()));
      }
    }

    void
    train_nic::finish()
    {
      if (finished_)
        return;

      networkinterface::finish();
      if (injector_)
        injector_->finish();
      if (mem_)
        mem_->finish();

      injector_ = 0;
      mem_ = 0;
    }

    netlink_block::netlink_block(double freq_hz, double bytes_per_cycle,
        double payload_fraction, double max_single_bw, bool enable_multipath,
        bool ignore_congestion) :
          traincongestionpoint(bytes_per_cycle, freq_hz, payload_fraction,
              "netlink", ""), max_single_bw_(max_single_bw * payload_fraction)
    {
      arbiter_ = infinite_buffer_arbitrator::construct(sstmac_env::params);
      arbiter_->set_bw(bw_bytes_);
      arbiter_->set_eventmodule(ptr(this));
      if (enable_multipath)
      {
        vc_alloc_ = new uncongested_vc_allocator;
      }
      else
      {
        vc_alloc_ = new inorder_vc_allocator;
      }

      enable_multipath_ = enable_multipath;
      ignore_congestion_ = ignore_congestion;
    }

    void
    netlink_block::finish()
    {
      if (finished_)
        return;

      traincongestionpoint::finish();

      if (injector_)
      {
        injector_->finish();
        injector_ = 0;
      }
      ejector_map::iterator it, end = ejectors_.end();
      for (it = ejectors_.end(); it != end; ++it)
      {
        it->second->finish();
      }
      ejectors_.clear();
    }

    netlink_block::ptr
    netlink_block::clone() const
    {
      ptr ret = new netlink_block(freq_hz_, num_bytes_per_send_,
          payload_fraction_, max_single_bw_ / payload_fraction_,
          enable_multipath_, ignore_congestion_);
      return ret;
    }

    void
    netlink_block::handle(const sst_message::ptr &msg)
    {
      packettrain::ptr train = safe_cast(msg,packettrain);
      train->throttle_bw(max_single_bw_);
      send(train);
    }

    void
    netlink_block::add_injector(const eventmodule::ptr &mod)
    {
      injector_ = mod;
    }

    void
    netlink_block::add_ejector(const nodeaddress::ptr &addr,
        const eventmodule::ptr &mod)
    {
      if (!mod)
      {
        sst_throw(nullerror, "netlinkblock::add_ejector: null ejector");
      }
      ejectors_[addr] = mod;
    }

    void
    netlink_block::do_send(const timestamp& t_start,
        const packettrain::ptr &train)
    {
      timestamp t_send = t_start + timestamp(1e-9); //assume super low latency
      ejector_map::iterator it = ejectors_.find(train->toaddr());
      if (it == ejectors_.end())
      {
        schedule(t_send, injector_, train);
      }
      else
      {
        schedule(t_send, it->second, train); //outgoing
      }
    }

    netlink_block::ptr
    netlink_block::construct(double freq_hz, double bytes_per_cycle,
        double payload_fraction, double max_single_bw, bool enable_multipath,
        bool ignore_congestion)
    {
      return new netlink_block(freq_hz, bytes_per_cycle, payload_fraction,
          max_single_bw, enable_multipath, ignore_congestion);
    }

    void
    netlink_block::ack(const timestamp &t, const packettrain::ptr &train)
    {
      //do nothing
    }

  }
} // end of namespace sstmac.
