#ifndef sstmac_hardware_nic_TRAIN_NIC_H
#define sstmac_hardware_nic_TRAIN_NIC_H

#include <sstmac/hardware/nic/networkinterface.h>
#include <sstmac/software/libraries/compute/lib_compute_memmove.h>
#include <sstmac/hardware/network/interconnect/interconnect.h>
#include <sstmac/hardware/packettrain/trainstation.h>
#include <sstmac/hardware/memory/train_memorymodel.h>
#include <sstmac/hardware/packettrain/trainswitch.h>

namespace sstmac
{
  namespace hw
  {

    /**
     @class multipath_nic_injector
     Class for allowing NICs to have multiple injection paths for, e.g.
     RDMA separate from simple datagram sends
     */
    class multipath_nic_injector : public trainpath
    {
    private:
      trainlink::ptr rdma_link_;
      trainlink::ptr udp_link_;

    protected:
      multipath_nic_injector(const trainlink::ptr& rdma_link,
          const trainlink::ptr& udp_link);

    public:
      typedef boost::intrusive_ptr<multipath_nic_injector> ptr;

      static ptr
          construct(const trainlink::ptr& rdma_link,
              const trainlink::ptr& udp_link);

      /**
       Based on message type, pick the appropriate injection link and send the train
       @param train The train to be sent
       */
      void
      send(const packettrain::ptr& train);

      void
      set_acker(const eventhandler::ptr& ack);

      virtual void
      finish();

      virtual void
      set_eventmanager(const eventmanager::ptr &m);

      virtual void
      handle(const sst_message::ptr &msg);

    };

    /**
     @class train_nic
     Network interface compatible with sending packet trains
     */
    class train_nic :
        public networkinterface
    {

    protected:
      train_memorymodel::ptr mem_;
      multipath_nic_injector::ptr injector_;
      multipath_nic_injector::ptr ejector_;
      trainstation::ptr node_;

      timestamp rdma_inj_latency_;
      double rdma_inj_bw_;
      timestamp udp_inj_latency_;
      double udp_inj_bw_;
      bool inj_enable_multipath_;
      bool inj_ignore_congestion_;
      bool ej_enable_multipath_;
      bool ej_ignore_congestion_;

      output_bw_arbitrator::ptr rdma_arb_template_;
      output_bw_arbitrator::ptr udp_arb_template_;

      train_nic(const output_bw_arbitrator::ptr& rdma_arb,
          const timestamp& rdma_inj_lat, double rdma_inj_bw,
          const output_bw_arbitrator::ptr& udp_arb,
          const timestamp& udp_inj_lat, double udp_inj_bw,
          bool inj_multipath, bool inj_congestion,
          bool ej_multi, bool ej_cong);

      virtual void
      recv(const sst_message::ptr& netmsg);

    public:

      typedef boost::intrusive_ptr<train_nic> ptr;

      static train_nic::ptr
      construct(const sim_parameters::ptr& params);

      /// Goodbye.
      virtual
      ~train_nic() throw ();

      virtual networkinterface::ptr
      clone(const node::ptr &parent,
          const sstmac::hw::interconnect::ptr &interconn,
          const sim_parameters::ptr& params) const;

      virtual void
      send(const sst_message::ptr &payload);

      void connect(connection_type_t ty, const connectable::ptr &mod);

      std::string
      to_string() const
      {
        return "train_nic";
      }

      virtual void
      init(const node::ptr &parent, const interconnectPtr& interconn);

      /**
       Set up the injection/ejection links to the switch the NIC is connected to
       @param sw The switch that injects/ejects
       */
      void
      set_inj_switch(const trainswitch::ptr& sw);

      void
      set_ej_switch(const trainswitch::ptr& sw);

      void
      validate_msg_supported(const sst_message::ptr& msg) const;

      virtual void
      set_eventmanager(const eventmanager::ptr &m);

      virtual void
      finish();

    };

    /**
     @class netlink_block
     A class representing a common congestion point for multiple nodes
     connected to a switch.  If used, this allows nodes on the same switch
     to exchange messages without actually using the switch.
     */
    class netlink_block : public traincongestionpoint
    {
    protected:
      double max_single_bw_;

      typedef boost::unordered_map<nodeaddress::ptr, eventmodule::ptr>
          ejector_map;

      ejector_map ejectors_;

      eventmodule::ptr injector_;

      netlink_block(double freq_hz, double bytes_per_cycle,
          double payload_fraction, double max_single_bw, bool enable_multipath,
          bool ignore_congestion);

      void
      do_send(const timestamp& t_start, const packettrain::ptr &send);

    public:
      typedef boost::intrusive_ptr<netlink_block> ptr;

      static ptr
      construct(double freq_hz, double bytes_per_cycle,
          double payload_fraction, double max_single_bw, bool enable_multipath,
          bool ignore_congestion);

      void
      handle(const sst_message::ptr &msg);

      void
      ack(const timestamp &t, const packettrain::ptr &train);

      /**
       Set up the ejection link from the block to the nodes
       @param mod The node or module at a given compute node address
       */
      void
      add_ejector(const nodeaddress::ptr& addr, const eventmodule::ptr& mod);

      /**
       Set up the injection link from the block to the switch
       @param mod The switch or module the netlink block injects messages into
       */
      void
      add_injector(const eventmodule::ptr& mod);

      virtual void
      finish();

      virtual std::string
      to_string() const
      {
        return sstprintf("netlink %p", this);
      }

      virtual netlink_block::ptr
      clone() const;

    };

  }
} // end of namespace sstmac

#endif // TRAIN_NIC_H
