#ifndef TRAINLINK_H
#define TRAINLINK_H


#include <sstmac/hardware/packettrain/congestion.h>
#include <sstmac/common/stats/stat_logger.h>

namespace sstmac {
    namespace hw {

/**
  @class trainlink
  Implements a single physical link between two endpoints.
  Trains are sent from start endpoint to finish endpoint.
*/
class trainlink :
    public traincongestionpoint
{

protected:
  static int num_links_;

  int link_num_;

  eventmodule::ptr dest_;

  timestamp link_latency_;

  eventhandler::ptr acker_;

  /**
    Schedule ack for when the train finishes sending on the link
    @param time The time the train finishes sending and the ack can occur
    @param train The train being sent that needs to send an ack notification
  */
  void ack(const timestamp& time, const packettrain::ptr& train);

  /**
    Implementation of parent virtual method.  Actually performs
    the send operation transferring the train from one endpoint
    to the next
    @param t_start The time the train can begin sending
    @param train The train being sent
  */
  virtual void
  do_send(const timestamp& t_start, const packettrain::ptr& train);

  trainlink(const eventmodule::ptr& dest, const timestamp& link_latency,
      double num_bytes_per_send, double freq_hz, double payload_fraction,
      bool enable_multipath, bool ignore_congestion,
      const output_bw_arbitrator::ptr& arb,
      const std::string& link_type,
      const std::string& link_name);

public:

  typedef boost::intrusive_ptr<trainlink> ptr;

  virtual ~trainlink(){}

  /**
        Create a new link connection on a given outport for this switch
        @param dest The endpoint (switch/node/etc) the link sends to
        @param link_latency The hop latency for the link
        @param bytes_per_cycle The number of bytes the link can send each clock cycle
        @param link_freq_hz The number of cycles per second (in GHz).  Bandwidth is
                              bytes_per_cycle * link_freq
        @param payload_fraction The fraction (0.0-1.0) of bytes per packet
                                that comprise the payload as opposed to the
                                routing metadata overhead
        @param enable_multipath Whether the link should allow simultaneous packet trains
                                from the same message.  See #inorder_vc_allocator
                                vs #uncongested_vc_allocator.
        @param ignore_congestion Whether the link should monitor congestion
                                 and monitor bandwidth accordingly
        @param arb               The arbitrator that will handle allocating bw during congestion
        @param link_type         A string identifying the type of link.  Common values are
                                 network, memory, injection.
        @param link_name         A unique name for the link, e.g. for network link in 3D torus
                                 name might be "[1 0 3]->[1 0 4]" to identify +Z link for
                                 switch at x=1,y=0,z=3.
  */
  static ptr
  construct(const eventmodule::ptr& dest, const timestamp& link_latency,
      double bytes_per_cycle, double freq_hz, double payload_fraction,
      bool enable_multipath, bool ignore_congestion,
      const output_bw_arbitrator::ptr& arb,
      const std::string& link_type,
      const std::string& link_name = "");


  /**
        Create a new link connection on a given outport for this switch
        @param dest The endpoint (switch/node/etc) the link sends to
        @param enable_multipath Whether the link should allow simultaneous packet trains
                                from the same message.  See #inorder_vc_allocator
                                vs #uncongested_vc_allocator.
        @param ignore_congestion Whether the link should monitor congestion
                                 and monitor bandwidth accordingly
        @param lat              The hop latency for the link
        @param bw               The link bandwidth in bytes/sec
        @param payload_fraction The fraction (0.0-1.0) of bytes per packet
                                that comprise the payload as opposed to the
                                routing metadata overhead
        @param arb               The arbitrator that will handle allocating bw during congestion
        @param link_type         A string identifying the type of link.  Common values are
                                 network, memory, injection.
        @param link_name         A unique name for the link, e.g. for network link in 3D torus
                                 name might be "[1 0 3]->[1 0 4]" to identify +Z link for
                                 switch at x=1,y=0,z=3.
  */
  static ptr
  construct(const eventmodule::ptr& dest, bool enable_multipath, bool ignore_congestion,
            const timestamp& lat, double bw,
            double payload_fraction,
            const output_bw_arbitrator::ptr& arb,
            const std::string& link_type,
            const std::string& link_name = "");

  /**
    Cast message and pass to #send
    @param msg Incoming message (should cast to packettrain)
  */
  virtual void
  handle(const sst_message::ptr &msg);

  /**
    Configure the event handler for acking when a train finishes sending
    @param ack The acker. For network links, this is often a NIC or node
               handler to indicate that a message has finished injecting
               into the network.
  */
  void
  set_acker(const eventhandler::ptr& ack){
      acker_ = ack;
  }

  virtual void
  set_eventmanager(const eventmanager::ptr &m);

  timestamp
  link_latency() const
  {
    return link_latency_;
  }

  /**
    Compute the next time the link is ''free.'' The exact physical meaning
    will vary depending on how the link arbitrates bandwidth.
    However, higher valules indicate more congestion.
    @return The next uncongested time returned as a delay (relative from now) in seconds
  */
  virtual double
  next_free() const {
    return arbiter_->next_free();
  }

  virtual double
  get_util_bw() const {
    return arbiter_->get_util_bw();
  }

  int
  get_queue_length() const;

  virtual void finish();

  /**
    @return The unique link identifier
  */
  int
  num()
  {
    return link_num_;
  }

  std::string
  to_string() const;
};



    }
}

#endif // TRAINLINK_H
