#include <sstmac/hardware/packettrain/trainstation.h>
#include <sstmac/common/eventcallback.h>
#include <sstmac/hardware/packettrain/trainswitch.h>
#include <sstmac/hardware/packettrain/trainlink.h>
#include <sstmac/software/api/mpi/mpimessage.h>

namespace sstmac
{
  namespace hw
  {

    debug trainstation::dbg_;
    int trainstation::num_stations_ = 0;


    trainstation::trainstation(const eventhandler::ptr& handler) :
      num_(num_stations_++), handler_(handler)
    {
      if (!dbg_.is_real())
      {
        dbg_ = debug("<debug> train");
      }
    }

    trainstation::ptr
    trainstation::construct(const eventhandler::ptr& handler)
    {
      return new trainstation(handler);
    }

    std::string
    trainstation::to_string() const
    {
      return "train station";
    }

    void
    trainstation::handle(const sst_message::ptr &msg)
    {
      packettrain::ptr train = safe_cast(msg, packettrain);
      if (train->negligible_size()){
            finalize_train(train);
            return;
      }

      uint64_t& num_bytes_left = trains_[train->msg_id()];
      if (num_bytes_left == 0)
      {
        num_bytes_left = train->orig()->get_byte_length();
        schedule_unload(train);
      }
      else
      {
        schedule_unload(train);
      }
    }

    void
    trainstation::finish()
    {
        if (finished_)
            return;

        eventmodule::finish();
        eventmodule::ptr mod = boost::dynamic_pointer_cast<eventmodule>(handler_);
        if (mod)
            mod->finish();
        handler_ = 0;
    }

    void
    trainstation::schedule_unload(const packettrain::ptr& train)
    {
        if (!train->negligible_size()){
            SSTMAC_DEBUG << sstprintf("On station %d, scheduling unload %s for %8.4e sec from now\n",
                                      num_, train->to_string().c_str(), train->ser_delay());
        }

        schedule_delay(
            timestamp(train->ser_delay()),
            new_event(trainstation::ptr(this), &trainstation::unload_train, train));
    }

    void
    trainstation::finalize_train(const packettrain::ptr& train)
    {
        train->finalize_congestion_delay();
        if (train->orig()->has_callback())
        {
            //if (!train->negligible_size()){
                SSTMAC_DEBUG << sstprintf("On station %d, unloading %s to callback\n", num_,
                                          train->to_string().c_str());
            //}

          train->orig()->pop_callback()->handle(train->orig());
        }
        else
        {
            //if (!train->negligible_size()){
                SSTMAC_DEBUG << sstprintf("On station %d, unloading %s to handler: %s\n",
                                          num_, train->to_string().c_str(),
                                          train->orig()->to_string().c_str());
            //}
          handler_->handle(train->orig());
        }
    }

    void
    trainstation::unload_train(const packettrain::ptr& train)
    {
      std::map<uint64_t, uint64_t>::iterator it = trains_.find(train->msg_id());
      uint64_t& num_bytes_left = it->second;
      num_bytes_left -= train->num_bytes();
      //if (!train->negligible_size()){
          SSTMAC_DEBUG << sstprintf("On station %d, unload %s -> %lu bytes left\n",
                                    num_, train->to_string().c_str(), num_bytes_left);
      //}

      if (num_bytes_left == 0)
      {
          finalize_train(train);
          trains_.erase(it);
      }
      else if (train->has_next()){
         packettrain::ptr next = train->get_next();
         schedule_unload(next);
      }
    }

  }
}
