#ifndef TRAINSWITCH_H
#define TRAINSWITCH_H

#include <sstmac/common/eventmodule.h>
#include <sstmac/common/sim_parameters.h>
#include <sstmac/hardware/network/topology/routing/router.h>
#include <sstmac/hardware/network/switches/networkswitch.h>
#include <sstmac/hardware/packettrain/trainstation.h>
#include <sstmac/hardware/packettrain/trainlink.h>
#include <sstmac/hardware/packettrain/congestion.h>
#include <sstmac/common/vis/vis.h>

namespace sstmac
{
  namespace hw
  {
    class train_nic;
    typedef boost::intrusive_ptr<train_nic> train_nicPtr;

    /**
     @class trainswitch
     A switch in the network that arbitrates/routes packettrains
     to the next link in the network
     */
    class trainswitch : public networkswitch
    {

    protected:
      trainswitch(const trainconfig::ptr& config);

      trainconfig::ptr config_;
      timestamp latency_;
      double tile_bw_;
      bool vis_;

      boost::unordered_map<nodeaddress::ptr, eventmodule::ptr> ejectors_;

      boost::unordered_map<long, trainlink::ptr> outports_;

      //      dim            dir    vis obj
      boost::unordered_map<long, vis::vis_obj::ptr> linkvis_;

      static debug dbg_;

      /**
       Implementation of virtual method from parent class.  This performs
       the routing operation and sends the packet train off to the next
       link.
       @param t_start The (absolute) simulation time in seconds to begin the send
       @param train The train to send
       */
      void
      do_send(const timestamp& t_start, const packettrain::ptr& train);

    public:
      typedef boost::intrusive_ptr<trainswitch> ptr;

      vis::vis_display::ptr vis_parent_;

      static ptr
      construct(const sim_parameters::ptr& params);

      /**
       Create an exact copy of this switch, but with different router and network node address
       @param The router for the cloned switch
       @param The network node address for the cloned switch
       @return The new trainswitch
       */
      virtual networkswitch::ptr
      clone(router::ptr route, nodeaddress::ptr addr) const
      {
        ptr ret = new trainswitch(config_);
        clone_into(ret, addr, route);
        return ret;
      }

      int
      get_queue_length(int port) const;

      virtual void
      connect(int dim, int port, connection_type_t ty,
          const connectable::ptr& mod, double weight, int red);

      virtual void
      connect(const nodeaddress::ptr& addr, int port, connection_type_t ty,
          const connectable::ptr& mod);

      /**
       Cast message and pass to #send
       @param msg Incoming message (should cast to packettrain)
       */
      void
      handle(const sst_message::ptr &msg);

      virtual void
      set_link_vis(vis::vis_obj::ptr vo, int dim, int dir)
      {
        long port = router_->convert_to_outport(dim, dir);
        linkvis_[port] = vo;
      }

      /**
       Set the link to use when ejecting packets at their endpoint.  A trainswitch
       can have any number of ejectors, corresponding to the number of nodes
       per switch.
       @param addr The compute node address of the endpoint to eject to
       @param link The link to the compute node for ejection
       */
      void
      add_ejector(const nodeaddress::ptr& addr, const trainpath::ptr& link);

      virtual void
      set_eventmanager(const eventmanager::ptr &m);

      /**
       Send a network ack when the train leaves.  Currently, switch
       does nothing.
       @param t The time the ack can be sent
       @param train The train being sent that can be acked
       */
      void
      ack(const timestamp &t, const packettrain::ptr &train);

      virtual void
      finish();

      virtual std::string
      to_string() const;

      virtual
      ~trainswitch();

      virtual std::string
      get_interprocess_link_name() const
      {
        return "network_train_link_latency";
      }

    };

  }
}

#endif // TRAINSWITCH_H
