/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/common/ptr_type.h>

#include <sstmac/common/errors.h>
#include <sstmac/hardware/processor/instructionprocessor.h>
#include <sstmac/common/messages/eventdata.h>
#include <sstmac/hardware/processor/instructionprocessor-config.h>

#include <iostream>

namespace sstmac
{
  namespace hw
  {

    //
    // Constructor
    //
    instructionprocessor::instructionprocessor(
        const instructionprocessor_config::ptr& cfg,
        const eventhandler::ptr &completion, computescheduler::ptr sched) :
        processor(cfg, completion, sched),
        tmemop_rand_ns_(0), tflop_rand_ns_(0),
         config_(cfg)
    {
      flop_label_ = config_->flop_label();
      memop_label_ = config_->memop_label();
      memop_local_label_ = config_->memop_local_label();
      memop_rand_label_ = config_->memop_rand_label();
      tflop_ = config_->tflop();
      tmem_ = config_->tmemop();
      tmemloc_ = config_->tmemop_local();
      tmemrnd_ = config_->tmemop_rand();

      if(stat_pwr_->is_active())
        stat_pwr_->record(config_->get_pwr_static());

      // debug report
      if (dbg_.is_active())
        {
          dbg_ << "instructionprocessor: initializing\n";
          dbg_ << "  average flop service time:   " << (tflop_ * 1e9) << " ns\n";
          dbg_ << "  average memory service time: " << (tmem_ * 1e9) << " ns\n";
        }
    }

    //
    // Construct a new processor with the given average service times
    //
    instructionprocessor::ptr
    instructionprocessor::construct(const instructionprocessor_config::ptr& cfg,
        const eventhandler::ptr &completion, computescheduler::ptr sched)
    {
      instructionprocessor::ptr retval(
          new instructionprocessor(cfg, completion, sched));

      return retval;
    }

    processor::ptr
    instructionprocessor::clone() const
    {

      return construct(config_, completion_, csched_);

    }

    void
    instructionprocessor::set_flop_distribution(double stdev)
    {
        tflop_rand_ns_ = RNG::NormalDistribution::construct(tflop_ * 1e9, stdev * 1e9);
    }

    void
    instructionprocessor::set_memop_distribution(double stdev)
    {
        tmemop_rand_ns_ = RNG::NormalDistribution::construct(tflop_ * 1e9, stdev * 1e9);
    }

    timestamp
    instructionprocessor::compute_time(const sst_message::ptr &data, long threadid)
    {
        // compute execution time in seconds
        timestamp time;
        if (tflop_ > 0)
          {
            if (flop_label_.size() == 0)
              throw valueerror("missing flop data label");
            uint64_t nop = data->get_optional_event_value(flop_label_);
            if (nop){
                double tsec = tflop_rand_ns_ ? tflop_rand_ns_->value() * 1e-9 * nop : nop * tflop_;
                timestamp thistime(tsec);
                if (tsec > 0 && stat_energy_->is_active()){
                  stat_energy_->record(csched_->now(), nop * config_->get_pwr_flop(), thistime);
                  stat_proc_->record(csched_->now(), nop);
                }
                time += thistime;
            }
          }
        if (tmem_ > 0)
          {
            if (memop_label_.size() == 0)
              throw valueerror("missing memory op data label");
            uint64_t nop = data->get_optional_event_value(memop_label_);
            if (nop){
                double tsec = nop * tmem_;
                timestamp thistime(tsec);
                if (tsec > 0 && stat_energy_->is_active())
                  stat_energy_->record(csched_->now(), nop * config_->get_pwr_memop(), thistime);
                time += thistime;
            }
          }
        if (tmemloc_ > 0)
          {
            if (memop_local_label_.size() == 0)
              throw valueerror("missing local memory op data label");
            uint64_t nop = data->get_optional_event_value(memop_local_label_);
            if (nop){
                double tsec = tmemop_rand_ns_ ? tmemop_rand_ns_->value() * 1e-9 * nop : nop * tmemloc_;
                timestamp thistime(tsec);
                if (tsec > 0 && stat_energy_->is_active())
                  stat_energy_->record(csched_->now(), nop * config_->get_pwr_memop_local(), thistime);
                time += thistime;
            }
          }
        if (tmemrnd_ > 0)
          {
            if (memop_rand_label_.size() == 0)
              throw valueerror("missing random memory op data label");
            uint64_t nop = data->get_optional_event_value(memop_rand_label_);
            if (nop){
                double tsec = nop * tmemrnd_;
                timestamp thistime(tsec);
                if (tsec > 0 && stat_energy_->is_active())
                  stat_energy_->record(csched_->now(), nop * config_->get_pwr_memop_rand(), thistime);
                time += thistime;
            }
          }
        return time;
    }

    void
    instructionprocessor::compute(const sst_message::ptr& inst, long thread_id)
    {
      SSTMAC_DEBUG << "instructionprocessor::compute("
                   << (long) inst.get()
                   << ") on thread " << thread_id << "\n";
      timestamp time = compute_time(inst, thread_id);
      SSTMAC_DEBUG << "instructionprocessor: execution time on thread " << thread_id
                     << ": " << time.sec() * 1e6 << "us\n";
      request_compute(time, inst, thread_id);

  }

}
} // end of namespace sstmac

