/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/main/sstmac-parallel.h>
#include <sstmac/main/sstmac.h>

#include <sstmac/backends/native/sim_partition.h>
#include <sstmac/backends/native/mpiparallel/parallelprocess.h>
#include <mpi.h>

using namespace sstmac;

sim_parameters::ptr
parsim_manager::check_multisim(const opts &oo, sim_parameters::ptr params)
{
  if (params->is_multi_sim())
  {
    sst_throw(ssterror, "parsim_manager::check_multisim - you can't do parameter sweeps with PDES yet");
  }

  return params;
}

void
parsim_manager::run(const opts &oo, sim_parameters::ptr params)
{

  sstmac_env::run_mode_ = sstmac_env::DIST_PARALLEL;

  debug dbg_("<debug> sstmac | sim | simulation");

  int rank, size;
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &size);

  if (dbg_.is_active())
  {
    dbg_ << "process " << rank << " constructing network graph \n";
  }
  std::stringstream ss;
  ss << "partitioner -f " << oo.configfile
      << " -p launch_app1=hello -p myrank=" << rank;
  system(ss.str().c_str());

  if (dbg_.is_active())
  {
    dbg_ << "process " << rank << " partitioning network, among " << size
        << " processes \n";
  }

  ss.str("");
  ss << "partition" << rank << ".metis";
  std::string graphfile = ss.str();
  ss.str("");
  ss << "gpmetis -contig \"" << graphfile << "\" " << size;
  system(ss.str().c_str());

  ss.str("");
  ss << "partition" << rank << ".metis.part." << size;
  std::string partfile = ss.str();
  params->add_param_override("partitionfile", partfile);

  if (dbg_.is_active())
  {
    dbg_ << "process " << rank << " constructing partition \n";
  }

  // Read in partitions.
  const native::partition::ptr partition = native::partition::construct(
      params->get_param("partitionfile"), size);

  native::parallelprocess::ptr pp = native::parallelprocess::construct(params,
      partition, rank, size);

  timestamp done = pp->run(oo.stoptime);

  if (oo.mode != "notime")
  {
    dbg_
        << sstprintf("Estimated total runtime of %20.8f seconds\n", done.sec());
  }

  //cleanup
  std::string comm = "rm " + partfile;
  system(comm.c_str());
  comm = "rm " + graphfile;
  system(comm.c_str());

}

int
main(int argc, char **argv)
{

  MPI_Init(&argc, &argv);

  int rc;

  try
  {
    parsim_manager::ptr sim = parsim_manager::construct();
    rc = try_sstmacmain(argc, argv, sim);
  }
  catch (const std::exception &e)
  {
    std::cout << argv[0] << ": caught exception while setting up simulation:\n"
        << e.what() << "\n";
    return 1;
  }
  catch (...)
  {
    std::cout << argv[0]
        << ": caught unknown exception while setting up simulation\n";
    return 1;
  }

  MPI_Finalize();

  return rc;
}

