/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_AMI_ENVIRONMENT_H_INCLUDED
#define SSTMAC_SOFTWARE_AMI_ENVIRONMENT_H_INCLUDED

#include <sstmac/software/process/taskid.h>
#include <sstmac/common/nodeaddress.h>

#include <sstmac/common/ptr_type.h>

#include <sstmac/common/sstmac-config.h>

#include <sstmac/common/messages/sst_serializer.h>
#include <boost/unordered_map.hpp>

namespace sstmac
{
  namespace sw
  {

    class environment : public ptr_type
    {
    public:
      typedef boost::intrusive_ptr<environment> ptr;
      typedef boost::intrusive_ptr<const environment> const_ptr;

      // ------- constructor and boost stuff ----------
    public:
      static environment::ptr
      construct();

      virtual
      ~environment();

    protected:
      environment();

    protected:
      typedef boost::unordered_map<long, nodeaddress::ptr> map_type;

      map_type mapping_;

      std::list<nodeaddress::ptr> extras_;

      std::vector<std::string> hostnames_;

      std::map<std::string, std::string> shell_env_;

      std::map<std::string, nodeaddress::ptr> inet_addr_map_;

    public:

      virtual std::string
      to_string() const
      {
        return "environment";
      }

      void add_hostname(long tid, const std::string& name, const nodeaddress::ptr& addr);

      void
      setenv(const std::string& name, const std::string& value)
      {
          shell_env_[name] = value;
      }

      std::string
      getenv(const std::string& name) const;

      std::string
      get_hostname(const taskid& tid) const;

      nodeaddress::ptr
      get_node_address(const std::string& hostname) const;

      nodeaddress::ptr
      get_node_address(long id) const;

      const nodeaddress::ptr
      get_node_address(taskid &s) const
      {
        nodeaddress::ptr ret = mapping_.find(s.id)->second;
        return ret;
      }

      long
      get_app_count() const
      {
        return mapping_.size();
      }

      int
      get_node_count() const {
        std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp> nodes;

        map_type::const_iterator it, end = mapping_.end();
        for(it = mapping_.begin(); it != end; it++){
            nodes.insert(it->second);
        }
        return nodes.size();
      }

      void
      add_item(taskid& s, const nodeaddress::ptr &n)
      {
        mapping_[s.id] = n;
      }

      void
      remove_item(taskid& s, const nodeaddress::ptr &n)
      {
        mapping_.erase(s.id);
      }

      void
      add_extra(nodeaddress::ptr a)
      {
        extras_.push_back(a);
      }

      std::list<nodeaddress::ptr>
      get_extras() const
      {
        return extras_;
      }

      void
      serialize(sst_serializer* ser);

      static environment::ptr
      deserialize(sst_serializer* ser);
    };

  }
} // end of namespace sstmac

#endif
