/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/api/eiger/eigermodel.h>

#include <sstmac/common/errors.h>
#include <sstmac/common/fileio.h>

#include <math.h>

namespace sstmac
{

  debug eigermodel::eigerdbg_;

	//
	// Constructor
	//
	eigermodel::eigermodel( const std::string &model_fname) : model_fname_(model_fname)
	{
    if (!eigerdbg_.is_real()){
      eigerdbg_ = sstmac::debug("<debug> eiger", sstmac::logger::red);
      sstmac::logger::register_logger(&eigerdbg_);
    }

		SSTFileIO::open_file(in_, model_fname);
		if(!in_.is_open()){
				std::stringstream ss;
				ss << "eigermodel: failed to open model file " << model_fname << std::endl;
				throw valueerror(ss.str());
		}
    if(eigerdbg_.is_active()){ 
      eigerdbg_ << "Parsing Eiger model " << model_fname << "\n"; 
    }
		std::string line;
		getline(in_,line); // kill one line

		principal_components_ = boost::numeric::ublas::matrix<double>();
		weights_ = boost::numeric::ublas::vector<double>();
		in_ >> principal_components_ >> weights_;
		getline(in_,line); // kill one line

    if(eigerdbg_.is_active()){ 
      eigerdbg_ << "Eiger model (" << model_fname << "): pca size: " 
                << principal_components_.size1() << "," 
                << principal_components_.size2() 
                << ", # model functions: " << weights_.size() << "\n"; 
    }

		predictors_.resize(weights_.size());
		for(unsigned i=0; i < predictors_.size(); i++){
			std::getline(in_, line);
			std::istringstream iss(line);
			int op;
			bool fail = false;
			while(iss.good()){
				basis new_basis;
				iss >> op;
				switch(op){
					case 0:
						new_basis.f = identity_func;
						break;
					case 1:
						new_basis.f = power_func;
						if(!(iss >> new_basis.op1 >> new_basis.op2)){
							fail = true;
						}
						break;
					case 2:
						new_basis.f = crossterm_func;
						if(!(iss >> new_basis.op1 >> new_basis.op2)){
							fail = true;
						}
						break;
					case 3:
						new_basis.f = sqrt_func;
						if(!(iss >> new_basis.op1)){
							fail = true;
						}
						break;
					case 4:
						new_basis.f = log_func;
						if(!(iss >> new_basis.op1)){
							fail = true;
						}
						break;
					default:
						fail = true;
						break;
				}
				if(fail){
					std::stringstream ss;
					ss << "eigermodel: malformed function in model \"" << model_fname_ << "\"\n";
					throw valueerror(ss.str());
				}
				predictors_[i].push_back(new_basis);
			}
		}

		for(unsigned i=0; i < principal_components_.size1(); i++){
			std::getline(in_, line);
			input_metrics_.push_back(line);
		}
		
		in_.close();

	}

	//
	// Poll this model with the given set of input parameters
	//
	double
	eigermodel::poll(std::map<std::string,double> params){
		
		boost::numeric::ublas::vector<double> v = 
			boost::numeric::ublas::vector<double>(principal_components_.size1());

		std::vector<std::string> missing;

    if(eigerdbg_.is_active()){
      eigerdbg_ << "Polling Eiger model " << model_fname_ 
                << " with input vector: [";
    }
		unsigned i = 0;
		for(std::vector<std::string>::iterator it = input_metrics_.begin();
				it != input_metrics_.end(); it++){
			std::map<std::string,double>::iterator param = params.find(*it);
			if(param != params.end()){
				v(i) = param->second;
        if(eigerdbg_.is_active()){
          eigerdbg_ << *it << ": " << param->second << ", ";
        }
			}
			else{
				missing.push_back(*it);
			}
			i++;
		}
    if(eigerdbg_.is_active()){ eigerdbg_ << "]\n"; }

		if(missing.size() > 0){
			std::stringstream ss;
			ss << "eigermodel: missing input parameters to model \"" << model_fname_ << "\":\n";
			for(std::vector<std::string>::iterator it = missing.begin();
					it != missing.end(); it++){
				ss << "-->\t" << *it << "\n";
			}
			throw valueerror(ss.str());
		}

		boost::numeric::ublas::vector<double> inputs;
		inputs = prod(v,principal_components_);
		boost::numeric::ublas::vector<double> function_vals = 
			boost::numeric::ublas::vector<double>(predictors_.size());
		i = 0;
		for(std::vector<std::vector<basis> >::iterator it = predictors_.begin();
				it != predictors_.end(); it++, i++){
			double res = 1.0;
			for(std::vector<basis>::iterator bi = it->begin();
					bi != it->end(); bi++){
				res *= (*bi->f)(inputs,bi->op1,bi->op2);
			}
			function_vals(i) = res;
		}
		double final_res = fabs(inner_prod(weights_,function_vals));
    if(eigerdbg_.is_active()){
      eigerdbg_ << "Eiger model " << model_fname_ << " polled for " << final_res << "\n";
    }
		return final_res;

	}


	//
	// Construct a new eiger model from the given file
	//
	eigermodel::ptr
	eigermodel::construct(const std::string &model_fname){
		eigermodel::ptr ret = ptr(new eigermodel(model_fname));

		return ret;
	}

	double eigermodel::identity_func(boost::numeric::ublas::vector<double> v, double op1, double op2)
	{
		return 1.0;
	}

	double eigermodel::power_func(boost::numeric::ublas::vector<double> v, double op1, double op2)
	{
		// guarantee that integer ops stored in doubles are correctly converted to the nearest integer
		int int_op1 = floor(op1 + 0.5);
		//int int_op2 = (int) (op2 + 0.5);
		return v(int_op1) != 0.0 ? pow(fabs(v(int_op1)),op2) : 1.0;
	}

	double eigermodel::crossterm_func(boost::numeric::ublas::vector<double> v, double op1, double op2)
	{
		// guarantee that integer ops stored in doubles are correctly converted to the nearest integer
		int int_op1 = floor(op1 + 0.5);
		int int_op2 = floor(op2 + 0.5);
		return v(int_op1) * v(int_op2);
	}

	double eigermodel::sqrt_func(boost::numeric::ublas::vector<double> v, double op1, double op2)
	{
		int int_op1 = floor(op1 + 0.5);
		return sqrt(fabs(v(int_op1)));
	}

	double eigermodel::log_func(boost::numeric::ublas::vector<double> v, double op1, double op2)
	{
		int int_op1 = floor(op1 + 0.5);
		if(v(int_op1) == 0.0) return 1.0;
		return log(fabs(v(int_op1)))/log(2.0);
	}

	double eigermodel::hinge1_func(boost::numeric::ublas::vector<double> v, double op1, double op2)
	{
		// guarantee that integer ops stored in doubles are correctly converted to the nearest integer
		double int_op1 = floor(op1 + 0.5);
		double val = v(int_op1);
		return val > op2 ? val - op2 : 0.0;
	}
	
	double eigermodel::hinge2_func(boost::numeric::ublas::vector<double> v, double op1, double op2)
	{
		// guarantee that integer ops stored in doubles are correctly converted to the nearest integer
		double int_op1 = floor(op1 + 0.5);
		double val = v(int_op1);
		return val < op2 ? op2 - val : 0.0;
	}

} // end of namespace sstmac

