/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SUBPACKAGES_EIGER_EIGERMODEL_H_INCLUDED
#define SUBPACKAGES_EIGER_EIGERMODEL_H_INCLUDED

#include <sstmac/common/nocopy.h>

#include <sstmac/common/ptr_type.h>
#include <sstmac/common/sim_parameters.h>
#include <sstmac/common/units.h>
#include <sstmac/common/debug.h>

#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/io.hpp>

#include <iostream>
#include <fstream>
#include <map>

namespace sstmac
{

  /**
   * Container for Eiger performance models. This object parses a model
   * file, as well as returning the result of polling
   */
  class eigermodel : virtual public nocopy, public ptr_type
  {
  protected:
    /// Protected constructor.
    eigermodel(const std::string &model_fname);

  private:
    typedef double
    (*function_ptr)(boost::numeric::ublas::vector<double>, double, double);

    struct basis
    {
      function_ptr f;
      double op1;
      double op2;

      basis() :
        f(NULL), op1(0.0), op2(0.0)
      {
      }
    };

    std::string model_fname_;
    std::ifstream in_;

    boost::numeric::ublas::matrix<double> principal_components_;

    boost::numeric::ublas::vector<double> weights_;
    std::vector<std::vector<basis> > predictors_;

    std::vector<std::string> input_metrics_;

    static double
    identity_func(boost::numeric::ublas::vector<double> v, double op1,
        double op2);
    static double
    power_func(boost::numeric::ublas::vector<double> v, double op1, double op2);
    static double
    crossterm_func(boost::numeric::ublas::vector<double> v, double op1,
        double op2);
    static double
    sqrt_func(boost::numeric::ublas::vector<double> v, double op1, double op2);
    static double
    log_func(boost::numeric::ublas::vector<double> v, double op1, double op2);
    static double
        hinge1_func(boost::numeric::ublas::vector<double> v, double op1,
            double op2);
    static double
        hinge2_func(boost::numeric::ublas::vector<double> v, double op1,
            double op2);

    static debug eigerdbg_;

  public:
    typedef boost::intrusive_ptr<eigermodel> ptr;

    /// Create Eiger model from file.
    static eigermodel::ptr
    construct(const std::string &model_fname);

    /// Goodbye
    virtual
    ~eigermodel()
    {
    }

    virtual std::string
    to_string() const
    {
      return "eigermodel";
    }

    /// Poll the eiger model
    double
    poll(std::map<std::string, double> params);

  };

} // end of namespace sstmac

#endif
