/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/software/api/hpx/agas/agas_single.h>
#include <sstmac/software/api/hpx/parcels/resolve_gid.h>
#include <sstmac/software/api/hpx/parcels/resolve_gid_response.h>
#include <sstmac/software/api/hpx/parcels/new_gid.h>
#include <sstmac/software/api/hpx/parcels/new_gid_response.h>
#include <sstmac/common/errors.h>

namespace hpx
{

  void
  agas_single::handle(const sstmac::sst_message::ptr &msg)
  {

    if (msg)
      {
        if (msg->get_type() == sstmac::sst_message::COMPUTE)
          {

          }
        else
          {
            hpx::parcels::parcel::ptr par = boost::dynamic_pointer_cast<
                hpx::parcels::parcel>(msg);

            if (!par)
              {
                throw sstmac::ssterror(
                    "hpx::agas_single -- could not cast message to parcel");
              }

            switch (par->type())
              {
            case hpx::parcels::parcel::NEW_GID:
              {
                SSTMAC_DEBUG << "agas_server["
                      << os_->get_my_addr()->unique_id()
                      << "]: got NEW_GID message \n";
                hpx::parcels::new_gid::ptr newgidmsg =
                    boost::dynamic_pointer_cast<hpx::parcels::new_gid>(msg);
                globalid gid;
                unsigned long long waitid;
                bool found = agas_single::new_gid(newgidmsg->new_gid_type(),
                    gid, waitid, newgidmsg->origin());
                if (!found)
                  {
                    sst_throw(sstmac::ssterror, "hpx_agas_single::handle - new gid - can't create one here for some reason");
                  }

                hpx::parcels::new_gid_response::ptr resp =
                    hpx::parcels::new_gid_response::construct(gid,
                        newgidmsg->origin(), newgidmsg->wait_id(), os_->now());
                resp->toaddr_ = newgidmsg->origin();
                resp->fromaddr_ = locality_;
                resp->set_lib_name(libname_); // send to my peer on the other side
                os_->execute_kernel(sstmac::sw::AMI::AMI_COMM_SEND, resp);
                break;
              }

            case hpx::parcels::parcel::NEW_GID_RESPONSE:
              {
                SSTMAC_DEBUG << "agas_server["
                      << os_->get_my_addr()->unique_id()
                      << "]: got NEW_GID_RESPONSE message \n";
                hpx::parcels::new_gid_response::ptr
                    newgidmsg = boost::dynamic_pointer_cast<
                        hpx::parcels::new_gid_response>(msg);
                unsigned long long waitid = newgidmsg->wait_id();
                cache_[newgidmsg->gid().gid_] = locality_;
                if (newgid_waiting_.find(waitid) != newgid_waiting_.end())
                  {
                    newgid_value_[waitid] = newgidmsg->gid();
                    os_->unblock(newgid_waiting_[waitid]);
                  }
                else
                  {
                    threadman_->gid_created(newgidmsg->gid(), waitid);
                  }
                break;
              }
            case hpx::parcels::parcel::RESOLVE_GID:
              {
                SSTMAC_DEBUG << "agas_server["
                      << os_->get_my_addr()->unique_id()
                      << "]: got RESOLVE_GID message \n";
                hpx::parcels::resolve_gid::ptr newgidmsg =
                    boost::dynamic_pointer_cast<hpx::parcels::resolve_gid>(msg);
                sstmac::nodeaddress::ptr loc;
                unsigned long long waitid;
                bool found = agas_single::resolve_locality(newgidmsg->gid(),
                    loc, waitid);
                if (!found)
                  {
                    sst_throw(sstmac::ssterror, "hpx_agas_single::handle - resolve locality - didn't find it here for some reason");
                  }
                hpx::parcels::resolve_gid_response::ptr
                    resp =
                        hpx::parcels::resolve_gid_response::construct(
                            newgidmsg->gid(), loc, newgidmsg->wait_id(),
                            os_->now());
                resp->toaddr_ = newgidmsg->origin();
                resp->fromaddr_ = locality_;
                resp->set_lib_name(libname_); // send to my peer on the other side
                os_->execute_kernel(sstmac::sw::AMI::AMI_COMM_SEND, resp);
                break;
              }

            case hpx::parcels::parcel::RESOLVE_GID_RESPONSE:
              {
                SSTMAC_DEBUG << "agas_server["
                      << os_->get_my_addr()->unique_id()
                      << "]: got RESOLVE_GID_RESPONSE message \n";
                hpx::parcels::resolve_gid_response::ptr newgidmsg =
                    boost::dynamic_pointer_cast<
                        hpx::parcels::resolve_gid_response>(msg);
                unsigned long long waitid = newgidmsg->wait_id();

                cache_[newgidmsg->gid().gid_] = newgidmsg->where();

                threadman_->locality_resolved(newgidmsg->where(), waitid);
                break;
              }

            case hpx::parcels::parcel::UPDATE_LOCALITY:
              {
                SSTMAC_DEBUG << "agas_server["
                      << os_->get_my_addr()->unique_id()
                      << "]: got UPDATE_LOCALITY message \n";
                hpx::parcels::update_locality::ptr upmsg =
                    boost::dynamic_pointer_cast<hpx::parcels::update_locality>(
                        msg);

                if (server_here_)
                  {
                    globalid gid = upmsg->gid();
                    lva l = lvamap_[gid.gid_];
                    lva newl(gid.gid_, upmsg->upaddr(), l.etype());
                    lvamap_[gid.gid_] = newl;
                  }
                break;
              }

            default:
              sst_throw(sstmac::ssterror, "hpx::agas_single::handle - unknown message type %d", par->type());

              }
          }
      }
  }

  /* void
   agas_single::service_one()
   {
   if (pending_.size() > 0)
   {
   hpx::parcels::parcel::ptr par = pending_.front();
   pending_.pop();

   switch (par->type())
   {
   case hpx::parcels::parcel::RESOLVE_GID:
   {
   SSTMAC_DEBUG << "hpx::agas_single(" << locality_->to_string()
   << "): servicing resolve_gid \n";
   hpx::parcels::resolve_gid::ptr ngp = boost::dynamic_pointer_cast<
   hpx::parcels::resolve_gid>(par);
   if (lvamap_.find(ngp->gid().gid_) == lvamap_.end())
   {
   throw sstmac::ssterror(
   "hpx::agas_single::service_one() -- resolve-gid - can't find gid in lva map");
   }

   hpx::parcels::resolve_gid_response::ptr resp =
   hpx::parcels::resolve_gid_response::construct(ngp->gid(),
   locality_, ngp->wait_id(), os_->now());
   resp->set_lib_name(libname_);
   resp->toaddr_ = ngp->origin();
   os_->execute_kernel(sstmac::sw::AMI::AMI_COMM,
   sstmac::sw::AMI::AMI_COMM_SEND, resp);
   break;
   }

   case hpx::parcels::parcel::RESOLVE_GID_RESPONSE:
   {
   SSTMAC_DEBUG << "hpx::agas_single(" << locality_->to_string()
   << "): servicing resolve_gid_response \n";
   hpx::parcels::resolve_gid_response::ptr ngp =
   boost::dynamic_pointer_cast<hpx::parcels::resolve_gid_response>(
   par);
   if (waiting_.find(ngp->wait_id()) == waiting_.end())
   {
   std::cout << "wait id: " << ngp->wait_id() << "\n";
   throw sstmac::ssterror(
   "hpx::agas_single::service_one() - resolve_gid_response -- couldn't find waiter in map");
   }

   rx_ = par;
   os_->unblock(waiting_[ngp->wait_id()]);

   break;
   }

   default:
   throw sstmac::ssterror("hpx::agas_single -- unknown parcel type");

   }
   }
   }*/

  bool
  agas_single::new_gid(lva::ENTITY_TYPE t, globalid &gid,
      unsigned long long &waitid)
  {
    return new_gid(t, gid, waitid, locality_);
  }

  bool
  agas_single::new_gid(lva::ENTITY_TYPE t, globalid &gid,
      unsigned long long &waitid, sstmac::nodeaddress::ptr loc)
  {
    if (server_here_)
      {
        SSTMAC_DEBUG << "agas_server[" << os_->get_my_addr()->unique_id()
              << "]: new_gid on server \n";
        globalid ret(gid_count_++);
        lva l(ret, loc, t);
        lvamap_[ret.gid_] = l;
        gid = ret;
        return true;
      }
    else
      {
        SSTMAC_DEBUG << "agas_server[" << os_->get_my_addr()->unique_id()
              << "]: new_gid, need to ask server \n";

        long waiting = WAITING_ID++;
        hpx::parcels::new_gid::ptr msg = hpx::parcels::new_gid::construct(t,
            locality_, waiting, os_->now());
        msg->toaddr_ = server_node_;
        msg->fromaddr_ = locality_;
        msg->set_lib_name(libname_); // send to my peer on the other side
        os_->execute_kernel(sstmac::sw::AMI::AMI_COMM_SEND, msg);

        waitid = waiting;
        return false;
      }

  }

  void
  agas_single::gid_move_complete(globalid gid)
  {
    if (server_here_)
      {
        lva l = lvamap_[gid.gid_];
        lva newl(gid.gid_, locality_, l.etype());
        lvamap_[gid.gid_] = newl;
      }
    else
      {
        //update the server
        hpx::parcels::update_locality::ptr
            msg = hpx::parcels::update_locality::construct(gid, locality_,
                os_->now());
        msg->toaddr_ = server_node_;
        msg->fromaddr_ = locality_;
        msg->set_lib_name(libname_); // send to my peer on the other side
        os_->execute_kernel(sstmac::sw::AMI::AMI_COMM_SEND, msg);

        cache_[gid.gid_] = locality_;
      }
  }

  void
  agas_single::gid_moving(globalid gid, sstmac::nodeaddress::ptr to)
  {
    if (server_here_)
      {
        lva l = lvamap_[gid.gid_];
        lva newl(gid.gid_, to, l.etype());
        lvamap_[gid.gid_] = newl;
      }
    else
      {
        cache_[gid.gid_] = to;
      }
  }

  bool
  agas_single::resolve_locality(const globalid &gid,
      sstmac::nodeaddress::ptr &loc, unsigned long long &waitid)
  {

    if (server_here_)
      {
        SSTMAC_DEBUG << "agas_server[" << os_->get_my_addr()->unique_id()
              << "]: resolve_locality on server for gid " << gid.gid_ << "\n";
        if (lvamap_.find(gid.gid_) != lvamap_.end())
          {
            loc = lvamap_[gid.gid_].addr();
            return true;
          }
        else
          {
            sst_throw(sstmac::ssterror, "hpx_agas_single::resolve_locality: can't find gid %lu", gid.gid_);
          }
      }
    else if (cache_.find(gid.gid_) != cache_.end())
      {
        SSTMAC_DEBUG << "agas_server[" << os_->get_my_addr()->unique_id()
              << "]: resolve_locality found " << gid.gid_ << " in cache \n";

        loc = cache_[gid.gid_];
        return true;
      }
    else
      {
        SSTMAC_DEBUG << "agas_server[" << os_->get_my_addr()->unique_id()
              << "]: resolve_locality, need to ask server \n";
        long waiting = WAITING_ID++;
        hpx::parcels::resolve_gid::ptr msg =
            hpx::parcels::resolve_gid::construct(gid, locality_, waiting,
                os_->now());
        msg->toaddr_ = server_node_;
        msg->fromaddr_ = locality_;
        msg->set_lib_name(libname_); // send to my peer on the other side
        os_->execute_kernel(sstmac::sw::AMI::AMI_COMM_SEND, msg);
        // loc = server_node_;
        //  return true;
        waitid = waiting;
        return false;
      }

  }

}
