/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/software/api/hpx/hpx_app.h>
#include <sstmac/software/libraries/compute/sstmac_compute.h>
#include <sstmac/software/api/hpx/agas/agas_dist.h>
#include <sstmac/software/api/hpx/agas/agas_ideal.h>
#include <sstmac/software/api/hpx/agas/agas_single.h>

namespace hpx
{

  using namespace sstmac;

  bool hpx_app::gidinit_ = false;

  void
  hpx_app::init()
  {

    std::string libn = sstmac::sstmac_env::params->get_optional_param("hpx_agas", "hpx_agas_dist");
    //start the agas server if not already there
    sstmac::sw::library::ptr agas = os_->get_lib(libn);
    if (agas == sstmac::sw::library::ptr()) // only do one mpiserver per app per node
    {
      if(libn == "hpx_agas_dist"){
        agas = agas_dist::construct();
      }else if(libn == "hpx_agas_ideal"){
        agas = agas_ideal::construct();
      }else if(libn == "hpx_agas_single"){
        agas = agas_single::construct();
      }else{
        sst_throw(ssterror, "hpx_app::init - unknown agas server given: %s", libn.c_str());
      }

      os_->register_lib(agas);
      agas_server::ptr casted = boost::dynamic_pointer_cast<agas_server>(agas);
      casted->start();

    }
    agas_ = boost::dynamic_pointer_cast<agas_server>(agas);


    //start the thread manager if not already there
    sstmac::sw::library::ptr tman =
        os_->get_lib(hpx_thread_manager::lib_name());
    if (tman == sstmac::sw::library::ptr()) // only do one mpiserver per app per node
    {
      tman = hpx_thread_manager::construct(agas_);
      os_->register_lib(tman);
      hpx_thread_manager::ptr casted = boost::dynamic_pointer_cast<
          hpx_thread_manager>(tman);
      casted->start();

    }
    threadman_ = boost::dynamic_pointer_cast<hpx_thread_manager>(tman);
    register_actions();
    register_components();

    agas_->set_threadman(threadman_);

    threadman_->thread_is_starting(ptr(this));

  }

  hpx_thread_manager::ptr
  hpx_app::get_current_thread_manager()
  {
    thread::ptr t = sstmac::sw::OS_API::get_current_thread();
    hpx_app::ptr app = boost::dynamic_pointer_cast<hpx_app>(t);

    if (!app)
    {
      boost::intrusive_ptr<hpx_thread> pxth = boost::dynamic_pointer_cast<
          hpx_thread>(t);

      if (!pxth)
      {

        throw sstmac::ssterror(
            "hpx_app::get_current_thread_manager() - app didn't cast, make sure app is an hpx_app or an hpx_thread");
      }
      else
      {
        return pxth->get_thread_man();
      }
    }
    else
    {

      return app->threadman_;
    }
  }

  agas_server::ptr
  hpx_app::get_current_agas_server()
  {
    thread::ptr t = sstmac::sw::OS_API::get_current_thread();
    hpx_app::ptr app = boost::dynamic_pointer_cast<hpx_app>(t);

    if (!app)
    {
      boost::intrusive_ptr<hpx_thread> pxth = boost::dynamic_pointer_cast<
          hpx_thread>(t);

      if (!pxth)
      {

        throw sstmac::ssterror(
            "hpx_app::get_current_thread_manager() - app didn't cast, make sure app is an hpx_app or an hpx_thread");
      }
      else
      {
        return pxth->get_agas_server();
      }
    }
    else
    {

      return app->agas_;
    }
  }

  void
  hpx_app::hpx_block(sstmac::sw::key::ptr blocker)
  {

    thread::ptr t = sstmac::sw::OS_API::get_current_thread();
    hpx_app::ptr app = boost::dynamic_pointer_cast<hpx_app>(t);

    if (!app)
    {
      boost::intrusive_ptr<hpx_thread> pxth = boost::dynamic_pointer_cast<
          hpx_thread>(t);

      if (!pxth)
      {

        throw sstmac::ssterror(
            "hpx_app::hpx_block() - app didn't cast, make sure app is an hpx_app or an hpx_thread");
      }
      else
      {
        pxth->hpx_block(blocker);
      }
    }
    else
    {
      app->threadman_->thread_is_blocking(app, blocker);
    }

  }

  /* void
   hpx_app::hpx_thread_swap()
   {

   thread::ptr t = sstmac::sw::OS_API::get_current_thread();
   hpx_app::ptr app = boost::dynamic_pointer_cast<hpx_app>(t);

   if (!app)
   {
   boost::intrusive_ptr<hpx_thread> pxth = boost::dynamic_pointer_cast<
   hpx_thread>(t);

   if (!pxth)
   {

   throw sstmac::ssterror(
   "hpx_app::hpx_thread_swap() - app didn't cast, make sure app is an hpx_app or an hpx_thread");
   }
   else
   {
   pxth->compute(sstmac::timestamp(hpx_thread::THREAD_SWAP_OVERHEAD));
   }
   }
   else
   {
   app->compute(sstmac::timestamp(hpx_thread::THREAD_SWAP_OVERHEAD));

   }

   }*/

  void
  hpx_app::hpx_unblock(sstmac::sw::key::ptr blocker)
  {
    //SSTMAC_DEBUG << "hpx_app::hpx_unblock: enter\n";
    hpx_thread::ptr pxth;
    hpx_app::ptr app;

    //throw sstmac::ssterror(
    //    "hpx_app::hpx_unblock() - app didn't cast, make sure app is an hpx_app or an hpx_thread");

    sstmac::sw::key::blocking_t bl = blocker->get_blocking();
    if (bl.size() > 1)
    {
      sst_throw(ssterror, "can only handle one blocking thread per key, should fix this");
    }
    sstmac::sw::key::blocking_t::iterator it, end = bl.end();
    for (it = bl.begin(); it != end; it++)
    {
      app = boost::dynamic_pointer_cast<hpx_app>(it->second);
      if (!app)
      {
        pxth = boost::dynamic_pointer_cast<hpx_thread>(it->second);
        if (!pxth)
        {
          std::cout << "thing is: " << it->second->to_string() << "\n";
          throw sstmac::ssterror(
              "hpx_app::hpx_unblock() - tried to cast blockers in the key to hpx_threads, but didn't work");

        }

        pxth->hpx_unblock(blocker);
      }
      else
      {
        app->threadman_->thread_is_unblocking(app, blocker);
      }
    }

  }

  void
  hpx_app::hpx_send_msg(sstmac::sst_message::ptr msg)
  {
    thread::ptr t = sstmac::sw::OS_API::get_current_thread();
    hpx_app::ptr app = boost::dynamic_pointer_cast<hpx_app>(t);

    if (!app)
    {
      boost::intrusive_ptr<hpx_thread> pxth = boost::dynamic_pointer_cast<
          hpx_thread>(t);

      if (!pxth)
      {

        throw sstmac::ssterror(
            "hpx_app::hpx_send_msg() - app didn't cast, make sure app is an hpx_app or an hpx_thread");
      }
      else
      {
        pxth->hpx_send(msg);
      }
    }
    else
    {
      app->os_->execute_kernel(sstmac::sw::AMI::AMI_COMM_SEND, msg);
    }
  }

  sstmac::nodeaddress::ptr
  hpx_app::gbl_find_here()
  {
    thread::ptr t = sstmac::sw::OS_API::get_current_thread();
    hpx_app::ptr app = boost::dynamic_pointer_cast<hpx_app>(t);

    if (!app)
    {
      boost::intrusive_ptr<hpx_thread> pxth = boost::dynamic_pointer_cast<
          hpx_thread>(t);

      if (!pxth)
      {

        throw sstmac::ssterror(
            "hpx_app::hpx_send_msg() - app didn't cast, make sure app is an hpx_app or an hpx_thread");
      }
      else
      {
        return pxth->find_here();
      }
    }
    else
    {
      return app->find_here();
    }
  }

  sstmac::sw::environment::const_ptr
  hpx_app::get_env()
  {
    thread::ptr t = sstmac::sw::OS_API::get_current_thread();
    hpx_app::ptr app = boost::dynamic_pointer_cast<hpx_app>(t);

    if (!app)
    {
      boost::intrusive_ptr<hpx_thread> pxth = boost::dynamic_pointer_cast<
          hpx_thread>(t);

      if (!pxth)
      {

        throw sstmac::ssterror(
            "hpx_app::hpx_send_msg() - app didn't cast, make sure app is an hpx_app or an hpx_thread");
      }
      else
      {
        return pxth->get_env(app->id_.app_);
      }
    }
    else
    {
      return app->os_->get_env(app->id_.app_);
    }
  }

  sstmac::sw::OS_API::ptr
  hpx_app::get_os()
  {
    thread::ptr t = sstmac::sw::OS_API::get_current_thread();
    hpx_app::ptr app = boost::dynamic_pointer_cast<hpx_app>(t);

    if (!app)
    {
      boost::intrusive_ptr<hpx_thread> pxth = boost::dynamic_pointer_cast<
          hpx_thread>(t);

      if (!pxth)
      {

        throw sstmac::ssterror(
            "hpx_app::get_os() - app didn't cast, make sure app is an hpx_app or an hpx_thread");
      }
      else
      {
        return pxth->get_os();
      }
    }
    else
    {
      return app->os_;
    }
  }

  void
  hpx_app::compute(const sstmac::timestamp &ti)
  {
    thread::ptr t = sstmac::sw::OS_API::get_current_thread();
    hpx_app::ptr app = boost::dynamic_pointer_cast<hpx_app>(t);

    if (!app)
    {
      boost::intrusive_ptr<hpx_thread> pxth = boost::dynamic_pointer_cast<
          hpx_thread>(t);

      if (!pxth)
      {

        throw sstmac::ssterror(
            "hpx_app::get_os() - app didn't cast, make sure app is an hpx_app or an hpx_thread");
      }
      else
      {
        return pxth->compute(ti);
      }
    }
    else
    {
      app->app::compute(ti);
    }
  }

  void
  hpx_app::compute_loop(long long from, long long to, long long numlines)
  {
    thread::ptr t = sstmac::sw::OS_API::get_current_thread();
    hpx_app::ptr app = boost::dynamic_pointer_cast<hpx_app>(t);

    if (!app)
    {
      boost::intrusive_ptr<hpx_thread> pxth = boost::dynamic_pointer_cast<
          hpx_thread>(t);

      if (!pxth)
      {

        throw sstmac::ssterror(
            "hpx_app::get_os() - app didn't cast, make sure app is an hpx_app or an hpx_thread");
      }
      else
      {
        return pxth->HPX_compute_loop(from, to, numlines);
      }
    }
    else
    {
      SSTMAC_compute_loop(from, to, numlines);
    }
  }

  void
  hpx_app::compute_loop2(long long from1, long long to1, long long from2, long long to2, long long numlines)
  {
    thread::ptr t = sstmac::sw::OS_API::get_current_thread();
    hpx_app::ptr app = boost::dynamic_pointer_cast<hpx_app>(t);

    if (!app)
    {
      boost::intrusive_ptr<hpx_thread> pxth = boost::dynamic_pointer_cast<
          hpx_thread>(t);

      if (!pxth)
      {

        throw sstmac::ssterror(
            "hpx_app::get_os() - app didn't cast, make sure app is an hpx_app or an hpx_thread");
      }
      else
      {
        return pxth->HPX_compute_loop2(from1, to1, from2, to2, numlines);
      }
    }
    else
    {
      SSTMAC_compute_loop2(from1, to1, from2, to2, numlines);
    }
  }

  void
  hpx_app::compute_loop3(long long from1, long long to1, long long from2, long long to2, long long from3,
      long long to3, long long numlines)
  {
    thread::ptr t = sstmac::sw::OS_API::get_current_thread();
    hpx_app::ptr app = boost::dynamic_pointer_cast<hpx_app>(t);

    if (!app)
    {
      boost::intrusive_ptr<hpx_thread> pxth = boost::dynamic_pointer_cast<
          hpx_thread>(t);

      if (!pxth)
      {

        throw sstmac::ssterror(
            "hpx_app::get_os() - app didn't cast, make sure app is an hpx_app or an hpx_thread");
      }
      else
      {
        return pxth->HPX_compute_loop3(from1, to1, from2, to2, from3, to3,
            numlines);
      }
    }
    else
    {
      SSTMAC_compute_loop3(from1, to1, from2, to2, from3, to3, numlines);
    }
  }

  void
  hpx_app::compute_loop4(long long from1, long long to1, long long from2, long long to2, long long from3,
      long long to3, long long from4, long long to4, long long numlines)
  {
    thread::ptr t = sstmac::sw::OS_API::get_current_thread();
    hpx_app::ptr app = boost::dynamic_pointer_cast<hpx_app>(t);

    if (!app)
    {
      boost::intrusive_ptr<hpx_thread> pxth = boost::dynamic_pointer_cast<
          hpx_thread>(t);

      if (!pxth)
      {

        throw sstmac::ssterror(
            "hpx_app::get_os() - app didn't cast, make sure app is an hpx_app or an hpx_thread");
      }
      else
      {
        return pxth->HPX_compute_loop4(from1, to1, from2, to2, from3, to3,
            from4, to4, numlines);
      }
    }
    else
    {
      SSTMAC_compute_loop4(from1, to1, from2, to2, from3, to3, from4, to4,
          numlines);
    }
  }

  void
  hpx_app::hpx_register_lib(sstmac::sw::library::ptr lib)
  {
    thread::ptr t = sstmac::sw::OS_API::get_current_thread();
    hpx_app::ptr app = boost::dynamic_pointer_cast<hpx_app>(t);

    if (!app)
    {
      boost::intrusive_ptr<hpx_thread> pxth = boost::dynamic_pointer_cast<
          hpx_thread>(t);

      if (!pxth)
      {

        throw sstmac::ssterror(
            "hpx_app::get_os() - app didn't cast, make sure app is an hpx_app or an hpx_thread");
      }
      else
      {
        return pxth->register_lib(lib);
      }
    }
    else
    {
      app->app::register_lib(lib);
    }
  }
}
