/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_HPX_HPX_APP_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_HPX_HPX_APP_H_INCLUDED

#include <sstmac/software/process/app.h>
#include <sstmac/software/api/hpx/runtime/actions/hpx_action.h>
#include <sstmac/software/api/hpx/runtime/hpx_thread_manager.h>
#include <sstmac/software/api/hpx/agas/agas_server.h>
#include <sstmac/software/ami/environment.h>
#include <sstmac/software/api/hpx/runtime/hpx_runnable.h>

namespace hpx
{

  /**
   * A base type for applications that can run via mpi.
   * Intended to be safe for multiple inheritance (for applications
   * that use MPI and some other form of communication).
   */
  class hpx_app : public sstmac::sw::app, public hpx_runnable
  {

  protected:

    hpx_thread_manager::ptr threadman_;
    agas_server::ptr agas_;

    static bool gidinit_;

    hpx_app(sstmac::sw::softwareid id) :
      app(id)
    {
    }

    hpx_app(const sstmac::sim_parameters::ptr& params)
        : app(params)
    {
    }

    virtual void
    hpx_main() = 0;

    template<typename ActName>
      void
      hpx_register_action(std::string actname)
      {
        boost::intrusive_ptr<ActName> actptr = ActName::construct();
        threadman_->register_action<ActName> (actname, actptr);
      }

    void
    init();

    virtual void
    register_actions() = 0;

    virtual void
    register_components() = 0;

  public:

    typedef boost::intrusive_ptr<hpx_app> ptr;

    /// Goodbye.
    virtual
    ~hpx_app()
    {
    }

    virtual void
    skeleton_main()
    {
      //if (id_.task_.id == 0)
      //  {
      //globalid::set_num_localities(os_->get_env(id_.app_)->get_app_count());
      //    globalid::set_num_localities(os_->get_env(id_.app_)->get_node_count());
      //  }

      if (!gidinit_)
      {
        globalid::set_num_localities(os_->get_env(id_.app_)->get_app_count());
        gidinit_ = true;
      }

      init();

      if (id_.task_.id == 0)
      {

        //globalid::set_num_localities(os_->get_env(id_.app_)->get_node_count());
        hpx_main();
        threadman_->kill(true);
      }
    }

    virtual std::string
    to_string() const
    {
      return "hpx_app";
    }

    //convenience wrapper
    sstmac::nodeaddress::ptr
    find_here()
    {
      return os_->get_my_addr();
    }

    static hpx_thread_manager::ptr
    get_current_thread_manager();

    static agas_server::ptr
    get_current_agas_server();

    static void
    hpx_block(sstmac::sw::key::ptr blocker);

    //static void
    //hpx_thread_swap();

    static void
    hpx_unblock(sstmac::sw::key::ptr unblock);

    static void
    hpx_send_msg(sstmac::sst_message::ptr msg);

    static sstmac::nodeaddress::ptr
    gbl_find_here();

    static sstmac::sw::environment::const_ptr
    get_env();

    static sstmac::sw::OS_API::ptr
    get_os();

    static void
    compute(const sstmac::timestamp &t);

    static void
    compute_loop(long long from, long long to, long long numlines);

    static void
    compute_loop2(long long from1, long long to1, long long from2, long long to2, long long numlines);

    static void
    compute_loop3(long long from1, long long to1, long long from2, long long to2, long long from3, long long to3,
        long long numlines);

    static void
    compute_loop4(long long from1, long long to1, long long from2, long long to2, long long from3, long long to3,
        long long from4, long long to4, long long numlines);

    static void
    hpx_register_lib(sstmac::sw::library::ptr lib);

  };

} // end of namespace sstmac.

#endif
