/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_HPX_LCOS_AND_GATE_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_HPX_LCOS_AND_GATE_H_INCLUDED

#include <sstmac/software/api/hpx/lcos/future.h>
#include <sstmac/software/api/hpx/runtime/hpx_thread.h>
#include <sstmac/software/api/hpx/runtime/actions/hpx_action_params.h>
#include <sstmac/software/api/hpx/runtime/hpx_thread_manager.h>
#include <sstmac/software/api/hpx/hpx_app.h>
namespace hpx
{
  namespace lcos
  {

    class and_gate : public lco
    {
    public:
      typedef boost::intrusive_ptr<and_gate> ptr;

      // -------- useful stuff ----------- //

    protected:

      struct Gen
      {
        size_t generation_;
        std::vector<bool> ports_;
        hpx::lcos::future<int>::ptr future_;
        bool is_init_;
      };

      size_t gen_count_;
      size_t current_gen_num_;
      boost::unordered_map<size_t, Gen*> current_gen_;

      boost::unordered_map<long, std::list<sstmac::sw::key::ptr> > sync_keys_;

    public:

      /**
       * Returns a promise that can be waited on for the gate to finish.
       */
      boost::intrusive_ptr<future<int> >
      get_future(size_t s, size_t *gen)
      {
        SSTMAC_DEBUG << "hpx::lcos::and_gate::get_future \n";
        *gen = gen_count_++;
        boost::intrusive_ptr<future<int> > ret;
        if (current_gen_.find(*gen) == current_gen_.end())
        {
          Gen* g = new Gen;
          g->generation_ = *gen;
          g->future_ = future<int>::construct();
          g->ports_.resize(s, false);
          g->is_init_ = true;
          current_gen_[*gen] = g;
          ret = current_gen_[*gen]->future_;
        }
        else if (!current_gen_[*gen]->is_init_)
        {
          current_gen_[*gen]->ports_.resize(s, false);
          current_gen_[*gen]->is_init_ = true;
          ret = current_gen_[*gen]->future_;
          check_complete();
        }

        return ret;
      }

      /**
       *  Make sure the current generation is the one passed in, block otherwise
       */
      void
      synchronize(size_t gen, std::string func)
      {
        SSTMAC_DEBUG << "hpx::lcos::and_gate::synchronize...";
        if (gen < current_gen_num_)
        {
          sst_throw(ssterror, "and_gate::synchronize - input gen %d is less than current %d, which is strange", gen, current_gen_num_);
        }
        else if (gen > current_gen_num_)
        {
          SSTMAC_DEBUG << "blocking \n";
          sstmac::sw::key::category cat("HPX::and_gate::" + func);
          sstmac::sw::key::ptr k = sstmac::sw::key::construct(
              cat);
         // k->set_event(sstmac::sw::key::mpi);
          sync_keys_[gen].push_back(k);
          hpx_app::hpx_block(k);
        }
        else
        {
          SSTMAC_DEBUG << "not blocking, already synced \n";
        }

      }

      /**
       * Like Star Trek
       */
      size_t
      next_generation()
      {
        SSTMAC_DEBUG << "hpx::lcos::and_gate::next_generation \n";
        return ++current_gen_num_;
      }

      void
      check_complete()
      {
        SSTMAC_DEBUG << "hpx::lcos::and_gate::check_complete...";
        Gen * g = current_gen_[current_gen_num_];
        bool done = true;
        for (int i = 0; i < g->ports_.size() && done; i++)
        {
          done = done && g->ports_[i];
        }

        if (done)
        {
          SSTMAC_DEBUG << "done \n";
          sstmac::valuepayload<int>::const_ptr fake =
              sstmac::valuepayload<int>::construct(1);
          g->future_->set_value(fake, hpx_app::get_os());
          delete g;
          current_gen_.erase(current_gen_num_);
          current_gen_num_++;
          if (sync_keys_[current_gen_num_].size() > 0)
          {
            size_t sz = sync_keys_[current_gen_num_].size();
            for (int i = 0; i < sz; i++)
            {
              hpx_app::hpx_unblock(sync_keys_[current_gen_num_].front());
              sync_keys_[current_gen_num_].pop_front();
            }
          }
          sync_keys_.erase(current_gen_num_);
        }else{
          SSTMAC_DEBUG << "still waiting \n";
        }
      }
      /**
       *  Indicate that a port of the and_gate is finished, possibly unblocking
       *  the future and anything waiting on the synchronize (by going to the next generation)
       */
      void
      set(size_t w)
      {
        SSTMAC_DEBUG << "hpx::lcos::and_gate::set \n";
        if (current_gen_.find(current_gen_num_) == current_gen_.end())
        {
          //sst_throw(ssterror, "and_gate::set - couldn't find current gen %d when setting port %d", current_gen_num_, w);
          Gen *newg = new Gen;
          newg->generation_ = current_gen_num_;
          newg->future_ = future<int>::construct();
          newg->is_init_ = false;

          current_gen_[current_gen_num_] = newg;
        }
        Gen * g = current_gen_[current_gen_num_];

        //check for completion
        if (g->is_init_)
        {
          g->ports_[w] = true;
          check_complete();
        }
        else
        {
          if (g->ports_.size() < w + 1)
          {
            g->ports_.resize(w + 1, false);
          }

          g->ports_[w] = true;
        }
      }

    protected:
      // ptr self_;

      and_gate()
      {
        current_gen_num_ = 0;
        gen_count_ = 0;
        // sync_key_ = sstmac::sw::key::construct();
      }

    public:

      static ptr
      construct()
      {
        return ptr(new and_gate());
      }

      virtual
      ~and_gate()
      {
      }

    };

  }
}

#endif
