/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstream>

#include <sstmac/software/api/hpx/runtime/hpx_thread.h>
#include <sstmac/software/process/os_api.h>
#include <sstmac/software/api/hpx/runtime/hpx_thread_manager.h>

namespace hpx
{
  sstmac::debug hpx_thread::dbg_;
  long hpx_thread::hpx_thread_cnt_ = 0;

  sstmac::stat_time_window_sum::ptr hpx_thread::num_threads_;

  hpx_thread::hpx_thread(boost::intrusive_ptr<hpx_thread_manager> man,
      boost::intrusive_ptr<agas_server> agas, sstmac::sw::softwareid sid) : thread(sid),
    threadman_(man), agas_(agas)
  {
    if (!dbg_.is_real())
      {
        dbg_ = sstmac::debug("<debug> hpx | hpx_thread", sstmac::logger::cyan);
        sstmac::logger::register_logger(&dbg_);
      }

    if (!num_threads_)
      {
        num_threads_ = sstmac::stat_time_window_sum::construct(
            "<stats> hpx_threads", "num_threads.csv");
        sstmac::eventmanager::register_stat(num_threads_);
      }
    num_threads_->record(man->now(), 1);

    hpx_thread_id_ = hpx_thread_cnt_++;

    //blocking_on_blocker_ = false;
    //dont_block_ = false;
  }

  boost::intrusive_ptr<hpx_thread_manager>
  hpx_thread::get_thread_man()
  {
    return threadman_;
  }

  hpx_thread::~hpx_thread()
  {

  }

  boost::intrusive_ptr<agas_server>
  hpx_thread::get_agas_server()
  {
    return agas_;
  }

  void
  hpx_thread::hpx_block(sstmac::sw::key::ptr blocker)
  {
    /* SSTMAC_DEBUG << "hpx_block:enter \n";
     core_->thread_done();
     SSTMAC_DEBUG << "hpx_block:after thread_done \n";

     if (!dont_block_)
     {
     blocking_on_blocker_ = true;
     os_->block(blocker);
     blocking_on_blocker_ = false;

     }
     dont_block_ = false;

     SSTMAC_DEBUG << "hpx_block:after block \n";
     threadman_->set_me_to_core(ptr(this), false);
     SSTMAC_DEBUG << "hpx_block:after set_me_to_core \n";
     double tm = core_->thread_start(ptr(this), false);
     SSTMAC_DEBUG << "hpx_block:after thread_start \n";
     //core_->add_thread(ptr(this));

     if (tm > 0)
     {
     libsleep_->sleep(sstmac::timestamp(tm));
     }

     SSTMAC_DEBUG << "hpx_block:after swap overhead \n";*/
    threadman_->thread_is_blocking(ptr(this), blocker);
  }

  /*void
   hpx_thread::set_core(boost::intrusive_ptr<coremodel> c)
   {
   core_ = c;
   }

   boost::intrusive_ptr<coremodel>
   hpx_thread::get_core()
   {
   return core_;
   }
   */
  void
  hpx_thread::hpx_unblock(sstmac::sw::key::ptr blocker)
  {
     SSTMAC_DEBUG << "hpx_unblock:enter \n";
     /*if (blocking_on_blocker_)
     {
     SSTMAC_DEBUG << "hpx_block:is blocking, unblock him \n";
     os_->unblock(blocker);
     }
     else
     {
     SSTMAC_DEBUG << "hpx_block:was not blocking \n";
     dont_block_ = true;
     }*/
    threadman_->thread_is_unblocking(ptr(this), blocker);
  }

  void
  hpx_thread::hpx_send(sstmac::sst_message::ptr msg)
  {
    os_->execute_kernel(sstmac::sw::AMI::AMI_COMM_SEND, msg);
  }

  sstmac::nodeaddress::ptr
  hpx_thread::find_here()
  {
    return os_->get_my_addr();
  }

  sstmac::sw::environment::const_ptr
  hpx_thread::get_env(sstmac::sw::appid aid)
  {
    return os_->get_env(aid);
  }

  sstmac::sw::OS_API::ptr
  hpx_thread::get_os()
  {
    return os_;
  }

  void
  hpx_thread::compute(const sstmac::timestamp &t)
  {
    if (!libcomp_)
      {
        throw sstmac::ssterror(
            "hpx_thread::compute -- compute lib has not been initialized");
      }
    libcomp_->compute(t);
  }

  void
  hpx_thread::register_lib(sstmac::sw::library::ptr lib)
  {
    os_->register_lib(lib);
  }

  // ----------- compute loop functions ------------- //
  void
  hpx_thread::HPX_compute_loop(long long from, long long to, long long numlines)
  {
    if (!libloops_)
      {
        std::stringstream ss;
        ss << "hpx_thread::lib_compute_loops-" << hpx_thread_id_;
        libloops_ = sstmac::sw::lib_compute_loops::construct(ss.str());
        os_->register_lib(libloops_);
      }

    libloops_->compute_loop(from, to, numlines, thread::tid().id);
  }

  void
  hpx_thread::HPX_compute_loop2(long long from1, long long to1, long long from2, long long to2,
      long long numlines)
  {
    if (!libloops_)
      {
        std::stringstream ss;
        ss << "hpx_thread::lib_compute_loops-" << hpx_thread_id_;
        libloops_ = sstmac::sw::lib_compute_loops::construct(ss.str());
        os_->register_lib(libloops_);
      }

    libloops_->compute_loop2(from1, to1, from2, to2, numlines, thread::tid().id);
  }

  void
  hpx_thread::HPX_compute_loop3(long long from1, long long to1, long long from2, long long to2,
      long long from3, long long to3, long long numlines)
  {
    if (!libloops_)
      {
        std::stringstream ss;
        ss << "hpx_thread::lib_compute_loops-" << hpx_thread_id_;
        libloops_ = sstmac::sw::lib_compute_loops::construct(ss.str());
        os_->register_lib(libloops_);
      }

    libloops_->compute_loop3(from1, to1, from2, to2, from3, to3, numlines,
        thread::tid().id);
  }

  void
  hpx_thread::HPX_compute_loop4(long long from1, long long to1, long long from2, long long to2,
      long long from3, long long to3, long long from4, long long to4, long long numlines)
  {
    if (!libloops_)
      {
        std::stringstream ss;
        ss << "hpx_thread::lib_compute_loops-" << hpx_thread_id_;
        libloops_ = sstmac::sw::lib_compute_loops::construct(ss.str());
        os_->register_lib(libloops_);
      }
    libloops_->compute_loop4(from1, to1, from2, to2, from3, to3, from4, to4,
        numlines, thread::tid().id);
  }

}
