/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/software/api/hpx/runtime/hpx_thread_manager.h>
#include <sstmac/software/api/hpx/runtime/actions/hpx_action.h>
#include <sstmac/software/api/hpx/components/hpx_component.h>

#include <sstmac/software/api/hpx/parcels/remote_action.h>
#include <sstmac/software/api/hpx/parcels/remote_action_result.h>
#include <sstmac/software/api/hpx/parcels/move_component.h>
#include <sstmac/software/api/hpx/lcos/remote_callback.h>
#include <sstmac/software/api/hpx/runtime/hpx_thread.h>
#include <sstmac/software/api/hpx/runtime/hpx_thread_start.h>

#include <sstmac/software/api/hpx/runtime/hpx_random_balancer.h>

namespace hpx
{
  long hpx_thread_manager::TRANS_ID_CNT = 0;
  sstmac::debug hpx_thread_manager::dbg_;

  using namespace sstmac;
  using namespace sstmac::sw;

  event_trace::ptr hpx_thread_manager::stat_threadcores_ = 0;

  hpx_thread_manager::hpx_thread_manager(agas_server::ptr agas) :
    agas_(agas)
  {
    libname_ = lib_name();

    if (!dbg_.is_real())
    {
      dbg_ = sstmac::debug("<debug> hpx | hpx_thread_manager",
          sstmac::logger::blue);
      sstmac::logger::register_logger(&dbg_);
    }

  }

  hpx_thread_manager::~hpx_thread_manager()
  {
  }

  void
  hpx_thread_manager::start()
  {
    service::start();

    int c = sstmac::sstmac_env::params->get_int_param("node_cores");
    swaps_per_core_ = sstmac::sstmac_env::params->get_int_param(
        "hpx_swaps_per_core");
    for (int i = 0; i < c; i++)
    {
      cores_[i] = coremodel::construct(i, swaps_per_core_);
    }
    busy_cores_ = 0;

    //compkey_ = sstmac::sw::key::construct();
    libcomp_ = sstmac::sw::lib_compute_time::construct(
        "hpx_thread_manager-lib-comp");
    //libcomp_->init_handler(sstmac::sw::unblockhandler::construct(os_, compkey_));
    //libcomp_->init_waiter(sstmac::sw::blockwaiter::construct(os_, compkey_));
    register_lib(libcomp_);

    selfptr_ = boost::dynamic_pointer_cast<hpx_thread_manager>(ptr(this));

    std::string lbname = sstmac::sstmac_env::params->get_optional_param(
        "hpx_load_balancer", "none");

    if (lbname == "none")
    {
      lb_ = hpx_load_balancer::construct();
    }
    else if (lbname == "random")
    {
      lb_ = hpx_random_balancer::construct();
    }

    if (sstmac::sstmac_env::params->has_param("hpx_load_balancer"))
    {
      lb_time_ = sstmac_env::params->get_time_param("hpx_lb_time");
      lb_msg_ = timed_message::construct(lb_time_);
      lb_ev_ = sstmac::new_event(ptr(this), &hpx_thread_manager::handle,
          lb_msg_);
      os_->get_eventmanager()->schedule(now() + lb_time_, lb_ev_);
    }

    if (!stat_threadcores_)
    {
      stat_threadcores_ = event_trace::construct("<trace> hpx",
          "hpx_threadcores.dat");
      logger::register_logger(stat_threadcores_.get());
    }

    SSTMAC_DEBUG << "hpx_thread_manager started on node "
          << os_->get_my_addr()->to_string() << "\n";
  }

  /**
   * This function should be called by the main thread with true for sendtoothers, false for others
   */
  void
  hpx_thread_manager::kill(bool sendtoothers)
  {
    if (lb_msg_)
    {
      SSTMAC_DEBUG << "hpx_thread_manager[" << os_->get_my_addr()->to_string()
            << "]: killing others...\n";
      if (sendtoothers)
      {
        sstmac::sw::environment::const_ptr env = os_->get_env(
            sstmac::sw::appid(1));

        for (long i = 0; i < env->get_app_count(); i++)
        {
          sstmac::nodeaddress::ptr to = env->get_node_address(i);
          hpx::parcels::kill_parcel::ptr k =
              hpx::parcels::kill_parcel::construct();
          k->toaddr_ = to;
          k->fromaddr_ = os_->get_my_addr();
          k->set_lib_name(libname_);
          os_->execute_kernel(AMI::AMI_COMM_SEND, k);
        }
      }
      lb_ev_->cancel();
    }

  }

  void
  hpx_thread_manager::handle(const sstmac::sst_message::ptr &msg)
  {

    if (msg->get_type() == sstmac::sst_message::SEND_ACK)
    {
      return;
    }

    if (msg == lb_msg_)
    {
      SSTMAC_DEBUG << "hpx_thread_manager[" << os_->get_my_addr()->to_string()
            << "]: load balancing...\n";
      typedef boost::unordered_map<hpx::components::hpx_component::ptr,
          nodeaddress::ptr> changetype;
      changetype change;
      lb_->balance(components_, os_->get_env(appid(1)), change);
      changetype::iterator it, end = change.end();
      for (it = change.begin(); it != end; it++)
      {
        SSTMAC_DEBUG << "sending " << it->first->gid().gid_ << " to "
              << it->second->to_string() << "\n";
        hpx::parcels::move_component::ptr mc =
            hpx::parcels::move_component::construct(it->first, now());
        mc->toaddr_ = it->second;
        mc->fromaddr_ = os_->get_my_addr();
        mc->set_lib_name(libname_);
        agas_->gid_moving(it->first->gid(), it->second);
        it->first->unregister_libs();
        os_->execute_kernel(AMI::AMI_COMM_SEND, mc);
      }

      SSTMAC_DEBUG << "done load balancing \n";
      os_->get_eventmanager()->schedule(now() + lb_time_, lb_ev_);

      return;
    }

    hpx_thread_start::ptr ts = boost::dynamic_pointer_cast<hpx_thread_start>(
        msg);
    if (ts)
    {
      SSTMAC_DEBUG << "hpx_thread_manager[" << os_->get_my_addr()->to_string()
            << "]: starting thread " << ts->run_->to_string() << " on core  "
            << ts->run_->get_core() << "\n";
      //start thread here
      cores_[ts->run_->get_core()]->thread_start(now());
      cores_[ts->run_->get_core()]->clear_from_swap(ts->run_);
      os_->unblock(ts->run_->get_blocker());

      thread_scheduled_ = false;
      try_thread_q();
      return;
    }
    hpx::parcels::parcel::ptr parc = boost::dynamic_pointer_cast<
        hpx::parcels::parcel>(msg);

    if (!parc)
    {
      throw sstmac::ssterror(
          "hpx_thread_manager::handle(...) - can't cast msg to a parcel");
    }

    switch (parc->type())
    {

    case hpx::parcels::parcel::KILL:
    {
      SSTMAC_DEBUG << "hpx_thread_manager[" << os_->get_my_addr()->to_string()
            << "]: received kill msg \n";
      kill(false);
      break;
    }

    case hpx::parcels::parcel::REMOTE_ACTION:
    {

      hpx::parcels::remote_action::ptr rem = boost::dynamic_pointer_cast<
          hpx::parcels::remote_action>(parc);
      SSTMAC_DEBUG << "hpx_thread_manager[" << os_->get_my_addr()->to_string()
            << "]: -- invoking action " << rem->which()
            << " from remote call, returning to "
            << rem->returnto()->to_string() << "\n";
      hpx::lcos::remote_callback::ptr sendback =
          hpx::lcos::remote_callback::construct(rem->trans_id(),
              rem->returnto());
      if (rem->is_gid())
      {
        sstmac::nodeaddress::ptr loc;
        unsigned long long waitid = -1;
        bool found = agas_->resolve_locality(rem->gid(), loc, waitid);
        if (found && is_local(loc))
        {
          this->start_action(rem->which(), rem->params(), sendback, rem->gid());
        }
        else if (found)
        {
          SSTMAC_DEBUG << "gid is not here, forwarding to " << loc->to_string()
                << "\n";
          rem->toaddr_ = loc;
          rem->fromaddr_ = os_->get_my_addr();
          os_->execute_kernel(AMI::AMI_COMM_SEND, rem);
        }
        else
        {
          sst_throw(sstmac::ssterror, "hpx_thread_manager::handle - should have found locality on remote action");
        }
      }
      else
      {
        this->start_action(rem->which(), rem->params(), sendback,
            os_->get_my_addr());
      }
      break;

    }

    case hpx::parcels::parcel::REMOTE_ACTION_RESULT:
    {
      SSTMAC_DEBUG << "hpx_thread_manager[" << os_->get_my_addr()->to_string()
            << "]: -- got a result back from remote action \n";
      hpx::parcels::remote_action_result::ptr rem =
          boost::dynamic_pointer_cast<hpx::parcels::remote_action_result>(parc);
      if (rem_actions_.find(rem->trans_id()) == rem_actions_.end())
      {
        throw sstmac::ssterror(
            "hpx_thread_manager::handle() -- transaction id not found in saved remote action invocations");
      }
      hpx::lcos::base_lco_result::ptr lco = rem_actions_[rem->trans_id()];
      rem_actions_.erase(rem->trans_id());
      lco->set_value(rem->result(), os_);

      break;
    }

    case hpx::parcels::parcel::MOVE_COMPONENT:
    {

      hpx::parcels::move_component::ptr mc = boost::dynamic_pointer_cast<
          hpx::parcels::move_component>(parc);
      hpx::components::hpx_component::ptr comp = mc->get_comp();
      SSTMAC_DEBUG << "hpx_thread_manager[" << os_->get_my_addr()->to_string()
            << "]: component " << comp->gid().gid_ << " arrived here \n";
      comp->set_curr_locality(os_);
      comp->reregister_libs();
      agas_->gid_move_complete(comp->gid().gid_);
      components_[comp->gid().gid_] = comp;
      break;
    }

    default:
      std::cout << "parcel number: " << parc->type() << "\n";
      throw sstmac::ssterror(
          "hpx_thread_manager::handle(..) -- unknown parcel type");
      break;
    }

  }

  globalid
  hpx_thread_manager::create_component(std::string name)
  {
    SSTMAC_DEBUG << "hpx_thread_manager[" << os_->get_my_addr()->to_string()
          << "]: -- creating component... ";
    if (comp_factory_.find(name) == comp_factory_.end())
    {
      std::cout << "component name: " << name << "\n";
      throw sstmac::ssterror(
          "hpx_thread_manager::create_component() - component has not been registered yet");
    }

    globalid newid = agas_->new_gid(lva::COMPONENT);

    SSTMAC_DEBUG << "gid is " << newid.gid_ << "\n";
    hpx::components::hpx_component::ptr comp =
        comp_factory_[name]->clone(newid);
    comp->set_curr_locality(os_);
    components_[newid.gid_] = comp;
    return newid;
  }

  /**
   * This function is called when either a thread comes alive from being blocked,
   * or has just been created.  Either way, the thread is "ready to run"
   */
  void
  hpx_thread_manager::set_me_to_core(boost::intrusive_ptr<hpx_thread> t,
      bool creating)
  {

    /* SSTMAC_DEBUG << "hpx_thread_manager(" << os_->get_my_addr()->to_string()
     << "): set_me_to_core...";
     if (creating)
     { //this is easy, pick the core with the shortest time-to-run, since
     // we have to pay the creation/swap cost anyway
     int bestcore = -1;
     int mincore = 1000000;
     for (int i = 0; i < cores_.size(); i++)
     {
     int d = cores_[i]->get_q_depth();
     if (d < mincore)
     {
     mincore = d;
     bestcore = i;
     }
     }
     SSTMAC_DEBUG << "being created, found core " << bestcore << "\n";

     cores_[bestcore]->add_thread(t, os_->now());
     t->set_core(cores_[bestcore]);
     t->set_affinity(bestcore);
     }
     else
     {
     //if there's room in the swap-registers, leave it there (do nothing)
     if (t->get_core()->get_q_depth() > swaps_per_core_ + 1)
     {
     t->get_core()->clear_spot_in_swaps(t);

     // else find a new core
     int bestcore = -1;
     int mincore = 1000000;
     for (int i = 0; i < cores_.size(); i++)
     {
     int d = cores_[i]->get_q_depth();
     if (d < mincore)
     {
     mincore = d;
     bestcore = i;
     }
     }

     SSTMAC_DEBUG << "swapping to new core, found core " << bestcore
     << "\n";
     cores_[bestcore]->add_thread(t, os_->now());
     t->set_core(cores_[bestcore]);
     t->set_affinity(bestcore);
     }
     else
     {
     SSTMAC_DEBUG << "staying put in core " << t->get_core()->id()
     << "\n";
     t->get_core()->add_thread(t, os_->now());
     }

     }*/
  }

  void
  hpx_thread_manager::thread_is_blocking(hpx_runnable::ptr t,
      sstmac::sw::key::ptr blocker)
  {
    SSTMAC_DEBUG2 << "hpx_thread_manager[" << os_->get_my_addr()->to_string()
          << "]: thread_is_blocking on core " << t->get_core() << " \n";
    cores_[t->get_core()]->set_busy(false);
    cores_[t->get_core()]->thread_stop(now());
    cores_[t->get_core()]->put_in_swaps(t);
    busy_cores_--;

    try_thread_q();
    t->set_blocker(blocker);
    os_->block(blocker);
  }

  void
  hpx_thread_manager::thread_is_unblocking(hpx_runnable::ptr t,
      sstmac::sw::key::ptr blocker)
  {
    SSTMAC_DEBUG2 << "hpx_thread_manager[" << os_->get_my_addr()->to_string()
          << "]: thread_is_unblocking \n";
    //os_->unblock(blocker);
    // t->set_blocker(blocker);

    ready_.push(t);
    try_thread_q();
  }

  static sstmac::sw::key::category thread_key_cat("HPX::thread");
  void
  hpx_thread_manager::thread_is_starting(hpx_runnable::ptr t)
  {
    SSTMAC_DEBUG2 << "hpx_thread_manager[" << os_->get_my_addr()->to_string()
          << "]: thread_is_starting \n";
    if (allthreads_.find(t) != allthreads_.end())
    {
      sst_throw(ssterror, "hpx_thread_manager::thread_is_starting -- that's strange, i already have the thread");
    }

    allthreads_[t] = -1;

    sstmac::sw::key::ptr wait = sstmac::sw::key::construct(thread_key_cat);
    t->set_blocker(wait);
    //wait->set_event(sstmac::sw::key::general);
    ready_.push(t);
    try_thread_q();
    os_->block(wait);
    //wait->set_event(sstmac::sw::key::mpi);
  }

  void
  hpx_thread_manager::thread_is_finished(hpx_runnable::ptr t)
  {
    SSTMAC_DEBUG2 << "hpx_thread_manager[" << os_->get_my_addr()->to_string()
          << "]: thread_is_finished \n";

    if (allthreads_.find(t) == allthreads_.end())
    {
      sst_throw(ssterror, "hpx_thread_manager::thread_is_finished -- that's strange, i don't have the thread");
    }

    allthreads_.erase(t);

    cores_[t->get_core()]->set_busy(false);
    cores_[t->get_core()]->thread_stop(now());
    busy_cores_--;

    try_thread_q();
  }

  void
  hpx_thread_manager::try_thread_q()
  {
    SSTMAC_DEBUG2 << "hpx_thread_manager[" << os_->get_my_addr()->to_string()
          << "]: try_thread_q() \n";
    while (busy_cores_ < (int) cores_.size() && ready_.size() > 0)
    { //we're good, lets schedule
      hpx_runnable::ptr run = ready_.front();
      ready_.pop();

      for (int i = 0; i < (int) cores_.size(); i++)
      {
        if (!cores_[i]->get_busy())
        {
          cores_[i]->set_busy(true);
          run->set_core(i);

          sstmac::sw::thread::ptr t = boost::dynamic_pointer_cast<
              sstmac::sw::thread>(run);

          if (!run)
          {
            sst_throw(ssterror, "hpx_thread_manager::try_thread_q - couldn't cast to thread");
          }

          t->set_affinity(i);

          double d;

          if (run->is_starting())
          {
            d = hpx_thread::THREAD_CREATION_OVERHEAD;
            run->started();
          }
          else if (cores_[i]->in_swaps(run))
          {
            d = hpx_thread::THREAD_SWAP_FAST_OVERHEAD;
          }
          else
          {
            d = hpx_thread::THREAD_SWAP_FROM_MEM_OVERHEAD;
          }

          if (stat_threadcores_->is_active())
          {
            stat_threadcores_->collect(i,
                "setting-thread-to-core",
                os_->get_my_addr(),
                t->get_affinity(), t->get_id(), 0,
                0, (long) now().ticks_int64(), 0);
          }
          timestamp delay = timestamp(d);

          hpx_thread_start::ptr start = hpx_thread_start::construct(run, delay);

          os_->execute_kernel(sstmac::sw::AMI::AMI_COMP_SLEEP, start);

          busy_cores_++;
          break;
        }
      }

    }
  }

  void
  hpx_thread_manager::start_action(std::string actname,
      hpx_action_params::ptr params,
      boost::intrusive_ptr<hpx::lcos::base_lco_result> prom,
      sstmac::nodeaddress::ptr locality)
  {
    SSTMAC_DEBUG << "hpx_thread_manager[" << os_->get_my_addr()->to_string()
          << "]: -- starting action " << actname << " with locality "
          << locality->to_string() << " \n";
    globalid gid;
    do_action(locality, actname, params, prom, false, gid);
  }

  void
  hpx_thread_manager::start_action(std::string actname,
      hpx_action_params::ptr params,
      boost::intrusive_ptr<hpx::lcos::base_lco_result> prom, globalid gid)
  {
    SSTMAC_DEBUG << "hpx_thread_manager[" << os_->get_my_addr()->to_string()
          << "]: -- starting action " << actname << " with globalid "
          << gid.gid_ << "\n";
    sstmac::nodeaddress::ptr loc;
    unsigned long long waitid = -1;
    bool found = agas_->resolve_locality(gid, loc, waitid);

    if (found)
    {
      do_action(loc, actname, params, prom, true, gid);
    }
    else
    {
      action_info i;
      i.actname_ = actname;
      i.params_ = params;
      i.prom_ = prom;
      i.gid_ = gid;
      i.on_comp_ = true;

      locres_waiting_[waitid] = i;
    }

  }

  void
  hpx_thread_manager::gid_created(globalid gid, unsigned long long waitid)
  {
    sst_throw(unimplementederror, "hpx_thread_maanger::gid_created");
  }

  void
  hpx_thread_manager::locality_resolved(sstmac::nodeaddress::ptr loc,
      unsigned long long waitid)
  {
    if (locres_waiting_.find(waitid) == locres_waiting_.end())
    {
      sst_throw(ssterror, "hpx_thread_manager::locality_resolved: cannot find waitid %l", waitid);
    }

    do_action(loc, locres_waiting_[waitid].actname_,
        locres_waiting_[waitid].params_, locres_waiting_[waitid].prom_,
        locres_waiting_[waitid].on_comp_, locres_waiting_[waitid].gid_);

    locres_waiting_.erase(waitid);
  }

  void
  hpx_thread_manager::do_action(sstmac::nodeaddress::ptr loc,
      std::string actname, hpx_action_params::ptr params,
      boost::intrusive_ptr<hpx::lcos::base_lco_result> prom, bool on_comp,
      globalid gid)
  {
    if (is_local(loc))
    {
      SSTMAC_DEBUG << "hpx_thread_manager[" << os_->get_my_addr()->to_string()
            << "]: -- action is local, at " << loc->to_string() << " \n";
      if (actions_.find(actname) == actions_.end())
      {
        std::cout
            << "hpx_thread_manager::start_action - can't find action in map,  make sure it was registered \n";
        sst_throw(ssterror,
            "hpx_thread_manager::start_action - can't find action in map,  make sure it was registered");
      }

      boost::intrusive_ptr<actions::hpx_action> actlib =
          boost::dynamic_pointer_cast<actions::hpx_action>(actions_[actname]);

      if (!actlib)
      {
        std::cout
            << "hpx_thread_manager::start_action() - problem casting found action to hpx_action\n";
        sst_throw(ssterror,
            "hpx_thread_manager::start_action() - problem casting found action to hpx_action");
      }

      softwareid sid(appid(1), taskid(loc->unique_id()), threadid(0));
      boost::intrusive_ptr<hpx_action_thread> pxth;
      if (on_comp)
      {

        if (components_.find(gid.gid_) == components_.end())
        {
          std::cout
              << "hpx_thread_manager::start_action() -- component not found here, something wrong with agas \n";
          sst_throw(ssterror,
              "hpx_thread_manager::start_action() -- component not found here, something wrong with agas");
        }

        hpx::components::hpx_component::ptr comp = components_[gid.gid_];

        if (!comp)
        {
          std::cout
              << "hpx_thread_manager::start_action() -- component didn't cast right \n";
          sst_throw(ssterror,
              "hpx_thread_manager::start_action() -- component didn't cast right");
        }

        if (actioncompthreads_.find(actname) == actioncompthreads_.end()
            || actioncompthreads_[actname].find(gid.gid_)
                == actioncompthreads_[actname].end())
        {
          pxth = hpx_action_thread::construct(selfptr_, agas_, actlib, params,
              prom, comp, sid);
          actioncompthreads_[actname][gid.gid_] = pxth;
          SSTMAC_DEBUG << "hpx_thread_manager["
                << os_->get_my_addr()->to_string()
                << "]: -- starting action thread \n";
          os_->start_thread(actioncompthreads_[actname][gid.gid_]);
        }
        else
        {
          SSTMAC_DEBUG << "found action thread, adding to it \n";
          actioncompthreads_[actname][gid.gid_]->add_action(params, prom);
        }

      }
      else
      {
        if (actionthreads_.find(actname) == actionthreads_.end())
        {
          pxth = hpx_action_thread::construct(selfptr_, agas_, actlib, params,
              prom, sid);
          actionthreads_[actname] = pxth;
          SSTMAC_DEBUG << "hpx_thread_manager["
                << os_->get_my_addr()->to_string()
                << "]: -- starting action thread \n";
          os_->start_thread(actionthreads_[actname]);
        }
        else
        {
          SSTMAC_DEBUG << "found action thread, adding to it \n";
          actionthreads_[actname]->add_action(params, prom);
        }

      }

    }
    else
    {
      SSTMAC_DEBUG << "hpx_thread_manager[" << os_->get_my_addr()->to_string()
            << "]: -- action is remote, at " << loc->to_string() << " \n";
      long trans = TRANS_ID_CNT++;
      hpx::parcels::remote_action::ptr msg;

      if (on_comp)
      {
        msg = hpx::parcels::remote_action::construct(actname, params, trans,
            gid, os_->get_my_addr(), os_->now());
      }
      else
      {
        msg = hpx::parcels::remote_action::construct(actname, params, trans,
            os_->get_my_addr(), os_->now());
      }
      msg->toaddr_ = loc;
      msg->set_lib_name(libname_);
      rem_actions_[trans] = prom;
      os_->execute_kernel(sstmac::sw::AMI::AMI_COMM_SEND, msg);
    }
  }

}
