/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_HPX_RUNTIME_HPX_THREAD_MANAGER_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_HPX_RUNTIME_HPX_THREAD_MANAGER_H_INCLUDED

#include <sstmac/software/services/service.h>
#include <sstmac/software/libraries/compute/lib_compute_time.h>
#include <sstmac/software/api/hpx/runtime/actions/hpx_action_params.h>
#include <sstmac/software/api/hpx/runtime/hpx_action_thread.h>
#include <sstmac/software/libraries/unblockhandler.h>
#include <sstmac/software/process/blockwaiter.h>
#include <sstmac/software/api/hpx/agas/agas_server.h>
#include <map>
#include <sstmac/software/api/hpx/agas/globalid.h>
#include <boost/unordered_map.hpp>
#include <sstmac/software/api/hpx/runtime/actions/hpx_action.h>
#include <sstmac/software/api/hpx/runtime/hpx_runnable.h>
#include <sstmac/software/api/hpx/hardware/coremodel.h>

#include <sstmac/software/api/hpx/runtime/hpx_load_balancer.h>

#include <sstmac/common/stats/event_trace.h>

namespace hpx
{

  namespace components
  {
    class hpx_component;
  }

  class hpx_thread;

  class hpx_thread_manager : public sstmac::sw::service
  {

    struct action_info
    {
      std::string actname_;
      hpx_action_params::ptr params_;
      boost::intrusive_ptr<hpx::lcos::base_lco_result> prom_;
      bool on_comp_;
      globalid gid_;
    };

  public:
    typedef boost::intrusive_ptr<hpx_thread_manager> ptr;

  protected:

    static sstmac::debug dbg_;

    boost::unordered_map<std::string, actions::hpx_action::ptr> actions_;
    boost::unordered_map<std::string, hpx_action_thread::ptr> actionthreads_;
    boost::unordered_map<std::string, boost::unordered_map<long,
        hpx_action_thread::ptr> > actioncompthreads_;

    sstmac::sw::lib_compute_time::ptr libcomp_;
    //sstmac::sw::key::ptr compkey_;

    agas_server::ptr agas_;
    boost::unordered_map<unsigned long long, action_info> locres_waiting_;

    ptr selfptr_;

    hpx_load_balancer::ptr lb_;
    sstmac::sst_message::ptr lb_msg_;
    sstmac::timestamp lb_time_;
    sstmac::event::ptr lb_ev_;

    boost::unordered_map<std::string, boost::intrusive_ptr<
        hpx::components::hpx_component> > comp_factory_;
    boost::unordered_map<long, boost::intrusive_ptr<
        hpx::components::hpx_component> > components_;

    static long TRANS_ID_CNT;
    boost::unordered_map<long, hpx::lcos::base_lco_result::ptr> rem_actions_;

    hpx_thread_manager(agas_server::ptr agas);

    std::map<int, coremodel::ptr> cores_;
    int swaps_per_core_;

    std::queue<hpx_runnable::ptr> ready_;
    boost::unordered_map<hpx_runnable::ptr, int> allthreads_;
    std::map<int, hpx_runnable::ptr> allcores_;
    int busy_cores_;
    bool thread_scheduled_;

    static sstmac::event_trace::ptr stat_threadcores_;

    void
    do_action(sstmac::nodeaddress::ptr loc, std::string actname,
        hpx_action_params::ptr params,
        boost::intrusive_ptr<hpx::lcos::base_lco_result> prom, bool on_comp,
        globalid gid);

  public:

    static ptr
    construct(agas_server::ptr agas)
    {
      return ptr(new hpx_thread_manager(agas));
    }

    virtual
    ~hpx_thread_manager();

    virtual void
    kill(bool sendtoothers);

    static std::string
    lib_name()
    {
      return "hpx_thread_manager";
    }

    sstmac::timestamp
    now()
    {
      return os_->now();
    }

    void
    set_me_to_core(boost::intrusive_ptr<hpx_thread> t, bool creating);

    void
    thread_is_blocking(hpx_runnable::ptr t, sstmac::sw::key::ptr blocker);

    void
    thread_is_unblocking(hpx_runnable::ptr t, sstmac::sw::key::ptr blocker);

    void
    thread_is_finished(hpx_runnable::ptr t);

    void
    thread_is_starting(hpx_runnable::ptr t);

    void
    try_thread_q();

    virtual void
    start();

    virtual void
    finish()
    {
      if (ready_.size() > 0)
      {
        sst_throw(sstmac::ssterror, "yo homies, thread manager gots more threads ready");
      }
    }

    virtual sstmac::sw::service::ptr
    clone()
    {
      return construct(agas_);
    }

    virtual void
    handle(const sstmac::sst_message::ptr &msg);

    globalid
    create_component(std::string name);

    template<typename ActName>
      void
      register_action(std::string name, boost::intrusive_ptr<ActName> actptr)
      {
        /* if (actions_.find(name) != actions_.end())
         {
         std::cout << "Action name: " << name << "\n";
         throw sstmac::ssterror(
         "hpx_thread_manager::register_action() -- action already registered here");
         }*/

        if (actions_.find(name) == actions_.end())
        {
          actions_[name] = actptr;
        }

        // os_->register_lib(actptr);
      }

    template<typename CompName>
      void
      register_component(std::string name,
          boost::intrusive_ptr<CompName> compptr)
      {
        /*if (comp_factory_.find(name) != comp_factory_.end())
         {
         std::cout << "Component name: " << name << "\n";
         throw sstmac::ssterror(
         "hpx_thread_manager::register_component() -- component already registered here");
         }*/

        if (comp_factory_.find(name) == comp_factory_.end())
        {
          comp_factory_[name] = compptr;
        }

        compptr->register_actions(selfptr_);
      }

    bool
    is_local(sstmac::nodeaddress::ptr loc)
    {
      return loc->equals(os_->get_my_addr());
    }

    void
    start_action(std::string actname, hpx_action_params::ptr params,
        boost::intrusive_ptr<hpx::lcos::base_lco_result> prom,
        sstmac::nodeaddress::ptr locality);

    void
    start_action(std::string actname, hpx_action_params::ptr params,
        boost::intrusive_ptr<hpx::lcos::base_lco_result> prom, globalid gid);

    void
    locality_resolved(sstmac::nodeaddress::ptr loc, unsigned long long waitid);

    void
    gid_created(globalid gid, unsigned long long waitid);

  };

}

#endif
