/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_COMM_MPICOMMFACTORY_PENDING_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_COMM_MPICOMMFACTORY_PENDING_H_INCLUDED

#include <sstmac/software/api/mpi/mpi_comm/mpicommfactory.h>
#include <sstmac/software/process/key.h>
#include <sstmac/software/api/mpi/mpi_comm/mpicommdata.h>
#include <sstmac/common/nocopy.h>
#include <sstmac/common/debug.h>

namespace sstmac { namespace sw {





  ////
  ///Manage information associated with pending requests for construction of
  ////MPI communicators.
  ////
  class mpicommfactory::pending : virtual public nocopy, public ptr_type {
    /// The key associated with this request.
    key::ptr key_;

    /// The scheduler to whom the MPI tasks belong.
   //// boost::shared_ptr<scheduler> scheduler_;

  protected:
    /// The identifier for the mpi group we're building.
    mpicommid id_;

    /// The group we are using to build from.
    mpicomm::ptr template_;

    /// Constructor.
    pending(const mpicommid &new_id////, const boost::shared_ptr<scheduler>&
        );

    /// Block current caller.
    void block();

    /// Unblock all callers.
    void unblock();

  public:
    virtual std::string to_string() const{
            return "pending";
          }

    typedef boost::intrusive_ptr<pending> ptr;
    /// Goodbye.
    virtual ~pending() {}
  };

 ////
  ////Supervise calls corresponding to MPI_Comm_create.
 ////
  class mpicommfactory::pending_create : public mpicommfactory::pending {
    /// The tasks (jobs) to be included in this group.
    std::vector<taskid> target_peers_;

    /// The nodes involved in the group.
   //// std::vector<nodeid> target_nodes_;
    
    /// Mapping from node index in the input group to node indices
    /// in the newly created group.
    /// Set to -1 for all nodes that are not members of the new group.
    std::vector<mpiid> index_mapping_;

    /// The number of processes that have entered the add(...) method.
    size_t added_;

    /// The shared mpicommdata.  Gets constructed when everybody
    /// has called add.
    mpicommdata::ptr commdata_;

  protected:
    /// Constructor.
    /// \param new_id is the identifier for the new group.
    pending_create(const mpicommid &new_id////, const boost::shared_ptr<scheduler>& sched
        );

  public:
    typedef boost::intrusive_ptr<pending_create> ptr;

    /// Goodbye.
    virtual ~pending_create() {}

    /// Construct a new object of this type.
    static pending_create::ptr
    construct(const mpicommid &new_id////, const boost::shared_ptr<scheduler>& sched
        );

    /// Get a new communicator for the target group.
    /// Blocks the caller until the group is complete.
    /// All members in the old group must make this call; nodes that
    /// are not in the new group will get mpicomm::null_comm.
    /// \throw valueerror if the members don't match earlier requests.
    mpicomm::ptr build(const mpicomm::ptr &caller, const std::vector<mpiid> &members);
  };

 ////
  /////Supervise calls corresponding to MPI_Comm_split.
 /////
  class mpicommfactory::pending_split : public mpicommfactory::pending {
    /// Nested type to manage the key, taskid, and nodeid for each color.
    class keylist : public ptr_type
    {
      /// This is how we store the location of each rank.
      struct location
      {
        location() :
          the_task(-1),
          ////the_node(-1),
              final_rank(-1)
        {
        }
        location(const taskid &tid////, const nodeid &nid
        ) :
          the_task(tid)////, the_node(nid)
        {
        }
        taskid the_task;
        ////nodeid the_node;
        /// This rank gets populated on first call to get_mpicomm,
        /// after which the list is locked.
        mpiid final_rank;
      };

      /// The nodes (keys) encountered so far.
      /// The keylist is complete whenever the ranks are densely populated
      /// from zero to size-1.
      typedef std::map<int, location> rankmap_t;
      rankmap_t ranks_;

      /// The communicator id.
      mpicommid id_;

      /// When get_mpicomm::ptr is called the first time, this shared state
      /// gets constructed.  After that, no more nodes can be inserted.
      mpicommdata::ptr shared_;

    protected:
      /// Hello.
      keylist(const mpicommid &id);

    public:
      typedef boost::intrusive_ptr<mpicommfactory::pending_split::keylist> ptr;

      virtual std::string to_string() const{
              return "keylist";
            }

      /// Goodbye.
      virtual
      ~keylist();

      /// Construct a new keylist.
      static keylist::ptr
      construct(const mpicommid &id);

      /// Return true if no ranks have been added yet.
      bool
      empty() const;

      /// Return true if shared state has been constructed (no more insert
      /// calls can be made).
      bool
      locked() const;

      /// Test whether the given rank is already defined in the keylist.
      /// Returns true for negative keys (since they correspond to comm_null).
      bool
      contains(int key) const;

      /// Insert a new rank in the keylist.
      /// \throw illformederror if(this->locked() == true).
      /// \throw valueerror if(key >= 0 && contains(key) == true).
      void
      insert(int key, const taskid &tid, nodeaddress::ptr &nid);

      /// Get a communicator corresponding to the given key.
      /// Locks the keylist (locked() will return false after this call).
      /// \throw uninitializederror if(key >= 0 && contains(key) == false).
      /// \throw valueerror if the key is not defined in the keylist.
      mpicomm::ptr
      get_mpicomm(int key);
    };

    /// This is where the mechanics of aggregating entries by color
    /// and building the communicators takes place.
    std::map<int, boost::intrusive_ptr<keylist> > colors_;

    /// Nodes still pending.
    int pending_;

    /// The maximum number of communicators we're allowed to generate.
    int maxsize_;

    /// The identifier for the next communicator constructed.
    mpicommid current_id_;

  protected:
    /// Constructor.
    pending_split(const mpicommid &new_id, int maxsize////,
                  ////const boost::shared_ptr<scheduler>& sched
                  );

    debug dbg_;
  public:
    typedef boost::intrusive_ptr<pending_split> ptr;

    /// Goodbye.
    virtual ~pending_split() {}
    
    /// Hello.
    static pending_split::ptr
    construct(const mpicommid &new_id, int maxsize////,
              ////const boost::shared_ptr<scheduler>& sched
              );

    /// Add a participant.  If the call is not complete yet, the caller
    /// gets context-switched out pending completion of the call.
    mpicomm::ptr build(const mpicomm::ptr &caller, int color, int key);
  };


} } // end of namespace sstmac

#endif
