/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/api/mpi/mpi_comm/mpicommfactory.h>
#include <sstmac/software/api/mpi/mpi_comm/mpicommfactory-pending.h>
#include <sstmac/software/ami/environment.h>
#include <sstmac/software/api/mpi/mpiapi.h>
#include <sstmac/software/api/mpi/mpi_comm/mpicomm.h>
#include <sstmac/software/api/mpi/mpi_comm/mpicomm_cart.h>
#include <sstmac/common/counter.h>
#include <sstmac/common/richpair.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>
//#include <sstmac/common/messages/valuepayload.h>
#include <sstmac/software/api/mpi/mpipayload.h>
#include <iterator>
#include <sstmac/software/api/mpi/sstmac_mpi.h>
#include <sstmac/software/api/mpi/mpitypes.h>
namespace sstmac
{
  namespace sw
  {

    using boost::dynamic_pointer_cast;

    debug mpicommfactory::dbg_;

    std::map<std::pair<appid, mpicommid>, tasklist::ptr>
        mpicommfactory::tlists_;
    // mpicommid mpicommfactory::next_id_(1);

    //
    // Build comm_world using information retrieved from the environment.
    //
    mpicommfactory::mpicommfactory(appid aid, mpiapi::ptr parent) :
      parent_(parent), aid_(aid), mpirun_np_(0), next_id_(1)
    {
      if (!dbg_.is_real())
      {
        dbg_ = debug("<debug> mpi | mpicommfactory", logger::yellow);
        logger::register_logger(&dbg_);
      }

      //next_id_ = mpicommid(1);
    }

    //
    // Build a new factory.
    //
    mpicommfactory::ptr
    mpicommfactory::construct(appid aid, mpiapi::ptr parent)
    {
      mpicommfactory::ptr retval(new mpicommfactory(aid, parent));
      return retval;
    }

    //
    // Goodbye.
    //
    mpicommfactory::~mpicommfactory()
    {
    }

    //
    // Initialize.
    //
    void
    mpicommfactory::init(environment::const_ptr& env, const mpiid &rank)
    {

      //mpirun_np_ = mpirun_np;
      next_id_.id = 1;

      mpirun_np_ = env->get_app_count();

      SSTMAC_DEBUG << "MPI comm factory[" << rank.id << "]: app count is "
            << mpirun_np_ << "\n";

      const mpicommid cid(0);
      std::pair<appid, mpicommid> index = std::make_pair<appid, mpicommid>(
          aid_, cid);

      /*  if (tlists_.find(index) == tlists_.end())
       {
       std::vector<taskid> dumm;

       for (int i = 0; i < mpirun_np_; i++)
       {
       dumm.push_back(taskid(i));
       }

       tasklist::ptr tl = tasklist::construct(dumm);

       tlists_[index] = tl;
       }*/

      mpigroup::ptr g = mpigroup::construct(mpirun_np_);

      worldcomm_ = mpicomm::construct(mpicommid(MPI_COMM_WORLD), rank, g, env,
          aid_);

      std::vector<taskid> selfp;
      selfp.push_back(taskid(rank.id));

      tasklist::ptr sl = tasklist::construct(selfp);
      mpigroup::ptr g2 = mpigroup::construct(sl);
      selfcomm_ = mpicomm::construct(mpicommid(MPI_COMM_SELF), mpiid(0), g2,
          env, aid_);

      splittype_ = mpitype("split_type", mpitype::mpi_int, 3, 1, 1, true, MPI_COMBINER_CONTIGUOUS);
    }

    void
    mpicommfactory::finalize()
    {
      parent_ = 0;
      worldcomm_ = 0;
      selfcomm_ = 0;
    }

    //
    // Get the world communicator.
    //
    const mpicomm::ptr
    mpicommfactory::get_world()
    {
      return worldcomm_;
    }

    //
    // Get the self communicator.
    //
    const mpicomm::ptr
    mpicommfactory::get_self()
    {
      return selfcomm_;
    }

    //
    // Duplicate a communicator.
    //
    mpicomm::ptr
    mpicommfactory::comm_dup(const mpicomm::ptr &caller)
    {
      //std::cerr << "mpicommfactory::comm_dup(" << caller << ").  Input comm id=" << caller->id() << ", size=" << caller->size() << "\n";
      //   std::vector<mpiid> participants;
      //  participants.reserve(caller->size().id);
      //  std::generate_n(std::back_inserter(participants), caller->size().id,
      //      counter<mpiid> (mpiid(0)));
      mpicomm::ptr ret = this->comm_create(caller, caller->group_);

      ret->dup_keyvals(caller);
      return ret;
    }

    //
    // Make the given mpiid refer to a newly created communicator.
    //
    mpicomm::ptr
    mpicommfactory::comm_create(const mpicomm::ptr &caller, mpigroup::ptr group)
    {

      //first reduce so we know everyone is on board, which is the necessary barrier
      //payload::const_ptr load = valuepayload<int>::construct(next_id_.id);
      payload::const_ptr load = mpipayload::construct(&next_id_.id,
          mpitype::mpi_int, 1, false);
      payload::const_ptr result;
      parent_->allreduce(1, mpitype::mpi_int, mpiop::max, caller, load, result,
          false);

      mpipayload::const_ptr casted = boost::dynamic_pointer_cast<
          const mpipayload>(result);
      int returndata;
      casted->extract(&returndata, 1, mpitype::mpi_int);

      SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_->rank().id
            << "]: we all decided on id: " << returndata << "\n";

      mpicommid cid = mpicommid(returndata);

      std::pair<appid, mpicommid> index = std::make_pair<appid, mpicommid>(
          aid_, cid);

      /* if (tlists_.find(index) == tlists_.end())
       {
       // std::vector<taskid> dumm;

       SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_->rank().id
       << "]: creating comm " << cid.id << " of size "
       << group->size() << "\n";

       //  for (int i = 0; i < participants.size(); i++)
       //    {
       //      SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_->rank().id
       //            << "]: adding " << participants[i].id << " to the mix \n";
       //      dumm.push_back(caller->peer_task(participants[i]));
       //    }

       tlists_[index] = group->list();
       }
       else
       {
       //if(!tlists_[index]->equals(participants)){
       //    return mpicomm::comm_null;  // this is erroneous behavior, if the participants don't match for different tasks
       // }
       }*/

      //now find my rank
      mpiid newrank = group->rank_of_task(caller->my_task());

      // if(newrank.id == 0){
      next_id_.id = cid.id + 1;
      SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_->rank().id
            << "]: incrementing next_id to " << next_id_.id << "\n";
      // }
      SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_->rank().id
            << "]: my new rank is " << newrank.id << "\n";
      if (newrank.id >= 0)
        return mpicomm::construct(cid, newrank, group, worldcomm_->env_, aid_);
      else
        return mpicomm::comm_null;
    }

    //
    // MPI_Comm_split.
    //
    mpicomm::ptr
    mpicommfactory::comm_split(const mpicomm::ptr &caller, int color, int key)
    {
      SSTMAC_DEBUG << "mpicommfactory[" << worldcomm_->rank().id
            << "]::comm_split(" << caller->to_string() << ", " << color << ", "
            << key << ")" << "\n";

      int mydata[3];
      mydata[0] = next_id_.id;
      mydata[1] = color;
      mydata[2] = key;

      payload::const_ptr mine = mpipayload::construct(mydata, splittype_, 1,
          false);
      //   valuepayload<richpair<int, richpair<int, int> > >::construct(
      //      std::make_pair<int, richpair<int, int> >(next_id_.id,
      //         std::make_pair<int, int>(color, key)));

      std::vector<payload::const_ptr> result;

      int mynextcommid = -1;
      int mynewrank = -1;

      parent_->allgather(1, splittype_, caller->size().id, splittype_, caller,
          mine, result, false);

      if (color >= 0)
      {

        mpicommid cid = mpicommid(-1);
        std::map<int, std::multimap<int, int> > colormap;
        for (uint i = 0; i < result.size(); i++)
        {
          int thisdata[3];
          mpipayload::const_ptr thisload = boost::dynamic_pointer_cast<const
              mpipayload>(result[i]);
          thisload->extract(thisdata, 1, splittype_);
          //  valuepayload<richpair<int, richpair<int, int> > >::const_ptr casted =
          //     boost::dynamic_pointer_cast<const valuepayload<richpair<int,
          //         richpair<int, int> > > >(result.at(i));


          SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_->rank().id
                << "]: split: creating color map: " << thisdata[1] << ", "
                << thisdata[2] << "\n";

          colormap[thisdata[1]].insert(std::pair<int, int>(thisdata[2], i));

          if (thisdata[0] > cid.id)
          {
            cid = mpicommid(thisdata[0]);
          }
        }

        std::map<int, std::multimap<int, int> >::iterator it, end =
            colormap.end();

        SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_->rank().id
              << "]: ended up with cid " << cid.id << "\n";

        //mpicommid nxt = next_id_.id;
        for (it = colormap.begin(); it != end; it++)
        {
          if (it->first >= 0)
          {
            std::multimap<int, int> keys = it->second;

            //redistribute it to everyone
            //parent_->bcast(1, mpitype::mpi_int, mpiid(0), caller, result);
            SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_->rank().id
                  << "]: split: considering : " << cid.id << "\n";

            // mpicommid cid = mpicommid(next_id_.id);

            if (it->first == color)
            {
              mynextcommid = cid.id;
            }
            std::pair<appid, mpicommid> index =
                std::make_pair<appid, mpicommid>(aid_, cid);
            if (tlists_.find(index) == tlists_.end())
            {
              std::vector<taskid> dumm;
              SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_->rank().id
                    << "]: split: creating comm " << cid.id << ": ";

              int num = 0;
              std::multimap<int, int>::iterator it2, lookahead, end2 =
                  keys.end();
              for (it2 = keys.begin(); it2 != end2;)
              {
                lookahead = it2;
                lookahead++;

                std::map<int, int> keysort;
                keysort[it2->second] = it2->second;
                while (lookahead != end2 && lookahead->first == it2->first)
                {
                  keysort[lookahead->second] = lookahead->second;
                  lookahead++;
                }

                std::map<int, int>::iterator it3, end3 = keysort.end();
                for (it3 = keysort.begin(); it3 != end3; it3++)
                {
                  SSTMAC_DEBUG << caller->peer_task(mpiid(it3->second)) << ", ";
                  dumm.push_back(caller->peer_task(mpiid(it3->second)));
                }
                it2 = lookahead;
              }

              SSTMAC_DEBUG << "\n comm is of size: " << dumm.size() << "\n";

              tlists_[index] = tasklist::construct(dumm);
            }

            // next_id_.id = next_id_.id + 1;
            //nxt++;
            cid.id = cid.id + 1;
          }

        }

        next_id_.id = cid.id;

        std::pair<appid, mpicommid> index = std::make_pair<appid, mpicommid>(
            aid_, mpicommid(mynextcommid));
        if (tlists_.find(index) == tlists_.end())
        {
          throw rangeerror("mpicommfactory: did not find task list");
        }
        tasklist::ptr tlist = tlists_[index];

        for (uint i = 0; i < tlist->size(); i++)
        {
          if (tlist->at(i) == caller->my_task())
          {
            mynewrank = i;
          }
        }

        if (mynewrank < 0)
        {
          std::cerr << "my task: " << caller->my_task().to_string() << "\n";
          std::cerr << "task list for index: " << index.first.id << ", "
              << index.second.id << "\n";
          for (uint i = 0; i < tlist->size(); i++)
          {
            std::cerr << "task: " << tlist->at(i).to_string() << "\n";
          }
          throw ssterror(
              "mpicommfactory: didn't find task in list, can't assign rank");
        }

        //parent_->barrier(caller);

        //if (color >= 0)
        return mpicomm::construct(index.second, mpiid(mynewrank),
            mpigroup::construct(tlist), worldcomm_->env_, aid_);
      }
      else
        return mpicomm::comm_null;

      /* valuepayload<richpair<int, int> >::const_ptr load = valuepayload<richpair<int, int> >::construct(std::make_pair<int, int>(color, key));
       std::vector<payload::const_ptr> result;

       int mynextcommid = -1;
       int mynewrank = -1;

       parent_->gather(1, mpitype::mpi_int, caller->size().id, mpitype::mpi_int, mpiid(0), caller, load, result);

       mpitag colltag = caller->next_collective_tag();

       if(caller->rank().id == 0){
       std::map<int, vector<int> > colormap;
       for(int i = 0; i < result.size(); i++){
       valuepayload<richpair<int, int> >::const_ptr casted = boost::dynamic_pointer_cast<const valuepayload<richpair<int, int> > >(result.at(i));


       pair<int, int> p = casted->data();

       SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_->rank().id << "]: split: creating color map: " << p.first << ", " << p.second << "\n";

       if(colormap[p.first].size() == 0){
       colormap[p.first].reserve(caller->size().id);
       colormap[p.first].resize(caller->size().id, -1);
       }
       colormap[p.first][i] = p.second;
       }

       std::map<int, vector<int> >::iterator it, end = colormap.end();


       std::vector<mpirequest_t> reqs;
       for(it = colormap.begin(); it != end; it++){
       vector<int> keys = it->second;

       SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_->rank().id << "]: split: considering : " << next_id_.id << "\n";

       mpicommid cid = mpicommid(next_id_.id);

       if(it->first == color){
       mynextcommid = next_id_.id;
       }
       pair<appid, mpicommid> index = make_pair<appid, mpicommid>(aid_, cid);
       if(tlists_.find(index) == tlists_.end()){
       std::vector<taskid> dumm;

       for(int i = 0; i < keys.size(); i++){
       if(keys[i] >= 0)
       dumm.push_back(caller->peer_task(mpiid(keys[i])));
       }

       SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_->rank().id << "]: split: creating comm " << cid.id << " of size " << dumm.size() << "\n";

       tlists_[index] = tasklist::construct(dumm);
       }

       payload::const_ptr nextid = valuepayload<int>::construct(next_id_.id);
       for(int i = 0; i < keys.size(); i++){
       if(keys[i] >= 0){
       mpirequest_t req;
       parent_->isend(1, mpitype::mpi_int, mpiid(i), colltag, caller, req, nextid);
       reqs.push_back(req);
       }
       }

       next_id_.id++;

       }

       std::vector<const_mpistatus_t> stats;
       parent_->waitall(reqs, stats);

       }else{
       const_mpistatus_t stat;

       SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_->rank().id << "]: split: at old rank " << caller->rank().id  << "..receiving new comm id \n";

       parent_->recv(1, mpitype::mpi_int, mpiid(0), colltag, caller, stat);

       valuepayload<int>::const_ptr casted = boost::dynamic_pointer_cast<const valuepayload<int> >(stat->content());

       SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_->rank().id << "]: split: at old rank " << caller->rank().id << ", I received mpicommid of " << casted->data() << "\n";
       mynextcommid = casted->data();
       }

       pair<appid, mpicommid> index = make_pair<appid, mpicommid>(aid_, mpicommid(mynextcommid));
       if(tlists_.find(index) == tlists_.end()){
       throw rangeerror("mpicommfactory: did not find task list");
       }
       tasklist::ptr tlist = tlists_[index];

       for(int i = 0;i < tlist->size(); i++){
       if(tlist->at(i) == caller->my_task()){
       mynewrank = i;
       }
       }

       if(color >= 0)
       return mpicomm(index.second, mpiid(mynewrank), tlist, worldcomm_->env_);
       else
       return mpicomm::comm_null;*/

    }

    mpicomm::ptr
    mpicommfactory::create_cart(const mpicomm::ptr &caller, int ndims,
        int *dims, int *periods, int reorder)
    {
      // std::vector<mpiid> participants;
      //  participants.reserve(caller->size().id);
      //  std::generate_n(std::back_inserter(participants), caller->size().id,
      //      counter<mpiid> (mpiid(0)));

      //first reduce so we know everyone is on board, which is the necessary barrier
      payload::const_ptr load = mpipayload::construct(&next_id_.id,
          mpitype::mpi_int, 1, false);
      payload::const_ptr result;
      parent_->allreduce(1, mpitype::mpi_int, mpiop::max, caller, load, result,
          false);

      mpipayload::const_ptr casted = boost::dynamic_pointer_cast<
          const mpipayload>(result);
      int returndata;
      casted->extract(&returndata, 1, mpitype::mpi_int);

      SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_->rank().id
            << "]: we all decided on id: " << returndata << "\n";

      mpicommid cid = mpicommid(returndata);

      std::pair<appid, mpicommid> index = std::make_pair<appid, mpicommid>(
          aid_, cid);

      /* if (tlists_.find(index) == tlists_.end())
       {
       std::vector<taskid> dumm;

       SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_->rank().id
       << "]: creating comm " << cid.id << " of size "
       << participants.size() << "\n";

       for (int i = 0; i < participants.size(); i++)
       {
       SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_->rank().id
       << "]: adding " << participants[i].id << " to the mix \n";
       dumm.push_back(caller->peer_task(participants[i]));
       }

       tlists_[index] = tasklist::construct(dumm);
       }
       else
       {
       //if(!tlists_[index]->equals(participants)){
       //    return mpicomm::comm_null;  // this is erroneous behavior, if the participants don't match for different tasks
       // }
       }*/

      //now find my rank
      mpiid newrank = caller->group_->rank_of_task(caller->my_task());

      // if(newrank.id == 0){
      next_id_.id = cid.id + 1;
      SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_->rank().id
            << "]: incrementing next_id to " << next_id_.id << "\n";
      // }
      SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_->rank().id
            << "]: my new rank is " << newrank.id << "\n";
      if (newrank.id >= 0)
        return mpicomm_cart::construct(cid, newrank, caller->group_,
            worldcomm_->env_, aid_, ndims, dims, periods, reorder);
      else
        return mpicomm::comm_null;
    }

  }
} // end of namespace sstmac
