#include <sstmac/software/api/mpi/mpi_protocol/mpiprotocol.h>
#include <sstmac/software/api/mpi/mpi_queue/mpiqueue-sendrequest.h>
#include <sstmac/software/api/mpi/mpiserver.h>

namespace sstmac
{
  namespace sw
  {

    bool
    eager1_rdma::send_needs_completion_ack() const
    {
      return false;
    }

    bool
    eager1_rdma::send_needs_rendezvous_ack() const
    {
      return false;
    }

    bool
    eager1_rdma::send_needs_nic_ack() const
    {
      return false;
    }

    bool
    eager1_rdma::send_needs_eager_ack() const
    {
      return true;
    }

    void
    eager1_rdma::send_header(const mpiqueue::ptr &queue,
        const mpimessage::ptr &msg)
    {
        SSTMACBacktrace("MPI Eager 1 Protocol: Send RDMA Header");
        SSTMAC_DEBUG << "mpiqueue[" << queue->id_string()
            << "]:: send_eager1_rdma_header done \n";

      queue->buffered_send(msg);

      /** the send will have copied into a temp buffer so we can 'ack' the buffer for now */
      mpiqueue::send_needs_ack_t::iterator it, end =
          queue->send_needs_eager_ack_.end();
      for (it = queue->send_needs_eager_ack_.begin(); it != end; ++it)
      {
        //horrible hack for now
        if ((*it)->matches(msg))
        {
          SSTMAC_DEBUG << "eager1 nic ack match found \n";
          // Match.
          mpiqueue::sendrequest::ptr sreq = *it;
          sreq->complete(msg);
          /** we can complete the send request and allow
           the program to continue, but we have to wait
           for buffer resources to be completed by the rdma get */
          sreq->wait_for_buffer();
          return;
        }
      }
      sst_throw(illformederror, "eager1 protocol could not find request to complete");
    }

    void
    eager1_rdma::handle_header(const mpiqueue::ptr &queue,
        const mpimessage::ptr &msg)
    {
      SSTMACBacktrace("MPI Eager 1 Protocol: Handle RDMA Header");
      if (msg->cat() == mpimessage::onesided)
      {
      }
      else
      {
        mpiqueue::recvrequest::ptr req =
            queue->find_pending_request(msg, false);
        if (req)
        { //we can post an RDMA get request direct to the buffer
          SSTMAC_DEBUG << "mpiqueue[" << queue->id_string()
                << "]:: eager1 got lucky and found matching request\n";
          //make sure to put the request back in, but alert it
          //that it should expect a data payload next time
          queue->waiting_message_.push_front(req);
          req->set_seqnum(msg->seqnum()); //associate the messages
        }
        else
        {
          SSTMAC_DEBUG << "mpiqueue[" << queue->id_string()
                << "]:: eager1 is unlucky and did not find matching request\n";
          msg->set_protocol(mpiprotocol::eager1_rdma_doublecpy_protocol);
        }
        queue->notify_probes(msg);

        /** we need to send an RDMA GET request back to the original node */
        msg->mutate(sst_message::GNI_HEADER_TO_RDMA_GET_REQ);
        // this has already been received by mpi in sequence
        // make sure mpi still handles this since it won't match
        // the current sequence number
        msg->set_ignore_seqnum(true);

        queue->post_rdma(msg);

      }
      SSTMAC_DEBUG << "mpiqueue[" << queue->id_string()
            << "]:: handle_eager1_rdma_header done \n";

    }

    void
    eager1_rdma::recv_header(const mpiqueue::ptr &queue,
        const mpimessage::ptr &msg, const mpiqueue::recvrequest::ptr& req)
    {
      sst_throw(illformederror, "eager1_rdma protocol should not have recv-request handle header");
    }

    void
    eager1_rdma::recv_payload(const mpiqueue::ptr &queue,
        const mpimessage::ptr &msg, const mpiqueue::recvrequest::ptr& req)
    {
    }

    void
    eager1_rdma_singlecpy::handle_payload(const mpiqueue::ptr &queue,
        const mpimessage::ptr &msg)
    {
      SSTMACBacktrace("MPI Eager 1 Protocol: Handle RDMA Paylod");
      SSTMAC_DEBUG << "mpiqueue[" << queue->id_string()
            << "]: incoming eager1 singlecpy payload\n";

      if (msg->cat() == mpimessage::onesided)
      {
        SSTMAC_DEBUG << "mpiqueue[" << queue->id_string()
                   << "]: handling onesided data\n";
        queue->handle_put(msg);
      }
      else
      {
        mpiqueue::recvrequest::ptr req = queue->find_waiting_request(msg);
        if (!req)
          throw illformederror(
              "running singlecpy eager1 protocol, but no matching request found");
        //We have RDMA GOT direct into the buffer - just complete
        req->handle(msg);
      }
    }

    void
    eager1_rdma_doublecpy::handle_payload(const mpiqueue::ptr &queue,
        const mpimessage::ptr &msg)
    {
      SSTMACBacktrace("MPI Eager 1 Protocol: Handle RDMA Paylod");
      SSTMAC_DEBUG << "mpiqueue[" << queue->id_string()
            << "]: incoming eager1 doublecpy payload\n";
      //the recv request is expecting a user message - update the message to match
      mpiqueue::recvrequest::ptr req = queue->find_pending_request(msg, true); //if not found, add to need_recv
      if (req) //just finish things off by copying buffers
      {
        SSTMAC_DEBUG << "mpiqueue[" << queue->id_string()
              << "]: found matching eager1 doublecpy recv\n";
        queue->buffered_recv(msg, req);
      }
    }

    bool
    eager1_rdma::handshake_only() const
    {
      //initial send is handshake only
      return true;
    }

  }
}
