#include <sstmac/software/api/mpi/mpi_queue/servicethread-mpiqueue.h>
#include <sstmac/software/api/mpi/mpiserver.h>
#include <sstmac/software/api/mpi/mpiapi.h>
#include <sstmac/software/api/mpi/mpi_implementation/mpi_implementation.h>
#include <sstmac/software/api/mpi/mpi_protocol/mpiprotocol.h>
#include <sstmac/software/process/operatingsystem.h>

namespace sstmac
{
  namespace sw
  {
    SSTRegister("service", mpiqueue, service_thread_mpiqueue,
                "Run MPI as a separate service thread distinct from the application.");

    service_thread_mpiqueue::ptr
    service_thread_mpiqueue::construct(const sim_parameters::ptr& params)
    {
        return new service_thread_mpiqueue;
    }

    service_thread_mpiqueue::service_thread_mpiqueue()
        : mpiqueue()
    {

    }

    service_thread_mpiqueue::~service_thread_mpiqueue() throw()
    {
    }

    timestamp
    service_thread_mpiqueue::progress_loop(const mpirequest_t& req)
    {
        //only block if the machine model supports nic send acks
        if (!req->is_complete())
          return os_->block(req->get_key());

        return os_->now();
    }

    void
    service_thread_mpiqueue::start_progress_loop(const std::vector<mpirequest_t>& req)
    {
        for (uint i = 0; i < req.size(); i++){
            req[i]->get_key()->add_blocker(os_->get_current_context());
        }

        key::ptr thekey = key::construct(mpiapi::default_key_category);
        os_->block(thekey);
    }

    void
    service_thread_mpiqueue::start_progress_loop(const std::vector<mpirequest_t>& req, timestamp timeout)
    {
        for (uint i = 0; i < req.size(); i++){
            req[i]->get_key()->add_blocker(os_->get_current_context());
        }

        timestamp done = os_->now() + timeout;

        key::ptr thekey = key::construct(mpiapi::default_key_category);
        event::ptr awake = new_event(ptr(this), &service_thread_mpiqueue::check_timeout, done, thekey);
        server_->schedule(done, awake);
        os_->block(thekey);
    }

    void
    service_thread_mpiqueue::finish_progress_loop(const std::vector<mpirequest_t>& req)
    {
        //cleanup
        for (uint i = 0; i < req.size(); i++){
            if (req[i]){
                req[i]->get_key()->remove_blocker(os_->get_current_context());
            }
        }
    }

    void
    service_thread_mpiqueue::check_timeout(const timestamp &t, const key::ptr &thekey)
    {
        if (os_->now() >= t){
            //we have timed out
            os_->unblock(thekey);
        }
        else {
            //do nothing
        }
    }

    void
    service_thread_mpiqueue::buffer_unexpected(const mpimessage::ptr &msg)
    {
    }

    void
    service_thread_mpiqueue::incoming_message(const mpimessage::ptr &message)
    {
        mpiqueue::incoming_message(message);
    }

    eventhandler::ptr
    service_thread_mpiqueue::progress_done_handler(const OS_API::ptr &os, const mpirequest_t &req)
    {
        return unblockhandler::construct(os, req->get_key());
    }

    void
    service_thread_mpiqueue::post_header(const mpimessage::ptr &msg)
    {
        //I have no way to model delay ... just send it
        server_->send(msg);
    }

    void
    service_thread_mpiqueue::post_rdma(const mpimessage::ptr &msg)
    {
        server_->send(msg);
    }

    void
    service_thread_mpiqueue::do_send(const mpimessage::ptr &mess)
    {
      event::ptr ev = new_event(mess->get_protocol(), &mpiprotocol::send_header, ptr(this), mess);
      server_->schedule_now(ev);
    }

    void
    service_thread_mpiqueue::do_recv(const mpiqueue::recvrequest::ptr &req)
    {
      event::ptr ev = new_event(ptr(this), &mpiqueue::start_recv, req);
      server_->schedule_now(ev);
    }

    void
    service_thread_mpiqueue::buffered_recv(const mpimessage::ptr &msg, const mpiqueue::recvrequest::ptr &req)
    {
        SSTMACBacktrace("MPI_buffered_recv");
        server_->buffered_recv(msg, req);
    }

    void
    service_thread_mpiqueue::buffered_send(const mpimessage::ptr &msg)
    {
        SSTMACBacktrace("MPI_buffered_send");
        server_->buffered_send(msg);
    }

  }
} // end of namespace sstmac

