#include <sstmac/software/api/mpi/mpi_queue/userthread-mpiqueue.h>
#include <sstmac/software/api/mpi/mpiserver.h>
#include <sstmac/software/api/mpi/mpiapi.h>
#include <sstmac/software/api/mpi/mpi_implementation/mpi_implementation.h>
#include <sstmac/software/api/mpi/mpi_protocol/mpiprotocol.h>
#include <sstmac/software/process/operatingsystem.h>
#include <sstmac/common/sstmac_env.h>

namespace sstmac 
{
    namespace sw
    {

    SSTRegister("user", mpiqueue, user_thread_mpiqueue,
                "Run MPI as a library within the application.");

    //
    // Hello.
    //
    user_thread_mpiqueue::ptr
    user_thread_mpiqueue::construct(const sim_parameters::ptr& params)
    {
        /** sstkeyword { gui=1us; docstring=Time it takes at software level to post RDMA transaction.; } */
        timestamp post_rdma_delay = sstmac_env::params->get_time_param("mpiqueue_post_rdma_delay");
        /** sstkeyword { gui=0.5us;
            docstring=Time it takes at software level to post a short header using
            low-latency short-message mailboxes.;
        } */
        timestamp post_header_delay = sstmac_env::params->get_time_param("mpiqueue_post_header_delay");
        /** sstkeyword {
            docstring=Time it takes at software level to poll for an incoming message.;
        } */
        timestamp poll_delay = sstmac_env::params->get_optional_time_param("mpiqueue_poll_delay", timestamp(0));
        return new user_thread_mpiqueue(post_rdma_delay,
                                        post_header_delay,
                                        poll_delay);
    }

    user_thread_mpiqueue::user_thread_mpiqueue(
            timestamp post_rdma_delay,
            timestamp post_header_delay,
            timestamp poll_delay)
        : mpiqueue(),
          is_polling_(false),
          blocking_key_(0),
          incoming_msg_(0),
          post_rdma_delay_(post_rdma_delay),
          post_header_delay_(post_header_delay),
          poll_delay_(poll_delay)
    {
    }

    user_thread_mpiqueue::~user_thread_mpiqueue() throw()
    {
    }

    void user_thread_mpiqueue::clear_pending()
    {
        if (pending_msgs_.empty())
            return;

        SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
                     << "]::clear pending messages\n";

        std::list<mpimessage::ptr>::iterator it;
        for (it=pending_msgs_.begin(); it != pending_msgs_.end(); ++it){
            SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
                         << "]::handle pending message "
                         << (*it)->to_string()
                         << "\n";
            if (poll_delay_.ticks_int64())
                user_lib_time_->compute(poll_delay_);
            mpiqueue::incoming_message(*it);
        }
        pending_msgs_.clear();
    }

    timestamp
    user_thread_mpiqueue::progress_loop(const mpirequest_t& req)
    {
        clear_pending();

        if (req->is_complete())
            return os_->now();

        SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
                     << "]::entering progress loop\n";

        SSTMACBacktrace("MPI Queue Poll");
        blocking_key_ = req->get_key();
        while (!req->is_complete()){
            SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
                         << "]::blocking in progress loop on key "
                         << (long) blocking_key_.get()
                         << " on thread " << os_->get_current_threadid()
                         << "\n";
            is_polling_ = true;
            os_->block(blocking_key_);
            is_polling_ = false;
            SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
                         << "]::continuing progress loop on "
                         << " incoming msg " << incoming_msg_->to_string()
                         << "\n";
            if (poll_delay_.ticks_int64())
                user_lib_time_->compute(poll_delay_);
            mpiqueue::incoming_message(incoming_msg_);
            clear_pending();
        }
        SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
                     << "]::finishing progress loop on key "
                     << (long) blocking_key_.get() << "\n";

        blocking_key_ = 0;
        incoming_msg_ = 0;

        return os_->now();
    }

    bool
    user_thread_mpiqueue::at_least_one_complete(const std::vector<mpirequest_t>& req)
    {
        SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
                     << "]: checking if any of "
                     << req.size() << "requests is done\n";
        for (int i=0; i < (int) req.size(); ++i){
            if (req[i] && req[i]->is_complete()){
                SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
                             << "]: request " << i
                             << " is done\n";
                req[i]->get_key()->remove_blocker(os_->get_current_context());
                return true;
            }
        }
        return false;
    }

    eventhandler::ptr
    user_thread_mpiqueue::progress_done_handler(const OS_API::ptr &os, const mpirequest_t &req)
    {
        return donothinghandler::construct();
    }

    void
    user_thread_mpiqueue::start_progress_loop(const std::vector<mpirequest_t>& req)
    {
        clear_pending();

        SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
                     << "]::starting progress loop\n";

        for (uint i = 0; i < req.size(); i++){
            req[i]->get_key()->add_blocker(os_->get_current_context());
        }

        blocking_key_ = key::construct(mpiapi::default_key_category);
        while (!at_least_one_complete(req)){
            SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
                         << "]::blocking in progress loop on key "
                         << blocking_key_.get() << "\n";
            is_polling_ = true;
            os_->block(blocking_key_);
            SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
                         << "]::continuing progress loop on "
                         << " incoming msg " << incoming_msg_->to_string()
                         << "\n";
            is_polling_ = false;
            clear_pending();
            mpiqueue::incoming_message(incoming_msg_);
        }
        SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
                     << "]::finishing progress loop on key "
                     << (long) blocking_key_.get() << "\n";
        blocking_key_ = 0;
        incoming_msg_ = 0;
    }

    void
    user_thread_mpiqueue::start_progress_loop(const std::vector<mpirequest_t>& req, timestamp timeout)
    {
        sst_throw(unimplementederror, "user_thread_mpiqueue::start_progress_loop on vector of requeusts");
    }

    void
    user_thread_mpiqueue::finish_progress_loop(const std::vector<mpirequest_t>& req)
    {
        for (uint i = 0; i < req.size(); i++){
            if (req[i]){
                req[i]->get_key()->remove_blocker(os_->get_current_context());
            }
        }
    }

    void
    user_thread_mpiqueue::do_send(const mpimessage::ptr &mess)
    {
        if (os_->get_current_threadid() == thread::main_thread){
            sst_throw(illformederror, "user_thread_mpiqueue::do_send: on main thread");
        }
        mess->get_protocol()->send_header(ptr(this), mess);
    }

    void
    user_thread_mpiqueue::do_recv(const mpiqueue::recvrequest::ptr &req)
    {
        if (os_->get_current_threadid() == thread::main_thread){
            sst_throw(illformederror, "user_thread_mpiqueue::do_recv: on main thread");
        }
        mpiqueue::start_recv(req);
    }

    void
    user_thread_mpiqueue::buffer_unexpected(const mpimessage::ptr &msg)
    {
        SSTMACBacktrace("MPI Queue Buffer Unexpected Message");
        if (os_->get_current_threadid() == thread::main_thread){
            sst_throw(illformederror, "buffering on main thread");
        }
        user_lib_mem_->copy(msg->payload_bytes(), thread::nic_thread, msg->dest_thread(os_));
    }

    void
    user_thread_mpiqueue::incoming_message(const mpimessage::ptr &message)
    {
        if (os_->get_current_threadid() != -1){
            sst_throw(illformederror, "user_thread_mpiqueue::incoming message not on main thread");
        }
        if (is_polling_){
            //the thread is ready to receive
            incoming_msg_ = message;
            SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
                         << "]::unblocking on key " << (long) blocking_key_.get()
                         <<" \n";
            if (!blocking_key_->still_blocked()){
                sst_throw(illformederror, "user_thread_mpiqueue::incoming_message: key is no longer blocked");
            }

            os_->unblock(blocking_key_);
        }
        else {
            SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
                         << "]::incoming_message " << message->to_string()
                         << " Add to pending since not in progress loop\n";

            //this is not polling in a progress loop
            pending_msgs_.push_back(message);
        }

    }

    void
    user_thread_mpiqueue::buffered_recv(const mpimessage::ptr &msg, const mpiqueue::recvrequest::ptr &req)
    {
        SSTMACBacktrace("MPI_buffered_recv");
        user_lib_mem_->copy(msg->payload_bytes(), thread::nic_thread, msg->dest_thread(os_));
        req->handle(msg);
    }

    void
    user_thread_mpiqueue::buffered_send(const mpimessage::ptr &msg)
    {
        SSTMACBacktrace("MPI_buffered_send");
        // we may be sending a header or a payload
        // regardless, we need to copy the actual payload
        user_lib_mem_->copy(msg->payload_bytes(), msg->source_thread(os_), thread::nic_thread);
        post_header(msg);
    }

    void
    user_thread_mpiqueue::post_header(const mpimessage::ptr &msg)
    {
        SSTMACBacktrace("MPI Queue Post Header");
        if (post_header_delay_.ticks_int64()){
            user_lib_time_->compute(post_header_delay_);
        }
        server_->send(msg);
    }

    void
    user_thread_mpiqueue::post_rdma(const mpimessage::ptr &msg)
    {
        SSTMACBacktrace("MPI Queue Post RDMA Request");
        if (post_rdma_delay_.ticks_int64()){
            user_lib_time_->compute(post_rdma_delay_);
        }
        server_->send(msg);
    }

    }
}
