/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/api/mpi/mpi_strategies/mpi_collective/engines/mpibtreescatter.h>
#include <sstmac/software/api/mpi/mpi_queue/mpiqueue.h>
#include <sstmac/software/api/mpi/mpistatus.h>
#include <sstmac/software/api/mpi/mpi_strategies/mpi_collective/mpicollpayload.h>
#include <sstmac/common/eventhandler.h>

#include <sstmac/software/api/mpi/mpi_strategies/mpi_iterator/btscatteriterator.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>
#include <boost/lexical_cast.hpp>

namespace sstmac { namespace sw
{

  // Pointer types.
  
  
  using boost::dynamic_pointer_cast;
  using boost::lexical_cast;

  //
  // Given an absolute rank, return the relative rank.
  //
  int
  mpibtreescatter::relrank(const mpiid &therank) const
  {
    return ((therank.id - root_.id + comm_->size().id) % comm_->size().id);
  }

  //
  // Given a relative rank, return the absolute rank.
  //
  mpiid
  mpibtreescatter::absrank(int therank) const
  {
    return mpiid((therank + root_.id) % comm_->size().id);
  }

  //
  // Start our zero or one receive..
  //
  void
  mpibtreescatter::init_recv()
  {

      SSTMAC_DEBUG << "mpibtreescatter[" << comm_->rank().id << "]::init_recv(" << ")\n";
    // Figure out where we are receiving from.
    const int relative_rank = relrank(comm_->rank());
    if (relative_rank == 0)
      {
        // We are root -- charge straight to start_send.
        this->init_sends();
        return;
      }
    int keymask = 1;
    while (!(relative_rank & keymask))
      {
        keymask <<= 1;
      }
    const int relative_source = relative_rank - keymask;
    int endrecv = relative_rank + keymask;
    if (endrecv > comm_->size().id)
      endrecv = comm_->size().id;
    const int entries = endrecv - relative_rank;
    const int length = entries * recvcount_;
    ++pending_recvs_;
    this->start_recv(length, recvtype_, tag_, absrank(relative_source));
  }

  //
  // Start our zero or more sends.
  //
  void
  mpibtreescatter::init_sends()
  {

      SSTMAC_DEBUG << "mpibtreescatter[" << comm_->rank().id << "]::init_sends("
          << ")\n";
    // We should never get here with a pending receive.
    if (pending_recvs_)
      throw illformederror("mpibtreescatter::init_sends:  Pending receive.");
    const int relative_rank = relrank(comm_->rank());
    const int size = comm_->size().id;
    int keymask = 1;
    if (relative_rank == 0)
      {
        // Root node -- we are interested in one-past MSB
        int v = size;
        int shift = 0;
        while (v)
          {
            ++shift;
            v >>= 1;
          }
        keymask = 1 << shift;
      }
    else
      {
        while (!(relative_rank & keymask))
          {
            keymask <<= 1;
          }

          SSTMAC_DEBUG << " -- node " << comm_->rank().id << " with relative rank "
              << relative_rank << " has keymask " << keymask << "\n";
      }
    // Figure out initial send/receive pair (if any).
    int localmask = keymask >> 1;
    int endsend = relative_rank + keymask;
    if (endsend > size)
      {
        endsend = size;
      }
    // And start.
    while (localmask > 0)
      {
        int relative_dest = relative_rank + localmask;
        if (relative_dest >= size)
          {
            localmask >>= 1;
            continue;
          }
        const int entries = endsend - relative_dest;
        const int length = entries * sendcount_;
        ++pending_sends_;
        mpiid rrank = absrank(relative_dest);
        this->start_send(length, sendtype_, tag_, rrank, content_);
        endsend = relative_dest;
        localmask >>= 1;
      }
    // Some of the nodes are not doing any sends.
    if (pending_sends_ == 0 && pending_recvs_ == 0){
        if (content_){
            mpicollective::complete(content_->get_content()[comm_->rank().id]);
        }
        else {
            mpicollective::complete(payload::const_ptr());
        }
    }

  }

  //
  // Hi.
  //
  mpibtreescatter::mpibtreescatter(const mpirequest::ptr& thekey,
      const mpiqueue::ptr& queue, int sendcount,
      const mpitype &sendtype, int recvcount, const mpitype &recvtype,
      const mpiid &root, const mpitag &tag, const mpicomm::ptr &comm,
      const std::vector<payload::const_ptr > &cc,
          eventhandler::ptr completion) :
    mpicollective(thekey, queue, tag, comm, completion), sendcount_(sendcount),
        sendtype_(sendtype), recvcount_(recvcount), recvtype_(recvtype), root_(
            root), pending_sends_(0),
        pending_recvs_(0)
  {
    static const std::string here("mpibtreescatter");
    if (cc.size()){
        content_ = mpicollpayload::construct(cc);
    }

  }

  //
  // Callback method to indicate that a send operation has completed.
  //
  void
  mpibtreescatter::send_complete(int count, const mpitype &type,
      const mpitag &tag, const mpiid &dest)
  {

      SSTMAC_DEBUG << "mpibtreescatter[" << comm_->to_string() << "]::send_complete("
          << ", " << count << ", " << type.id << ", " << tag.id << ", "
          << dest.id << ")\n";
    --pending_sends_;
    if (!pending_sends_)
      {
        // All done.
          SSTMAC_DEBUG << " -- sends complete on node " << comm_->rank().id << "\n";
          if (content_)
            mpicollective::complete(content_->get_content()[comm_->rank().id]);
          else
            mpicollective::complete(content_);
      }
  }

  //
  // Callback method to indicate that a receive operation has completed.
  //
  void
  mpibtreescatter::recv_complete(int count,
      const mpistatus_t& stat)
  {

      SSTMAC_DEBUG << "mpibtreescatter[" << comm_->to_string() << "]::recv_complete("
          << ", " << count << ", " << stat << ")\n";
      if (content_){
          mpicollpayload::const_ptr incoming = dynamic_pointer_cast<const mpicollpayload> (stat->content());
          content_ = mpicollpayload::construct(incoming);
      }

    --pending_recvs_;
    if (!pending_recvs_)
      this->init_sends();
  }

  //
  // Completion of a copy event.
  //
  void
  mpibtreescatter::copy_complete(int count, const mpitype &type,
      const mpitag &tag, const mpiid &source)
  {

      SSTMAC_DEBUG << "mpibtreescatter[" << comm_->to_string() << "]::copy_done(" << ", "
          << count << ", " << type.id << ", " << tag.id << ", " << source.id
          << ")\n";
    --pending_recvs_;
    if (!pending_recvs_)
      this->init_sends();
  }

  //
  // Hello.
  //
  mpibtreescatter::ptr
  mpibtreescatter::construct(const mpirequest::ptr& thekey,
      const mpiqueue::ptr& queue, int sendcount,
      const mpitype &sendtype, int recvcount, const mpitype &recvtype,
      const mpiid &root, const mpitag &tag, const mpicomm::ptr &comm,
      const std::vector<payload::const_ptr > &cc,
          eventhandler::ptr completion)
  {
    mpibtreescatter::ptr retval(new mpibtreescatter(thekey, queue,
        sendcount, sendtype, recvcount, recvtype, root, tag, comm, cc,
        completion));

    return retval;
  }

  //
  // Goodbye.
  //
  mpibtreescatter::~mpibtreescatter() throw ()
  {
  }

  //
  // Start this kernel.
  //
  void
  mpibtreescatter::start()
  {

      SSTMAC_DEBUG << "mpibtreescatter[" << comm_->to_string() << "]::start(" << ", "
          << ")\n";
    mpicollective::start();
    this->init_recv();
  }

} } // end of namespace sstmac
