/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/api/mpi/mpi_strategies/mpi_collective/engines/mpidirectscatterv.h>

#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>
#include <boost/lexical_cast.hpp>

namespace sstmac { namespace sw {

  // Pointer types.
  
  
  using boost::dynamic_pointer_cast;
  using boost::lexical_cast;
  
  //
  // Hi.
  //
  mpidirectscatterv::mpidirectscatterv(const mpirequest_t& thekey,
                                       const mpiqueue::ptr& queue,
                                       const std::vector<int> &sendcount,
                                       const mpitype &sendtype,
                                       int recvcount, const mpitype &recvtype,
                                       const mpiid &root, const mpitag &tag,
                                       const mpicomm::ptr &comm,
                                       const std::vector<payload::const_ptr >& cont,
                                       eventhandler::ptr completion) :
    mpicollective(thekey, queue, tag, comm, completion),
    sendcount_(sendcount), sendtype_(sendtype),
    recvcount_(recvcount), recvtype_(recvtype),
    root_(root),
    pending_sends_(0), pending_recvs_(0)
  {
    static const std::string here("mpibdirectscatterv");
    if(comm->rank() == root) {

      content_ = cont;
    }
  }
  
  //
  // Callback method to indicate that a send operation has completed.
  //
  void mpidirectscatterv::send_complete(int count,
                                       const mpitype &type, const mpitag &tag,
                                       const mpiid &dest)
  {
    SSTMAC_DEBUG << "mpidirectscatterv[" << comm_->rank().id << "]::send_complete("  << ", " << count
              << ", " << type.to_string() << ", " << tag.to_string() << ", " << dest.to_string() << ")\n";
    --pending_sends_;
    if(! pending_sends_)
      mpicollective::complete(content_[comm_->rank().id]);
  }

  //
  // Callback method to indicate that a receive operation has completed.
  //
  void mpidirectscatterv::recv_complete( int count,
                                        const mpistatus_t& stat)
  {
    SSTMAC_DEBUG << "mpidirectscatterv[" << comm_->rank().id << "]::recv_complete(" << ", " << count
              << ", " << stat << ")\n";
    payload::const_ptr load = stat->content();


    --pending_recvs_;
        if(! pending_recvs_)
          mpicollective::complete(load);
  }

  //
  // Callback to signal copy completion (only on clients).
  //
  void mpidirectscatterv::copy_complete(int count,
                                        const mpitype &type, const mpitag &tag,
                                        const mpiid &source)
  {
    SSTMAC_DEBUG << "mpidirectscatterv[" << comm_->rank().id << "]::copy_done("  << ", " << count << ", "
              << type.to_string() << ", " << tag.to_string() << ", " << source.to_string() << ")\n";
    //--pending_recvs_;
    //if(! pending_recvs_)
    //  mpicollective::complete(content_->get(comm_->rank()));
  }

  //
  // Hello.
  //
  mpidirectscatterv::ptr
  mpidirectscatterv::construct(const mpirequest_t& thekey,
                               const mpiqueue::ptr& queue,
                               const std::vector<int> &sendcount,
                               const mpitype &sendtype,
                               int recvcount, const mpitype &recvtype,
                               const mpiid &root, const mpitag &tag,
                               const mpicomm::ptr &comm,
                               const std::vector<payload::const_ptr >& content,
                               eventhandler::ptr completion)
  {
    mpidirectscatterv::ptr retval
      (new mpidirectscatterv(thekey, queue, sendcount, sendtype,
                             recvcount, recvtype, root, tag, comm, content, completion));

    return retval;
  }

  //
  // Goodbye.
  //
  mpidirectscatterv::~mpidirectscatterv() throw() {
  }

  //
  // Start this kernel.
  //
  void mpidirectscatterv::start()
  {
    SSTMAC_DEBUG << "mpidirectscatterv[" << comm_->rank().id << "]::start()\n";
    mpicollective::start();
    if(comm_->rank() == root_) {
      // Root node -- fire off a bunch of sends.
      mpiid size = comm_->size();
      pending_sends_ = comm_->size().id - 1;
      if(pending_sends_ == 0) {
        mpicollective::complete(content_[comm_->rank().id]);
      }
      else {
        for(mpiid dest(0); dest < size; ++dest) {
          if(dest != root_) {
            this->start_send(sendcount_.at(dest.id), sendtype_, tag_, dest,
                             content_.at(dest.id));
          }
        }
      }
    }
    else {
      // Client node -- fire off a single receive.
      pending_recvs_ = 1;
      this->start_recv(recvcount_, recvtype_, tag_, root_);
    }
  }


} } // end of namespace sstmac
