/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/api/mpi/mpi_strategies/mpi_collective/engines/mpilinearbarrierengine.h>
#include <sstmac/software/api/mpi/mpi_queue/mpiqueue.h>
#include <sstmac/software/api/mpi/mpistatus.h>
#include <sstmac/common/messages/payload.h>
#include <sstmac/software/api/mpi/mpi_strategies/mpi_iterator/topologyiterator.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>

namespace sstmac
{
  namespace sw
  {

    //
    // Fire off the next set of send and receive operations.
    //
    void
    mpilinearbarrierengine::sendrecv(bool firsttime)
    {
      SSTMAC_DEBUG << "mpilinearbarrierengine[" << comm_->rank().id
            << "]:: sendrecv \n";

      if (comm_->rank() == root_)
        {
          // Fire off all sends and receives.
          for (long it = 0; it < comm_->size().id; ++it)
            {
              if (firsttime)
                {
                  if (it != root_.id)
                    {
                      ++pending_recvs_;
                      this->start_recv(1, mpitype::mpi_byte, tag_, mpiid(it));
                    }
                }
              else
                {
                  if (it != root_.id)
                    {
                      ++pending_sends_;
                      this->start_send(1, mpitype::mpi_byte, tag_, mpiid(it),
                          payload::null());
                    }
                }
            }
        }
      else
        {
          if (firsttime)
            {
              ++pending_recvs_;
              this->start_recv(1, mpitype::mpi_byte, tag_, root_);

              ++pending_sends_;
              this->start_send(1, mpitype::mpi_byte, tag_, root_,
                  payload::null());
            }
        }

      // If we have no pending sends or receives, we are done.
      if (pending_sends_ == 0 && pending_recvs_ == 0)
        {
          mpicollective::complete(payload::const_ptr());
        }
    }

    //
    // Callback method to indicate that a send operation has completed.
    //
    void
    mpilinearbarrierengine::send_complete(int count, const mpitype &type,
        const mpitag &tag, const mpiid &dest)
    {
      SSTMAC_DEBUG << "mpilinearbarrierengine[" << comm_->rank().id
            << "]::send_complete(" << count << ", " << type.to_string() << ", "
            << tag.to_string() << ", " << dest.to_string() << ")\n";
      --pending_sends_;
      if (!pending_sends_)
        {
          if (comm_->rank() == root_)
            {
              mpicollective::complete(payload::const_ptr());
            }
          else
            {
              this->sendrecv(false);
            }
        }
    }

    //
    // Callback method to indicate that a receive operation has completed.
    //
    void
    mpilinearbarrierengine::recv_complete(int count, const mpistatus_t& stat)
    {
      SSTMAC_DEBUG << "mpilinearbarrierengine[" << comm_->rank().id
            << "]::recv_complete(" << count << ", " << ")\n";
      --pending_recvs_;
      if (!pending_recvs_)
        {
          this->sendrecv(false);
        }
    }

    //
    // Copy complete.
    //
    void
    mpilinearbarrierengine::copy_complete(int, const mpitype&, const mpitag&,
        const mpiid&)
    {
      throw illformederror("mpilinearbarrierengine::copy_complete:  "
          "MPI barrier doesn't copy data.");
    }

    //
    // Hi.
    //
    mpilinearbarrierengine::mpilinearbarrierengine(
        const topologyiterator::ptr& iter, const mpirequest::ptr& the_key,
        const mpiqueue::ptr& queue, const mpitag &tag, const mpicomm::ptr &comm,
        const mpiid& root, eventhandler::ptr completion) :
        mpicollective(the_key, queue, tag, comm, completion), iter_(iter),
        pending_sends_(0), pending_recvs_(0), root_(root)
    {
      if (tag_ == mpitag::any_tag)
        throw valueerror("mpilinearbarrierengine(...)"
            ":  refusing to send using mpitag::any_tag.");
    }

    //
    // Hi.
    //
    mpilinearbarrierengine::ptr
    mpilinearbarrierengine::construct(const topologyiterator::ptr& iter,
        const mpirequest::ptr& the_key, const mpiqueue::ptr& queue,
        const mpitag &tag, const mpicomm::ptr &comm, const mpiid& root,
        eventhandler::ptr completion)
    {
      mpilinearbarrierengine::ptr retval(
          new mpilinearbarrierengine(iter, the_key, queue, tag, comm, root,
              completion));

      return retval;
    }

    //
    // Goodbye.
    //
    mpilinearbarrierengine::~mpilinearbarrierengine() throw ()
    {
    }

    //
    // Start this kernel.
    //
    void
    mpilinearbarrierengine::start()
    {
      SSTMAC_DEBUG << "mpilinearbarrierengine[" << comm_->rank().id
            << "]::start \n";
      mpicollective::start();
      this->sendrecv(true);
    }

  }
} // end of namespace sstmac

