#include <sstmac/software/api/mpi/mpi_strategies/mpi_collective/engines/mpilogtwovirt.h>

namespace sstmac {
namespace sw {

int
mpilogtwovirtallreduce::two_to_power(int n)
{
    int p2 = 1;
    for (int i=0; i < n; ++i)
        p2 *= 2;
    return p2;
}

int mpilogtwovirtallreduce::log_base_two(int n)
{
    int log2 = 0;
    int remainder = n;
    while (remainder > 1){
        ++log2;
        remainder /= 2;
    }
    return log2;
}

void
mpilogtwovirtallreduce::configure_round(int me, int round,
                                        int& count, int& partner)
{
    int delta = two_to_power(round);
    int group_mod_divider = delta * 2;
    int my_group_rank = me % group_mod_divider;
    count = count_ / group_mod_divider;

    bool power_bottom = my_group_rank < delta;
    if (power_bottom)
        partner = me + delta;
    else
        partner = me - delta;
}

int
mpilogtwovirtallreduce::virtual_rank_to_actual_rank(int rank)
{
    if (rank < nproc_){
        return rank;
    }
    //I exceed max by
    int max_excess = rank - nproc_ + 1;

    int delta_exchange = 1;
    while (delta_exchange < max_excess){
        delta_exchange *= 2;
    }

    //figure out the first round something real tries to exchange with this
    int partner = nproc_;
    int count = 0;
    int round = 0;
    while (partner >= nproc_){
        configure_round(rank, round, count, partner);
        ++round;
    }

    return partner;
}

void
mpilogtwovirtallreduce::do_run_round(int round, int me, int virtual_me,
                                     std::list<exchange>& to_do)
{
    //printf("Round %d for virtual node %d real node %d\n",
    //       round, virtual_me, my_rank_);
    reducing_round_ = round < (num_rounds_ / 2);
    int effective_round = reducing_round_ ? round : num_rounds_ - round - 1;
    int virtual_partner, count;
    configure_round(virtual_me, effective_round, count, virtual_partner);
    int partner = virtual_partner;
    if (partner >= nproc_){
        partner = virtual_rank_to_actual_rank(partner);
    }

    // you don't need to map two virtual nodes to each other
    if (virtual_me >= nproc_ && virtual_partner >= nproc_){
    }
    else if (partner == me){
        //make sure I take on this role going forward
        my_roles_.insert(virtual_partner);
        //printf("Self-exchange %d<->%d (%d<->%d) in round %d count %d\n",
        //       virtual_me, virtual_partner, me, partner, round, count);
    }
    else{
        ++pending_sends_;
        ++pending_recvs_;

        /**



        */
        to_do.push_back(exchange(partner,count));
        //printf("Exchange %d<->%d (%d<->%d) in round %d count %d\n",
        //       virtual_me, virtual_partner, me, partner, round, count);
    }
}

void
mpilogtwovirtallreduce::copy_complete(int count, const mpitype &type,
                                      const mpitag &tag, const mpiid &source)
{
    sst_throw(unimplementederror, "mpilogwovirtallreduce::copy_complete");
}


void
mpilogtwovirtallreduce::run_next_round()
{
    //printf("Running round %d on node %d with %d virtual roles\n",
    //       round_, my_rank_, my_roles_.size());

    std::set<int> current_roles = my_roles_;
    std::set<int>::iterator it, end = current_roles.end();
    std::list<exchange> to_do;
    for (it=current_roles.begin(); it != end; ++it){
        int virtual_me = *it;
        do_run_round(round_, my_rank_, virtual_me, to_do);
    }

    if (to_do.size() == 0){
        //we didn't actually do anything
        ++round_;
        if (round_ == num_rounds_){
            mpicollective::complete(payload::const_ptr());
        }
        else{
            run_next_round();
        }
    }
    else {
        std::list<exchange>::iterator it, end = to_do.end();
        for (it=to_do.begin(); it != end; ++it){
            exchange& e = *it;
            start_send(e.count, type_, tag_, mpiid(e.partner));
            start_recv(e.count, type_, tag_, mpiid(e.partner));
        }
    }
}

void
mpilogtwovirtallreduce::send_complete(
        int count, const mpitype &type,
        const mpitag &tag, const mpiid &dest)
{
    --pending_sends_;
    if (round_ == num_rounds_ && pending_sends_ == 0)
            mpicollective::complete(payload::const_ptr());
}

void
mpilogtwovirtallreduce::recv_complete(int count, const mpistatus_t &stat)
{
    --pending_recvs_;
    if (reducing_round_){
        //nothing - payloads not yet supported
    }

    if (pending_recvs_ == 0){
        //this round of recvs is complete
        ++round_;
        if (round_ == num_rounds_){
            if (pending_sends_ == 0)
                mpicollective::complete(payload::const_ptr());
        }
        else {
            run_next_round();
        }
    }


}

mpilogtwovirtallreduce::mpilogtwovirtallreduce(
        const mpirequest::ptr &thekey,
        const mpiqueue::ptr &queue,
        int count, const mpitype &datatype,
        const mpiop &op, const mpiid &root,
        const mpitag &tag, const mpicomm::ptr &comm,
        const payload::const_ptr &content,
        const eventhandler::ptr &completion)
    : mpicollective(thekey, queue, tag, comm, completion),
    count_(count), type_(datatype),
    op_(op), root_(root),
    pending_sends_(0),
    pending_recvs_(0),
    reducing_round_(false),
    complete_lock_(false),
    nproc_(comm->size().id),
    my_rank_(comm->my_task().id),
      round_(0),
      num_rounds_(0)
{
    num_rounds_ = log_base_two(nproc_);
    int proc_test = two_to_power(num_rounds_);
    if (nproc_ != proc_test) //we don't have even power of two
        ++num_rounds_;

    //we need to fan out and fan back in
    num_rounds_ *= 2;
    my_roles_.insert(my_rank_);
}

mpilogtwovirtallreduce::ptr
mpilogtwovirtallreduce::construct(
    const mpirequest::ptr& thekey, const mpiqueue::ptr& queue,
    int count, const mpitype &datatype, const mpiop &op, const mpiid &root,
    const mpitag &tag, const mpicomm::ptr &comm,
    const payload::const_ptr& content,
    const eventhandler::ptr& completion)
{
  return new mpilogtwovirtallreduce(thekey, queue, count,
                                    datatype, op, root,
                                    tag, comm, content, completion);
}

void
mpilogtwovirtallreduce::start()
{
    run_next_round();
}

  }
}

