/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/api/mpi/mpi_strategies/mpi_collective/engines/mpiringallgather.h>

#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>

namespace sstmac
{
  namespace sw
  {

    // Pointer types.

    using boost::dynamic_pointer_cast;

    //
    // Hello.
    //
    mpiringallgather::mpiringallgather(const mpirequest::ptr& thekey,
        const mpiqueue::ptr& queue, int sendcnt, const mpitype &sendtype,
        int recvcnt, const mpitype &recvtype, const mpitag &tag,
        const mpicomm::ptr &comm, const payload::const_ptr& content,
        eventhandler::ptr completion) :
      mpicollective(thekey, queue, tag, comm, completion), sendcnt_(sendcnt),
          sendtype_(sendtype), recvcnt_(recvcnt), recvtype_(recvtype),
          iteration_(0), target_iterations_(comm->size().id - 1),
          source_((comm->size().id + comm->rank().id - 1) % comm->size().id),
          dest_((comm->rank().id + 1) % comm->size().id), pending_sends_(0),
          pending_recvs_(0)
    {

      if (content)
      {
        std::vector<payload::const_ptr> vals(comm->size().id, payload::const_ptr());
        vals[comm->rank().id] = content;
     //   for (int i = 0; i < comm->size().id; i++)
      //  {
       //   if (i == comm->rank().id)
        //    vals.push_back(content);
        //  else
        //    vals.push_back(payload::const_ptr());
       // }

        content_ = mpicollpayload::construct(vals);
      }
      else
      {
        content_ = mpicollpayload::ptr();
      }

    }

    //
    // Callback method to indicate that a send operation has completed.
    //
    void
    mpiringallgather::send_complete(int count, const mpitype &type,
        const mpitag &tag, const mpiid &dest)
    {
      SSTMAC_DEBUG
            << sstprintf(
                "on rank %d: mpiringallgather::send_complete(count=%d,tag=%d,src=%d,dest=%d)\n",
                comm_->rank().id, count, tag.id, comm_->rank().id, dest.id);
      --pending_sends_;
      if (pending_sends_ == 0 && pending_recvs_ == 0)
      {
        mpicollective::complete(content_);
      }
    }

    //
    // Callback method to indicate that a receive operation has completed.
    //
    void
    mpiringallgather::recv_complete(int count, const mpistatus_t& stat)
    {

      SSTMAC_DEBUG
            << sstprintf(
                "on rank %d: mpiringallgather::recv_complete(count=%d,tag=%d,src=%d,dest=%d)\n",
                comm_->rank().id, count, stat->tag().id, stat->source().id,
                comm_->rank().id);
      ++iteration_;

      payload::const_ptr pay = stat->content();
      if (pay)
      {
        mpicollpayload::const_ptr tmp = dynamic_pointer_cast<
            const mpicollpayload> (stat->content()->clone());
        //mpicollpayload::ptr other = mpicollpayload::construct(
        //    dynamic_pointer_cast<const mpicollpayload> (stat->content()->clone()));
        if (tmp)
        {
          mpicollpayload::ptr other = mpicollpayload::construct(tmp);

          std::vector<payload::const_ptr> realload = other->get_content();

          std::vector<payload::const_ptr>::iterator it, end = realload.end();

          int i = 0;
          for (it = realload.begin(); it != end; it++)
          {
            if (*it)
            {
              SSTMAC_DEBUG << "mpiringallgather(" << comm_->rank().id
                    << "]: setting content for position " << i << "\n";

              content_->set_content(*it, i);
            }
            i++;
          }
        }
      }

      /** DEPRECATED JJW 01/09/2012 Do not chain recvs anymore.
       All sends/recvs are now started at the beginning. This
       generates a massive call stack that breaks the simulator
       */
      if (iteration_ < target_iterations_)
      {
        // Next send/receive pair.
        ++pending_recvs_;
        this->start_recv(recvcnt_, recvtype_, tag_, source_);
        /** JJW 01/09/2012 Even though we don't chain recvs,
         we should still chain sends. This won't generate
         a massive call stack */
        ++pending_sends_;
        this->start_send(sendcnt_, sendtype_, tag_, dest_, content_);
      }
      --pending_recvs_;

      if (pending_sends_ == 0 && pending_recvs_ == 0)
      {

        mpicollective::complete(content_);
      }

    }

    //
    // Callback to signal copy completion (ignored).
    //
    void
    mpiringallgather::copy_complete(int count, const mpitype &type,
        const mpitag &tag, const mpiid &source)
    {
      SSTMAC_DEBUG << "mpiringallgather::copy_complete(" << ", " << count
            << ", " << type.id << ", " << tag.id << ", " << source.id << ")\n";
    }

    //
    // Hello.
    //
    mpiringallgather::ptr
    mpiringallgather::construct(const mpirequest::ptr& thekey,
        const mpiqueue::ptr& queue, int sendcnt, const mpitype &sendtype,
        int recvcnt, const mpitype &recvtype, const mpitag &tag,
        const mpicomm::ptr &comm, const payload::const_ptr& content,
        eventhandler::ptr completion)
    {
      mpiringallgather::ptr retval(
          new mpiringallgather(thekey, queue, sendcnt, sendtype, recvcnt,
              recvtype, tag, comm, content, completion));

      return retval;
    }

    //
    // Goodbye.
    //
    mpiringallgather::~mpiringallgather() throw ()
    {
    }

    //
    // Get busy child.
    //
    void
    mpiringallgather::start()
    {
      SSTMAC_DEBUG << sstprintf("mpiringallgather::start(tag=%d)\n", tag_.id);
      mpicollective::start();
      // Catch situations where comm_->size() is 1.
      if (target_iterations_ == 0)
      {

        mpicollective::complete(content_);
      }
      else
      {

        /** Start all the recvs at once to avoid 
         a giant handle_recv recursive call stack that
         breaks the simulator */
        iteration_ = 0;
        pending_sends_ = pending_recvs_ = 0;

        ++pending_recvs_;
        this->start_recv(recvcnt_, recvtype_, tag_, source_);

        /** The payload content_ gets updated after each recv.  Thus,
         we do not want to start all sends at once.  Each successive
         recv will launch the next send */
        ++pending_sends_;
        this->start_send(sendcnt_, sendtype_, tag_, dest_, content_);
      }
    }

  }
} // end of namespace sstmac
