/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_STRATEGIES_MPI_COLLECTIVE_ENGINES_MPIRINGALLGATHERV_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_STRATEGIES_MPI_COLLECTIVE_ENGINES_MPIRINGALLGATHERV_H_INCLUDED

#include <sstmac/software/api/mpi/mpi_strategies/mpi_collective/mpicollective.h>
#include <sstmac/software/api/mpi/mpi_strategies/mpi_collective/mpicollpayload.h>

namespace sstmac { namespace sw {


  /**
   * Perform an allgatherv using a ring.
   * Each node performs size-1 iterations where it receives data from node
   * rank-1 and sends to node rank+1 (where both ranks wrap around).
   */
  class mpiringallgatherv : public mpicollective {
    /// Private nested type to hold send and receive counts.
    struct srpair {
      explicit srpair(int send = -1, int recv = -1) :
        sendcnt(send), recvcnt(recv)
      {}
      int sendcnt;
      int recvcnt;
    };
    
    /// The number of elements sent to each node.
    int sendcnt_;
    /// The type we're sending.
    mpitype sendtype_;
    /// The number of elements received from each node.
    std::vector<int> recvcnt_;
    /// The type we are receiving.
    mpitype recvtype_;

    /// The remaining iteration rounds (we pop_front on each iteration).
    std::deque<srpair> iteration_;
    /// The node we will be receiving from.
    const mpiid source_;
    /// The node we will be sending to.
    const mpiid dest_;
    /// The number of pending sends (usually zero or one).
    int pending_sends_;
    /// The number of pending receives (usually zero or one).
    int pending_recvs_;

    mpicollpayload::ptr content_;

  protected:
    /// Hello.
    mpiringallgatherv(const mpirequest::ptr& thekey,
                      const mpiqueue::ptr& queue,
                      int sendcnt, const mpitype &sendtype,
                      const std::vector<int> &recvcnt, const mpitype &recvtype,
                      const mpitag &tag, const mpicomm::ptr &comm,
                      const payload::const_ptr& content,
                      eventhandler::ptr completion);

    /// Callback method to indicate that a send operation has completed.
    virtual void send_complete(
                               int count,
                               const mpitype &type,
                               const mpitag &tag,
                               const mpiid &dest);

    /// Callback method to indicate that a receive operation has completed.
    virtual void recv_complete(
                               int count,
                               const mpistatus_t& stat);

    /// Callback to signal copy completion.
    void copy_complete(
                       int count,
                       const mpitype &type,
                       const mpitag &tag,
                       const mpiid &source);

  public:
    typedef boost::intrusive_ptr<mpiringallgatherv> ptr;

    std::string to_string() const {
        return sstprintf("mpiringallgatherv(tag=%d)", tag_.id);
    }

    /// Hello.
    static mpiringallgatherv::ptr
    construct(const mpirequest_t& thekey,
              const mpiqueue::ptr& queue,
              int sendcnt, const mpitype &sendtype,
              const std::vector<int> &recvcnt, const mpitype &recvtype,
              const mpitag &tag, const mpicomm::ptr &comm,
              const payload::const_ptr& content,
              eventhandler::ptr completion);
    /// Goodbye.
    virtual ~mpiringallgatherv() throw();

    /// Get busy child.
    virtual void start();
  };

} } // end of namespace sstmac

#endif
