#include <sstmac/software/api/mpi/mpi_strategies/mpi_collective/engines/mpisimultaneousbtreescan.h>

namespace sstmac {
    namespace sw {

mpi_simultaneous_btree_scan_engine::mpi_simultaneous_btree_scan_engine(
  const mpirequest::ptr& thekey,
  const mpiqueue::ptr& queue,
  int count, const mpitype &type,
  const mpiop &op,
  const mpitag &tag, const mpicomm::ptr &comm,
  const payload::const_ptr& content,
  const eventhandler::ptr& completion
        ) : mpicollective(thekey, queue, tag, comm, completion),
    count_(count), type_(type), op_(op), content_(content),
    send_round_(0), send_two_to_the_k_(1),
    recv_round_(0), recv_two_to_the_k_(1),
    me_(comm->my_task().id),
    nproc_(comm->size().id),
    pending_sends_(0), pending_recvs_(0),
    send_done_(false), recv_done_(false)
{
}

void
mpi_simultaneous_btree_scan_engine::next_send()
{
    int dst = me_ + send_two_to_the_k_;
    ++send_round_;
    send_two_to_the_k_ *= 2;
    if (dst < nproc_){
        SSTMAC_DEBUG << sstprintf("mpi_simultaneous_btree_scan: send on %d to %d in round %d \n",
                                  me_, dst, send_round_ - 1);
        ++pending_sends_;
        start_send(count_, type_, tag_, mpiid(dst), content_);
    }
    else {
        SSTMAC_DEBUG << sstprintf("mpi_simultaneous_btree_scan: send done on %d \n", me_);
        send_done_ = true;
    }
}

void
mpi_simultaneous_btree_scan_engine::next_recv()
{
    int src = me_ - recv_two_to_the_k_;
    ++recv_round_;
    recv_two_to_the_k_ *= 2;
    if (src >= 0){
        SSTMAC_DEBUG << sstprintf("mpi_simultaneous_btree_scan: recv on %d from %d in round %d \n",
                                  me_, src, recv_round_ - 1);
        ++pending_recvs_;
        start_recv(count_, type_, tag_, mpiid(src));
    }
    else {
        SSTMAC_DEBUG << sstprintf("mpi_simultaneous_btree_scan: recv done on %d \n", me_);
        recv_done_ = true;
    }
}

void
mpi_simultaneous_btree_scan_engine::sendrecv()
{
    next_send();
    next_recv();

    if (recv_done_) //finish all my sends
    {
        while (!send_done_)
        {
            next_send();
        }
    }
}

void
mpi_simultaneous_btree_scan_engine::send_complete(int count, const mpitype &type, const mpitag &tag, const mpiid &dest)
{
    --pending_sends_;
    SSTMAC_DEBUG << sstprintf("mpi_simultaneous_btree_scan %d: send complete on %d to %d with %d pending sends %d pending recvs\n",
                              me_, dest.id, pending_sends_, pending_recvs_);


    if (pending_sends_ == 0 && pending_recvs_ == 0)
        complete(content_);

}

void
mpi_simultaneous_btree_scan_engine::recv_complete(int count, const mpistatus_t& stat)
{
    payload::const_ptr load = stat->content();
    if (load)
        content_ = combine_content(content_, load, op_, comm_->rank_.id, stat->source().id);

    //each recv triggers the next sendrecv
    sendrecv();

    --pending_recvs_;
    SSTMAC_DEBUG << sstprintf("mpi_simultaneous_btree_scan %d: recv complete on %d from %d  with %d pending sends %d pending recvs \n",
                              me_, stat->source().id, pending_sends_, pending_recvs_);

    if (pending_recvs_ == 0 && pending_sends_ == 0)
        complete(content_);

}

void
mpi_simultaneous_btree_scan_engine::start()
{
    if (nproc_ == 1)
        mpicollective::complete(content_);
    else
        sendrecv();
}

void
mpi_simultaneous_btree_scan_engine::copy_complete(int count, const mpitype &type, const mpitag &tag, const mpiid &source)
{
    sst_throw(illformederror, "cannot call copy_complete on btree scan");
}

mpi_simultaneous_btree_scan_engine::~mpi_simultaneous_btree_scan_engine() throw()
{
}

mpi_simultaneous_btree_scan_engine::ptr
mpi_simultaneous_btree_scan_engine::construct(
    const mpirequest::ptr &thekey,
    const mpiqueue::ptr &queue,
    int count, const mpitype &type,
    const mpiop &op,
    const mpitag &tag,
    const mpicomm::ptr &comm,
    const payload::const_ptr &content,
    const eventhandler::ptr &completion)
{
    return new mpi_simultaneous_btree_scan_engine(thekey, queue, count, type, op, tag, comm, content, completion);
}

std::string
mpi_simultaneous_btree_scan_engine::to_string() const
{
    return "mpi_scan";
}

    }
}
