/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/api/mpi/mpiapi.h>
#include <sstmac/software/api/mpi/mpi_strategies/mpi_collective/mpicollective.h>
#include <sstmac/software/api/mpi/mpi_strategies/mpi_collective/mpicollective-sender.h>
#include <sstmac/software/api/mpi/mpi_strategies/mpi_collective/mpicollective-recver.h>
#include <sstmac/software/api/mpi/mpi_strategies/mpi_collective/mpicollective-funneledcopy.h>
#include <sstmac/software/api/mpi/mpi_strategies/mpi_collective/mpicollective-immediatecopy.h>
#include <sstmac/software/api/mpi/mpistatus.h>
//#include <sstmac/software/api/mpi/mpi_strategies/mpi_collective/mpicollpayload.h>
#include <sstmac/software/api/mpi/mpi_queue/mpiqueue.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>
#include <sstmac/common/messages/valuepayload.h>
#include <sstmac/software/api/mpi/mpipayload.h>

namespace sstmac
{
  namespace sw
  {

    // Pointer types.

    debug mpicollective::dbg_;

    //
    // Hi.
    //
    mpicollective::mpicollective(const mpirequest::ptr& the_key,
        const mpiqueue::ptr& queue, const mpitag& tag,
        const mpicomm::ptr &comm, const eventhandler::ptr& completion) :
      comm_(comm), tag_(tag), completion_(completion), queue_(queue),
          req_(the_key), status_(mpistatus::construct())

    {
      if (!dbg_.is_real())
      {
        dbg_ = debug("<debug> mpicollective | mpi | library", logger::cyan);
        logger::register_logger(&dbg_);
      }
    }

    //
    // Request a send.
    //
    void
    mpicollective::start_send(int count, const mpitype &type,
        const mpitag &tag, const mpiid &dest, const payload::const_ptr& content)
    {
      SSTMAC_DEBUG
            << sstprintf(
                "on rank %d: mpicollective::start_send(count=%d,tag=%d,src=%d,dest=%d)\n",
                comm_->rank().id, count, tag.id, comm_->rank().id, dest.id);

      mpirequest_t req = mpirequest::construct(key::construct(mpiapi::default_key_category));
      eventhandler::ptr snd = sender::construct(mpicollective::ptr(this),
          status_, count, type, tag, dest);
      mpiqueue::sendinfo sinfo;
      queue_->send(req, count, type, dest, tag, comm_, sinfo,
          mpimessage::collective, snd, content);
    }

    //
    // Request a recv.
    //
    void
    mpicollective::start_recv(int count, const mpitype &type,
        const mpitag &tag, const mpiid &source)
    {
      SSTMAC_DEBUG
            << sstprintf(
                "on rank %d: mpicollective::start_recv(count=%d,tag=%d,src=%d,dest=%d)\n",
                comm_->rank().id, count, tag.id, source.id, comm_->rank().id);

      mpirequest_t req = mpirequest::construct(key::construct(mpiapi::default_key_category));
      eventhandler::ptr rcv = recver::construct(mpicollective::ptr(this),
          count, type, tag, source);

      queue_->recv(req, count, type, source, tag, comm_,
          mpimessage::collective, rcv);
    }

    //
    // Request a funneled copy.
    //
    void
    mpicollective::start_funneled_copy(int count, const mpitype &type,
        const mpitag &tag, const mpiid &partner)
    {
      /** DEPRECATED JJW 01/15/2012
       Using funneled copies creates a mass memory leak
       */
      throw unimplementederror(
          "mpi collective funneled copy is a deprecated feature and "
            "requires a refactoring before being used.");

      if (funnel_ == 0)
      {
        throw nullerror(
            "mpicollective::start_funneled_copy: Null funnel object. "
              "Did you forget to call mpicollective::start?");
      }
      funnel_->copy(count, type, tag, partner);
    }

    //
    // Request an immediate copy.
    //
    void
    mpicollective::start_immediate_copy(int count, const mpitype &type,
        const mpitag &tag, const mpiid &partner)
    {
      // eventhandler::ptr copier = immediatecopy::construct(
      //     this->get_self(), count, type, tag, partner);
      //  node_->copy(node_->now(), count * type.size, copier);
    }

    //
    // Goodbye.
    //
    mpicollective::~mpicollective() throw ()
    {
    }

    //
    // Get busy child.
    //
    void
    mpicollective::start()
    {
      SSTMAC_DEBUG << "Starting mpicollective " << to_string() << " on rank "
            << comm_->rank().id << "\n";

      /** DEPRECATED JJW 01/15/2012
       The funneled copy currently never gets used and it is
       creating a circular reference the keeps the collective
       from getting deleted.  It should not be up to us
       to keep the funnel alive.  It should be up to the
       pending messages for the funnel to keep it alive

       funnel_ = funneledcopy::construct(mpicollective::ptr(this));
       */
    }

    void
    mpicollective::complete(const payload::const_ptr& content)
    {
      SSTMAC_DEBUG << "Completing mpicollective " << to_string() << " on rank "
            << comm_->rank().id << " with content " << content << "\n";
      status_->set_content(content);
      mpimessage::ptr fake = mpimessage::construct(content, mpimessage::fake);

      req_->complete(fake);
      completion_->handle(fake);
    }

    const payload::const_ptr
    mpicollective::combine_content(const payload::const_ptr &p1,
        const payload::const_ptr &p2, mpiop &op, int rankhere, int rankincoming)
    {
      payload::const_ptr ret;

      SSTMAC_DEBUG << "mpi_collective, combining " << p1->to_string() << " and " << p2->to_string() << "\n";

      if (op.label == "sum")
      {
        ret = p1->add(p2);
      }
      else if (op.label == "prod")
      {
        ret = p1->prod(p2);
      }
      else if (op.label == "min")
      {
        ret = p1->min(p2);
      }
      else if (op.label == "max")
      {
        ret = p1->max(p2);
      }
      else if (op.label == "minloc")
      {
        mpipayload::const_ptr pay1 = boost::dynamic_pointer_cast<
            const mpipayload>(p1);

        ret = pay1->minloc(p2);
      }
      else if (op.label == "maxloc")
      {
        mpipayload::const_ptr pay1 = boost::dynamic_pointer_cast<
            const mpipayload>(p1);

        ret = pay1->maxloc(p2);
      }
      else if (op.label == "lor")
      {
        ret = p1->logical_or(p2);
      }
      else if (op.label == "lxor")
      {
        ret = p1->logical_xor(p2);
      }
      else if (op.label == "land")
      {
        ret = p1->logical_and(p2);
      }
      else if (op.label == "band")
      {
        ret = p1->bitwise_and(p2);
      }
      else if (op.label == "bxor")
      {
        ret = p1->bitwise_xor(p2);
      }
      else if (op.label == "bor")
      {
        ret = p1->bitwise_or(p2);
      }
      else if (op.label == "user")
      {
        payload::const_ptr temp1;
        payload::const_ptr temp2;

        //need to respect commutative property
        if (op.commute_ || rankhere > rankincoming)
        {
          temp1 = p1;
          temp2 = p2;
        }
        else
        {
          temp1 = p2;
          temp2 = p1;
        }

        mpipayload::const_ptr cast = boost::dynamic_pointer_cast<
            const mpipayload>(temp1);
        if (!cast)
        {
          sst_throw(valueerror, "mpicollective::combine_content - not an mpipayload in user op, no can do");
        }
        ret = cast->user_op(temp2, op);
      }
      else
      {
        sst_throw(valueerror, "mpicollective: mpiop %s not supported yet",
            op.label.c_str());
      }

      SSTMAC_DEBUG << "...got " << ret->to_string() << "\n";

      return ret;
    }

  }
} // end of namespace sstmac
