/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/api/mpi/mpi_strategies/mpi_iterator/btscatteriterator.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>

namespace sstmac { namespace sw {

  // Pointer types.
  

  //
  // Hi.
  //
  btscatteriterator::btscatteriterator(int rank, int size, int root) :
    src_(-1), dst_(-1), mask_(1),
    rank_(rank), size_(size), root_(root)
  {
    static const std::string here("btscatteriterator(...)");
    // Sanity check.
    if(size_ <= 0)
      throw valueerror(here + ":  Invalid zero or negative) size.");
    if(rank_ < 0 || rank >= size_)
      throw valueerror(here + ":  Invalid rank.");
    if(root_ < 0 || root >= size_)
      throw valueerror(here + ":  Invalid root.");
    //
    // Compute where we will be receiving data from.
    int relative_rank = (rank_ >= root_ ? rank_-root_ : rank_-root_+size_);
    while(mask_ < size_) {
      if(relative_rank & mask_) {
        src_ = rank_ - mask_;
        if(src_ < 0) src_ += size_;
        break;
      }
      mask_ <<= 1;
    }
    // The root node needs to set up for the first send.
    if(src_ < 0) {
      mask_ >>= 1;
      while(mask_ > 0) {
        if(relative_rank + mask_ < size_) {
          dst_ = rank_ + mask_;
          if(dst_ >= size_)
            dst_ -= size_;
          break;
        }
        mask_ >>= 1;
      }
    }
  }
  
  //
  // Hello.
  //
  btscatteriterator::ptr
  btscatteriterator::construct(int rank, int size, int root) {
    return btscatteriterator::ptr(new btscatteriterator(rank,size,root));
  }

  //
  // Goodbye.
  //
  btscatteriterator::~btscatteriterator() throw() {
  }

  //
  // Increment the iterator.
  //
  bool btscatteriterator::next() {
  //  if(debugger::active())
  //    debug() << "btscatteriterator[rank="<<rank_<<",size="<<size_<<",root="
   //           <<root_<<"]::next()\n";
    if(! this->active()) {
      throw iteratorerror("btscatteriterator::next(): Iterator is not active.");
    }
    src_ = -1;
    int relative_rank = (rank_ >= root_ ? rank_-root_ : rank_-root_+size_);
    mask_ >>= 1;
    while(mask_ > 0) {
      if(relative_rank + mask_ < size_) {
        dst_ = rank_ + mask_;
        if(dst_ >= size_)
          dst_ -= size_;
        break;
      }
      mask_ >>= 1;
    }
    return this->active();
  }

  //
  // Test whether this iterator is still active (has a valid neighbor).
  //
  bool btscatteriterator::active() const {
    return (0 < mask_ && mask_ < size_);
  }

  //
  // Are we sending?
  //
  bool btscatteriterator::sending() const {
    return (this->active() && dst_ >= 0);
  }

  //
  // Are we receiving?
  //
  bool btscatteriterator::receiving() const {
    return (this->active() && src_ >= 0);
  }

  //
  // Return the index of the node we are receiving from.
  //
  int btscatteriterator::source() const {
  //  if(debugger::active())
  //    debug() << "btscatteriterator[rank="<<rank_<<",size="<<size_<<",root="
  //            <<root_<<"]::source()\n";
    if(! this->receiving()) {
      throw iteratorerror("btscatteriterator::source():  Not receiving.");
    }
 //   if(debugger::active())
 //     debug() << "btscatteriterator::source:  Returning " << src_ << "\n";
    return src_;
  }

  //
  // Return the index of the node we are sending to.
  //
  int btscatteriterator::dest() const {
  //  if(debugger::active())
  //    debug() << "btscatteriterator[rank="<<rank_<<",size="<<size_<<",root="
   //           <<root_<<"]::dest()\n";
    if(! this->sending()) {
      throw iteratorerror("btscatteriterator::dest():  Not sending.");
    }
  //  if(debugger::active())
  //    debug() << "btscatteriterator::dest:  Returning " << dst_ << "\n";
    return dst_;
  }

} } // end of namespace sstmac
