/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/api/mpi/mpi_strategies/mpi_iterator/hfmiterator.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>

namespace sstmac { namespace sw {

  // Pointer types.
  

  //
  // Hello.
  //
  hfmiterator::hfmiterator(int rank, int size) :
    mask_(0), rank_(rank), size_(size), source_(-1), dest_(-1)
  {
    const std::string here("hfmiterator(...)");
    if(rank_ < 0 || rank >= size_)
      throw valueerror(here + ":  Invalid rank.");
    if(size_ < 0)
      throw valueerror(here + ":  Invalid size.");
    if(this->active())
      this->next();
  }

  //
  // Construct a new iterator.
  //
  hfmiterator::ptr hfmiterator::construct(int rank, int size) {
    return hfmiterator::ptr(new hfmiterator(rank, size));
  }

  //
  // Goodbye.
  //
  hfmiterator::~hfmiterator() throw() {
  }

  //
  // Increment the iterator.
  //
  bool hfmiterator::next() {
    if(! this->active())
      throw iteratorerror("hfmiterator::next():  Iterator is not active.");
    // Go.
    if(mask_ > 0) {
      mask_ <<= 1;
    }
    else {
      // First entry.
      mask_ = 1;
    }
    dest_ = (rank_ + mask_) % size_;
    source_ = (rank_ - mask_ + size_) % size_;
    return this->active();
  }

  //
  // Test whether this iterator is still active (has valid neighbors).
  //
  bool hfmiterator::active() const {
    return (mask_ ? 
            (mask_ > 0 && size_ > 0 && mask_ < size_ &&
             dest_ >= 0 && dest_ < size_ &&
             source_ >= 0 && source_ < size_) :
            true);
  }

  //
  // Return the index of the node we are receiving from.
  //
  int hfmiterator::source() const {
    if(! this->active()) {
      throw iteratorerror("hfmiterator::source():  Iterator is not active.");
    }
    return source_;
  }

  //
  // Return the index of the node we are sending to.
  //
  int hfmiterator::dest() const {
    if(! this->active()) {
      throw iteratorerror("hfmiterator::dest():  Iterator is not active.");
    }
    return dest_;
  }


} } // end of namespace sstmac
