/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/api/mpi/mpi_strategies/mpistrategy.h>
#include <sstmac/software/api/mpi/mpi_strategies/mpicorestrategies.h>
#include <sstmac/common/errors.h>

namespace sstmac { namespace sw {

  

  // Private test for whether a shared_ptr is null.
  template<typename T>
  void nonnull(const boost::intrusive_ptr<T> &ptr, const char *caller) {
    if(ptr == 0) {
      throw nullerror(std::string(caller) + ":  null strategy object.");
    }
  }

  //
  // Default construction.  Builds default implementations for
  // all the individual strategy types.
  //
  mpistrategy::mpistrategy() :
    bcast_(mpicorebcast::construct()),
    reduce_(mpicorereduce::construct()),
    allreduce_(mpicoreallreduce::construct()),
    gather_(mpicoregather::construct()),
    gatherv_(mpicoregatherv::construct()),
    allgather_(mpicoreallgather::construct()),
    allgatherv_(mpicoreallgatherv::construct()),
    alltoall_(mpicorealltoall::construct()),
    alltoallv_(mpicorealltoallv::construct()),
    scatter_(mpicorescatter::construct()),
    scatterv_(mpicorescatterv::construct()),
    reduce_scatter_(mpicorereduce_scatter::construct()),
    barrier_(mpicorebarrier::construct(mpicorebarrier::HFM)),
    scan_(mpicorescan::construct())
  {
    // Make the sends and receives shared by default.
    send_ = bsend_ = mpicoresend::construct();
    rsend_ = mpicorersend::construct();
    ssend_ = mpicoressend::construct();
    recv_ = mpicorerecv::construct();
  }

  //
  // Static build method.
  //
  mpistrategy::ptr mpistrategy::construct() {
    return mpistrategy::ptr(new mpistrategy);
  }

  // Assign a non-null strategy to send.
  void mpistrategy::set_send(const mpisendstrategy::ptr& st) {
    nonnull(st, "mpistrategy::set_send");
    send_ = st;
  }

  /// Assign a non-null strategy to bsend.
  void mpistrategy::set_bsend(const mpisendstrategy::ptr& st) {
    nonnull(st, "mpistrategy::set_bsend");
    bsend_ = st;
  }

  /// Assign a non-null strategy to ssend.
  void mpistrategy::set_ssend(const mpisendstrategy::ptr& st) {
    nonnull(st, "mpistrategy::set_ssend");
    ssend_ = st;
  }

  /// Assign a non-null strategy to rsend.
  void mpistrategy::set_rsend(const mpisendstrategy::ptr& st) {
    nonnull(st, "mpistrategy::set_rsend");
    rsend_ = st;
  }

  /// Assign a non-null strategy to recv.
  void mpistrategy::set_recv(const mpirecvstrategy::ptr& st) {
    nonnull(st, "mpistrategy::set_recv");
    recv_ = st;
  }

  // Assign a non-null strategy to bcast.
  void mpistrategy::set_bcast(const mpibcaststrategy::ptr& st) {
    nonnull(st, "mpistrategy::set_bcast");
    bcast_ = st;
  }

  // Assign a non-null strategy to reduce.
  void mpistrategy::set_reduce(const mpireducestrategy::ptr& st) {
    nonnull(st, "mpistrategy::set_reduce");
    reduce_ = st;
  }

  // Assign a non-null strategy to allreduce.
  void mpistrategy::set_allreduce(const mpiallreducestrategy::ptr& st) {
    nonnull(st, "mpistrategy::set_allreduce");
    allreduce_ = st;
  }

  // Assign a non-null strategy to gather.
  void mpistrategy::set_gather(const mpigatherstrategy::ptr& st) {
    nonnull(st, "mpistrategy::set_gather");
    gather_ = st;
  }

  // Assign a non-null strategy to gatherv.
  void mpistrategy::set_gatherv(const mpigathervstrategy::ptr& st) {
    nonnull(st, "mpistrategy::set_gatherv");
    gatherv_ = st;
  }

  // Assign a non-null strategy to allgather.
  void mpistrategy::set_allgather(const mpiallgatherstrategy::ptr& st) {
    nonnull(st, "mpistrategy::set_allgather");
    allgather_ = st;
  }

  // Assign a non-null strategy to allgatherv.
  void mpistrategy::set_allgatherv(const mpiallgathervstrategy::ptr& st) {
    nonnull(st, "mpistrategy::set_allgatherv");
    allgatherv_ = st;
  }

  // Assign a non-null strategy to alltoall.
  void mpistrategy::set_alltoall(const mpialltoallstrategy::ptr& st) {
    nonnull(st, "mpistrategy::set_alltoall");
    alltoall_ = st;
  }

  // Assign a non-null strategy to alltoallv.
  void mpistrategy::set_alltoallv(const mpialltoallvstrategy::ptr& st) {
    nonnull(st, "mpistrategy::set_alltoallv");
    alltoallv_ = st;
  }

  // Assign a non-null strategy to scatter.
  void mpistrategy::set_scatter(const mpiscatterstrategy::ptr& st) {
    nonnull(st, "mpistrategy::set_scatter");
    scatter_ = st;
  }

  // Assign a non-null strategy to scatterv.
  void mpistrategy::set_scatterv(const mpiscattervstrategy::ptr& st) {
    nonnull(st, "mpistrategy::set_scatterv");
    scatterv_ = st;
  }

  // Assign a non-null strategy to reduce_scatter.
  void mpistrategy::set_reduce_scatter(const mpireduce_scatterstrategy::ptr& st) {
    nonnull(st, "mpistrategy::set_reduce_scatter");
    reduce_scatter_ = st;
  }

  // Assign a non-null strategy to synchronization barriers.
  void mpistrategy::set_barrier(const mpibarrierstrategy::ptr& st) {
    nonnull(st, "mpistrategy::set_barrier");
    barrier_ = st;
  }

  // Assign a non-null strategy to scan operations.
  void mpistrategy::set_scan(const mpiscanstrategy::ptr& st) {
    nonnull(st, "mpistrategy::set_scan");
    scan_ = st;
  }

} } // end of namespace sstmac
