/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_STRATEGIES_MPISTRATEGY_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_STRATEGIES_MPISTRATEGY_H_INCLUDED

#include <sstmac/software/api/mpi/mpi_strategies/mpistrategies.h>
#include <sstmac/common/ptr_type.h>

namespace sstmac { namespace sw {

  /**
   * A full ensemble of strategies for an MPI communication environment.
   * All the individual strategies have (hopefully reasonable) defaults.
   *
   * All operations are preseumed to be non-blocking.  The choice of whether
   * (or when) to block is left to the caller.
   */
  class mpistrategy : public ptr_type {
  private:
    /// The strategy objects for non-blocking sends.
    mpisendstrategy::ptr send_, bsend_, ssend_, rsend_;
    /// The strategy object for blocking receives.
    mpirecvstrategy::ptr recv_;
    /// The strategy object for broadcasts.
    mpibcaststrategy::ptr bcast_;
    /// The strategy object for reduce operations.
    mpireducestrategy::ptr reduce_;
    /// The strategy object for allreduce.  Should this be fused with reduce?
    mpiallreducestrategy::ptr allreduce_;
    /// The strategy object for gather operations.
    mpigatherstrategy::ptr gather_;
    /// The strategy object for gatherv operations.
    mpigathervstrategy::ptr gatherv_;
    /// The strategy object for allgather operations.
    mpiallgatherstrategy::ptr allgather_;
    /// The strategy object for allgatherv operations.
    mpiallgathervstrategy::ptr allgatherv_;
    /// The strategy object for alltoall operations.
    mpialltoallstrategy::ptr alltoall_;
    /// The strategy object for alltoallv operations.
    mpialltoallvstrategy::ptr alltoallv_;
    /// The strategy object for scatter operations.
    mpiscatterstrategy::ptr scatter_;
    /// The strategy object for scatterv operations.
    mpiscattervstrategy::ptr scatterv_;
    /// The strategy object for fused reduce_scatter operations.
    mpireduce_scatterstrategy::ptr reduce_scatter_;
    // The strategy object for synchronization barriers.
    mpibarrierstrategy::ptr barrier_;
    /// The strategy object for scan operations.
    mpiscanstrategy::ptr scan_;

  protected:
    /// Default construction.  Builds default implementations for
    /// all the individual strategy types.
    mpistrategy();
    
  public:
    virtual std::string to_string() const{
            return "mpistrategy";
          }

    typedef boost::intrusive_ptr<mpistrategy> ptr;

    /// Build me an object.
    static mpistrategy::ptr construct();

    /// The strategy objects for sends.
    /// The set_* methods throw mpistrategy_error if the strategy is NULL.
    mpisendstrategy::ptr send() { return send_; }
    const mpisendstrategy::ptr send() const { return send_; }
    void set_send(const mpisendstrategy::ptr& st);
    mpisendstrategy::ptr bsend() { return bsend_; }
    const mpisendstrategy::ptr bsend() const { return bsend_; }
    void set_bsend(const mpisendstrategy::ptr&);
    mpisendstrategy::ptr ssend() { return ssend_; }
    const mpisendstrategy::ptr ssend() const { return ssend_; }
    void set_ssend(const mpisendstrategy::ptr& st);
    mpisendstrategy::ptr rsend() { return rsend_; }
    const mpisendstrategy::ptr rsend() const { return rsend_; }
    void set_rsend(const mpisendstrategy::ptr& st);

    /// The strategy object for receives.
    mpirecvstrategy::ptr recv() { return recv_; }
    const mpirecvstrategy::ptr recv() const { return recv_; }
    void set_recv(const mpirecvstrategy::ptr& st);

    /// The strategy object for broadcasts.
    mpibcaststrategy::ptr bcast() { return bcast_; }
    const mpibcaststrategy::ptr bcast() const { return bcast_; }
    void set_bcast(const mpibcaststrategy::ptr& st);

    /// The strategy object for reduce operations.
    mpireducestrategy::ptr reduce() { return reduce_; }
    const mpireducestrategy::ptr reduce() const {
      return reduce_;
    }
    void set_reduce(const mpireducestrategy::ptr& st);

    /// The strategy object for allreduce.
    mpiallreducestrategy::ptr allreduce() { return allreduce_; }
    const mpiallreducestrategy::ptr allreduce() const {
      return allreduce_;
    }
    void set_allreduce(const mpiallreducestrategy::ptr& st);

    /// The strategy object for gather operations.
    mpigatherstrategy::ptr gather() { return gather_; }
    const mpigatherstrategy::ptr gather() const {return gather_;}
    void set_gather(const mpigatherstrategy::ptr& st);

    /// The strategy object for gatherv operations.
    mpigathervstrategy::ptr gatherv() { return gatherv_; }
    const mpigathervstrategy::ptr gatherv() const {
      return gatherv_;
    }
    void set_gatherv(const mpigathervstrategy::ptr& st);

    /// The strategy object for allgather operations.
    mpiallgatherstrategy::ptr allgather() { return allgather_; }
    const mpiallgatherstrategy::ptr allgather() const {
      return allgather_;
    }
    void set_allgather(const mpiallgatherstrategy::ptr& st);

    /// The strategy object for allgatherv operations.
    mpiallgathervstrategy::ptr allgatherv() {return allgatherv_;}
    const mpiallgathervstrategy::ptr allgatherv() const {
      return allgatherv_;
    }
    void set_allgatherv(const mpiallgathervstrategy::ptr& st);

    /// The strategy object for alltoall operations.
    mpialltoallstrategy::ptr alltoall() { return alltoall_; }
    const mpialltoallstrategy::ptr alltoall() const {
      return alltoall_;
    }
    void set_alltoall(const mpialltoallstrategy::ptr& st);

    /// The strategy object for alltoallv operations.
    mpialltoallvstrategy::ptr alltoallv() { return alltoallv_; }
    const mpialltoallvstrategy::ptr alltoallv() const {
      return alltoallv_;
    }
    void set_alltoallv(const mpialltoallvstrategy::ptr& st);

    /// The strategy object for scatter operations.
    mpiscatterstrategy::ptr scatter() {
      return scatter_;
    }
    const mpiscatterstrategy::ptr scatter() const {
      return scatter_;
    }
    void set_scatter(const mpiscatterstrategy::ptr& st);

    /// The strategy object for scatterv operations.
    mpiscattervstrategy::ptr scatterv() {
      return scatterv_;
    }
    const mpiscattervstrategy::ptr scatterv() const {
      return scatterv_;
    }
    void set_scatterv(const mpiscattervstrategy::ptr& st);

    /// The strategy object for fused reduce_scatter operations.
    mpireduce_scatterstrategy::ptr reduce_scatter() {
      return reduce_scatter_;
    }
    const mpireduce_scatterstrategy::ptr reduce_scatter() const {
      return reduce_scatter_;
    }
    void set_reduce_scatter(const mpireduce_scatterstrategy::ptr& st);

    /// The strategy object for synchronization barriers.
    mpibarrierstrategy::ptr barrier() { return barrier_; }
    const mpibarrierstrategy::ptr barrier() const {
      return barrier_;
    }
    void set_barrier(const mpibarrierstrategy::ptr& st);

    /// Thestrategy object for scan operations.
    mpiscanstrategy::ptr scan() { return scan_; }
    const mpiscanstrategy::ptr scan() const { return scan_; }
    void set_scan(const mpiscanstrategy::ptr& st);
  };
} } // end of namespace sstmac.

#endif
