/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_TYPES_MPITYPE_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_TYPES_MPITYPE_H_INCLUDED

#include <sstmac/common/types.h>
#include <iosfwd>
#include <sstmac/common/ptr_type.h>
#include <sstmac/common/messages/sst_serializer.h>
#include <sstmac/common/messages/serializable.h>
namespace sstmac
{
  namespace sw
  {

    class pairdata;
    class vecdata;
    class inddata;

    /// MPI datatypes.
    class mpitype
    {
    public:
      enum TYPE_TYPE
      {
        PRIM, PAIR, VEC, IND, NONE
      };

      mpitype();

      //primitive datatype
          mpitype(const std::string &labelit, const int sizeit, int align,
              int comb);

      //pair of primitives datatype
      mpitype(const std::string &labelit, const mpitype &b1, const mpitype &b2,
          int size, int comb);

      //vector datatype
      mpitype(const std::string &labelit, const mpitype &base, int count,
          int block, int str, bool in_elem, int comb);

      //indexed datatype
      mpitype(const std::string &labelit, boost::intrusive_ptr<inddata> dat,
          int sz, int ext, int numint, int numdtype, int comb);

      mpitype(const mpitype &other);

      void
      deep_copy(const mpitype& other);

      // id gets assigned automatically by the constructor.
      unsigned long id;
      std::string label;
      mutable long native_id; //what the C interface thinks it is

    protected:

      TYPE_TYPE type_;

      int alignment_; //required for primitive types

      boost::intrusive_ptr<pairdata> pdata_;
      boost::intrusive_ptr<vecdata> vdata_;
      boost::intrusive_ptr<inddata> idata_;

      int size; //this is the packed size !!!
      size_t extent_; //holds the extent, as defined by the MPI standard

      //stuff for get_envelope
      int combiner_;
      int numints_;
      int numaddr_;
      int numdtype_;

    public:

      void
      serialize_order(sst_serializer* ser);

      virtual
      ~mpitype();

      mpitype
      clone() const;

      TYPE_TYPE
      get_type() const
      {
        return type_;
      }

      int
      get_comb() const
      {
        return combiner_;
      }
      void
      set_comb(int c)
      {
        combiner_ = c;
      }

      int
      get_align() const
      {
        return alignment_;
      }

      int
      bytes_to_elements(size_t bytes) const;

      int
      get_packed_size() const;

      //returns the marker if available, true lb if not
      int
      get_lb() const;

      //disregards the marker, returns lowest displacement
      int
      get_true_lb() const;

      //returns the marker if available, true ub if not
      int
      get_ub() const;

      //returns the difference between markers if available, true extent if not
      int
      get_extent() const;

      //returns the actual difference between min and max displacements (ignore markers)
      int
      get_true_extent() const;

      void
      get_envelope(int* numint, int* numaddr, int* numdtype, int* comb);

      void
      get_contents(int* ints, int* addr, long* dtypes);

      virtual void
      pack(const void *inbuf, int incount, void *outbuf, int outsize,
          int *position) const;

      virtual void
      unpack(const void *inbuf, int insize, void *outbuf, int outcount,
          int *position) const;

      virtual std::string
      to_string() const;

      static const mpitype mpi_null;
      static const mpitype mpi_char;
      static const mpitype mpi_unsigned_char;
      static const mpitype mpi_signed_char;
      static const mpitype mpi_wchar;
      static const mpitype mpi_unsigned_long_long;
      static const mpitype mpi_lb;
      static const mpitype mpi_ub;
      static const mpitype mpi_byte;
      static const mpitype mpi_short;
      static const mpitype mpi_unsigned_short;
      static const mpitype mpi_int;
      static const mpitype mpi_unsigned;
      static const mpitype mpi_long;
      static const mpitype mpi_unsigned_long;
      static const mpitype mpi_float;
      static const mpitype mpi_double;
      static const mpitype mpi_long_double;
      static const mpitype mpi_long_long_int;
      static const mpitype mpi_long_long;
      static const mpitype mpi_packed;
      static const mpitype mpi_float_int;
      static const mpitype mpi_double_int;
      static const mpitype mpi_long_int;
      static const mpitype mpi_short_int;
      static const mpitype mpi_2int;
      static const mpitype mpi_long_double_int;
      static const mpitype mpi_complex;
      static const mpitype mpi_complex8;
      static const mpitype mpi_complex16;
      static const mpitype mpi_complex32;
      static const mpitype mpi_double_complex;
      static const mpitype mpi_logical;
      static const mpitype mpi_real;
      static const mpitype mpi_real4;
      static const mpitype mpi_real8;
      static const mpitype mpi_real16;
      static const mpitype mpi_double_precision;
      static const mpitype mpi_integer;
      static const mpitype mpi_integer1;
      static const mpitype mpi_integer2;
      static const mpitype mpi_integer4;
      static const mpitype mpi_integer8;
      static const mpitype mpi_2integer;
      static const mpitype mpi_2complex;
      static const mpitype mpi_2double_complex;
      static const mpitype mpi_2real;
      static const mpitype mpi_2double_precision;
      static const mpitype mpi_character;

      mpitype&
      operator=(const mpitype& other);
    };

    class pairdata : public serializable, public serializable_type<pairdata>
    {
    ImplementSerializable(pairdata)
    protected:
      pairdata()
      {
      }
    public:
      typedef boost::intrusive_ptr<pairdata> ptr;
      static ptr
      construct()
      {
        return ptr(new pairdata());
      }

      virtual
      ~pairdata()
      {
      }
      virtual std::string
      to_string() const
      {
        return "pairdata";
      }
      mpitype base1_;
      mpitype base2_;

      void
      serialize_order(sst_serializer* ser);
    };

    class vecdata :  public serializable, public serializable_type<
        vecdata>
    {
    ImplementSerializable(vecdata)
    protected:
      vecdata()
      {
      }
    public:
      typedef boost::intrusive_ptr<vecdata> ptr;
      static ptr
      construct()
      {
        return ptr(new vecdata());
      }

      virtual
      ~vecdata()
      {
      }
      virtual std::string
      to_string() const
      {
        return "vecdata";
      }
      mpitype base_;
      int count_;
      int blocklen_;
      int stride_; //always in bytes!!!!

      void
      serialize_order(sst_serializer* ser);
    };

    struct ind_block
    {
      mpitype base;
      int disp; ///always in bytes!!!!
      int num;
    };

    struct inddata :  public serializable, public serializable_type<
        inddata>
    {
    ImplementSerializable(inddata)
    protected:
      inddata()
      {
      }
    public:
      typedef boost::intrusive_ptr<inddata> ptr;
      static ptr
      construct()
      {
        return ptr(new inddata());
      }

      virtual
      ~inddata()
      {
      }
      virtual std::string
      to_string() const
      {
        return "inddata";
      }
      // int maxdisp_; //in bytes
      //  int maxblockind_;
      int mindisp_; //in bytes, true lb
      int maxbyte_; //top byte, true ub
      int ub_;
      int lb_;
      boost::unordered_map<int, ind_block> blocks;

      void
      serialize_order(sst_serializer* ser);
    };

    inline bool
    operator<(const mpitype &a, const mpitype &b)
    {
      return (a.id < b.id);
    }
    bool
    operator==(const mpitype &a, const mpitype &b);
    bool
    operator!=(const mpitype &a, const mpitype &b);
    std::ostream&
    operator<<(std::ostream &os, const mpitype &type);

  }

} // end of namespace sstmac

#endif
