/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstream>

#include <sstmac/software/api/mpi/mpi_queue/mpiqueue.h>
#include <sstmac/software/api/mpi/mpi_queue/userthread-mpiqueue.h>
#include <sstmac/software/api/mpi/mpi_queue/servicethread-mpiqueue.h>

#include <sstmac/software/api/mpi/mpiapi.h>
#include <sstmac/software/api/mpi/mpiapi-persistent.h>
#include <sstmac/software/api/mpi/mpistatus.h>
#include <sstmac/software/api/mpi/mpitracer.h>
#include <sstmac/software/api/mpi/mpirequest.h>
#include <sstmac/software/api/mpi/mpi_strategies/mpi_collective/mpicollpayload.h>
#include <sstmac/software/api/mpi/sstmac_mpi.h>
#include <sstmac/hardware/node/node.h>

#include <sstmac/software/process/graphviz.h>

#include <sstmac/software/process/os_api.h>
#include <sstmac/software/api/mpi/mpi_strategies/mpistrategy.h>
#include <sstmac/software/api/mpi/mpi_strategies/mpicoreallreduce.h>
#include <sstmac/software/api/mpi/mpi_comm/mpicommfactory.h>
#include <sstmac/software/api/mpi/mpi_comm/mpicomm_cart.h>

#include <sstmac/software/services/launch/instantlaunch.h>

#include <sstmac/software/api/mpi/mpiserver.h>
#include <sstmac/common/errors.h>

#include <sstmac/software/libraries/unblockhandler.h>

#include <sstmac/software/api/mpi/dumpitracer.h>
#include <sstmac/software/process/operatingsystem.h>
#include <sstmac/software/process/api.h>

#include <sstmac/hardware/network/topology/regular_topology.h>

#include <sstmac/software/api/mpi/mpitypes.h>

#include <time.h>

#include <boost/lexical_cast.hpp>

ImplementAPI(sstmac::sw, mpiapi, "mpi")
;

namespace sstmac
{
  namespace sw
  {
    SSTRegister("mpi", api, mpiapi, "Create bindings for MPI runtime");

    key::category mpiapi::default_key_category("MPI");
    key::category mpiapi::poll_key_category("MPI Poll");
    key::category mpiapi::memcpy_key_category("MPI Memcpy");

    // Pointer types.

    using boost::dynamic_pointer_cast;
    using boost::lexical_cast;

    debug mpiapi::dbg_;

    boost::unordered_map<int64_t, std::string> mpiapi::hostmap_;

    std::string mpiapi::tracername_ = "dumpitracer";
    mpitracer::ptr mpiapi::parent_tracer_ = 0;

    mpiapi::ptr
    sstmac_mpi()
    {
      thread::ptr t = OS_API::get_current_thread();
      return t->get_api<mpiapi> ();
    }

    bool
    mpiapi::tracer_active() const
    {
      return tracer_ && tracer_->is_active();
    }

    // Catch ugly null pointers.
    template<typename T>
      void
      nonnull(const boost::intrusive_ptr<T> &p, const char *caller)
      {
        if (p == 0)
          throw nullerror(caller);
      }

    //
    // Catch calls on an un-initialized MPI environment.
    //
    inline void
    mpiapi::assert_initialized(const char *caller) const
    {
      if (status_ != is_initialized)
      {
        throw illformederror(
            std::string(caller) + ":  mpiapi is not active (initialized).");
      }
    }

    // Test whether a communicator is valid for sending or receiving.
    inline void
    validcomm(const mpicomm::ptr &comm, const char *here)
    {
      if (comm == mpicomm::comm_null)
        throw valueerror(
            std::string(here) + ": Can't communicate on MPI_COMM_NULL.");
    }

    // Test whether we are sending a valid number of elements.
    inline void
    validcount(int count, const char *here)
    {
      if (count < 0)
        throw valueerror(std::string(here) + ": Negative array length.");
    }

    // Test whether the send type is acceptable.
    inline void
    validtype(const mpitype& type, const char *here)
    {
      if (type.get_packed_size() <= 0)
        throw valueerror(
            std::string(here) + ":  MPI type \"" + type.label
                + "\" does not have size greater than zero.");
    }

    // Test whether the given id is valid for sending.
    inline void
    validsendid(const mpiid &target, const mpicomm::ptr &comm, const char *here)
    {
      if (target.id < 0 || target >= comm->size())
        throw valueerror(
            std::string(here) + ": id: " + comm->rank().to_string()
                + ": Invalid destination rank " + lexical_cast<std::string> (
                target));
    }

    // Test whether the given id is valid for receiving.
    inline void
    validrecvid(const mpiid &source, const mpicomm::ptr &comm, const char *here)
    {
      if ((source.id < 0 || source >= comm->size()) && (source
          != mpiid::any_source))
      {
        std::cerr << "Failing on source rank" << source << "\n";
        throw valueerror(
            std::string(here) + ": Invalid source rank" + lexical_cast<
                std::string> (source));
      }
    }

    // Test wheter the given tag is valid for sending.
    inline void
    validsendtag(const mpitag &tag, const char *here)
    {
      if (tag.id < 0)
        throw valueerror(std::string(here) + ": Send-tag must be >= 0.");
    }

    // Test wheter the given tag is valid for receiving.
    inline void
    validrecvtag(const mpitag &tag, const char *here)
    {
      if ((tag.id < 0) && (tag != mpitag::any_tag))
        throw valueerror(
            std::string(here)
                + ": Receive-tag must be >= 0 or mpitag::any_tag.");
    }

    // Test paramters for sending a single element.
    inline void
    validsend(int count, const mpitype& sendtype, const mpiid &target,
        const mpitag &tag, const mpicomm::ptr &comm, const char *here)
    {
      validcount(count, here);
      // validtype(sendtype, here);
      validcomm(comm, here);
      validsendid(target, comm, here);
      validsendtag(tag, here);
    }

    // Test paramters for receiving a single element.
    inline void
    validrecv(int count, const mpitype& recvtype, const mpiid &source,
        const mpitag &tag, const mpicomm::ptr &comm, const char *here)
    {
      validcount(count, here);
      // validtype(recvtype, here);
      validcomm(comm, here);
      validrecvid(source, comm, here);
      validrecvtag(tag, here);
    }

    // Test paramters for sending a vector of elements.
    inline void
    validsends(const std::vector<int> &counts, const mpitype& sendtype,
        const mpiid &target, const mpitag &tag, const mpicomm::ptr &comm,
        const char *here)
    {
      std::vector<int>::const_iterator it, end = counts.end();
      for (it = counts.begin(); it != end; ++it)
        validcount(*it, here);
      // validtype(sendtype, here);
      validcomm(comm, here);
      validsendid(target, comm, here);
      validsendtag(tag, here);
    }

    // Test paramters for receiving a vector of elements.
    inline void
    validrecvs(const std::vector<int> &counts, const mpitype& recvtype,
        const mpiid &source, const mpitag &tag, const mpicomm::ptr &comm,
        const char *here)
    {
      std::vector<int>::const_iterator it, end = counts.end();
      for (it = counts.begin(); it != end; ++it)
        validcount(*it, here);
      //  validtype(recvtype, here);
      validcomm(comm, here);
      validrecvid(source, comm, here);
      validrecvtag(tag, here);
    }

    // Test arguments for a broadcast.
    inline void
    validbcast(int count, const mpitype &sendtype, const mpiid &root,
        const mpitag &tag, const mpicomm::ptr &comm, const char *here)
    {
      validcount(count, here);
      //   validtype(sendtype, here);
      validcomm(comm, here);
      validsendid(root, comm, here);
      validsendtag(tag, here);
    }

    // Test whether all the given ranks are valid in a communicator.
    inline void
    validranks(const std::vector<mpiid> &rank, const mpicomm::ptr &comm,
        const char *here)
    {
      const int ub = comm->size().id;
      std::vector<mpiid>::const_iterator it, end = rank.end();
      for (it = rank.begin(); it != end; ++it)
      {
        if (it->id < 0 || it->id >= ub)
        {
          throw valueerror(std::string(here) + ":  Invalid rank found.");
        }
      }
    }

    //
    // Build a new mpiapi.
    //
    mpiapi::mpiapi(const sim_parameters::ptr& params) :
      api(mpiapi::default_key_category),
      saved_payload_(payload::null()),
      status_(is_fresh)
    {
      if (!dbg_.is_real())
      {
        dbg_ = debug("<debug> mpi | mpiapi");
        logger::register_logger(&dbg_);
      }

      if (tracername_ == "")
      {
        std::stringstream timestr;
        time_t t = time(NULL);
        timestr << "dumpi-trace-";
        timestr << t;
        tracername_ = timestr.str();
      }

      if (!parent_tracer_)
      {
        parent_tracer_ = dumpitracer::construct(tracername_, "localhost"); //pass a null hostname
      }
    }

    void
    mpiapi::init_factory_params(const sim_parameters::ptr &params)
    {
      // Turn payloads on/off
      if (params->has_param("MPI_buffer_ignore"))
      {
        /**
            sstkeyword {
                gui=true;
                docstring=Whether to emulate payloads in the simulator.ENDL
                If false, send/recv pairs do not actually send/recv buffer data.ENDL
                Only time is simulated, making simulation much faster.ENDL
                If true,  send/recv pairs actually deliver the correct buffer.ENDL
                This makes the simulation much slower and uses A LOT of memory.;
            }
        */
        bool ignore = params->get_bool_param("MPI_buffer_ignore");
        if (ignore)
          MPI_Disable_Payloads();
        else
          MPI_Enable_Payloads();
      }

      strategy_ = mpistrategy::construct();
      /**
        sstkeyword {
            gui=rabenseifner;
            docstring=The algorithm to be used for performing all reduce.;
            rabenseifner=The standard MPI rabenseifner, halving reduce.;
            wilke=Modification of Rabenseifner.  Creates virtual 2^N network to improve performance
                  for non-powers of 2.;
        }
      */
      std::string allreduce_type = params->get_optional_param("mpi_allreduce",
          "rabenseifner");
      if (allreduce_type == "wilke")
      {
        strategy_->set_allreduce(mpiwilkeallreduce::construct());
      }
      api::init_factory_params(params);
    }

    void
    mpiapi::init(const softwareid& id)
    {
      api::init(id);
      process_manager::init(id);
      id_ = id;
      rank_ = mpiid(id.task_.id);

      if (id_.app_.id > 0)
      {
        SSTMAC_DEBUG << "constructed mpiapi with rank " << rank_.id
              << " for app " << id_.app_.id << "\n";
      }
      libname_ = "mpiapi" + id.to_string();
    }

    void
    mpiapi::init_os(const OS_API::ptr &os)
    {
      api::init_os(os);
      process_manager::init_os(os);
    }

    void
    mpiapi::clear_statics()
    {
      parent_tracer_ = 0;
    }

    //
    // Build a new object.
    //
    mpiapi::ptr
    mpiapi::construct(const sim_parameters::ptr& params)
    {
      return new mpiapi(params);
    }

    //
    // Adieu.
    //
    mpiapi::~mpiapi()
    {
    }

    //
    // Get the world.
    //
    const mpicomm::ptr
    mpiapi::comm_world() const
    {
      assert_initialized("MPI");
      return worldcomm_;
    }

    //
    // Get self.
    //
    const mpicomm::ptr
    mpiapi::comm_self() const
    {
      return selfcomm_;
    }

    void
    mpiapi::attr_put(mpicomm::ptr comm, keyval::ptr k, void* val)
    {
      comm->set_keyval(k, val);
    }

    void
    mpiapi::attr_get(mpicomm::ptr comm, keyval::ptr k, void* val, int* flag)
    {
      comm->get_keyval(k, val, flag);
    }

    /*
     * Methods exposing MPI calls and other key methods to
     * derived objects.
     */

    //
    // Get count.
    //
    int
    mpiapi::get_count(const_mpistatus_t status, const mpitype &type)
    {
      assert_initialized("mpiapi::get_count");
      nonnull(status, "mpiapi::get_count:  Null status pointer.");
      int count = status->bytes_received() / type.get_packed_size();
      if (tracer_active())
        tracer_->get_count(os_->now(), os_->now(), status, type, count);
      return count;
    }

    /* Set up and tear down */

    //
    // Test whether MPI has bee initialized.
    //
    bool
    mpiapi::initialized()
    {
      if (tracer_active())
        tracer_->initialized(os_->now(), os_->now(), status_ == is_initialized);
      return (status_ == is_initialized);
    }

    //
    // Test whether MPI has been finalized.
    //
    bool
    mpiapi::finalized()
    {
      if (tracer_active())
        tracer_->finalized(os_->now(), os_->now(), status_ == is_finalized);
      return (status_ == is_finalized);
    }

    void
    mpiapi::unregister()
    {
      //must get libname first! queue->unregister clears server
      const std::string& libname = queue_->server()->get_lib_name();

      queue_->unregister(id_);

      library::ptr lib = os_->get_lib(libname);
      if (lib)
      {

        mpiserver::ptr mpiserv = boost::dynamic_pointer_cast<mpiserver>(lib);
        mpiserv->unregister(id_);

        /** These are now handled by unregister conditionally
         only if all mpiqueues have been unregistered
         mpiserv->unregister_all_libs();
         os_->unregister_lib(mpiserv);
         */
      }
      else if (queue_)
      {
        sst_throw(illformederror, "mpiapi::unregister has non-null mpiqueue, but a null mpiserver");
      }
      queue_ = 0;
    }

    //
    // Initialize MPI.
    //
    timestamp
    mpiapi::init()
    {
      SSTMACBacktrace("MPI_Init");
      start_api_call();

      if (!os_)
      {
        throw nullerror("mpiapi::init - os has not been initialized yet");
      }

      environment::const_ptr env = os_->get_env(id_.app_);
      if (!env)
      {
        throw nullerror("mpiapi::init - no environment found");
      }

      std::string server_libname = sstprintf("mpiserver-%d", id_.app_.id);
      library::ptr mpilib = os_->get_lib(server_libname);

      if (mpilib == library::ptr()) // only do one mpiserver per app per node
      {
        mpiserver::ptr mpiserv = mpiserver::construct(id_.app_.id);
        mpilib = mpiserv;
        os_->register_ser(mpiserv);
        mpiserv->start();
      }
      mpiserver::ptr mpiserv = boost::dynamic_pointer_cast<mpiserver>(mpilib);

      /**
        sstkeyword {
            docstring=Whether MPI runs as an asynchronous progress thread [service] or
            blocks the application [user] thread.;
        }
      */
      queue_ = mpiqueue_factory::get_optional_param("mpiqueue_thread_type",
          "service", sstmac_env::params, mpiserv);
      queue_->init(id_);

      /**
        sstkeyword {
            docstring=The underlying protocols/hardware features that MPI supports.;
        }
      */
      mpi_implementation::ptr mpi_impl =
          mpi_implementation_factory::get_optional_param("mpi_implementation",
              "basic", sstmac_env::params);
      queue_->set_implementation(mpi_impl);

      AMI* ami = os_->getAMI(AMI::AMI_COMPUTE);
      sstmac::hw::node::ptr mynode = dynamic_cast<sstmac::hw::node*> (ami);
      if (!mynode)
      {
        throw oserror("MPI should only be run on AMIs inheriting from node");
      }
      queue_->set_eventmanager(mynode->get_eventmanager());

      mpiserv->add_task(id_, queue_, rank_.id);

      //we are all done with mpiserv - clear for help track mem leaks
      mpiserv = 0;
      mpilib = 0;

#if 0
      if (sstmac_env::params->has_param("mpi_delay"))
      {
        timestamp t = sstmac_env::params->get_time_param("mpi_delay");
        sst_message::ptr sleepmsg = sleep_message::construct(get_lib_name(),
            timestamp(t));
        os_->execute_kernel(AMI::AMI_COMP_SLEEP, sleepmsg);
        wait_this(sleepmsg);
      }
#endif

      comm_factory_ = mpicommfactory::construct(id_.app_, mpiapi::ptr(this));
      comm_factory_->init(env, rank_);

      worldcomm_ = comm_factory_->get_world();
      selfcomm_ = comm_factory_->get_self();

      status_ = is_initialized;

      if (worldcomm_->rank().id == 0)
      {
        starttime_ = os_->now();
      }

      if (parent_tracer_->is_active())
      {
        SSTMAC_DEBUG << "mpiapi: parent tracer is active...\n";
        std::string nodefile;
        if (sstmac_env::params->has_param("launch_dumpi_mapname"))
        {
          nodefile = sstmac_env::params->reread_param("launch_dumpi_mapname");
        }
        else
        {
          throw inputerror(
              "dumpi tracer cannot run without receiving nodemap file via launch_dumpi_mapname parameter");
        }

        sstmac::hw::topology::ptr top = allocationstrategy::get_topology();
        if (top == 0)
        {
          throw illformederror(
              "allocationstrategy static topology was not configured");
        }

        std::map<std::string, std::vector<int> > hostmap;
        hostname_allocation::read_map_file("mpiapi: ", nodefile, hostmap);
        std::map<std::string, std::vector<int> >::iterator it, end =
            hostmap.end();

        // this is really only compatible with regular topologies
        hw::regular_topology::ptr regtop = safe_cast(top, hw::regular_topology);
        for (it = hostmap.begin(); it != end; ++it)
        {
          const std::vector<int> coords = it->second;
          int64_t nodeindex = regtop->get_node_id(coords);
          hostmap_[nodeindex] = it->first;
        }
      }

      /** Do this before activating the tracer. We don't want to trace BOTH
       the barrier AND the MPI_Init */
      timestamp before = os_->now();
      barrier(worldcomm_, false);

      if (parent_tracer_->is_active() && !tracer_)
      {
        int rank = rank_.id;
        taskid tid(rank);
        int nodeindex = env->get_node_address(tid)->unique_id();
        std::string hostname = hostmap_[nodeindex];

        if (hostname == "")
        {
          throw illformederror(
              "unable to find nodes in nodenum map for dumpi tracer");
        }

        tracer_ = dumpitracer::construct(tracername_, hostname);

        SSTMAC_DEBUG << "mpiapi: tracer active: " << tracer_active() << "\n";

      }

      timestamp after = os_->now();
      if (tracer_active())
        tracer_->init(before, after, worldcomm_, selfcomm_);

      end_api_call();

      return os_->now();

    }

    //
    // Finalize MPI.
    //
    timestamp
    mpiapi::finalize()
    {
      SSTMACBacktrace("MPI_Finalize");
      start_api_call();

      mpitracer::ptr tmptracer = tracer_;
      tracer_ = 0; //nuke the tracers so we don't double count barriers
      timestamp before = os_->now();
      barrier(worldcomm_, false);
      timestamp after = os_->now();

      if (tmptracer && tmptracer->is_active())
        tmptracer->finalize(before, after);

      status_ = is_finalized;

      if (worldcomm_->rank().id == 0)
      {
        runtime_ = os_->now() - starttime_;
        debug mpicheck("<debug> mpicheck");
        if (mpicheck.is_active())
        {
          mpicheck << "----- MPI API ----------" << "\n";
          //mpicheck << "App: " << appname_ << "\n";
          mpicheck << "ID: " << id_.to_string() << "\n";
          mpicheck << "At sim time: " << os_->now().sec() << "\n";
          mpicheck << "App runtime: " << runtime_.sec() << "\n";
          mpicheck << "Rank 0 is passed the barrier in finalize," << "\n";
          mpicheck << "which means the simulation can probably be" << "\n";
          mpicheck << "killed, if it's not already dead." << "\n";
          mpicheck << "------------------------ " << "\n";
        }
        sstmac_env::end_gracefully();
      }
      comm_factory_->finalize();

      end_api_call();

      return os_->now();
    }

    //
    // Get current time.
    //
    timestamp
    mpiapi::wtime()
    {
      return os_->now();
    }

    /* Create and destroy communicators. */

    //
    // Duplicate a communicator.
    //
    timestamp
    mpiapi::comm_dup(const mpicomm::ptr &input, mpicomm::ptr &output)
    {
      SSTMACBacktrace("MPI_Comm_dup");
      start_api_call();

      assert_initialized("mpiapi::comm_dup");
      timestamp entry = os_->now();
      output = comm_factory_->comm_dup(input);
      if (tracer_active())
        tracer_->comm_dup(entry, os_->now(), input, output);
      end_api_call();
      return os_->now();

      throw unimplementederror("mpiapi::comm_dup not implemented");
    }

    //
    // Create a communicator containing a subset of an existing comm.
    //
    timestamp
    mpiapi::comm_create(const mpicomm::ptr &input, mpigroup::ptr group,
        mpicomm::ptr &output)
    {
      SSTMACBacktrace("MPI_Comm_create");
      start_api_call();

      static const char* here = "mpiapi::comm_create";
      assert_initialized(here);
      timestamp entry = os_->now();
      // validranks(members, input, here);
      output = comm_factory_->comm_create(input, group);
      if (tracer_active())
      {
        std::vector<mpiid> members;
        for (uint i = 0; i < group->size(); i++)
        {
          members.push_back(input->get_group()->rank_of_task(group->at(i)));
        }
        tracer_->comm_create(entry, os_->now(), input, members, output);
      }
      end_api_call();
      return os_->now();

      throw unimplementederror("mpiapi::comm_create not implemented");
    }

    //
    // Split a communicator.  This one is a little weird.
    //
    timestamp
    mpiapi::comm_split(const mpicomm::ptr &incomm, int color, int key,
        mpicomm::ptr &outcomm)
    {
      SSTMACBacktrace("MPI_Comm_split");
      start_api_call();

      static const char* here = "mpiapi::comm_split";
      //std::cerr << here << "(" << incomm << ", " << color << ", "
      //      << key << ", &outcomm)\n";
      assert_initialized(here);
      outcomm = comm_factory_->comm_split(incomm, color, key);
      if (tracer_active())
        tracer_->comm_split(os_->now(), os_->now(), incomm, color, key, outcomm);
      end_api_call();
      return os_->now();

      throw unimplementederror("mpiapi::comm_split not implemented");
    }

    //
    // Destroy a communicator.
    //
    timestamp
    mpiapi::comm_free(mpicomm::ptr comm)
    {
      SSTMACBacktrace("MPI_Comm_free");
      start_api_call();

      assert_initialized("mpiapi::comm_create");
      if (tracer_active())
        tracer_->comm_free(os_->now(), os_->now(), comm);
      end_api_call();
      return os_->now();

      throw unimplementederror("mpiapi::comm_free not implemented");
    }

    void
    mpiapi::mpi_cart_create(mpicomm::ptr &comm, int ndims, int *dims,
        int *periods, int reorder, mpicomm::ptr &outcomm)
    {
      SSTMACBacktrace("MPI_Cart_create");
      start_api_call();

      static const char* here = "mpiapi::mpi_cart_create";
      assert_initialized(here);
      outcomm = comm_factory_->create_cart(comm, ndims, dims, periods, reorder);
      end_api_call();
    }

    void
    mpiapi::mpi_cart_get(mpicomm::ptr &comm, int maxdims, int *dims,
        int *periods, int *coords)
    {
      SSTMACBacktrace("MPI_Cart_get");
      start_api_call();

      mpicomm_cart::ptr c = boost::dynamic_pointer_cast<mpicomm_cart>(comm);

      if (!c)
      {
        throw ssterror(
            "mpiapi::mpi_cart_get() - mpi comm did not cast to mpicomm_cart");
      }

      for (int i = 0; i < maxdims; i++)
      {
        dims[i] = c->get_dim(i);
        periods[i] = c->get_period(i);
      }

      c->set_coords(c->rank().id, coords);
      end_api_call();
    }

    void
    mpiapi::mpi_cartdim_get(mpicomm::ptr &comm, int *ndims)
    {
      SSTMACBacktrace("MPI_Cartdim_get");
      start_api_call();

      mpicomm_cart::ptr c = boost::dynamic_pointer_cast<mpicomm_cart>(comm);

      if (!c)
      {
        throw ssterror(
            "mpiapi::mpi_cart_get() - mpi comm did not cast to mpicomm_cart");
      }
      *ndims = c->get_ndims();
      end_api_call();
    }

    void
    mpiapi::mpi_cart_rank(mpicomm::ptr comm, int *coords, int *rank)
    {
      SSTMACBacktrace("MPI_Cart_rank");
      start_api_call();

      mpicomm_cart::ptr c = boost::dynamic_pointer_cast<mpicomm_cart>(comm);

      if (!c)
      {
        throw ssterror(
            "mpiapi::mpi_cart_rank() - mpi comm did not cast to mpicomm_cart");
      }

      *rank = c->get_rank(coords);
      end_api_call();
    }

    void
    mpiapi::mpi_cart_shift(mpicomm::ptr comm, int direction, int displ,
        int *source, int *dest)
    {
      SSTMACBacktrace("MPI_Cart_shift");
      start_api_call();

      mpicomm_cart::ptr c = boost::dynamic_pointer_cast<mpicomm_cart>(comm);

      if (!c)
      {
        throw ssterror(
            "mpiapi::mpi_cart_shift() - mpi comm did not cast to mpicomm_cart");
      }

      *source = c->shift(direction, -1 * displ);
      *dest = c->shift(direction, displ);
      end_api_call();
    }

    void
    mpiapi::mpi_cart_coords(mpicomm::ptr comm, int rank, int maxdims,
        int *coords)
    {
      SSTMACBacktrace("MPI_Cart_coords");
      start_api_call();

      mpicomm_cart::ptr c = boost::dynamic_pointer_cast<mpicomm_cart>(comm);

      if (!c)
      {
        throw ssterror(
            "mpiapi::mpi_cart_coords() - mpi comm did not cast to mpicomm_cart");
      }

      c->set_coords(rank, coords);
      end_api_call();
    }

    /* Basic point-to-point operations. */

    //
    // Blocking send.
    //
    timestamp
    mpiapi::send(int count, const mpitype &type, const mpiid &target,
        const mpitag &tag, const mpicomm::ptr &comm,
        const payload::const_ptr& content)
    {
      SSTMACBacktrace("MPI_Send");
      start_api_call();

      static const char *here = "mpiapi::send";
      assert_initialized(here);
      validsend(count, type, target, tag, comm, here);
      timestamp entry = os_->now();
      key::ptr block_key = key::construct(mpiapi::default_key_category);

      mpirequest_t req = mpirequest::construct(block_key);
      this->do_isend(count, type, target, tag, comm, req, content);

      this->queue_->progress_loop(req);

      if (tracer_active())
        tracer_->send(entry, os_->now(), count, type, target, tag, comm);
      end_api_call();
      return os_->now();
    }

    //
    // Blocking send using a user-defined buffer (see buffer_attach).
    //
    timestamp
    mpiapi::bsend(int count, const mpitype &type, const mpiid &target,
        const mpitag &tag, const mpicomm::ptr &comm,
        const payload::const_ptr& content)
    {
      SSTMACBacktrace("MPI_Bsend");
      start_api_call();

      static const char *here = "mpiapi::bsend";
      assert_initialized(here);
      validsend(count, type, target, tag, comm, here);
      timestamp entry = os_->now();
      mpirequest_t req = mpirequest::construct(key::construct(mpiapi::default_key_category));
      pending_bsends_.push_back(req);
      this->do_ibsend(count, type, target, tag, comm, req, content);
      //timestamp finish = queue_->progress_loop(req);
      timestamp finish = os_->now(); ///  GH: -------- added this, because a bsend can be unmatched
      if (tracer_active())
        tracer_->bsend(entry, finish, count, type, target, tag, comm);
      end_api_call();
      return finish;
    }

    //
    // Blocking syncronous send.
    //
    timestamp
    mpiapi::ssend(int count, const mpitype &type, const mpiid &target,
        const mpitag &tag, const mpicomm::ptr &comm,
        const payload::const_ptr& content)
    {
      SSTMACBacktrace("MPI_Ssend");
      start_api_call();

      static const char *here = "mpiapi::comm_ssend";
      assert_initialized(here);
      validsend(count, type, target, tag, comm, here);
      mpirequest_t req = mpirequest::construct(key::construct(mpiapi::default_key_category));
      timestamp entry = os_->now();
      this->do_issend(count, type, target, tag, comm, req, content);
      timestamp finish = queue_->progress_loop(req);
      if (tracer_active())
        tracer_->ssend(entry, finish, count, type, target, tag, comm);
      end_api_call();
      return finish;
    }

    //
    // Ready send (as in "ready or not, here comes the data").
    timestamp
    mpiapi::rsend(int count, const mpitype &type, const mpiid &target,
        const mpitag &tag, const mpicomm::ptr &comm,
        const payload::const_ptr& content)
    {
      SSTMACBacktrace("MPI_Rsend");
      static const char *here = "mpiapi::rsend";
      assert_initialized(here);
      validsend(count, type, target, tag, comm, here);
      mpirequest_t req = mpirequest::construct(key::construct(mpiapi::default_key_category));
      timestamp entry = os_->now();
      this->do_irsend(count, type, target, tag, comm, req, content);
      timestamp finish = queue_->progress_loop(req);
      if (tracer_active())
        tracer_->ssend(entry, finish, count, type, target, tag, comm);
      return finish;
    }

    //
    // Nonblocking send.
    //
    timestamp
    mpiapi::isend(int count, const mpitype &type, const mpiid &dest,
        const mpitag &tag, const mpicomm::ptr &comm, mpirequest_t &req,
        const payload::const_ptr& content)
    {
      SSTMACBacktrace("MPI_Isend");
      start_api_call();

      static const char *here = "mpiapi::isend";
      assert_initialized(here);
      validsend(count, type, dest, tag, comm, here);
      timestamp entry = os_->now();
      req = mpirequest::construct(key::construct(mpiapi::default_key_category));
      this->do_isend(count, type, dest, tag, comm, req, content);
      timestamp finish = os_->now();
      if (tracer_active())
        tracer_->isend(entry, finish, count, type, dest, tag, comm, req);
      end_api_call();
      return finish;
    }

    /// Nonblocking send.
    /// wait for this request has the same completion as send.
    /// \return time when the request has been registered (likely current time).
    timestamp
    mpiapi::do_isend(int count, const mpitype &type, const mpiid &dest,
        const mpitag &tag, const mpicomm::ptr &comm, const mpirequest_t &req,
        const payload::const_ptr& content)
    {
      SSTMAC_DEBUG << "mpiapi[" << rank_.id << "]::do_isend(" << count << ","
            << type.label << "," << dest.id << ", " << tag.id << ", "
            << comm->id().id << ", " << "[req]" << ")\n";

      //// shared_ptr<mpikernel> ker = strategy_->send()->build(req, queue_, count,
      ////     type, dest, tag, comm, content, magic_);

      strategy_->send()->execute(req, queue_, count, type, dest, tag, comm,
          content, os_);

      ////  dbg_ << "mpiapi::isend: adding send kernel.\n";
      ////os_->add(os_->now(), ker);
      return os_->now();
    }

    //
    // Nonblocking user-buffered send.
    //
    timestamp
    mpiapi::ibsend(int count, const mpitype &type, const mpiid &dest,
        const mpitag &tag, const mpicomm::ptr &comm, mpirequest_t &req,
        const payload::const_ptr& content)
    {
      SSTMACBacktrace("MPI_Ibsend");
      start_api_call();

      static const char *here = "mpiapi::ibsend";
      assert_initialized(here);
      validsend(count, type, dest, tag, comm, here);
      timestamp entry = os_->now();
      req = mpirequest::construct(key::construct(mpiapi::default_key_category));
      this->do_ibsend(count, type, dest, tag, comm, req, content);
      timestamp finish = os_->now();
      if (tracer_active())
        tracer_->ibsend(entry, finish, count, type, dest, tag, comm, req);
      end_api_call();
      return finish;
    }

    //
    // Nonblocking user-buffered send.
    //
    timestamp
    mpiapi::do_ibsend(int count, const mpitype &type, const mpiid &dest,
        const mpitag &tag, const mpicomm::ptr &comm, const mpirequest_t &req,
        const payload::const_ptr& content)
    {
      //std::cerr << "WARNING:  mpiapi::ibsend does not use "
      //    << "user-defined buffer\n";
      return this->do_isend(count, type, dest, tag, comm, req, content);
    }

    //
    // Nonblocking synchronous send.
    //
    timestamp
    mpiapi::issend(int count, const mpitype &type, const mpiid &dest,
        const mpitag &tag, const mpicomm::ptr &comm, mpirequest_t &req,
        const payload::const_ptr& content)
    {
      SSTMACBacktrace("MPI_Issend");
      start_api_call();

      static const char *here = "mpiapi::issend";
      assert_initialized(here);
      validsend(count, type, dest, tag, comm, here);
      timestamp entry = os_->now();
      req = mpirequest::construct(key::construct(mpiapi::default_key_category));
      this->do_issend(count, type, dest, tag, comm, req, content);
      timestamp finish = os_->now();
      if (tracer_active())
        tracer_->issend(entry, finish, count, type, dest, tag, comm, req);
      end_api_call();
      return finish;
    }

    //
    // Nonblocking synchronous send.
    //
    timestamp
    mpiapi::do_issend(int count, const mpitype &type, const mpiid &dest,
        const mpitag &tag, const mpicomm::ptr &comm, const mpirequest_t &req,
        const payload::const_ptr& content)
    {
      strategy_->ssend()->execute(req, queue_, count, type, dest, tag, comm,
          content, os_);

      return os_->now();
    }

    //
    // Nonblocking ready send.
    //
    timestamp
    mpiapi::irsend(int count, const mpitype &type, const mpiid &dest,
        const mpitag &tag, const mpicomm::ptr &comm, mpirequest_t &req,
        const payload::const_ptr& content)
    {
      SSTMACBacktrace("MPI_Irsend");
      start_api_call();

      static const char *here = "mpiapi::irsend";
      assert_initialized(here);
      validsend(count, type, dest, tag, comm, here);
      timestamp entry = os_->now();
      req = mpirequest::construct(key::construct(mpiapi::default_key_category));
      this->do_irsend(count, type, dest, tag, comm, req, content);
      timestamp finish = os_->now();
      if (tracer_active())
        tracer_->irsend(entry, finish, count, type, dest, tag, comm, req);
      end_api_call();
      return finish;
    }

    //
    // Nonblocking ready send.
    //
    timestamp
    mpiapi::do_irsend(int count, const mpitype &type, const mpiid &dest,
        const mpitag &tag, const mpicomm::ptr &comm, const mpirequest_t &req,
        const payload::const_ptr& content)
    {
      strategy_->rsend()->execute(req, queue_, count, type, dest, tag, comm,
          content, os_);

      return os_->now();
    }

    //
    // Synchronous receive.
    //
    timestamp
    mpiapi::recv(int count, const mpitype &type, const mpiid &source,
        const mpitag &tag, const mpicomm::ptr &comm, const_mpistatus_t &stat)
    {
      SSTMACBacktrace("MPI_Recv");
      start_api_call();

      static const char *here = "mpiapi::recv";
      assert_initialized(here);
      validrecv(count, type, source, tag, comm, here);
      timestamp entry = os_->now();
      key::ptr block_key = key::construct(mpiapi::default_key_category);

      mpirequest_t req = mpirequest::construct(block_key);
      this->do_irecv(count, type, source, tag, comm, req);
      timestamp finish = queue_->progress_loop(req);

      if (tracer_active())
        tracer_->recv(entry, finish, count, type, source, tag, comm, stat);
      stat = req->get_status();
      end_api_call();
      return finish;
    }

    //
    // Nonblocking receive
    //
    timestamp
    mpiapi::irecv(int count, const mpitype &type, const mpiid &source,
        const mpitag &tag, const mpicomm::ptr &comm, mpirequest_t &req)
    {
      SSTMACBacktrace("MPI_Irecv");
      start_api_call();

      static const char *here = "mpiapi::irecv";
      assert_initialized(here);
      validrecv(count, type, source, tag, comm, here);

      timestamp entry = os_->now();

      req = mpirequest::construct(key::construct(mpiapi::default_key_category));

      this->do_irecv(count, type, source, tag, comm, req);
      timestamp finish = os_->now();
      if (tracer_active())
        tracer_->irecv(entry, finish, count, type, source, tag, comm, req);
      end_api_call();
      return finish;
    }

    //
    // Asynchronous receive.
    //
    timestamp
    mpiapi::do_irecv(int count, const mpitype &type, const mpiid &source,
        const mpitag &tag, const mpicomm::ptr &comm, const mpirequest_t &req)
    {
      static const char *here = "mpiapi::irecv";
      assert_initialized(here);
      validrecv(count, type, source, tag, comm, here);
      SSTMAC_DEBUG << "mpiapi[" << rank_.id << "]::do_irecv(" << count << ","
            << type.label << "," << source.id << ", " << tag.id << ", "
            << comm->id().id << ", " << "[req]" << ")\n";

      strategy_->recv()->execute(req, queue_, count, type, source, tag, comm,
          os_);

      return os_->now();
    }

    //
    // Send/receieve exchange.
    //
    timestamp
    mpiapi::sendrecv(int sendcount, const mpitype &sendtype,
        const mpiid &target, const mpitag &sendtag, int recvcount,
        const mpitype &recvtype, const mpiid &source, const mpitag &recvtag,
        const mpicomm::ptr &comm, const_mpistatus_t &stat,
        const payload::const_ptr& content)
    {
      SSTMACBacktrace("MPI_Sendrecv");
      start_api_call();

      static const char *here = "mpiapi::sendrecv";
      assert_initialized(here);
      validsend(sendcount, sendtype, target, sendtag, comm, here);
      validrecv(recvcount, recvtype, source, recvtag, comm, here);
      timestamp entry = os_->now();
      mpirequest_t req1 = mpirequest::construct(key::construct(mpiapi::default_key_category));
      mpirequest_t req2 = mpirequest::construct(key::construct(mpiapi::default_key_category));

      this->do_isend(sendcount, sendtype, target, sendtag, comm, req1, content);
      this->do_irecv(recvcount, recvtype, source, recvtag, comm, req2);
      std::vector<mpirequest_t> reqs;
      reqs.push_back(req1);
      reqs.push_back(req2);
      std::vector<const_mpistatus_t> stats;
      waitall(reqs, stats, false);
      stat = dynamic_pointer_cast<const mpistatus> (req2->get_status());
      timestamp fini = os_->now();
      if (tracer_active())
        tracer_->sendrecv(entry, fini, sendcount, sendtype, target, sendtag,
            recvcount, recvtype, source, recvtag, comm, stat);
      end_api_call();
      return fini;
    }

    //
    // Send and receive using a single buffer.
    //
    timestamp
    mpiapi::sendrecv_replace(int count, const mpitype &sendtype,
        const mpiid &dest, const mpitag &sendtag, const mpiid &source,
        const mpitag &recvtag, const mpicomm::ptr &comm,
        const_mpistatus_t &stat)
    {
      static const char *here = "mpiapi::sendrecv";
      assert_initialized(here);
      validsend(count, sendtype, dest, sendtag, comm, here);
      validrecv(count, sendtype, source, recvtag, comm, here);
      timestamp entry = os_->now();
      mpirequest_t req1 = mpirequest::construct(key::construct(mpiapi::default_key_category));
      mpirequest_t req2 = mpirequest::construct(key::construct(mpiapi::default_key_category));
      payload::const_ptr cont;
      this->do_isend(count, sendtype, dest, sendtag, comm, req1, cont);
      this->do_irecv(count, sendtype, source, recvtag, comm, req2);
      std::vector<mpirequest_t> reqs;
      reqs.push_back(req1);
      reqs.push_back(req2);
      std::vector<const_mpistatus_t> stats;
      waitall(reqs, stats, false);
      stat = dynamic_pointer_cast<const mpistatus> (req2->get_status());
      timestamp fini = os_->now();
      if (tracer_active())
        tracer_->sendrecv_replace(entry, fini, count, sendtype, dest, sendtag,
            source, recvtag, comm, stat);
      return fini;
    }

    /* Stepwise point-to-point (not implemented, but
     * should be fairly straight-forward to complete). */

    //
    // Initiate a send.
    //
    timestamp
    mpiapi::send_init(int count, const mpitype &type, const mpiid &target,
        const mpitag &tag, const mpicomm::ptr &comm, mpirequest_t &req,
        void* buf)
    {
      static const char *here = "mpiapi::send_init";
      assert_initialized(here);
      req = persistentsend::construct(mpiapi::ptr(this), &mpiapi::do_isend,
          count, type, target, tag, comm, buf);
      if (tracer_active())
        tracer_->send_init(os_->now(), os_->now(), count, type, target, tag,
            comm, req);
      return this->wtime();
    }

    //
    // Initiate a buffered send.
    //
    timestamp
    mpiapi::bsend_init(int count, const mpitype &type, const mpiid &target,
        const mpitag &tag, const mpicomm::ptr &comm, mpirequest_t &req,
        void* buf)
    {
      static const char *here = "mpiapi::bsend_init";
      assert_initialized(here);
      validsend(count, type, target, tag, comm, here);
      req = persistentsend::construct(mpiapi::ptr(this), &mpiapi::do_ibsend,
          count, type, target, tag, comm, buf);
      return this->wtime();
    }

    //
    // Initiate a synchronous send.
    //
    timestamp
    mpiapi::ssend_init(int count, const mpitype &type, const mpiid &target,
        const mpitag &tag, const mpicomm::ptr &comm, mpirequest_t &req,
        void* buf)
    {
      static const char *here = "mpiapli::ssend_init";
      assert_initialized(here);
      validsend(count, type, target, tag, comm, here);
      req = persistentsend::construct(mpiapi::ptr(this), &mpiapi::do_issend,
          count, type, target, tag, comm, buf);
      return this->wtime();
    }

    //
    // Initiate a ready send.
    //
    timestamp
    mpiapi::rsend_init(int count, const mpitype &type, const mpiid &target,
        const mpitag &tag, const mpicomm::ptr &comm, mpirequest_t &req,
        void* buf)
    {
      static const char *here = "mpiapi::rsend_init";
      assert_initialized(here);
      validsend(count, type, target, tag, comm, here);
      req = persistentsend::construct(mpiapi::ptr(this), &mpiapi::do_irsend,
          count, type, target, tag, comm, buf);
      return this->wtime();
    }

    //
    // Initiate a receive.
    //
    timestamp
    mpiapi::recv_init(int count, const mpitype &type, const mpiid &target,
        const mpitag &tag, const mpicomm::ptr &comm, mpirequest_t &req)
    {
      static const char *here = "mpiapi::recv_init";
      assert_initialized(here);
      validrecv(count, type, target, tag, comm, here);
      req = persistentrecv::construct(mpiapi::ptr(this), count, type, target,
          tag, comm);
      if (tracer_active())
        tracer_->recv_init(os_->now(), os_->now(), count, type, target, tag,
            comm, req);
      return this->wtime();
    }

    //
    // Start a previously initiated request.
    //
    timestamp
    mpiapi::start(mpirequest_t req)
    {
      SSTMACBacktrace("MPI_Start");
      start_api_call();
      persistent::ptr pers = dynamic_pointer_cast<persistent> (req);
      if (!pers)
        throw illformederror("mpiapi::start:  Incorrect request handle type.");
      timestamp entry = os_->now();
      pers->start();
      timestamp fini = os_->now();
      if (tracer_active())
        tracer_->start(entry, fini, req);
      end_api_call();
      return this->wtime();
    }

    //
    // Start a collection of previously initiated requests.
    //
    timestamp
    mpiapi::startall(std::vector<mpirequest_t> &requests)
    {
      SSTMACBacktrace("MPI_Startall");
      start_api_call();
      typedef std::vector<mpirequest_t>::iterator iterator_t;
      for (iterator_t it = requests.begin(); it != requests.end(); ++it)
        this->start(*it);
      end_api_call();
      return this->wtime();
    }

    //
    // Cancel a request that had been previously initiated.
    //
    timestamp
    mpiapi::cancel(mpirequest_t req)
    {
      SSTMACBacktrace("MPI_Cancel");
      start_api_call();
      assert_initialized("mpiapi::cancel");
      if (!req->is_complete())
      {
        req->cancel();
      }
      end_api_call();
      return this->wtime();
    }

    //
    // Test whether a request has been cancelled (takes no time).
    //
    timestamp
    mpiapi::request_free(mpirequest_t req)
    {
      assert_initialized("mpiapi::request_free");

      return this->wtime();
    }

    //
    // Test whether a request has been cancelled (takes no time).
    //
    bool
    mpiapi::test_cancelled(const const_mpistatus_t &req)
    {
      assert_initialized("mpiapi::test_cancelled");
      throw unimplementederror("mpiapi::test_cancelled:  Not implemented yet.");
    }

    /* Completion of outstanding requests */

    //
    // Wait for a request to finish.
    //
    timestamp
    mpiapi::wait(mpirequest_t req, const_mpistatus_t &stat)
    {
      SSTMACBacktrace("MPI_Wait");
      start_api_call();
      assert_initialized("mpiapi::wait");
      timestamp entry = os_->now();
      timestamp fini = entry;
      timestamp finish;
      if (req.get() == NULL || req->is_cancelled())
      {
        // Null request is instantly complete.
        finish = os_->now();
      }
      else if (req->is_persistent())
      {
        persistent::ptr pers = boost::dynamic_pointer_cast<persistent>(req);
        if (!pers->started_)
        {
          mpistatus::ptr retstat = mpistatus::construct();
          retstat->set_tag(mpitag::any_tag);
          retstat->set_source(mpiid::any_source);
          retstat->set_count(0);
          stat = retstat;
          finish = os_->now();
        }
        else
        {
          finish = queue_->progress_loop(req);
          stat = req->get_status();
        }
      }
      else
      {
        finish = queue_->progress_loop(req);
        stat = req->get_status();
      }
      if (tracer_active())
        tracer_->wait(entry, finish, req, stat);
      end_api_call();
      return fini;
    }

    //
    // Wait for all the given requests to finish.
    //
    timestamp
    mpiapi::waitall(std::vector<mpirequest_t> &req,
        std::vector<const_mpistatus_t> &statuses, bool tracethis)
    {
      SSTMACBacktrace("MPI_Waitall");
      start_api_call();

      assert_initialized("mpiapi::waitall");
      timestamp entry = os_->now();
      timestamp fini;
      if (!req.empty())
      {
        typedef std::vector<mpirequest_t>::iterator iterator_t;
        key::ptr thekey;

        statuses.clear();
        for (iterator_t it = req.begin(); it != req.end(); ++it)
        {
          mpirequest_t& req = *it;
          if (!req)
          {
            SSTMAC_DEBUG << "mpiapi::waitall: null request, skipping \n";
          }
          else if (req->is_cancelled())
          {
            SSTMAC_DEBUG << "mpiapi::waitall: request cancelled, skipping \n";
            statuses.push_back(mpistatus::const_ptr());
          }
          else
          {
            if (!req->is_complete())
            {
              SSTMAC_DEBUG << "mpiapi[" << worldcomm_->rank().id
                    << "]: req not finished, blocking \n";
              fini = queue_->progress_loop(req);
            }
            else
            {
              SSTMAC_DEBUG << "mpiapi[" << worldcomm_->rank().id
                    << "]: req already finished, continuing \n";
            }

            const_mpistatus_t stat;
            if (req.get() != NULL)
            {
              stat = dynamic_pointer_cast<const mpistatus> (req->get_status());
              if (stat == 0)
              {
                sst_throw(illformederror, "mpiapi::waitall: "
                    "Wait request returned null mpistatus object.");
              }
            }
            statuses.push_back(stat);
          }
        }
      }
      else
      {
        statuses.clear();
      }

      if (tracethis && tracer_active())
        tracer_->waitall(entry, fini, req, statuses);
      end_api_call();
      return fini;
    }

    //
    // Wait for at least one of the given requests to complete.
    //
    timestamp
    mpiapi::waitany(std::vector<mpirequest_t> &req, int &index,
        const_mpistatus_t &stat)
    {
      SSTMACBacktrace("MPI_Waitany");
      start_api_call();
      assert_initialized("mpiapi::waitany");
      timestamp entry = os_->now(), fini = entry;
      if (!req.empty())
      {

        index = -1;
        for (uint i = 0; i < req.size(); i++)
        {
          if (req[i]->is_complete())
          {
            SSTMAC_DEBUG << "mpiapi[" << worldcomm_->rank().id << "]: request "
                  << i << " already complete\n";
            index = i;
            break;
          }
        }

        if (index < 0)
        {
          queue_->start_progress_loop(req);
          for (uint i = 0; i < req.size(); i++)
          {
            if (!req[i]->get_key()->still_blocked())
            {
              //this is the one
              SSTMAC_DEBUG << "mpiapi[" << worldcomm_->rank().id
                    << "]: unblocked by request " << i << "\n";
              index = i;
              break;
            }
          }
        }

        if (index >= 0)
        {
          stat = req.at(index)->get_status();
          if (stat == 0)
          {
            sst_throw(illformederror, "mpiapi::waitany:  "
                "Wait request returned null mpistatus object.");
          }
        }
      }
      else
      {
        SSTMAC_DEBUG << "mpiapi[" << worldcomm_->rank().id
              << "]: waitany- no requests found \n";
        index = -1;
      }

      queue_->finish_progress_loop(req);

      fini = os_->now();
      if (tracer_active())
        tracer_->waitany(entry, fini, req, index, stat);
      end_api_call();
      return fini;
    }

    //
    // Wait for at least one of the given requests to complete, or for timeout amount of time, whichever comes first.
    //
    timestamp
    mpiapi::waitany(std::vector<mpirequest_t> &req, int &index,
        const_mpistatus_t &stat, timestamp timeout)
    {
      SSTMACBacktrace("MPI_Waitany");
      start_api_call();
      assert_initialized("mpiapi::waitany-timeout");
      timestamp entry = os_->now(), fini = entry;
      sst_message::ptr sleep = sleep_message::construct(libname_, timeout);
      if (!req.empty())
      {
        key::ptr thekey = key::construct(mpiapi::default_key_category);

        index = -1;
        for (uint i = 0; i < req.size(); i++)
        {
          if (!req[i])
          {
            sst_throw(ssterror, "mpiapi::waitany: null request object");
          }
          if (req[i]->is_complete())
          {
            SSTMAC_DEBUG << "mpiapi[" << worldcomm_->rank().id << "]: request "
                  << i << " already complete\n";
            index = i;
            break;
          }
        }

        if (index < 0)
        {
          queue_->start_progress_loop(req, timeout);

          for (uint i = 0; i < req.size(); i++)
          {
            if (!req[i]->get_key()->still_blocked())
            {
              //this is the one
              SSTMAC_DEBUG << "mpiapi[" << worldcomm_->rank().id
                    << "]: unblocked by request " << i << "\n";
              index = i;
              break;
            }
          }
        }

        if (index >= 0)
        {
          stat = req.at(index)->get_status();
          if (stat == 0)
          {
            sst_throw(illformederror, "mpiapi::waitany:  Wait request returned null mpistatus");
          }
        }

      }
      else
      {
        SSTMAC_DEBUG << "mpiapi[" << worldcomm_->rank().id
              << "]: waitany- no requests found \n";
        index = -1;
      }

      queue_->finish_progress_loop(req);

      fini = os_->now();
      if (tracer_active())
        tracer_->waitany(entry, fini, req, index, stat);
      end_api_call();
      return fini;
    }

    //
    // Wait for some of the given requests to complete.
    //
    timestamp
    mpiapi::waitsome(std::vector<mpirequest_t> &req, std::vector<int> &indices,
        std::vector<const_mpistatus_t> &statuses)
    {
      SSTMACBacktrace("MPI_Waitsome");
      start_api_call();
      assert_initialized("mpiapi::waitany");
      timestamp entry = os_->now(), fini = entry;
      if (!req.empty())
      {
        indices.clear();
        for (uint i = 0; i < req.size(); i++)
        {
          if (req[i] != 0)
          {
            if (req[i]->is_complete())
            {
              SSTMAC_DEBUG << "mpiapi[" << worldcomm_->rank().id
                    << "]: request " << i << " already complete\n";
              indices.push_back(i);
            }
            //req[i]->get_key()->add_blocker(os_->get_current_context());
          }
        }

        if (indices.size() == 0)
        {
          queue_->start_progress_loop(req);
          for (uint i = 0; i < req.size(); i++)
          {
            if (req[i] != 0)
            {
              if (!req[i]->get_key()->still_blocked())
              {
                //this is the one
                SSTMAC_DEBUG << "mpiapi[" << worldcomm_->rank().id
                      << "]: unblocked by request " << i << "\n";
                indices.push_back(i);
              }
            }
            else
            {
              indices.push_back(i);
            }
          }
        }

        if (indices.size() > 0)
        {
          for (uint i = 0; i < indices.size(); i++)
          {
            mpirequest_t& mpireq = req.at(indices[i]);
            if (mpireq != 0)
            {
              const_mpistatus_t stat = req.at(indices[i])->get_status();
              if (stat == 0)
              {
                sst_throw(illformederror,"mpiapi::waitany:  Wait request returned null mpistatus");
              }
              statuses.push_back(stat);
            }
          }
        }
      }

      //cleanup
      queue_->finish_progress_loop(req);

      fini = os_->now();
      if (tracer_active())
        tracer_->waitsome(entry, fini, req, indices, statuses);
      end_api_call();
      return fini;
    }

    /* Testing of outstanding requests */

    /// Test whether a request has finished.
    /// \return the completion time for the call (current time).
    timestamp
    mpiapi::test(mpirequest_t req, bool &flag, const_mpistatus_t &stat)
    {
      SSTMACBacktrace("MPI_Test");
      start_api_call();
      assert_initialized("mpiapi::test");
      timestamp entry = os_->now(), fini = entry;
      if (!req)
      {
        flag = true;
        stat = const_mpistatus_t();
      }
      else
      {
        if (req->is_persistent())
        {
          //persistent::ptr pers = dynamic_pointer_cast<persistent> (req);
          flag = true;
          mpistatus::ptr retstat = mpistatus::construct();
          retstat->set_tag(mpitag::any_tag);
          retstat->set_source(mpiid::any_source);
          retstat->set_count(0);
          stat = retstat;
        }
        else
        {

          flag = req->is_complete();
          if (flag)
          {

            stat = dynamic_pointer_cast<const mpistatus> (req->get_status());
          }
        }
      }
      if (tracer_active())
        tracer_->test(entry, fini, req, flag, stat);
      end_api_call();
      return fini;
    }

    //
    // Test whether all the given requests have finished.
    //
    timestamp
    mpiapi::testall(std::vector<mpirequest_t> &req, bool &flag,
        std::vector<const_mpistatus_t> &stats)
    {
      SSTMACBacktrace("MPI_Testall");
      start_api_call();
      assert_initialized("mpiapi::testall");
      timestamp entry = os_->now();
      if (!req.empty())
      {
        flag = true;
        for (size_t id = 0; id < req.size(); ++id)
        {
          if (req.at(id).get() != NULL && !req.at(id)->is_complete())
          {
            flag = false;
            break;
          }
        }
        if (flag)
        {
          // All requests have completed.
          stats.resize(req.size());
          for (size_t id = 0; id < req.size(); ++id)
          {
            if (req.at(id).get() != NULL)
            {

              stats.at(id) = dynamic_pointer_cast<const mpistatus> (
                  req.at(id)->get_status());
            }
            else
              stats.at(id) = const_mpistatus_t();
          }
        }
      }
      else
      {
        stats.clear();
        flag = false;
      }
      timestamp fini = os_->now();
      if (tracer_active())
        tracer_->testall(entry, fini, req, flag, stats);
      end_api_call();
      return fini;
    }

    //
    // Test whether at least one of the given requests has completed.
    //
    timestamp
    mpiapi::testany(std::vector<mpirequest_t> &req, int &index, bool &flag,
        const_mpistatus_t &stat)
    {
      SSTMACBacktrace("MPI_Testany");
      start_api_call();
      assert_initialized("mpiapi::testany");
      timestamp entry = os_->now();
      if (!req.empty())
      {
        flag = false;
        for (size_t id = 0; id < req.size(); ++id)
        {
          if (req.at(id).get() != NULL && req.at(id)->is_complete())
          {

            index = id;
            flag = true;
            stat = dynamic_pointer_cast<const mpistatus> (
                req.at(id)->get_status());
            break;
          }
        }
      }
      else
      {
        flag = false;
      }
      timestamp fini = os_->now();
      if (tracer_active())
        tracer_->testany(entry, fini, req, index, flag, stat);
      end_api_call();
      return fini;
    }

    //
    // Test whether some of the given requests have completed.
    //
    timestamp
    mpiapi::testsome(std::vector<mpirequest_t> &req, std::vector<int> &indices,
        std::vector<const_mpistatus_t> &statuses)
    {
      SSTMACBacktrace("MPI_Testsome");
      start_api_call();
      assert_initialized("mpiapi::testsome");
      timestamp entry = os_->now();
      if (!req.empty())
      {
        indices.clear();
        statuses.clear();
        for (size_t id = 0; id < req.size(); ++id)
        {
          if (req.at(id).get() != NULL && req.at(id)->is_complete())
          {

            indices.push_back(id);
            statuses.push_back(
                dynamic_pointer_cast<const mpistatus> (req.at(id)->get_status()));
          }
        }
      }
      else
      {
        indices.clear();
        statuses.clear();
      }
      timestamp fini = os_->now();
      if (tracer_active())
        tracer_->testsome(entry, fini, req, indices, statuses);
      end_api_call();
      return fini;
    }

    //
    // Blocking probe for a message that matches the given signature.
    //
    timestamp
    mpiapi::probe(const mpiid &source, const mpitag &tag,
        const mpicomm::ptr &comm, const_mpistatus_t &stat)
    {
      SSTMACBacktrace("MPI_Probe");
      start_api_call();
      assert_initialized("mpiapi::probe");
      timestamp entry = os_->now();
      key::ptr thekey = key::construct(mpiapi::default_key_category);
      mpirequest_t req = mpirequest::construct(thekey);
      eventhandler::ptr done = queue_->progress_done_handler(os_, req);
      if (source == mpiid::proc_null)
      {
        mpistatus_t statout = mpistatus::construct();
        statout->set_count(0);
        statout->set_source(source);
        statout->set_tag(mpitag::any_tag);
        stat = statout;
      }
      else
      {

        queue_->probe(req, comm, source, tag, done);
        queue_->progress_loop(req);
        if (req->get_status()->get_count() > 0)
        { //just return this status
          stat = req->get_status();
        }
        else
        {
          mpistatus_t newstat = mpistatus::construct();
          newstat->set_commid(req->get_status()->commid());
          newstat->set_source(req->get_status()->source());
          newstat->set_dest(req->get_status()->dest());
          newstat->set_tag(req->get_status()->tag());
          newstat->set_send_type(req->get_status()->send_type());
          newstat->set_count(-req->get_status()->get_count());
          stat = newstat;
        }
      }
      timestamp fini = os_->now();
      if (tracer_active())
        tracer_->probe(entry, fini, source, tag, comm, stat);
      end_api_call();
      return fini;
    }

    //
    // Non-blocking probe for messages.
    //
    timestamp
    mpiapi::iprobe(const mpiid &source, const mpitag &tag,
        const mpicomm::ptr &comm, bool &flag, const_mpistatus_t &stat)
    {
      SSTMACBacktrace("MPI_Iprobe");
      start_api_call();
      assert_initialized("mpiapi::iprobe");
      timestamp entry = os_->now();
      if (source == mpiid::proc_null)
      {
        flag = true;
        mpistatus_t statout = mpistatus::construct();
        statout->set_count(0);
        statout->set_source(source);
        statout->set_tag(mpitag::any_tag);
        stat = statout;
      }
      else
      {
        mpistatus_t statout = queue_->iprobe(comm, source, tag);
        if (statout != 0)
        {
          flag = true;
          if (statout->get_count() > 0)
          { //just return this status
            stat = statout;
          }
          else
          {
            mpistatus_t newstat = mpistatus::construct();
            newstat->set_commid(statout->commid());
            newstat->set_source(statout->source());
            newstat->set_dest(statout->dest());
            newstat->set_tag(statout->tag());
            newstat->set_send_type(statout->send_type());
            newstat->set_count(-statout->get_count());
            stat = newstat;
          }
          stat = statout;
        }
        else
        {
          flag = false;
        }
      }
      timestamp fini = os_->now();
      if (tracer_active())
        tracer_->iprobe(entry, fini, source, tag, comm, flag, stat);
      end_api_call();
      return fini;
    }

    /* Mucking with MPI state. */

    //
    // Attach a user-defined buffer for *bsend operations.
    //
    timestamp
    mpiapi::buffer_attach(int bytes)
    {
      assert_initialized("mpiapi::buffer_attach");
      return os_->now();
    }

    //
    // Detach the user-defined buffer.
    //
    timestamp
    mpiapi::buffer_detach(int &bytes)
    {
      SSTMACBacktrace("MPI_Buffer_detach");
      start_api_call();
      assert_initialized("mpiapi::buffer_detach");
      std::vector<const_mpistatus_t> stats;
      waitall(pending_bsends_, stats, false);
      pending_bsends_.clear();
      end_api_call();
      return os_->now();
    }

    /* Collective operations */
    //
    // Set a barrier.
    //
    timestamp
    mpiapi::barrier(const mpicomm::ptr &comm, bool tracethis)
    {
      SSTMACBacktrace("MPI_Barrier");
      start_api_call();
      assert_initialized("mpiapi::barrier");
      mpitag barriertag = comm->next_collective_tag();
      SSTMAC_DEBUG << "mpiapi[" << worldcomm_->rank().id
            << "]: barrier: next tag = " << barriertag.id << "\n";
      timestamp entry = os_->now();
      mpirequest_t req = mpirequest::construct(key::construct(mpiapi::default_key_category));

      strategy_->barrier()->execute(req, queue_, barriertag, comm, os_);

      timestamp fini = queue_->progress_loop(req);

      if (tracethis && tracer_active())
        tracer_->barrier(entry, fini, comm);
      end_api_call();
      return fini;
    }

    //
    // Broadcast data without payload
    //
    timestamp
    mpiapi::bcast(int count, const mpitype &type, const mpiid &root,
        const mpicomm::ptr &comm)
    {
      payload::const_ptr content = payload::null();
      return this->bcast(count, type, root, comm, content);
    }

    //
    // Broadcast data with payload.
    //
    timestamp
    mpiapi::bcast(int count, const mpitype &type, const mpiid &root,
        const mpicomm::ptr &comm, payload::const_ptr &content)
    {
      SSTMACBacktrace("MPI_Bcast");
      start_api_call();
      static const char *here = "mpiapi::bcast";
      mpitag bcasttag = comm->next_collective_tag();
      assert_initialized(here);
      validbcast(count, type, root, bcasttag, comm, here);
      timestamp entry = os_->now();

      SSTMAC_DEBUG << "mpiapi(worldrank=" << worldcomm_->rank().id
            << ", commrank= " << comm->rank().id << "):: starting bcast \n";

      mpirequest_t req = mpirequest::construct(key::construct(mpiapi::default_key_category));
      strategy_->bcast()->execute(req, queue_, count, type, root, bcasttag,
          comm, content, os_);

      timestamp fini = queue_->progress_loop(req);

      mpistatus::const_ptr stat = req->get_status();

      if (!stat)
        throw illformederror("mpiapi::bcast: NULL status pointer after bcast.");

      content = stat->content();

      if (tracer_active())
        tracer_->bcast(entry, fini, count, type, root, comm);
      end_api_call();
      return fini;
    }

    //
    // Scatter data uniformly.
    //
    timestamp
    mpiapi::scatter(int sendcount, const mpitype &sendtype, int recvcount,
        const mpitype &recvtype, const mpiid &root, const mpicomm::ptr &comm)
    {
      std::vector<payload::const_ptr> vals;

      //if (comm->rank() == root)
      //  {
      //    for (mpiid id = mpiid(0); id < comm->size(); ++id)
      //     vals.push_back(payload::null());
      // }

      payload::const_ptr result;
      timestamp fini = this->scatter(sendcount, sendtype, recvcount, recvtype,
          root, comm, vals, result, false);
      return fini;
    }

    //
    // Scatter data with payload.
    //
    timestamp
    mpiapi::scatter(int sendcount, const mpitype &sendtype, int recvcount,
        const mpitype &recvtype, const mpiid &root, const mpicomm::ptr &comm,
        const std::vector<payload::const_ptr> &cc, payload::const_ptr &result,
        bool usingpayload)
    {
      SSTMACBacktrace("MPI_Scatter");
      start_api_call();
      static const char *here = "mpiapi::scatter";
      mpitag scattertag = comm->next_collective_tag();
      assert_initialized(here);

      if (comm->rank() == root)
      {
        if (usingpayload && comm->size().id != (int) cc.size())
        {
          sst_throw(valueerror, "mpiapi: scatter: content size %lu does not match comm size %d",
              cc.size(), comm->size().id);
        }
        validsend(sendcount, sendtype, root, scattertag, comm, here);
      }
      else
      {
        validrecv(recvcount, recvtype, root, scattertag, comm, here);
      }

      timestamp entry = os_->now();
      mpirequest_t req = mpirequest::construct(key::construct(mpiapi::default_key_category));
      strategy_->scatter()->execute(req, queue_, sendcount, sendtype,
          recvcount, recvtype, root, scattertag, comm, cc, os_);

      timestamp fini = queue_->progress_loop(req);

      mpistatus::const_ptr stat = dynamic_pointer_cast<const mpistatus> (
          req->get_status());
      if (!stat)
        throw illformederror("mpiapi::scatter: NULL status pointer.");

      payload::const_ptr collpay = boost::dynamic_pointer_cast<const payload>(
          stat->content());

      if (usingpayload && !collpay)
        throw valueerror("mpiapi:scatter - null mpicollpayload");

      result = collpay;

      if (tracer_active())
        tracer_->scatter(entry, fini, sendcount, sendtype, recvcount, recvtype,
            root, comm);
      end_api_call();
      return fini;
    }

    //
    // Scatter data non-uniformly.
    //
    timestamp
    mpiapi::scatterv(const std::vector<int> &sendcounts,
        const mpitype &sendtype, int recvcount, const mpitype &recvtype,
        const mpiid &root, const mpicomm::ptr &comm)
    {
      std::vector<payload::const_ptr> vals;

      if (comm->rank() == root)
      {
        for (mpiid id = mpiid(0); id < comm->size(); ++id)
          vals.push_back(payload::null());
      }

      payload::const_ptr result;
      timestamp fini = this->scatterv(sendcounts, sendtype, recvcount,
          recvtype, root, comm, vals, result, false);
      return fini;
    }

    //
    // Scatter data non-uniformly.
    //
    timestamp
    mpiapi::scatterv(const std::vector<int> &sendcounts,
        const mpitype &sendtype, int recvcount, const mpitype &recvtype,
        const mpiid &root, const mpicomm::ptr &comm,
        const std::vector<payload::const_ptr> &cc, payload::const_ptr &result,
        bool usingpayload)
    {
      SSTMACBacktrace("MPI_Scatterv");
      start_api_call();
      static const char *here = "mpiapi::scatter";
      mpitag scattertag = comm->next_collective_tag();
      assert_initialized(here);
      if (comm->rank() == root)
      {
        if (comm->size().id != (int) cc.size())
        {
          throw valueerror(
              "mpiapi: scatter - content size does not match comm size");
        }
        validsends(sendcounts, sendtype, root, scattertag, comm, here);
      }
      else
        validrecv(recvcount, recvtype, root, scattertag, comm, here);
      //
      timestamp entry = os_->now();
      mpirequest_t req = mpirequest::construct(key::construct(mpiapi::default_key_category));
      strategy_->scatterv()->execute(req, queue_, sendcounts, sendtype,
          recvcount, recvtype, root, scattertag, comm, cc, os_);

      timestamp fini = queue_->progress_loop(req);

      mpistatus::const_ptr stat = dynamic_pointer_cast<const mpistatus> (
          req->get_status());
      if (!stat)
        throw illformederror("mpiapi::scatter: NULL status pointer.");
      payload::const_ptr collpay = boost::dynamic_pointer_cast<const payload>(
          stat->content());

      if (usingpayload && !collpay)
        throw valueerror("mpiapi:scatter - null mpicollpayload");

      result = collpay;

      if (tracer_active())
        tracer_->scatterv(entry, fini, sendcounts, sendtype, recvcount,
            recvtype, root, comm);
      end_api_call();
      return fini;
    }

    //
    // Gather uniformly distributed data.
    //
    timestamp
    mpiapi::gather(int sendcount, const mpitype &sendtype, int recvcount,
        const mpitype &recvtype, const mpiid &root, const mpicomm::ptr &comm)
    {
      std::vector<payload::const_ptr> rv;
      return gather(sendcount, sendtype, recvcount, recvtype, root, comm,
          payload::null(), rv);
    }

    //
    // Gather uniformly distributed data.
    //
    timestamp
    mpiapi::gather(int sendcount, const mpitype &sendtype, int recvcount,
        const mpitype &recvtype, const mpiid &root, const mpicomm::ptr &comm,
        const payload::const_ptr& content, std::vector<payload::const_ptr> &cc)
    {
      SSTMACBacktrace("MPI_Gather");
      start_api_call();
      static const char *here = "mpiapi::gather";
      mpitag gathertag = comm->next_collective_tag();
      assert_initialized(here);
      if (comm->rank() == root)
        validrecv(recvcount, recvtype, root, gathertag, comm, here);
      else
        validsend(sendcount, sendtype, root, gathertag, comm, here);

      //
      timestamp entry = os_->now();
      mpirequest_t req = mpirequest::construct(key::construct(mpiapi::default_key_category));
      strategy_->gather()->execute(req, queue_, sendcount, sendtype, recvcount,
          recvtype, root, gathertag, comm, content, os_);

      timestamp fini = queue_->progress_loop(req);

      if (comm->rank() == root)
      {
        mpistatus::const_ptr stat = dynamic_pointer_cast<const mpistatus> (
            req->get_status());
        if (!stat)
          throw illformederror("mpiapi::gather: NULL status pointer.");

        if (content)
        {
          mpicollpayload::const_ptr coll = dynamic_pointer_cast<
              const mpicollpayload> (stat->content());
          if (!coll)
            throw illformederror("mpiapi::gather:  NULL mpicollpayload.");
          cc = coll->get_content();
        }

      }
      if (tracer_active())
        tracer_->gather(entry, fini, sendcount, sendtype, recvcount, recvtype,
            root, comm);
      end_api_call();
      return fini;
    }

    //
    // Gather non-uniformly distributed data.
    //
    timestamp
    mpiapi::gatherv(int sendcount, const mpitype &sendtype,
        const std::vector<int> &recvcnts, const mpitype &recvtype,
        const mpiid &root, const mpicomm::ptr &comm)
    {
      std::vector<payload::const_ptr> cc;
      return gatherv(sendcount, sendtype, recvcnts, recvtype, root, comm,
          payload::null(), cc);
    }

    //
    // Gather non-uniformly distributed data.
    //
    timestamp
    mpiapi::gatherv(int sendcount, const mpitype &sendtype,
        const std::vector<int> &recvcnts, const mpitype &recvtype,
        const mpiid &root, const mpicomm::ptr &comm,
        const payload::const_ptr &cc, std::vector<payload::const_ptr> &result)
    {
      SSTMACBacktrace("MPI_Scatterv");
      start_api_call();
      static const char *here = "mpiapi::gatherv";
      mpitag scattertag = comm->next_collective_tag();
      assert_initialized(here);
      if (comm->rank() == root)
        validrecvs(recvcnts, recvtype, root, scattertag, comm, here);
      else
        validsend(sendcount, sendtype, root, scattertag, comm, here);
      //
      timestamp entry = os_->now();
      mpirequest_t req = mpirequest::construct(key::construct(mpiapi::default_key_category));
      strategy_->gatherv()->execute(req, queue_, sendcount, sendtype, recvcnts,
          recvtype, root, scattertag, comm, cc, os_);

      timestamp fini = queue_->progress_loop(req);

      if (comm->rank() == root)
      {
        mpistatus::const_ptr stat = dynamic_pointer_cast<const mpistatus> (
            req->get_status());
        if (!stat)
          throw illformederror("mpiapi::gather: NULL status pointer.");
        mpicollpayload::const_ptr coll = dynamic_pointer_cast<
            const mpicollpayload> (stat->content());
        if (!coll)
          throw illformederror("mpiapi::gather:  NULL mpicollpayload.");

        result = coll->get_content();

      }
      if (tracer_active())
        tracer_->gatherv(entry, fini, sendcount, sendtype, recvcnts, recvtype,
            root, comm);
      end_api_call();
      return fini;
    }

    //
    // Gather uniformly distributed data and make it available to all nodes.
    //
    timestamp
    mpiapi::allgather(int sendcount, const mpitype &sendtype, int recvcount,
        const mpitype &recvtype, const mpicomm::ptr &comm, bool tracethis)
    {
      std::vector<payload::const_ptr> rv;
      return allgather(sendcount, sendtype, recvcount, recvtype, comm,
          payload::null(), rv, tracethis);
    }

    //
    // Gather uniformly distributed data and make it available to all nodes.
    //
    timestamp
    mpiapi::allgather(int sendcount, const mpitype &sendtype, int recvcount,
        const mpitype &recvtype, const mpicomm::ptr &comm,
        const payload::const_ptr& content,
        std::vector<payload::const_ptr> &result, bool tracethis)
    {
      SSTMACBacktrace("MPI_Allgather");
      start_api_call();
      static const char *here = "mpiapi::allgather";
      mpitag mytag = comm->next_collective_tag();
      assert_initialized(here);
      validsend(sendcount, sendtype, comm->rank(), mytag, comm, here);
      validrecv(recvcount, recvtype, comm->rank(), mytag, comm, here);
      //
      timestamp entry = os_->now();
      mpirequest_t req = mpirequest::construct(key::construct(mpiapi::default_key_category));
      strategy_->allgather()->execute(req, queue_, sendcount, sendtype,
          recvcount, recvtype, mytag, comm, content, os_);

      timestamp fini = queue_->progress_loop(req);

      mpistatus::const_ptr stat = req->get_status();
      if (!stat)
      {
        sst_throw(illformederror,
            "mpiapi::allgather: NULL status pointer for %ld",
            (long) stat.get());
      }

      if (content)
      {
        mpicollpayload::const_ptr coll = dynamic_pointer_cast<
            const mpicollpayload> (stat->content());
        if (!coll)
          throw illformederror("mpiapi::allgather:  NULL mpicollpayload.");
        result = coll->get_content();
      }

      if (tracethis && tracer_active())
        tracer_->allgather(entry, fini, sendcount, sendtype, recvcount,
            recvtype, comm);
      end_api_call();
      return fini;
    }

    //
    // Gather non-uniformly distributed data onto all nodes.
    //
    timestamp
    mpiapi::allgatherv(int sendcnt, const mpitype &sendtype,
        const std::vector<int> &recvcnts, const mpitype &recvtype,
        const mpicomm::ptr &comm)
    {
      std::vector<payload::const_ptr> rv;
      return allgatherv(sendcnt, sendtype, recvcnts, recvtype, comm,
          payload::null(), rv);
    }

    //
    // Gather non-uniformly distributed data onto all nodes.
    //
    timestamp
    mpiapi::allgatherv(int sendcnt, const mpitype &sendtype,
        const std::vector<int> &recvcnts, const mpitype &recvtype,
        const mpicomm::ptr &comm, const payload::const_ptr& content,
        std::vector<payload::const_ptr> &result)
    {
      SSTMACBacktrace("MPI_Allgatherv");
      start_api_call();
      static const char *here = "mpiapi::allgatherv";
      mpitag mytag = comm->next_collective_tag();
      assert_initialized(here);
      validsend(sendcnt, sendtype, comm->rank(), mytag, comm, here);
      validsends(recvcnts, recvtype, comm->rank(), mytag, comm, here);
      //
      timestamp entry = os_->now();
      mpirequest_t req = mpirequest::construct(key::construct(mpiapi::default_key_category));

      strategy_->allgatherv()->execute(req, queue_, sendcnt, sendtype,
          recvcnts, recvtype, mytag, comm, content, os_);

      timestamp fini = queue_->progress_loop(req);

      mpistatus::const_ptr stat = dynamic_pointer_cast<const mpistatus> (
          req->get_status());
      if (!stat)
        throw illformederror("mpiapi::allgatherv: NULL status pointer.");

      mpicollpayload::const_ptr coll = dynamic_pointer_cast<
          const mpicollpayload> (stat->content());

      if (!coll)
        throw illformederror("mpiapi::allgatherv:  NULL mpicollpayload.");

      result = coll->get_content();

      if (tracer_active())
        tracer_->allgatherv(entry, fini, sendcnt, sendtype, recvcnts, recvtype,
            comm);
      end_api_call();
      return fini;
    }

    //
    // Send uniformly distributed data from all to all processes.
    //
    timestamp
    mpiapi::alltoall(int sendcount, const mpitype &sendtype, int recvcount,
        const mpitype &recvtype, const mpicomm::ptr &comm)
    {
      /** Create a mass of null content. The all to all later on expects a payload 
       for each node. */
      std::vector<payload::const_ptr> cn(comm->size().id, payload::null());
      std::vector<payload::const_ptr> rv;
      return this->alltoall(sendcount, sendtype, recvcount, recvtype, comm, cn,
          rv);
    }

    //
    // Send uniformly distributed data from all to all processes.
    //
    timestamp
    mpiapi::alltoall(int sendcount, const mpitype &sendtype, int recvcount,
        const mpitype &recvtype, const mpicomm::ptr &comm,
        const std::vector<payload::const_ptr>& content,
        std::vector<payload::const_ptr> &result)
    {
      SSTMACBacktrace("MPI_Alltoall");
      start_api_call();
      static const char *here = "mpiapi::alltoall";
      mpitag a2atag = comm->next_collective_tag();
      assert_initialized(here);
      validsend(sendcount, sendtype, comm->rank(), a2atag, comm, here);
      validrecv(recvcount, recvtype, comm->rank(), a2atag, comm, here);
      //
      timestamp entry = os_->now();
      mpirequest_t req = mpirequest::construct(key::construct(mpiapi::default_key_category));

      strategy_->alltoall()->execute(req, queue_, sendcount, sendtype,
          recvcount, recvtype, a2atag, comm, content, os_);

      timestamp fini = queue_->progress_loop(req);

      mpistatus::const_ptr stat = dynamic_pointer_cast<const mpistatus> (
          req->get_status());

      if (!stat)
        throw illformederror("mpiapi::alltoall: NULL status pointer.");

      mpicollpayload::const_ptr pay =
          dynamic_pointer_cast<const mpicollpayload> (stat->content());

      if (!pay)
        throw illformederror("mpiapi::alltoall:  NULL mpicollpayload.");

      result = pay->get_content();

      if (tracer_active())
        tracer_->alltoall(entry, fini, sendcount, sendtype, recvcount,
            recvtype, comm);
      end_api_call();
      return fini;
    }

    //
    // Send non-uniformly distributed data from all to all processors.
    //
    timestamp
    mpiapi::alltoallv(const std::vector<int> &sendcnts,
        const mpitype &sendtype, const std::vector<int> &recvcnts,
        const mpitype &recvtype, const mpicomm::ptr &comm)
    {
      std::vector<payload::const_ptr> cn(comm->size().id, payload::const_ptr());
      std::vector<payload::const_ptr> rv(comm->size().id, payload::const_ptr());
      return alltoallv(sendcnts, sendtype, recvcnts, recvtype, comm, cn, rv);
    }

    //
    // Send non-uniformly distributed data from all to all processors.
    //
    timestamp
    mpiapi::alltoallv(const std::vector<int> &sendcnts,
        const mpitype &sendtype, const std::vector<int> &recvcnts,
        const mpitype &recvtype, const mpicomm::ptr &comm,
        const std::vector<payload::const_ptr>& content,
        std::vector<payload::const_ptr> &result)
    {
      SSTMACBacktrace("MPI_Alltoallv");
      start_api_call();
      static const char *here = "mpiapi::alltoallv";
      mpitag a2avtag = comm->next_collective_tag();
      assert_initialized(here);
      validsends(sendcnts, sendtype, comm->rank(), a2avtag, comm, here);
      validrecvs(recvcnts, recvtype, comm->rank(), a2avtag, comm, here);
      //
      timestamp entry = os_->now();
      mpirequest_t req = mpirequest::construct(key::construct(mpiapi::default_key_category));
      strategy_->alltoallv()->execute(req, queue_, sendcnts, sendtype,
          recvcnts, recvtype, a2avtag, comm, content, os_);

      timestamp fini = queue_->progress_loop(req);

      mpistatus::const_ptr stat = dynamic_pointer_cast<const mpistatus> (
          req->get_status());
      if (!stat)
        throw illformederror("mpiapi::alltoall: NULL status pointer.");

      mpicollpayload::const_ptr pay =
          dynamic_pointer_cast<const mpicollpayload> (stat->content());

      if (!pay)
        throw illformederror("mpiapi::alltoall:  NULL mpicollpayload.");

      result = pay->get_content();
      if (tracer_active())
        tracer_->alltoallv(entry, fini, sendcnts, sendtype, recvcnts, recvtype,
            comm);
      end_api_call();
      return fini;
    }

    //
    // Reduce data from all nodes onto the root node.
    //
    timestamp
    mpiapi::reduce(int count, const mpitype &type, const mpiop &op,
        const mpiid &root, const mpicomm::ptr &comm)
    {
      payload::const_ptr result;
      return reduce(count, type, op, root, comm, payload::null(), result);
    }

    //
    // Reduce data from all nodes onto the root node.
    //
    timestamp
    mpiapi::reduce(int count, const mpitype &type, const mpiop &op,
        const mpiid &root, const mpicomm::ptr &comm,
        const payload::const_ptr& content, payload::const_ptr &result)
    {
      SSTMACBacktrace("MPI_Reduce");
      start_api_call();
      static const char *here = "mpiapi::reduce";
      mpitag reducetag = comm->next_collective_tag();
      assert_initialized(here);
      validsend(count, type, comm->rank(), reducetag, comm, here);

      //
      timestamp entry = os_->now();
      mpirequest_t req = mpirequest::construct(key::construct(mpiapi::default_key_category));
      strategy_->reduce()->execute(req, queue_, count, type, op, root,
          reducetag, comm, content, os_);

      timestamp fini = queue_->progress_loop(req);

      if (comm->rank() == root)
      {
        mpistatus::const_ptr stat = req->get_status();
        if (!stat)
          throw illformederror(
              "mpiapi::bcast: NULL status pointer after bcast.");

        result = stat->content();
      }
      else
      {
        result = payload::const_ptr();
      }
      if (tracer_active())
        tracer_->reduce(entry, fini, count, type, op, root, comm);
      end_api_call();
      return fini;
    }

    //
    // Reduce data from all nodes to all nodes.
    //
    timestamp
    mpiapi::allreduce(int count, const mpitype &type, const mpiop &op,
        const mpicomm::ptr &comm, bool tracethis)
    {
      payload::const_ptr result;
      return allreduce(count, type, op, comm, payload::null(), result,
          tracethis);
    }

    timestamp
    mpiapi::allreduce(int count, const mpitype &type, const mpiop &op,
        const mpicomm::ptr &comm, const payload::const_ptr& content,
        payload::const_ptr &result, bool tracethis)
    {
      SSTMACBacktrace("MPI_Allreduce");
      start_api_call();
      static const char *here = "mpiapi::allreduce";
      mpitag allreducetag = comm->next_collective_tag();
      assert_initialized(here);
      validsend(count, type, comm->rank(), allreducetag, comm, here);
      //
      timestamp entry = os_->now();
      mpirequest_t req = mpirequest::construct(key::construct(mpiapi::default_key_category));
      strategy_->allreduce()->execute(req, queue_, count, type, op,
          allreducetag, comm, content, os_);

      timestamp fini = queue_->progress_loop(req);

      mpistatus::const_ptr stat = req->get_status();

      if (!stat)
        throw illformederror("mpiapi::bcast: NULL status pointer after bcast.");

      result = stat->content();

      if (tracethis && tracer_active())
        tracer_->allreduce(entry, fini, count, type, op, comm);
      end_api_call();
      return fini;
    }

    //
    // Reduce data and scatter results.
    //
    timestamp
    mpiapi::reduce_scatter(const std::vector<int> &recvcnts,
        const mpitype &type, const mpiop &op, const mpicomm::ptr &comm)
    {
      payload::const_ptr content = payload::null();
      payload::const_ptr result = payload::null();
      return reduce_scatter(recvcnts, type, op, comm, content, result);
    }

    //
    // Reduce data and scatter results.
    //
    timestamp
    mpiapi::reduce_scatter(const std::vector<int> &recvcnts,
        const mpitype &type, const mpiop &op, const mpicomm::ptr &comm,
        const payload::const_ptr& content, payload::const_ptr &result)
    {
      SSTMACBacktrace("MPI_Reducescatter");
      start_api_call();
      static const char *here = "mpiapi::reduce_scatter";
      static mpitag rscattag(0xFFAA1A);
      assert_initialized(here);
      validsends(recvcnts, type, comm->rank(), rscattag, comm, here);
      //
      timestamp entry = os_->now();
      mpirequest_t req = mpirequest::construct(key::construct(mpiapi::default_key_category));
      strategy_->reduce_scatter()->execute(req, queue_, recvcnts, type, op,
          rscattag, comm, content, os_);

      timestamp fini = queue_->progress_loop(req);

      mpistatus::const_ptr stat = req->get_status();
      if (!stat)
        throw illformederror("mpiapi::reduce_scatter: NULL status pointer.");
      result = stat->content();
      if (!result)
        throw illformederror("mpiapi::reduce_scatter:  NULL mpicollpayload.");
      if (tracer_active())
        tracer_->reduce_scatter(entry, fini, recvcnts, type, op, comm);
      end_api_call();
      return fini;
    }

    //
    // Compute partial reductions on data.
    //
    timestamp
    mpiapi::scan(int count, const mpitype &type, const mpiop &op,
        const mpicomm::ptr &comm)
    {
      payload::const_ptr content = payload::null();
      payload::const_ptr result = payload::null();
      return this->scan(count, type, op, comm, content, result);
    }

    //
    // Compute partial reductions and aggregate data.
    //
    timestamp
    mpiapi::scan(int count, const mpitype &type, const mpiop &op,
        const mpicomm::ptr &comm, const payload::const_ptr& content,
        payload::const_ptr &result)
    {
      SSTMACBacktrace("MPI_Scan");
      start_api_call();

      mpitag scantag = comm->next_collective_tag();
      assert_initialized("mpiapi::scan");
      validsend(count, type, comm->rank(), scantag, comm, "mpiapi::scan");
      //
      timestamp entry = os_->now();
      mpirequest_t req = mpirequest::construct(key::construct(mpiapi::default_key_category));
      strategy_->scan()->execute(req, queue_, count, type, op, scantag, comm,
          content, os_);

      timestamp fini = queue_->progress_loop(req);

      mpistatus::const_ptr stat = dynamic_pointer_cast<const mpistatus> (
          req->get_status());
      if (!stat)
        throw illformederror("mpiapi::scan: NULL status pointer.");
      result = stat->content();

      if (tracer_active())
        tracer_->scan(entry, fini, count, type, op, comm);
      end_api_call();
      return fini;
    }

    //----------------------------------------------------------------
    // --- MPI Derived Datatype (experimental)
    //----------------------------------------------------------------

    //
    // Creates a contiguous datatype
    //
    timestamp
    mpiapi::type_contiguous(int count, const mpitype &old_type,
        mpitype &new_type)
    {

      mpitype ret = mpitype("contiguous-" + old_type.label, old_type, count, 1,
          1, true, MPI_COMBINER_CONTIGUOUS);

      new_type = ret;
      derived_types_[new_type.id] = new_type;
      return os_->now();
    }

    /// Creates a vector (strided) datatype
    timestamp
    mpiapi::type_vector(int count, int blocklength, int stride,
        const mpitype &old_type, mpitype &new_type, bool stride_in_elem)
    {

      int comb = (stride_in_elem) ? MPI_COMBINER_VECTOR : MPI_COMBINER_HVECTOR;
      std::stringstream ss;
      ss << "vector-" << old_type.label << "\n";
      mpitype ret = mpitype(ss.str(), old_type, count, blocklength, stride,
          stride_in_elem, comb);

      new_type = ret;
      derived_types_[new_type.id] = new_type;
      return os_->now();
    }

    //
    // Creates a struct datatype
    //
    timestamp
    mpiapi::type_struct(const int count, const std::vector<int> &blocklens,
        const std::vector<int> &indices, const std::vector<mpitype> &old_types,
        mpitype &newtype)
    {
      if (count > 0)
      {

        inddata::ptr idata = inddata::construct();
        int maxdisp = -1;
        int maxindex = 0;
        int typesizes = 0;
        int maxsize = 0;
        int mindisp = INT_MAX;
        int index = 0;
        int ub = -1;
        int lb = -1;
        int align = 0;
        bool ubset = false;
        bool lbset = false;
        for (int i = 0; i < count; i++)
        {
          if (old_types[i] == mpitype::mpi_lb)
          {
            lb = indices[i];
            lbset = true;
          }
          else if (old_types[i] == mpitype::mpi_ub)
          {
            ub = indices[i];
            ubset = true;

            //if (ub > maxdisp)
            // {
            //  maxdisp = ub;
            // maxindex = i;
            // maxsize = 0;
            // }

          }
          else if (blocklens[i] > 0)
          {
            idata->blocks[index] = ind_block();
            idata->blocks[index].base = old_types[i];
            idata->blocks[index].disp = indices[i];
            idata->blocks[index].num = blocklens[i];
            typesizes += old_types[i].get_packed_size() * blocklens[i];
            if (indices[i] > maxdisp)
            {
              maxdisp = indices[i];
              maxindex = i;
              maxsize = old_types[i].get_extent();
            }

            if (indices[i] < mindisp)
            {
              mindisp = indices[i];
            }

            if (old_types[i].get_type() == mpitype::PRIM)
            {
              align = std::max(align, old_types[i].get_align());
            }
            index++;
          }
        }

        if (!lbset)
        {
          lb = mindisp;
        }
        if (!ubset)
        {
          if (index == 0)
          { //there was no data
            ub = 0;
          }
          else
            ub = maxdisp + maxsize * blocklens[maxindex];
          if (align > 0)
          {
            while ((double) ub / (double) align != ub / align)
            {
              ub++;
            }
          }

        }
        idata->ub_ = ub;
        idata->lb_ = lb;
        idata->mindisp_ = mindisp;
        idata->maxbyte_ = maxdisp + maxsize * blocklens[maxindex];

        mpitype ret = mpitype("struct", idata, typesizes, ub - lb,
            index * 2 + 1, index, MPI_COMBINER_STRUCT);

        SSTMAC_DEBUG << "mpiapi: building struct, maxdisp: " << maxdisp
              << ", maxsize: " << maxsize << ", block len of max: "
              << blocklens[maxindex] << ", ub: " << ub << ", lb: " << lb
              << ", extent: " << ret.get_extent() << "\n";

        newtype = ret;
        derived_types_[newtype.id] = newtype;
      }
      else
      {
        inddata::ptr idata = inddata::construct();
        idata->ub_ = 0;
        idata->lb_ = 0;
        idata->mindisp_ = 0;
        idata->maxbyte_ = 0;

        mpitype ret = mpitype("struct", idata, 0, 0, 1, 0, MPI_COMBINER_STRUCT);

        derived_types_[ret.id] = ret;
        newtype = ret;
      }
      return os_->now();
    }

    int
    mpiapi::type_size(const mpitype &type)
    {
      return type.get_packed_size();
    }

    timestamp
    mpiapi::type_dup(const mpitype &intype, mpitype &outtype)
    {
      outtype = intype.clone();
      if (outtype.get_comb() == MPI_COMBINER_NAMED)
        outtype.set_comb(MPI_COMBINER_DUP);
      derived_types_[outtype.id] = outtype;
      return this->wtime();
    }

    timestamp
    mpiapi::type_set_name(long id, const std::string &name)
    {
      if (derived_types_.find(id) != derived_types_.end())
      {
        derived_types_[id].label = name;
      }
      return this->wtime();
    }

    timestamp
    mpiapi::type_indexed(int count, int lens[], const std::vector<int> &ind,
        const mpitype &intype, mpitype &outtype, bool in_elem, int comb)
    {

      if (count > 0)
      {
        inddata::ptr idata = inddata::construct();
        int maxdisp = -1;
        int maxindex = 0;
        int mindisp = INT_MAX;
        int index = 0;
        int typesizes = 0;
        for (int i = 0; i < count; i++)
        {
          if (lens[i] > 0)
          {
            int bytesdisp;
            if (in_elem)
              bytesdisp = ind[i] * intype.get_extent();
            else
              bytesdisp = ind[i];

            idata->blocks[index] = ind_block();
            idata->blocks[index].base = intype;
            idata->blocks[index].disp = bytesdisp;
            idata->blocks[index].num = lens[i];
            typesizes += lens[i] * intype.get_packed_size();
            //  size += intype.size * lens[i];
            if (bytesdisp > maxdisp)
            {
              maxdisp = bytesdisp;
              maxindex = i;
            }

            if (bytesdisp < mindisp)
            {
              mindisp = bytesdisp;
            }
            index++;
          }

        }

        idata->ub_ = maxdisp + intype.get_extent() * lens[maxindex];
        idata->lb_ = mindisp;
        idata->mindisp_ = mindisp;
        idata->maxbyte_ = maxdisp + intype.get_extent() * lens[maxindex];

        mpitype ret = mpitype(intype.label, idata, typesizes,
            idata->ub_ - idata->lb_, index * 2 + 1, 1, comb);

        SSTMAC_DEBUG << "mpiapi: building indexed, maxdisp: " << maxdisp
              << ", maxsize: " << idata->ub_ << ", block len of max: "
              << lens[maxindex] << ", size: " << ret.get_packed_size()
              << ", extent: " << ret.get_extent() << "\n";

        outtype = ret;
        derived_types_[outtype.id] = outtype;

      }
      else
      {
        inddata::ptr idata = inddata::construct();
        idata->ub_ = 0;
        idata->lb_ = 0;
        idata->mindisp_ = 0;
        idata->maxbyte_ = 0;

        mpitype ret = mpitype("struct", idata, 0, 0, 1, 0, comb);
        derived_types_[ret.id] = ret;
        outtype = ret;
      }

      return os_->now();
    }

    //
    // A datatype object has to be committed before use in communication.
    //
    timestamp
    mpiapi::type_commit(const mpitype& type)
    {
      commited_[type.id] = true;
      return os_->now();
    }

    //
    // Mark datatype for deallocation.
    //
    timestamp
    mpiapi::type_free(const mpitype& type)
    {
      commited_[type.id] = false;
      return os_->now();
    }

    //
    // Get the derived mpitype mapped to an id
    //
    const mpitype&
    mpiapi::get_derived_type(unsigned long id)
    {
      boost::unordered_map<long, mpitype>::iterator iter;
      iter = derived_types_.find(id);
      if (iter == derived_types_.end())
      {
        throw invalidkeyerror("derived type not found");
      }
      return derived_types_[id];
    }

    const mpiop&
    mpiapi::get_created_op(unsigned long id)
    {
      if (created_op_.find(id) == created_op_.end())
      {
        sst_throw(valueerror, "mpiapi::get_created_op - could not find id %lu", id);
      }

      return created_op_[id];
    }

    timestamp
    mpiapi::op_create(void
    (*func)(void *, void *, int *, long *), int comm, mpiop &output)
    {
      mpiop o = mpiop(func, comm);
      created_op_[o.id] = o;
      output = o;
      return os_->now();
    }

    timestamp
    mpiapi::op_free(long id)
    {
      created_op_.erase(id);
      return os_->now();
    }

    timestamp
    mpiapi::win_create(void* base, size_t size, int disp, mpiinfo::ptr inf,
        const mpicomm::ptr &comm, mpi_window::ptr &win, long id)
    {
      SSTMACBacktrace("MPI_Win_create");
      mpi_window::ptr ret = mpi_window::construct(base, size, disp, inf, comm,
          id);

      barrier(comm, false);

      win = ret;

      queue_->win_create(ret);
      end_api_call();
      return os_->now();
    }

    timestamp
    mpiapi::win_free(mpi_window::ptr win)
    {
      static const char *here = "mpiapi::win_put";
      assert_initialized(here);

      queue_->win_free(win);

      return os_->now();
    }

    timestamp
    mpiapi::win_get(mpirequest_t &req, int origin_count,
        const mpitype &origin_datatype, const mpiid &target_rank,
        MPI_Aint target_disp, int target_count, const mpitype &target_datatype,
        const mpi_window::ptr &win)
    {
      SSTMACBacktrace("MPI_Win_get");
      start_api_call();
      static const char *here = "mpiapi::win_put";
      assert_initialized(here);

      timestamp entry = os_->now();
      key::ptr block_key = key::construct(mpiapi::default_key_category);

      req = mpirequest::construct(block_key);
      eventhandler::ptr completion = queue_->progress_done_handler(os_, req);

      mpi_rma_message::op_info op;
      op.disp_ = target_disp;
      op.count_ = target_datatype.get_packed_size() * target_count;
      op.type_ = mpi_rma_message::get;
      op.epoch_ = win->get_current_epoch();
      op.target_ = target_rank;

      queue_->get(req, origin_count, origin_datatype, target_rank, win->comm(),
          win, op, completion);

      pending_rmas_[win][req] = op;
      end_api_call();
      return os_->now();
    }

    timestamp
    mpiapi::win_put(mpirequest_t &req, const payload::const_ptr &load,
        int target_count, const mpitype &target_datatype,
        const mpiid &target_rank, long long target_disp,
        const mpi_window::ptr &win)
    {
      SSTMACBacktrace("MPI_Win_put");
      start_api_call();
      static const char *here = "mpiapi::win_put";
      assert_initialized(here);

      timestamp entry = os_->now();
      key::ptr block_key = key::construct(mpiapi::default_key_category);

      req = mpirequest::construct(block_key);
      eventhandler::ptr completion = queue_->progress_done_handler(os_, req);

      mpi_rma_message::op_info op;
      op.disp_ = target_disp;
      op.count_ = target_datatype.get_packed_size() * target_count;
      op.type_ = mpi_rma_message::put;
      op.epoch_ = win->get_current_epoch();
      op.target_ = target_rank;

      queue_->put(req, target_count, target_datatype, target_rank, win->comm(),
          win, op, completion, load);

      pending_rmas_[win][req] = op;
      end_api_call();
      return os_->now();
    }

    timestamp
    mpiapi::win_accumulate(mpirequest_t &req, const payload::const_ptr &load,
        int target_count, const mpitype &target_datatype,
        const mpiid &target_rank, long long target_disp, const mpiop &o,
        const mpi_window::ptr &win)
    {
      SSTMACBacktrace("MPI_Accumulate");
      start_api_call();
      static const char *here = "mpiapi::win_accumulate";
      assert_initialized(here);

      timestamp entry = os_->now();
      key::ptr block_key = key::construct(mpiapi::default_key_category);

      req = mpirequest::construct(block_key);
      eventhandler::ptr completion = queue_->progress_done_handler(os_, req);

      mpi_rma_message::op_info op;
      op.disp_ = target_disp;
      op.count_ = target_datatype.get_packed_size() * target_count;
      op.type_ = mpi_rma_message::acc;
      op.epoch_ = win->get_current_epoch();
      op.target_ = target_rank;
      op.op_ = o;

      queue_->put(req, target_count, target_datatype, target_rank, win->comm(),
          win, op, completion, load);

      pending_rmas_[win][req] = op;
      end_api_call();
      return os_->now();
    }

    timestamp
    mpiapi::win_fence(const mpi_window::ptr &win, bool barr,
        boost::unordered_map<mpirequest_t, payload::const_ptr> &loads)
    {
      SSTMACBacktrace("MPI_Win_fence");
      start_api_call();
      static const char *here = "mpiapi::win_fence";
      assert_initialized(here);

      std::queue<mpirequest_t> toremove;

      boost::unordered_map<mpirequest_t, mpi_rma_message::op_info>::iterator
          it, end = pending_rmas_[win].end();
      for (it = pending_rmas_[win].begin(); it != end; it++)
      {
        if (it->second.epoch_ == win->get_current_epoch())
        {
          if (!(it->first)->is_complete())
          {
            os_->block((it->first)->get_key());
          }
          mpistatus::const_ptr stat = (it->first)->get_status();

          if (it->second.type_ == mpi_rma_message::get)
          {

            if (!stat)
              throw illformederror(
                  "mpiapi::win_fence: NULL status pointer after get.");

            loads[it->first] = stat->content();
          }
          toremove.push(it->first);
        }

      }

      while (toremove.size() > 0)
      {
        pending_rmas_[win].erase(toremove.front());
        toremove.pop();
      }

      if (barr)
        barrier(win->comm(), false);

      win->next_epoch();
      end_api_call();
      return os_->now();
    }

    timestamp
    mpiapi::win_lock(LOCK_TYPE t, const mpiid &rank, WIN_FLAGS flag,
        const mpi_window::ptr &win)
    {
      SSTMACBacktrace("MPI_Win_lock");
      start_api_call();
      static const char *here = "mpiapi::win_lock";
      assert_initialized(here);

      // if(flag != mpi_win_mode_nocheck_){
      // queue_->win_lock(t, rank, win);
      // }

      end_api_call();
      return os_->now();
    }

    timestamp
    mpiapi::win_unlock(const mpiid &rank, const mpi_window::ptr &win,
        boost::unordered_map<mpirequest_t, payload::const_ptr> &loads)
    {
      SSTMACBacktrace("MPI_Win_unlock");
      start_api_call();
      static const char *here = "mpiapi::win_unlock";
      assert_initialized(here);

      std::queue<mpirequest_t> toremove;

      boost::unordered_map<mpirequest_t, mpi_rma_message::op_info>::iterator
          it, end = pending_rmas_[win].end();
      for (it = pending_rmas_[win].begin(); it != end; it++)
      {
        if (it->second.epoch_ == win->get_current_epoch() && it->second.target_
            == rank)
        {
          if (!(it->first)->is_complete())
          {
            os_->block((it->first)->get_key());
          }
          mpistatus::const_ptr stat = (it->first)->get_status();

          if (it->second.type_ == mpi_rma_message::get)
          {

            if (!stat)
              throw illformederror(
                  "mpiapi::win_fence: NULL status pointer after get.");

            loads[it->first] = stat->content();
          }

          toremove.push(it->first);
        }

      }

      while (toremove.size() > 0)
      {
        pending_rmas_[win].erase(toremove.front());
        toremove.pop();
      }

      win->next_epoch();
      end_api_call();
      return os_->now();

    }
  }
} // end of namespace sstmac

