/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_MPI_MPIAPI_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_MPI_MPIAPI_H_INCLUDED

#include <sstmac/common/messages/payload.h>
#include <sstmac/common/sim_parameters.h>

#include <sstmac/software/libraries/library.h>
#include <sstmac/software/api/api.h>

#include <sstmac/software/api/mpi/mpi_comm/mpicomm.h>
#include <sstmac/software/api/mpi/mpi_types/mpiid.h>
#include <sstmac/software/api/mpi/mpi_types/mpitag.h>
#include <sstmac/software/api/mpi/mpi_types/mpitype.h>
#include <sstmac/software/api/mpi/mpi_types/mpiop.h>
#include <sstmac/software/api/mpi/mpi_types/mpiinfo.h>
#include <sstmac/software/api/mpi/rma/mpi_window.h>

#include <sstmac/software/process/softwareid.h>

#include <sstmac/software/api/mpi/mpirequest.h>
#include <sstmac/software/api/mpi/mpistatus.h>
#include <sstmac/software/api/mpi/mpi_queue/mpiqueue.h>
#include <sstmac/software/api/mpi/mpi_strategies/mpistrategy.h>
#include <sstmac/software/api/mpi/mpi_comm/mpicommfactory.h>
#include <sstmac/software/api/mpi/mpitracer.h>

#include <sstmac/software/services/launch/allocation/hostname_allocation.h>

#include <sstmac/common/debug.h>
#include <sstmac/common/factories/factory.h>

#include <sstmac/software/process/key.h>
#include <sstmac/software/process/pmi.h>

namespace sstmac
{
  namespace sw
  {

    // Forward declaration of relevant types.

    ////class nodeinterface;
    class OS_API;

   // class key;
    class mpitracer;

    /**
     * All the joy of using MPI from an application (or rather, a thread).
     */
    class mpiapi :
        public api,
        public process_manager
    {
      /// This friendship declaration can probably be eliminated later.
      friend class mpilaunch;

      /// Nested classes to take care of persistent communications.
    public:
      class persistent;
    private:
      class persistentsend;
      class persistentrecv;

    public:
      static key::category default_key_category;
      static key::category poll_key_category;
      static key::category memcpy_key_category;

      typedef boost::intrusive_ptr<mpiapi> ptr;

      static std::string tracername_;

      int
      save_payload(payload::const_ptr load)
      {
        int ret = has_saved_payload();
        saved_payload_ = load;
        return ret;
      }

      payload::const_ptr
      get_saved_payload()
      {
        return saved_payload_;
      }

      bool
      has_saved_payload()
      {
        int ret = (saved_payload_ != payload::null());
        return ret;
      }

      void
      unregister();

      static void
      clear_statics();

    protected:
      static debug dbg_;

      static boost::unordered_map<int64_t, std::string> hostmap_;

      softwareid id_;

      payload::const_ptr saved_payload_;

    private:
      bool
      tracer_active() const;

      static mpitracer::ptr parent_tracer_;

      /// The MPI call tracer.  If this is NULL, no tracing will be performed.
      mpitracer::ptr tracer_;

      /// The MPI server.
      mpiqueue::ptr queue_;

      //// My MPI index in the world.
      mpiid rank_;

      /// The strategy object that gets consulted for how to do anything.
      mpistrategy::ptr strategy_;

      /// The builder for mpi communicators.
      mpicommfactory::ptr comm_factory_;

      /// Ensure that collective operations are given unique tags
      /// to protect shared back-end data (magic payload info).

      /// The state of this object (initialized or not).
      enum
      {
        is_fresh, is_initialized, is_finalizing, is_finalized
      } status_;

      /// MPI_COMM_WORLD.
      mpicomm::ptr worldcomm_;

      /// MPI_COMM_SELF.
      mpicomm::ptr selfcomm_;


      std::vector<mpirequest_t> pending_bsends_;

      /// Ensure that MPI is properly initialized.
      inline void
      assert_initialized(const char *caller) const;

      /// Nonblocking send -- actual implementation.
      /// wait for this request has the same completion as send.
      /// \return time when the request has been registered (likely current time).
      timestamp
      do_isend(int count, const mpitype &type, const mpiid &target,
          const mpitag &tag, const mpicomm::ptr &comm, const mpirequest_t &req,
          const payload::const_ptr& content);

      /// Nonblocking send -- actual implementation.
      /// wait for this request has the same completion as send.
      /// \return time when the request has been registered (likely current time).
      timestamp
      do_ibsend(int count, const mpitype &type, const mpiid &target,
          const mpitag &tag, const mpicomm::ptr &comm, const mpirequest_t &req,
          const payload::const_ptr& content);

      /// Nonblocking send -- actual implementation.
      /// wait for this request has the same completion as send.
      /// \return time when the request has been registered (likely current time).
      timestamp
      do_issend(int count, const mpitype &type, const mpiid &target,
          const mpitag &tag, const mpicomm::ptr &comm, const mpirequest_t &req,
          const payload::const_ptr& content);

      /// Nonblocking send -- actual implementation.
      /// wait for this request has the same completion as send.
      /// \return time when the request has been registered (likely current time).
      timestamp
      do_irsend(int count, const mpitype &type, const mpiid &target,
          const mpitag &tag, const mpicomm::ptr &comm, const mpirequest_t &req,
          const payload::const_ptr& content);

      /// Asynchronous receive -- actual implementation.
      /// wait for this request has the same completion as recv.
      /// \return time at which the receive request has been registered.
      timestamp
      do_irecv(int count, const mpitype &type, const mpiid &source,
          const mpitag &tag, const mpicomm::ptr &comm, const mpirequest_t &req);

    protected:
      /// Build a new mpiapi.
      mpiapi(const sim_parameters::ptr& params);

    public:
      /// Hello.
      static mpiapi::ptr
      construct(const sim_parameters::ptr& params);

      /// Adieu.
      virtual
      ~mpiapi();

      void
      init(const softwareid& id);

      void
      init_os(const OS_API::ptr &os);

      void
      init_factory_params(const sim_parameters::ptr &params);

      timestamp starttime_;
      timestamp runtime_;

      /// Direct access to the mpiqueue
      mpiqueue::ptr
      queue()
      {
        return queue_;
      }

      /*
       * Methods exposing MPI calls and other key methods to
       * derived objects.
       */

      /// Get the world communicator.
      const mpicomm::ptr
      comm_world() const;

      /// Get the self communicator.
      const mpicomm::ptr
      comm_self() const;

      /* Utility functions to mirror MPI API functionality */

      /// The rank of this node in a communicator.
      inline mpiid
      comm_rank(const mpicomm::ptr &comm)
      {
        return comm->rank();
      }

      /// The size of a communicator.
      inline mpiid
      comm_size(const mpicomm::ptr &comm)
      {
        return comm->size();
      }

      void
      attr_put(mpicomm::ptr comm, keyval::ptr k, void* val);

      void
      attr_get(mpicomm::ptr comm, keyval::ptr k, void* val, int *flag);

      void
      mpi_cart_create(mpicomm::ptr &comm, int ndims, int *dims, int *periods,
          int reorder, mpicomm::ptr &outcomm);

      void
      mpi_cart_get(mpicomm::ptr &comm, int maxdims, int *dims, int *periods,
          int *coords);

      void
      mpi_cartdim_get(mpicomm::ptr &comm, int *ndims);

      void
      mpi_cart_rank(mpicomm::ptr comm, int *coords, int *rank);

      void
      mpi_cart_shift(mpicomm::ptr comm, int direction, int displ, int *source,
          int *dest);

      void
      mpi_cart_coords(mpicomm::ptr comm, int rank, int maxdims, int *coords);

      /// Get the number of elements received assuming a specific type.
      int
      get_count(const_mpistatus_t status, const mpitype &type);

      /// Split a communicator.  This one is a little weird.
      timestamp
      comm_split(const mpicomm::ptr &incomm, int color, int key,
          mpicomm::ptr &outcomm);

      /// Get type size.
      int
      type_size(const mpitype &type);

      /* Set up and tear down */

      /// Test whether MPI has bee initialized.
      bool
      initialized();

      /// Test whether MPI has been finalized.
      bool
      finalized();

      /// Initialize MPI.  Must be the first MPI call made other than
      /// initialized or finalized.
      timestamp
      init();

      /// Finalize MPI.  Must be the last MPI call other than
      /// initialized or finalized.
      timestamp
      finalize();

      /// Get current time.
      timestamp
      wtime();

      /* Create and destroy communicators. */

      /// Duplicate a communicator.
      timestamp
      comm_dup(const mpicomm::ptr &input, mpicomm::ptr &output);

      /// Create a communicator containing a subset of an existing comm.
      timestamp
      comm_create(const mpicomm::ptr &input, mpigroup::ptr group,
          mpicomm::ptr &output);

      /// Destroy a communicator.  This is currently a noop, but should later
      /// mark the communicator invalid so erroneous program behavior can be
      /// detected.
      timestamp
      comm_free(mpicomm::ptr input);

      /* Basic point-to-point operations. */

      /// Blocking send.
      /// \return time at which the (imaginary) input buffer can be reused.
      timestamp
      send(int count, const mpitype &type, const mpiid &target,
          const mpitag &tag, const mpicomm::ptr &comm,
          const payload::const_ptr& content = payload::null());

      /// Blocking send using a user-defined buffer (see buffer_attach).
      /// \return time at which it is safe to use the (hypothetical) input array.
      timestamp
      bsend(int count, const mpitype &type, const mpiid &target,
          const mpitag &tag, const mpicomm::ptr &comm,
          const payload::const_ptr& content = payload::null());

      /// Blocking syncronous send.
      /// \return time when a matching recv (or irecv) has acknowledged receipt.
      timestamp
      ssend(int count, const mpitype &type, const mpiid &target,
          const mpitag &tag, const mpicomm::ptr &comm,
          const payload::const_ptr& content = payload::null());

      /// Ready send (as in "ready or not, here comes the data").
      /// \return time when it is safe to use the input array.
      timestamp
      rsend(int count, const mpitype &type, const mpiid &target,
          const mpitag &tag, const mpicomm::ptr &comm,
          const payload::const_ptr& content = payload::null());

      /// Nonblocking send.
      /// wait for this request has the same completion as send.
      /// \return time when the request has been registered (likely current time).
      timestamp
      isend(int count, const mpitype &type, const mpiid &target,
          const mpitag &tag, const mpicomm::ptr &comm, mpirequest_t &req,
          const payload::const_ptr& content = payload::null());

      /// Nonblocking user-buffered send.
      /// wait for this request has the same completion as bsend.
      /// \return time when the request has been registered (likely current time).
      timestamp
      ibsend(int count, const mpitype &type, const mpiid &target,
          const mpitag &tag, const mpicomm::ptr &comm, mpirequest_t &req,
          const payload::const_ptr& content = payload::null());

      /// Nonblocking synchronous send.
      /// wait for this request has the same completion as ssend.
      timestamp
      issend(int count, const mpitype &type, const mpiid &target,
          const mpitag &tag, const mpicomm::ptr &comm, mpirequest_t &req,
          const payload::const_ptr& content = payload::null());

      /// Nonblocking ready send.
      /// wait for this request has the same completion as rsend.
      timestamp
      irsend(int count, const mpitype &type, const mpiid &target,
          const mpitag &tag, const mpicomm::ptr &comm, mpirequest_t &req,
          const payload::const_ptr& content = payload::null());

      /// Blocking receive.
      /// Note that the message content (may be null) is available
      /// via stat.content().
      /// \return time at which the input data is available for use.
      timestamp
      recv(int count, const mpitype &type, const mpiid &source,
          const mpitag &tag, const mpicomm::ptr &comm, const_mpistatus_t &stat);

      /// Asynchronous receive.
      /// wait for this request has the same completion as recv.
      /// \return time at which the receive request has been registered.
      timestamp
      irecv(int count, const mpitype &type, const mpiid &source,
          const mpitag &tag, const mpicomm::ptr &comm, mpirequest_t &req);

      /// Send/receieve exchange.
      /// \return time at which the request has been completed.
      timestamp
      sendrecv(int sendcount, const mpitype &sendtype, const mpiid &target,
          const mpitag &sendtag, int recvcount, const mpitype &recvtype,
          const mpiid &source, const mpitag &recvtag, const mpicomm::ptr &comm,
          const_mpistatus_t &stat,
          const payload::const_ptr& content = payload::null());

      /// Send and receive using a single buffer.
      timestamp
      sendrecv_replace(int count, const mpitype &type, const mpiid &dest,
          const mpitag &sendtag, const mpiid &source, const mpitag &recvtag,
          const mpicomm::ptr &comm, const_mpistatus_t &stat);

      /* Stepwise point-to-point (not implemented, but
       * should be fairly straight-forward to complete). */

      /// Initiate a send.
      timestamp
      send_init(int count, const mpitype &type, const mpiid &target,
          const mpitag &tag, const mpicomm::ptr &comm, mpirequest_t &req,
           void* buf = NULL);

      /// Initiate a buffered send.
      timestamp
      bsend_init(int count, const mpitype &type, const mpiid &target,
          const mpitag &tag, const mpicomm::ptr &comm, mpirequest_t &req,
           void* buf = NULL);

      /// Initiate a synchronous send.
      timestamp
      ssend_init(int count, const mpitype &type, const mpiid &target,
          const mpitag &tag, const mpicomm::ptr &comm, mpirequest_t &req,
           void* buf = NULL);

      /// Initiate a ready send.
      timestamp
      rsend_init(int count, const mpitype &type, const mpiid &target,
          const mpitag &tag, const mpicomm::ptr &comm, mpirequest_t &req,
           void* buf = NULL);

      /// Initiate a receive.
      timestamp
      recv_init(int count, const mpitype &type, const mpiid &target,
          const mpitag &tag, const mpicomm::ptr &comm, mpirequest_t &req);

      /// Start a previously initiated request.
      timestamp
      start(mpirequest_t req);

      /// Start a collection of previously initiated requests.
      timestamp
      startall(std::vector<mpirequest_t> &requests);

      /// Cancel a request that had been previously initiated.
      timestamp
      cancel(mpirequest_t req);

      /// Test whether a request has been cancelled (takes no time).
      bool
      test_cancelled(const const_mpistatus_t &req);

      /* Completion of outstanding requests */

      /// Wait for a request to finish.
      /// \return the completion time for the call (current time).
      timestamp
      wait(mpirequest_t req, const_mpistatus_t &stat);

      /// Wait for all the given requests to finish.
      /// \return the completion time for the call (current time).
      timestamp
      waitall(std::vector<mpirequest_t> &req,
          std::vector<const_mpistatus_t> &statuses, bool tracethis = true);

      /// Wait for at least one of the given requests to complete.
      /// \param index is the index of the first request identified as complete.
      ///        This does not have to be the request with the earliest completion
      /// \return the completion time for the call (current time).
      timestamp
      waitany(std::vector<mpirequest_t> &req, int &index,
          const_mpistatus_t &stat);

      timestamp
      waitany(std::vector<mpirequest_t> &req, int &index,
          const_mpistatus_t &stat, timestamp timeout);

      /// Wait for some of the given requests to complete.
      /// \param indices are the request indices that have completed.
      /// \return the completion time for the call (current time).
      timestamp
      waitsome(std::vector<mpirequest_t> &req, std::vector<int> &indices,
          std::vector<const_mpistatus_t> &statuses);

      /* Testing of outstanding requests */

      timestamp
      request_free(mpirequest_t req);

      /// Test whether a request has finished.
      /// \return the completion time for the call (current time).
      timestamp
      test(mpirequest_t req, bool &flag, const_mpistatus_t &stat);

      /// Test whether all the given requests have finished.
      /// \return the completion time for the call (current time).
      timestamp
      testall(std::vector<mpirequest_t> &req, bool &flag,
          std::vector<const_mpistatus_t> &statuses);

      /// Test whether at least one of the given requests has completed.
      /// \param index is the index of the first request identified as complete.
      ///        This does not have to be the request with the earliest completion
      /// \return the completion time for the call (current time).
      timestamp
      testany(std::vector<mpirequest_t> &req, int &index, bool &flag,
          const_mpistatus_t &stat);

      /// Test whether some of the given requests have completed.
      /// \param indices are the request indices that have completed.
      /// \return the completion time for the call (current time).
      timestamp
      testsome(std::vector<mpirequest_t> &req, std::vector<int> &indices,
          std::vector<const_mpistatus_t> &statuses);

      /// Blocking probe for a message that matches the given signature.
      /// \return current time when the probe succeeds.
      timestamp
      probe(const mpiid &source, const mpitag &tag, const mpicomm::ptr &comm,
          const_mpistatus_t &stat);

      /// Non-blocking probe for messages.
      /// \return current time when the call completes.
      timestamp
      iprobe(const mpiid &source, const mpitag &tag, const mpicomm::ptr &comm,
          bool &flag, const_mpistatus_t &stat);

      /* Mucking with MPI state. */

      /// Attach a user-defined buffer for *bsend operations.
      timestamp
      buffer_attach(int bytes);

      /// Detach the user-defined buffer.
      /// \param bytes is the previously assigned buffer size.
      timestamp
      buffer_detach(int &bytes);

      // I am omitting all the keyval stuff; it doesn't have much function
      // in a simulator and has a significant implementation complexity.

      /* Collective operations */

      /// Set a barrier.
      /// \return the time at which this node was released from the barrier.
      timestamp
      barrier(const mpicomm::ptr &comm, bool tracethis = true);

      /// Broadcast data with no payload.
      timestamp
      bcast(int count, const mpitype &type, const mpiid &root,
          const mpicomm::ptr &comm);

      /// Broadcast data and send/receive payload.
      /// On root, the content will be unchanged.
      /// Other nodes will get a copy of root's payload.
      timestamp
      bcast(int count, const mpitype &type, const mpiid &root,
          const mpicomm::ptr &comm, payload::const_ptr &content);

      /// Scatter data uniformly.
      timestamp
      scatter(int sendcount, const mpitype &sendtype, int recvcount,
          const mpitype &recvtype, const mpiid &root, const mpicomm::ptr &comm);

      /// Scatter data uniformly and send payload.
      /// The content at root will be sent to all nodes.
      /// The 'result' pointer will be updated on all nodes to
      /// point to a copy of root's content.
      timestamp
      scatter(int sendcount, const mpitype &sendtype, int recvcount,
          const mpitype &recvtype, const mpiid &root, const mpicomm::ptr &comm,
          const std::vector<payload::const_ptr> &cc,
          payload::const_ptr &result, bool usingpayload = true);

      /// Scatter data non-uniformly.
      timestamp
      scatterv(const std::vector<int> &sendcounts, const mpitype &sendtype,
          int recvcount, const mpitype &recvtype, const mpiid &root,
          const mpicomm::ptr &comm);

      /// Scatter data non-uniformly.
      /// The content at root will be sent to all nodes.
      /// The 'result' pointer will be updated on all nodes to
      /// point to a copy of root's content.
      timestamp
      scatterv(const std::vector<int> &sendcounts, const mpitype &sendtype,
          int recvcount, const mpitype &recvtype, const mpiid &root,
          const mpicomm::ptr &comm,
          const std::vector<payload::const_ptr>& content,
          payload::const_ptr &result, bool usingpayload = true);

      /// Gather uniformly distributed data.
      timestamp
      gather(int sendcount, const mpitype &sendtype, int recvcount,
          const mpitype &recvtype, const mpiid &root, const mpicomm::ptr &comm);

      /// Gather uniformly distributed data and payload.
      /// The 'result' at root will have pointers to copies of
      /// all node's content.
      timestamp
      gather(int sendcount, const mpitype &sendtype, int recvcount,
          const mpitype &recvtype, const mpiid &root, const mpicomm::ptr &comm,
          const payload::const_ptr& content,
          std::vector<payload::const_ptr> &cc);

      /// Gather non-uniformly distributed data.
      timestamp
      gatherv(int sendcount, const mpitype &sendtype,
          const std::vector<int> &recvcnts, const mpitype &recvtype,
          const mpiid &root, const mpicomm::ptr &comm);

      /// Gather non-uniformly distributed data and payload.
      /// The 'result' at root will have pointers to copies of
      /// all node's content.
      timestamp
      gatherv(int sendcount, const mpitype &sendtype,
          const std::vector<int> &recvcnts, const mpitype &recvtype,
          const mpiid &root, const mpicomm::ptr &comm,
          const payload::const_ptr& content,
          std::vector<payload::const_ptr> &result);

      /// Gather uniformly distributed data and make it available to all nodes.
      timestamp
      allgather(int sendcount, const mpitype &sendtype, int recvcount,
          const mpitype &recvtype, const mpicomm::ptr &comm,
          bool tracethis = true);

      /// Gather uniformly distributed data and make it available to all nodes.
      timestamp
      allgather(int sendcount, const mpitype &sendtype, int recvcount,
          const mpitype &recvtype, const mpicomm::ptr &comm,
          const payload::const_ptr& content,
          std::vector<payload::const_ptr> &result, bool tracethis = true);

      /// Gather non-uniformly distributed data onto all nodes.
      timestamp
      allgatherv(int sendcount, const mpitype &sendtype,
          const std::vector<int> &recvcnts, const mpitype &recvtype,
          const mpicomm::ptr &comm);

      /// Gather non-uniformly distributed data onto all nodes.
      timestamp
      allgatherv(int sendcount, const mpitype &sendtype,
          const std::vector<int> &recvcnts, const mpitype &recvtype,
          const mpicomm::ptr &comm, const payload::const_ptr& content,
          std::vector<payload::const_ptr> &result);

      /// Send uniformly distributed data from all to all processes.
      timestamp
      alltoall(int sendcount, const mpitype &sendtype, int recvcount,
          const mpitype &recvtype, const mpicomm::ptr &comm);

      /// Send uniformly distributed data from all to all processes.
      timestamp
      alltoall(int sendcount, const mpitype &sendtype, int recvcount,
          const mpitype &recvtype, const mpicomm::ptr &comm,
          const std::vector<payload::const_ptr>& content,
          std::vector<payload::const_ptr> &result);

      /// Send non-uniformly distributed data from all to all processors.
      timestamp
      alltoallv(const std::vector<int> &sendcnts, const mpitype &sendtype,
          const std::vector<int> &recvcnts, const mpitype &recvtype,
          const mpicomm::ptr &comm);

      /// Send non-uniformly distributed data from all to all processors.
      timestamp
      alltoallv(const std::vector<int> &sendcnts, const mpitype &sendtype,
          const std::vector<int> &recvcnts, const mpitype &recvtype,
          const mpicomm::ptr &comm,
          const std::vector<payload::const_ptr>& content,
          std::vector<payload::const_ptr> &result);

      /// Reduce data from all nodes onto the root node.
      timestamp
      reduce(int count, const mpitype &type, const mpiop &op,
          const mpiid &root, const mpicomm::ptr &comm);

      /// Reduce data from all nodes onto the root node and give root node
      /// access to everybody's payload.
      /// On root node, the result pointer will be rewritten to point to
      /// an object of type mpicollpayload.  On other nodes, this
      /// pointer will be set to null.
      timestamp
      reduce(int count, const mpitype &type, const mpiop &op,
          const mpiid &root, const mpicomm::ptr &comm,
          const payload::const_ptr& content, payload::const_ptr &result);

      /// Reduce data from all nodes to all nodes.
      timestamp
      allreduce(int count, const mpitype &type, const mpiop &op,
          const mpicomm::ptr &comm, bool tracethis = true);

      /// Reduce data from all nodes to all nodes and give all nodes access
      /// to the payload posted by each node.  The result pointer will point to
      /// an object of type mpicollpayload.
      timestamp
      allreduce(int count, const mpitype &type, const mpiop &op,
          const mpicomm::ptr &comm, const payload::const_ptr& content,
          payload::const_ptr &result, bool tracethis = true);

      /// Reduce data and scatter results.
      timestamp
      reduce_scatter(const std::vector<int> &recvcnts, const mpitype &type,
          const mpiop &op, const mpicomm::ptr &comm);

      /// Reduce data and scatter results.
      /// All nodes will get a pointer to all payload.
      timestamp
      reduce_scatter(const std::vector<int> &recvcnts, const mpitype &type,
          const mpiop &op, const mpicomm::ptr &comm,
          const payload::const_ptr& content, payload::const_ptr &result);

      /// Compute partial reductions on data.
      timestamp
      scan(int count, const mpitype &type, const mpiop &op,
          const mpicomm::ptr &comm);

      /// Compute partial reductions on data and share data.
      /// Each node will have as a result its own payload and that of
      /// all lower-ranked nodes
      timestamp
      scan(int count, const mpitype &type, const mpiop &op,
          const mpicomm::ptr &comm, const payload::const_ptr& content,
          payload::const_ptr &result);

      //----------------------------------------------------------------
      // --- MPI Derived Datatype
      // --- Presently, payloads won't work with derived datatypes
      //----------------------------------------------------------------

    private:
      /// Map derived datatypes to an id
      boost::unordered_map<long, mpitype> derived_types_;
      boost::unordered_map<long, bool> commited_;

      boost::unordered_map<long, mpiop> created_op_;
    public:

      /// Duplicate a type.  Easy enough.
      timestamp
      type_dup(const mpitype &intype, mpitype &outtype);

      timestamp
      type_set_name(long id, const std::string &name);

      /// Create an indexed type.  Yes, lets recreate C++ with MPI function calls.
      timestamp
      type_indexed(int count, int _blocklens_[], const std::vector<int> &_indices_,
          const mpitype &intype, mpitype &outtype, bool in_elem, int comb);

      /// Creates a contiguous datatype
      timestamp
      type_contiguous(int count, const mpitype &old_type, mpitype &new_type);

      /// Creates a vector (strided) datatype
      timestamp
      type_vector(int count, int blocklength, int stride,
          const mpitype &old_type, mpitype &new_type, bool stride_in_elem);

      /// Creates a struct datatype
      timestamp
      type_struct(const int count, const std::vector<int> &blocklens,
          const std::vector<int> &indices,
          const std::vector<mpitype> &old_types, mpitype &newtype);

      /// A datatype object has to be committed before use in communication.
      timestamp
      type_commit(const mpitype &type);

      /// Mark datatype for deallocation.
      timestamp
      type_free(const mpitype &type);

      /// Get the derived mpitype mapped to an id
      const mpitype&
      get_derived_type(unsigned long id);

      const mpiop&
      get_created_op(unsigned long id);

      timestamp
      op_create(void (*func) ( void *, void *, int *, long * ), int comm, mpiop &output);

      timestamp
      op_free(long id);

      //----------------------------------------------------------------
      // --- MPI-2 RMA / windows
      //----------------------------------------------------------------

    public:

    protected:

      nested_type<mpi_window::ptr, mpirequest_t, mpi_rma_message::op_info>::unordered_map
          pending_rmas_;

    public:

      enum LOCK_TYPE
      {
        mpi_lock_shared_ = 0, mpi_lock_exclusive_
      };
      enum WIN_FLAGS
      {
        mpi_flag_none_ = 0, mpi_win_mode_nocheck_
      };

      timestamp
      win_create(void* base, size_t size, int disp, mpiinfo::ptr inf,
          const mpicomm::ptr &comm, mpi_window::ptr &win, long id);

      timestamp
      win_free(mpi_window::ptr win);

      timestamp
      win_get(mpirequest_t &req, int origin_count,
          const mpitype &origin_datatype, const mpiid &target_rank,
          long long target_disp, int target_count,
          const mpitype &target_datatype, const mpi_window::ptr &win);

      timestamp
      win_put(mpirequest_t &req, const payload::const_ptr &load,
          int target_count, const mpitype &target_datatype,
          const mpiid &target_rank, long long target_disp,
          const mpi_window::ptr &win);

      timestamp
      win_accumulate(mpirequest_t &req, const payload::const_ptr &load,
          int target_count, const mpitype &target_datatype,
          const mpiid &target_rank, long long target_disp, const mpiop & op,
          const mpi_window::ptr &win);

      timestamp
      win_fence(const mpi_window::ptr &win, bool barr,
          boost::unordered_map<mpirequest_t, payload::const_ptr> &loads);

      timestamp
      win_unlock(const mpiid &rank, const mpi_window::ptr &win,
          boost::unordered_map<mpirequest_t, payload::const_ptr> &loads);

      timestamp
      win_lock(LOCK_TYPE t, const mpiid &rank, WIN_FLAGS flag,
          const mpi_window::ptr &win);

    };

    mpiapi::ptr
    sstmac_mpi();

  }
} // end of namespace sstmac

#endif
