/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/api/mpi/mpimessage.h>
#include <sstmac/software/api/mpi/mpistatus.h>
#include <sstmac/common/messages/payload.h>
#include <sstmac/software/api/mpi/mpi_protocol/mpiprotocol.h>
#include <sstmac/common/sstmac_env.h>
#include <sstmac/software/process/operatingsystem.h>

#include <sstmac/common/debug.h>
#include <sstmac/common/errors.h>
#include <stdlib.h>
#include <sstream>

DeclareSerializable(sstmac::sw::mpimessage)
  ;

namespace sstmac
{
  namespace sw
  {

    // Pointer types.

    //
    // Hello.
    //
    mpimessage::mpimessage(const std::string& libn, const timestamp& time,
        int64_t envelope, int64_t mintrans, int count, const mpitype &type,
        const mpiid &source, const mpiid &dest, const mpitag &tag,
        const mpicommid &commid, int seqnum, const messageid &msgid,
        bool ssend, category cat, const taskid &source_task,
        const taskid &dest_task, const appid &aid, content_type_t content_type,
        const payload::const_ptr& content, const mpiprotocol::ptr& protocol,
        MESSAGE_TYPES ty) :
      sst_message(ty), library_message(libn), timed_message(time),
          envelope_(envelope), mintrans_(mintrans), count_(count), type_(type),
          source_(source), dest_(dest), tag_(tag), commid_(commid),
          seqnum_(seqnum), msgid_(msgid), cat_(cat),
          content_type_(content_type), ssend_(ssend),
          source_task_(source_task), dest_task_(dest_task), aid_(aid),
          content_(content), ignore_seqnum_(false), bytes_(0), delay_(0)
    {
      if (envelope_ < 0)
        envelope_ = 0;
      if (mintrans_ < 0)
        mintrans_ = 0;

      protocol_ = protocol->get_prot_id();

      recompute_bytes();
    }

    void
    mpimessage::recompute_bytes()
    {
      // Pretend that envelope info is 12 bytes.
      // A handshake request is, therefore, only 12 bytes.
      if (count_ > 0)
      {
        int64_t sbytes = envelope_ + count_ * int64_t(type_.get_packed_size());
        bytes_ = sbytes > mintrans_ ? sbytes : mintrans_;
      }
      else
      {
        bytes_ = (envelope_ > mintrans_ ? envelope_ : mintrans_);
      }
    }

    mpimessage::mpimessage(content_type_t content_type) :
      sst_message(MPI), library_message(""), timed_message(timestamp(0)),
          envelope_(0), mintrans_(0), count_(0), commid_(-1),
          content_type_(content_type), content_(0),
          protocol_(mpiprotocol::EAGER0_SOCKET), ignore_seqnum_(false),
          delay_(0)
    {
      recompute_bytes();
    }

    mpimessage::mpimessage(const payload::const_ptr& content,
        content_type_t content_type) :
      sst_message(MPI), library_message(""), timed_message(timestamp(0)),
          envelope_(0), mintrans_(0), count_(0), commid_(-1),
          content_type_(content_type), content_(content),
          protocol_(mpiprotocol::EAGER0_SOCKET), ignore_seqnum_(false),
          delay_(0)
    {
      recompute_bytes();
    }

    mpimessage::mpimessage() :
      sst_message(MPI), library_message(""), timed_message(timestamp(0)),
          commid_(-1)
    {

    }

    //
    // Construct a new message.
    //
    mpimessage::ptr
    mpimessage::construct(const std::string& libn, const timestamp& time,
        int64_t envelope, int64_t mintrans, int count, const mpitype &type,
        const mpiid &source, const mpiid &dest, const mpitag &tag,
        const mpicommid &commid, int seqnum, const messageid &msgid,
        bool ssend, category cat, const taskid &source_task,
        const taskid &dest_task, const appid &aid, content_type_t content_type,
        const payload::const_ptr& content, const mpiprotocol::ptr& protocol,
        MESSAGE_TYPES ty)
    {

      mpimessage::ptr retval(
          new mpimessage(libn, time, envelope, mintrans, count, type, source,
              dest, tag, commid, seqnum, msgid, ssend, cat, source_task,
              dest_task, aid, content_type, content, protocol, ty));
      return retval;
    }

    sst_message::ptr
    mpimessage::clone() const
    {
      return deep_clone();
    }

    mpimessage::ptr
    mpimessage::deep_clone() const
    {
      mpimessage::ptr ret = mpimessage::construct(lib_name_, time_, envelope_,
          mintrans_, count_, type_, source_, dest_, tag_, commid_, seqnum_,
          msgid_, ssend_, cat_, source_task_, dest_task_, aid_, content_type_,
          content_, get_protocol());

      ret->toaddr_ = toaddr_;
      ret->fromaddr_ = fromaddr_;
      ret->ignore_seqnum_ = ignore_seqnum_;
      return ret;
    }

    sst_message::ptr
    mpimessage::clone(MESSAGE_TYPES ty) const
    {
      mpimessage::ptr ret = deep_clone();
      ret->msgtype_ = ty;
      return ret;
    }

    //
    // Goodbye.
    //
    mpimessage::~mpimessage() throw ()
    {
    }

    //
    // Create a status object.
    //
    mpistatus::ptr
    mpimessage::build_status() const
    {
      mpistatus::ptr stat = mpistatus::construct();
      if (content_type_ != fake)
      {
        stat->set_commid(this->commid());
        stat->set_source(this->source());
        stat->set_dest(this->dest());
        stat->set_tag(this->tag());
        stat->set_send_type(this->type());
        stat->set_count(this->count_ * this->type_.get_packed_size());
      }
      stat->set_content(this->content());
      return stat;
    }

#define enumcase(x) case x: return #x;
    const char*
    mpimessage::str(category cat)
    {
      switch (cat)
      {
      enumcase(user);
      enumcase(collective);
      enumcase(onesided);
      }
      sst_throw(valueerror, "ERROR: unknown mpi category %d", cat);
    }

    const char*
    mpimessage::str(content_type_t content_type)
    {
      switch (content_type)
      {
      enumcase(data)
        ;
      enumcase(header)
        ;
      enumcase(rendezvous_ack)
        ;
      enumcase(completion_ack)
        ;
      enumcase(fake)
        ;
      }
      sst_throw(valueerror, "ERROR: unknown mpi content type %d", content_type);
    }

    std::string
    mpimessage::to_string() const
    {
      std::stringstream ss;
      ss << "mpimessage(" << "msgtype=" << sst_message::tostr(msgtype_)
          << ", env=" << envelope_ << ", mintrans=" << mintrans_ << ", count="
          << count_ << ", type=" << type_.to_string() << ", src="
          << source_.to_string() << ", dst=" << dest_.to_string() << ", tag="
          << tag_.to_string() << ", commid" << commid_.to_string();
      if (ignore_seqnum_)
        ss << ", seq=(ignored)" << seqnum_;
      else
        ss << ", seq=" << seqnum_;
      ss << ", ssend=" << ssend_ << ", category=" << str(cat_) << ", content="
          << str(content_type_) << ", srctsk=" << source_task_.to_string()
          << ", dsttsk=" << dest_task_.to_string() << ", protocol="
          << get_protocol()->to_string() << ", load=" << ((content_ == 0) ? "null"
          : content_->to_string()) << ")";
      return ss.str();
    }

    void
    mpimessage::serialize_order(sst_serializer* ser)
    {
      //sst_message::serialize_order(ser);
      network_message::serialize_order(ser);

      //child class stuff, dangerous
      ser->do_primitive(lib_name_);
      ser->do_primitive(time_);

      ser->do_primitive(envelope_);
      ser->do_primitive(mintrans_);
      ser->do_primitive(count_);
    //  ser->do_primitive(type_.id);
      type_.serialize_order(ser);
      ser->do_primitive(source_.id);
      ser->do_primitive(dest_.id);
      ser->do_primitive(tag_.id);
      ser->do_primitive(commid_.id);
      ser->do_primitive(seqnum_);
      ser->do_primitive(msgid_.id);
      ser->do_enum(cat_);
      ser->do_enum(content_type_);
      ser->do_primitive(ssend_);
      ser->do_primitive(source_task_.id);
      ser->do_primitive(dest_task_.id);
      ser->do_primitive(aid_.id);
      ser->do_primitive(ignore_seqnum_);
      ser->do_primitive(bytes_);
      ser->do_primitive(delay_);
      ser->do_enum(protocol_);
      ser->do_const_obj(content_);
    }



    long
    mpimessage::payload_bytes() const
    {
      return abs(count_) * type_.get_packed_size();
    }

    void
    mpimessage::mutate(mutation_t mut)
    {
      if (mut == sst_message::RDMA_GET_REQ_TO_RSP)
      {
        if (count_ > 0)
          throw illformederror(
              "mpimessage: invalid rdma request generation from real payload");
        count_ = -count_;
        msgtype_ = RDMA_GET_RSP;
        content_type_ = mpimessage::data;
        //set_needs_ack(false);
        reverse();
        bytes_ = type_.get_packed_size() * count_;
        //no MPI overhead on this - just the payload
        //recompute_bytes();
      }
      else if (mut == sst_message::GNI_HEADER_TO_RDMA_GET_REQ)
      {
        msgtype_ = RDMA_GET_REQ;
        //set_needs_ack(false);
        reverse();
      }
      else if (mut == sst_message::RDMA_PUT_REQ_TO_RSP)
      {
        if (count_ < 0)
          throw illformederror(
              "mpimessage: invalid rdma put request generation");
        count_ = -count_;
        msgtype_ = RDMA_PUT_RSP;
        content_type_ = mpimessage::completion_ack;
        //set_needs_ack(false);
        reverse();

        //no MPI overhead on this - just the payload
        recompute_bytes();
      }
      else
      {
        sst_throw(valueerror, "invalid mutation type received by mpimessage: %s", sst_message::tostr(mut));
      }
    }

    void
    mpimessage::reverse()
    {
      //also flip the addresses
      nodeaddress::ptr dst = fromaddr_;
      nodeaddress::ptr src = toaddr_;
      toaddr_ = dst;
      fromaddr_ = src;

      taskid srcid = source_task_;
      taskid dstid = dest_task_;
      source_task_ = dstid;
      dest_task_ = srcid;
    }

    void
    mpimessage::header_to_rendezvous_ack()
    {
      reverse();
      content_type_ = mpimessage::rendezvous_ack;
    }

    void
    mpimessage::rendezvous_ack_to_payload()
    {
      reverse();
      content_type_ = mpimessage::data;
      count_ = -count_;
      recompute_bytes();
    }

    void
    mpimessage::payload_to_completion_ack()
    {
      reverse();
      content_type_ = mpimessage::completion_ack;
      count_ = -count_;
      recompute_bytes();
    }

    mpiprotocol::ptr
    mpimessage::get_protocol() const
    {
      return mpiprotocol::get_protocol_object((mpiprotocol::PROTOCOL_ID)protocol_);
    }

    void
    mpimessage::set_protocol(const mpiprotocol::ptr& protocol)
    {
      protocol_ = protocol->get_prot_id();
    }

    long
    mpimessage::source_thread(const operatingsystem::ptr& os) const
    {
      return os->get_task_threadid(source_task_);
    }

    long
    mpimessage::dest_thread(const operatingsystem::ptr& os) const
    {
      return os->get_task_threadid(dest_task_);
    }

    void
    mpimessage::set_ignore_seqnum(bool flag)
    {
      ignore_seqnum_ = flag;
    }

    bool
    mpimessage::ignore_seqnum() const
    {
      return ignore_seqnum_;
    }

  }
} // end of namespace sstmac
