/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_API_MPI_MPIPAYLOAD_H_INCLUDED
#define SSTMAC_SOFTWARE_API_MPI_MPIPAYLOAD_H_INCLUDED

#include <sstream>

#include <sstmac/common/config.h>
#include <sstmac/common/messages/payload.h>
#include <sstmac/software/api/mpi/mpi_types/mpitype.h>
#include <sstmac/software/api/mpi/mpi_types/mpiop.h>

namespace sstmac
{
  namespace sw
  {
    /**
     * Network payload consisting of a boost multiarray.
     *
     */
    class mpipayload : public payload, public serializable_type<mpipayload>
    {
    ImplementSerializable(mpipayload)

      void* buf_;
      size_t bufsize_; //in bytes
      mpitype type_;

      int cnt_;

      bool wrapped_;
      bool given_;
      bool real_;

      /// Construction time.
      mpipayload(void* data, mpitype t, int count, bool usewrapper);
      //internal construction
      mpipayload(void* data, mpitype t, int count);

      mpipayload();

    public:
      typedef boost::intrusive_ptr<mpipayload> ptr;
      typedef boost::intrusive_ptr<const mpipayload> const_ptr;

      static const_ptr
      construct(void* data, mpitype t, int count, bool usewrapper)
      {
        return const_ptr(new mpipayload(data, t, count, usewrapper));
      }

      static const_ptr
      construct_given_buffer(void* data, mpitype t, int count)
      {
        return const_ptr(new mpipayload(data, t, count));
      }

      static ptr
      construct()
      {
        return ptr(new mpipayload());
      }

      virtual long
      get_byte_length() const
      {
        return bufsize_;
      }

      virtual
      ~mpipayload()
      {
        if ((given_ || !wrapped_) && bufsize_ > 0)
          free(buf_);
      }

      /// Clone this object.
      payload::const_ptr
      clone() const
      {
        return const_ptr(new mpipayload(buf_, type_, cnt_, wrapped_));
      }

      virtual serializable::ptr
      serialization_clone() const
      {
        void* newbuf = malloc(bufsize_);
        memcpy(newbuf, buf_, bufsize_);
        return ptr(new mpipayload(newbuf, type_, cnt_));
      }

      void
      extract(void* data, int count, mpitype t) const
      {
        if (count > 0 && bufsize_ > 0)
        {
          int pos = 0;
          t.unpack(buf_, bufsize_, data, count, &pos);
        }
      }

      //extract everything
      void
      extract(void* data) const
      {
        if (cnt_ > 0 && bufsize_ > 0)
        {
          int pos = 0;
          type_.unpack(buf_, bufsize_, data, cnt_, &pos);
        }
      }

      int
      get_count() const
      {
        return cnt_;
      }

      mpitype
      get_type() const
      {
        return type_;
      }

    protected:

      struct addstruct
      {
        template<typename Type>
          void
          go(void* newbuf, const void* arg1, const void* arg2, int ind, Type f) const
          {
            Type* cast1 = (Type*) arg1;
            Type* cast2 = (Type*) arg2;
            Type* tocast = (Type*) newbuf;
            tocast[ind] = cast1[ind] + cast2[ind];
          }
      };

      struct prodstruct
      {
        template<typename Type>
          void
          go(void* newbuf, const void* arg1, const void* arg2, int ind, Type f) const
          {
            Type* cast1 = (Type*) arg1;
            Type* cast2 = (Type*) arg2;
            Type* tocast = (Type*) newbuf;
            tocast[ind] = cast1[ind] * cast2[ind];
          }
      };

      struct minstruct
      {
        template<typename Type>
          void
          go(void* newbuf, const void* arg1, const void* arg2, int ind, Type f) const
          {
            Type* cast1 = (Type*) arg1;
            Type* cast2 = (Type*) arg2;
            Type* tocast = (Type*) newbuf;
            tocast[ind] = (cast1[ind] > cast2[ind]) ? cast2[ind] : cast1[ind];
          }
      };

      struct maxstruct
      {
        template<typename Type>
          void
          go(void* newbuf, const void* arg1, const void* arg2, int ind, Type f) const
          {
            Type* cast1 = (Type*) arg1;
            Type* cast2 = (Type*) arg2;
            Type* tocast = (Type*) newbuf;
            tocast[ind] = (cast1[ind] > cast2[ind]) ? cast1[ind] : cast2[ind];
          }
      };

      struct replacestruct
      {
        template<typename Type>
          void
          go(void* newbuf, const void* arg1, const void* arg2, int ind, Type f) const
          {
            Type* cast1 = (Type*) arg1;
            Type* cast2 = (Type*) arg2;
            Type* tocast = (Type*) newbuf;
            tocast[ind] = cast2[ind];
          }
      };

      struct minlocstruct
      {
        template<typename Type1, typename Type2>
          void
          go(void* newbuf, const void* arg1, const void* arg2, int ind,
              Type1 f, Type2 g) const
          {
            struct mystruct
            {
              Type1 a;
              Type2 b;
            };
            mystruct* cast1 = (mystruct*) arg1;
            mystruct* cast2 = (mystruct*) arg2;
            mystruct* tocast = (mystruct*) newbuf;
            if (cast1[ind].a < cast2[ind].a)
            {
              tocast[ind].a = cast1[ind].a;
              tocast[ind].b = cast1[ind].b;
            }
            else
            {
              tocast[ind].a = cast2[ind].a;
              tocast[ind].b = cast2[ind].b;
            }
          }
      };

      struct maxlocstruct
      {
        template<typename Type1, typename Type2>
          void
          go(void* newbuf, const void* arg1, const void* arg2, int ind,
              Type1 f, Type2 g) const
          {
            struct mystruct
            {
              Type1 a;
              Type2 b;
            };
            mystruct* cast1 = (mystruct*) arg1;
            mystruct* cast2 = (mystruct*) arg2;
            mystruct* tocast = (mystruct*) newbuf;
            if (cast1[ind].a > cast2[ind].a)
            {
              tocast[ind].a = cast1[ind].a;
              tocast[ind].b = cast1[ind].b;
            }
            else
            {
              tocast[ind].a = cast2[ind].a;
              tocast[ind].b = cast2[ind].b;
            }
          }
      };

      struct orstruct
      {
        template<typename Type>
          void
          go(void* newbuf, const void* arg1, const void* arg2, int ind, Type f) const
          {
            Type* cast1 = (Type*) arg1;
            Type* cast2 = (Type*) arg2;
            Type* tocast = (Type*) newbuf;
            tocast[ind] = cast1[ind] || cast2[ind];
          }
      };

      struct xorstruct
      {
        template<typename Type>
          void
          go(void* newbuf, const void* arg1, const void* arg2, int ind, Type f) const
          {
            Type* cast1 = (Type*) arg1;
            Type* cast2 = (Type*) arg2;
            Type* tocast = (Type*) newbuf;
            tocast[ind] = (cast1[ind] && !cast2[ind]) || (!cast1[ind]
                && cast2[ind]);
          }
      };

      struct andstruct
      {
        template<typename Type>
          void
          go(void* newbuf, const void* arg1, const void* arg2, int ind, Type f) const
          {
            Type* cast1 = (Type*) arg1;
            Type* cast2 = (Type*) arg2;
            Type* tocast = (Type*) newbuf;
            tocast[ind] = cast1[ind] && cast2[ind];
          }
      };

      struct borstruct
      {
        template<typename Type>
          void
          go(void* newbuf, const void* arg1, const void* arg2, int ind, Type f) const
          {
            Type* cast1 = (Type*) arg1;
            Type* cast2 = (Type*) arg2;
            Type* tocast = (Type*) newbuf;
            tocast[ind] = cast1[ind] | cast2[ind];
          }
      };

      struct bxorstruct
      {
        template<typename Type>
          void
          go(void* newbuf, const void* arg1, const void* arg2, int ind, Type f) const
          {
            Type* cast1 = (Type*) arg1;
            Type* cast2 = (Type*) arg2;
            Type* tocast = (Type*) newbuf;
            tocast[ind] = (cast1[ind] ^ cast2[ind]);
          }
      };

      struct bandstruct
      {
        template<typename Type>
          void
          go(void* newbuf, const void* arg1, const void* arg2, int ind, Type f) const
          {
            Type* cast1 = (Type*) arg1;
            Type* cast2 = (Type*) arg2;
            Type* tocast = (Type*) newbuf;
            tocast[ind] = cast1[ind] & cast2[ind];
          }
      };

      template<typename T>
        payload::const_ptr
        go_go_gadget_operation(payload::const_ptr other) const
        {
          if (type_.get_type() != mpitype::PRIM)
          {
            sst_throw(unimplementederror, "mpipayload:: operation on derived type");
          }

          mpipayload::const_ptr oth = boost::dynamic_pointer_cast<
              const mpipayload>(other);

          if (!oth)
          {
            sst_throw(valueerror, "mpipayload::operation - payload didn't cast to mpipayload: %s", other->to_string().c_str());
          }

          T thing;

          void* newbuf = (void*) malloc(bufsize_);
          int ind = 0;

          const char* othbuf = (const char*) oth->buf_;
          const char* mybuf = (const char*) buf_;
          for (int i = 0; i < cnt_; i++)
          {
            if (type_ == mpitype::mpi_int || type_ == mpitype::mpi_integer)
            {
              int f;
              thing.go(newbuf, buf_, oth->buf_, i, f);
            }
            else if (type_ == mpitype::mpi_long)
            {
              long f;
              thing.go(newbuf, buf_, oth->buf_, i, f);
            }
            else if (type_ == mpitype::mpi_long_long || type_
                == mpitype::mpi_long_long_int)
            {
              long long f;
              thing.go(newbuf, buf_, oth->buf_, i, f);
            }
            else if (type_ == mpitype::mpi_double)
            {
              double f;
              thing.go(newbuf, buf_, oth->buf_, i, f);
            }
            else if (type_ == mpitype::mpi_short)
            {
              short f;
              thing.go(newbuf, buf_, oth->buf_, i, f);
            }
            else if (type_ == mpitype::mpi_char || type_ == mpitype::mpi_byte)
            {
              char f;
              thing.go(newbuf, buf_, oth->buf_, i, f);
            }
            else if (type_ == mpitype::mpi_signed_char)
            {
              signed char f;
              thing.go(newbuf, buf_, oth->buf_, i, f);
            }

            else if (type_ == mpitype::mpi_unsigned_char)
            {
              signed char f;
              thing.go(newbuf, buf_, oth->buf_, i, f);
            }
            else if (type_ == mpitype::mpi_float)
            {
              float f;
              thing.go(newbuf, buf_, oth->buf_, i, f);
            }
            else if (type_ == mpitype::mpi_unsigned)
            {
              unsigned int f;
              thing.go(newbuf, buf_, oth->buf_, i, f);
            }
            else if (type_ == mpitype::mpi_unsigned_long)
            {
              unsigned long f;
              thing.go(newbuf, buf_, oth->buf_, i, f);
            }
            else if (type_ == mpitype::mpi_unsigned_short)
            {
              unsigned short f;
              thing.go(newbuf, buf_, oth->buf_, i, f);
            }
            else if (type_ == mpitype::mpi_unsigned_char)
            {
              unsigned char f;
              thing.go(newbuf, buf_, oth->buf_, i, f);
            }
            else if (type_ == mpitype::mpi_unsigned_long_long)
            {
              unsigned long long f;
              thing.go(newbuf, buf_, oth->buf_, i, f);
            }
            else if (type_ == mpitype::mpi_long_double)
            {
              long double f;
              thing.go(newbuf, buf_, oth->buf_, i, f);
            }
            else
            {
              sst_throw(unimplementederror, "mpipayload::operation - %s not supported", type_.to_string().c_str());
            }
            ind += type_.get_packed_size();
          }

          return mpipayload::construct_given_buffer(newbuf, type_, cnt_);
        }

      template<typename T>
        payload::const_ptr
        go_go_bitwise_operation(payload::const_ptr other) const
        {
          if (type_.get_type() != mpitype::PRIM)
          {
            sst_throw(unimplementederror, "mpipayload:: operation on derived type");
          }

          mpipayload::const_ptr oth = boost::dynamic_pointer_cast<
              const mpipayload>(other);

          if (!oth)
          {
            sst_throw(valueerror, "mpipayload::operation - payload didn't cast to mpipayload: %s", other->to_string().c_str());
          }

          T thing;

          void* newbuf = (void*) malloc(bufsize_);

          int ind = 0;

          const char* othbuf = (const char*) oth->buf_;
          const char* mybuf = (const char*) buf_;
          for (int i = 0; i < cnt_; i++)
          {
            if (type_ == mpitype::mpi_int || type_ == mpitype::mpi_integer)
            {
              int f;
              thing.go(newbuf, buf_, oth->buf_, i, f);
            }
            else if (type_ == mpitype::mpi_long)
            {
              long f;
              thing.go(newbuf, buf_, oth->buf_, i, f);
            }
            else if (type_ == mpitype::mpi_long_long || type_
                == mpitype::mpi_long_long_int)
            {
              long long f;
              thing.go(newbuf, buf_, oth->buf_, i, f);
            }

            else if (type_ == mpitype::mpi_short)
            {
              short f;
              thing.go(newbuf, buf_, oth->buf_, i, f);
            }
            else if (type_ == mpitype::mpi_char || type_ == mpitype::mpi_byte)
            {
              char f;
              thing.go(newbuf, buf_, oth->buf_, i, f);
            }

            else if (type_ == mpitype::mpi_signed_char)
            {
              signed char f;
              thing.go(newbuf, buf_, oth->buf_, i, f);
            }

            else if (type_ == mpitype::mpi_unsigned_char)
            {
              signed char f;
              thing.go(newbuf, buf_, oth->buf_, i, f);
            }

            else if (type_ == mpitype::mpi_unsigned)
            {
              unsigned int f;
              thing.go(newbuf, buf_, oth->buf_, i, f);
            }
            else if (type_ == mpitype::mpi_unsigned_long)
            {
              unsigned long f;
              thing.go(newbuf, buf_, oth->buf_, i, f);
            }
            else if (type_ == mpitype::mpi_unsigned_short)
            {
              unsigned short f;
              thing.go(newbuf, buf_, oth->buf_, i, f);
            }
            else if (type_ == mpitype::mpi_unsigned_char)
            {
              unsigned char f;
              thing.go(newbuf, buf_, oth->buf_, i, f);
            }
            else if (type_ == mpitype::mpi_unsigned_long_long)
            {
              unsigned long long f;
              thing.go(newbuf, buf_, oth->buf_, i, f);
            }
            else
            {
              sst_throw(unimplementederror, "mpipayload::bitwise_operation - %s not supported", type_.to_string().c_str());
            }
            ind += type_.get_packed_size();
          }

          return mpipayload::construct_given_buffer(newbuf, type_, cnt_);
        }

      template<typename T>
        payload::const_ptr
        go_go_pair_operation(payload::const_ptr other) const
        {
          if (type_.get_type() != mpitype::PAIR)
          {
            sst_throw(unimplementederror, "mpipayload:: trying pair operation on non-pair");
          }

          mpipayload::const_ptr oth = boost::dynamic_pointer_cast<
              const mpipayload>(other);

          if (!oth)
          {
            sst_throw(valueerror, "mpipayload::operation - payload didn't cast to mpipayload: %s", other->to_string().c_str());
          }

          T thing;

          void* newbuf = (void*) malloc(cnt_ * type_.get_extent());
          int ind = 0;

          void* unpack1 = (void*) malloc(cnt_ * type_.get_extent());
          void* unpack2 = (void*) malloc(cnt_ * type_.get_extent());

          int pos1 = 0, pos2 = 0;
          type_.unpack(buf_, cnt_ * type_.get_extent(), unpack1, cnt_, &pos1);
          type_.unpack(oth->buf_, cnt_ * type_.get_extent(), unpack2, cnt_,
              &pos2);

          const char* othbuf = (const char*) oth->buf_;
          const char* mybuf = (const char*) buf_;
          for (int i = 0; i < cnt_; i++)
          {
            if (type_ == mpitype::mpi_2int)
            {
              int f;
              int g;
              thing.go(newbuf, unpack1, unpack2, i, f, g);
            }
            else if (type_ == mpitype::mpi_long_int)
            {

              long f;
              int g;
              thing.go(newbuf, unpack1, unpack2, i, f, g);
            }
            else if (type_ == mpitype::mpi_short_int)
            {
              short f;
              int g;
              thing.go(newbuf, unpack1, unpack2, i, f, g);
            }

            else if (type_ == mpitype::mpi_float_int)
            {
              float f;
              int g;
              thing.go(newbuf, unpack1, unpack2, i, f, g);
            }
            else if (type_ == mpitype::mpi_double_int)
            {
              double f;
              int g;
              thing.go(newbuf, unpack1, unpack2, i, f, g);
            }
            else if (type_ == mpitype::mpi_long_double_int)
            {
              long double f;
              int g;
              thing.go(newbuf, unpack1, unpack2, i, f, g);
            }

            else
            {
              sst_throw(unimplementederror, "mpipayload::pair_operation - %s not supported", type_.to_string().c_str());
            }
            ind += type_.get_packed_size();
          }

          payload::const_ptr p = mpipayload::construct(newbuf, type_, cnt_,
              false);

          free(newbuf);
          free(unpack1);
          free(unpack2);
          return p;
        }

    public:
      /**
       * Add operator
       * @param other
       * @return a smart pointer to a new valuepayload object
       */
      virtual const payload::const_ptr
      add(const payload::const_ptr &other) const
      {
        return go_go_gadget_operation<addstruct> (other);
      }

      virtual const payload::const_ptr
      prod(const payload::const_ptr &other) const
      {
        return go_go_gadget_operation<prodstruct> (other);
      }

      virtual const payload::const_ptr
      user_op(const payload::const_ptr &other, const mpiop &op) const
      {
        mpipayload::const_ptr oth = boost::dynamic_pointer_cast<
            const mpipayload>(other);

        if (!oth)
        {
          sst_throw(valueerror, "mpipayload::operation - payload didn't cast to mpipayload: %s", other->to_string().c_str());
        }

        void* newbuf = (void*) malloc(cnt_ * type_.get_extent());
        void* otherbuf = (void*) malloc(cnt_ * type_.get_extent());

        int pos1 = 0, pos2 = 0;
        type_.unpack(buf_, cnt_ * type_.get_extent(), newbuf, cnt_, &pos1);
        type_.unpack(oth->buf_, cnt_ * type_.get_extent(), otherbuf, cnt_,
            &pos2);

        int c = cnt_;
        long t = type_.native_id;
        (*op.userfunc_)(otherbuf, newbuf, &c, &t);
        payload::const_ptr p =
            mpipayload::construct(newbuf, type_, cnt_, false);
        free(newbuf);
        free(otherbuf);
        return p;
      }

      /**
       * Less than comparator
       * @param other
       * @return
       */
      virtual const payload::const_ptr
      min(const payload::const_ptr &other) const
      {
        return go_go_gadget_operation<minstruct> (other);
      }

      virtual const payload::const_ptr
      minloc(const payload::const_ptr &other) const
      {
        return go_go_pair_operation<minlocstruct> (other);
      }

      virtual const payload::const_ptr
      maxloc(const payload::const_ptr &other) const
      {
        return go_go_pair_operation<maxlocstruct> (other);
      }

      /**
       * Greater than comparator
       * @param other
       * @return
       */
      virtual const payload::const_ptr
      max(const payload::const_ptr &other) const
      {
        return go_go_gadget_operation<maxstruct> (other);
      }

      virtual const payload::const_ptr
      replace(const payload::const_ptr &other) const
      {
        return go_go_gadget_operation<replacestruct> (other);
      }

      /**
       * Equals comparator
       * @param other
       * @return
       */
      virtual bool
      equals(const payload::const_ptr &other) const
      {
        mpipayload::const_ptr oth = boost::dynamic_pointer_cast<
            const mpipayload>(other);

        bool ret = true;
        char* othbuf = (char*) oth->buf_;
        char* mybuf = (char*) buf_;
        for (int i = 0; i < cnt_ * type_.get_packed_size() && ret; i++)
        {
          ret = ret && (mybuf[i] == othbuf[i]);
        }

        return ret;
      }

      /**
       * Logical or comparator
       * @param other
       * @return
       */
      virtual const payload::const_ptr
      logical_or(const payload::const_ptr &other) const
      {
        return go_go_gadget_operation<orstruct> (other);
      }

      /**
       * Logical xor comparator
       * @param other
       * @return
       */
      virtual const payload::const_ptr
      logical_xor(const payload::const_ptr &other) const
      {
        return go_go_gadget_operation<xorstruct> (other);
      }

      /**
       * Logical and comparator
       * @param other
       * @return
       */
      virtual const payload::const_ptr
      logical_and(const payload::const_ptr &other) const
      {
        return go_go_gadget_operation<andstruct> (other);
      }

      virtual const payload::const_ptr
      bitwise_or(const payload::const_ptr &other) const
      {
        return go_go_bitwise_operation<borstruct> (other);
      }

      /**
       * Logical xor comparator
       * @param other
       * @return
       */
      virtual const payload::const_ptr
      bitwise_xor(const payload::const_ptr &other) const
      {
        return go_go_bitwise_operation<bxorstruct> (other);
      }

      /**
       * Logical and comparator
       * @param other
       * @return
       */
      virtual const payload::const_ptr
      bitwise_and(const payload::const_ptr &other) const
      {
        return go_go_bitwise_operation<bandstruct> (other);
      }

      virtual void
      serialize_order(sst_serializer* ser);

      /**
       * Strinfier
       * @return a std::string description
       */
      virtual std::string
      to_string() const;

    };

  } // end of namespace sstmac
}
#endif
