/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/api/mpi/mpiserver.h>
#include <sstmac/software/process/operatingsystem.h>
#include <sstmac/software/api/mpi/mpimessage.h>
#include <sstmac/common/messages/sst_message.h>
#include <sstmac/software/api/mpi/mpi_queue/mpiqueue.h>
#include <sstmac/common/eventcallback.h>
#include <sstmac/software/api/mpi/mpi_protocol/mpiprotocol.h>
#include <sstmac/common/errors.h>

namespace sstmac
{
  namespace sw
  {

    using boost::dynamic_pointer_cast;

    debug mpiserver::dbg_;

    mpiserver::mpiserver(int appnum) : sender_(0)
    {
      if (!dbg_.is_real())
        {
          dbg_ = debug("<debug> mpiserver | mpi | library", logger::blue);
          logger::register_logger(&dbg_);
        }
      libname_ = sstprintf("mpiserver-%d", appnum);
    }

    mpiserver::~mpiserver() throw ()
    {
    }

    mpiserver::ptr
    mpiserver::construct(int appnum)
    {
        return new mpiserver(appnum);
    }

    void
    mpiserver::start()
    {
        service::start();
        sender_ = msg_callback(mpiserver::ptr(this), &mpiserver::start_send);
        lib_mem_ = lib_compute_memmove::construct(libname_ + "-lib-mem");
        lib_mem_->init_as_callback_lib();
        register_lib(lib_mem_);
    }

    void
    mpiserver::handle(const sst_message::ptr &msg)
    {
      SSTMAC_DEBUG << "mpiserver[" << os_->get_my_addr()->to_string()
          << "]: handle message " << msg->to_string() << " \n";
      mpimessage::ptr mpi = boost::dynamic_pointer_cast<mpimessage>(msg);
      handle_recv(mpi);
    }

    //
    // Define a task for this server.  Each task can only be defined once
    //
    void
    mpiserver::add_task(const softwareid &sid, const mpiqueue::ptr& peer,
        int rank)
    {
      if (peer == 0)
        {
          throw nullerror("mpiserver::add_task:  null peer pointer.");
        }
      queue_t::iterator it = queue_.find(sid);
      if (it != queue_.end())
        {
          throw illformederror(
              "mpiserver::add_task:  Attempt to redefine peer.");
        }
      queue_[sid] = peer;

      sids_[sid.task_.id] = sid;
    }

    void
    mpiserver::unregister(const softwareid &sid)
    {
      queue_t::iterator it = queue_.find(sid);
      if (it == queue_.end())
        {
          throw illformederror(
              "mpiserver::unregister -- can't find sid in queue list.");
        }

      queue_.erase(sid);

      if (queue_.empty())
      {
          unregister_all_libs();
          os_->unregister_lib(service::ptr(this));
          //kill the callback to avoid circ reference
          sender_ = 0;
      }
    }

    //
    // Test whether the given task is defined on this node.
    //
    bool
    mpiserver::has_task(const softwareid &tid) const
    {
      return (queue_.find(tid) != queue_.end());
    }

    //
    // Send data.
    //
    void
    mpiserver::send(const mpimessage::ptr& payload)
    {
        handle_send(payload);
    }

    void
    mpiserver::buffered_send(const mpimessage::ptr& payload)
    {
        SSTMAC_DEBUG << "mpiserver: buffered send\n";
        long src_thread = fullos_->get_task_threadid(payload->source_task());
        eventcallback::ptr cb = new_callback(mpiserver::ptr(this), &mpiserver::start_send, payload);

      /** For now, assume that message stays in cache before being sent */
      lib_mem_->read(payload->get_byte_length(), src_thread, cb);
    }

    void
    mpiserver::start_send(const sst_message::ptr& msg)
    {
      handle_send(boost::dynamic_pointer_cast<mpimessage>(msg));
    }

    void
    mpiserver::handle_intranode_send(const mpimessage::ptr& payload)
    {
      taskid recver = payload->dest_task();
      softwareid sid(payload->app(), recver);
      queue_t::iterator it = queue_.find(sid);
      if (it != queue_.end()){
          //schedule the recv
          mpiqueue::ptr queue = it->second;
          event::ptr ev = new_event(queue, &mpiqueue::incoming_message, payload);
          schedule_now(ev);

          // Transfer between processes on local node.
          SSTMAC_DEBUG << "mpiserver: copy to local task \n";

          //push a nic ack to the source also
          if (payload->get_needs_ack()){
              mpimessage::ptr nicack = boost::dynamic_pointer_cast<mpimessage>(
                  payload->clone(sst_message::SEND_ACK));

              taskid sender = payload->source_task();
              softwareid sid2(payload->app(), sender);
              queue_t::iterator it2 = queue_.find(sid2);
              if (it2 != queue_.end()){
                  it2->second->handle_nic_ack(nicack);
              }
          }
      }
      else {
          /** The queue is not yet available... wait a little bit and try again */
          eventcallback::ptr cb = new_callback(mpiserver::ptr(this),
              &mpiserver::handle_intranode_send, payload);
          timestamp delay(1e-3);
          schedule(now() + delay, cb, payload);
      }
    }

    void
    mpiserver::handle_internode_send(const mpimessage::ptr& payload)
    {

      // Network transfer.
      SSTMAC_DEBUG << "mpiserver: executing os kernel to send remote \n";
      library::os_->execute_kernel(AMI::AMI_COMM_SEND, payload);
    }

    void
    mpiserver::handle_send(const mpimessage::ptr& payload)
    {
      SSTMAC_DEBUG << "mpiserver[" << os_->get_my_addr()->to_string()
          << "]: sending...\n";
      if (payload->intranode())
        {
          handle_intranode_send(payload);
        }
      else
        {
          handle_internode_send(payload);
        }
    }

    void
    mpiserver::buffered_recv(const mpimessage::ptr& msg, const mpiqueue::recvrequest::ptr &req)
    {
        eventcallback::ptr cb = new_callback(req, &mpiqueue::recvrequest::handle, msg);

      long dst_thread = fullos_->get_task_threadid(msg->dest_task());
      /** This was pushed into memory by the NIC to a temp buffer.  Therefore,
       set the src thread as the main thread... let's guess and be optimistic
       for now that this is in cache */
      lib_mem_->write(msg->get_byte_length(), dst_thread, cb);
    }

    //
    // Complete a receive.
    void
    mpiserver::handle_recv(const mpimessage::ptr& mess)
    {
      SSTMAC_DEBUG << "mpiserver[" << os_->get_my_addr()->to_string()
          << "]: handling recv - " << mess->to_string() << "\n";
      if (mess == 0)
        {
          throw illformederror("mpiserver::incoming_message:  Message is not "
            "of type mpimessage.");
        }

      queue_t::iterator it;
      if (mess->get_type() == sst_message::SEND_ACK)
        {
          taskid sender = mess->source_task();
          softwareid sid(mess->app(), sender);
          SSTMAC_DEBUG << "mpiserver[" << os_->get_my_addr()->to_string()
              << "]: checking for sender \n";
          it = queue_.find(sid);
        }
      else
        {
          taskid recver = mess->dest_task();
          softwareid sid(mess->app(), recver);
          SSTMAC_DEBUG << "mpiserver[" << os_->get_my_addr()->to_string()
              << "]: checking for receiver \n";
          it = queue_.find(sid);
        }

      if (it == queue_.end())
        {

          std::cerr << "mpiserver[" << os_->get_my_addr()->to_string()
              << "]::incoming_message:  "
              << "For app " << mess->app() << ","
              << " failed to find receiver " << mess->dest_task()
              << " or sender " << mess->source_task()
              << " in the task queues."
              << " My tasks are: [ ";
          for (queue_t::iterator a = queue_.begin(); a != queue_.end(); ++a){
              std::cerr << a->first << ", ";
          }
          std::cerr << "\n" << mess->to_string() << "\n";
          std::cerr << "]\n";
          sst_throw(illformederror, "mpiserver::incoming_message:  Message for task"
                                    " not registered with this server.");
        }
      else
        {
          SSTMAC_DEBUG << "mpiserver[" << os_->get_my_addr()->to_string()
                       << "]: delivering to queue "
                       << it->first.to_string()
                       << "\n";
          it->second->incoming_message(mess);
        }
    }

  }
} // end of namespace sstmac
