/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_MPI_MPISERVER_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_MPI_MPISERVER_H_INCLUDED

#include <sstmac/software/process/softwareid.h>
#include <sstmac/software/services/service.h>
#include <sstmac/common/nocopy.h>
#include <sstmac/common/sstmac-config.h>

#include <sstmac/common/messages/sst_message.h>
#include <sstmac/software/api/mpi/mpimessage.h>

#include <sstmac/software/libraries/compute/lib_compute_memmove.h>

#include <sstmac/common/ptr_type.h>
#include <sstmac/common/stats/stats_common.h>
#include <sstmac/common/eventmodule.h>
#include <sstmac/software/api/mpi/mpi_queue/mpiqueue-recvrequest.h>

namespace sstmac
{
  namespace sw
  {

    /// Important types.
    class mpiqueue;

    /**
     * A server that acts as a single entry/exit point for
     * mpi messages for a node.
     * this can have multiple tasks (ranks) attached to it.
     */
    class mpiserver :
        virtual public service,
        virtual public nocopy,
        public eventmodule
    {
        friend class mpiqueue;

    public:
      typedef boost::intrusive_ptr<mpiserver> ptr;

    private:

      /// The message queue for each of the peers residing on this node
      /// (indexed by their task id).
      /// These objects handle unfulfilled inbound and outbound requests,
      /// ack-ing messages, etc.
      typedef std::map<softwareid, boost::intrusive_ptr<mpiqueue> > queue_t;
      queue_t queue_;

      eventcallback::ptr sender_;

    protected:
      /// Hello.

      /**
        @param poll_delay The "polling" delay between handling requests for the MPI server
        @param nthread    The number of concurrent requests the MPI server can handle
                          simultaneously
      */
      mpiserver(int appnum);

      virtual void
      handle_recv(const mpimessage::ptr& msg);

      virtual void
      handle_send(const mpimessage::ptr& msg);

      void start_send(const sst_message::ptr& msg);

      virtual void
      handle_internode_send(const mpimessage::ptr& msg);

      virtual void
      handle_intranode_send(const mpimessage::ptr& msg);

      static debug dbg_;

      lib_compute_memmove::ptr lib_mem_;

      std::map<int, softwareid> sids_;

      int appnum_;

    public:
      void buffered_send(const mpimessage::ptr& msg);

      void buffered_recv(const mpimessage::ptr& msg, const mpiqueue::recvrequest::ptr& req);

      /// Hello.
      static ptr construct(int appnum);

      virtual service::ptr
      clone()
      {
        return construct(appnum_);
      }

      /// Goodbye.
      virtual ~mpiserver() throw ();

      virtual void start();

      /// Define a task for this server.  Each task can only be defined once
      virtual void
      add_task(const softwareid &tid,
          const boost::intrusive_ptr<mpiqueue>& peer, int rank);

      virtual void
      unregister(const softwareid &sid);

      /// Test whether the given task is defined on this node.
      virtual bool
      has_task(const softwareid &tid) const;

      /// Send data
      /// The backend has the "right" to handle local transfers differently
      /// than network transfers.
      /// The data will be sent exactly as is, it is up to the queue to
      /// decide on when to do handshakes, ready-sends, etc.
      void
      send(const mpimessage::ptr& payload);

      /// the mpi queue should be able to find this out, to tell
      /// if it should complete sends immediately or not
      bool
      nic_send_ack_supported()
      {
        return library::os_->kernel_supported(AMI::AMI_COMM_NIC_SEND_ACK);
      }

      lib_compute_memmove::ptr get_lib_mem() const { return lib_mem_; }

      virtual void
      handle(const sst_message::ptr &msg);

      virtual std::string
      to_string() const
      {
          return get_lib_name();
      }

      timestamp
      now()
      {
        return os_->now();
      }

    };

  }
} // end of namespace sstmac.

#endif
