/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_MPI_MPISTATUS_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_MPI_MPISTATUS_H_INCLUDED


#include <sstmac/software/api/mpi/mpi_comm/mpicomm.h>
#include <sstmac/software/api/mpi/mpi_types/mpiid.h>
#include <sstmac/software/api/mpi/mpi_types/mpitag.h>
#include <sstmac/software/api/mpi/mpi_types/mpitype.h>
#include <sstmac/common/messages/payload.h>

#include <sstmac/common/ptr_type.h>

namespace sstmac { namespace sw
{

  // Forward declaration of important types.
  //class payload;

  /**
   * The status of an MPI message (sent or received).
   */
  class mpistatus : public ptr_type //: public status
  {
    /// The communicator involved in this communication.
    mpicommid commid_;
    /// The source and destination for this communication.
    mpiid source_, dest_;
    /// The tag that was used.
    mpitag tag_;
    /// The datatype that was communicated.
    mpitype send_type_;
    mutable mpitype recv_type_;
    /// The content sent (may be null).
    payload::const_ptr content_;

    /// The amount of bytes I communicated.
    int64_t sent_, recvd_;

    long count_;

  protected:
    /// Hello.
    mpistatus();

    /// Copy constructor.
    mpistatus(const mpistatus &m);

  public:

    virtual std::string to_string() const{
            return "mpistatus";
          }

    typedef boost::intrusive_ptr<mpistatus> ptr;
    typedef boost::intrusive_ptr<const mpistatus> const_ptr;

    /// Make me an object.
    static mpistatus::ptr
    construct();

    /// Goodbye.
    virtual
    ~mpistatus() throw ();

    long
    get_count() const { return count_;}

    void
    set_count(long c) {count_ = c; }

    /// Get a copy of this status object.
    mpistatus::ptr
    clone() const;

    /// The communicator that was used.
    const mpicommid&
    commid() const;

    /// Register the communicator used.
    void
    set_commid(const mpicommid &c);

    /// Transmissions originated from here:
    const mpiid&
    source() const;

    /// Define the source.
    void
    set_source(const mpiid &s);

    /// Transmissions ended up here:
    const mpiid&
    dest() const;

    /// Define the destination.
    void
    set_dest(const mpiid &d);

    /// This was the tag used:
    const mpitag&
    tag() const;

    /// Define the tag.
    void
    set_tag(const mpitag &t);

    /// This is the type that was sent:
    const mpitype&
    send_type() const;

    /// Define the type.
    void
    set_send_type(const mpitype &t);

    /// This is the type that was received.
    const mpitype&
    recv_type() const;

    /// Define the received type.
    void
    set_recv_type(const mpitype &t) const;

    /// This is the content.
    payload::const_ptr
    content() const;

    /// Set the content.
    void
    set_content(const payload::const_ptr& cc);

    /// Access bytes sent.
    int64_t
    bytes_sent() const;

    /// Increment the number of bytes sent.
    void
    add_bytes_sent(int64_t bytes);

    /// Access bytes received.
    int64_t
    bytes_received() const;

    /// Increment the number of bytes received.
    void
    add_bytes_received(int64_t bytes);

  };

  typedef mpistatus::ptr mpistatus_t;
  typedef mpistatus::const_ptr const_mpistatus_t;

} } // end of namespace sstmac.

#endif
