/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/api/mpi/rma/mpi_rma_message.h>

#include <sstmac/software/api/mpi/mpi_protocol/mpiprotocol.h>

#include <sstmac/software/process/operatingsystem.h>

#include <sstmac/common/debug.h>
#include <sstmac/common/errors.h>
#include <stdlib.h>
#include <sstream>

namespace sstmac
{
  namespace sw
  {

    // Pointer types.

    //
    // Hello.
    //
    mpi_rma_message::mpi_rma_message(const std::string& libn,
        const timestamp& time, int64_t envelope, int64_t mintrans, int count,
        const mpitype &type, const mpiid &source, const mpiid &dest,
        const mpicommid &commid, int seqnum, const messageid &msgid,
        category cat, const taskid &source_task, const taskid &dest_task,
        const appid &aid, content_type_t content_type,
        const payload::const_ptr& content, const mpiprotocol::ptr& protocol,
        MESSAGE_TYPES ty, rmatype_t rmat, int win, const op_info &info) :
          mpimessage(libn, time, envelope, mintrans, count, type, source, dest,
              mpitag(-1), commid, seqnum, msgid, false, cat, source_task,
              dest_task, aid, content_type, content, protocol, ty),
          rmatype_(rmat), win_(win), info_(info)
    {

    }

    mpi_rma_message::mpi_rma_message(content_type_t content_type,
        rmatype_t rmat, int win, const op_info &info) :
      mpimessage(content_type), rmatype_(rmat), win_(win), info_(info)
    {

    }

    mpi_rma_message::mpi_rma_message(const payload::const_ptr& content,
        content_type_t content_type, rmatype_t rmat, int win,
        const op_info &info) :
      mpimessage(content, content_type), rmatype_(rmat), win_(win), info_(info)
    {

    }

    //
    // Construct a new message.
    //
    mpi_rma_message::ptr
    mpi_rma_message::construct(const std::string& libn, const timestamp& time,
        int64_t envelope, int64_t mintrans, int count, const mpitype &type,
        const mpiid &source, const mpiid &dest, const mpicommid &commid,
        int seqnum, const messageid &msgid, category cat,
        const taskid &source_task, const taskid &dest_task, const appid &aid,
        content_type_t content_type, const payload::const_ptr& content,
        const mpiprotocol::ptr& protocol, rmatype_t rmat, int win,
        const op_info &info, MESSAGE_TYPES ty)
    {

      mpi_rma_message::ptr retval(
          new mpi_rma_message(libn, time, envelope, mintrans, count, type,
              source, dest, commid, seqnum, msgid, cat, source_task, dest_task,
              aid, content_type, content, protocol, ty, rmat, win, info));
      return retval;
    }

    mpimessage::ptr
    mpi_rma_message::deep_clone() const
    {
      mpi_rma_message::ptr ret = mpi_rma_message::construct(lib_name_, time_,
          envelope_, mintrans_, count_, type_, source_, dest_, commid_,
          seqnum_, msgid_, cat_, source_task_, dest_task_, aid_, content_type_,
          content_, get_protocol(), rmatype_, win_, info_);

      ret->toaddr_ = toaddr_;
      ret->fromaddr_ = fromaddr_;
      ret->ignore_seqnum_ = ignore_seqnum_;
      return ret;
    }

    std::string
    mpi_rma_message::to_string() const
    {
      std::string b = mpimessage::to_string();

      return "(rma)" + b;
    }

    mpistatus::ptr
    mpi_rma_message::build_status() const
    {
      mpistatus::ptr stat = mpistatus::construct();

      stat->set_content(this->content());
      return stat;
    }

  }
} // end of namespace sstmac
