/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2011 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/api/mpi/sstmac_mpi.h>

#include <sstmac/software/api/mpi/mpistatus.h>
#include <sstmac/software/api/mpi/mpirequest.h>
#include <sstmac/software/api/mpi/mpi_types/mpitype.h>
#include <sstmac/software/api/mpi/mpi_types/mpiop.h>
#include <sstmac/software/api/mpi/mpi_types/mpiid.h>
#include <sstmac/software/api/mpi/mpi_types/mpitag.h>
#include <sstmac/software/api/mpi/mpi_comm/mpigroup.h>
#include <sstmac/software/api/mpi/mpi_types/mpiinfo.h>
#include <sstmac/software/api/mpi/rma/mpi_window.h>
//#include <sstmac/software/api/mpi/mpiapi.h>

#include <sstmac/software/api/mpi/mpiapi-persistent.h>

#include <boost/intrusive_ptr.hpp>
#include <sstmac/software/process/thread.h>
#include <sstmac/software/api/mpi/mpiapp.h>
#include <sstmac/software/api/mpi/mpiapi.h>
#include <iostream>

//#include <sstmac/common/messages/payload.h>
//#include <sstmac/common/messages/valuepayload.h>
//#include <sstmac/common/messages/pairpayload.h>
//#include <sstmac/common/messages/vectorpayload.h>
#include <sstmac/software/api/mpi/mpipayload.h>

#include <boost/unordered_map.hpp>

// global symbol isn't found unless this code is in sstmac namespace???
// tried "using namespace sstmac;", symbol not found


#define MAX_TYPE_COUNT 10

namespace sstmac
{
  namespace sw
  {
    class EnsureMPIFinalize
    {
    private:
      bool initialized_;
      bool finalized_;

    public:
      EnsureMPIFinalize() :
        initialized_(false), finalized_(false)
      {
      }

      void
      set_initialized()
      {
        initialized_ = true;
      }

      void
      set_finalized()
      {
        finalized_ = true;
      }

      ~EnsureMPIFinalize()
      {
        if (initialized_ && !finalized_)
        {
          std::cerr << "Program terminated without calling MPI_Finalize"
              << std::endl;
          abort();
        }
      }
    };
    static EnsureMPIFinalize ensure_mpi_finalize;

    bool has_been_initialized = false;

    static bool payloads_ = true;

    typedef boost::unordered_map<MPI_Comm, mpicomm::ptr> comm_ptr_map;
    static std::vector<comm_ptr_map> comm_ptr_maps;

    typedef boost::unordered_map<MPI_Group, mpigroup::ptr> group_ptr_map;
    static std::vector<group_ptr_map> group_ptr_maps;
    static std::vector<MPI_Group> group_counters;

    typedef boost::unordered_map<MPI_Comm, MPI_Group> comm_grp_map;
    static std::vector<comm_grp_map> comm_groups;

    typedef boost::unordered_map<MPI_Info, mpiinfo::ptr> info_map;
    static std::vector<info_map> infos;
    static std::vector<MPI_Info> info_counters;

    typedef boost::unordered_map<MPI_Win, mpi_window::ptr> win_map;
    static std::vector<win_map> windows;
    static std::vector<MPI_Win> win_counters;

    static boost::unordered_map<int, std::pair<void*, int> > attached_buffers;

    struct win_rma_info
    {
      void* buffer;
      int count;
      MPI_Datatype datatype;
      mpi_window::ptr win;
      enum rmatype
      {
        get, put, acc
      };
      rmatype type;
      mpiop op;
    };

    static boost::unordered_map<mpirequest_t, win_rma_info> pending_rmas;

    typedef boost::unordered_map<MPI_Request, mpirequest_t> req_ptr_map;
    static std::vector<req_ptr_map> req_ptr_maps;
    static std::vector<MPI_Request> req_counters;

    typedef boost::unordered_map<MPI_Comm, MPI_Errhandler> err_handler_map;
    static err_handler_map err_handlers;

    typedef boost::unordered_map<MPI_Win, MPI_Errhandler> win_err_handler_map;
    static win_err_handler_map win_err_handlers;

    static boost::unordered_map<int, keyval::ptr> keyvals;

    typedef boost::unordered_map<MPI_Request, void*> buf_map_t;
    std::vector<buf_map_t> buffers;

    static debug dbg_;

    static mpicomm::ptr
    get_comm(int rank, MPI_Comm comm)
    {
      boost::unordered_map<MPI_Comm, mpicomm::ptr>& cmap = comm_ptr_maps[rank];
      boost::unordered_map<MPI_Comm, mpicomm::ptr>::iterator it = cmap.find(
          comm);
      if (it == cmap.end())
      {
        sst_throw(ssterror, "could not find mpi communicator %d for rank %d", comm, rank);
      }
      return it->second;
    }

    static mpiinfo::ptr
    get_info(int rank, MPI_Info inf)
    {
      info_map& imap = infos[rank];
      info_map::iterator it = imap.find(inf);
      if (it == imap.end())
      {
        sst_throw(ssterror, "could not find mpi info %d for rank %d", inf, rank);
      }
      return it->second;
    }

    static mpi_window::ptr
    get_window(int rank, MPI_Win win)
    {
      win_map& wmap = windows[rank];
      win_map::iterator it = wmap.find(win);
      if (it == wmap.end())
      {
        sst_throw(ssterror, "could not find mpi window %d for rank %d", win, rank);
      }
      return it->second;
    }

    MPI_Info
    add_info_ptr(int rank, const mpiinfo::ptr& ptr)
    {
      MPI_Info inf = info_counters[rank]++;
      infos[rank][inf] = ptr;
      return inf;
    }

    MPI_Win
    add_win_ptr(int rank, const mpi_window::ptr& ptr, MPI_Win win)
    {
      windows[rank][win] = ptr;
      return win;
    }

    static mpigroup::ptr
    get_group(int rank, MPI_Group grp)
    {
      boost::unordered_map<MPI_Group, mpigroup::ptr>& gmap =
          group_ptr_maps[rank];
      boost::unordered_map<MPI_Group, mpigroup::ptr>::iterator it = gmap.find(
          grp);
      if (it == gmap.end())
      {
        sst_throw(ssterror, "could not find mpi group %d for rank %d", grp, rank);
      }
      return it->second;
    }

    static mpirequest_t
    get_request(int rank, MPI_Request req)
    {
      boost::unordered_map<MPI_Request, mpirequest_t>& rmap =
          req_ptr_maps[rank];
      boost::unordered_map<MPI_Request, mpirequest_t>::iterator it = rmap.find(
          req);
      if (it == rmap.end())
      {
        sst_throw(ssterror, "could not find mpi request %d for rank %d", req, rank);
      }
      return it->second;
    }

    MPI_Comm
    add_comm_ptr(int rank, const mpicomm::ptr& ptr)
    {
      MPI_Comm comm = ptr->id().id;
      comm_ptr_maps[rank][comm] = ptr;
      return comm;
    }

    void
    erase_comm_ptr(int rank, MPI_Comm comm)
    {
      if (comm != MPI_COMM_WORLD && comm != MPI_COMM_SELF && comm != MPI_COMM_NULL)
      {
        comm_ptr_map& cmap = comm_ptr_maps[rank];
        comm_ptr_map::iterator it = cmap.find(comm);
        if (it == cmap.end())
        {
          sst_throw(ssterror, "could not find mpi communicator %d for rank %d", comm, rank);
        }
        cmap.erase(it);
      }
    }

    MPI_Group
    add_group_ptr(int rank, const mpigroup::ptr& ptr)
    {
      MPI_Group grp = group_counters[rank]++;
      group_ptr_maps[rank][grp] = ptr;
      return grp;
    }

    void
    erase_group_ptr(int rank, MPI_Group grp)
    {
      if (grp != MPI_GROUP_EMPTY && grp != comm_groups[rank][MPI_COMM_WORLD] && grp != comm_groups[rank][MPI_COMM_SELF])
      {
        group_ptr_map& rmap = group_ptr_maps[rank];
        group_ptr_map::iterator it = rmap.find(grp);
        if (it == rmap.end())
        {
          sst_throw(ssterror, "could not find mpi group %d for rank %d", grp, rank);
        }
        rmap.erase(it);
      }
    }

    MPI_Request
    add_request_ptr(int rank, mpirequest_t ptr)
    {
      MPI_Request req = req_counters[rank]++;
      req_ptr_maps[rank][req] = ptr;
      return req;
    }

    void
    erase_request_ptr(int rank, MPI_Request req)
    {
      req_ptr_map& rmap = req_ptr_maps[rank];
      req_ptr_map::iterator it = rmap.find(req);
      if (it == rmap.end())
      {
        sst_throw(ssterror, "could not find mpi request %d for rank %d", req, rank);
      }
      rmap.erase(it);
    }

    void
    add_comm_grp(int rank, MPI_Comm comm, MPI_Group grp)
    {
      comm_groups[rank][comm] = grp;
    }

    MPI_Group
    get_comm_grp(int rank, MPI_Comm comm)
    {
      comm_grp_map& cmap = comm_groups[rank];
      comm_grp_map::iterator it = cmap.find(comm);
      if (it == cmap.end())
      {
        sst_throw(ssterror, "could not find mpi group for comm %d for rank %d", comm, rank);
      }
      return it->second;
    }

  }
}

using namespace sstmac::sw;
using namespace sstmac;

extern "C" void MPI_Disable_Payloads()
{
  payloads_ = false;
}
extern "C" void MPI_Enable_Payloads()
{
  payloads_ = true;
}

/*
 * ---------------------------------------------------------------------------
 * ------------------ Validity checkers -----------------------
 * ---------------------------------------------------------------------------
 */

void
check_key(int key, const char* here)
{
  if (keyvals.find(key) == keyvals.end())
  {
    sst_throw(ssterror, "sstmac_mpi::check_key() - could not find keyval in key_map: %s", here);
  }
}

/*
 * ---------------------------------------------------------------------------
 * --------------- Utility functions to convert to sstmac --------------------
 * ---------------------------------------------------------------------------
 */

static mpiid
get_mpi_source(int src)
{
  switch (src)
  {
  case MPI_PROC_NULL:
    return mpiid::proc_null;
  case MPI_ANY_SOURCE:
    return mpiid::any_source;
  default:
    return mpiid(src);
  }
}

static mpitag
get_mpi_tag(int tag)
{
  switch (tag)
  {
  case MPI_ANY_TAG:
    return mpitag::any_tag;
  default:
    return mpitag(tag);
  }
}

mpiapi::ptr
get_current_mpi()
{
  thread::ptr t = OS_API::get_current_thread();
  return t->get_api<mpiapi> ();
}

void
init_mpitypes()
{
  mpitype::mpi_char.native_id = MPI_CHAR;
  mpitype::mpi_unsigned_char.native_id = MPI_UNSIGNED_CHAR;

  mpitype::mpi_signed_char.native_id = MPI_SIGNED_CHAR;

  mpitype::mpi_byte.native_id = MPI_BYTE;

  mpitype::mpi_double.native_id = MPI_DOUBLE;

  mpitype::mpi_int.native_id = MPI_INT;

  mpitype::mpi_unsigned.native_id = MPI_UNSIGNED;

  mpitype::mpi_short.native_id = MPI_SHORT;

  mpitype::mpi_unsigned_short.native_id = MPI_UNSIGNED_SHORT;

  mpitype::mpi_long.native_id = MPI_LONG;

  mpitype::mpi_long_long_int.native_id = MPI_LONG_LONG_INT;

  mpitype::mpi_unsigned_long.native_id = MPI_UNSIGNED_LONG;

  mpitype::mpi_float.native_id = MPI_FLOAT;

  mpitype::mpi_double_int.native_id = MPI_DOUBLE_INT;

  mpitype::mpi_2int.native_id = MPI_2INT;

  mpitype::mpi_float_int.native_id = MPI_FLOAT_INT;

  mpitype::mpi_long_int.native_id = MPI_LONG_INT;

  mpitype::mpi_short_int.native_id = MPI_SHORT_INT;

  mpitype::mpi_long_double.native_id = MPI_LONG_DOUBLE;

  mpitype::mpi_long_double_int.native_id = MPI_LONG_DOUBLE_INT;

  mpitype::mpi_packed.native_id = MPI_PACKED;

}

mpitype
get_mpi_type(MPI_Datatype t)
{
  switch (t)
  {
  case MPI_CHAR:
    return mpitype::mpi_char;
  case MPI_UNSIGNED_CHAR:
    return mpitype::mpi_unsigned_char;
  case MPI_SIGNED_CHAR:
    return mpitype::mpi_signed_char;
  case MPI_WCHAR:
    return mpitype::mpi_wchar;
  case MPI_BYTE:
    return mpitype::mpi_byte;
  case MPI_DOUBLE:
    return mpitype::mpi_double;
  case MPI_INT:
    return mpitype::mpi_int;
  case MPI_UNSIGNED:
    return mpitype::mpi_unsigned;
  case MPI_SHORT:
    return mpitype::mpi_short;
  case MPI_UNSIGNED_SHORT:
    return mpitype::mpi_unsigned_short;
  case MPI_LONG:
    return mpitype::mpi_long;
  case MPI_LONG_LONG_INT:
    return mpitype::mpi_long_long_int;
  case MPI_UNSIGNED_LONG:
    return mpitype::mpi_unsigned_long;
  case MPI_UNSIGNED_LONG_LONG:
      return mpitype::mpi_unsigned_long_long;
  case MPI_FLOAT:
    return mpitype::mpi_float;

  case MPI_DOUBLE_INT:
    return mpitype::mpi_double_int;

  case MPI_2INT:
    return mpitype::mpi_2int;

  case MPI_FLOAT_INT:
    return mpitype::mpi_float_int;

  case MPI_LONG_INT:
    return mpitype::mpi_long_int;

  case MPI_SHORT_INT:
    return mpitype::mpi_short_int;

  case MPI_LONG_DOUBLE:
    return mpitype::mpi_long_double;

  case MPI_LONG_DOUBLE_INT:
    return mpitype::mpi_long_double_int;

  case MPI_INT8_T:
    return mpitype::mpi_char;

  case MPI_INT16_T:
    return mpitype::mpi_short;

  case MPI_INT32_T:
    return mpitype::mpi_int;

  case MPI_INT64_T:
    return mpitype::mpi_long;

  case MPI_UINT8_T:
    return mpitype::mpi_unsigned_char;

  case MPI_UINT16_T:
    return mpitype::mpi_unsigned_short;

  case MPI_UINT32_T:
    return mpitype::mpi_unsigned;

  case MPI_UINT64_T:
    return mpitype::mpi_unsigned_long;

  case MPI_LB:
    return mpitype::mpi_lb;

  case MPI_UB:
    return mpitype::mpi_ub;

  case MPI_PACKED:
    return mpitype::mpi_packed;

  case MPI_INTEGER1:
    return mpitype::mpi_integer1;

  case MPI_INTEGER2:
    return mpitype::mpi_integer2;

  case MPI_INTEGER4:
    return mpitype::mpi_integer4;

  case MPI_INTEGER8:
    return mpitype::mpi_integer8;

  case MPI_REAL4:
    return mpitype::mpi_real4;

  case MPI_REAL8:
    return mpitype::mpi_real8;

  case MPI_REAL16:
    return mpitype::mpi_real16;

  case MPI_INTEGER:
    return mpitype::mpi_integer;

  case MPI_2INTEGER:
    return mpitype::mpi_2integer;

  case MPI_2REAL:
    return mpitype::mpi_2real;

  case MPI_2DOUBLE_PRECISION:
    return mpitype::mpi_2double_precision;

  case MPI_REAL:
    return mpitype::mpi_real;
  case MPI_COMPLEX:
    return mpitype::mpi_complex;

  case MPI_DOUBLE_COMPLEX:
    return mpitype::mpi_double_complex;

  case MPI_LOGICAL:
    return mpitype::mpi_logical;

  case MPI_CHARACTER:
    return mpitype::mpi_character;

  case MPI_DOUBLE_PRECISION:
    return mpitype::mpi_double_precision;

  case MPI_COMPLEX8:
    return mpitype::mpi_complex8;

  case MPI_COMPLEX16:
    return mpitype::mpi_complex16;

  case MPI_COMPLEX32:
    return mpitype::mpi_complex32;

  default:
    mpiapi::ptr app = get_current_mpi();
    try
    {
      return app->get_derived_type(t);
    }
    catch (...)
    {
      sst_throw(sstmac::ssterror, "sstmac_mpi::get_mpi_type() - unknown type %d", t);
    }
  }
}

MPI_Datatype
get_mpi_datatype(mpitype t)
{
  if (t == mpitype::mpi_char)
    return MPI_CHAR;
  else if (t == mpitype::mpi_unsigned_char)
    return MPI_UNSIGNED_CHAR;
  else if (t == mpitype::mpi_signed_char)
    return MPI_SIGNED_CHAR;
  else if (t == mpitype::mpi_byte)
    return MPI_BYTE;
  else if (t == mpitype::mpi_int)
    return MPI_INT;
  else if (t == mpitype::mpi_unsigned)
    return MPI_UNSIGNED;
  else if (t == mpitype::mpi_short)
    return MPI_SHORT;
  else if (t == mpitype::mpi_unsigned_short)
    return MPI_UNSIGNED_SHORT;
  else if (t == mpitype::mpi_double)
    return MPI_DOUBLE;
  else if (t == mpitype::mpi_long)
    return MPI_LONG;
  else if (t == mpitype::mpi_unsigned_long)
    return MPI_UNSIGNED_LONG;
  else if (t == mpitype::mpi_float)
    return MPI_FLOAT;

  else if (t == mpitype::mpi_double_int)
    return MPI_DOUBLE_INT;

  else if (t == mpitype::mpi_2int)
    return MPI_2INT;

  else if (t == mpitype::mpi_float_int)
    return MPI_FLOAT_INT;

  else if (t == mpitype::mpi_long_int)
    return MPI_LONG_INT;

  else if (t == mpitype::mpi_short_int)
    return MPI_SHORT_INT;

  else if (t == mpitype::mpi_long_double_int)
    return MPI_LONG_DOUBLE_INT;
  else if (t == mpitype::mpi_long_double)
    return MPI_LONG_DOUBLE;

  else if (t == mpitype::mpi_char)
    return MPI_INT8_T;

  else if (t == mpitype::mpi_short)
    return MPI_INT16_T;

  else if (t == mpitype::mpi_int)
    return MPI_INT32_T;

  else if (t == mpitype::mpi_long)
    return MPI_INT64_T;

  else if (t == mpitype::mpi_unsigned_char)
    return MPI_UINT8_T;

  else if (t == mpitype::mpi_unsigned_short)
    return MPI_UINT16_T;

  else if (t == mpitype::mpi_unsigned)
    return MPI_UINT32_T;

  else if (t == mpitype::mpi_unsigned_long)
    return MPI_UINT64_T;
  else
  {
    throw sstmac::ssterror("sstmac_mpi::get_mpi_datatype() - unknown type");
  }
}


mpiop
get_mpi_op(int o)
{
  switch (o)
  {
  case MPI_MAX:
    return mpiop::max;
  case MPI_MIN:
    return mpiop::min;
  case MPI_SUM:
    return mpiop::sum;
  case MPI_PROD:
    return mpiop::prod;
  case MPI_LAND:
    return mpiop::land;
  case MPI_BAND:
    return mpiop::band;
  case MPI_LOR:
    return mpiop::lor;
  case MPI_BOR:
    return mpiop::bor;
  case MPI_LXOR:
    return mpiop::lxor;
  case MPI_BXOR:
    return mpiop::bxor;
  case MPI_MAXLOC:
    return mpiop::maxloc;
  case MPI_MINLOC:
    return mpiop::minloc;
  case MPI_REPLACE:
    return mpiop::replace;
  default:
    mpiapi::ptr app = get_current_mpi();
    try
    {
      return app->get_created_op(o);
    }
    catch (...)
    {
      sst_throw(sstmac::ssterror, "sstmac_mpi::get_mpi_op() - unknown operation %d", o);
    }
  }
}

// ----- These functions were added to support SST/macro payloads,
// ----- should be called before and after mpi calls that use real payloads


payload::const_ptr
get_mpi_payload_func(const void* buf, int offset, int count, MPI_Datatype type,
    bool usewrapper)
{
  if (buf == NULL || !payloads_ || count == 0)
  {
    SSTMAC_DEBUG << "SSTMAC_MPI: get_mpi_payload -- no payload \n";
    return payload::null();
  }
  else
  {
    mpitype t = get_mpi_type(type);
    char* casted = (char*) buf;
    casted += (offset * t.get_extent());

    return mpipayload::construct(casted, t, count, usewrapper);
  }
}

payload::const_ptr
get_mpi_payload(const void* buf, int offset, int count, MPI_Datatype type)
{
  return get_mpi_payload_func(buf, offset, count, type, false);
}

bool
get_mpi_vector_payload_func(const void* buf, int nrank, int count,
    MPI_Datatype data_type, std::vector<payload::const_ptr>& ret,
    bool usewrapper)
{
  if (buf == NULL || !payloads_)
  {
    SSTMAC_DEBUG << "SSTMAC_MPI: get_mpi_vector_payload -- no payload \n";
    //we need to make sure the ret is at least the correct size
    ret.resize(nrank);
    return false;
  }
  else
  {
    mpitype t = get_mpi_type(data_type);
    ret.resize(nrank);
    char* casted = (char*) buf;
    for (int i = 0; i < nrank; i++)
    {
      ret[i] = mpipayload::construct(casted + i * count * t.get_extent(), t, count,
          usewrapper);
    }
    return true;
  }
}

bool
get_mpi_vector_payload(const void* buf, int nrank, int count,
    MPI_Datatype data_type, std::vector<payload::const_ptr>& ret)
{
  return get_mpi_vector_payload_func(buf, nrank, count, data_type, ret, false);
}



void
set_mpi_payload(void* buf, int offset, int count, mpitype type,
    payload::const_ptr load)
{
  if (buf == NULL || load == payload::null() || !payloads_)
  {
    SSTMAC_DEBUG << "SSTMAC_MPI: set_mpi_payload -- no payload : payloads_ = "
          << payloads_ << ", buf = " << buf << ", load = " << load << " \n";
    return;
  }
  else
  {
    mpipayload::const_ptr mpiload = boost::dynamic_pointer_cast<
        const mpipayload>(load);
    char* casted = (char*) buf;
    casted += (offset * type.get_extent());
    mpiload->extract(casted, count, type);

  }
}

void
set_mpi_payload(void* buf, int offset, int count, MPI_Datatype type,
    payload::const_ptr load){
  mpitype t = get_mpi_type(type);
  set_mpi_payload(buf, offset, count, t, load);
}

void
set_mpi_payload(void* buf, int count, MPI_Datatype data_type,
    std::vector<payload::const_ptr> load, const char* here)
{
  if (buf == NULL || !payloads_)
  {
    SSTMAC_DEBUG << "SSTMAC_MPI: set_mpi_payload -- no payload \n";
    return;
  }
  else
  {
    mpitype t = get_mpi_type(data_type);
    char* casted = (char*) buf;
    for (int i = 0; i < load.size(); i++)
    {
      mpipayload::const_ptr mpiload = boost::dynamic_pointer_cast<
          const mpipayload>(load[i]);

      int cnt = mpiload->get_count() * mpiload->get_type().get_packed_size() / t.get_packed_size();

      mpiload->extract(casted, cnt, t);

      casted += (cnt * t.get_extent());

      SSTMAC_DEBUG2 << "set_mpi_payload with vector: advancing buffer " << (cnt * t.get_extent()) << " bytes \n";
    }

  }
}

void
set_mpi_vector_payload(void* buf, int nrank, int count, MPI_Datatype data_type,
    std::vector<payload::const_ptr> load)
{
  if (buf == NULL || !payloads_)
  {
    return;
  }
  else
  {
    if ((int) load.size() != nrank)
    {
      throw ssterror(
          "sstmac_mpi::set_mpi_payload(vector) - count does not match vector size");
    }
    mpitype t = get_mpi_type(data_type);
    char* casted = (char*) buf;
    for (int i = 0; i < load.size(); i++)
    {
      mpipayload::const_ptr mpiload = boost::dynamic_pointer_cast<
          const mpipayload>(load[i]);

      int cnt = mpiload->get_count() * mpiload->get_type().get_packed_size() / t.get_packed_size();

      mpiload->extract(casted, cnt, t);

      casted += (cnt * t.get_extent());
    }
  }
}

void
get_displaced_payload(void * buf, int nranks, int *sendcnts, int *displs,
    MPI_Datatype datatype, std::vector<payload::const_ptr> &load)
{
  for (int i = 0; i < nranks; ++i)
    load.push_back(get_mpi_payload(buf, displs[i], sendcnts[i], datatype));
}

void
set_displaced_payload(void * buf, int nranks, int *recvcnts, int *displs,
    MPI_Datatype datatype, std::vector<payload::const_ptr> load)
{
  for (int i = 0; i < nranks; ++i)
    set_mpi_payload(buf, displs[i], recvcnts[i], datatype, load[i]);
}

extern "C" void MPI_Print(const char* msg)
{
  SSTMAC_DEBUG << msg << "\n";
}

/*
 * ---------------------------------------------------------------------------
 * ------------------ The real MPI functions ---------------------------------
 * ---------------------------------------------------------------------------
 */

int* const MPI_UNWEIGHTED = (int* const ) 0x1;

/// this guy is useful for libraries
extern "C" long MPI_NodeAddress(int rank, MPI_Comm c)
{
  static const char* here = "MPI_NodeAddress";
  SSTMAC_DEBUG << here << "\n";
  mpicomm::ptr com = get_comm(rank, c);
  mpiid r(rank);
  return com->node_at(r)->unique_id();
}

extern "C" long MPI_Taskid(int rank, MPI_Comm c)
{
  static const char* here = "MPI_NodeAddress";
  SSTMAC_DEBUG << here << "\n";
  mpicomm::ptr com = get_comm(rank, c);
  mpiid r(rank);
  return com->peer_task(r).id;
}

extern "C" void MPI_Abort(MPI_Comm c, int code)
{
  sst_throw(ssterror, "sstmac_mpi::MPI_Abort with code %d -- this function is not actually implemented properly", code);
}

extern "C" int MPI_Init(int *argc, char ***argv)
{
  if (!dbg_.is_real())
  {
    dbg_ = sstmac::debug("<debug> sstmac_mpi", sstmac::debug::magenta);
    sstmac::logger::register_logger(&dbg_);
  }
  SSTMAC_DEBUG << "MPI_Init \n";
  mpiapi::ptr mpi = get_current_mpi();
  mpi->init();

  mpicomm::ptr world = mpi->comm_world();
  world->set_name("MPI_COMM_WORLD");

  mpicomm::ptr self = mpi->comm_self();
  self->set_name("MPI_COMM_SELF");
  if (comm_ptr_maps.empty())
  {
    int nproc = world->size().id;
    comm_ptr_maps.resize(nproc);
    group_ptr_maps.resize(nproc);
    group_counters.resize(nproc, 0);
    req_ptr_maps.resize(nproc);
    req_counters.resize(nproc, 0);
    infos.resize(nproc);
    windows.resize(nproc);
    info_counters.resize(nproc, 0);
    win_counters.resize(nproc, 0);
    comm_groups.resize(nproc);
    buffers.resize(nproc);

    init_mpitypes();
  }

  int rank = world->rank().id;
  MPI_Comm worldcomm = add_comm_ptr(world->rank().id, world);
  MPI_Group worldgrp = add_group_ptr(rank, world->get_group());
  add_comm_grp(rank, worldcomm, worldgrp);

  MPI_Comm selfcomm = add_comm_ptr(world->rank().id, self);
  MPI_Group selfgrp = add_group_ptr(rank, self->get_group());
  add_comm_grp(rank, selfcomm, selfgrp);

  //empty group
  std::vector<taskid> vec;
  tasklist::ptr t = tasklist::construct(vec);
  mpigroup::ptr newgrp = mpigroup::construct(t);
  group_ptr_maps[rank][MPI_GROUP_EMPTY] = newgrp;


  err_handlers[MPI_COMM_WORLD] = MPI_ERRORS_ARE_FATAL;
  err_handlers[MPI_COMM_SELF] = MPI_ERRORS_ARE_FATAL;

  ensure_mpi_finalize.set_initialized();

  return MPI_SUCCESS; //yes, error handling would be nice
}

extern "C" int MPI_Initialized(int *flag)
{
  SSTMAC_DEBUG << "MPI_Initialized \n";
  mpiapi::ptr app = get_current_mpi();
  *flag = (int) app->initialized();
  return MPI_SUCCESS; //yes, error handling would be nice
}

extern "C" int MPI_Finalize()
{
  SSTMAC_DEBUG << "MPI_Finalize \n";
  mpiapi::ptr app = get_current_mpi();
  mpicomm::ptr world = app->comm_world();
  int rank = world->rank().id;
  app->finalize();

  //erase_comm_ptr(world->rank().id, world->id().id);

  //these all need to get cleared
  //otherwise cxa_finalize can freak out
  //with memory errors
  ensure_mpi_finalize.set_finalized();
  comm_ptr_maps.clear();
  group_ptr_maps.clear();
  comm_groups.clear();
  infos.clear();
  windows.clear();
  pending_rmas.clear();
  req_ptr_maps.clear();
  keyvals.clear();

  return MPI_SUCCESS; //yes, error handling would be nice
}

static void
complete_wait(int rank, const mpirequest_t& req, const_mpistatus_t& stat,
    MPI_Request* request, MPI_Status* status)
{
  buf_map_t& buf_map = buffers[rank];
  buf_map_t::iterator it = buf_map.find(*request);

  if (!req->is_cancelled())
  {
    if (status != MPI_STATUS_IGNORE && stat)
    {
      status->MPI_SOURCE = stat->source().id;
      status->MPI_TAG = stat->tag().id;
      status->MPI_ERROR = 0;
      status->COUNT = stat->get_count();
      status->CANCELLED = false;
    }

    if (it != buf_map.end())
    {
      MPI_Recover_Payload(it->second, *request);
    }
  }else{
    status->CANCELLED = true;
  }
  if (!req->is_persistent())
  {
    MPI_Request_free(request);
    if (it != buf_map.end())
      buf_map.erase(*request);
    *request = MPI_REQUEST_NULL;
  }else{
    mpiapi::persistent::ptr pers = boost::dynamic_pointer_cast<mpiapi::persistent>(req);
    pers->started_ = false;
  }

}

extern "C" int MPI_Wait(MPI_Request *request, MPI_Status *status)
{
  static const char* here = "MPI_Wait";
  SSTMAC_DEBUG << here << "\n";
  if (request == 0)
  {
    sst_throw(ssterror, "sstmac_mpi::MPI_Wait: how dare you pass a null pointer to a request \n");
  }
  else if (*request >= 0)
  {
    mpiapi::ptr app = get_current_mpi();

    int rank = app->comm_world()->rank().id;
    mpirequest_t req = get_request(rank, *request);
    const_mpistatus_t stat;
    if (!req->is_cancelled())
    {

      app->wait(req, stat);
    }
    complete_wait(rank, req, stat, request, status);

  }
  return MPI_SUCCESS; //yes, error handling would be nice
}

extern "C" int MPI_Waitall(int count, MPI_Request array_of_requests[],
    MPI_Status array_of_statuses[])
{
  static const char* here = "MPI_Waitall";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  SSTMAC_DEBUG << here << "(" << mpi->comm_rank(mpi->comm_world()).id << ") \n";
  std::vector<const_mpistatus_t> sst_stats;
  std::vector<mpirequest_t> sst_reqs;
  std::vector<MPI_Request*> mpi_reqs;
  std::vector<int> indices;
  for (int i = 0; i < count; i++)
  {
    if (array_of_requests[i] != MPI_REQUEST_NULL)
    {
      sst_reqs.push_back(get_request(rank, array_of_requests[i]));
      mpi_reqs.push_back(&array_of_requests[i]);
      sst_stats.push_back(const_mpistatus_t());
      indices.push_back(i);
    }
  }

  mpi->waitall(sst_reqs, sst_stats, true);

  int num_completed = mpi_reqs.size();
  for (int i = 0; i < num_completed; ++i)
  {
    int idx = indices[i];
    MPI_Status* status =
        array_of_statuses == MPI_STATUSES_IGNORE ? MPI_STATUS_IGNORE
            : &array_of_statuses[idx];
    complete_wait(rank, sst_reqs[i], sst_stats[i], mpi_reqs[i], status);

  }

  return MPI_SUCCESS; //yes, error handling would be nice
}

extern "C" int MPI_Waitany_timeout(int count, MPI_Request array_of_requests[],
    int *index, MPI_Status *status, double timeout)
{
  static const char* here = "MPI_Waitany";

  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  SSTMAC_DEBUG << here << "(" << mpi->comm_rank(mpi->comm_world()).id << ") \n";
  const_mpistatus_t stat;
  std::vector<mpirequest_t> sst_reqs;
  std::vector<MPI_Request*> mpi_reqs;
  for (int i = 0; i < count; i++)
  {

    if (array_of_requests[i] != MPI_REQUEST_NULL)
    {
      SSTMAC_DEBUG << "adding request " << array_of_requests[i]
            << " to the list \n";
      sst_reqs.push_back(get_request(rank, array_of_requests[i]));
      mpi_reqs.push_back(&array_of_requests[i]);
    }
  }
  if (mpi_reqs.size() > 0)
  {
    int ind;
    mpi->waitany(sst_reqs, ind, stat, timestamp(timeout));

    if (ind >= 0)
    {
      complete_wait(rank, sst_reqs[ind], stat, mpi_reqs[ind], status);
      for (int i = 0; i < count; i++)
      {
        if (*(mpi_reqs[ind]) == (array_of_requests[i]))
        {
          *index = i;
          break;
        }
      }
    }
  }else{
    *index = MPI_UNDEFINED;
  }

  return MPI_SUCCESS; //yes, error handling would be nice
}

extern "C" int MPI_Waitany(int count, MPI_Request array_of_requests[],
    int *index, MPI_Status *status)
{
  static const char* here = "MPI_Waitany";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;

  SSTMAC_DEBUG << here << "(" << mpi->comm_rank(mpi->comm_world()).id << ") \n";
  const_mpistatus_t stat;
  std::vector<mpirequest_t> sst_reqs;
  std::vector<MPI_Request*> mpi_reqs;
  for (int i = 0; i < count; i++)
  {
    if (array_of_requests[i] != MPI_REQUEST_NULL)
    {

      SSTMAC_DEBUG << "adding request " << array_of_requests[i]
            << " to the list \n";
      sst_reqs.push_back(get_request(rank, array_of_requests[i]));
      mpi_reqs.push_back(&array_of_requests[i]);
    }
  }
  if (mpi_reqs.size() > 0)
  {
    int ind;
    mpi->waitany(sst_reqs, ind, stat);

    for (int i = 0; i < count; i++)
    {
      if (*(mpi_reqs[ind]) == (array_of_requests[i]))
      {
        *index = i;
        break;
      }
    }

    complete_wait(rank, sst_reqs[ind], stat, mpi_reqs[ind], status);
  }else{
    *index = MPI_UNDEFINED;
  }

  return MPI_SUCCESS; //yes, error handling would be nice
}

extern "C" int MPI_Waitsome(int incount, MPI_Request array_of_requests[],
    int *outcount, int array_of_indices[], MPI_Status array_of_statuses[])
{
  static const char* here = "MPI_Waitsome";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  std::vector<mpirequest_t> sst_reqs;
  for (int i = 0; i < incount; i++)
  {
    if (array_of_requests[i] != MPI_REQUEST_NULL)
    {
      sst_reqs.push_back(get_request(rank, array_of_requests[i]));
    }
  }

  std::vector<int> ind;
  std::vector<const_mpistatus_t> sst_stats;
  mpi->waitsome(sst_reqs, ind, sst_stats);

  *outcount = sst_stats.size();
  for (uint i = 0; i < ind.size(); i++)
  {
    int idx = ind[i];
    MPI_Request* mpi_req = &array_of_requests[idx];
    MPI_Status* mpi_status =
        array_of_statuses == MPI_STATUSES_IGNORE ? MPI_STATUS_IGNORE
            : &array_of_statuses[idx];
    complete_wait(rank, sst_reqs[idx], sst_stats[i], mpi_req, mpi_status);
    array_of_indices[i] = idx;
  }

  return MPI_SUCCESS; //yes, error handling would be nice
}

extern "C" int MPI_Get_count(MPI_Status *status, MPI_Datatype datatype,
    int *count)
{
  static const char* here = "MPI_Get_count";
  SSTMAC_DEBUG << here << "\n";
  // throw sstmac::ssterror("MPI_Get_count is unimplemented");
  mpitype t = get_mpi_type(datatype);
  if(t.get_packed_size() > 0){
    *count = status->COUNT / t.get_packed_size();
    if(*count * t.get_packed_size() != status->COUNT)
      *count = MPI_UNDEFINED;  //not a round number of things
  }
  else
    *count = 0;


  return MPI_SUCCESS; //yes, error handling would be nice
}

extern "C" int MPI_Test(MPI_Request *request, int *flag, MPI_Status *status)
{
  static const char* here = "MPI_Test";
  SSTMAC_DEBUG << here << "\n";
  if (*request != MPI_REQUEST_NULL)
  {
    mpiapi::ptr mpi = get_current_mpi();
    int rank = mpi->comm_world()->rank().id;
    bool f;
    mpirequest_t req = get_request(rank, *request);
    const_mpistatus_t stat;
    mpi->test(req, f, stat);
    if (f)
    {
      complete_wait(rank, req, stat, request, status);
    }
    *flag = f;
  }else{
    *flag = true;
    status->COUNT = 0;
    status->MPI_SOURCE = MPI_ANY_SOURCE;
    status->MPI_TAG = MPI_ANY_TAG;
  }
  return MPI_SUCCESS; //yes, error handling would be nice
}

extern "C" int MPI_Probe(int source, int tag, MPI_Comm comm, MPI_Status *status)
{
  static const char* here = "MPI_Probe";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  const_mpistatus_t stat;
  mpicomm::ptr commptr = get_comm(rank, comm);
  mpi->probe(get_mpi_source(source), get_mpi_tag(tag), commptr, stat);
  status->MPI_SOURCE = stat->source().id;
  status->MPI_TAG = stat->tag().id;
  status->COUNT = stat->get_count();
  status->MPI_ERROR = 0;
  status->CANCELLED = false;
  return MPI_SUCCESS; //yes, error handling would be nice
}

extern "C" int MPI_Cancel(MPI_Request *request)
{
  static const char* here = "MPI_Cancel";
  SSTMAC_DEBUG << here << "\n";
  if (request >= 0)
  {
    mpiapi::ptr mpi = get_current_mpi();
    int rank = mpi->comm_world()->rank().id;
    mpirequest_t req = get_request(rank, *request);
    mpi->cancel(req);
    //erase_request_ptr(rank, *request);
  }
  return MPI_SUCCESS; //yes, error handling would be nice
}

extern "C" int MPI_Request_free(MPI_Request *request)
{
  static const char* here = "MPI_Request_free";
  SSTMAC_DEBUG << here << "\n";
  if (*request != MPI_REQUEST_NULL)
  {
    mpiapi::ptr mpi = get_current_mpi();
    int rank = mpi->comm_world()->rank().id;
    erase_request_ptr(rank, *request);
    *request = MPI_REQUEST_NULL;
  }
  return MPI_SUCCESS;
}

extern "C" int MPI_Comm_dup(MPI_Comm comm, MPI_Comm *newcomm)
{
  static const char* here = "MPI_Comm_dup";
  SSTMAC_DEBUG << here << "(src comm= " << comm << ")\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);
  mpicomm::ptr ret;
  mpi->comm_dup(commptr, ret);
  *newcomm = add_comm_ptr(rank, ret);
  MPI_Group g = add_group_ptr(rank, ret->get_group());
  add_comm_grp(rank, *newcomm, g);
  SSTMAC_DEBUG << "MPI_Comm_dup: src comm: " << comm << " created comm "
        << *newcomm << "\n";
  err_handlers[*newcomm] = err_handlers[comm];
  return MPI_SUCCESS;
}

extern "C" int MPI_Comm_rank(MPI_Comm comm, int *rankptr)
{
  static const char* here = "MPI_Comm_rank";
  SSTMAC_DEBUG << here << "\n";
  int rank = get_current_mpi()->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);
  *rankptr = commptr->rank().id;
  return MPI_SUCCESS;
}

extern "C" int MPI_Comm_split(MPI_Comm comm, int color, int key,
    MPI_Comm *newcomm)
{
  static const char* here = "MPI_Comm_split";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);
  SSTMAC_DEBUG << "[rank=" << rank << "] " << here << "(" << comm << ", "
        << color << ", " << key << ") \n";

  mpicomm::ptr ret;
  mpi->comm_split(commptr, color, key, ret);
  if (ret != mpicomm::comm_null)
  {
    *newcomm = add_comm_ptr(rank, ret);
    MPI_Group g = add_group_ptr(rank, ret->get_group());
    add_comm_grp(rank, *newcomm, g);
  }
  else
    *newcomm = MPI_COMM_NULL;
  return MPI_SUCCESS;
}

extern "C" int MPI_Comm_size(MPI_Comm comm, int *size)
{
  static const char* here = "MPI_Comm_size";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);
  *size = commptr->size().id;
  return MPI_SUCCESS;
}

extern "C" int MPI_Comm_free(MPI_Comm *comm)
{
  static const char* here = "MPI_Comm_free";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, *comm);
  mpi->comm_free(commptr);
  erase_comm_ptr(rank, *comm);
  // MPI_Group g = get_comm_grp(rank, *comm);
  // erase_group_ptr(rank, g);
  err_handlers.erase(*comm);
  *comm = MPI_COMM_NULL;
  return MPI_SUCCESS;
}

extern "C" int MPI_Comm_create(MPI_Comm comm, MPI_Group group,
    MPI_Comm *newcomm)
{
  static const char* here = "MPI_Comm_create";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);
  mpigroup::ptr grp_ptr = get_group(rank, group);
  mpicomm::ptr ret;
  mpi->comm_create(commptr, grp_ptr, ret);
  if(ret == mpicomm::comm_null){
    *newcomm = MPI_COMM_NULL;
  }else{
    *newcomm = add_comm_ptr(rank, ret);
    err_handlers[*newcomm] = err_handlers[comm];
  }
  return MPI_SUCCESS;
}

extern "C" int MPI_Comm_group(MPI_Comm comm, MPI_Group *group)
{
  static const char* here = "MPI_Comm_group";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  *group = get_comm_grp(rank, comm);
  mpicomm::ptr commptr = get_comm(rank, comm);
  mpigroup::ptr grp_ptr = commptr->get_group();
  return MPI_SUCCESS;
}

extern "C" int MPI_Group_incl(MPI_Group group, int n, int *ranks,
    MPI_Group *newgroup)
{
  static const char* here = "MPI_Group_incl";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpigroup::ptr grp_ptr = get_group(rank, group);
  std::vector<taskid> vec(n);

  if (n > grp_ptr->size())
  {
    sst_throw(illformederror,
        "MPI_Group_incl: invalid group size %d", n);
  }

  for (int i = 0; i < n; i++)
  {
    vec[i] = grp_ptr->at(ranks[i]);
  }

  if (vec.size() > 0)
  {
    tasklist::ptr t = tasklist::construct(vec);
    mpigroup::ptr newgrp = mpigroup::construct(t);
    *newgroup = add_group_ptr(rank, newgrp);
  }
  else
  {
    *newgroup = MPI_GROUP_EMPTY;
  }
  return MPI_SUCCESS;
}
extern "C" int MPI_Group_free(MPI_Group *group)
{
  static const char* here = "MPI_Group_free";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;

  erase_group_ptr(rank, *group);
  *group = MPI_GROUP_NULL;
  return MPI_SUCCESS;
}

typedef timestamp
(mpiapi::*mpiapi_send_fxn)(int count, const mpitype &type, const mpiid &target,
    const mpitag &tag, const mpicomm::ptr &comm,
    const payload::const_ptr& content);

static int
mpi_blocking_send(void *buf, int count, MPI_Datatype datatype, int dest,
    int tag, MPI_Comm comm, mpiapi_send_fxn fxn)
{
  static const char* here = "MPI_Send";
  SSTMAC_DEBUG << here << "-" << (long) buf << ", " << count << ", "
        << datatype << ", " << dest << ", " << tag << ", " << comm << "\n";

  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);

  if (dest == MPI_PROC_NULL)
  {
    return MPI_SUCCESS;
  }

  sstmac::payload::const_ptr load = get_mpi_payload(buf, 0, count, datatype);

  (mpi.get()->*fxn)(count, get_mpi_type(datatype), mpiid(dest), mpitag(tag),
      commptr, load);

  return MPI_SUCCESS;
}

extern "C" int MPI_Ssend(void *buf, int count, MPI_Datatype datatype, int dest,
    int tag, MPI_Comm comm)
{
  return mpi_blocking_send(buf, count, datatype, dest, tag, comm,
      &mpiapi::ssend);
}

extern "C" int MPI_Rsend(void *buf, int count, MPI_Datatype datatype, int dest,
    int tag, MPI_Comm comm)
{
  return mpi_blocking_send(buf, count, datatype, dest, tag, comm,
      &mpiapi::rsend);
}

extern "C" int MPI_Bsend(void *buf, int count, MPI_Datatype datatype, int dest,
    int tag, MPI_Comm comm)
{
  return mpi_blocking_send(buf, count, datatype, dest, tag, comm,
      &mpiapi::bsend);
}

extern "C" int MPI_Send(void *buf, int count, MPI_Datatype datatype, int dest,
    int tag, MPI_Comm comm)
{
  return mpi_blocking_send(buf, count, datatype, dest, tag, comm, &mpiapi::send);
}

extern "C" int MPI_Sendrecv(void *sendbuf, int sendcount,
    MPI_Datatype sendtype, int dest, int sendtag, void *recvbuf, int recvcount,
    MPI_Datatype recvtype, int source, int recvtag, MPI_Comm comm,
    MPI_Status* status)
{
  static const char* here = "MPI_Sendrecv";
  SSTMAC_DEBUG << here << "\n";

  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);
  const_mpistatus_t stat;

  if (dest == MPI_PROC_NULL || source == MPI_PROC_NULL)
  {
    return MPI_SUCCESS;
  }

  payload::const_ptr load = get_mpi_payload(sendbuf, 0, sendcount, sendtype);

  mpi->sendrecv(sendcount, get_mpi_type(sendtype), mpiid(dest),
      mpitag(sendtag), recvcount, get_mpi_type(recvtype), get_mpi_source(source),
      mpitag(recvtag), commptr, stat, load);

  set_mpi_payload(recvbuf, 0, recvcount, recvtype, stat->content());

  if (status)
  {
    status->MPI_SOURCE = stat->source().id;
    status->MPI_TAG = stat->tag().id;
    status->MPI_ERROR = 0;
    status->COUNT = stat->get_count();
    status->CANCELLED = false;
  }

  return MPI_SUCCESS;
}

typedef timestamp
(mpiapi::*mpiapi_isend_fxn)(int count, const mpitype &type,
    const mpiid &target, const mpitag &tag, const mpicomm::ptr &comm,
    mpirequest_t &req, const payload::const_ptr& content);

static MPI_Request
mpi_immediate_send(const void *buf, int count, MPI_Datatype datatype, int dest,
    int tag, MPI_Comm comm, MPI_Request *request, mpiapi_isend_fxn fxn)
{
  static const char* here = "MPI_Isend";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);
  mpirequest_t req;

  if (dest == MPI_PROC_NULL)
  {
    return MPI_SUCCESS;
  }

  sstmac::payload::const_ptr load = get_mpi_payload(buf, 0, count, datatype);

  (mpi.get()->*fxn)(count, get_mpi_type(datatype), mpiid(dest), mpitag(tag),
      commptr, req, load);
  *request = add_request_ptr(rank, req);

  SSTMAC_DEBUG << "MPI_Isend(" << mpi->comm_rank(commptr).id
        << "): returning request " << *request << " key="
        << (long) req->get_key().get() << "\n";

  return MPI_SUCCESS;
}

extern "C" int MPI_Isend(void *buf, int count, MPI_Datatype datatype, int dest,
    int tag, MPI_Comm comm, MPI_Request *request)
{
  return mpi_immediate_send(buf, count, datatype, dest, tag, comm, request,
      &mpiapi::isend);
}

extern "C" int MPI_Ibsend(void *buf, int count, MPI_Datatype datatype,
    int dest, int tag, MPI_Comm comm, MPI_Request *request)
{
  return mpi_immediate_send(buf, count, datatype, dest, tag, comm, request,
      &mpiapi::ibsend);
}

extern "C" int MPI_Issend(void *buf, int count, MPI_Datatype datatype,
    int dest, int tag, MPI_Comm comm, MPI_Request *request)
{
  return mpi_immediate_send(buf, count, datatype, dest, tag, comm, request,
      &mpiapi::issend);
}

extern "C" int MPI_Irsend(void *buf, int count, MPI_Datatype datatype,
    int dest, int tag, MPI_Comm comm, MPI_Request *request)
{
  return mpi_immediate_send(buf, count, datatype, dest, tag, comm, request,
      &mpiapi::irsend);
}

extern "C" int MPI_Recv(void *buf, int count, MPI_Datatype datatype,
    int source, int tag, MPI_Comm comm, MPI_Status *status)
{
  static const char* here = "MPI_Recv";
  SSTMAC_DEBUG << here << "-" << (long) buf << ", " << count << ", "
        << datatype << ", " << source << ", " << tag << ", " << comm << "\n";

  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);

  if (source == MPI_PROC_NULL)
  {
    return MPI_SUCCESS;
  }

  const_mpistatus_t stat;
  mpi->recv(count, get_mpi_type(datatype), get_mpi_source(source), mpitag(tag),
      commptr, stat);
  set_mpi_payload(buf, 0, count, datatype, stat->content());

  if (status)
  {
    status->MPI_SOURCE = stat->source().id;
    status->MPI_TAG = stat->tag().id;
    status->MPI_ERROR = 0;
    status->COUNT = stat->get_count();
    status->CANCELLED = false;
  }

  return MPI_SUCCESS;
}

extern "C" int MPI_Irecv(void *buf, int count, MPI_Datatype datatype,
    int source, int tag, MPI_Comm comm, MPI_Request *request)
{
  static const char* here = "MPI_Irecv";
  SSTMAC_DEBUG << here << "\n";

  if (source == MPI_PROC_NULL)
  {
    return MPI_SUCCESS;
  }

  //  if (buf != NULL)
  //   {
  //     throw ssterror(
  //         "MPI_Irecv: payload buffer does not work with Irecv, you must pass NULL and call MPI_Recover_payload(request)");
  //   }
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);

  mpirequest_t req;
  mpi->irecv(count, get_mpi_type(datatype), get_mpi_source(source), mpitag(tag),
      commptr, req);
  *request = add_request_ptr(rank, req);
  buffers[rank][*request] = buf;

  SSTMAC_DEBUG << "MPI_Irecv(" << mpi->comm_rank(commptr).id
        << "): returning request " << *request << " key="
        << (long) req->get_key().get() << "\n";

  return MPI_SUCCESS;
}

extern "C" int MPI_Recover_Payload(void *buf, MPI_Request request)
{
  static const char* here = "MPI_Recover_Payload";
  SSTMAC_DEBUG << here << "(" << (long) buf << ", " << request << ") \n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpirequest_t req = get_request(rank, request);

  if (!req->is_complete())
  {
    return MPI_ERR_PENDING;
  }
  else
  {
    const_mpistatus_t stat = req->get_status();
    if (!stat)
    {
      throw ssterror("MPI_Recover_Payload: stat recovered from request is null");
    }

    if (stat->recv_type().get_packed_size() > 0)
    {
      set_mpi_payload(buf, 0,
          ceil((double)stat->get_count() / (double)stat->recv_type().get_packed_size()),
          stat->recv_type(), stat->content());
    }
    return MPI_SUCCESS;
  }
}

extern "C" int MPI_Allreduce(void *sendbuf, void *recvbuf, int count,
    MPI_Datatype datatype, MPI_Op op, MPI_Comm comm)
{
  static const char* here = "MPI_Allreduce";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);

  sstmac::payload::const_ptr load;
  if (sendbuf == MPI_IN_PLACE)
  {
    load = get_mpi_payload(recvbuf, 0, count, datatype);
  }
  else
  {
    load = get_mpi_payload(sendbuf, 0, count, datatype);
  }
  SSTMAC_DEBUG << "rank " << mpi->comm_world()->rank().id << ": using comm "
        << comm << ", which has size " << commptr->size().id << "\n";
  if (load)
    SSTMAC_DEBUG << "rank " << mpi->comm_world()->rank().id << ": reducing "
          << load->to_string() << "\n";
  sstmac::payload::const_ptr res;
  mpi->allreduce(count, get_mpi_type(datatype), get_mpi_op(op), commptr, load,
      res, true);
  if (res)
    SSTMAC_DEBUG << "rank " << mpi->comm_world()->rank().id
          << ": ended up with " << res->to_string() << "\n";
  set_mpi_payload(recvbuf, 0, count, datatype, res);
  return MPI_SUCCESS;
}

extern "C" int MPI_Reduce(void *sendbuf, void *recvbuf, int count,
    MPI_Datatype datatype, MPI_Op op, int root, MPI_Comm comm)
{
  static const char* here = "MPI_Reduce";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);

  sstmac::payload::const_ptr load;
  if (sendbuf == MPI_IN_PLACE)
  {
    load = get_mpi_payload(recvbuf, 0, count, datatype);
  }
  else
  {
    load = get_mpi_payload(sendbuf, 0, count, datatype);
  }
  sstmac::payload::const_ptr res;
  mpi->reduce(count, get_mpi_type(datatype), get_mpi_op(op), mpiid(root),
      commptr, load, res);
  if (mpi->comm_rank(commptr).id == root)
  {
    set_mpi_payload(recvbuf, 0, count, datatype, res);
  }
  return MPI_SUCCESS;
}

extern "C" int MPI_Barrier(MPI_Comm comm)
{
  static const char* here = "MPI_Barrier";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);
  mpi->barrier(commptr);
  return MPI_SUCCESS;
}

extern "C" int MPI_Bcast(void *buffer, int count, MPI_Datatype datatype,
    int root, MPI_Comm comm)
{
  static const char* here = "MPI_Bcast";

  int rank;
  MPI_Comm_rank(comm, &rank);

  SSTMAC_DEBUG << here << "(rank= " << rank << ", root= " << root << ", comm= "
        << comm << ")\n";
  mpiapi::ptr mpi = get_current_mpi();
  int world_rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(world_rank, comm);

  if (root == MPI_PROC_NULL)
  {
    return MPI_SUCCESS;
  }

  sstmac::payload::const_ptr load;
  if (rank == root)
  {

    load = get_mpi_payload(buffer, 0, count, datatype);
    SSTMAC_DEBUG << "rank " << root << " constructed payload "
          << (load ? load->to_string() : "NULL") << "\n";
  }

  mpi->bcast(count, get_mpi_type(datatype), mpiid(root), commptr, load);

  if (rank != root)
  {
    set_mpi_payload(buffer, 0, count, datatype, load);
  }

  return MPI_SUCCESS;
}

extern "C" int MPI_Scan(void *sendbuf, void *recvbuf, int count,
    MPI_Datatype datatype, MPI_Op op, MPI_Comm comm)
{
  static const char* here = "MPI_Scan";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);
  sstmac::payload::const_ptr load =
      get_mpi_payload(sendbuf, 0, count, datatype);
  sstmac::payload::const_ptr res;
  mpi->scan(count, get_mpi_type(datatype), get_mpi_op(op), commptr, load, res);
  set_mpi_payload(recvbuf, 0, count, datatype, res);
  return MPI_SUCCESS;
}

extern "C" int MPI_Gather(void *sendbuf, int sendcnt, MPI_Datatype sendtype,
    void *recvbuf, int recvcnt, MPI_Datatype recvtype, int root, MPI_Comm comm)
{
  static const char* here = "MPI_Gather";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);
  int commrank = commptr->rank_.id;
  sstmac::payload::const_ptr load;
  MPI_Datatype thesendtype = sendtype;
  if (sendbuf == MPI_IN_PLACE)
  {
    sendcnt = recvcnt;
    sendtype = recvtype;
    // my in place buffer begins at my rank offset
    int offset = commrank * sendcnt;
    load = get_mpi_payload(recvbuf, offset, sendcnt, sendtype);
  }
  else
  {
    int offset = 0;
    load = get_mpi_payload(sendbuf, offset, sendcnt, sendtype);
  }

  mpitype rtype = mpitype::mpi_null;

  if(root == commrank){
    rtype = get_mpi_type(recvtype);
  }

  std::vector<sstmac::payload::const_ptr> cc;
  mpi->gather(sendcnt, get_mpi_type(sendtype), recvcnt, rtype,
      sstmac::sw::mpiid(root), commptr, load, cc);
  if (mpi->comm_rank(commptr).id == root)
  {
    set_mpi_payload(recvbuf, recvcnt, recvtype, cc, here);
  }
  return MPI_SUCCESS;
}

extern "C" int MPI_Gatherv(void *sendbuf, int sendcnt, MPI_Datatype sendtype,
    void *recvbuf, int *recvcnts, int *displs, MPI_Datatype recvtype, int root,
    MPI_Comm comm)
{
  static const char* here = "MPI_Gatherv";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);
  int commrank = commptr->rank_.id;
  mpiid group_size = mpi->comm_size(commptr);
  std::vector<int> cxx_recvcnts(group_size.id, 0);
  if (rank == root)
  {
    for (int i = 0; i < group_size.id; ++i)
      cxx_recvcnts[i] = recvcnts[i];
  }

  payload::const_ptr sendload;
  std::vector<sstmac::payload::const_ptr> recvload;

  if (sendbuf == MPI_IN_PLACE)
  {
    sendload = get_mpi_payload(recvbuf, displs[commptr->rank().id], sendcnt,
        sendtype);
  }
  else
  {
    sendload = get_mpi_payload(sendbuf, 0, sendcnt, sendtype);
  }
  //type<sstmac::payload::const_ptr>::vector recvload(sendload.size);

  mpitype rtype = mpitype::mpi_null;

   if(root == commrank){
     rtype = get_mpi_type(recvtype);
   }

  mpi->gatherv(sendcnt, get_mpi_type(sendtype), cxx_recvcnts,
     rtype, mpiid(root), commptr, sendload, recvload);

  if (commptr->rank().id == root)
    set_displaced_payload(recvbuf, commptr->size().id, recvcnts, displs,
        recvtype, recvload);
  return MPI_SUCCESS; //yes, error handling would be nice
}

extern "C" int MPI_Allgather(void *sendbuf, int sendcount,
    MPI_Datatype sendtype, void *recvbuf, int recvcount, MPI_Datatype recvtype,
    MPI_Comm comm)
{
  static const char* here = "MPI_Allgather";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);
  int commrank = commptr->rank_.id;
  sstmac::payload::const_ptr load;
  mpitype thesendtype;

  if (sendbuf == MPI_IN_PLACE)
  {
    load = get_mpi_payload(recvbuf, commrank * recvcount, recvcount, recvtype);
    thesendtype = get_mpi_type(recvtype);
    sendcount = recvcount;
    sendtype = recvtype;
  }
  else
  {
    load = get_mpi_payload(sendbuf, 0, sendcount, sendtype);
    thesendtype = get_mpi_type(sendtype);
  }
  std::vector<sstmac::payload::const_ptr> cc;
  mpi->allgather(sendcount, thesendtype, recvcount, get_mpi_type(recvtype),
      commptr, load, cc, true);
  set_mpi_payload(recvbuf, recvcount, recvtype, cc, here);
  return MPI_SUCCESS;
}

extern "C" int MPI_Allgatherv(void *sendbuf, int sendcount,
    MPI_Datatype sendtype, void *recvbuf, int *recvcounts, int *displs,
    MPI_Datatype recvtype, MPI_Comm comm)
{
  static const char* here = "MPI_Allgatherv";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);
  int commrank = commptr->rank_.id;

  mpiid group_size = mpi->comm_size(commptr);
  std::vector<int> cxx_recvcnts(group_size.id);
  for (int i = 0; i < group_size.id; ++i)
    cxx_recvcnts[i] = recvcounts[i];

  sstmac::payload::const_ptr sendload;
  if (sendbuf == MPI_IN_PLACE)
  {
    sendload = get_mpi_payload(recvbuf, displs[commrank], cxx_recvcnts[commrank],
        recvtype);
    sendcount = cxx_recvcnts[commrank];
    sendtype = recvtype;
  }
  else
  {
    sendload = get_mpi_payload(sendbuf, 0, sendcount, sendtype);
  }
  std::vector<sstmac::payload::const_ptr> recvload;

  mpi->allgatherv(sendcount, get_mpi_type(sendtype), cxx_recvcnts,
      get_mpi_type(recvtype), commptr, sendload, recvload);

  set_displaced_payload(recvbuf, commptr->size().id, recvcounts, displs,
      recvtype, recvload);
  return MPI_SUCCESS; //yes, error handling would be nice
}

extern "C" int MPI_Scatter(void *sendbuf, int sendcnt, MPI_Datatype sendtype,
    void *recvbuf, int recvcnt, MPI_Datatype recvtype, int root, MPI_Comm comm)
{
  static const char* here = "MPI_Scatter";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);
  sstmac::payload::const_ptr load;
  std::vector<sstmac::payload::const_ptr> cc(commptr->size().id);
  bool nonnull_payload = true;

  mpitype stype, rtype;
  int scnt, rcnt;

  if (commptr->rank().id == root)
  {
    if (recvbuf == MPI_IN_PLACE)
    {
      rtype = mpitype::mpi_null;
      rcnt = 0;
    }
    else
    {
      rtype = get_mpi_type(recvtype);
      rcnt = recvcnt;
    }
    stype = get_mpi_type(sendtype);
    scnt = sendcnt;

    nonnull_payload = get_mpi_vector_payload(sendbuf, commptr->size().id, scnt,
        sendtype, cc);
  }
  else
  {
    stype = mpitype::mpi_null;
    scnt = 0;
    rtype = get_mpi_type(recvtype);
    rcnt = recvcnt;
  }

  //vector payloads were actually built
  mpi->scatter(scnt, stype, rcnt, rtype, sstmac::sw::mpiid(root), commptr, cc,
      load);

  if (recvbuf != MPI_IN_PLACE)
  {

    set_mpi_payload(recvbuf, 0, recvcnt, recvtype, load);
  }
  return MPI_SUCCESS;
}

extern "C" int MPI_Scatterv(void *sendbuf, int *sendcnts, int *displs,
    MPI_Datatype sendtype, void *recvbuf, int recvcnt, MPI_Datatype recvtype,
    int root, MPI_Comm comm)
{
  static const char* here = "MPI_Scatterv";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);

  mpiid group_size = mpi->comm_size(commptr);
  std::vector<int> cxx_sendcnts(group_size.id);
  for (int i = 0; i < group_size.id; ++i)
    cxx_sendcnts[i] = sendcnts[i];

  sstmac::payload::const_ptr load;
  std::vector<sstmac::payload::const_ptr> cc;
  if (commptr->rank().id == root)
  {
    get_displaced_payload(sendbuf, group_size.id, sendcnts, displs, sendtype,
        cc);
  }

  if (sendbuf != NULL || recvbuf != NULL)
  { //vector payloads were actually built
    mpi->scatterv(cxx_sendcnts, get_mpi_type(sendtype), recvcnt,
        get_mpi_type(recvtype), sstmac::sw::mpiid(root), commptr, cc, load);
  }
  else
  { //ignore the vector payloads
    mpi->scatterv(cxx_sendcnts, get_mpi_type(sendtype), recvcnt,
        get_mpi_type(recvtype), sstmac::sw::mpiid(root), commptr);
  }
  set_mpi_payload(recvbuf, 0, recvcnt, recvtype, load);
  return MPI_SUCCESS;
}

extern "C" int MPI_Alltoall(void *sendbuf, int sendcount,
    MPI_Datatype sendtype, void *recvbuf, int recvcount, MPI_Datatype recvtype,
    MPI_Comm comm)
{
  static const char* here = "MPI_Alltoall";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);
  std::vector<sstmac::payload::const_ptr> load;
  std::vector<sstmac::payload::const_ptr> cc;
  bool nonnull_payload = get_mpi_vector_payload(sendbuf, commptr->size().id,
      sendcount, sendtype, cc);
  mpi->alltoall(sendcount, get_mpi_type(sendtype), recvcount,
      get_mpi_type(recvtype), commptr, cc, load);
  set_mpi_vector_payload(recvbuf, commptr->size().id, recvcount, recvtype, load);
  return MPI_SUCCESS;
}

extern "C" int MPI_Alltoallv(void *sendbuf, int *sendcnts, int *sdispls,
    MPI_Datatype sendtype, void *recvbuf, int *recvcnts, int *rdispls,
    MPI_Datatype recvtype, MPI_Comm comm)
{
  static const char* here = "MPI_Alltoallv";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);

  mpiid group_size = mpi->comm_size(commptr);
  std::vector<int> cxx_sendcnts(group_size.id);
  std::vector<int> cxx_recvcnts(group_size.id);
  for (int i = 0; i < group_size.id; ++i)
  {
    cxx_sendcnts[i] = sendcnts[i];
    cxx_recvcnts[i] = recvcnts[i];
  }

  std::vector<payload::const_ptr> sendload;
  get_displaced_payload(sendbuf, commptr->size().id, sendcnts, sdispls,
      sendtype, sendload);
  std::vector<sstmac::payload::const_ptr> recvload;

  mpi->alltoallv(cxx_sendcnts, get_mpi_type(sendtype), cxx_recvcnts,
      get_mpi_type(recvtype), commptr, sendload, recvload);

  set_displaced_payload(recvbuf, commptr->size().id, recvcnts, rdispls,
      recvtype, recvload);
  return MPI_SUCCESS; //yes, error handling would be nice
}

extern "C" double MPI_Wtime()
{
  mpiapi::ptr app = get_current_mpi();

  return app->wtime().sec(); //yes, error handling would be nice
}

extern "C" double MPI_Wtick()
{
  return timestamp::tick_interval_sec();
}

//----------------------------------------------------------------
// --- MPI Cart functions
//----------------------------------------------------------------
extern "C" int MPI_Cart_create(MPI_Comm comm, int ndims, int *dims,
    int *periods, int reorder, MPI_Comm *outcomm)
{
  static const char* here = "MPI_Cart_create";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);
  mpicomm::ptr newc;
  mpi->mpi_cart_create(commptr, ndims, dims, periods, reorder, newc);
  *outcomm = add_comm_ptr(rank, newc);
  return MPI_SUCCESS;
}

extern "C" int MPI_Cart_get(MPI_Comm comm, int maxdims, int *dims,
    int *periods, int *coords)
{
  static const char* here = "MPI_Cart_get";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);
  mpi->mpi_cart_get(commptr, maxdims, dims, periods, coords);
  return MPI_SUCCESS;
}

extern "C" int MPI_Cart_rank(MPI_Comm comm, int *coords, int *rankptr)
{
  static const char* here = "MPI_Cart_rank";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);
  mpi->mpi_cart_rank(commptr, coords, rankptr);
  return MPI_SUCCESS;
}

extern "C" int MPI_Cart_shift(MPI_Comm comm, int direction, int displ,
    int *source, int *dest)
{
  static const char* here = "MPI_Cart_shift";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);
  mpi->mpi_cart_shift(commptr, direction, displ, source, dest);
  if (*source == mpicomm::proc_null)
  {
    *source = MPI_PROC_NULL;
  }
  if (*dest == mpicomm::proc_null)
  {
    *dest = MPI_PROC_NULL;
  }
  return MPI_SUCCESS;
}

extern "C" int MPI_Cart_coords(MPI_Comm comm, int rank, int maxdims,
    int *coords)
{
  static const char* here = "MPI_Cart_coords";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int world_rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(world_rank, comm);
  mpi->mpi_cart_coords(commptr, rank, maxdims, coords);
  return MPI_SUCCESS;
}

extern "C" int MPI_Address(void *location, MPI_Aint *address)
{
  address = (MPI_Aint*) location;
  return MPI_SUCCESS; //yes, error handling would be nice
}

extern "C" int MPI_Iprobe(int source, int tag, MPI_Comm comm, int *flag,
    MPI_Status *status)
{
  static const char* here = "MPI_Iprobe";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);
  const_mpistatus_t stat;
  mpiid src = get_mpi_source(source);
  mpitag t = get_mpi_tag(tag);


  bool fl;
  mpi->iprobe(src,t, commptr, fl, stat);
  *flag = (int) fl;
  if (fl)
  {
    status->MPI_SOURCE = stat->source().id;
    status->MPI_TAG = stat->tag().id;
    status->COUNT = stat->get_count();
    status->CANCELLED = false;
  }
  status->MPI_ERROR = 0;
  return MPI_SUCCESS; //yes, error handling would be nice
}

typedef timestamp
(mpiapi::*mpiapi_send_init_fxn)(int count, const mpitype &type,
    const mpiid &target, const mpitag &tag, const mpicomm::ptr &comm,
    mpirequest_t &req, void* buf);

static int
mpi_init_send(void *buf, int count, MPI_Datatype datatype, int dest, int tag,
    MPI_Comm comm, MPI_Request *request, mpiapi_send_init_fxn fxn)
{
  static const char* here = "MPI_Send_init";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);
  mpirequest_t req;
  //sstmac::payload::const_ptr load = get_mpi_payload(buf, 0, count, datatype);

  (mpi.get()->*fxn)(count, get_mpi_type(datatype), mpiid(dest), mpitag(tag),
      commptr, req, buf);

  *request = add_request_ptr(rank, req);

  return MPI_SUCCESS;
}

extern "C" int MPI_Send_init(void *buf, int count, MPI_Datatype datatype,
    int dest, int tag, MPI_Comm comm, MPI_Request *request)
{
  return mpi_init_send(buf, count, datatype, dest, tag, comm, request,
      &mpiapi::send_init);
}

extern "C" int MPI_Bsend_init(void *buf, int count, MPI_Datatype datatype,
    int dest, int tag, MPI_Comm comm, MPI_Request *request)
{
  return mpi_init_send(buf, count, datatype, dest, tag, comm, request,
      &mpiapi::bsend_init);
}

extern "C" int MPI_Rsend_init(void *buf, int count, MPI_Datatype datatype,
    int dest, int tag, MPI_Comm comm, MPI_Request *request)
{
  return mpi_init_send(buf, count, datatype, dest, tag, comm, request,
      &mpiapi::rsend_init);
}

extern "C" int MPI_Ssend_init(void *buf, int count, MPI_Datatype datatype,
    int dest, int tag, MPI_Comm comm, MPI_Request *request)
{
  return mpi_init_send(buf, count, datatype, dest, tag, comm, request,
      &mpiapi::ssend_init);
}

extern "C" int MPI_Recv_init(void *buf, int count, MPI_Datatype datatype,
    int source, int tag, MPI_Comm comm, MPI_Request *request)
{
  static const char* here = "MPI_Recv_init";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);
  mpirequest_t req;

  if (source == MPI_PROC_NULL)
  {
    return MPI_SUCCESS;
  }

  mpi->recv_init(count, get_mpi_type(datatype), get_mpi_source(source),
      mpitag(tag), commptr, req);
  *request = add_request_ptr(rank, req);
  buffers[rank][*request] = buf;

  return MPI_SUCCESS;
}

extern "C" int MPI_Start(MPI_Request *request)
{
  static const char* here = "MPI_Start";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;

  mpirequest_t req = get_request(rank, *request);

  mpi->start(req);

  return MPI_SUCCESS;
}

extern "C" int MPI_Startall(int count, MPI_Request array_of_requests[])
{
  static const char* here = "MPI_Startall";
  SSTMAC_DEBUG << here << "\n";
  std::vector<mpirequest_t> reqs;
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  for (int i = 0; i < count; i++)
  {
    MPI_Request r = array_of_requests[i];
    mpirequest_t req = get_request(rank, r);
    reqs.push_back(req);
  }

  mpi->startall(reqs);

  return MPI_SUCCESS;
}

extern "C" int MPI_Testall(int count, MPI_Request array_of_requests[],
    int *flag, MPI_Status array_of_statuses[])
{
  static const char* here = "MPI_Testall";
  SSTMAC_DEBUG << here << "\n";

  //we are actually going to avoid testall for code reuse
  int tmp_flag;
  int cumul_flag = 1; //set to true for now
  for (int i = 0; i < count; ++i)
  {
    MPI_Status* status =
        array_of_statuses == MPI_STATUS_IGNORE ? MPI_STATUS_IGNORE
            : &array_of_statuses[i];
    MPI_Test(&array_of_requests[i], &tmp_flag, status);
    cumul_flag = cumul_flag && tmp_flag; //only return true if ALL tests complete
  }
  *flag = cumul_flag;
  return MPI_SUCCESS;
}

//----------------------------------------------------------------
// --- MPI Derived Datatype
// --- Presently, payloads won't work with derived datatypes
//----------------------------------------------------------------

extern "C" int MPI_Type_contiguous(int count, MPI_Datatype old_type,
    MPI_Datatype *new_type_p)
{
  mpiapi::ptr mpi = get_current_mpi();
  mpitype cxx_new;
  mpi->type_contiguous(count, get_mpi_type(old_type), cxx_new);
  *new_type_p = cxx_new.id;
  return MPI_SUCCESS; //yes, error handling would be nice
}

extern "C" int MPI_Type_vector(int count, int blocklength, int stride,
    MPI_Datatype old_type, MPI_Datatype *newtype_p)
{
  mpiapi::ptr mpi = get_current_mpi();
  mpitype cxx_new;
  mpi->type_vector(count, blocklength, stride, get_mpi_type(old_type), cxx_new,
      true);
  *newtype_p = cxx_new.id;
  return MPI_SUCCESS; //yes, error handling would be nice
}

extern "C" int MPI_Type_hvector(int count, int blocklength, MPI_Aint stride,
    MPI_Datatype old_type, MPI_Datatype *newtype_p)
{
  mpiapi::ptr mpi = get_current_mpi();
  mpitype cxx_new;
  mpi->type_vector(count, blocklength, (int) stride, get_mpi_type(old_type),
      cxx_new, false);
  *newtype_p = cxx_new.id;
  return MPI_SUCCESS; //yes, error handling would be nice
}

extern "C" int MPI_Type_struct(int count, int blocklens[], MPI_Aint indices[],
    MPI_Datatype old_types[], MPI_Datatype *newtype)
{
  mpiapi::ptr mpi = get_current_mpi();
  std::vector<int> cxx_blocklens(count);
  std::vector<mpitype> cxx_old_types(count);
  std::vector<int> cxx_ind(count);
  for (int i = 0; i < count; ++i)
  {
    cxx_blocklens[i] = blocklens[i];
    cxx_ind[i] = indices[i];
    cxx_old_types[i] = get_mpi_type(old_types[i]);
  }
  mpitype cxx_new;
  mpi->type_struct(count, cxx_blocklens, cxx_ind, cxx_old_types, cxx_new);
  *newtype = cxx_new.id;
  return MPI_SUCCESS; //yes, error handling would be nice
}

extern "C" int MPI_Type_commit(MPI_Datatype *datatype)
{
  mpiapi::ptr mpi = get_current_mpi();
  mpi->type_commit(get_mpi_type(*datatype));
  return MPI_SUCCESS; //yes, error handling would be nice
}

extern "C" int MPI_Type_free(MPI_Datatype *datatype)
{
  mpiapi::ptr mpi = get_current_mpi();
  mpi->type_free(get_mpi_type(*datatype));
  return MPI_SUCCESS; //yes, error handling would be nice
}

//----------------------------------------------------------------
// --- Misc functions added to support parmetis and libtopomap
//----------------------------------------------------------------

extern "C" MPI_Comm
MPI_Comm_f2c(MPI_Fint comm)
{
  return comm;
}
extern "C" MPI_Fint
MPI_Comm_c2f(MPI_Comm comm)
{
  return comm;
}

extern "C" int
MPI_Keyval_create(MPI_Copy_function *copy_fn, MPI_Delete_function *delete_fn,
    int *key, void *extra_state)
{
  static const char* here = "MPI_Keyval_create";
  SSTMAC_DEBUG << here << "\n";
  keyval::ptr k = keyval::construct(keyval::get_new_key(), copy_fn, delete_fn,
      extra_state);

  keyvals[k->key()] = k;
  *key = k->key();

  return MPI_SUCCESS;
}

extern "C" int
MPI_Attr_put(MPI_Comm comm, int key, void *attr_value)
{
  static const char* here = "MPI_Attr_put";
  SSTMAC_DEBUG << here << "\n";
  check_key(key, here);
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);

  mpi->attr_put(commptr, keyvals[key], attr_value);

  return MPI_SUCCESS;
}

extern "C" int
MPI_Attr_get(MPI_Comm comm, int keyval, void *attr_value, int *flag)
{
  static const char* here = "MPI_Attr_get";
  SSTMAC_DEBUG << here << "\n";
  check_key(keyval, here);
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);

  mpi->attr_get(commptr, keyvals[keyval], attr_value, flag);

  return (*flag == 0) ? MPI_SUCCESS : MPI_ERR_UNKNOWN;
}

template<typename T>
  void
  move_some_stuff(T* dest, T* from, size_t fromoffset, size_t num)
  {
    memmove(dest, &from[fromoffset], num);
  }

/**
 * GH: This function does not call the Helgi MPI, it uses other functions in this file.
 * It is not optimized, and is definitely not as efficient as the real thing.
 * I just wanted to get it working.
 */
extern "C" int
MPI_Reduce_scatter(void *sendbuf, void *recvbuf, int *recvcnts,
    MPI_Datatype datatype, MPI_Op op, MPI_Comm comm)
{

  static const char* here = "MPI_Reduce_scatter";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int r = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(r, comm);
  int size = commptr->size().id;
  int rank = commptr->rank().id;

  std::vector<int> counts;
  long total = 0;
  long myindex = 0;
  for (int i = 0; i < size; i++)
  {
    counts.push_back(recvcnts[i]);
    if (i == rank)
    {
      myindex = total;
    }
    total += recvcnts[i];
  }

  void* temp = NULL;
  size_t typesize;
  if (sendbuf != NULL)
  {

    switch (datatype)
    {
    case MPI_INT:
      typesize = sizeof(int);
      break;

    case MPI_SHORT:
      typesize = sizeof(short);
      break;

    case MPI_DOUBLE:
      typesize = sizeof(double);
      break;

    case MPI_LONG:
      typesize = sizeof(long);
      break;

    default:
      sst_throw(ssterror, "MPI_Reduce_scatter: error creating temp buffer");
    }

    temp = malloc(typesize * total);
  }

  MPI_Allreduce(sendbuf, temp, total, datatype, op, comm);

  if (recvbuf != NULL)
  {
    //for(int i = 0; i < recvcnts[rank]; i++){
    // recvbuf[i] = temp[myindex + i];
    switch (datatype)
    {
    case MPI_INT:
      move_some_stuff<int> ((int*) recvbuf, (int*) temp, myindex,
          typesize * recvcnts[rank]);
      break;

    case MPI_SHORT:
      move_some_stuff<short> ((short*) recvbuf, (short*) temp, myindex,
          typesize * recvcnts[rank]);
      break;

    case MPI_DOUBLE:
      move_some_stuff<double> ((double*) recvbuf, (double*) temp, myindex,
          typesize * recvcnts[rank]);
      break;

    case MPI_LONG:
      move_some_stuff<long> ((long*) recvbuf, (long*) temp, myindex,
          typesize * recvcnts[rank]);
      break;
      // }
    }
  }

  if (temp != NULL)
  {
    free(temp);
  }

  /*  std::vector<int> counts;
   int total = 0;
   for (int i = 0; i < size; i++)
   {
   counts.push_back(recvcnts[i]);
   total += recvcnts[i];
   }

   sstmac::payload::const_ptr load =
   get_mpi_payload(sendbuf, 0, total, datatype);

   sstmac::payload::const_ptr res;
   mpi->reduce_scatter(counts, get_mpi_type(datatype), get_mpi_op(op), commptr,
   load, res);
   set_mpi_payload(recvbuf, recvcnts[commptr->rank().id], 0, datatype, res);*/

  return MPI_SUCCESS;
}

extern "C" int
MPI_Cartdim_get(MPI_Comm comm, int *ndims)
{
  static const char* here = "MPI_Cartdim_get";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);
  mpi->mpi_cartdim_get(commptr, ndims);
  return MPI_SUCCESS;
}

extern "C" int
MPI_Topo_test(MPI_Comm comm, int *topo_type)
{
  static const char* here = "MPI_Topo_test";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);

  switch (commptr->get_topo_type())
  {
  case mpicomm::TOPO_NONE:
    *topo_type = MPI_UNDEFINED;
    break;

  case mpicomm::TOPO_GRAPH:
    *topo_type = MPI_GRAPH;
    break;

  case mpicomm::TOPO_CART:
    *topo_type = MPI_CART;
    break;

  default:
    sst_throw(ssterror, "MPI_Topo_test: mpicomm is of unknown type");

  }

  return MPI_SUCCESS;
}

extern "C" int MPI_Group_translate_ranks(MPI_Group group1, int n, int *ranks1,
    MPI_Group group2, int *ranks2)
{
  static const char* here = "MPI_Group_translate_ranks";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;

  mpigroup::ptr g1 = get_group(rank, group1);
  mpigroup::ptr g2 = get_group(rank, group2);

  for (int i = 0; i < n; i++)
  {
    ranks2[i] = MPI_UNDEFINED;

    if(ranks1[i] == MPI_PROC_NULL){
       ranks2[i] = ranks1[i];
    }
    else
    {

      for (uint j = 0; j < g2->size(); j++)
      {

        if (g2->at(j) == g1->at(ranks1[i]))
        {
          ranks2[i] = j;
          break;
        }

      }
    }
  }
  return MPI_SUCCESS;
}

extern "C" int
MPI_Graph_neighbors(MPI_Comm comm, int rank, int maxneighbors, int *neighbors)
{
  sst_throw(unimplementederror, "MPI_Graph_neighbors not implemented");
}

extern "C" int
MPI_Graph_neighbors_count(MPI_Comm comm, int rank, int *nneighbors)
{
  sst_throw(unimplementederror, "MPI_Graph_neighbors_count not implemented");
}

/**
 @param group [in] group (handle)
 @param n [in] number of triplets in array ranges (integer)
 @param ranges [in] a one-dimensional array of integer triplets, of the form (first rank, last rank, stride) indicating ranks in group or processes to be included in newgroup.
 @param newgroup [out] new group derived from above, in the order defined by ranges (handle)
 */
extern "C" int MPI_Group_range_incl(MPI_Group group, int n, int ranges[][3],
    MPI_Group *newgroup)
{
  static const char* here = "MPI_Group_range_incl";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpigroup::ptr grp_ptr = get_group(rank, group);
  std::vector<taskid> vec;

  if (n > grp_ptr->size())
  {
    sst_throw(illformederror,
        "MPI_Group_range_incl: invalid group size %d", n);
  }

  for (int i = 0; i < n; i++)
  {
    int first = ranges[i][0];
    int last = ranges[i][1];
    int stride = ranges[i][2];
    if(stride == 0){
      sst_throw(valueerror, "sstmac_mpi::MPI_Group_range_incl - got a zero stride at position %d", i);
    }
    int num = (last - first) / stride;
    SSTMAC_DEBUG << "first: " << first << ", last: " << last << ", stride: " << stride << ", num: " << num << "\n";
    for(int j = 0; j <= num; j++){
      int rank = first + j * stride;
      vec.push_back(grp_ptr->at(rank));
    }
  }

  if (vec.size() > 0)
  {
    SSTMAC_DEBUG << "returning group of size " << vec.size() << " \n";
    tasklist::ptr t = tasklist::construct(vec);
    mpigroup::ptr newgrp = mpigroup::construct(t);
    *newgroup = add_group_ptr(rank, newgrp);
  }
  else
  {
    SSTMAC_DEBUG << "nothing in vector, returning empty group \n";
    *newgroup = MPI_GROUP_EMPTY;
  }
  return MPI_SUCCESS;
}

/**
 @param keyval [in] Frees the integer key value (integer)
 */
extern "C" int MPI_Keyval_free(int *keyval)
{
  sst_throw(unimplementederror, "MPI_Keyval_free");
  return MPI_SUCCESS;
}

/**
 @param result [out] integer which is MPI_IDENT if the contexts and groups are the same, MPI_CONGRUENT if different contexts but identical groups, MPI_SIMILAR if different contexts but similar groups, and MPI_UNEQUAL otherwise
 */
extern "C" int MPI_Comm_compare(MPI_Comm comm1, MPI_Comm comm2, int *result)
{
  sst_throw(unimplementederror, "MPI_Comm_compare");
  return MPI_SUCCESS;
}

/**
 @param datatype [in] datatype (handle)
 @param extent [out] datatype extent (integer)
 */
extern "C" int MPI_Type_extent(MPI_Datatype datatype, MPI_Aint *extent)
{
  static const char* here = "MPI_Type_extent";
  SSTMAC_DEBUG << here << "\n";
  //for now assume no padding and that type=size
  mpitype ty = get_mpi_type(datatype);
  *extent = ty.get_extent();
  return MPI_SUCCESS;
}

/**
 @param type [in] datatype (handle)
 @param newtype [out] copy of type (handle)
 */
extern "C" int MPI_Type_dup(MPI_Datatype datatype, MPI_Datatype *newtype)
{
  static const char* here = "MPI_Type_dup";
  SSTMAC_DEBUG << here << "\n";
  mpitype ty = get_mpi_type(datatype);
  mpitype newty;
  mpiapi::ptr mpi = get_current_mpi();
  mpi->type_dup(ty, newty);
  *newtype = newty.id;
  return MPI_SUCCESS;
}

/**
 @param type [in] datatype whose identifier is to be set (handle) _type
 @param name [in] the character string which is remembered as the name (string)
 */
extern "C" int MPI_Type_set_name(MPI_Datatype type, char *typename_)
{
  static const char* here = "MPI_Type_set_name";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  mpi->type_set_name(type, std::string(typename_));

  return MPI_SUCCESS;
}

/**
 @param count [in] number of blocks -- also number of entries in indices and blocklens
 @param blocklens [in] number of elements in each block (array of nonnegative integers)
 @param indices [in] displacement of each block in multiples of old_type (array of integers) _old
 @param type [in] old datatype (handle)
 @param newtype [out] new datatype (handle)
 */
extern "C" int MPI_Type_indexed(int count, int _blocklens_[], int _indices_[],
    MPI_Datatype old_type, MPI_Datatype *newtype)
{
  static const char* here = "MPI_Type_indexed";
  SSTMAC_DEBUG << here << "\n";
  mpitype ty = get_mpi_type(old_type);
  mpitype newty;
  mpiapi::ptr mpi = get_current_mpi();
  std::vector<int> ind(count);
  for (int i = 0; i < count; i++)
  {
    ind[i] = _indices_[i];
  }
  mpi->type_indexed(count, _blocklens_, ind, ty, newty, true, MPI_COMBINER_INDEXED);
  *newtype = newty.id;
  return MPI_SUCCESS;
}

/**
 @param datatype [in] datatype (handle)
 @param size [out] datatype size (integer)
 */
extern "C" int MPI_Type_size(MPI_Datatype datatype, int *size)
{
  static const char* here = "MPI_Type_size";
  SSTMAC_DEBUG << here << "\n";
  mpitype ty = get_mpi_type(datatype);
  *size = ty.get_packed_size();
  return MPI_SUCCESS;
}

/**
 @param type [in] datatype whose name is to be returned (handle) _type
 @param name [out] the name previously stored on the datatype, or a empty string if no such name exists (string)
 @param resultlen [out] length of returned name (integer)
 */
extern "C" int MPI_Type_get_name(MPI_Datatype datatype, char *type_name,
    int *resultlen)
{
  static const char* here = "MPI_Type_get_name";
  SSTMAC_DEBUG << here << "\n";
  mpitype ty = get_mpi_type(datatype);
  std::string name = ty.label;

  std::copy(name.begin(), name.end(), type_name);
  type_name[name.size()] = '\0'; // don't forget the terminating 0


  *resultlen = name.size();
  return MPI_SUCCESS;
}

/**
 @param comm [in] Local (intra)communicator _local
 @param leader [in] Rank in local_comm of leader (often 0) _peer
 @param comm [in] Local (intra)communicator _local
 @param leader [in] Rank in local_comm of leader (often 0) _peer_comm _
 [in] Communicator used to communicate between a designated process in the other communicator. Significant only at the process in local_comm with rank local_leader. _remote_leader _
 [in] Rank in peer_comm of remote leader (often 0)
 @param tag [in] Message tag to use in constructing intercommunicator; if multiple MPI_Intercomm_creates are being made, they should use different tags (more precisely, ensure that the local and remote leaders are using different tags for each MPI_intercomm_create). _comm
 @param out [out] Created intercommunicator
 */
extern "C" int MPI_Intercomm_create(MPI_Comm local_comm, int local_leader,
    MPI_Comm peer_comm, int remote_leader, int tag, MPI_Comm *newintercomm)
{
  sst_throw(unimplementederror, "MPI_Intercomm_create");
  return MPI_SUCCESS;
}

/**
 @param group [in] group (handle)
 @param n [in] number of elements in array ranks (integer)
 @param ranks [in] array of integer ranks in group not to appear in newgroup
 @param newgroup [out] new group derived from above, preserving the order defined by group (handle)
 */
extern "C" int MPI_Group_excl(MPI_Group group, int n, int *ranks,
    MPI_Group *newgroup)
{
  static const char* here = "MPI_Group_excl";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpigroup::ptr grp_ptr = get_group(rank, group);
  std::vector<taskid> vec;
  std::vector<taskid> retvec;

  if (n > grp_ptr->size())
  {
    sst_throw(illformederror,
        "MPI_Group_excll: invalid rank array size %d (group is of size %d)", n, grp_ptr->size());
  }

  tasklist::const_ptr tl = grp_ptr->list();
  if (!tl)
  { // comm world doesn't have a peer list, have to create it (ug)
    vec.resize(grp_ptr->size());
    for (int i = 0; i < grp_ptr->size(); i++)
    {
      vec[i] = taskid(i);
    }
  }
  else
  {
    vec = tl->get_vector();
  }

  for (int i = 0; i < n; i++)
  {
      vec[ranks[i]] = taskid(-1);
  }

  for (int i = 0; i < vec.size(); i++)
  {
    if (vec[i].id >= 0)
    {
      retvec.push_back(taskid(i));
    }
  }

  if (retvec.size() > 0)
  {
    tasklist::ptr t = tasklist::construct(retvec);
    mpigroup::ptr newgrp = mpigroup::construct(t);
    *newgroup = add_group_ptr(rank, newgrp);
  }
  else
  {
    *newgroup = MPI_GROUP_EMPTY;
  }
  return MPI_SUCCESS;
}

/**
 @param comm [in] communicator (handle)
 @param size [out] number of processes in the remote group of comm (integer)
 */
extern "C" int MPI_Comm_remote_size(MPI_Comm comm, int *size)
{
  sst_throw(unimplementederror, "MPI_Comm_remote_size");
  return MPI_SUCCESS;
}

/**
 @param errorcode [in] Error code returned by an MPI routine
 @param errorclass [out] Error class associated with errorcode
 */
extern "C" int MPI_Error_class(int errorcode, int *errorclass)
{
  sst_throw(unimplementederror, "MPI_Error_class");
  return MPI_SUCCESS;
}

/**
 @param errorcode [in] Error code returned by an MPI routine or an MPI error class
 @param string [out] Text that corresponds to the errorcode
 @param resultlen [out] Length of string
 */
extern "C" int MPI_Error_string(int errorcode, char *string, int *resultlen)
{
  sst_throw(unimplementederror, "MPI_Error_string");
  return MPI_SUCCESS;
}

/**
 @param buffer [in] initial buffer address (choice)
 @param size [in] buffer size, in bytes (integer)
 */
extern "C" int MPI_Buffer_attach(void *buffer, int size)
{
  static const char* here = "MPI_Buffer_attach";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  mpi->buffer_attach(size);
  int rank = mpi->comm_world()->rank().id;
  attached_buffers[rank] = std::make_pair<void*, int>(buffer, size);
  return MPI_SUCCESS;
}

/**
 @param buffer [out] initial buffer address (choice)
 @param size [out] buffer size, in bytes (integer)
 */
extern "C" int MPI_Buffer_detach(void *buffer, int *size)
{
  static const char* here = "MPI_Buffer_detach";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  mpi->buffer_detach(*size);
  int rank = mpi->comm_world()->rank().id;
  void** cast = (void**)buffer;
  *cast = attached_buffers[rank].first;
  *size = attached_buffers[rank].second;
  attached_buffers[rank] = std::make_pair<void*, int>(NULL, 0);
  return MPI_SUCCESS;
}

/**
 @param count [in] list length (integer) _array_of
 @param requests [in] array of requests (array of handles)
 @param index [out] index of operation that completed, or MPI_UNDEFINED if none completed (integer)
 @param flag [out] true if one of the operations is complete (logical)
 @param status [out] status object (Status). May be MPI_STATUS_IGNORE.
 */
extern "C" int MPI_Testany(int count, MPI_Request _array_of_requests_[],
    int *index, int *flag, MPI_Status *status)
{
  static const char* here = "MPI_Testany";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;

  SSTMAC_DEBUG << here << "(" << mpi->comm_rank(mpi->comm_world()).id << ") \n";
  const_mpistatus_t stat;
  std::vector<mpirequest_t> sst_reqs;
  std::vector<MPI_Request*> mpi_reqs;
  for (int i = 0; i < count; i++)
  {
    if (_array_of_requests_[i] != MPI_REQUEST_NULL)
    {

      SSTMAC_DEBUG << "adding request " << _array_of_requests_[i]
            << " to the list \n";
      sst_reqs.push_back(get_request(rank, _array_of_requests_[i]));
      mpi_reqs.push_back(&_array_of_requests_[i]);
    }
  }
  if (mpi_reqs.size() > 0)
  {
    int ind;
    bool f;
    mpi->testany(sst_reqs, ind, f, stat);
    *flag = f;

    if(f){
    for (int i = 0; i < count; i++)
      {
        if (*(mpi_reqs[ind]) == (_array_of_requests_[i]))
        {
          *index = i;
          break;
        }
      }

      complete_wait(rank, sst_reqs[ind], stat, mpi_reqs[ind], status);
    }else
      *index = MPI_UNDEFINED;
  }
  else
  {
    *index = MPI_UNDEFINED;
    *flag = true;
  }

  return MPI_SUCCESS; //yes, error handling would be nice
}

/**
 @param incount [in] length of array_of_requests (integer) _array_of
 @param requests [in] array of requests (array of handles)
 @param outcount [out] number of completed requests (integer) _array_of
 @param indices [out] array of indices of operations that completed (array of integers) _array_of
 @param statuses [out] array of status objects for operations that completed (array of Status). May be MPI_STATUSES_IGNORE.
 */
extern "C" int MPI_Testsome(int incount, MPI_Request _array_of_requests_[],
    int *outcount, int _array_of_indices_[], MPI_Status _array_of_statuses_[])
{
  static const char* here = "MPI_Testsome";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  std::vector<mpirequest_t> sst_reqs;
  for (int i = 0; i < incount; i++)
  {
    if (_array_of_requests_[i] != MPI_REQUEST_NULL)
    {
      sst_reqs.push_back(get_request(rank, _array_of_requests_[i]));
    }
  }

  if (sst_reqs.size() > 0)
  {
    std::vector<int> ind;
    std::vector<const_mpistatus_t> sst_stats;
    mpi->testsome(sst_reqs, ind, sst_stats);

    *outcount = sst_stats.size();
    for (uint i = 0; i < ind.size(); i++)
    {
      int idx = ind[i];
      MPI_Request* mpi_req = &_array_of_requests_[idx];
      MPI_Status* mpi_status =
          _array_of_statuses_ == MPI_STATUSES_IGNORE ? MPI_STATUS_IGNORE
              : &_array_of_statuses_[idx];
      complete_wait(rank, sst_reqs[idx], sst_stats[i], mpi_req, mpi_status);
      _array_of_indices_[i] = idx;
    }
  }
  else
  {
    *outcount = MPI_UNDEFINED;
  }

  return MPI_SUCCESS; //yes, error handling would be nice
}

/**
 @param status [in] status object (Status)
 @param flag [out] true if the request was cancelled, false otherwise (logical)
 */
extern "C" int MPI_Test_cancelled(MPI_Status *status, int *flag)
{
  static const char* here = "MPI_Test_cancelled";
  SSTMAC_DEBUG << here << "\n";
  if (status)
    *flag = status->CANCELLED;
  return MPI_SUCCESS;
}

/**
 @param buf [in/out] initial address of send and receive buffer (choice)
 @param count [in] number of elements in send and receive buffer (integer)
 @param datatype [in] type of elements in send and receive buffer (handle)
 @param dest [in] rank of destination (integer)
 @param sendtag [in] send message tag (integer)
 @param source [in] rank of source (integer)
 @param recvtag [in] receive message tag (integer)
 @param comm [in] communicator (handle)
 @param status [out] status object (Status)
 */
extern "C" int MPI_Sendrecv_replace(void *buf, int count,
    MPI_Datatype datatype, int dest, int sendtag, int source, int recvtag,
    MPI_Comm comm, MPI_Status *status)
{
  sst_throw(unimplementederror, "MPI_Sendrecv_replace");
  return MPI_SUCCESS;
}

/**
 @param count [in] number of blocks -- also number of entries in indices and blocklens
 @param blocklens [in] number of elements in each block (array of nonnegative integers)
 @param indices [in] byte displacement of each block (array of MPI_Aint) _old
 @param type [in] old datatype (handle)
 @param newtype [out] new datatype (handle)
 */
extern "C" int MPI_Type_hindexed(int count, int _blocklens_[],
    MPI_Aint _indices_[], MPI_Datatype old_type, MPI_Datatype *newtype)
{
  static const char* here = "MPI_Type_hindexed";
  SSTMAC_DEBUG << here << "\n";
  mpitype ty = get_mpi_type(old_type);
  mpitype newty;
  mpiapi::ptr mpi = get_current_mpi();
  std::vector<int> ind(count);
  for (int i = 0; i < count; i++)
  {
    ind[i] = _indices_[i];
  }
  mpi->type_indexed(count, _blocklens_, ind, ty, newty, false, MPI_COMBINER_HINDEXED);
  *newtype = newty.id;
  return MPI_SUCCESS;
}

/**
 @param datatype [in] datatype (handle)
 @param displacement [out] displacement of lower bound from origin, in bytes (integer)
 */
extern "C" int MPI_Type_lb(MPI_Datatype datatype, MPI_Aint *displacement)
{
  static const char* here = "MPI_Type_lb";
  SSTMAC_DEBUG << here << "\n";
  mpitype ty = get_mpi_type(datatype);
  *displacement = ty.get_lb();
  return MPI_SUCCESS;
}

/**
 @param datatype [in] datatype (handle)
 @param displacement [out] displacement of upper bound from origin, in bytes (integer)
 */
extern "C" int MPI_Type_ub(MPI_Datatype datatype, MPI_Aint *displacement)
{
  static const char* here = "MPI_Type_ub";
  SSTMAC_DEBUG << here << "\n";
  mpitype ty = get_mpi_type(datatype);
  *displacement = ty.get_ub();
  return MPI_SUCCESS;
}

/**
 @param status [in] status
 @param datatype [in] datatype
 @param elements [out] elements
 */
extern "C" int MPI_Get_elements(MPI_Status *status, MPI_Datatype datatype,
    int *elements)
{
  static const char* here = "MPI_Get_elements";
  SSTMAC_DEBUG << here << "\n";
  // throw sstmac::ssterror("MPI_Get_count is unimplemented");
  mpitype t = get_mpi_type(datatype);
  *elements = t.bytes_to_elements(status->COUNT);
  return MPI_SUCCESS; //yes, error handling would be nice
}

/**
 @param inbuf [in] input buffer
 @param incount [in] input count
 @param datatype [in] datatype
 @param outbuf [in] output buffer
 @param outcount [in] output count
 @param position [in] position
 @param comm [in] communicator
 */
extern "C" int MPI_Pack(void *inbuf, int incount, MPI_Datatype datatype,
    void *outbuf, int outsize, int *position, MPI_Comm comm)
{
  static const char* here = "MPI_Pack";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  mpitype ty = get_mpi_type(datatype);
  ty.pack(inbuf, incount, outbuf, outsize, position);
  return MPI_SUCCESS;
}

/**
 @param inbuf [in] input buffer start (choice)
 @param insize [in] size of input buffer, in bytes (integer)
 @param position [in] current position in bytes (integer)
 @param outbuf [out] output buffer start (choice)
 @param outcount [in] number of items to be unpacked (integer)
 @param datatype [in] datatype of each output data item (handle)
 @param comm [in] communicator for packed message (handle)
 */
extern "C" int MPI_Unpack(void *inbuf, int insize, int *position, void *outbuf,
    int outcount, MPI_Datatype datatype, MPI_Comm comm)
{
  static const char* here = "MPI_Unpack";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  mpitype ty = get_mpi_type(datatype);
  ty.unpack(inbuf, insize, outbuf, outcount, position);
  return MPI_SUCCESS;
}

/**
 @param incount [in] count argument to packing call (integer)
 @param datatype [in] datatype argument to packing call (handle)
 @param comm [in] communicator argument to packing call (handle)
 @param size [out] upper bound on size of packed message, in bytes (integer)
 */
extern "C" int MPI_Pack_size(int incount, MPI_Datatype datatype, MPI_Comm comm,
    int *size)
{
  static const char* here = "MPI_Pack_size";
  SSTMAC_DEBUG << here << "\n";
  mpitype ty = get_mpi_type(datatype);
  *size = incount * ty.get_packed_size();
  return MPI_SUCCESS;
}

/**
 @param function [in] user defined function (function)
 @param commute [in] true if commutative; false otherwise. (logical)
 @param op [out] operation (handle)
 */
extern "C" int MPI_Op_create(MPI_User_function *function, int commute,
    MPI_Op *op)
{
  static const char* here = "MPI_Op_create";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  mpiop otype;
  mpi->op_create(function, commute, otype);
  *op = otype.id;
  return MPI_SUCCESS;
}

/**
 @param op [in] operation (handle)
 */
extern "C" int MPI_Op_free(MPI_Op *op)
{
  static const char* here = "MPI_Op_free";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();

  mpi->op_free(*op);
  *op = MPI_OP_NULL;
  return MPI_SUCCESS;
}

/**
 @param group [in] group (handle) Output Parameter:
 @param size [out] number of processes in the group (integer)
 */
extern "C" int MPI_Group_size(MPI_Group group, int *size)
{
  static const char* here = "MPI_Group_size";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpigroup::ptr grp = get_group(rank, group);
  *size = grp->size();

  return MPI_SUCCESS;
}

/**
 @param group [in] group (handle)
 @param rank [out] rank of the calling process in group, or MPI_UNDEFINED if the process is not a member (integer)
 */
extern "C" int MPI_Group_rank(MPI_Group group, int *rank)
{
  static const char* here = "MPI_Group_size";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int r = mpi->comm_world()->rank().id;
  mpigroup::ptr grp = get_group(r, group);
  *rank = grp->rank_of_task(sstmac::sw::taskid(r)).id;
  if(*rank == -1){
    *rank = MPI_UNDEFINED;
  }

  return MPI_SUCCESS;
}

/**
 @param result [out] integer which is MPI_IDENT if the order and members of the two groups are the same, MPI_SIMILAR if only the members are the same, and MPI_UNEQUAL otherwise
 */
extern "C" int MPI_Group_compare(MPI_Group group1, MPI_Group group2,
    int *result)
{
  static const char* here = "MPI_Group_compare";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpigroup::ptr grp_ptr1 = get_group(rank, group1);
  mpigroup::ptr grp_ptr2 = get_group(rank, group2);
  boost::unordered_map<int, int> check;

  if (group1 == group2)
  { //that was easy
    *result = MPI_IDENT;
    return MPI_SUCCESS;
  }

  if(grp_ptr1->size() != grp_ptr2->size()){
    *result = MPI_UNEQUAL;
    return MPI_SUCCESS;
  }

  for (int i = 0; i < grp_ptr1->size(); i++)
  {
    check[grp_ptr1->at(i).id] = i;
  }

  *result = MPI_IDENT;
  for (int i = 0; i < grp_ptr2->size(); i++)
  {
    if (check.find(grp_ptr2->at(i).id) == check.end())
    {
      *result = MPI_UNEQUAL;
      return MPI_SUCCESS;
    }
    else if (check[grp_ptr2->at(i).id] != i)
    {
      *result = MPI_SIMILAR;
    }

  }

  return MPI_SUCCESS;
}

/**
 @param newgroup [out] union group (handle)
 */
extern "C" int MPI_Group_union(MPI_Group group1, MPI_Group group2,
    MPI_Group *newgroup)
{
  static const char* here = "MPI_Group_union";
   SSTMAC_DEBUG << here << "\n";
   mpiapi::ptr mpi = get_current_mpi();
   int rank = mpi->comm_world()->rank().id;
   mpigroup::ptr grp_ptr1 = get_group(rank, group1);
   mpigroup::ptr grp_ptr2 = get_group(rank, group2);
   std::vector<taskid> vec;
   std::map<int, int> check;


   for(int i = 0; i < grp_ptr1->size(); i++){
     check[grp_ptr1->at(i).id] = grp_ptr1->at(i).id;
   }

  for (int i = 0; i < grp_ptr2->size(); i++)
  {
    check[grp_ptr2->at(i).id] = grp_ptr2->at(i).id;
  }

  std::map<int, int>::iterator it, end = check.end();
   for (it = check.begin(); it != end; it++)
   {
     vec.push_back(taskid(it->second));
   }

   if (vec.size() > 0)
   {
     tasklist::ptr t = tasklist::construct(vec);
     mpigroup::ptr newgrp = mpigroup::construct(t);
     *newgroup = add_group_ptr(rank, newgrp);
   }
   else
   {
     *newgroup = MPI_GROUP_EMPTY;
   }
   return MPI_SUCCESS;
}

/**
 @param newgroup [out] intersection group (handle)
 */
extern "C" int MPI_Group_intersection(MPI_Group group1, MPI_Group group2,
    MPI_Group *newgroup)
{
  static const char* here = "MPI_Group_intersection";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpigroup::ptr grp_ptr1 = get_group(rank, group1);
  mpigroup::ptr grp_ptr2 = get_group(rank, group2);
  std::vector<taskid> vec;
  boost::unordered_map<int, int> check;

  for (int i = 0; i < grp_ptr1->size(); i++)
  {
    check[grp_ptr1->at(i).id] = grp_ptr1->at(i).id;
  }

  for (int i = 0; i < grp_ptr2->size(); i++)
  {
    if (check.find(grp_ptr2->at(i).id) != check.end())
    {
      vec.push_back(grp_ptr2->at(i));
    }

  }

  if (vec.size() > 0)
  {
    tasklist::ptr t = tasklist::construct(vec);
    mpigroup::ptr newgrp = mpigroup::construct(t);
    *newgroup = add_group_ptr(rank, newgrp);
  }
  else
  {
    *newgroup = MPI_GROUP_EMPTY;
  }
  return MPI_SUCCESS;
}

/**
 @param newgroup [out] difference group (handle)
 */
extern "C" int MPI_Group_difference(MPI_Group group1, MPI_Group group2,
    MPI_Group *newgroup)
{
  sst_throw(unimplementederror, "MPI_Group_difference");
  return MPI_SUCCESS;
}

/**
 @param group [in] group (handle)
 @param n [in] number of elements in array ranks (integer)
 @param ranges [in] a one-dimensional array of integer triplets of the form (first rank, last rank, stride), indicating the ranks in group of processes to be excluded from the output group newgroup .
 @param newgroup [out] new group derived from above, preserving the order in group (handle)
 */
extern "C" int MPI_Group_range_excl(MPI_Group group, int n, int ranges[][3],
    MPI_Group *newgroup)
{
  static const char* here = "MPI_Group_range_excl";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpigroup::ptr grp_ptr = get_group(rank, group);
  std::vector<taskid> vec;
  std::vector<taskid> retvec;

  tasklist::const_ptr tl = grp_ptr->list();
  if (!tl)
  { // comm world doesn't have a peer list, have to create it (ug)
    vec.resize(grp_ptr->size());
    for(int i = 0; i < grp_ptr->size(); i++){
      vec[i] = taskid(i);
    }
  }
  else
  {
    vec = tl->get_vector();
  }

  for(int i = 0; i < n; i++){
    int first = ranges[i][0];
    int last = ranges[i][1];
    int stride = ranges[i][2];

    if(stride == 0){
      sst_throw(valueerror, "sstmac_mpi::MPI_Group_range_excl - found 0 stride at position %d", i);
    }
    int num = (last - first) / stride;
    for(int j  = 0; j <= num; j++ ){
      int rank = first + j * stride;
      vec[rank] = taskid(-1);
    }
  }


  for(int i = 0; i < vec.size(); i++){
    if(vec[i].id >= 0){
      retvec.push_back(taskid(i));
    }
  }

  if (retvec.size() > 0)
  {
    tasklist::ptr t = tasklist::construct(retvec);
    mpigroup::ptr newgrp = mpigroup::construct(t);
    *newgroup = add_group_ptr(rank, newgrp);
  }
  else
  {
    *newgroup = MPI_GROUP_EMPTY;
  }
  return MPI_SUCCESS;
}

/**
 @param comm [in] communicator to test (handle)
 @param flag [out] true if this is an inter-communicator(logical)
 */
extern "C" int MPI_Comm_test_inter(MPI_Comm comm, int *flag)
{
  sst_throw(unimplementederror, "MPI_Comm_test_inter");
  return MPI_SUCCESS;
}

/**
 @param comm [in] Communicator (must be an intercommunicator) (handle)
 @param group [out] remote group of communicator (handle)
 */
extern "C" int MPI_Comm_remote_group(MPI_Comm comm, MPI_Group *group)
{
  sst_throw(unimplementederror, "MPI_Comm_remote_group");
  return MPI_SUCCESS;
}

/**
 @param comm [in] Intercommunicator (handle)
 @param high [in] Used to order the groups within comm (logical) when creating the new communicator. This is a boolean value; the group that sets high true has its processes ordered _after_ the group that sets this value to false. If all processes in the intercommunicator provide the same value, the choice of which group is ordered first is arbitrary. _comm
 @param out [out] Created intracommunicator (handle)
 */
extern "C" int MPI_Intercomm_merge(MPI_Comm intercomm, int high,
    MPI_Comm *newintracomm)
{
  sst_throw(unimplementederror, "MPI_Intercomm_merge");
  return MPI_SUCCESS;
}

/**
 @param comm [in] communicator to which attribute is attached (handle)
 @param keyval [in] The key value of the deleted attribute (integer)
 */
extern "C" int MPI_Attr_delete(MPI_Comm comm, int keyval)
{
  sst_throw(unimplementederror, "MPI_Attr_delete");
  return MPI_SUCCESS;
}

/**
 @param nnodes [in] number of nodes in a grid (integer)
 @param ndims [in] number of cartesian dimensions (integer)
 @param dims [in/out] integer array of size ndims specifying the number of nodes in each dimension. A value of 0 indicates that [MPI_Dims_create](MPI_Dims_create.html) should fill in a suitable value.
 */
extern "C" int MPI_Dims_create(int nnodes, int ndims, int *dims)
{
  sst_throw(unimplementederror, "MPI_Dims_create");
  return MPI_SUCCESS;
}

/**
 @param old [in] input communicator without topology (handle)
 @param nnodes [in] number of nodes in graph (integer)
 @param index [in] array of integers describing node degrees (see below)
 @param edges [in] array of integers describing graph edges (see below)
 @param reorder [in] ranking may be reordered (true) or not (false) (logical) _comm
 @param graph [out] communicator with graph topology added (handle)
 */
extern "C" int MPI_Graph_create(MPI_Comm comm_old, int nnodes, int *index,
    int *edges, int reorder, MPI_Comm *comm_graph)
{
  sst_throw(unimplementederror, "MPI_Graph_create");
  return MPI_SUCCESS;
}

/**
 @param comm [in] communicator for group with graph structure (handle)
 @param nnodes [out] number of nodes in graph (integer)
 @param nedges [out] number of edges in graph (integer)
 */
extern "C" int MPI_Graphdims_get(MPI_Comm comm, int *nnodes, int *nedges)
{
  sst_throw(unimplementederror, "MPI_Graphdims_get");
  return MPI_SUCCESS;
}

/**
 @param comm [in] communicator with graph structure (handle)
 @param maxindex [in] length of vector index in the calling program (integer)
 @param maxedges [in] length of vector edges in the calling program (integer)
 @param index [out] array of integers containing the graph structure (for details see the definition of MPI_GRAPH_CREATE)
 @param edges [out] array of integers containing the graph structure
 */
extern "C" int MPI_Graph_get(MPI_Comm comm, int maxindex, int maxedges,
    int *index, int *edges)
{
  sst_throw(unimplementederror, "MPI_Graph_get");
  return MPI_SUCCESS;
}

/**
 @param comm [in] communicator with cartesian structure (handle) _remain
 @param dims [in] the ith entry of remain_dims specifies whether the ith dimension is kept in the subgrid (true) or is dropped (false) (logical vector)
 @param newcomm [out] communicator containing the subgrid that includes the calling process (handle)
 */
extern "C" int MPI_Cart_sub(MPI_Comm comm, int *remain_dims, MPI_Comm *newcomm)
{
  sst_throw(unimplementederror, "MPI_Cart_sub");
  return MPI_SUCCESS;
}

/**
 @param comm [in] input communicator (handle)
 @param ndims [in] number of dimensions of Cartesian structure (integer)
 @param dims [in] integer array of size ndims specifying the number of processes in each coordinate direction
 @param periods [in] logical array of size ndims specifying the periodicity specification in each coordinate direction
 @param newrank [out] reordered rank of the calling process; MPI_UNDEFINED if calling process does not belong to grid (integer)
 */
extern "C" int MPI_Cart_map(MPI_Comm comm_old, int ndims, int *dims,
    int *periods, int *newrank)
{
  sst_throw(unimplementederror, "MPI_Cart_map");
  return MPI_SUCCESS;
}

/**
 @param comm [in] input communicator (handle)
 @param nnodes [in] number of graph nodes (integer)
 @param index [in] integer array specifying the graph structure, see MPI_GRAPH
 @param CREATE edges _
 [in] integer array specifying the graph structure
 @param newrank [out] reordered rank of the calling process; MPI_UNDEFINED if the calling process does not belong to graph (integer)
 */
extern "C" int MPI_Graph_map(MPI_Comm comm_old, int nnodes, int *index,
    int *edges, int *newrank)
{
  sst_throw(unimplementederror, "MPI_Graph_map");
  return MPI_SUCCESS;
}

/**
 @param name [out] A unique specifier for the actual (as opposed to virtual) node. This must be an array of size at least MPI_MAX_PROCESSOR_NAME.
 @param resultlen [out] Length (in characters) of the name
 */
extern "C" int MPI_Get_processor_name(char *name, int *resultlen)
{
  sst_throw(unimplementederror, "MPI_Get_processor_name");
  return MPI_SUCCESS;
}

/**
 @param version [out] Version of MPI
 @param subversion [out] Subversion of MPI
 */
extern "C" int MPI_Get_version(int *version, int *subversion)
{
  sst_throw(unimplementederror, "MPI_Get_version");
  return MPI_SUCCESS;
}

/**
 @param function [in] user defined error handling procedure
 @param errhandler [out] MPI error handler (handle)
 */
extern "C" int MPI_Errhandler_create(MPI_Handler_function *function,
    MPI_Errhandler *errhandler)
{
  sst_throw(unimplementederror, "MPI_Errhandler_create");
  return MPI_SUCCESS;
}

/**
 @param comm [in] communicator to get the error handler from (handle)
 @param errhandler [in] MPI error handler currently associated with communicator (handle)
 */
extern "C" int MPI_Errhandler_set(MPI_Comm comm, MPI_Errhandler errhandler)
{
  err_handlers[comm] = errhandler;
  return MPI_SUCCESS;
}

/**
 @param comm [in] communicator to get the error handler from (handle)
 @param errhandler [out] MPI error handler currently associated with communicator (handle)
 */
extern "C" int MPI_Errhandler_get(MPI_Comm comm, MPI_Errhandler *errhandler)
{
  err_handler_map::iterator it = err_handlers.find(comm);
  if (it == err_handlers.end())
  {
    sst_throw(sstmac::valueerror, "Invalid communicator %d in MPI_Errhandler_get", comm);
  }
  *errhandler = it->second;
  return MPI_SUCCESS;
}

/**
 @param errhandler [in/out] MPI error handler (handle). Set to MPI_ERRHANDLER_NULL on exit.
 */
extern "C" int MPI_Errhandler_free(MPI_Errhandler *errhandler)
{
  sst_throw(unimplementederror, "MPI_Errhandler_free");
  return MPI_SUCCESS;
}

/**
 @param level [in] Profiling level _... _
 other arguments (see notes)
 */
extern "C" int MPI_Pcontrol(const int level, ...)
{
  sst_throw(unimplementederror, "MPI_Pcontrol");
  return MPI_SUCCESS;
}

/**
 @param name [in] a port name (string)
 */
extern "C" int MPI_Close_port(char *port_name)
{
  sst_throw(unimplementederror, "MPI_Close_port");
  return MPI_SUCCESS;
}

/**
 @param name [in] port name (string, used only on root)
 @param info [in] implementation-dependent information (handle, used only on root)
 @param root [in] rank in comm of root node (integer)
 @param comm [in] comm intracommunicator over which call is collective (handle)
 @param newcomm [out] intercommunicator with client as remote group (handle)
 */
extern "C" int MPI_Comm_accept(char *port_name, MPI_Info info, int root,
    MPI_Comm comm, MPI_Comm *newcomm)
{
  sst_throw(unimplementederror, "MPI_Comm_accept");
  return MPI_SUCCESS;
}

/**
 @param name [in] network address (string, used only on root)
 @param info [in] implementation-dependent information (handle, used only on root)
 @param root [in] rank in comm of root node (integer)
 @param comm [in] intracommunicator over which call is collective (handle)
 @param newcomm [out] intercommunicator with server as remote group (handle)
 */
extern "C" int MPI_Comm_connect(char *port_name, MPI_Info info, int root,
    MPI_Comm comm, MPI_Comm *newcomm)
{
  sst_throw(unimplementederror, "MPI_Comm_connect");
  return MPI_SUCCESS;
}

/**
 @param comm [in] communicator (handle)
 */
extern "C" int MPI_Comm_disconnect(MPI_Comm * comm)
{
  sst_throw(unimplementederror, "MPI_Comm_disconnect");
  return MPI_SUCCESS;
}

/**
 @param parent [out] the parent communicator (handle)
 */
extern "C" int MPI_Comm_get_parent(MPI_Comm *parent)
{
  sst_throw(unimplementederror, "MPI_Comm_get_parent");
  return MPI_SUCCESS;
}

/**
 @param fd [in] socket file descriptor
 @param intercomm [out] new intercommunicator (handle)
 */
extern "C" int MPI_Comm_join(int fd, MPI_Comm *intercomm)
{
  sst_throw(unimplementederror, "MPI_Comm_join");
  return MPI_SUCCESS;
}

/**
 @param command [in] name of program to be spawned (string, significant only at root)
 @param argv [in] arguments to command (array of strings, significant only at root)
 @param maxprocs [in] maximum number of processes to start (integer, significant only at root)
 @param info [in] a set of key-value pairs telling the runtime system where and how to start the processes (handle, significant only at root)
 @param root [in] rank of process in which previous arguments are examined (integer)
 @param comm [in] intracommunicator containing group of spawning processes (handle)
 @param intercomm [out] intercommunicator between original group and the newly spawned group (handle) _array_of
 @param errcodes [out] one code per process (array of integer)
 */
extern "C" int MPI_Comm_spawn(char *command, char *_argv_[], int maxprocs,
    MPI_Info info, int root, MPI_Comm comm, MPI_Comm *intercomm,
    int _array_of_errcodes_[])
{
  sst_throw(unimplementederror, "MPI_Comm_spawn");
  return MPI_SUCCESS;
}

/**
 @param count [in] number of commands (positive integer, significant to MPI only at root _array_of
 @param commands [in] programs to be executed (array of strings, significant only at root) _array_of
 @param argv [in] arguments for commands (array of array of strings, significant only at root) _array_of
 @param maxprocs [in] maximum number of processes to start for each command (array of integer, significant only at root) _array_of
 @param info [in] info objects telling the runtime system where and how to start processes (array of handles, significant only at root)
 @param root [in] rank of process in which previous arguments are examined (integer)
 @param comm [in] intracommunicator containing group of spawning processes (handle)
 @param intercomm [out] intercommunicator between original group and newly spawned group (handle) _array_of
 @param errcodes [out] one error code per process (array of integer)
 */
extern "C" int MPI_Comm_spawn_multiple(int count, char *_array_of_commands_[],
    char* *_array_of_argv_[], int _array_of_maxprocs_[],
    MPI_Info _array_of_info_[], int root, MPI_Comm comm, MPI_Comm *intercomm,
    int _array_of_errcodes_[])
{
  sst_throw(unimplementederror, "MPI_Comm_spawn_multiple");
  return MPI_SUCCESS;
}

/**
 @param name [in] a service name (string)
 @param info [in] implementation-specific information (handle) _port
 @param name [in] a service name (string) _info _
 [in] implementation-specific information (handle) _port_name _
 [out] a port name (string)
 */
extern "C" int MPI_Lookup_name(char *service_name, MPI_Info info,
    char *port_name)
{
  sst_throw(unimplementederror, "MPI_Lookup_name");
  return MPI_SUCCESS;
}

/**
 @param info [in] implementation-specific information on how to establish a port for [MPI_Comm_accept](MPI_Comm_accept.html) (handle) _port
 @param name [out] newly established port (string)
 */
extern "C" int MPI_Open_port(MPI_Info info, char *port_name)
{
  sst_throw(unimplementederror, "MPI_Open_port");
  return MPI_SUCCESS;
}

/**
 @param name [in] a service name to associate with the port (string)
 @param info [in] implementation-specific information (handle) _port
 @param name [in] a service name to associate with the port (string) _info _
 [in] implementation-specific information (handle) _port_name _
 [in] a port name (string)
 */
extern "C" int MPI_Publish_name(char *service_name, MPI_Info info,
    char *port_name)
{
  sst_throw(unimplementederror, "MPI_Publish_name");
  return MPI_SUCCESS;
}

/**
 @param name [in] a service name (string)
 @param info [in] implementation-specific information (handle) _port
 @param name [in] a service name (string) _info _
 [in] implementation-specific information (handle) _port_name _
 [in] a port name (string)
 */
extern "C" int MPI_Unpublish_name(char *service_name, MPI_Info info,
    char *port_name)
{
  sst_throw(unimplementederror, "MPI_Unpublish_name");
  return MPI_SUCCESS;
}

/**
 @param addr [in] initial address of buffer (choice) _origin
 @param count [in] number of entries in buffer (nonnegative integer) _origin
 @param datatype [in] datatype of each buffer entry (handle) _target
 @param rank [in] rank of target (nonnegative integer) _target
 @param disp [in] displacement from start of window to beginning of target buffer (nonnegative integer) _target
 @param count [in] number of entries in buffer (nonnegative integer) _origin
 @param datatype [in] datatype of each buffer entry (handle) _target_rank _
 [in] rank of target (nonnegative integer) _target_disp _
 [in] displacement from start of window to beginning of target buffer (nonnegative integer) _target_count _
 [in] number of entries in target buffer (nonnegative integer) _target_datatype _
 [in] datatype of each entry in target buffer (handle)
 @param op [in] predefined reduce operation (handle)
 @param win [in] window object (handle)
 */
extern "C" int MPI_Accumulate(void *origin_addr, int origin_count,
    MPI_Datatype origin_datatype, int target_rank, MPI_Aint target_disp,
    int target_count, MPI_Datatype target_datatype, MPI_Op op, MPI_Win win)
{
  static const char* here = "MPI_Accumulate";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpi_window::ptr winptr = get_window(rank, win);
  sstmac::payload::const_ptr load = get_mpi_payload_func(origin_addr, 0,
      origin_count, origin_datatype, true);

  mpitype ttype = get_mpi_type(target_datatype);
  mpiid trank(target_rank);
  mpirequest_t req;
  win_rma_info info;

  info.buffer = origin_addr;
  info.count = origin_count;
  info.datatype = origin_datatype;
  info.win = winptr;
  info.type = win_rma_info::acc;
  info.op = get_mpi_op(op);

  mpi->win_accumulate(req, load, target_count, ttype, trank, target_disp, info.op, winptr);

  pending_rmas[req] = info;

  return MPI_SUCCESS;
}

/**
 @param win [in] window object (handle)
 */
extern "C" int MPI_Win_complete(MPI_Win win)
{
  sst_throw(unimplementederror, "MPI_Win_complete");
  return MPI_SUCCESS;
}

/**
 @param base [in] initial address of window (choice)
 @param size [in] size of window in bytes (nonnegative integer) _disp
 @param unit [in] local unit size for displacements, in bytes (positive integer)
 @param info [in] info argument (handle)
 @param comm [in] communicator (handle)
 @param win [out] window object returned by the call (handle)
 */
extern "C" int MPI_Win_create(void *base, MPI_Aint size, int disp_unit,
    MPI_Info info, MPI_Comm comm, MPI_Win *win)
{
  static const char* here = "MPI_Win_create";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);
  mpi_window::ptr winptr;
  mpiinfo::ptr infptr;
  MPI_Win newwin = win_counters[rank]++;
  if (info != MPI_INFO_NULL)
  {
    infptr = get_info(rank, info);
  }
  mpi->win_create(base, size, disp_unit, infptr, commptr, winptr, newwin);

  if (winptr)
  {
    *win = add_win_ptr(rank, winptr, newwin);
  }

  return MPI_SUCCESS;
}

/**
 @param assert [in] program assertion (integer)
 @param win [in] window object (handle)
 */
extern "C" int MPI_Win_fence(int assert, MPI_Win win)
{
  static const char* here = "MPI_Win_fence";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpi_window::ptr winptr = get_window(rank, win);

  boost::unordered_map<mpirequest_t, payload::const_ptr> loads;

  bool barr = (assert & MPI_MODE_NOPRECEDE) == 0;
  mpi->win_fence(winptr, barr, loads);

  //set payloads for gets
  boost::unordered_map<mpirequest_t, payload::const_ptr>::iterator it, end =
      loads.end();
  for (it = loads.begin(); it != end; it++)
  {
    if (!it->first)
    {
      sst_throw(ssterror, "sstmac_mpi::MPI_Win_fence - null request");
    }
    if (it->second)
    {
      if (pending_rmas.find(it->first) == pending_rmas.end())
      {
        std::cout << "request: " << it->first->to_string() << "\n";
        sst_throw(ssterror, "sstmac_mpi::MPI_Win_fence - didn't find get with request");
      }
      win_rma_info i = pending_rmas[it->first];
      if (i.type == win_rma_info::get)
      {
        set_mpi_payload(i.buffer, 0, i.count, i.datatype, it->second);
      }
      pending_rmas.erase(it->first);
    }
  }

  return MPI_SUCCESS;
}

/**
 @param win [in] window object (handle)
 */
extern "C" int MPI_Win_free(MPI_Win *win)
{
  static const char* here = "MPI_Win_free";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpi_window::ptr winptr = get_window(rank, *win);

  mpi->win_free(winptr);

  *win = MPI_WIN_NULL;

  return MPI_SUCCESS;
}

/**
 @param win [in] window object (handle)
 @param group [out] group of processes which share access to the window (handle)
 */
extern "C" int MPI_Win_get_group(MPI_Win win, MPI_Group *group)
{
  sst_throw(unimplementederror, "MPI_Win_get_group");
  return MPI_SUCCESS;
}

/**
 @param type [in] Indicates whether other processes may access the target window at the same time (if MPI_LOCK_SHARED) or not (MPI_LOCK_EXCLUSIVE)
 @param rank [in] rank of locked window (nonnegative integer)
 @param assert [in] Used to optimize this call; zero may be used as a default. See notes. (integer)
 @param win [in] window object (handle)
 */
extern "C" int MPI_Win_lock(int lock_type, int r, int assert, MPI_Win win)
{
  static const char* here = "MPI_Win_lock";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpi_window::ptr winptr = get_window(rank, win);

  mpiapi::LOCK_TYPE t;
  if (lock_type == MPI_LOCK_SHARED)
  {
    t = mpiapi::mpi_lock_shared_;
  }
  else if (lock_type == MPI_LOCK_EXCLUSIVE)
  {
    t = mpiapi::mpi_lock_exclusive_;
  }
  else
  {
    sst_throw(ssterror, "MPI_Win_lock:: unknown lock type %d", lock_type);
  }

  mpiapi::WIN_FLAGS f;
  if (assert == MPI_MODE_NOCHECK)
  {
    f = mpiapi::mpi_win_mode_nocheck_;
  }
  else
  {
    f = mpiapi::mpi_flag_none_;
  }

  mpi->win_lock(t, mpiid(r), f, winptr);
  return MPI_SUCCESS;
}

/**
 @param group [in] group of origin processes (handle)
 @param assert [in] Used to optimize this call; zero may be used as a default. See notes. (integer)
 @param win [in] window object (handle)
 */
extern "C" int MPI_Win_post(MPI_Group group, int assert, MPI_Win win)
{
  sst_throw(unimplementederror, "MPI_Win_post");
  return MPI_SUCCESS;
}

/**
 @param group [in] group of target processes (handle)
 @param assert [in] Used to optimize this call; zero may be used as a default. See notes. (integer)
 @param win [in] window object (handle)
 */
extern "C" int MPI_Win_start(MPI_Group group, int assert, MPI_Win win)
{
  sst_throw(unimplementederror, "MPI_Win_start");
  return MPI_SUCCESS;
}

/**
 @param win [in] window object (handle)
 @param flag [out] success flag (logical)
 */
extern "C" int MPI_Win_test(MPI_Win win, int *flag)
{
  sst_throw(unimplementederror, "MPI_Win_test");
  return MPI_SUCCESS;
}

/**
 @param rank [in] rank of window (nonnegative integer)
 @param win [in] window object (handle)
 */
extern "C" int MPI_Win_unlock(int r, MPI_Win win)
{
  static const char* here = "MPI_Win_unlock";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpi_window::ptr winptr = get_window(rank, win);

  boost::unordered_map<mpirequest_t, payload::const_ptr> loads;

  mpi->win_unlock(mpiid(r), winptr, loads);

  //set payloads for gets
  boost::unordered_map<mpirequest_t, payload::const_ptr>::iterator it, end =
      loads.end();
  for (it = loads.begin(); it != end; it++)
  {
    if (!it->first)
    {
      sst_throw(ssterror, "sstmac_mpi::MPI_Win_fence - null request");
    }
    if (it->second)
    {
      if (pending_rmas.find(it->first) == pending_rmas.end())
      {
        std::cout << "request: " << it->first->to_string() << "\n";
        sst_throw(ssterror, "sstmac_mpi::MPI_Win_fence - didn't find get with request");
      }
      win_rma_info i = pending_rmas[it->first];
      if (i.type == win_rma_info::get)
      {
        set_mpi_payload(i.buffer, 0, i.count, i.datatype, it->second);
      }
      pending_rmas.erase(it->first);
    }
  }

  return MPI_SUCCESS;
}

/**
 @param win [in] window object (handle)
 */
extern "C" int MPI_Win_wait(MPI_Win win)
{
  sst_throw(unimplementederror, "MPI_Win_wait");
  return MPI_SUCCESS;
}

/**
 @param sendbuf [in] starting address of send buffer (choice)
 @param sendcounts [in] integer array equal to the group size specifying the number of elements to send to each processor (integer)
 @param sdispls [in] integer array (of length group size). Entry j specifies the displacement in bytes (relative to sendbuf) from which to take the outgoing data destined for process j
 @param sendtypes [in] array of datatypes (of length group size). Entry j specifies the type of data to send to process j (handle)
 @param recvbuf [out] address of receive buffer (choice)
 @param recvcounts [in] integer array equal to the group size specifying the number of elements that can be received from each processor (integer)
 @param rdispls [in] integer array (of length group size). Entry i specifies the displacement in bytes (relative to recvbuf) at which to place the incoming data from process i
 @param recvtypes [in] array of datatypes (of length group size). Entry i specifies the type of data received from process i (handle)
 @param comm [in] communicator (handle)
 */
extern "C" int MPI_Alltoallw(void *sendbuf, int *sendcnts, int *sdispls,
    MPI_Datatype *sendtypes, void *recvbuf, int *recvcnts, int *rdispls,
    MPI_Datatype *recvtypes, MPI_Comm comm)
{
  sst_throw(unimplementederror, "MPI_Alltoallw");
  return MPI_SUCCESS;
}

/**
 @param sendbuf [in] starting address of send buffer (choice)
 @param recvbuf [out] starting address of receive buffer (choice)
 @param count [in] number of elements in input buffer (integer)
 @param datatype [in] data type of elements of input buffer (handle)
 @param op [in] operation (handle)
 @param comm [in] communicator (handle)
 */
extern "C" int MPI_Exscan(void *sendbuf, void *recvbuf, int count,
    MPI_Datatype datatype, MPI_Op op, MPI_Comm comm)
{
  sst_throw(unimplementederror, "MPI_Exscan");
  return MPI_SUCCESS;
}

/**
 @param errorclass [out] New error class
 */
extern "C" int MPI_Add_error_class(int *errorclass)
{
  sst_throw(unimplementederror, "MPI_Add_error_class");
  return MPI_SUCCESS;
}

/**
 @param errorclass [in] Error class to add an error code.
 @param errorcode [out] New error code for this error class.
 */
extern "C" int MPI_Add_error_code(int errorclass, int *errorcode)
{
  sst_throw(unimplementederror, "MPI_Add_error_code");
  return MPI_SUCCESS;
}

/**
 @param errorcode [in] error code or class (integer)
 @param string [in] text corresponding to errorcode (string)
 */
extern "C" int MPI_Add_error_string(int errorcode, char *string)
{
  sst_throw(unimplementederror, "MPI_Add_error_string");
  return MPI_SUCCESS;
}

/**
 @param comm [in] communicator with error handler (handle)
 @param errorcode [in] error code (integer)
 */
extern "C" int MPI_Comm_call_errhandler(MPI_Comm comm, int errorcode)
{
  sst_throw(unimplementederror, "MPI_Comm_call_errhandler");
  return MPI_SUCCESS;
}

/**
 @param fn [in] Copy callback function for keyval _comm_delete_attr
 @param fn [in] Copy callback function for keyval _comm_delete_attr_fn _
 [in] Delete callback function for keyval

 _comm
 @param keyval [out] key value for future access (integer)  _extra
 @param state [in] Extra state for callback functions
 */
extern "C" int MPI_Comm_create_keyval(
    MPI_Comm_copy_attr_function *comm_copy_attr_fn,
    MPI_Comm_delete_attr_function *comm_delete_attr_fn, int *comm_keyval,
    void *extra_state)
{
  sst_throw(unimplementederror, "MPI_Comm_create_keyval");
  return MPI_SUCCESS;
}

/**
 @param comm [in] communicator to which attribute is attached (handle) _comm
 @param keyval [in] The key value of the deleted attribute (integer)
 */
extern "C" int MPI_Comm_delete_attr(MPI_Comm comm, int commkeyval_)
{
  sst_throw(unimplementederror, "MPI_Comm_delete_attr");
  return MPI_SUCCESS;
}

/**
 @param keyval [in] Frees the integer key value (integer)
 */
extern "C" int MPI_Comm_free_keyval(int *comm_keyval)
{
  sst_throw(unimplementederror, "MPI_Comm_free_keyval");
  return MPI_SUCCESS;
}

/**
 @param comm [in] communicator to which attribute is attached (handle) _comm
 @param keyval [in] key value (integer) _attribute
 @param val [out] attribute value, unless flag = false
 @param flag [out] true if an attribute value was extracted; false if no attribute is associated with the key
 */
extern "C" int MPI_Comm_get_attr(MPI_Comm comm, int commkeyval_,
    void *attribute_val, int *flag)
{
  sst_throw(unimplementederror, "MPI_Comm_get_attr");
  return MPI_SUCCESS;
}

/**
 @param comm [in] Communicator to get name of (handle) _comm
 @param name [out] On output, contains the name of the communicator. It must be an array of size at least MPI_MAX_OBJECT_NAME.
 @param resultlen [out] Number of characters in name
 */
extern "C" int MPI_Comm_get_name(MPI_Comm comm, char *commname_, int *resultlen)
{
  static const char* here = "MPI_Comm_set_name";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);
  std::string n = commptr->get_name();
  strcpy(commname_, n.c_str());
  *resultlen = n.size();
  return MPI_SUCCESS;
}

/**
 @param comm [in] communicator to which attribute will be attached (handle)
 @param keyval [in] key value, as returned by [MPI_Comm_create_keyval](MPI_Comm_create_keyval.html) (integer) _attribute
 @param val [in] attribute value
 */
extern "C" int MPI_Comm_set_attr(MPI_Comm comm, int commkeyval_,
    void *attribute_val)
{
  sst_throw(unimplementederror, "MPI_Comm_set_attr");
  return MPI_SUCCESS;
}

/**
 @param comm [in] communicator to name (handle) _comm
 @param name [in] Name for communicator (string)
 */
extern "C" int MPI_Comm_set_name(MPI_Comm comm, char *commname_)
{
  static const char* here = "MPI_Comm_set_name";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpicomm::ptr commptr = get_comm(rank, comm);
  commptr->set_name(commname_);
  return MPI_SUCCESS;
}

/**
 @param fh [in] MPI file with error handler (handle)
 @param errorcode [in] error code (integer)
 */
extern "C" int MPI_File_call_errhandler(MPI_File fh, int errorcode)
{
  sst_throw(unimplementederror, "MPI_File_call_errhandler");
  return MPI_SUCCESS;
}

/**
 @param request [in] Generalized request to mark as complete
 */
extern "C" int MPI_Grequest_complete(MPI_Request request)
{
  sst_throw(unimplementederror, "MPI_Grequest_complete");
  return MPI_SUCCESS;
}

/**
 @param fn [in] callback function invoked when request status is queried (function) _free
 @param fn [in] callback function invoked when request status is queried (function) _free
 @param fn [in] callback function invoked when request status is queried (function) _free_fn _
 [in] callback function invoked when request is freed (function) _cancel_fn _
 [in] callback function invoked when request is cancelled (function) _extra
 @param state [in] Extra state passed to the above functions.
 @param request [out] Generalized request (handle)
 */
extern "C" int MPI_Grequest_start(MPI_Grequest_query_function *query_fn,
    MPI_Grequest_free_function *free_fn,
    MPI_Grequest_cancel_function *cancel_fn, void *extra_state,
    MPI_Request *request)
{
  sst_throw(unimplementederror, "MPI_Grequest_start");
  return MPI_SUCCESS;
}

/**
 @param argc [in] Pointer to the number of arguments
 @param argv [in] Pointer to the argument vector
 @param required [in] Level of desired thread support
 @param provided [out] Level of provided thread support
 */
extern "C" int MPI_Init_thread(int *argc, char ***argv, int required,
    int *provided)
{
  *provided = required;
  return MPI_Init(argc, argv);
}

/**
 @param flag [out] Flag is true if [MPI_Init](MPI_Init.html) or [MPI_Init_thread](MPI_Init_thread.html) has been called by this thread and false otherwise. (logical)
 */
extern "C" int MPI_Is_thread_main(int *flag)
{
  sst_throw(unimplementederror, "MPI_Is_thread_main");
  return MPI_SUCCESS;
}

/**
 @param provided [out] Level of thread support provided. This is the same value that was returned in the provided argument in [MPI_Init_thread](MPI_Init_thread.html).
 */
extern "C" int MPI_Query_thread(int *provided)
{
  sst_throw(unimplementederror, "MPI_Query_thread");
  return MPI_SUCCESS;
}

/**
 @param status [in] status to associate cancel flag with (Status)
 @param flag [in] if true indicates request was cancelled (logical)
 */
extern "C" int MPI_Status_set_cancelled(MPI_Status *status, int flag)
{
  sst_throw(unimplementederror, "MPI_Status_set_cancelled");
  return MPI_SUCCESS;
}

/**
 @param status [in] status to associate count with (Status)
 @param datatype [in] datatype associated with count (handle)
 @param count [in] number of elements to associate with status (integer)
 */
extern "C" int MPI_Status_set_elements(MPI_Status *status,
    MPI_Datatype datatype, int count)
{
  sst_throw(unimplementederror, "MPI_Status_set_elements");
  return MPI_SUCCESS;
}

/**
 @param fn [in] copy callback function for type_keyval (function) _type_delete_attr
 @param fn [in] copy callback function for type_keyval (function) _type_delete_attr_fn _
 [in] delete callback function for type_keyval (function) _type
 @param keyval [out] key value for future access (integer) _extra
 @param state [in] extra state for callback functions
 */
extern "C" int MPI_Type_create_keyval(
    MPI_Type_copy_attr_function *type_copy_attr_fn,
    MPI_Type_delete_attr_function *type_delete_attr_fn, int *type_keyval,
    void *extra_state)
{
  sst_throw(unimplementederror, "MPI_Type_create_keyval");
  return MPI_SUCCESS;
}

/**
 @param type [in] MPI datatype to which attribute is attached (handle) _type
 @param keyval [in] The key value of the deleted attribute (integer)
 */
extern "C" int MPI_Type_delete_attr(MPI_Datatype type, int typekeyval_)
{
  sst_throw(unimplementederror, "MPI_Type_delete_attr");
  return MPI_SUCCESS;
}

/**
 @param keyval [in] Frees the integer key value (integer)
 */
extern "C" int MPI_Type_free_keyval(int *type_keyval)
{
  sst_throw(unimplementederror, "MPI_Type_free_keyval");
  return MPI_SUCCESS;
}

/**
 @param type [in] datatype to which the attribute is attached (handle) _type
 @param keyval [in] key value (integer) _attribute
 @param val [out] attribute value, unless flag = false
 @param flag [out] false if no attribute is associated with the key (logical)
 */
extern "C" int MPI_Type_get_attr(MPI_Datatype type, int typekeyval_,
    void *attribute_val, int *flag)
{
  sst_throw(unimplementederror, "MPI_Type_get_attr");
  return MPI_SUCCESS;
}

/**
 @param datatype [in] datatype _max
 @param integers [in] max integers _max
 @param addresses [in] max addresses _max
 @param datatypes [in] max datatypes _array_of
 @param integers [in] max integers _max
 @param addresses [in] max addresses _max
 @param datatypes [in] max datatypes _array_of_integers _
 [out] integers _array_of_addresses _
 [out] addresses _array_of_datatypes _
 [out] datatypes
 */
extern "C" int MPI_Type_get_contents(MPI_Datatype datatype, int max_integers,
    int max_addresses, int max_datatypes, int _array_of_integers_[],
    MPI_Aint _array_of_addresses_[], MPI_Datatype _array_of_datatypes_[])
{
  static const char* here = "MPI_Type_get_contents";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  mpitype t = get_mpi_type(datatype);
  t.get_contents(_array_of_integers_, (int*)_array_of_addresses_,  (long*)_array_of_datatypes_);
  return MPI_SUCCESS;
}

/**
 @param datatype [in] datatype _num
 @param integers [out] num integers _num
 @param addresses [out] num addresses _num
 @param datatypes [out] num datatypes
 @param combiner [out] combiner
 */
extern "C" int MPI_Type_get_envelope(MPI_Datatype datatype, int *num_integers,
    int *num_addresses, int *num_datatypes, int *combiner)
{
  static const char* here = "MPI_Type_get_envelope";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  mpitype t = get_mpi_type(datatype);
  t.get_envelope(num_integers, num_addresses, num_datatypes, combiner);
  return MPI_SUCCESS;
}

/**
 @param type [in] MPI Datatype to which attribute will be attached (handle)
 @param keyval [in] key value, as returned by [MPI_Type_create_keyval](MPI_Type_create_keyval.html) (integer) _attribute
 @param val [in] attribute value
 */
extern "C" int MPI_Type_set_attr(MPI_Datatype type, int typekeyval_,
    void *attribute_val)
{
  sst_throw(unimplementederror, "MPI_Type_set_attr");
  return MPI_SUCCESS;
}

/**
 @param typeclass [in] generic type specifier (integer)
 @param size [in] size, in bytes, of representation (integer)
 @param datatype [out] datatype with correct type, size (handle)
 */
extern "C" int MPI_Type_match_size(int typeclass, int size,
    MPI_Datatype *datatype)
{
  sst_throw(unimplementederror, "MPI_Type_match_size");
  return MPI_SUCCESS;
}

/**
 @param win [in] window with error handler (handle)
 @param errorcode [in] error code (integer)
 */
extern "C" int MPI_Win_call_errhandler(MPI_Win win, int errorcode)
{
  sst_throw(unimplementederror, "MPI_Win_call_errhandler");
  return MPI_SUCCESS;
}

/**
 @param fn [in] copy callback function for win_keyval (function) _win_delete_attr
 @param fn [in] copy callback function for win_keyval (function) _win_delete_attr_fn _
 [in] delete callback function for win_keyval (function) _win
 @param keyval [out] key value for future access (integer) _extra
 @param state [in] extra state for callback functions
 */
extern "C" int MPI_Win_create_keyval(
    MPI_Win_copy_attr_function *win_copy_attr_fn,
    MPI_Win_delete_attr_function *win_delete_attr_fn, int *win_keyval,
    void *extra_state)
{
  sst_throw(unimplementederror, "MPI_Win_create_keyval");
  return MPI_SUCCESS;
}

/**
 @param win [in] window from which the attribute is deleted (handle) _win
 @param keyval [in] key value (integer)
 */
extern "C" int MPI_Win_delete_attr(MPI_Win win, int winkeyval_)
{
  sst_throw(unimplementederror, "MPI_Win_delete_attr");
  return MPI_SUCCESS;
}

/**
 @param keyval [in] key value (integer)
 */
extern "C" int MPI_Win_free_keyval(int *win_keyval)
{
  sst_throw(unimplementederror, "MPI_Win_free_keyval");
  return MPI_SUCCESS;
}

/**
 @param win [in] window to which the attribute is attached (handle) _win
 @param keyval [in] key value (integer) _attribute
 @param val [out] attribute value, unless flag is false
 @param flag [out] false if no attribute is associated with the key (logical)
 */
extern "C" int MPI_Win_get_attr(MPI_Win win, int winkeyval_,
    void *attribute_val, int *flag)
{
  sst_throw(unimplementederror, "MPI_Win_get_attr");
  return MPI_SUCCESS;
}

/**
 @param win [in] window whose name is to be returned (handle) _win
 @param name [out] the name previously stored on the window, or a empty string if no such name exists (string)
 @param resultlen [out] length of returned name (integer)
 */
extern "C" int MPI_Win_get_name(MPI_Win win, char *winname_, int *resultlen)
{
  sst_throw(unimplementederror, "MPI_Win_get_name");
  return MPI_SUCCESS;
}

/**
 @param win [in] MPI window object to which attribute will be attached (handle)
 @param keyval [in] key value, as returned by [MPI_Win_create_keyval](MPI_Win_create_keyval.html) (integer) _attribute
 @param val [in] attribute value
 */
extern "C" int MPI_Win_set_attr(MPI_Win win, int winkeyval_,
    void *attribute_val)
{
  sst_throw(unimplementederror, "MPI_Win_set_attr");
  return MPI_SUCCESS;
}

/**
 @param win [in] window whose identifier is to be set (handle) _win
 @param name [in] the character string which is remembered as the name (string)
 */
extern "C" int MPI_Win_set_name(MPI_Win win, char *winname_)
{
  sst_throw(unimplementederror, "MPI_Win_set_name");
  return MPI_SUCCESS;
}

/**
 @param fh [in] C file handle (handle)
 */
extern "C" int MPI_File_c2f(MPI_File mpi_fh)
{
  sst_throw(unimplementederror, "MPI_File_c2f");
  return MPI_SUCCESS;
}

/**
 @param size [in] size of memory segment in bytes (nonnegative integer)
 @param info [in] info argument (handle)
 @param baseptr [out] pointer to a pointer to the beginning of the memory segment allocated
 */
extern "C" int MPI_Alloc_mem(MPI_Aint size, MPI_Info info, void *baseptr)
{
  void** cast = (void**) baseptr;
  *cast = malloc(size);
  return MPI_SUCCESS;
}

/**
 @param function user defined error handling procedure (function)
 @param errhandler MPI error handler (handle)
 */
extern "C" int MPI_Comm_create_errhandler(MPI_Comm_errhandler_fn *function,
    MPI_Errhandler *errhandler)
{
 // sst_throw(unimplementederror, "MPI_Comm_create_errhandler");
  return MPI_SUCCESS;
}

/**
 @param comm [in] communicator (handle)
 @param errhandler [out] handler currently associated with communicator (handle)
 */
extern "C" int MPI_Comm_get_errhandler(MPI_Comm comm,
    MPI_Errhandler *errhandler)
{
  return MPI_Errhandler_get(comm, errhandler);
}

/**
 @param comm [in] communicator (handle)
 @param errhandler [in] new error handler for communicator (handle)
 */
extern "C" int MPI_Comm_set_errhandler(MPI_Comm comm, MPI_Errhandler errhandler)
{
  return MPI_Errhandler_set(comm, errhandler);
}

/**
 @param function [in] user defined error handling procedure (function)
 @param errhandler [out] MPI error handler (handle)
 */
extern "C" int MPI_File_create_errhandler(
    MPI_File_errhandler_function *function, MPI_Errhandler *errhandler)
{
  sst_throw(unimplementederror, "MPI_File_create_errhandler");
  return MPI_SUCCESS;
}

/**
 @param file [in] MPI file (handle)
 @param errhandler [out] handler currently associated with file (handle)
 */
extern "C" int MPI_File_get_errhandler(MPI_File file,
    MPI_Errhandler *errhandler)
{
  sst_throw(unimplementederror, "MPI_File_get_errhandler");
  return MPI_SUCCESS;
}

/**
 @param file [in] MPI file (handle)
 @param errhandler [in] new error handler for file (handle)
 */
extern "C" int MPI_File_set_errhandler(MPI_File file, MPI_Errhandler errhandler)
{
  sst_throw(unimplementederror, "MPI_File_set_errhandler");
  return MPI_SUCCESS;
}

/**
 @param flag [out] Flag is true if [MPI_Finalize](MPI_Finalize.html) has been called and false otherwise. (logical)
 */
extern "C" int MPI_Finalized(int *flag)
{
  static const char* here = "MPI_Finalized";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  *flag = mpi->finalized();
  return MPI_SUCCESS;
}

/**
 @param base [in] initial address of memory segment allocated by MPI_ALLOC_MEM (choice)
 */
extern "C" int MPI_Free_mem(void *base)
{
  free(base);
  return MPI_SUCCESS;
}

/**
 @param location [in] location in caller memory (choice)
 @param address [out] address of location (address)
 */
extern "C" int MPI_Get_address(void *location, MPI_Aint *address)
{
  *address = (MPI_Aint)location;
  return MPI_SUCCESS;
}

/**
 @param info [out] info object created (handle)
 */
extern "C" int MPI_Info_create(MPI_Info *info)
{
  sst_throw(unimplementederror, "MPI_Info_create");
  return MPI_SUCCESS;
}

/**
 @param info [in] info object (handle)
 @param key [in] key (string)
 */
extern "C" int MPI_Info_delete(MPI_Info info, char *key)
{
  sst_throw(unimplementederror, "MPI_Info_delete");
  return MPI_SUCCESS;
}

/**
 @param info [in] info object (handle)
 @param newinfo [out] duplicate of info object (handle)
 */
extern "C" int MPI_Info_dup(MPI_Info info, MPI_Info *newinfo)
{
  sst_throw(unimplementederror, "MPI_Info_dup");
  return MPI_SUCCESS;
}

/**
 @param info [in] info object to be freed (handle)
 */
extern "C" int MPI_Info_free(MPI_Info *info)
{
  sst_throw(unimplementederror, "MPI_Info_free");
  return MPI_SUCCESS;
}

/**
 @param info [in] info object (handle)
 @param key [in] key (string)
 @param valuelen [in] length of value argument (integer)
 @param value [out] value (string)
 @param flag [out] true if key defined, false if not (boolean)
 */
extern "C" int MPI_Info_get(MPI_Info info, char *key, int valuelen,
    char *value, int *flag)
{
  sst_throw(unimplementederror, "MPI_Info_get");
  return MPI_SUCCESS;
}

/**
 @param info [in] info object (handle)
 @param nkeys [out] number of defined keys (integer)
 */
extern "C" int MPI_Info_get_nkeys(MPI_Info info, int *nkeys)
{
  sst_throw(unimplementederror, "MPI_Info_get_nkeys");
  return MPI_SUCCESS;
}

/**
 @param info [in] info object (handle)
 @param n [in] key number (integer) _key_
 [out] key (string). The maximum number of characters is MPI_MAX_INFO_KEY.
 */
extern "C" int MPI_Info_get_nthkey(MPI_Info info, int n, char *key)
{
  sst_throw(unimplementederror, "MPI_Info_get_nthkey");
  return MPI_SUCCESS;
}

/**
 @param info [in] info object (handle)
 @param key [in] key (string)
 @param valuelen [out] length of value argument (integer)
 @param flag [out] true if key defined, false if not (boolean)
 */
extern "C" int MPI_Info_get_valuelen(MPI_Info info, char *key, int *valuelen,
    int *flag)
{
  sst_throw(unimplementederror, "MPI_Info_get_valuelen");
  return MPI_SUCCESS;
}

/**
 @param info [in] info object (handle)
 @param key [in] key (string)
 @param value [in] value (string)
 */
extern "C" int MPI_Info_set(MPI_Info info, char *key, char *value)
{
  sst_throw(unimplementederror, "MPI_Info_set");
  return MPI_SUCCESS;
}

/**
 @param datarep [in] data representation (string)
 @param inbuf [in] input buffer start (choice)
 @param incount [in] number of input data items (integer)
 @param datatype [in] datatype of each input data item (handle)
 @param outbuf [out] output buffer start (choice)
 @param outcount [in] output buffer size, in bytes (integer)
 @param position [in/out] current position in buffer, in bytes (integer)
 */
extern "C" int MPI_Pack_external(char *datarep, void *inbuf, int incount,
    MPI_Datatype datatype, void *outbuf, MPI_Aint outcount, MPI_Aint *position)
{
  sst_throw(unimplementederror, "MPI_Pack_external");
  return MPI_SUCCESS;
}

/**
 @param datarep [in] data representation (string)
 @param incount [in] number of input data items (integer)
 @param datatype [in] datatype of each input data item (handle)
 @param size [out] output buffer size, in bytes (integer)
 */
extern "C" int MPI_Pack_external_size(char *datarep, int incount,
    MPI_Datatype datatype, MPI_Aint *size)
{
  sst_throw(unimplementederror, "MPI_Pack_external_size");
  return MPI_SUCCESS;
}

/**
 @param request [in] request (handle)
 @param flag [out] true if operation has completed (logical)
 @param status [out] status object (Status). May be MPI_STATUS_IGNORE.
 */
extern "C" int MPI_Request_get_status(MPI_Request request, int *flag,
    MPI_Status *status)
{
  static const char* here = "MPI_Request_get_status";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();

  int rank = mpi->comm_world()->rank().id;

  mpirequest_t req = get_request(rank, request);


  if (!req->is_cancelled())
  {
    if (req->is_complete())
    {
      const_mpistatus_t stat = req->get_status();
      if (status != MPI_STATUS_IGNORE && stat)
      {
        status->MPI_SOURCE = stat->source().id;
        status->MPI_TAG = stat->tag().id;
        status->MPI_ERROR = 0;
        status->COUNT = stat->get_count();
        status->CANCELLED = false;
      }
      *flag = true;
    }else
      *flag = false;
  }
  else
  {
    *flag = true;
    status->CANCELLED = true;
  }
  return MPI_SUCCESS;
}

/**
 @param size [in] size of process group (positive integer)
 @param rank [in] rank in process group (nonnegative integer)
 @param ndims [in] number of array dimensions as well as process grid dimensions (positive integer) _array_of
 @param gsizes [in] number of elements of type oldtype in each dimension of global array (array of positive integers) _array_of
 @param distribs [in] distribution of array in each dimension (array of state) _array_of
 @param dargs [in] distribution argument in each dimension (array of positive integers) _array_of
 @param psizes [in] size of process grid in each dimension (array of positive integers)
 @param order [in] array storage order flag (state)
 @param oldtype [in] old datatype (handle)
 @param newtype [out] new datatype (handle)
 */
extern "C" int MPI_Type_create_darray(int size, int rank, int ndims,
    int _array_of_gsizes_[], int _array_of_distribs_[], int _array_of_dargs_[],
    int _array_of_psizes_[], int order, MPI_Datatype oldtype,
    MPI_Datatype *newtype)
{
  sst_throw(unimplementederror, "MPI_Type_create_darray");
  return MPI_SUCCESS;
}

/**
 @param count [in] number of blocks --- also number of entries in displacements and blocklengths (integer)
 @param blocklengths [in] number of elements in each block (array of nonnegative integers)
 @param displacements [in] byte displacement of each block (array of integer)
 @param oldtype [in] old datatype (handle)
 @param newtype [out] new datatype (handle)
 */
extern "C" int MPI_Type_create_hindexed(int count, int _blocklengths_[],
    MPI_Aint _displacements_[], MPI_Datatype oldtype, MPI_Datatype *newtype)
{
  sst_throw(unimplementederror, "MPI_Type_create_hindexed");
  return MPI_SUCCESS;
}

/**
 @param count [in] number of blocks (nonnegative integer)
 @param blocklength [in] number of elements in each block (nonnegative integer)
 @param stride [in] number of bytes between start of each block (integer)
 @param oldtype [in] old datatype (handle)
 @param newtype [out] new datatype (handle)
 */
extern "C" int MPI_Type_create_hvector(int count, int blocklength,
    MPI_Aint stride, MPI_Datatype oldtype, MPI_Datatype *newtype)
{
  sst_throw(unimplementederror, "MPI_Type_create_hvector");
  return MPI_SUCCESS;
}

/**
 @param count [in] length of array of displacements (integer)
 @param blocklength [in] size of block (integer) _array_of
 @param displacements [in] array of displacements (array of integer)
 @param oldtype [in] old datatype (handle)
 @param newtype [out] new datatype (handle)
 */
extern "C" int MPI_Type_create_indexed_block(int count, int blocklength,
    int _array_of_displacements_[], MPI_Datatype oldtype, MPI_Datatype *newtype)
{
  static const char* here = "MPI_Type_create_indexed_block";
  SSTMAC_DEBUG << here << "\n";
  mpitype ty = get_mpi_type(oldtype);
  mpitype newty;
  mpiapi::ptr mpi = get_current_mpi();
  int* blocklens = (int*) malloc(sizeof(int) * count);
  for (int i = 0; i < count; i++)
    blocklens[i] = blocklength;

  std::vector<int> ind(count);
  for (int i = 0; i < count; i++)
  {
    ind[i] = _array_of_displacements_[i];
  }
  mpi->type_indexed(count, blocklens, ind, ty, newty, true, MPI_COMBINER_INDEXED_BLOCK);
  free(blocklens);
  *newtype = newty.id;
  return MPI_SUCCESS;
  return MPI_SUCCESS;
}

/**
 @param oldtype [in] input datatype (handle)
 @param lb [in] new lower bound of datatype (integer)
 @param extent [in] new extent of datatype (integer)
 @param newtype [out] output datatype (handle)
 */
extern "C" int MPI_Type_create_resized(MPI_Datatype oldtype, MPI_Aint lb,
    MPI_Aint extent, MPI_Datatype *newtype)
{
  sst_throw(unimplementederror, "MPI_Type_create_resized");
  return MPI_SUCCESS;
}

/**
 @param count [in] number of blocks (integer) --- also number of entries in arrays array_of_types, array_of_displacements and array_of
 @param blocklengths array_of
 @param blocklength [in] number of elements in each block (array of integer) _array_of
 @param displacements [in] byte displacement of each block (array of integer) _array_of
 @param types [in] type of elements in each block (array of handles to datatype objects)
 @param newtype [out] new datatype (handle)
 */
extern "C" int MPI_Type_create_struct(int count, int _array_of_blocklengths_[],
    MPI_Aint _array_of_displacements_[], MPI_Datatype _array_of_types_[],
    MPI_Datatype *newtype)
{
  mpiapi::ptr mpi = get_current_mpi();
  std::vector<int> cxx_blocklens(count);
  std::vector<mpitype> cxx_old_types(count);
  std::vector<int> cxx_ind(count);
  for (int i = 0; i < count; ++i)
  {
    cxx_blocklens[i] = _array_of_blocklengths_[i];
    cxx_ind[i] = _array_of_displacements_[i];
    cxx_old_types[i] = get_mpi_type(_array_of_types_[i]);
  }
  mpitype cxx_new;
  mpi->type_struct(count, cxx_blocklens, cxx_ind, cxx_old_types, cxx_new);
  *newtype = cxx_new.id;
  return MPI_SUCCESS; //yes, error handling would be nice
}

/**
 @param ndims [in] number of array dimensions (positive integer) _array_of
 @param sizes [in] number of elements of type oldtype in each dimension of the full array (array of positive integers) _array_of
 @param subsizes [in] number of elements of type oldtype in each dimension of the subarray (array of positive integers) _array_of
 @param starts [in] starting coordinates of the subarray in each dimension (array of nonnegative integers)
 @param order [in] array storage order flag (state)
 @param oldtype [in] array element datatype (handle)
 @param newtype [out] new datatype (handle)
 */
extern "C" int MPI_Type_create_subarray(int ndims, int _array_of_sizes_[],
    int _array_of_subsizes_[], int _array_of_starts_[], int order,
    MPI_Datatype oldtype, MPI_Datatype *newtype)
{
  sst_throw(unimplementederror, "MPI_Type_create_subarray");
  return MPI_SUCCESS;
}

/**
 @param datatype [in] datatype to get information on (handle)
 @param lb [out] lower bound of datatype (integer)
 @param extent [out] extent of datatype (integer)
 */
extern "C" int MPI_Type_get_extent(MPI_Datatype datatype, MPI_Aint *lb,
    MPI_Aint *extent)
{
  static const char* here = "MPI_Type_get_extent";
  SSTMAC_DEBUG << here << "\n";
  //for now assume no padding and that type=size
  mpitype ty = get_mpi_type(datatype);
  *extent = ty.get_extent();
  *lb = ty.get_lb();
  return MPI_SUCCESS;
}

/**
 @param datatype [in] datatype to get information on (handle) _true
 @param lb [out] true lower bound of datatype (integer) _true
 @param extent [out] true size of datatype (integer)
 */
extern "C" int MPI_Type_get_true_extent(MPI_Datatype datatype,
    MPI_Aint *true_lb, MPI_Aint *true_extent)
{
  static const char* here = "MPI_Type_get_true_extent";
  SSTMAC_DEBUG << here << "\n";
  //for now assume no padding and that type=size
  mpitype ty = get_mpi_type(datatype);
  *true_lb = ty.get_true_lb();
  *true_extent = ty.get_true_extent();
  return MPI_SUCCESS;
}

/**
 @param datarep [in] data representation (string)
 @param inbuf [in] input buffer start (choice)
 @param insize [in] input buffer size, in bytes (integer)
 @param outcount [in] number of output data items (integer)
 @param datatype [in] datatype of output data item (handle)
 @param position [in/out] current position in buffer, in bytes (integer)
 @param outbuf [out] output buffer start (choice)
 */
extern "C" int MPI_Unpack_external(char *datarep, void *inbuf, MPI_Aint insize,
    MPI_Aint *position, void *outbuf, int outcount, MPI_Datatype datatype)
{
  sst_throw(unimplementederror, "MPI_Unpack_external");
  return MPI_SUCCESS;
}

/**
 @param function [in] user defined error handling procedure (function)
 @param errhandler [out] MPI error handler (handle)
 */
extern "C" int MPI_Win_create_errhandler(MPI_Win_errhandler_function *function,
    MPI_Errhandler *errhandler)
{
  sst_throw(unimplementederror, "MPI_Win_create_errhandler");
  return MPI_SUCCESS;
}

/**
 @param win [in] window (handle)
 @param errhandler [out] error handler currently associated with window (handle)
 */
extern "C" int MPI_Win_get_errhandler(MPI_Win win, MPI_Errhandler *errhandler)
{
  //sst_throw(unimplementederror, "MPI_Win_get_errhandler");
  *errhandler = win_err_handlers[win];
  return MPI_SUCCESS;
}

/**
 @param win [in] window (handle)
 @param errhandler [in] new error handler for window (handle)
 */
extern "C" int MPI_Win_set_errhandler(MPI_Win win, MPI_Errhandler errhandler)
{
  //sst_throw(unimplementederror, "MPI_Win_set_errhandler");
  win_err_handlers[win] = errhandler;
  return MPI_SUCCESS;
}

/**
 @param origin_addr [in] Address of the buffer in which to receive the data
 @param origin_count [in] number of entries in origin buffer (nonnegative integer)
 @param origin_datatype [in] datatype of each entry in origin buffer (handle)
 @param target_rank [in] rank of target (nonnegative integer)
 @param target_disp [in] displacement from window start to the beginning of the target buffer (nonnegative integer)
 @param target_count [in] number of entries in target buffer (nonnegative integer)
 @param target_datatype [in] datatype of each entry in target buffer (handle)
 @param win [in] window object used for communication (handle)
 */
extern "C" int MPI_Get(void *origin_addr, int origin_count,
    MPI_Datatype origin_datatype, int target_rank, MPI_Aint target_disp,
    int target_count, MPI_Datatype target_datatype, MPI_Win win)
{
  static const char* here = "MPI_Get";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpi_window::ptr winptr = get_window(rank, win);
  sstmac::payload::const_ptr load;
  mpitype otype = get_mpi_type(origin_datatype);
  mpitype ttype = get_mpi_type(target_datatype);
  mpiid trank(target_rank);
  mpirequest_t req;
  win_rma_info info;
  info.buffer = origin_addr;
  info.count = origin_count;
  info.datatype = origin_datatype;
  info.win = winptr;
  info.type = win_rma_info::get;

  mpi->win_get(req, origin_count, otype, trank, target_disp, target_count,
      ttype, winptr);

  pending_rmas[req] = info;

  return MPI_SUCCESS;
}

/**
 @param origin_addr [in] initial address of origin buffer (choice)
 @param origin_count [in] number of entries in origin buffer (nonnegative integer)
 @param origin_datatype [in] datatype of each entry in origin buffer (handle)
 @param target_rank [in] rank of target (nonnegative integer)
 @param target_disp [in] displacement from start of window to target buffer (nonnegative integer)
 @param target_count [in] number of entries in target buffer (nonnegative integer)
 @param target_datatype [in] datatype of each entry in target buffer (handle)
 @param win [in] window object used for communication (handle)
 */
extern "C" int MPI_Put(void *origin_addr, int origin_count,
    MPI_Datatype origin_datatype, int target_rank, MPI_Aint target_disp,
    int target_count, MPI_Datatype target_datatype, MPI_Win win)
{
  static const char* here = "MPI_Put";
  SSTMAC_DEBUG << here << "\n";
  mpiapi::ptr mpi = get_current_mpi();
  int rank = mpi->comm_world()->rank().id;
  mpi_window::ptr winptr = get_window(rank, win);
  sstmac::payload::const_ptr load = get_mpi_payload_func(origin_addr, 0,
      origin_count, origin_datatype, true);

  mpitype ttype = get_mpi_type(target_datatype);
  mpiid trank(target_rank);
  mpirequest_t req;
  win_rma_info info;

  info.buffer = origin_addr;
  info.count = origin_count;
  info.datatype = origin_datatype;
  info.win = winptr;
  info.type = win_rma_info::put;

  mpi->win_put(req, load, target_count, ttype, trank, target_disp, winptr);

  pending_rmas[req] = info;

  return MPI_SUCCESS;
}

/**
 @param sendbuf starting address of send buffer (choice)
 @param recvcount element count per block (non-negative integer)
 @param datatype data type of elements of input buffer (handle)
 @param op operation (handle)
 @param comm communicator (handle)
 @param recvbuf starting address of receive buffer (choice)
 */
extern "C" int MPI_Reduce_scatter_block(void *sendbuf, void *recvbuf,
    int recvcount, MPI_Datatype datatype, MPI_Op op, MPI_Comm comm)
{
  sst_throw(unimplementederror, "MPI_Reduce_scatter_block");
  return MPI_SUCCESS;
}

/**
 @param comm_old input communicator (handle)
 @param indegree size of sources and sourceweights arrays (non-negative integer)
 @param sources ranks of processes for which the calling process is a destination (array of non-negative integers)
 @param sourceweights weights of the edges into the calling process (array of non-negative integers or MPI_UNWEIGHTED)
 @param outdegree size of destinations and destweights arrays (non-negative integer)
 @param destinations ranks of processes for which the calling process is a source (array of non-negative integers)
 @param destweights weights of the edges out of the calling process (array of non-negative integers or MPI_UNWEIGHTED)
 @param info hints on optimization and interpretation of weights (handle)
 @param reorder the ranks may be reordered (true) or not (false) (logical)
 @param comm_dist_graph communicator with distributed graph topology (handle)
 */
int MPI_Dist_graph_create_adjacent(MPI_Comm comm_old, int indegree,
    int sources[], int sourceweights[], int outdegree, int destinations[],
    int destweights[], MPI_Info info, int reorder, MPI_Comm *comm_dist_graph)
{
  sst_throw(unimplementederror, "MPI_Dist_graph_create_adjacent");
  return MPI_SUCCESS;
}

/**
 @param comm_old input communicator (handle)
 @param n number of source nodes for which this process specifies edges (non-negative integer)
 @param sources array containing the n source nodes for which this process specifies edges (array of non-negative integers)
 @param degrees array specifying the number of destinations for each source node in the source node array (array of non-negative integers)
 @param destinations destination nodes for the source nodes in the source node array (array of non-negative integers)
 @param weights weights for source to destination edges (array of non-negative integers or MPI_UNWEIGHTED)
 @param info hints on optimization and interpretation of weights (handle)
 @param reorder the process may be reordered (true) or not (false) (logical)
 @param comm_dist_graph communicator with distributed graph topology added (handle)
 */
int MPI_Dist_graph_create(MPI_Comm comm_old, int n, int sources[],
    int degrees[], int destinations[], int weights[], MPI_Info info,
    int reorder, MPI_Comm *comm_dist_graph)
{
  sst_throw(unimplementederror, "MPI_Dist_graph_create_adjacent");
  return MPI_SUCCESS;
}

/**
 @param comm communicator with distributed graph topology (handle)
 @param indegree number of edges into this process (non-negative integer)
 @param outdegree number of edges out of this process (non-negative integer)
 @param weighted false if MPI_UNWEIGHTED was supplied during creation, true otherwise (logical)
 */
int MPI_Dist_graph_neighbors_count(MPI_Comm comm, int *indegree,
    int *outdegree, int *weighted)
{
  sst_throw(unimplementederror, "MPI_Dist_graph_neighbors_count");
  return MPI_SUCCESS;
}

/**
 @param comm communicator with distributed graph topology (handle)
 @param maxindegree size of sources and sourceweights arrays (non-negative integer)
 @param maxoutdegree size of destinations and destweights arrays (non-negative integer)
 @param sources processes for which the calling process is a destination (array of non-negative integers)
 @param sourceweights weights of the edges into the calling process (array of non-negative integers)
 @param destinations processes for which the calling process is a source (array of non-negative integers)
 @param destweights weights of the edges out of the calling process (array of non-negative integers)
 */
int MPI_Dist_graph_neighbors(MPI_Comm comm, int maxindegree, int sources[],
    int sourceweights[], int maxoutdegree, int destinations[],
    int destweights[])
{
  sst_throw(unimplementederror, "MPI_Dist_graph_neighbors");
  return MPI_SUCCESS;
}

