/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */
#ifndef SSTMAC_SHMEM_heap_H_
#define SSTMAC_SHMEM_heap_H_

#include <sstmac/common/ptr_type.h>
#include "shmem/memory/symmetric.h"
#include "shmem/memory/global.h"
#include <sstmac/common/debug.h>

#include "shmem/memory/wait.h"

namespace sstmac
{
  namespace sw
  {
    namespace shmem
    {
      class heap : public ptr_type
      {
        //                   local raw --> object
        typedef boost::unordered_map<void*, symmetric::ptr> rawmaptype;
        rawmaptype mems_;

        //this one is for remote accessing
        //                  addr num --> object
        typedef boost::unordered_map<long, symmetric::ptr> addrmaptype;
        addrmaptype addrmap_;
        heap();

        long addr_count_;

        static debug dbg_;

        boost::unordered_map<void*, wait_base::ptr> waiting_;

      public:
        typedef boost::intrusive_ptr<heap> ptr;

        static ptr
        construct()
        {
          return ptr(new heap());
        }

        virtual
        ~heap()
        {
        }

        virtual std::string
        to_string() const
        {
          return "heap";
        }

        symmetric::ptr
        shmalloc(size_t size, bool real = true);

        symmetric::ptr
        shmemalign(size_t size, size_t align, bool real = true);

        symmetric::ptr
        shrealloc(void* raw, size_t sz);

        void
        add_global(global_constructed* gptr)
        {
          void* thisraw = gptr->get_addr();
          SSTMAC_DEBUG << "shmem::heap::add_global() - ptr=" << (long) gptr
                << ", raw=" << (long) thisraw << "\n";

          symmetric::ptr sym = symmetric::construct(gptr->get_size(),
              addr_count_, thisraw);
          //global<T> g = (*gptr);

          mems_[thisraw] = sym;

          addrmap_[addr_count_] = sym;

          addr_count_++;
        }

        void
        shfree(void* addr);

        void
        free_by_addr(long addr);

        symmetric::ptr
        get_symmetric_by_addr(long addr);

        /**
         * Returns true if the given address is in this heap
         */
        bool
        find_symmetric_by_addr(long addr)
        {
          return addrmap_.find(addr) != addrmap_.end();
        }

        symmetric::ptr
        get_symmetric_by_raw(void* addr);

        void
        add_wait(wait_base::ptr w);

        bool
        check_wait(void* base);

        /**
         * looks up the symmetric offset.
         * @param raw
         * @param addr
         * @param off
         * @param throw_on_err
         * @return 1 if ok,0 if addr could not be found
         */
        template<typename T>
          bool
          get_addr_off(const T* raw, long *addr, int *off,
              bool throw_on_err = true)
          {
            SSTMAC_DEBUG << "shmem::heap::get_addr_off(" << (long) raw
                  << ") \n";
            if (raw == NULL)
            {
              *addr = -1;
              *off = -1;
              return true;
            }
            rawmaptype::iterator it, end = mems_.end();
            size_t r2 = (long) raw;
            for (it = mems_.begin(); it != end; it++)
            {
              SSTMAC_DEBUG2 << "have raw addr " << (long) it->first << "\n";
              long r1 = (long) it->first;

              SSTMAC_DEBUG2 << "raw casted to " << r2 << "\n";
              long upper = (long) ((char*) it->first + it->second->get_size());
              SSTMAC_DEBUG2 << "upper bound is " << upper << "\n";
              if (r1 <= r2 && upper > r2)
              {
                *addr = it->second->get_shmem_addr();
                T* cast = (T*) it->first;
                *off = (int) (raw - cast);
                return true;
              }
            }
            //std::cout << "raw: " << (long)raw << "\n";
            if (throw_on_err)
              sst_throw(ssterror, "shmem::heap::get_addr_off - raw %l not found", r2);
            return false;
          }

        /**
         * Returns the logical address if found, -1 if not
         * @param raw
         */
        long
        find_addr(void* raw)
        {
          SSTMAC_DEBUG << "shmem::heap::find_addr(" << (long) raw << ") \n";
          if (raw == NULL)
          {
            return -1;
          }

          rawmaptype::iterator it, end = mems_.end();
          size_t r2 = (long) raw;
          for (it = mems_.begin(); it != end; it++)
          {
            SSTMAC_DEBUG2 << "have raw addr " << (long) it->first << "\n";
            long r1 = (long) it->first;

            SSTMAC_DEBUG2 << "raw casted to " << r2 << "\n";
            long upper = (long) ((char*) it->first + it->second->get_size());
            SSTMAC_DEBUG2 << "upper bound is " << upper << "\n";
            if (r1 <= r2 && upper > r2)
            {
              return it->second->get_shmem_addr();

            }
          }

          return -1;
        }
      };

    }
  }
}
#endif /* heap_H_ */
