/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */
#ifndef SSTMAC_SHMEM_MESSAGES_ATOMIC_H_
#define SSTMAC_SHMEM_MESSAGES_ATOMIC_H_

#include <sstmac/software/api/openshmem/shmem/transfer_api.h>

namespace sstmac
{
  namespace sw
  {
    namespace shmem
    {
      class atomic : public ptr_type
      {
      public:
        enum ATOMIC_T
        {
          SWAP, CSWAP, FINC, FADD
        };

      protected:
        atomic(ATOMIC_T t) :
          type_(t)
        {

        }

        ATOMIC_T type_;

      public:
        typedef boost::intrusive_ptr<atomic> ptr;

        virtual
        ~atomic()
        {
        }

        virtual std::string
        to_string() const
        {
          return "atomic";
        }

        virtual ATOMIC_T
        get_atomic_type() const
        {
          return type_;
        }

        virtual void
        execute(transfer_api::ptr a, long addr, int offset) = 0;

        virtual long
        size_bytes() const = 0;
      };

      template<typename T>
        class swap : public atomic
        {
        protected:
          swap(T val) :
            atomic(atomic::SWAP), val_(val)
          {

          }

          T val_;

        public:
          typedef boost::intrusive_ptr<swap> ptr;

          static ptr
          construct(T val)
          {
            return ptr(new swap(val));
          }

          virtual
          ~swap()
          {
          }

          virtual long
          size_bytes() const
          {
            return sizeof(T);
          }

          virtual void
          execute(transfer_api::ptr a, long addr, int offset)
          {
            symmetric::ptr s = a->get_symmetric_by_addr(addr);

            void* r = s->get_raw();

            T* casted = (T*) r;
            casted = &casted[offset];
            *casted = val_;

            a->check_wait(casted);
          }
        };

      template<typename T>
        class cswap : public atomic
        {
        protected:
          cswap(T val, T cond) :
            atomic(atomic::CSWAP), val_(val), cond_(cond)
          {

          }

          T val_;
          T cond_;

        public:
          typedef boost::intrusive_ptr<cswap> ptr;

          static ptr
          construct(T val, T cond)
          {
            return ptr(new cswap(val, cond));
          }

          virtual
          ~cswap()
          {
          }

          virtual long
          size_bytes() const
          {
            return 2 * sizeof(T);
          }

          virtual void
          execute(transfer_api::ptr a, long addr, int offset)
          {
            symmetric::ptr s = a->get_symmetric_by_addr(addr);

            void* r = s->get_raw();

            T* casted = (T*) r;
            casted = &casted[offset];
            if (*casted == cond_)
            {
              *casted = val_;
              a->check_wait(casted);
            }

          }
        };

      template<typename T>
        class fadd : public atomic
        {
        protected:
          fadd(T val) :
            atomic(atomic::FADD), val_(val)
          {

          }

          T val_;

        public:
          typedef boost::intrusive_ptr<fadd> ptr;

          static ptr
          construct(T val)
          {
            return ptr(new fadd(val));
          }

          virtual
          ~fadd()
          {
          }

          virtual long
          size_bytes() const
          {
            return sizeof(T);
          }

          virtual void
          execute(transfer_api::ptr a, long addr, int offset)
          {
            symmetric::ptr s = a->get_symmetric_by_addr(addr);

            void* r = s->get_raw();

            T* casted = (T*) r;
            casted = &casted[offset];
            *casted = *casted + val_;

            a->check_wait(casted);
          }
        };

      template<typename T>
        class finc : public atomic
        {
        protected:
          finc() :
            atomic(atomic::FINC)
          {

          }

        public:
          typedef boost::intrusive_ptr<finc> ptr;

          static ptr
          construct()
          {
            return ptr(new finc());
          }

          virtual
          ~finc()
          {
          }

          virtual long
          size_bytes() const
          {
            return 0;
          }

          virtual void
          execute(transfer_api::ptr a, long addr, int offset)
          {
            symmetric::ptr s = a->get_symmetric_by_addr(addr);

            void* r = s->get_raw();

            T* casted = (T*) r;
            casted = &casted[offset];
            *casted = *casted + 1;

            a->check_wait(casted);
          }
        };
    }
  }
}

#endif
