/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */
#ifndef SSTMAC_SHMEM_GET_TRANSFER_H_
#define SSTMAC_SHMEM_GET_TRANSFER_H_

#include "shmem/messages/transfer.h"
#include <sstmac/common/messages/vectorpayload.h>
#include "shmem/messages/fake_data.h"

namespace sstmac
{
  namespace sw
  {
    namespace shmem
    {

      template<typename T>
        class get_transfer : public transfer
        {

        protected:

          size_t size_;
          long addr_;
          int offset_;
          long from_pe_;
          long to_pe_;
          ptrdiff_t stride_;

          get_transfer(long addr, int offset, long frompe, long tope, ptrdiff_t stride) :
            addr_(addr), offset_(offset), size_(1), from_pe_(frompe),
                to_pe_(tope), stride_(stride)
          {

          }

          get_transfer(long addr, int offset, size_t size, long frompe,
              long tope, ptrdiff_t stride) :
            addr_(addr), offset_(offset), size_(size), from_pe_(frompe),
                to_pe_(tope), stride_(stride)
          {

          }

        public:
          typedef boost::intrusive_ptr<get_transfer> ptr;

          static ptr
          construct(long addr, int offset, long frompe, long tope, int st)
          {
            return ptr(new get_transfer(addr, offset, frompe, tope, st));
          }

          static ptr
          construct(long addr, int offset, size_t size, long frompe, long tope, int st)
          {
            return ptr(new get_transfer(addr, offset, size, frompe, tope, st));
          }

          virtual std::string
          to_string() const
          {
            return "get_transfer";
          }

          virtual
          ~get_transfer()
          {

          }

          virtual long
          size_bytes() const
          {
            return 0;
          }

          virtual long
          get_remote_pe() const
          {
            return to_pe_;
          }

          virtual long
          get_local_pe() const
          {
            return from_pe_;
          }

          virtual TRANSFER_TYPES
          get_trans_type() const
          {
            return GET;
          }

          virtual void
          execute(transfer_api::ptr a) const
          {
            payload::const_ptr ret;
            if (addr_ >= 0)
            {
              symmetric::ptr s = a->get_symmetric_by_addr(addr_);

              void* r = s->get_raw();

              SSTMAC_DEBUG << "get_transfer: executing on addr " << addr_
                    << " which is raw " << (long) r << " with offset "
                    << offset_ << "\n";
              T* casted = (T*) r;
              //casted += offset_;
              ret = vector1payload<T, T*>::construct(&casted[offset_], size_, stride_);
            }
            else
            {
              ret = fake_data::construct(sizeof(T) * size_);
            }

            a->send_get_response(const_ptr(this), ret);
          }

        };

    }
  }
}
#endif
