/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */
#ifndef SSTMAC_SHMEM_OP_TRANSFER_H_
#define SSTMAC_SHMEM_OP_TRANSFER_H_

#include "shmem/messages/transfer.h"
#include <sstmac/common/messages/vectorpayload.h>

namespace sstmac
{
  namespace sw
  {
    namespace shmem
    {

      template<typename T>
        class op_transfer : public transfer
        {

        protected:
          payload::const_ptr val_;

          size_t size_;
          long me_;
          long targaddr_;
          int targoffset_;

          //   long pe_start_;
          //  int log_stride_;
          int pe_size_;

          long tope_;

          symmetric::OPS op_;

          op_transfer(long targ_addr, int targ_offset, payload::const_ptr val,
              size_t size, long tope, long me, symmetric::OPS op, int pe_size) :
            targaddr_(targ_addr), targoffset_(targ_offset), val_(val),
                size_(size), me_(me), op_(op), tope_(tope), pe_size_(pe_size)
          {

          }

        public:
          typedef boost::intrusive_ptr<op_transfer> ptr;

          static ptr
          construct(long targ_addr, int targ_offset, payload::const_ptr val,
              size_t size, long tope, long me, symmetric::OPS op, int pe_size)
          {
            return ptr(
                new op_transfer(targ_addr, targ_offset, val, size, tope, me,
                    op, pe_size));
          }

          virtual std::string
          to_string() const
          {
            return "op_transfer";
          }

          virtual
          ~op_transfer()
          {

          }

          virtual long
          size_bytes() const
          {
            return val_->get_byte_length();
          }

          virtual long
          get_remote_pe() const
          {
            return tope_;
          }

          virtual long
          get_local_pe() const
          {
            return me_;
          }

          virtual TRANSFER_TYPES
          get_trans_type() const
          {
            return ATOMIC;
          }

          virtual void
          execute(transfer_api::ptr a) const
          {
            SSTMAC_DEBUG << "op_transfer::execute ...";
            symmetric::ptr s;
            if (targoffset_ >= 0)
            {
              SSTMAC_DEBUG << " with addr " << targaddr_ << " and offset "
                    << targoffset_ << "\n";

              s = a->get_symmetric_by_addr(targaddr_);
              boost::intrusive_ptr<const vector1payload<T, T*> > cast =
                  boost::dynamic_pointer_cast<const vector1payload<T, T*> >(
                      val_);
              s->perform_math_op<T> (cast->data(), targoffset_, size_, op_);

            }
            else
            {
              SSTMAC_DEBUG << " on fake\n";
              s = a->get_fake_sym(size_);
              s->perform_fake_op(size_);
            }

            if (s->get_op_counter() == pe_size_)
            {
              SSTMAC_DEBUG << "done with op, unblocking \n";
              s->reset_op_counter();
              a->unblock();
            }
          }

        };

      template<typename T>
        class logical_op_transfer : public op_transfer<T>
        {

        protected:

          logical_op_transfer(long targ_addr, int targ_offset,
              payload::const_ptr val, size_t size, long tope, long me,
              symmetric::OPS op, int pe_size) :
                op_transfer<T> (targ_addr, targ_offset, val, size, tope, me,
                    op, pe_size)
          {

          }

        public:
          typedef boost::intrusive_ptr<logical_op_transfer> ptr;

          static ptr
          construct(long targ_addr, int targ_offset, payload::const_ptr val,
              size_t size, long tope, long me, symmetric::OPS op, int pe_size)
          {
            return ptr(
                new logical_op_transfer(targ_addr, targ_offset, val, size,
                    tope, me, op, pe_size));
          }

          virtual std::string
          to_string() const
          {
            return "logical_op_transfer";
          }

          virtual
          ~logical_op_transfer()
          {

          }

          virtual void
          execute(transfer_api::ptr a) const
          {
            //SSTMAC_DEBUG << "logical_op_transfer::execute ...";
            symmetric::ptr s;
            if (op_transfer<T>::targoffset_ >= 0)
            {
              //  SSTMAC_DEBUG << " with addr " << targaddr_ << " and offset "
              //        << targoffset_ << "\n";

              s = a->get_symmetric_by_addr(op_transfer<T>::targaddr_);
              boost::intrusive_ptr<const vector1payload<T, T*> > cast =
                  boost::dynamic_pointer_cast<const vector1payload<T, T*> >(
                      op_transfer<T>::val_);
              s->perform_logical_op<T> (cast->data(),
                  op_transfer<T>::targoffset_, op_transfer<T>::size_,
                  op_transfer<T>::op_);

            }
            else
            {
              //SSTMAC_DEBUG << " on fake\n";
              s = a->get_fake_sym(op_transfer<T>::size_);
              s->perform_fake_op(op_transfer<T>::size_);
            }

            if (s->get_op_counter() == op_transfer<T>::pe_size_)
            {
             // SSTMAC_DEBUG << "done with op, unblocking \n";
              s->reset_op_counter();
              a->unblock();
            }
          }

        };

    }
  }
}
#endif
