/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */
#ifndef SSTMAC_SHMEM_PUT_TRANSFER_H_
#define SSTMAC_SHMEM_PUT_TRANSFER_H_

#include "shmem/messages/transfer.h"

namespace sstmac
{
  namespace sw
  {
    namespace shmem
    {

      template<typename T>
        class put_transfer : public transfer
        {

        protected:

          const T* val_;
          size_t size_;
          long addr_;
          int offset_;
          long from_pe_;
          long to_pe_;
          long tstride_;

          T* alloc_;

          put_transfer(long addr, int offset, const T *val, size_t size,
              long frompe, long tope, ptrdiff_t tstride, ptrdiff_t sstride) :
            addr_(addr), offset_(offset), size_(size), from_pe_(frompe),
                to_pe_(tope), tstride_(tstride)
          {

            if (addr_ >= 0)
            {
              alloc_ = (T*) malloc(sizeof(T) * size);
              if (sstride == 1)
                memcpy(alloc_, val, sizeof(T) * size);
              else
              {
                for (size_t i = 0; i < size; i++)
                {
                  alloc_[i] = val[i * sstride];
                }
              }
              val_ = alloc_;
            }

          }

        public:
          typedef boost::intrusive_ptr<put_transfer> ptr;

          static ptr
          construct(long addr, int offset, T val, long frompe, long tope,
              ptrdiff_t tstride, ptrdiff_t sstride)
          {
            return ptr(
                new put_transfer(addr, offset, &val, 1, frompe, tope, tstride,
                    sstride));
          }

          static ptr
          construct(long addr, int offset, const T* val, size_t size,
              long frompe, long tope, ptrdiff_t tstride, ptrdiff_t sstride)
          {
            return ptr(
                new put_transfer(addr, offset, val, size, frompe, tope,
                    tstride, sstride));
          }

          virtual std::string
          to_string() const
          {
            return "put_transfer";
          }

          virtual
          ~put_transfer()
          {
            if (addr_ >= 0)
            {
              free(alloc_);
            }

          }

          virtual long
          size_bytes() const
          {
            return sizeof(T) * size_;
          }

          virtual long
          get_remote_pe() const
          {
            return to_pe_;
          }

          virtual long
          get_local_pe() const
          {
            return from_pe_;
          }

          virtual TRANSFER_TYPES
          get_trans_type() const
          {
            return PUT;
          }

          virtual void
          execute(transfer_api::ptr a) const
          {
            if (addr_ >= 0)
            {
              symmetric::ptr s = a->get_symmetric_by_addr(addr_);

              void* r = s->get_raw();

              SSTMAC_DEBUG << "put_transfer: executing on addr " << addr_
                    << " which is raw " << (long) r << " with offset "
                    << offset_ << "\n";
              T* casted = (T*) r;
              //casted += offset_;

              for (int i = 0; i < size_; i++)
              {
                SSTMAC_DEBUG << "setting val: " << val_[i] << " to "
                      << (long) (&casted[offset_ + i]) << "\n";
                casted[offset_ + i*tstride_] = val_[i];
                a->check_wait((void*)(&casted[offset_ + i*tstride_]));
              }



              if (dbg_.is_active())
              {
                SSTMAC_DEBUG << "--resulting symmetric variable of size "
                      << s->get_size() << " is: \n";
                for (int i = 0; i < s->get_size() / sizeof(T); i++)
                {
                  SSTMAC_DEBUG << "[" << i << "]: " << casted[i] << "\n";
                }
              }
            }

            a->send_put_complete(const_ptr(this));
          }

        };

    }
  }
}
#endif
