/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/software/api/pthread/pthread_impl.h>
#include <sstmac/software/process/thread.h>
#include <sstmac/software/process/app.h>
#include <sstmac/software/process/os_api.h>
#include <sstmac/software/api/pthread/pthread_runner.h>

using namespace sstmac;
using namespace sstmac::sw;

//               process context, ....
static nested_type<long, sstmac_pthread_t, thread::ptr>::unordered_map
    threadmap_;
static nested_type<long, thread::ptr, sstmac_pthread_t>::unordered_map
    mapthread_;
static boost::unordered_map<long, int> threadcounters_;

thread::ptr
get_current_thread()
{
  thread::ptr t = OS_API::get_current_thread();

  if (!t)
  {
    throw sstmac::ssterror(
        "sstcpp::get_current_thread - OS_API didn't return a valid thread, which is very strange indeed");
  }

  return t;
}

OS_API::ptr
get_current_os()
{
  thread::ptr t = get_current_thread();
  return t->get_os();
}

thread::ptr
get_thread(long ptxt, sstmac_pthread_t t)
{
  if (threadmap_[ptxt].find(t) == threadmap_[ptxt].end())
  {
    sst_throw(ssterror, "pthread::get_thread - can't find thread %d on process context %d", t, ptxt);
  }
  return threadmap_[ptxt][t];
}

int
get_next_threadid(long ptxt)
{
  if (threadcounters_.find(ptxt) == threadcounters_.end())
  {
    threadcounters_[ptxt] = 1;
  }
  return threadcounters_[ptxt]++;
}

/**
 * Notes:
 *   - attr currently not used
 */
extern "C" int
SSTMAC_pthread_create(sstmac_pthread_t * thread,
    const sstmac_pthread_attr_t * attr, void *
    (*start_routine)(void *), void * arg)
{
  thread::ptr current = get_current_thread();
  OS_API::ptr os = get_current_os();
  long ptxt = current->get_process_context()->get_key();
  threadid thid(get_next_threadid(ptxt));
  //std::cout << "task " << current->tid() << " creating pthread \n";
  softwareid newid(current->aid(), current->tid(), thid);
  pthread_runner::ptr tr = pthread_runner::construct(newid,
      current->get_parent_app(),
      start_routine, arg);
  threadmap_[ptxt][thid.id] = tr;
  mapthread_[ptxt][tr] = thid.id;
  *thread = thid.id;

  os->start_thread(tr);

  return 0;
}

extern "C" void
SSTMAC_pthread_exit(void *retval)
{
  thread::ptr current = get_current_thread();
  current->kill();

}

/**
 * Notes:
 *      - status currently not set, don't use it
 */
extern "C" int
SSTMAC_pthread_join(sstmac_pthread_t thread, void ** status)
{
  thread::ptr current = get_current_thread();
  OS_API::ptr os = get_current_os();
  long ptxt = current->get_process_context()->get_key();
  thread::ptr joiner = get_thread(ptxt, thread);
  os->join_thread(joiner);

  return 0;
}

extern "C" sstmac_pthread_t
SSTMAC_pthread_self()
{
  thread::ptr t = get_current_thread();
  long ptxt = t->get_process_context()->get_key();
  if (mapthread_[ptxt].find(t) == mapthread_[ptxt].end())
  {
    sst_throw(ssterror, "pthread::pthread_self - can't find self in mapthread, make sure you're calling this from a pthread");
  }
  return mapthread_[ptxt][t];
}

extern "C" int
SSTMAC_pthread_equal(sstmac_pthread_t thread_1, sstmac_pthread_t thread_2)
{
  sst_throw(unimplementederror, "pthread::pthread_equal not implemented");
}

extern "C" int
SSTMAC_pthread_mutex_init(sstmac_pthread_mutex_t * mutex,
    const sstmac_pthread_mutex_attr *attr)
{
  sst_throw(unimplementederror, "pthread::pthread_mutex_init not implemented");
}

extern "C" int
SSTMAC_pthread_mutex_destroy(sstmac_pthread_mutex_t * mutex)
{
  sst_throw(unimplementederror, "pthread::pthread_mutex_destroy not implemented");
}

extern "C" int
SSTMAC_pthread_mutex_lock(sstmac_pthread_mutex_t * mutex)
{
  sst_throw(unimplementederror, "pthread::pthread_mutex_lock not implemented");
}

extern "C" int
SSTMAC_pthread_mutex_trylock(sstmac_pthread_mutex_t * mutex)
{
  sst_throw(unimplementederror, "pthread::pthread_mutex_trylock not implemented");
}

extern "C" int
SSTMAC_pthread_mutex_unlock(sstmac_pthread_mutex_t * mutex)
{
  sst_throw(unimplementederror, "pthread::pthread_mutex_unlock not implemented");
}

extern "C" int
SSTMAC_pthread_cond_init(sstmac_pthread_cond_t * cond,
    const sstmac_pthread_cond_attr *attr)
{
  sst_throw(unimplementederror, "pthread::pthread_cond_init not implemented");
}

extern "C" int
SSTMAC_pthread_cond_destroy(sstmac_pthread_cond_t * cond)
{
  sst_throw(unimplementederror, "pthread::pthread_cond_destroy not implemented");
}

extern "C" int
SSTMAC_pthread_cond_wait(sstmac_pthread_cond_t * cond,
    sstmac_pthread_mutex_t * mutex)
{
  sst_throw(unimplementederror, "pthread::pthread_cond_wait not implemented");
}

extern "C" int
SSTMAC_pthread_cond_timedwait(sstmac_pthread_cond_t * cond,
    sstmac_pthread_mutex_t * mutex, const struct timespec * abstime)
{
  sst_throw(unimplementederror, "pthread::pthread_cond_timedwait not implemented");
}

extern "C" int
SSTMAC_pthread_cond_signal(sstmac_pthread_cond_t * cond)
{
  sst_throw(unimplementederror, "pthread::pthread_cond_signal not implemented");
}

extern "C" int
SSTMAC_pthread_cond_broadcast(sstmac_pthread_cond_t * cond)
{
  sst_throw(unimplementederror, "pthread::pthread_cond_broadcast not implemented");
}

extern "C" int
SSTMAC_pthread_once(sstmac_pthread_once_t * once_init, void
(*init_routine)(void))
{
  sst_throw(unimplementederror, "pthread::pthread_once not implemented");
}

extern "C" int
SSTMAC_pthread_key_create(sstmac_pthread_key_t * key, void
(* dest_routine)(void *))
{
  sst_throw(unimplementederror, "pthread::pthread_key_create not implemented");
}

extern "C" int
SSTMAC_pthread_key_delete(sstmac_pthread_key_t key)
{
  sst_throw(unimplementederror, "pthread::pthread_key_delete not implemented");
}

extern "C" int
SSTMAC_pthread_setspecific(sstmac_pthread_key_t key, const void * pointer)
{
  sst_throw(unimplementederror, "pthread::pthread_setspecific not implemented");
}

extern "C" void *
SSTMAC_pthread_getspecific(sstmac_pthread_key_t key)
{
  sst_throw(unimplementederror, "pthread::pthread_equal not implemented");
}

extern "C" void
SSTMAC_pthread_cleanup_push(void
(*routine)(void *), void *routine_arg)
{
  sst_throw(unimplementederror, "pthread::pthread_cleanup_push not implemented");
}

extern "C" void
SSTMAC_pthread_cleanup_pop(int execute)
{
  sst_throw(unimplementederror, "pthread::pthread_cleanup_pop not implemented");
}

extern "C" int
SSTMAC_pthread_attr_init(sstmac_pthread_attr_t *attr)
{
  sst_throw(unimplementederror, "pthread::pthread_attr_init not implemented");
}

extern "C" int
SSTMAC_pthread_attr_destroy(sstmac_pthread_attr_t *attr)
{
  sst_throw(unimplementederror, "pthread::pthread_attr_destroy not implemented");
}

extern "C" int
SSTMAC_pthread_attr_getstacksize(sstmac_pthread_attr_t *attr,
    size_t * stacksize)
{
  sst_throw(unimplementederror, "pthread::pthread_attr_getstacksize not implemented");
}

extern "C" int
SSTMAC_pthread_attr_setstacksize(sstmac_pthread_attr_t *attr, size_t stacksize)
{
  sst_throw(unimplementederror, "pthread::pthread_attr_setstacksize not implemented");
}

extern "C" int
SSTMAC_pthread_attr_getdetachstate(const sstmac_pthread_attr_t *attr,
    int *state)
{
  sst_throw(unimplementederror, "pthread::pthread_getdetachstate not implemented");
}

extern "C" int
SSTMAC_pthread_attr_setdetachstate(sstmac_pthread_attr_t *attr, int state)
{
  sst_throw(unimplementederror, "pthread::pthread_setdetachstate not implemented");
}
