#include <sstmac/software/api/simpmsg/simple_message.h>

namespace simpmsg
{

const int simple_message::ack_size = 16;
const int simple_message::header_size = 64;

simple_message::simple_message(long num_bytes)
    : library_message("simpmsg"),
    sst_message(NETWORK),
    event_(null_event),
    num_bytes_(num_bytes)
{
}

#define enumcase(x) case x: return #x
const char*
simple_message::tostr(event_t ev)
{
    switch(ev)
    {
        enumcase(send);
        enumcase(rdma_get_start);
        enumcase(rdma_put_start);
        enumcase(rdma_put_ack);
        enumcase(rdma_get_ack);
        enumcase(msg_arrived);
        enumcase(null_event);
    }
}

void
simple_message::clone_into(const simple_message::ptr& cln) const
{
    cln->event_ = event_;
    cln->src_ = src_;
    cln->dest_ = dest_;
    cln->lib_name_ = lib_name_;
    cln->msg_id_ = msg_id_;
}

void
simple_message::clone_into(MESSAGE_TYPES ty, const simple_message::ptr& cln) const
{
    if (ty == sst_message::SEND_ACK){
        cln->set_needs_ack(false);
        cln->num_bytes_ = ack_size;
        //cln->reverse();
    }
    cln->msgtype_ = ty;
}

sstmac::sst_message::ptr
simple_message::clone(MESSAGE_TYPES ty) const
{
    simple_message::ptr cln = new simple_message(num_bytes_);
    clone_into(cln);
    clone_into(ty, cln);
    return cln;
}

void
simple_message::mutate(mutation_t mut)
{
      if (mut == sst_message::RDMA_GET_REQ_TO_RSP)
      {
        msgtype_ = RDMA_GET_RSP;
        reverse();
        num_bytes_ = abs(num_bytes_);
      }
      else {
        sst_throw(sstmac::valueerror,
            "invalid mutation type for simple message %d",
            mut);
      }
}

void
simple_message::reverse()
{
  //also flip the addresses
  sstmac::nodeaddress::ptr src = fromaddr_;
  sstmac::nodeaddress::ptr dst = toaddr_;
  toaddr_ = src;
  fromaddr_ = dst;

  sstmac::sw::taskid srcid = src_;
  sstmac::sw::taskid dstid = dest_;
  src_ = dstid;
  dest_ = srcid;
}

long
simple_message::source_thread(const sstmac::sw::operatingsystem::ptr& os) const
{
  return os->get_task_threadid(src_);
}

long
simple_message::dest_thread(const sstmac::sw::operatingsystem::ptr& os) const
{
  return os->get_task_threadid(dest_);
}

simpmsg_queue::ptr
simpmsg_queue::construct()
{
    return new simpmsg_queue(sstmac::sw::operatingsystem::get_current_os());
}

simpmsg_queue::ptr
simpmsg_queue::construct(const sstmac::sw::operatingsystem::ptr &os)
{
    return new simpmsg_queue(os);
}

simpmsg_queue::simpmsg_queue(const sstmac::sw::operatingsystem::ptr &os)
    : os_(os)
{
}

simple_message::ptr
simpmsg_queue::poll_until_message()
{
    if (pending_messages_.empty()){
        sstmac::sw::key::ptr blocker = sstmac::sw::key::construct();
        blocked_keys_.push_back(blocker);
        os_->block(blocker);
    }

    //we have been unblocked
    simple_message::ptr msg = pending_messages_.front();
    pending_messages_.pop_front();
    return msg;
}

void
simpmsg_queue::put_message(const simple_message::ptr &msg)
{
    pending_messages_.push_back(msg);

    if (!blocked_keys_.empty()){
        //unblock the first pending thread
        sstmac::sw::key::ptr next_key = blocked_keys_.front();
        blocked_keys_.pop_front();
        os_->unblock(next_key);
    }
}

}
