#include <sstmac/software/api/simpmsg/sstmac_simpmsg_api.h>

ImplementAPI(simpmsg, simpmsg_api, "simpmsg")

using namespace sstmac;
using namespace sstmac::sw;

namespace simpmsg
{
    SSTRegister("simpmsg", api, simpmsg_api,
        "Create bindings for simple message runtime");

simpmsg_api::simpmsg_api()
 : msg_id_ticker_(0)
{
}

void
simpmsg_api::init()
{
    env_ = os_->get_env(sid_.app_);
    nproc_ = env_->get_app_count();

    library::ptr server_lib = os_->get_lib(server_libname_);
    simpmsg_server::ptr server;

    // only do one server per app per node
    if (server_lib == library::ptr()){
        server = simpmsg_server::construct(sid_.app_.id);
        os_->register_ser(server);
        server->start();
    }
    else {
        server = safe_cast(server_lib, simpmsg_server);
    }
    server->register_proc(rank_, ptr(this));

    queue_ = simpmsg_queue::construct(fullos_);
}

void
simpmsg_api::init_os(const OS_API::ptr &os)
{
    process_manager::init_os(os);
    api::init_os(os);
}

void
simpmsg_api::finalize()
{
}

void
simpmsg_api::init(const sstmac::sw::softwareid &sid)
{
    sid_ = sid;
    rank_ = sid.task_.id;
    libname_ = sstprintf("simpmsg_api_%d_%d",
                         sid.app_.id, sid.task_.id);
    server_libname_ = sstprintf("simpmsg_server_%d", sid_.app_.id);
    process_manager::init(sid);
    thread::ptr thr = OS_API::get_current_thread();
    app::ptr my_app = safe_cast(thr, app);
    my_app->compute(timestamp(1e-6));
}

void
simpmsg_api::send(int dst, const simple_message::ptr &msg)
{
    msg->set_src(rank_);
    msg->set_type(sst_message::NETWORK);
    msg->toaddr_ = env_->get_node_address(dst);
    msg->set_dest(dst);
    msg->set_lib_name(server_libname_);
    msg->set_event(simple_message::send);
    msg->set_msg_id(msg_id_ticker_++);
    msg->set_needs_ack(false);
    /**
    std::cout << sstprintf("Sending %p type %s, id %ld from %d to %d:%s\n",
                            msg.get(),
                            simple_message::tostr(msg->event()),
                            msg->msg_id(),
                            rank_, dst,
                            msg->toaddr_->to_string().c_str());
    */
    if (dst == rank_){
        //deliver to self
        queue_->put_message(msg);
    }
    else {
        library::os_->execute_kernel(AMI::AMI_COMM_SEND, msg);
    }

}

void
simpmsg_api::rdma_put(int dst, const simple_message::ptr &msg)
{
    msg->set_src(rank_);
    msg->set_type(sst_message::RDMA_PUT_REQ);
    msg->toaddr_ = env_->get_node_address(dst);
    msg->set_dest(dst);
    msg->set_lib_name(server_libname_);
    msg->set_event(simple_message::rdma_put_start);
    msg->set_msg_id(msg_id_ticker_++);
    msg->set_needs_ack(true);
    pending_acks_[msg->msg_id()] = simple_message::rdma_put_start;
    /**
    std::cout << sstprintf("RDMA put %p type %s, %ld from %d to %d\n",
                            msg.get(),
                            simple_message::tostr(msg->event()),
                            msg->msg_id(),
                            rank_, dst);
    */
    library::os_->execute_kernel(AMI::AMI_COMM_SEND, msg);
}

void
simpmsg_api::rdma_get(int dst, const simple_message::ptr &msg)
{
    msg->set_src(rank_);
    msg->set_type(sst_message::RDMA_GET_REQ);
    msg->set_virtual_num_bytes();
    msg->toaddr_ = env_->get_node_address(dst);
    msg->set_dest(dst);
    msg->set_lib_name(server_libname_);
    msg->set_event(simple_message::rdma_get_start);
    msg->set_msg_id(msg_id_ticker_++);
    msg->set_needs_ack(false);
    /**
    std::cout << sstprintf("RDMA %p get type %s, %ld from %d to %d\n",
                            msg.get(),
                            simple_message::tostr(msg->event()),
                            msg->msg_id(),
                            rank_, dst);
    */
    library::os_->execute_kernel(AMI::AMI_COMM_SEND, msg);
}

void
simpmsg_api::handle(const simple_message::ptr &msg)
{
    /**
    std::cout << sstprintf("Handle %p %s:%s, id %ld from %d on %d\n",
                            msg.get(),
                            sst_message::tostr(msg->get_type()),
                            simple_message::tostr(msg->event()),
                            msg->msg_id(),
                            msg->src(), rank_);
    */
    sst_message::MESSAGE_TYPES ty = msg->get_type();
    if (ty == sst_message::SEND_ACK){
        ack_map::iterator it = pending_acks_.find(msg->msg_id());
        if (it != pending_acks_.end()){
            simple_message::event_t outgoing_event = it->second;
            switch(outgoing_event)
            {
                case simple_message::rdma_put_start:
                    msg->set_event(simple_message::rdma_put_ack);
                    break;
                default:
                    sst_throw(valueerror,
                        "invalid simpmsg outgoing event %d being acked",
                        outgoing_event);
            }
            pending_acks_.erase(it);
        }
    }
    queue_->put_message(msg);
}

simple_message::ptr
simpmsg_api::poll_until_notification()
{
    simple_message::ptr msg = queue_->poll_until_message();
    return msg;
}

simpmsg_server::simpmsg_server(int appid)
    : appid_(appid)
{
    libname_ = sstprintf("simpmsg_server_%d", appid);
}

void
simpmsg_server::handle(const sst_message::ptr &msg)
{
    simple_message::ptr smsg = safe_cast(msg, simple_message);
    if (msg->get_type() == sst_message::SEND_ACK){
        get_proc(smsg->src())->handle(smsg);
    }
    else {
        get_proc(smsg->dest())->handle(smsg);
    }

}

simpmsg_server::ptr
simpmsg_server::construct(int appid)
{
    return new simpmsg_server(appid);
}

service::ptr
simpmsg_server::clone()
{
    return new simpmsg_server(appid_);
}

simpmsg_api::ptr
simpmsg_server::get_proc(int rank) const
{
    boost::unordered_map<int, simpmsg_api::ptr>::const_iterator
        it = procs_.find(rank);
    if (it == procs_.end()){
        it = procs_.begin();
        std::cerr << "Valid ranks for server are:\n";
        while (it != procs_.end())
        {
            std::cerr << it->first << std::endl;
            ++it;
        }

        sst_throw(valueerror,
            "simpmsg_server::get_proc: invalid rank %d for server",
            rank);
    }
    return it->second;
}

void
simpmsg_server::register_proc(int rank, const simpmsg_api::ptr &proc)
{
    procs_[rank] = proc;
}

}
