#include <sstmac/software/api/sockets/lib_sockets.h>
#include <sstmac/software/process/operatingsystem.h>

#define enumcase(x) case x: return #x;

ImplementAPI(sstmac::sw, socketapi, "socket");

namespace sstmac {
    namespace sw {

    SSTRegister("socket", api, socketapi, "Create bindings for sockets");

    debug socketapi::dbg_;
    debug socket::dbg_;
    taskid socket_message::nic_task(-1);

    socket_message::socket_message(
        int number, long bytes,
        type_t ty,
        const std::string &libname,
        const nodeaddress::ptr &toaddr,
        const nodeaddress::ptr &fromaddr)
        : network_message(toaddr, fromaddr),
        library_message(libname),
        number_(number),
        nbytes_(bytes),
        type_(ty)
    {
    }

    long
    socket_message::nbytes() const
    {
        return nbytes_;
    }

    int
    socket_message::number() const
    {
        return number_;
    }

    long
    socket_message::get_byte_length() const
    {
        return nbytes_;
    }

    socket_message::type_t
    socket_message::reverse(type_t ty) const
    {
        switch(ty){
            case client:
                return server;
            case server:
                return client;
        }
    }

    sst_message::ptr
    socket_message::clone(MESSAGE_TYPES t) const
    {
        type_t new_type = t == SEND_ACK ? reverse(type_) : type_;
        socket_message::ptr newmsg
            = socket_message::construct(
                number_, nbytes_,
                new_type,
                lib_name_,
                toaddr_,
                fromaddr_);
        newmsg->msgtype_ = t;
        return newmsg;
    }

    socket_message::ptr
    socket_message::construct(
        int socket, long bytes,
        type_t ty,
        const std::string &libname,
        const nodeaddress::ptr &toaddr,
        const nodeaddress::ptr &fromaddr)
    {
        return new socket_message(socket, bytes, ty, libname, toaddr, fromaddr);
    }

    socket::socket(int fd, const nodeaddress::ptr &my_addr,
                   socket_domain_t domain,
                   socket_type_t typ,
                   socket_protocol_t protocol,
                   const operatingsystem::ptr& os)
        :  number_(unbound), fd_(fd),
          domain_(domain), protocol_(protocol),
          type_(typ), my_addr_(my_addr),
          key_(key::construct()), os_(os),
          state_(UNBLOCKED), bytes_available(0)
    {
        if (!dbg_.is_real()){
            dbg_ = debug("<debug> socket");
        }
    }

    int
    socket::number() const
    {
        return number_;
    }

    void
    socket::bind(int number, const nodeaddress::ptr& addr)
    {
        if (number_ != unbound){
            sst_throw(valueerror, "socket %d is already bound.  cannot bind again", number_);
        }
        number_ = number;
        remote_addr_ = addr;
    }

    void
    socket::disconnect()
    {
        remote_addr_ = 0;
        number_ = unbound;
    }

    void
    socket::connect(int number, const nodeaddress::ptr& addr)
    {
        if (remote_addr_){
            sst_throw(valueerror, "socket %d is already connected", number_);
        }
        remote_addr_ = addr;
        number_ = number;
    }

    void
    socket::unaccept()
    {
        bytes_available = 0;
        //remote_addr_ = 0;
        pending_recv_.clear();
    }

    void
    socket::shutdown()
    {
        number_ = unbound;
        bytes_available = 0;
        remote_addr_ = 0;
    }

    std::string
    socket::to_string() const
    {
        return sstprintf("socket fd=%d on node %s", number_, my_addr_->to_string().c_str());
    }

    void
    socket::add_pending(const socket_message::ptr& msg)
    {
        bytes_available += msg->nbytes();
        pending_recv_.push_back(msg);
    }

    void
    socket::accept(const nodeaddress::ptr& addr)
    {
        remote_addr_ = addr;
        if (!remote_addr_){
            sst_throw(illformederror, "socket::accept: null address");
        }
    }

    nodeaddress::ptr
    socket::send_addr() const
    {
        return remote_addr_;
    }

    nodeaddress::ptr
    socket::my_addr() const
    {
        return my_addr_;
    }

    long
    socket::recv(long max_bytes)
    {
        long bytes_recved = 0;
        std::list<socket_message::ptr>::iterator it = pending_recv_.begin(), end = pending_recv_.end();
        while (it != end){
            std::list<socket_message::ptr>::iterator tmp = it++;
            socket_message::ptr msg = *tmp;
            long next_bytes = bytes_recved + msg->nbytes();
            if (next_bytes > max_bytes){ //can't go over!
                return bytes_recved;
            }
            bytes_recved += msg->nbytes();
            pending_recv_.erase(tmp);
            bytes_available -= msg->nbytes();
        }
        return bytes_recved;
    }

    socket::state_t
    socket::state() const
    {
        return state_;
    }

    int
    socket::fd() const
    {
        return fd_;
    }

    void
    socket::rebind()
    {
        //remote_addr_ = 0;
    }

    nodeaddress::ptr
    socket::recv_addr() const
    {
        return remote_addr_;
    }

    void
    socket::block(state_t state)
    {
        state_ = state;
        os_->block(key_);
    }

    void
    socket::unblock()
    {
        state_ = UNBLOCKED;
        os_->unblock(key_);
    }

    const char*
    socket::tostr(state_t state)
    {
        switch (state){
            enumcase(UNBLOCKED);
            enumcase(BLOCK_ON_ACCEPT);
            enumcase(BLOCK_ON_SEND);
            enumcase(BLOCK_ON_RECV);
        }
        return "";
    }

    socketapi::socketapi()
        : api(key::general),
          next_free_socket_(0),
          in_blocking_send_(false)
    {
        if (!dbg_.is_real()){
            dbg_ = debug("<debug> socket | socketapi");
        }
        init_as_multithreaded_lib();
    }

    socketapi::~socketapi()
    {
    }

    socket::ptr
    socketapi::get_socket(int fd) const
    {
        std::map<int, socket::ptr>::const_iterator it = all_sockets_.find(fd);
        if (it == all_sockets_.end()){
            sst_throw(valueerror, "socket %d does not exist yet in socketapi::recv");
        }
        return it->second;
    }

    socketapi::ptr
    socketapi::construct(const sim_parameters::ptr& params)
    {
        return new socketapi();
    }

    socket::ptr
    socketapi::get_connected_socket(int number) const
    {
        std::map<int, socket::ptr>::const_iterator it = connected_sockets_.find(number);
        if (it == connected_sockets_.end()){
            return socket::ptr();
        }
        return it->second;
    }

    socket::ptr
    socketapi::get_bound_socket(int number) const
    {
        std::map<int, socket::ptr>::const_iterator it = bound_sockets_.find(number);
        if (it == bound_sockets_.end()){
            return socket::ptr();
        }
        return it->second;
    }

    long
    socketapi::recv(int fd, long max_bytes, bool blocking)
    {
        SSTMAC_DEBUG << sstprintf("Starting %s receive of %ld bytes on socket %d\n",
                                  blocking ? "blocking" : "non-blocking",
                                  max_bytes, fd);
        socket::ptr sock = get_socket(fd);
        if (blocking && sock->bytes_available == 0){ //gotta recv something!
            SSTMAC_DEBUG << sstprintf("Blocking on socket %d until %ld bytes are available.  Only have %ld\n",
                                      fd, max_bytes, sock->bytes_available);
            sock->block(socket::BLOCK_ON_RECV);
            SSTMAC_DEBUG << sstprintf("Unblocking on socket %d now that %ld bytes are available\n",
                                      fd, sock->bytes_available);
        }
        long bytes = std::min(sock->bytes_available, max_bytes);

        long dstid = fullos_->get_current_threadid();
        sock->rebind();
        lib_mem_->read(bytes, thread::nic_thread, dstid);
        return bytes;
    }

    long
    socketapi::send(int fd, long bytes, bool wait_on_ack)
    {
        SSTMAC_DEBUG << sstprintf("Starting send of %ld bytes on socket %d\n", bytes, fd);
        socket::ptr sock = get_socket(fd);

        bool from_server = bound_sockets_.find(sock->number()) != bound_sockets_.end();
        socket_message::type_t going_to = from_server ?
            socket_message::client : socket_message::server;

        socket_message::ptr msg = socket_message::construct(
                                    sock->number(), bytes,
                                    going_to,
                                    server_libname_,
                                    sock->send_addr(),
                                    sock->my_addr());
        msg->set_needs_ack(wait_on_ack);
        long srcid = fullos_->get_current_threadid();
        lib_mem_->write(bytes, srcid, thread::nic_thread);
        os_->execute_kernel(AMI::AMI_COMM_SEND, msg);
        if (wait_on_ack){
            SSTMAC_DEBUG << sstprintf("Waiting for ack on socket %d\n", fd);
            sock->block(socket::BLOCK_ON_SEND);
            SSTMAC_DEBUG << sstprintf("Got ack on socket %d\n", fd);
        }
        return bytes;
    }

    void
    socketapi::init_os(const OS_API::ptr &os)
    {
        library::init_os(os);
        os_->register_lib(lib_mem_);

        library::ptr server_lib = os_->get_lib(server_libname_);
        // only do one server per app per node
        if (server_lib == library::ptr()){
            server_ = socket_server::construct(server_libname_);
            os_->register_ser(server_);
            server_->start();
        }
        else {
            server_ = safe_cast(server_lib, socket_server);
        }
    }

    void
    socketapi::close(int fd)
    {
        std::map<int, socket::ptr>::iterator it = all_sockets_.find(fd);
        if (it == all_sockets_.end()){
            sst_throw(valueerror, "cannot close socket %d. no such file descriptor could be found", fd);
        }
        all_sockets_.erase(it);
        free_sockets_.push_back(fd);
    }

    int
    socketapi::allocate_socket(socket_domain_t domain, socket_type_t typ, socket_protocol_t prot)
    {
        int fd;
        if (free_sockets_.empty()){
            fd = next_free_socket_++;

        }
        else {
            fd = free_sockets_.back();
            free_sockets_.pop_back();
        }
        nodeaddress::ptr addr = operatingsystem::get_current_node();
        socket::ptr newsock = new socket(fd, addr, domain, typ, prot, fullos_);
        all_sockets_[fd] = newsock;
        return fd;
    }

    void
    socketapi::incoming_message(const socket_message::ptr& msg)
    {
        int number = msg->number();

        SSTMAC_DEBUG << sstprintf(
            "Incoming message of type=%s size=%ld from addr=%ld on socket number=%d\n"
            "# Bound = %d, # Connected = %d\n",
            sst_message::tostr(msg->get_type()),
            msg->nbytes(),
            msg->fromaddr()->unique_id(), number,
            bound_sockets_.size(),
            connected_sockets_.size());

        if (msg->get_type() == sst_message::SEND_ACK){
            socket::ptr sock = get_connected_socket(number);
            if (!sock){
                sst_throw(illformederror, "received ack for unconnected send on socket number=%d", number);
            }
            else if (sock->state() != socket::BLOCK_ON_SEND){
                sst_throw(illformederror, "received ack for non-blocking send on socket number=%d", number);
            }
            else {
                sock->unblock();
            }
        }
        else {
            socket::ptr sock = get_bound_socket(number);
            if (!sock){
                SSTMAC_DEBUG << sstprintf("No bound socket found for number=%d\n", number);
                pending_recv_.push_back(msg);
                return;
            }
            else if (sock->recv_addr()->equals(msg->fromaddr())){
                sock->add_pending(msg);
                SSTMAC_DEBUG << sstprintf("Socket number=%d now has %ld bytes available\n",
                                          sock->number(), sock->bytes_available);
            }
            else {
                pending_recv_.push_back(msg);
            }

            SSTMAC_DEBUG << sstprintf("Checking socket on fd=%d number=%d in state=%s\n",
                                      sock->fd(), sock->number(), socket::tostr(sock->state()));


            if (sock->state() == socket::BLOCK_ON_ACCEPT){
                if (sock->recv_addr()->equals(msg->fromaddr()))
                {
                    SSTMAC_DEBUG << sstprintf("Unblocking accept on fd=%d number=%d on node=%ld\n",
                                             sock->fd(), sock->number(), msg->fromaddr()->unique_id());
                    sock->accept(msg->fromaddr());
                    sock->unblock();
                }
                else {
                    SSTMAC_DEBUG << sstprintf("Failing to unblock on fd=%d number=%d for needed addr=%ld on msg addr=%ld\n",
                                              sock->fd(), sock->number(), sock->recv_addr()->unique_id(), msg->fromaddr()->unique_id());
                }
            }
            else if (sock->state() == socket::BLOCK_ON_RECV)
            {
                SSTMAC_DEBUG << sstprintf("Socket fd=%d number=%d has %ld bytes available and needs %ld to unblock\n",
                                         sock->fd(), sock->number(),
                                         sock->bytes_available, sock->bytes_needed_to_unblock);
                sock->unblock();
            }

        }
    }

    nodeaddress::ptr
    socketapi::accept(int fd, const nodeaddress::ptr& addr, bool blocking)
    {
        socket::ptr sock = get_socket(fd);

        SSTMAC_DEBUG << sstprintf("%s accept on socket fd=%d number=%d for addr %ld\n",
                                  blocking ? "Blocking" : "Non-blocking",
                                  fd, sock->number(), addr->unique_id());

        { std::list<socket_message::ptr>::iterator it = pending_recv_.begin(), end = pending_recv_.end();
        for ( ; it != end; ++it){
            socket_message::ptr msg = *it;
            if (msg->number() == sock->number() && addr->equals(msg->fromaddr())){
                sock->accept(msg->fromaddr());
                break;
            }
        } }

        if (!sock->recv_addr()){
            if (blocking){
                sock->accept(addr);
                sock->block(socket::BLOCK_ON_ACCEPT);
            }
            else {
                return nodeaddress::ptr();
            }
        }

        { std::list<socket_message::ptr>::iterator it = pending_recv_.begin(), end = pending_recv_.end();
        for ( ; it != end; ++it){
            //socket_message::ptr msg = *it;
        } }

        SSTMAC_DEBUG << sstprintf("Accept on socket fd=%d number=%d found addr %ld\n",
                                  fd, sock->number(), sock->recv_addr()->unique_id());

        return sock->recv_addr();
    }

    void
    socketapi::unaccept(int fd)
    {
        socket::ptr sock = get_socket(fd);
        SSTMAC_DEBUG << sstprintf("Unaccept socket fd=%d number=%d\n",
                                  fd, sock->number());
        sock->unaccept();
    }

    void
    socketapi::connect(int fd, int number, const nodeaddress::ptr& addr)
    {
        SSTMAC_DEBUG << sstprintf("Connected socket fd=%d number=%d\n",
                                  fd, number);
        socket::ptr sock = get_socket(fd);
        sock->connect(number,addr);
        connected_sockets_[number] = sock;
        server_->open_socket(number, ptr(this));
    }

    void
    socketapi::disconnect(int fd)
    {
        socket::ptr sock = get_socket(fd);
        SSTMAC_DEBUG << sstprintf("Disconnected socket fd=%d number=%d\n",
                                  sock->fd(), sock->number());
        server_->close_socket(sock->number());
        sock->disconnect();
        connected_sockets_.erase(sock->number());
    }

    void
    socketapi::bind(int fd, int number, const nodeaddress::ptr& addr)
    {
        SSTMAC_DEBUG << sstprintf("Binding socket on fd %d number %d\n",
                                  fd, number);
        socket::ptr sock = get_socket(fd);
        sock->bind(number, addr);
        bound_sockets_[number] = sock;
        server_->bind_socket(number, ptr(this));
    }

    void
    socketapi::init(const softwareid &sid)
    {
        api::init(sid);
        id_ = sid;
        libname_ = "libsocket-" + sid.to_string();

        std::string mem_lib_name = sstprintf("%s-libmem", libname_.c_str());
        lib_mem_ = lib_compute_memmove::construct(mem_lib_name);
        lib_mem_->init_as_multithreaded_lib();

        server_libname_ = sstprintf("socket_server_daemon%d", sid.app_.id);
    }

    bool
    socketapi::supported() const
    {
        return os_->kernel_supported(AMI::AMI_COMM_SEND);
    }

    socket_server::socket_server(const std::string& libname)
    {
        libname_ = libname;
    }

    void
    socket_server::handle(const sst_message::ptr &msg)
    {
        socket_message::ptr sockmsg = safe_cast(msg, socket_message);
        if (sockmsg->type() == socket_message::client){
            get_client_socket(sockmsg->number())->incoming_message(sockmsg);
        }
        else {
            get_server_socket(sockmsg->number())->incoming_message(sockmsg);
        }

    }

    socket_server::ptr
    socket_server::construct(const std::string& libname)
    {
        return new socket_server(libname);
    }

    service::ptr
    socket_server::clone()
    {
        return new socket_server(libname_);
    }

    socketapi::ptr
    socket_server::get_client_socket(int number) const
    {
        boost::unordered_map<int, socketapi::ptr>::const_iterator
            it = client_sockets_.find(number);
        if (it == client_sockets_.end()){
            sst_throw(valueerror,
                "socket_server::get_client_socket: invalid socket %d",
                number);
        }
        return it->second;
    }

    socketapi::ptr
    socket_server::get_server_socket(int number) const
    {
        boost::unordered_map<int, socketapi::ptr>::const_iterator
            it = server_sockets_.find(number);
        if (it == server_sockets_.end()){
            sst_throw(valueerror,
                "socket_server::get_server_socket: invalid socket %d",
                number);
        }
        return it->second;
    }

    void
    socket_server::bind_socket(int number, const socketapi::ptr &proc)
    {
        socketapi::ptr& existing = server_sockets_[number];
        if (existing){
            sst_throw(illformederror,
                "socket_server::bind_socket: %d is already open",
                number);
        }
        existing = proc;
    }

    void
    socket_server::unbind_socket(int number)
    {
        boost::unordered_map<int,socketapi::ptr>::iterator
            it = server_sockets_.find(number);
        if (it == server_sockets_.end()){
            sst_throw(valueerror,
                "socket_server::unbind_socket: no socket on fd %d",
                number);
        }
        server_sockets_.erase(it);
    }

    void
    socket_server::open_socket(int number, const socketapi::ptr &proc)
    {
        socketapi::ptr& existing = client_sockets_[number];
        if (existing){
            sst_throw(illformederror,
                "socket_server::open_socket: %d is already open",
                number);
        }
        existing = proc;
    }

    void
    socket_server::close_socket(int number)
    {
        boost::unordered_map<int,socketapi::ptr>::iterator
            it = client_sockets_.find(number);
        if (it == client_sockets_.end()){
            sst_throw(valueerror,
                "socket_server::close_socket: no socket on fd %d",
                number);
        }
        client_sockets_.erase(it);
    }


    }
}
