#ifndef sstmac_software_libraries_socketapiS_H
#define sstmac_software_libraries_socketapiS_H

#include <sstmac/software/libraries/compute/lib_compute_memmove.h>
#include <sstmac/software/process/softwareid.h>
#include <sstmac/common/sstmac-config.h>
#include <sstmac/software/process/operatingsystem.h>
#include <sstmac/software/api/sockets/socketapi.h>
#include <sstmac/software/api/api.h>
#include <sstmac/hardware/network/network_message.h>
#include <sstmac/software/services/service.h>

namespace sstmac
{
  namespace sw
  {


    class socket_message :
        public hw::network_message,
        public library_message
    {
        public:
            typedef enum { client, server } type_t;

        private:
            int number_;

            long nbytes_;

            static taskid nic_task;

            type_t type_;

        protected:
            socket_message(
                int number, long bytes,
                type_t ty,
                const std::string& libname,
                const nodeaddress::ptr &toaddr,
                const nodeaddress::ptr &fromaddr);

        public:
            typedef boost::intrusive_ptr<socket_message> ptr;

            static ptr
            construct(
                int socket, long bytes,
                type_t ty,
                const std::string& libname,
                const nodeaddress::ptr &toaddr,
                const nodeaddress::ptr &fromaddr);

            int number() const;

            long nbytes() const;

            long get_byte_length() const;

            type_t type() const {
                return type_;
            }

            type_t reverse(type_t ty) const;

            void offset(long nbytes);

            sst_message::ptr clone(MESSAGE_TYPES t) const;

            const taskid& source_task() const { return nic_task; }

            const taskid& dest_task() const { return nic_task; }


    };

    class socket :
        virtual public ptr_type
    {
        // ------- constructor / boost stuff -------------//
        public:
          typedef enum { BLOCK_ON_ACCEPT, BLOCK_ON_RECV, BLOCK_ON_SEND, UNBLOCKED } state_t;

        private:
            int number_;

            int fd_;

            socket_domain_t domain_;

            socket_protocol_t protocol_;

            socket_type_t type_;

            nodeaddress::ptr my_addr_;

            nodeaddress::ptr remote_addr_;

            std::list<socket_message::ptr> pending_recv_;

            static debug dbg_;

            key::ptr key_;

            operatingsystem::ptr os_;

            state_t state_;

        public:
            typedef boost::intrusive_ptr<socket> ptr;

            socket(int fd, const nodeaddress::ptr& my_addr,
                   socket_domain_t domain, socket_type_t typ,
                   socket_protocol_t protocol,
                   const operatingsystem::ptr& os);

            std::string to_string() const;

            void connect(int number, const nodeaddress::ptr& addr);

            void disconnect();

            void bind(int number, const nodeaddress::ptr& addr);

            void shutdown();

            long recv(long max_bytes);

            nodeaddress::ptr send_addr() const;

            nodeaddress::ptr recv_addr() const;

            nodeaddress::ptr my_addr() const;

            void block(state_t state);

            void unblock();

            state_t state() const;

            void rebind();

            long bytes_available;

            long bytes_needed_to_unblock;

            static const long unblocked = -1;

            static const int unbound = -1;

            void accept(const nodeaddress::ptr& remote_addr);

            void unaccept();

            int fd() const;

            int number() const;

            void add_pending(const socket_message::ptr& msg);

            static const char* tostr(state_t state);

    };

    class socket_server;
    class socketapi :
        public api
    {
    protected:
      typedef boost::intrusive_ptr<socket_server> serverPtr;

      std::list<socket_message::ptr> pending_recv_;

      std::map<int,socket::ptr> all_sockets_;

      std::map<int,socket::ptr> bound_sockets_;

      std::map<int,socket::ptr> connected_sockets_;

      socketapi();

      std::vector<int> free_sockets_;

      int next_free_socket_;

      bool in_blocking_send_;

      lib_compute_memmove::ptr lib_mem_;

      softwareid id_;

      std::string server_libname_;

      serverPtr server_;

    public:
      typedef boost::intrusive_ptr<socketapi> ptr;

      virtual ~socketapi();

      static ptr
      construct(const sim_parameters::ptr& params);
      // --------------------------------------//

    public:
      void connect(int fd, int number, const nodeaddress::ptr& addr);

      void disconnect(int fd);

      void unaccept(int fd);

      void bind(int fd, int number, const nodeaddress::ptr& addr);

      long send(int fd, long bytes, bool wait_on_ack = true);

      long recv(int fd, long max_bytes, bool blocking);

      nodeaddress::ptr accept(int fd, const nodeaddress::ptr& addr, bool blocking);

      void close(int fd);

      virtual bool supported() const;

      void init(const softwareid &sid);

      void init_os(const OS_API::ptr &os);

      int allocate_socket(socket_domain_t domain, socket_type_t typ, socket_protocol_t prot);

      void incoming_message(const socket_message::ptr &msg);

      socket::ptr
      get_socket(int fd) const;

      socket::ptr
      get_bound_socket(int number) const;

      socket::ptr
      get_connected_socket(int number) const;

    protected:
      static debug dbg_;

    };

    class socket_server :
        public sstmac::sw::service
    {

        private:
            boost::unordered_map<int, socketapi::ptr> client_sockets_;
            boost::unordered_map<int, socketapi::ptr> server_sockets_;

            socketapi::ptr
            get_server_socket(int number) const;

            socketapi::ptr
            get_client_socket(int number) const;

        protected:
            socket_server(const std::string& libname);

        public:
            typedef boost::intrusive_ptr<socket_server> ptr;

            void handle(const sst_message::ptr& msg);

            void open_socket(int number, const socketapi::ptr& proc);
            void close_socket(int number);

            void bind_socket(int number, const socketapi::ptr& proc);
            void unbind_socket(int number);

            static ptr
            construct(const std::string& libname);

            sstmac::sw::service::ptr
            clone();
    };

  }
} //end of namespace sstmac

#endif // socketapiS_H
