/*  api.h                                                                    */

#ifndef SSTMAC_UPC_api_H_
#define SSTMAC_UPC_api_H_

#include <sstmac/common/ptr_type.h>
#include <sstmac/common/debug.h>
#include <sstmac/software/process/softwareid.h>
#include <sstmac/software/libraries/library.h>
#include <sstmac/software/libraries/compute/lib_compute_memmove.h>
#include <sstmac/common/messages/vectorpayload.h>
#include <sstmac/software/api/api.h>

#include "sstmac_upc/memory/shared_heap.h"
//#include "sstmac_upc/memory/global.h"
#include "sstmac_upc/memory/shared.h"
//#include "sstmac_upc/memory/shared_ptr.h" // --> see forward declaration below.
#include "sstmac_upc/memory/lock.h"
#include "sstmac_upc/messages/transfer.h"
#include "sstmac_upc/transfer_api.h"

#include "sstmac_upc/messages/get_transfer.h"
#include "sstmac_upc/messages/put_transfer.h"
#include "sstmac_upc/messages/op_transfer.h"

#include "sstmac_upc/messages/fake_data.h"

namespace sstmac
{
   namespace sw
   {
       namespace upc
       {
           // Forward declared to prevent circular dependence.
           template <typename T, size_t B>
           class shared_ptr;
           
           class api : public ::sstmac::sw::api, public virtual transfer_api
           {
               protected:
                   typedef ::sstmac::sw::api parent_api;
                   
                   static debug dbg_;
                   softwareid id_;
                   boost::intrusive_ptr<server> server_;
                   shared_heap::ptr heap_;
                   bool setup_;
                   
                   // lib_compute_memmove::ptr lib_mem_;
                   
                   key::ptr global_key_;
                   
                   boost::unordered_map<long, transfer::ptr> pending_puts_;
                   payload::const_ptr get_arrived_;
                   bool waiting_on_fence_;
                   
                   // Notify - Wait flags
                   bool inside_notify_;
                   static boost::unordered_map<key::ptr, server::ptr> notify_servmap_;
                   static long notify_count_;
                   static long wait_index_;
                   static int notify_key_1_;
                   static int notify_key_2_;
                   int  loc_notify_key_;
                   long loc_wait_index_;
                   
                   shared::ptr current_fake_;
                   
                   const static size_t UPC_MAX_BLOCKSIZE = 65536; // or whatever...
                   const static int DEFAULT_KEY = -42;
                   
                   api();
                   
                   
               public:
                   typedef boost::intrusive_ptr<api> ptr;
                   
                   static ptr
                   construct(const sim_parameters::ptr& params)
                   {
                       return new api;
                   }
                   
                   void
                   setup();
                   
                   void
                   init(const softwareid &sid);
                   
                   virtual
                   ~api();
                   
                   virtual std::string
                   to_string() const
                   {
                       return "api";
                   }
                   
                   virtual void
                   init_os(const OS_API::ptr& os);
                   
                   virtual void
                   put_done(long tid);
                   
                   virtual void
                   get_arrived(payload::const_ptr pay);
                   
                   virtual void
                   send_put_complete(transfer::const_ptr t) const;
                   
                   virtual void
                   send_get_response(transfer::const_ptr t, payload::const_ptr pay);
                   
                   virtual bool
                   check_shared_heap(shared::ptr sobj);
                   
                   virtual void
                   unblock();
                   
                   void
                   start_trans(transfer::ptr tran, bool block = true,
                               bool unblock_on_mem = true);
                   
                   void
                   init_shared()
                   {
                       if (shared::constructed_)
                       {
                           for (int i = 0; i < shared::constructed_->size(); i++)
                           {
                               shared::ptr sptr = (*shared::constructed_)[i];
                               heap_->add_shared(sptr);
                           }
                       }
                       
                       // shared::constructed_->clear();
                   }
                   
                   // UPC API calls
                   
                   size_t
                   threads();
                   
                   size_t
                   mythread();
                   
                   size_t
                   maxblocksize();
                   
                   // Synchronization
                   
                   void
                   notify();
                   
                   void
                   notify(int nkey);
                   
                   void
                   wait();
                   
                   void
                   wait(int wkey);
                   
                   void
                   barrier();
                   
                   void
                   barrier(int bkey);
                   
                   void
                   fence();
                   
                   // Shared 'sizeof' operators
                   
                   template <typename T, size_t B>
                   size_t
                   localsizeof(shared_ptr<T, B>& sobj)
                   {
                       SSTMACBacktrace("localsizeof");
                       start_api_call();
                       SSTMAC_DEBUG << "upc::api::localsizeof(" << sobj << ")\n";
                       end_api_call();
                       return sobj.localsizeof();
                   }
                   
                   template <typename T, size_t B>
                   size_t
                   blocksizeof(shared_ptr<T, B>& sobj)
                   {
                       SSTMACBacktrace("blocksizeof");
                       start_api_call();
                       SSTMAC_DEBUG << "upc::api::blocksizeof(" << sobj << ")\n";
                       end_api_call();
                       return sobj.blocksizeof();
                   }
                   
                   template <typename T, size_t B>
                   size_t
                   elemsizeof(shared_ptr<T, B>& sobj)
                   {
                       SSTMACBacktrace("elemsizeof");
                       start_api_call();
                       SSTMAC_DEBUG << "upc::api::elemsizeof(" << sobj << ")\n";
                       end_api_call();
                       return sobj.elemsizeof();
                   }
                   
                   size_t
                   localsizeof(std::string type_name);
                   
                   size_t
                   blocksizeof(std::string type_name);
                   
                   size_t
                   elemsizeof(std::string type_name);
                   
                   // UPC exit
                   
                   void
                   global_exit(int status);
                   
                   // Shared memory allocation
                   
                   shared::ptr
                   global_alloc(size_t nblocks, size_t nbytes)
                   {
                       SSTMACBacktrace("global_alloc");
                       start_api_call();
                       SSTMAC_DEBUG << "upc::api::global_alloc(" << nblocks << ", " << nbytes << ")\n";
                       if (nblocks * nbytes <= 0)
                           return shared::ptr();
                       
                       shared::ptr sobj = shared::construct(nblocks, nbytes);
                       heap_->add_shared(sobj);
                       
                       end_api_call();
                       return sobj;
                   }
                   
                   shared::ptr
                   all_alloc(size_t nblocks, size_t nbytes)
                   {
                       SSTMACBacktrace("all_alloc");
                       start_api_call();
                       SSTMAC_DEBUG << "upc::api::all_alloc(" << nblocks << ", " << nbytes << ")\n";
                       // Some fancy work to get all threads to return the same pointer...
                       static shared::ptr sobj;
                       static long count = 0;
                       
                       if (inside_notify_)
                       {
                           sst_throw(ssterror, "upc::api::all_alloc annot be called inside a notify-wait block.");
                       }
                       else
                       {
                           count++;
                           
                           if (count == 1)
                           {
                             if(nblocks*nbytes > 0)
                               sobj = shared::construct(nblocks, nbytes);
                           }
                           else if (count == threads())
                           {
                               count = 0;
                           }
                           
                           if(nblocks*nbytes > 0)
                             heap_->add_shared(sobj);
                       }
                       
                       barrier(-157);
                       
                       end_api_call();
                       return sobj;
                   }
                   
                   shared::ptr
                   alloc(size_t nbytes)
                   {
                       SSTMACBacktrace("alloc");
                       start_api_call();
                       SSTMAC_DEBUG << "upc::api::alloc(" << nbytes << ")\n";
                       if (nbytes <= 0)
                           return shared::ptr();
                       
                       shared::ptr sobj = shared::construct(1, nbytes, mythread());
                       heap_->add_shared(sobj);
                       
                       end_api_call();
                       return sobj;
                   }
                   
                   template <typename T, size_t B>
                   void
                   free(shared_ptr<T, B>& sptr)
                   {
                       SSTMACBacktrace("free");
                       start_api_call();
                       SSTMAC_DEBUG << "upc::api::free(" << sptr.to_string() << ")\n";
                       if (sptr == 0)
                           return;
                       
                       shared::ptr sobj = sptr.get_ref();
                       heap_->delete_shared(sobj);
                       
                       sptr = NULL;
                       
                       end_api_call();
                   }
                   
                   // Shared pointer operations
                   
                   template <typename T, size_t B>
                   size_t
                   threadof(shared_ptr<T, B>& sptr)
                   {
                       SSTMACBacktrace("threadof");
                       start_api_call();
                       SSTMAC_DEBUG << "upc::api::threadof(" << sptr.to_string() << sptr.get_mythread() << ")\n";
                       end_api_call();
                       return sptr.threadof();
                   }
                   
                   template <typename T, size_t B>
                   size_t
                   phaseof(shared_ptr<T, B>& sptr)
                   {
                       SSTMACBacktrace("phaseof");
                       start_api_call();
                       SSTMAC_DEBUG << "upc::api::phaseof(" << sptr.to_string() << sptr.get_mythread() << ")\n";
                       end_api_call();
                       return sptr.phaseof();
                   }
                   
                   template <typename T, size_t B>
                   size_t
                   addrfield(shared_ptr<T, B>& sptr)
                   {
                       SSTMACBacktrace("addrfield");
                       start_api_call();
                       SSTMAC_DEBUG << "upc::api::addrfield(" << sptr.to_string() << sptr.get_mythread() << ")\n";
                       end_api_call();
                       return sptr.addrfield();
                   }
                   
                   template <typename T, size_t B>
                   shared_ptr<T, B>&
                   resetphase(shared_ptr<T, B>& sptr)
                   {
                       SSTMACBacktrace("resetphase");
                       start_api_call();
                       SSTMAC_DEBUG << "upc::api::resetphase(" << sptr.to_string() << sptr.get_mythread() << ")\n";
                       shared_ptr<T, B> other(sptr);
                       other.resetphase();
                       end_api_call();
                       return other;
                   }
                   
                   size_t
                   affinitysize(size_t totalsize, size_t nbytes, size_t threadid);
                   
                   // Shared lock functions.
                   
                   lock *
                   global_lock_alloc();
                   
                   lock *
                   all_lock_alloc();
                   
                   void
                   lock_free(lock *lck);
                   
                   void
                   lock_wait(lock *lck);
                   
                   int
                   lock_attempt(lock *lck);
                   
                   void
                   unlock(lock *lck);
                   
                   // Memory operations.
                   
                   template <typename T, size_t B>
                   void
                   memcpy(shared_ptr<T, B>& dst, shared_ptr<T, B>& src, size_t n)
                   {
                       // less than efficient way to do a copy...
                       size_t len = n/sizeof(T);
                       T t_array[len];
                       upc_get(t_array, src, len);
                       upc_put(dst, t_array, len);
                   }
                   
                   template <typename T, size_t B>
                   void
                   memget(void *dst, shared_ptr<T, B>& src, size_t n)
                   {
                       T* t_dst = (T *)dst;
                       upc_get(t_dst, src, n/sizeof(T));
                   }
                   
                   template <typename T, size_t B>
                   void
                   memput(shared_ptr<T, B>& dst, const void *src, size_t n)
                   {
                       const T* t_src = (const T*)src;
                       upc_put(dst, t_src, n/sizeof(T));
                   }
                   
                   template <typename T, size_t B>
                   void
                   memset(shared_ptr<T, B>& dst, int c, size_t n)
                   {
                       // rather bad way of doing a memset...
                       unsigned char c_array[n];
                       std::memset(c_array, c, n);
                       T* t_src = (T *)c_array;
                       upc_put(dst, t_src, n/sizeof(T));
                   }
                   
                   // Generic put and get routines.
                   
                   template <typename T, size_t B>
                   void
                   upc_put(shared_ptr<T, B>& dst, T* src, size_t size)
                   {
                       SSTMAC_DEBUG << "upc::api::upc_put(" << size << ")\n";
                       boost::intrusive_ptr<put_transfer<T> > tran =
                            boost::dynamic_pointer_cast<put_transfer<T> > (dst.build_put(src, mythread(), size));
                       pending_puts_[tran->get_tid()] = tran;
                       start_trans(tran, true, true);
                   }
                   
                   template <typename T, size_t B>
                   void
                   upc_get(T *dst, shared_ptr<T, B>& src, size_t size)
                   {
                       SSTMAC_DEBUG << "upc::api::upc_get(" << size << ")\n";
                       boost::intrusive_ptr<get_transfer<T> > tran =
                                boost::dynamic_pointer_cast<get_transfer<T> > (src.build_get(mythread(), size));
                            
                       start_trans(tran, true, false);
                       if (src != NULL && dst != NULL)
                       {
                           boost::intrusive_ptr<const vector1payload<T, T*> > cast =
                                    boost::dynamic_pointer_cast<const vector1payload<T, T*> >(get_arrived_);
                           const T *array = cast->data();
                           for (long i = 0; i < size; ++i)
                               dst[i] = array[i];
                           get_arrived_ = payload::ptr();
                       }
                   }
           };
           
       }
   }
}

#endif
