/*  fake_data.h                                                              */

#ifndef SSTMAC_UPC_fake_data_H_
#define SSTMAC_UPC_fake_data_H_

#include <sstmac/common/messages/payload.h>

namespace sstmac
{
  namespace sw
  {
    namespace upc
    {

      /**
       * Pure abstract base for payload that can be associated with a network send.
       */
      class fake_data :
        public payload,
        public serializable_type<fake_data>
      {
      ImplementSerializable(fake_data)

      protected:
        size_t size_;

        /**
         * Empty protected constructor to enforce using construct()
         */
        fake_data(size_t s) :
          size_(s)
        {

        }
        fake_data() :
          size_(0)
        {

        }
      public:

        typedef boost::intrusive_ptr<fake_data> ptr;
        typedef boost::intrusive_ptr<const fake_data> const_ptr;

        /// Goodbye.
        virtual
        ~fake_data()
        {
        }

        static ptr
        construct(size_t s)
        {
          return ptr(new fake_data(s));
        }

        static ptr
        construct()
        {
          return ptr(new fake_data());
        }

        /**
         * Create a copy of this payload
         * @return A smart pointer to a new payload object
         */
        virtual payload::const_ptr
        clone() const
        {
          return construct(size_);
        }

        /**
         * Stringify
         * @return String description
         */
        virtual std::string
        to_string() const
        {
          return "fake_data";
        }

        virtual long
        get_byte_length() const
        {
          return size_;
        }

        /**
         * Pure virtual add operator
         * @param other Payload object to add to this one
         * @return A smart pointer to a new object which is the sum of this and other
         */
        virtual const payload::const_ptr
        add(const payload::const_ptr &other) const
        {
          return const_ptr(this);
        }

        virtual const payload::const_ptr
        prod(const payload::const_ptr &other) const
        {
          return const_ptr(this);
        }

        /**
         * Pure virtual less than (<) operator
         * @param other Payload to compare to
         * @return true if this payload is considered less than other
         */
        virtual const payload::const_ptr
        min(const payload::const_ptr &other) const
        {
          return const_ptr(this);
        }

        /**
         * Pure virtual greater than (>) operator
         * @param other Payload to compare to
         * @return true if this payload is considered greather than other
         */
        virtual const payload::const_ptr
        max(const payload::const_ptr &other) const
        {
          return const_ptr(this);
        }

        /**
         * Pure virtual equals operator
         * @param other Payload to compare to
         * @return true if this payload is considered equal to the other
         */
        virtual bool
        equals(const payload::const_ptr &other) const
        {
          sst_throw(ssterror, "fake_data::equals - i have no way of knowing");
        }

        /**
         * Pure virtual logical or operator
         * @param other Payload to compare to
         * @return true if this either payload is considered true
         */
        virtual const payload::const_ptr
        logical_or(const payload::const_ptr &other) const
        {
          return const_ptr(this);
        }

        /**
         * Pure virtual logical xor operator
         * @param other Payload to compare to
         * @return true if exactly one payload is considered true
         */
        virtual const payload::const_ptr
        logical_xor(const payload::const_ptr &other) const
        {
          return const_ptr(this);
        }

        /**
         * Pure virtual logical and operator
         * @param other Payload to compare to
         * @return true if both payloads are considered true
         */
        virtual const payload::const_ptr
        logical_and(const payload::const_ptr &other) const
        {
          return const_ptr(this);
        }

        virtual const payload::const_ptr
        bitwise_or(const payload::const_ptr &other) const
        {
          return const_ptr(this);
        }

        /**
         * Pure virtual logical xor operator
         * @param other Payload to compare to
         * @return true if exactly one payload is considered true
         */
        virtual const payload::const_ptr
        bitwise_xor(const payload::const_ptr &other) const
        {
          return const_ptr(this);
        }

        /**
         * Pure virtual logical and operator
         * @param other Payload to compare to
         * @return true if both payloads are considered true
         */
        virtual const payload::const_ptr
        bitwise_and(const payload::const_ptr &other) const
        {
          return const_ptr(this);
        }


        /**
         * Serialize this message during parallel simulation.
         * @param ser The serializer to use
         */
        virtual void
        serialize_order(sst_serializer* ser)
        {
          ser->do_primitive(size_);
        }


      };
    } // end of namespace sstmac
  }
}

#endif
