/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/library.h>
#include <sstmac/software/process/operatingsystem.h>
#include <sstmac/common/sstmac_env.h>
#include <sstmac/software/libraries/libwaiters.h>

namespace sstmac
{
  namespace sw
  {

    static std::map<std::string, library_builder*>* library_builders_ = 0;

    void
    library::register_builder(const std::string &libname,
        library_builder *builder)
    {
      if (!library_builders_)
      {
        library_builders_ = new std::map<std::string, library_builder*>;
      }
      (*library_builders_)[libname] = builder;
    }

    library_builder*
    library::get_builder(const std::string &libname)
    {
      if (!library_builders_){
        sst_throw(valueerror, "no library builders registered");
      }

      library_builder* builder = (*library_builders_)[libname];
      if (!builder){
        sst_throw(valueerror, "name %s is not known for constructing library", libname.c_str());
      }
      return builder;
    }

    library::ptr
    library::construct_lib(const std::string& libname, const softwareid& sid)
    {
      library_builder* builder = get_builder(libname);
      std::string sid_libname = libname + sid.to_string();
      library::ptr lib = builder->construct(sid_libname);
      return lib;
    }

    library::ptr
    library::construct_lib(const std::string& libname)
    {
      library_builder* builder = get_builder(libname);
      library::ptr lib = builder->construct(libname);
      return lib;
    }

    library::library() :
      self_initialized_(false)
    {
    }

    library::~library()
    {
    }

    void
    library::wait_this(const sst_message::ptr &msg)
    {
      if (!waiter_){
        sst_throw(libraryerror,
            "library::wait: waiter has not been initialized. make sure that init_os was also called");
      }
      waiter_->wait_msg(msg);
    }

    sst_message::ptr
    library::wait_num_type(sst_message::MESSAGE_TYPES t, long num)
    {
      if (!waiter_){
        sst_throw(libraryerror,
            "library::wait: waiter has not been initialized. make sure that init_os was also called");
      }
      return waiter_->wait_msg_type(t, num);
    }

    sst_message::ptr
    library::wait_type(sst_message::MESSAGE_TYPES t)
    {
        return wait_num_type(t, 1);
    }

    void
    library::unregister_all_libs()
    {
      std::list<library::ptr>::iterator it, end = otherlibs_.end();
      for (it = otherlibs_.begin(); it != end; it++)
      {
        (*it)->unregister_all_libs();
        os_->unregister_lib(*it);
      }
      otherlibs_.clear();
    }

    void
    library::register_child_libs()
    {
    }

    void
    library::cancel_wait_this(const sst_message::ptr &msg)
    {
    }

    void
    library::register_lib(const library::ptr &lib)
    {
      if (os_)
        os_->register_lib(lib);
      else
        throw valueerror("library::register_lib - os not initialized yet");

      otherlibs_.push_back(lib);
    }

    operatingsystem::ptr
    library::get_full_os() const
    {
      return fullos_;
    }

    void
    library::init_os(const OS_API::ptr& os)
    {
      os_ = os;
      fullos_ = boost::dynamic_pointer_cast<operatingsystem>(os);

      if (!waiter_)
        waiter_ = serial_single_msg_lib_waiter::construct(key_cat_);

      waiter_->init_os(fullos_);

      if (sstmac_env::params)
      {
        consume_params(sstmac_env::params);
      }
    }

    void
    library::consume_params(const sim_parameters::ptr& params)
    {
      //do nothing by default
    }

    OS_API::ptr
    library::get_os() const
    {
      return os_;
    }

    void
    library::incoming_message(const sst_message::ptr &msg)
    {
      if (!waiter_) {
        sst_throw(libraryerror,
            "library::incoming_message: waiter has not been initialized. make sure that init_os was also called");
      }
      waiter_->incoming_msg(msg);
    }

    void
    library::init_as_callback_lib()
    {
        init_waiter(callback_lib_waiter::construct());
    }

    void
    library::init_as_multithreaded_lib()
    {
        init_waiter(multithreaded_blocking_lib_waiter::construct(key_cat_));
    }

    void
    library_waiter::wait_msg(const sst_message::ptr &msg){
        sst_throw(unimplementederror,
            "waiter %s is not compatible with wait_msg");
    }

    sst_message::ptr
    library_waiter::wait_msg_type(sst_message::MESSAGE_TYPES ty, int count){
        sst_throw(unimplementederror,
            "waiter %s is not compatible with wait_msg_type");
    }


  }
} //end namespace

