#include "libtopomap.hpp"

/* greedily maps a logical topology graph (ltg) to a physical topology
 * graph (ptg) with nprocs[i] processors per vertex i. Results in a
 * mapping from ltg to ptg in mapping. The greedy algorithm begins at
 * the vertex start */
int TPM_Map_greedy(int r, TPM_Graph *ptg_ref, TPM_Graph_ewgts *ptgc_ref, TPM_Graph_vwgts *nprocs_ref, TPM_Graph *ltg_ref, TPM_Graph_ewgts *weights_ref, TPM_Mapping *mapping_ref, int start) {
  // turn refs into normal objects
  TPM_Graph& ptg = *ptg_ref;
  TPM_Graph_ewgts& ptgc = *ptgc_ref; // ptg edge capacities!
  TPM_Graph& ltg = *ltg_ref;
  TPM_Graph_ewgts& ltgw = *weights_ref; // ltg edge weights!
  TPM_Graph_vwgts& nprocs = *nprocs_ref;
  TPM_Mapping& mapping = *mapping_ref;

  // number of slots in ptg (nprocs) must be identical to number of
  // nodes in ltg! This is just a correctness check!
#ifdef CHECKS
  if(std::accumulate(nprocs.begin(), nprocs.end(), 0) != ltg.size()) {
    if(!r) printf("number of slots does not match (nprocs: %i, ltg.size(): %i!\n", std::accumulate(nprocs.begin(), nprocs.end(), 0), ltg.size());
    MPI_Abort(MPI_COMM_WORLD, 5);
  }
#endif

//printf("ltgw: %i, ltg:%i \n", ltgw.size(), ltg.size());

  TPM_Graph_dewgts w(ptg.size());
  // init edge weights in ptg to max(nprocs)*n^2 to force shortest paths!
  int maxw=1;
  for(int i=0; i<ltgw.size(); ++i) if(ltgw[i].size() > 0) maxw=std::max(maxw, *std::max_element(ltgw[i].begin(), ltgw[i].end()) );
  double init = (*std::max_element(nprocs.begin(), nprocs.end())) * maxw * ptg.size()*ptg.size();
  assert(init > 0); // check for overflow
  for(int i=0; i<ptg.size(); ++i) w[i].resize(ptg[i].size(), init);

  TPM_Graph_dists ptdists(ptg.size());
  TPM_Graph_preds ptpreds(ptg.size());


  std::vector<bool> mapped(ltg.size(), false);
  std::vector<bool> discovered(ltg.size(), false);

  /********************************************************
   * The actual mapping code
   ********************************************************/

  int current=start; // the current vertex in ptg to be mapped to
  int num_map=0; // the number of mapped vertices in ltg
  while(num_map < ltg.size()) {
    /* find vertex in ltg with most edges (most edge weights) that is
     * not mapped yet and map it to next free vertex in ptg (starting
     * with myself ) */
    // TODO: linear for now, should be done with sort ...
    int maxedges=-1;
    int maxvert=-1;
    for(int i=0; i<ltg.size(); ++i) {
      if(mapped[i]) continue;
      int thisedges=0;
      for(int j=0; j<ltg[i].size(); ++j) {
        assert(i<ltgw.size() && j<ltgw[i].size());
        thisedges+=ltgw[i][j]; // edge weight of logical edge
      }
      //printf("%i -- %i\n", i, ltg[i].size());
      if(thisedges > maxedges) {
        maxedges=thisedges;
        maxvert=i;
      }
    }

    DBG4(if(!r) printf("most 'expensive' vertex in logical topology graph: %i (cost: %i)\n", maxvert, maxedges));

    // if current vertex in ptg does not have free slots, choose different
    // (pretty much random) one -- should this pick a particular one?
    while(nprocs[current] <= 0) { current = (current+1)%nprocs.size(); }

    DBG4(if(!r) printf("mapped %i to %i\n", maxvert, current));
    discovered[maxvert] = true;
    mapped[maxvert] = true;
    num_map++;
    mapping[maxvert] = current; // map rank maxvert to current
    nprocs[current]--; // one of the procs of current is occupied now
    /* add all edges of this vertex that lead to not-discovered vertices
     * to prio queue */
    std::priority_queue<std::pair<double,int>, std::vector<std::pair<double,int> >, max_compare_func > Q;
    for(int j=0; j<ltg[maxvert].size(); ++j) {
      if(!discovered[ltg[maxvert][j]]) {
        Q.push(std::make_pair(ltgw[maxvert][j], ltg[maxvert][j]));
        discovered[ltg[maxvert][j]] = 1;
      }
    }

    /* while Q not empty find most expensive edge in queue */
    while(!Q.empty()) {
      // take the most expensive edge out of the queue
      std::pair<double,int> Qel = Q.top();
      Q.pop();
      int target = Qel.second;

      DBG4(if(!r) printf("most expensive edge in topology graph leads to: %i (cost: %i)\n", target, Qel.first));

      /* find next vertex that is as close as possible (minimum path from
       * current but still has available slots (vertex weight is not 0)
       * map the target of the edge from previous step to it */
      TPM_SSSP(&ptg, &w, current, &ptdists, &ptpreds);
      double smallestdist=std::numeric_limits<double>::max();
      int smallestdistvert=-1;
      // again linear time, should be logarithmic!
      for(int j=0; j<ptdists.size(); ++j) if(smallestdist > ptdists[j] && nprocs[j] > 0) {
        smallestdist = ptdists[j];
        smallestdistvert = j;
      }
      DBG4(if(!r) printf("vertex with smallest distance from current (%i) is %i (dist %.2f)\n", current, smallestdistvert, smallestdist/ptg.size()/ptg.size()));

      // map most expensive process to the nearest physical neighbor!
      DBG4(if(!r) printf("mapped %i to %i\n", target, smallestdistvert));
      nprocs[smallestdistvert]--;
      mapping[target] = smallestdistvert;
      mapped[target] = true;
      num_map++;

      // update occupied edges in graph
      int v=smallestdistvert;
      while(v != current) {
        // find position of v in weights map (same as in h)
        int pos; for(pos=0; ptg[ptpreds[v]][pos] != v; pos++);
        assert(w.size() > ptpreds[v]);
        assert(w[ptpreds[v]].size() > pos);
        w[ptpreds[v]][pos] += Qel.first/ptgc[ptpreds[v]][pos];
        //printf("increasing edge weight on edge %i -> %i to %i\n", pred[v], v, w[pred[v]][pos]-nh*nh);
        v = ptpreds[v];
      }

      /* add all edges of just mapped vertex (target) that lead to
       * not-discovered vertices to prio queue */
      for(int j=0; j<ltg[target].size(); ++j) {
        if(!discovered[ltg[target][j]]) {
          Q.push(std::make_pair(ltgw[target][j], ltg[target][j]));
          discovered[ltg[target][j]] = 1;
        }
      }
    }
  }
  return 0;
}
