#include "libtopomap.hpp"
#include <sstmac/software/libraries/parmetis/parmetis.h>

/* uses the reverse Cuthill Mckee algorithm top optimize the mapping.
 * This approach applies rcm to the topology graph as well as the
 * problem graph and then maps them together. This of course only works
 * if both graphs have the same cardinality! */
int TPM_Map_rcm(int r, TPM_Graph *ptg_ref, TPM_Graph_vwgts *nprocs_ref, TPM_Graph *ltg_ref, TPM_Mapping *mapping_ref, int start) {
  // turn refs into normal objects
  TPM_Graph& ptg = *ptg_ref;
  TPM_Graph& ltg = *ltg_ref;
  TPM_Graph_vwgts& nprocs = *nprocs_ref;
  TPM_Mapping& mapping = *mapping_ref;

  //MPI_Barrier(MPI_COMM_WORLD);

  // if cardinalities don't match
  TPM_Graph rtg;
  TPM_Mapping rtg2ptgmap(ptg.size()); // map to translate rtg vertices to ptg vertices
  int use_rtg = 0; // decide which graph to use for RCM of ptg
  if(ptg.size()!=ltg.size()) {
    // figure out which subgraph is actually used ... gaah, it is a
    // Steiner problem and the rcm doesn't quite work as it seems :-(
    // as a simple fix, just remove all unused (nprocs == 0) vertices (and
    // all neighboring edges) from ptg! This might disconnect ptg but RCM
    // does not care :-)!
    assert(ptg.size() == nprocs.size());
    // to remove from ptg, we build a new graph called rtg (RCM Topology Graph)
    TPM_Mapping rtgmap(ptg.size(), -1); // new numbers in rtg
    int cnt=0;
    for(int i=0; i<ptg.size(); ++i) {
      if(nprocs[i] != 0) {
        rtgmap[i] = cnt;
        rtg2ptgmap[cnt] = i;
        cnt++;
      }
    }

    for(int i=0; i<ptg.size(); ++i) if(nprocs[i] != 0) {
      // translate all adjacent vertices!
      std::vector<int> adj;
      for(int j=0; j<ptg[i].size(); ++j) {
        int vert = ptg[i][j];
        if(rtgmap[j] != -1) adj.push_back(rtgmap[j]);
      }
      rtg.push_back(adj);
    }
    use_rtg = 1;
  }

  // there should not be more than one proc in nprocs
#ifdef CHECKS
  for(int i=0; i<nprocs.size(); ++i) if(nprocs[i]>1)  {
    printf("[%i] nprocs at pos %i > 1 (%i)!\n", r, i, nprocs[i]);
    MPI_Abort(MPI_COMM_WORLD, 3);
  }
#endif

  std::vector<int> rcm_ptg_map;
  if(use_rtg) {
    rcm_ptg_map.resize(rtg.size());
    int rtgn=0; for(int i=0; i<rtg.size(); ++i) rtgn+=rtg[i].size();
    std::vector<idx_t>  xadj(rtg.size()+1), // CSR index
                          adjncy(rtgn); // CSR list
    for(int i=0; i<(rtg.size()+1); i++) if(i==0) xadj[i]=0; else xadj[i]=xadj[i-1]+rtg[i-1].size();
    int pos=0; for(int i=0; i<rtg.size(); i++) for(int j=0; j<rtg[i].size(); ++j) { adjncy[pos++] = rtg[i][j]; assert(rtg[i][j]<rtg.size()); }

    std::vector<signed char> mask(rtg.size(), 1);
    std::vector<int> degs(rtg.size());
    TPM_genrcmi(rtg.size(), 0, &xadj[0], &adjncy[0], &rcm_ptg_map[0], &mask[0], &degs[0]);
  } else {
    rcm_ptg_map.resize(ptg.size());
    int ptgn=0; for(int i=0; i<ptg.size(); ++i) ptgn+=ptg[i].size();
    std::vector<idx_t>  xadj(ptg.size()+1), // CSR index
                          adjncy(ptgn); // CSR list
    for(int i=0; i<(ptg.size()+1); i++) if(i==0) xadj[i]=0; else xadj[i]=xadj[i-1]+ptg[i-1].size();
    int pos=0; for(int i=0; i<ptg.size(); i++) for(int j=0; j<ptg[i].size(); ++j) { adjncy[pos++] = ptg[i][j]; assert(ptg[i][j]<ptg.size()); }

    std::vector<signed char> mask(ptg.size(), 1);
    std::vector<int> degs(ptg.size());
    TPM_genrcmi(ptg.size(), 0, &xadj[0], &adjncy[0], &rcm_ptg_map[0], &mask[0], &degs[0]);
  }

  //if(!r) { printf("ptg map: "); for(int i=0; i<rcm_ptg_map.size(); ++i) printf("%i ", rcm_ptg_map[i]); printf("\n"); }

  std::vector<int> rcm_ltg_map(ltg.size());
  {
    int ltgn=0; for(int i=0; i<ltg.size(); ++i) ltgn+=ltg[i].size();
    std::vector<idx_t>  xadj(ltg.size()+1), // CSR index
                          adjncy(ltgn); // CSR list
    for(int i=0; i<(ltg.size()+1); i++) if(i==0) xadj[i]=0; else xadj[i]=xadj[i-1]+ltg[i-1].size();
    int pos=0; for(int i=0; i<ltg.size(); i++) for(int j=0; j<ltg[i].size(); ++j) { adjncy[pos++] = ltg[i][j]; assert(ltg[i][j]<ltg.size()); }

    std::vector<signed char> mask(ltg.size(), 1);
    std::vector<int> degs(ltg.size());
    TPM_genrcmi(ltg.size(), 0, &xadj[0], &adjncy[0], &rcm_ltg_map[0], &mask[0], &degs[0]);
  }

  //if(!r) { printf("ltg map: "); for(int i=0; i<rcm_ltg_map.size(); ++i) printf("%i ", rcm_ltg_map[i]); printf("\n"); }

  // translate mappings
  if(use_rtg) {
    for(int i=0; i<ltg.size(); ++i) mapping[rcm_ltg_map[i]] = rtg2ptgmap[rcm_ptg_map[i]]; // TODO: not 100% sure that this is right
  } else {
    for(int i=0; i<ltg.size(); ++i) mapping[rcm_ltg_map[i]] = rcm_ptg_map[i]; // TODO: not sure either
  }

  //if(!r) { printf("final mapping: "); for(int i=0; i<mapping.size(); ++i) printf("%i ", mapping[i]); printf("\n"); }
  return 0;
}
