/* This file contains all the routing stuff ...
 */

#include "libtopomap.hpp"

/**
 * This call initializes the routing subsystem. If the TPM_ROUTING_FILE
 * environment variable is not set, then it uses simple SSSP routing.
 * If it is set, then it reads the file and decides based on the header
 * what the format is (how to read it.
 */
void TPM_RoutesInit(int print, TPM_Routes *r) {

  char *env = getenv("TPM_ROUTING_FILE");
  if(env != NULL) {
    TPM_RoutesRead(r, env);
    if(print) printf("using route file %s\n", env);
  } else {
    (*r).type = SSSP;
    if(print) printf("using SSSP routing\n", env);
  }

}

/* this implements the Dijkstra algorithm */
void TPM_SSSP(TPM_Graph *g_ref, TPM_Graph_dewgts *w_ref, int src, TPM_Graph_dists *dist_ref, TPM_Graph_preds *pred_ref) {
  TPM_Graph& g= *g_ref; // the graph!
  TPM_Graph_dewgts& w= *w_ref; // edge weights
  TPM_Graph_dists& dist= *dist_ref; // distance vector
  TPM_Graph_preds& pred = *pred_ref; // predecessor map

  std::vector<bool> finished(g.size(),false);

  // infinite initial distance and invalid predecessor
  std::fill(dist.begin(), dist.end(), std::numeric_limits<double>::max());
  std::fill(pred.begin(), pred.end(), -1);

  // Dijkstra queue
  std::priority_queue<std::pair<double /* dist */,int /* index */>, std::vector<std::pair<double,int> >, dijk_compare_func > Q;
  std::pair <double,int> Qel;

  dist[src] = 0;
  Q.push(std::pair <double,int> (dist[src], src));

  while(!Q.empty()) {
    // take the cheapest vertex out of the queue
    Qel = Q.top();
    Q.pop();
    int u = Qel.second;
    DBG1(printf("finished vertex %i at dist %.2f\n", u, dist[u]));
    // we need to do this because we can't change elements in the queue, thus, we just throw
    // more in every time we find a shorter path (so each vertex might be more than once in the queue)
    if(!finished[u]) {
      finished[u]=true;
      // investigate all neighbors of current vertex u
      for(int i=0; i<g[u].size(); ++i) {
        int v = g[u][i];
        double vweight = w[u][i];
        // did we discover a shorter route?
        if(!finished[v] && dist[u] + vweight < dist[v]) {
          dist[v] = dist[u] + vweight;
          pred[v] = u;
          DBG1(printf("discovered path to %i and dist %.2f\n", v, dist[v]));
          Q.push(std::pair <double,int> (dist[v], v));
        }
      }
    }
  }
}

/**
 * This function returns the route backwards, from the target to the destination. Mostly for historic reasons!
 */
void TPM_RouteGet(TPM_Graph *g_ref, TPM_Graph_dewgts *w_ref, TPM_Routes *routes_ref,
                  int src, int dst, std::vector<int> *path_ref) {
  TPM_Graph& g= *g_ref; // the graph!
  TPM_Graph_dewgts& w= *w_ref; // edge weights
  TPM_Routes& routes= *routes_ref; // distance vector
  std::vector<int>& path = *path_ref; // predecessor map

  if(src == dst) return;

  std::vector<int> tmp; // temp array
  if(routes.type == SSSP) {
    TPM_Graph_dists dists(g.size());
    TPM_Graph_preds preds(g.size());

    TPM_SSSP(&g, &w, src, &dists, &preds);

    // printf("route get: src: %i, dst: %i\n", src, dst);
    // walk predecessor array backwards and store to tmp
    int v=dst;
    while(v != src) {
      tmp.push_back(v);
      v = preds[v];
    }

    // tmp is in wrong order, revert it!
    for(int i=tmp.size()-1; i>0; --i) {
      path.push_back(tmp[i]);
    }
    path.push_back(dst);
  } else if(routes.type == COMPXYZ) {
    //printf("XYZ routing\n");
    std::map<std::pair<int,int>, std::vector<int>* >::iterator route_iter = routes.routes.find( std::make_pair(src, dst));
    if(route_iter != routes.routes.end()) { // if we have a special route ...
      std::vector<int>* rt = route_iter->second;
      printf("found a special route for %i->%i (length %lu): ", src, dst, path.size());
     for(int i=0; i<rt->size(); ++i) {
        printf(" %i", (*rt)[i]);
        path.push_back((*rt)[i]); // could use std::copy
      }
      printf("\n");
    } else { // just do normal XYZ routing, fist X, then Y, then Z (all shortest)
      xyzcoords_t srcc = routes.rank_to_xyz[src];
      xyzcoords_t dstc = routes.rank_to_xyz[dst];
      xyzcoords_t curc = srcc;
      xyzcoords_t max = routes.xyzmaxdims;

      // distances in all directions
      int x_dist = std::min( (dstc.x + max.x - srcc.x) % max.x, (srcc.x + max.x - dstc.x) % max.x );
      if((srcc.x + x_dist) % max.x != dstc.x) x_dist = -x_dist;
      int y_dist = std::min( (dstc.y + max.y - srcc.y) % max.y, (srcc.y + max.y - dstc.y) % max.y );
      if((srcc.y + y_dist) % max.y != dstc.y) y_dist = -y_dist;
      int z_dist = std::min( (dstc.z + max.z - srcc.z) % max.z, (srcc.z + max.z - dstc.z) % max.z );
      if((srcc.z + z_dist) % max.z != dstc.z) z_dist = -z_dist;

      printf("search standard XYZ route for %i->%i (%i,%i,%i)->(%i,%i,%i) - dists: %i, %i, %i\n", src, dst, srcc.x, srcc.y, srcc.z, dstc.x, dstc.y, dstc.z, x_dist, y_dist, z_dist);
      // walk in X
      while(x_dist > 0) {
        curc.x++;
        path.push_back(routes.xyz_to_rank[curc]);
        x_dist--;
      }
      while(x_dist < 0) {
        curc.x--;
        path.push_back(routes.xyz_to_rank[curc]);
        x_dist++;
      }
      // walk in Y
      while(y_dist > 0) {
        curc.y++;
        path.push_back(routes.xyz_to_rank[curc]);
        y_dist--;
      }
      while(y_dist < 0) {
        curc.y--;
        path.push_back(routes.xyz_to_rank[curc]);
        y_dist++;
      }
      // walk in Z
      while(z_dist > 0) {
        curc.z++;
        path.push_back(routes.xyz_to_rank[curc]);
        printf("adding: (%i,%i,%i) - %i\n", curc.x, curc.y, curc.z, routes.xyz_to_rank[curc]);
        z_dist--;
      }
      while(z_dist < 0) {
        curc.z--;
        path.push_back(routes.xyz_to_rank[curc]);
        z_dist++;
      }
      printf("found standard XYZ route for %i->%i (length %i): ", src, dst, path.size());
      for(int i=0; i<path.size(); ++i) {
        printf(" %i", path[i]);
      }
      printf("\n");
    }
  } else if(routes.type == COMPCRAYXE6) {
    //printf("Cray XYZ routing\n");
    std::map<std::pair<int,int>, std::vector<int>* >::iterator route_iter = routes.routes.find( std::make_pair(src, dst));
    if(route_iter != routes.routes.end()) { // if we have a special route ...
      std::vector<int>* rt = route_iter->second;
      printf("found a special route for %i->%i (length %i): ", src, dst, path.size());
     for(int i=0; i<rt->size(); ++i) {
        printf(" %i", (*rt)[i]);
        path.push_back((*rt)[i]); // could use std::copy
      }
      printf("\n");
    } else { // just do normal XYZ routing, fist X, then Y, then Z (all shortest)
      xyzcoords_t max = routes.xyzmaxdims;
      int srcgemini = routes.xe6gemini[src]; // this is the gemini where the src is attached
      int dstgemini = routes.xe6gemini[dst]; // this is the gemini where the dst is attached
      xyzcoords_t srcc = routes.rank_to_xyz[src]; // src gemini's XYZ coords
      xyzcoords_t dstc = routes.rank_to_xyz[dst]; // dst gemini's XYZ coords
      xyzcoords_t curc = srcc; // the current location in the torus

      // we have to walk to the src gemini first (only one choice)
      path.push_back(srcgemini);

      // distances in all directions
      int x_dist = std::min( (dstc.x + max.x - srcc.x) % max.x, (srcc.x + max.x - dstc.x) % max.x );
      if((srcc.x + x_dist) % max.x != dstc.x) x_dist = -x_dist;
      int y_dist = std::min( (dstc.y + max.y - srcc.y) % max.y, (srcc.y + max.y - dstc.y) % max.y );
      if((srcc.y + y_dist) % max.y != dstc.y) y_dist = -y_dist;
      int z_dist = std::min( (dstc.z + max.z - srcc.z) % max.z, (srcc.z + max.z - dstc.z) % max.z );
      if((srcc.z + z_dist) % max.z != dstc.z) z_dist = -z_dist;

      //printf("search standard XYZ route for %i (%i)->%i (%i) (%i,%i,%i)->(%i,%i,%i) - dists: %i, %i, %i\n", src, srcgemini, dst, dstgemini, srcc.x, srcc.y, srcc.z, dstc.x, dstc.y, dstc.z, x_dist, y_dist, z_dist);
      // walk in X
      while(x_dist > 0) {
        curc.x++;
        if (curc.x > max.x - 1) curc.x = 0;
        path.push_back(routes.xyz_to_rank[curc]);
        x_dist--;
      }
      while(x_dist < 0) {
        curc.x--;
        if (curc.x < 0) curc.x = max.x - 1;
        path.push_back(routes.xyz_to_rank[curc]);
        x_dist++;
      }
      // walk in Y
      while(y_dist > 0) {
        curc.y++;
        if (curc.y > max.y - 1) curc.y = 0;
        path.push_back(routes.xyz_to_rank[curc]);
        y_dist--;
      }
      while(y_dist < 0) {
        curc.y--;
        if (curc.y < 0) curc.y = max.y - 1;
        path.push_back(routes.xyz_to_rank[curc]);
        y_dist++;
      }
      // walk in Z
      while(z_dist > 0) {
        curc.z++;
        if (curc.z > max.z - 1) curc.z = 0;
        path.push_back(routes.xyz_to_rank[curc]);
        //printf("adding: (%i,%i,%i) - %i\n", curc.x, curc.y, curc.z, routes.xyz_to_rank[curc]);
        z_dist--;
      }
      while(z_dist < 0) {
        curc.z--;
        if (curc.z < 0) curc.z = max.z - 1;
        path.push_back(routes.xyz_to_rank[curc]);
        z_dist++;
      }
      //printf("found standard XYZ route for %i->%i (length %i): ", src, dst, path.size());
    }

    // we have to walk to the dst last (only one choice)
    path.push_back(dst);

    //for(int i=0; i<path.size(); ++i) {
    //  printf(" %i", path[i]);
    //}
    //printf("\n");
  } else if(routes.type == UNCOMPRESSED) {
    //printf("Cray XYZ routing\n");
    std::map<std::pair<int,int>, std::vector<int>* >::iterator route_iter = routes.routes.find( std::make_pair(src, dst));
    if(route_iter != routes.routes.end()) { // if we have a special route ...
      std::vector<int>* rt = route_iter->second;
      //printf("found a special route for %i->%i (length %i): ", src, dst, path.size());
      for(int i=0; i<rt->size(); ++i) {
        //printf(" %i", (*rt)[i]);
        path.push_back((*rt)[i]); // could use std::copy
      }
      //printf("\n");
    } else {
      printf("route %i->%i not found!\n", src, dst);
    }
  } else {
    // probably invoke RouteGetNextHop here until we're at the destination!
    printf("routing mode not supported\n");
  }
}

