/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_PROCESS_APP_H_INCLUDED
#define SSTMAC_SOFTWARE_PROCESS_APP_H_INCLUDED

#include <sstmac/software/libraries/compute/lib_compute_inst.h>
#include <sstmac/software/libraries/compute/lib_compute_time.h>
#include <sstmac/software/libraries/compute/lib_compute_memmove.h>
#include <sstmac/software/libraries/compute/lib_sleep.h>
#include <sstmac/software/libraries/compute/lib_compute_loops.h>
#include <sstmac/common/ptr_type.h>
#include <sstmac/common/factories/factory.h>
#include <sstmac/software/process/thread.h>
#include <sstmac/common/sim_parameters.h>

namespace sstmac
{
  namespace sw
  {
    class operatingsystem;
    class launchinfo;
    /**
     * The app derived class adds to the thread base class by providing
     * facilities to allow applications to simulate computation.
     * Messaging models are supported through an api class,
     * which are stored by the app
     */
    class app :
            public thread,
            public factory_type
    {
      lib_compute_inst::ptr computeinst_;
      lib_compute_time::ptr computetime_;
      lib_compute_memmove::ptr computememmove_;
      lib_sleep::ptr sleeplib_;
      lib_compute_loops::ptr computeloops_;

    protected:
      sim_parameters::ptr params_;

      boost::intrusive_ptr<launchinfo> linfo_;

      app(const sim_parameters::ptr& params);

      app(softwareid id);

      static debug dbg_;

      softwareid id_;

      static std::vector<char**> argv_for_app_;

      static std::vector<int> argc_for_app_;

      void init_argv();

      virtual void init_mem_lib();

    public:
      typedef boost::intrusive_ptr<app> ptr;

      softwareid
      get_sid() const;

      int
      get_appnum() const;

      app::ptr
      get_parent_app() const {
        return ptr(const_cast<app*>(this));
      }

      ptr_type::ptr
      build_api(int aid, const std::string &name);

      virtual void
      init_os(const OS_API::ptr &os);

      void
      set_launch_info(const boost::intrusive_ptr<launchinfo> &linf);

      boost::intrusive_ptr<launchinfo>
      get_launch_info();

      // convenience functions for apps
      // public to allow C interface
      void
      compute(timestamp time);

      void
      compute_inst(const eventdata::ptr &cmsg);

      void
      compute_block_read(long bytes);

      void
      compute_block_write(long bytes);

      void
      compute_block_memcpy(long bytes);

      lib_compute_loops::ptr
      get_compute_loops_lib();

      void
      sleep(timestamp time);

      std::string
      get_compute_name();

      std::string
      get_compute_inst_name();

      std::string
      get_sleep_name();

      std::string
      get_compute_memmove_name();

      /// Goodbye.
      virtual ~app();

      char** get_argv();

      int get_argc();
        
      virtual void
      consume_params(const sim_parameters::ptr &params) = 0;

      virtual app::ptr
      clone(softwareid newid) = 0;

      //called when killing the app, in case you want to check or clean anything up before destructor
      virtual void kill();

      sim_parameters::ptr parameters();

      virtual void
      init_libs(sim_parameters::ptr params);

      virtual void
      skeleton_main() = 0;

      virtual void run();

    };

    /** utility function for computing stuff */
    void compute_time(double tsec);

    class app_factory : public SSTFactory<app>
    {

    public:
      static void
      print_apps();

      static void
      clear_apps();

    };

  }
} // end of namespace sstmac.

#define SSTRegisterApp(appstr, appname, ...) \
    SSTRegister(appstr, ::sstmac::sw::app, appname, __VA_ARGS__)

#define sstmac_register_app(cls) \
static int cls##_main(int argc, char** argv); \
class cls :  \
    public ::sstmac::sw::app \
{ \
    public: \
      typedef ::boost::intrusive_ptr<cls> ptr; \
      typedef ::boost::intrusive_ptr<const cls> const_ptr; \
    protected: \
      cls(::sstmac::sw::softwareid id) \
          : ::sstmac::sw::app(id) \
      { \
      } \
     cls(const sstmac::sim_parameters::ptr& params) \
          : app(params) \
      { \
      } \
public: \
  ~cls() throw () {} \
\
  static cls::ptr \
  construct(::sstmac::sw::softwareid id) { \
      return new cls(id); \
  } \
\
  static ptr \
  construct(const ::sstmac::sim_parameters::ptr& params) \
  { \
    return new cls(params); \
  } \
  ::sstmac::sw::app::ptr \
  clone(::sstmac::sw::softwareid newid) { \
      return construct(newid); \
  } \
\
  void consume_params(const ::sstmac::sim_parameters::ptr &params) \
  { \
  } \
    \
  std::string to_string() const { \
    return "new skeleton"; \
  } \
\
  void skeleton_main() { \
      int argc = get_argc(); \
      char** argv = get_argv(); \
      cls##_main(argc, argv); \
  } \
}; \
SSTRegister(#cls, ::sstmac::sw::app, cls);


#endif
