/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <stdlib.h>
#include <sstream>

#include <sstmac/common/sstmac-config.h>
#include <sstmac/common/config.h>

#include <sstmac/software/process/operatingsystem.h>

#include <sstmac/software/process/app.h>
#include <sstmac/software/process/key.h>
#include <sstmac/software/ami/environment.h>
#include <sstmac/common/messages/sst_message.h>
#include <sstmac/common/eventcallback.h>

#include <sstmac/common/errors.h>

#include <sstmac/software/threading/threading_pth.h>
#include <sstmac/software/threading/threading_pthread.h>
#include <sstmac/software/threading/threading_ucontext.h>

#include <sstmac/software/services/service.h>
#include <sstmac/hardware/processor/computescheduler.h>
#include <sstmac/software/process/graphviz.h>
#include <sstmac/software/process/ftq.h>

#include <sstmac/software/services/launch/launcher.h>
#include <sstmac/hardware/node/node.h>

#include <sstmac/software/services/launch/messages/complete_message.h>

namespace sstmac
{
  namespace sw
  {

    thread::ptr OS_API::current_thread_ = NULL;

    GraphViz::ptr operatingsystem::call_graph_ = 0;

    FTQCalendar::ptr operatingsystem::ftq_trace_ = 0;


    event_trace::ptr operatingsystem::event_trace_ = 0;

    //true to add guard bands - I don't want to debug this otherwise
    stackalloc operatingsystem::stackalloc_;

    std::list<thread::ptr> operatingsystem::to_delete_;

    nested_type<int, int, threading_interface*>::map operatingsystem::context_map_;

    operatingsystem* operatingsystem::current_os_ = 0;

    appid operatingsystem::current_aid_(-1);

    taskid operatingsystem::current_tid_(-1);

    debug operatingsystem::dbg_;

    bool statics_done = false;

    operatingsystem::operatingsystem(hardware_descriptor::ptr &hw) :
      ncores_(hw->get_cores()), current_thread_id_(thread::main_thread)
    {
      if (!dbg_.is_real())
        {
          dbg_ = debug("<debug> os | operatingsystem");
          logger::register_logger(&dbg_);
        }

      if (!call_graph_)
        {
          call_graph_ = GraphViz::construct();
          logger::register_logger(call_graph_.get());
        }

      if (!ftq_trace_)
      {
          ftq_trace_ = FTQCalendar::construct();
          logger::register_logger(ftq_trace_.get());
          if (ftq_trace_->is_active()){
              timestamp epoch_time = sstmac_env::params->get_time_param("ftq_epoch");
              bool show_visualization = sstmac_env::params->get_optional_bool_param("ftq_vis_show", false);
              bool show_vis_thread_chart = sstmac_env::params->get_optional_bool_param("ftq_vis_show_thread_chart", false);
              std::string show_vis_thread_chart_type = sstmac_env::params->get_optional_param("ftq_vis_show_thread_chart_type", "bar");
              bool show_vis_thread_chart_separate = sstmac_env::params->get_optional_bool_param("ftq_vis_show_thread_chart_separate", true);
              bool show_vis_summary_chart = sstmac_env::params->get_optional_bool_param("ftq_vis_show_summary_chart", false);
              int64_t epoch_nticks = epoch_time.ticks_int64();
              ftq_trace_->init(
                      epoch_nticks,
                      show_visualization,
                      show_vis_thread_chart,
                      show_vis_thread_chart_type,
                      show_vis_thread_chart_separate,
                      show_vis_summary_chart);
          }
      }

      if(!event_trace_){
        event_trace_ = event_trace::construct("<trace> events", "eventrace.dat");
        logger::register_logger(event_trace_.get());
      }


      restarting_ = false;

      if (!stackalloc_.initialized())
        {
          int stacksize = sstmac_env::params->get_optional_int_param(
              "stack_size", 1 << 17);
          int chunksize = sstmac_env::params->get_optional_int_param(
              "stack_chunk_size", 1 << 21);
          bool mprot = sstmac_env::params->get_optional_bool_param(
              "stack_protect", false);
          stackalloc_.init(stacksize, chunksize, mprot);
        }

      std::vector<std::string> libs_needed;
      if (sstmac_env::params->has_param("os_startup_libs")){
          sstmac_env::params->get_vector_param("os_startup_libs", libs_needed);
      }

      std::vector<std::string>::const_iterator it, end = libs_needed.end();
      for (it=libs_needed.begin(); it != end; ++it)
      {
          const std::string& libname = *it;
          library::ptr lib = library::construct_lib(libname);
          register_lib(lib);
      }

      const char *threading_pchar = getenv("SSTMAC_THREADING");
      if (threading_pchar)
        {
          // Different variants of getenv return different std::strings.
          // Adjust the threading_string to point to just the value.
          if (!strncmp("SSTMAC_THREADING=", threading_pchar, 17))
            {
              threading_pchar = strchr(threading_pchar, '=');
            }
          if (*threading_pchar == '=')
            threading_pchar++;
          std::string threading_string(threading_pchar);
          if (threading_string == "pth")
            {
#       if defined(SSTMAC_HAVE_PTH)
              des_context_ = new threading_pth();
#       else
              throw valueerror(
                  "operatingsystem: SSTMAC_THREADING=pth is not supported");
#       endif
            }
          else if (threading_string == "pthread")
            {
#       if defined(SSTMAC_HAVE_PTHREAD)
              des_context_ = new threading_pthread();
#       else
              throw valueerror(
                  "operatingsystem: SSTMAC_THREADING=pthread is not supported");
#       endif
            }
          else if (threading_string == "ucontext")
            {
#       if defined(SSTMAC_HAVE_UCONTEXT)
              des_context_ = new threading_ucontext();
#       else
              std::cerr << "SSTMAC_THREADING=ucontext is not supported"
                  << std::endl;
              abort();
              //throw valueerror(
              //    "operatingsystem: SSTMAC_THREADING=ucontext is not supported");
#       endif
            }
          else
            {
              sst_throw(valueerror, "operatingsystem: invalid value %s"
                  " for SSTMAC_THREADING environmental variable.\n"
                  "choose one of pth, pthread, ucontext.\n", threading_string.c_str());
            }
        }
      else
        {
#     if defined(SSTMAC_USE_PTH)
          des_context_ = new threading_pth();
#     elif defined(SSTMAC_USE_PTHREAD)
          des_context_ = new threading_pthread();
#     elif defined(SSTMAC_USE_UCONTEXT)
          des_context_ = new threading_ucontext();
#     else
#       error "no default threading library specified"
#     endif
        }

      des_context_->init_context();

      threadstack_.push(thread_data_t(des_context_, NULL));
      current_thread_ = threadstack_.top().second;
    }

    operatingsystem::~operatingsystem()
    {
      if (des_context_)
        {
          delete des_context_;
        }
    }

    void
    operatingsystem::clear_statics()
    {
      call_graph_ = 0;
      stackalloc_.clear();
    }

    //
    // Allocate an object of this type.
    //
    operatingsystem::ptr
    operatingsystem::construct(hardware_descriptor::ptr &hw)
    {
      operatingsystem::ptr retval(new operatingsystem(hw));

      return retval;
    }

    void
    operatingsystem::execute_kernel(AMI::AMI_HW_FUNC func,
        const sst_message::ptr& data)
    {
      AMI* ami = getAMI(AMI::AMI_HW);
      if (ami == NULL)
        sst_throw(valueerror, "operatingsystem: can't find AMI %s", AMI::tostr(func));
      ami->execute_kernel(func, data);
    }

    ptr_type::ptr
    operatingsystem::execute_kernel(AMI::AMI_HW_FUNC func)
    {
      AMI* ami = getAMI(AMI::AMI_HW);
      if (ami == NULL)
        sst_throw(valueerror, "operatingsystem: can't find AMI %s", AMI::tostr(func));
      return ami->execute_kernel(func);
    }

    void
    operatingsystem::execute_kernel(AMI::AMI_DISP_FUNC func,
        const sst_message::ptr& data)
    {
      AMI* ami = getAMI(AMI::AMI_DISP);
      if (ami == NULL)
        sst_throw(valueerror, "operatingsystem: can't find AMI %s", AMI::tostr(func));
      ami->execute_kernel(func, data);
    }

    ptr_type::ptr
    operatingsystem::execute_kernel(AMI::AMI_DISP_FUNC func)
    {
      AMI* ami = getAMI(AMI::AMI_DISP);
      if (ami == NULL)
        sst_throw(valueerror, "operatingsystem: can't find AMI %s", AMI::tostr(func));
      return ami->execute_kernel(func);
    }

    void
    operatingsystem::execute_kernel(AMI::AMI_QUERY_FUNC func,
        const sst_message::ptr& data)
    {
      AMI* ami = getAMI(AMI::AMI_QUERY);
      if (ami == NULL)
        sst_throw(valueerror, "operatingsystem: can't find AMI %s", AMI::tostr(func));
      ami->execute_kernel(func, data);
    }

    ptr_type::ptr
    operatingsystem::execute_kernel(AMI::AMI_QUERY_FUNC func)
    {
      AMI* ami = getAMI(AMI::AMI_QUERY);
      if (ami == NULL)
        sst_throw(valueerror, "operatingsystem: can't find AMI %s", AMI::tostr(func));
      return ami->execute_kernel(func);
    }

    void
    operatingsystem::execute_kernel(AMI::AMI_COMP_FUNC func,
        const sst_message::ptr& data)
    {
      AMI* ami = getAMI(AMI::AMI_COMPUTE);
      if (ami == NULL)
        sst_throw(valueerror, "operatingsystem: can't find AMI %s", AMI::tostr(func));
      ami->execute_kernel(func, data);
    }

    ptr_type::ptr
    operatingsystem::execute_kernel(AMI::AMI_COMP_FUNC func)
    {
      AMI* ami = getAMI(AMI::AMI_COMPUTE);
      if (ami == NULL)
        sst_throw(valueerror, "operatingsystem: can't find AMI %s", AMI::tostr(func));
      return ami->execute_kernel(func);
    }

    void
    operatingsystem::execute_kernel(AMI::AMI_COMM_FUNC func,
        const sst_message::ptr& data)
    {
      AMI* ami = getAMI(AMI::AMI_COMM);
      if (ami == NULL)
        sst_throw(valueerror, "operatingsystem: can't find AMI %s", AMI::tostr(func));
      ami->execute_kernel(func, data);
    }

    ptr_type::ptr
    operatingsystem::execute_kernel(AMI::AMI_COMM_FUNC func)
    {
      AMI* ami = getAMI(AMI::AMI_COMM);
      if (ami == NULL)
        sst_throw(valueerror, "operatingsystem: can't find AMI %s", AMI::tostr(func));
      return ami->execute_kernel(func);
    }

    bool
    operatingsystem::kernel_supported(AMI::AMI_COMP_FUNC func) const
    {
      AMI* ami = getAMI(AMI::AMI_COMPUTE);
      if (ami == NULL)
        {
          return false;
        }
      return ami->kernel_supported(func);
    }

    bool
    operatingsystem::kernel_supported(AMI::AMI_COMM_FUNC func) const
    {
      AMI* ami = getAMI(AMI::AMI_COMM);
      if (ami == NULL)
        {
          return false;
        }
      return ami->kernel_supported(func);
    }

    bool
    operatingsystem::kernel_supported(AMI::AMI_DISP_FUNC func) const
    {
      AMI* ami = getAMI(AMI::AMI_DISP);
      if (ami == NULL)
        {
          return false;
        }
      return ami->kernel_supported(func);
    }

    bool
    operatingsystem::kernel_supported(AMI::AMI_QUERY_FUNC func) const
    {
      AMI* ami = getAMI(AMI::AMI_QUERY);
      if (ami == NULL)
        {
          return false;
        }
      return ami->kernel_supported(func);
    }

    bool
    operatingsystem::kernel_supported(AMI::AMI_HW_FUNC func) const
    {
      AMI* ami = getAMI(AMI::AMI_HW);
      if (ami == NULL)
        {
          return false;
        }
      return ami->kernel_supported(func);
    }

    AMI*
    operatingsystem::getAMI(AMI::AMI_CAT cat) const
    {
      std::map<AMI::AMI_CAT, AMI*>::const_iterator it = amis_.find(cat);
      if (it == amis_.end())
        {
          return NULL;
        }

      return it->second;
    }

    void
    operatingsystem::setAMI(AMI::AMI_CAT cat, AMI* a)
    {
      amis_[cat] = a;
    }

    // ------- THREADING functions ----------

    operatingsystem*
    operatingsystem::get_current_os()
    {
      return current_os_;
    }

    void
    operatingsystem::switch_to_context(int aid, int tid)
    {
#if ENABLE_DEBUG_SWAP
        thread_data_t from = current_os_->threadstack_.top();
        threading_interface* thr = context_map_[aid][tid];
        current_os_->threadstack_.push(thread_data_t(thr, NULL));
        from.first->swap_context(thr);
#else
      sst_throw(illformederror, "operating system is not configured to track code positions");
#endif
    }

    environment::ptr
    operatingsystem::get_current_env()
    {
      return launcher::get_env(current_aid_);
    }

    library::ptr
    operatingsystem::get_current_library(const std::string &name)
    {
      return current_os_->get_lib(name);
    }

    appid
    operatingsystem::get_current_aid()
    {
      return current_aid_;
    }

    taskid
    operatingsystem::get_current_tid()
    {
      return current_tid_;
    }

    nodeaddress::ptr
    operatingsystem::get_current_node()
    {
      nodeaddress::ptr addr = launcher::get_node(current_aid_, current_tid_);
      return addr;
    }

    void
    operatingsystem::switch_to_thread(thread_data_t tothread)
    {

      long old_id = current_thread_id_;
      SSTMAC_DEBUG << "os::switch_thread from " << current_thread_id_ << " to "
            << tothread.second->get_id() << "\n";

      thread_data_t from = threadstack_.top();
      threadstack_.push(tothread);

      SSTMAC_DEBUG << "os: node address: " << get_my_addr()->to_string()
            << "\n";
      SSTMAC_DEBUG << "os: size of threadstack: " << threadstack_.size()
            << "\n";

      thread::ptr next_thread = tothread.second;
      current_thread_id_ = next_thread->get_id();

      int strict_affinity = next_thread->get_affinity();
      thread_scheduler_->schedule_affinity(next_thread->get_id(),
          strict_affinity);

      current_thread_ = tothread.second;
      current_thread_id_ = get_current_threadid();
      current_os_ = this;
      current_aid_ = current_thread_->aid();
      current_tid_ = current_thread_->tid();

      from.first->swap_context(tothread.first);

      /** back to main thread */
      current_thread_id_ = old_id;
      SSTMAC_DEBUG << "os: node address: " << get_my_addr()->to_string()
            << "\n";
      SSTMAC_DEBUG << "os: size of threadstack: " << threadstack_.size()
            << "\n";
      SSTMAC_DEBUG << "os::switch_thread : back to " << current_thread_id_
            << "\n";
    }

    void
    operatingsystem::cleanup()
    {
      if (to_delete_.empty())
        return;

      std::list<thread::ptr>::iterator it, end = to_delete_.end();
      std::list<thread::ptr>::iterator it2, end2 = threads_.end();
      thread::ptr thr;
      for (it = to_delete_.begin(); it != end; it++)
        {
          thr = *it;
          thr->destroy();
          for (it2 = threads_.begin(); it2 != end2; it2++)
            {
              if (*it2 == thr)
                {
                  threads_.erase(it2);
                  break;
                }
            }
        }
      to_delete_.clear();
    }

    void
    operatingsystem::print_libs(std::ostream &os) const
    {
      std::map<std::string, library::ptr>::const_iterator it = libs_.begin(), end =
          libs_.end();
      os << "available libraries: \n";
      for (; it != end; ++it)
        {
          os << it->first << "\n";
        }
    }

    void
    operatingsystem::finish()
    {
      std::map<std::string, library::ptr>::iterator it, end = libs_.end();
      for (it = libs_.begin(); it != end; it++)
        {
          service::ptr ser = boost::dynamic_pointer_cast<service>(it->second);
          if (ser)
            {
              ser->finish();
            }
        }
      thread_scheduler_ = 0;
    }

    timestamp
    operatingsystem::block(const key::ptr &req)
    {
      if (req == 0)
        {
          throw nullerror("operatingsystem::block:  Refusing to block "
            "pending a null key pointer.");
        }
      // OK, we're ready to roll.
      //blocked_.insert(std::make_pair(req, current_));
      if (threadstack_.size() == 1)
        {
          throw ssterror(
              "OS: trying to block the DES thread (bottom of the stack)");
        }
      thread_data_t top = threadstack_.top();
      req->blocking_.insert(top);
      // blocked_count_++;
      SSTMAC_DEBUG << "operatingsystem::block: " << "blocking thread "
            << current_thread_id_ << " with req=" << (long) req.get()
            << " req->blocking_.size()=" << req->blocking_.size() << "\n";
      //<< ", leaving " << active_.size() << " contexts pending.\n";
      // swap_context(*current_, caller_context_);

      int64_t before_ticks = now().ticks_int64();

      threadstack_.pop();
      current_thread_ = threadstack_.top().second;
      /** Block the currently running thread - and start the next waiting thread */
      top.first->swap_context(threadstack_.top().first);

      int64_t after_ticks = now().ticks_int64();
      int64_t delta_ticks = after_ticks - before_ticks;
      if (call_graph_->is_active()){
          call_graph_->count_trace(delta_ticks, current_thread_);
      }
      if (ftq_trace_->is_active()){
        ftq_trace_->collect(req->event_typeid(), current_aid_.id, current_tid_.id, before_ticks, delta_ticks);
      }
      if(event_trace_->is_active()){
        event_trace_->collect(
                    req->event_typeid(),
                    req->get_name(),
                    get_my_addr(),
                    current_thread_->get_affinity(),
                    current_thread_->threadid_,
                    current_aid_.id, current_tid_.id,
                    before_ticks, delta_ticks);
      }

      //SSTMAC_DEBUG << "os: this: " << (long) this << "\n";
      SSTMAC_DEBUG << "os: node address: " << get_my_addr()->to_string()
            << "\n";
      SSTMAC_DEBUG << "operatingsystem::block: " << " done blocking thread "
            << current_thread_id_ << ", size of threadstack: "
            << threadstack_.size() << "\n";

      return now();
    }

    timestamp
    operatingsystem::unblock(const key::ptr &req)
    {
      SSTMAC_DEBUG << "operatingsystem::unblock: " << " unblocking "
            << req->blocking_.size() << " threads \n";

      key::blocking_t::iterator it, end = req->blocking_.end();
      key::blocking_t templist;

      //save to a temp list first, so we can clear the key before we start switching to threads
      for (it = req->blocking_.begin(); it != end; ++it)
        {
          templist.insert(*it);
        }

      req->blocking_.clear();

      for (it = templist.begin(); it != templist.end(); ++it)
        {
          thread::ptr td = (*it).second;
          //std::cout << "switching to thread " << td->get_id()
          //          <<  " on key " << (long) req.get()
          //           << "\n";
          switch_to_thread((*it));
        }

      return now();
    }

    void
    operatingsystem::start_thread(const thread::ptr &t)
    {
#if HAVE_GRAPHVIZ
        {
            void** stack = call_graph_->allocate_trace();
            t->set_backtrace(stack);
        }
#endif
      add_thread(t);
#if ENABLE_DEBUG_SWAP
      int aid = t->aid().id;
      int tid = t->tid().id;
      context_map_[aid][tid] = t->context_;
#endif
      switch_to_thread(thread_data_t(t->context_, t));

    }

    void
    operatingsystem::join_thread(const thread::ptr &t){
      SSTMAC_DEBUG << "operatingsystem::join thread \n";
      if(t->get_state() != thread::DONE){
          key::ptr k = key::construct();
          t->joiners_.push(k);
          block(k);
      }
    }

    void
    operatingsystem::complete_thread(bool succ)
    {
      thread::ptr thr_todelete = threadstack_.top().second;
      threading_interface* current_tinfo = threadstack_.top().first;

      //if any threads waiting on the join, unblock them
      while (thr_todelete->joiners_.size() > 0)
        {
          SSTMAC_DEBUG << "unblocking joiner \n";
          unblock(thr_todelete->joiners_.front());
          //to_awake_.push(thr_todelete->joiners_.front());
          thr_todelete->joiners_.pop();
        }

      threadstack_.pop();

      SSTMAC_DEBUG << "os::complete: " << "Context " << (long) current_tinfo
            << "\n";

      current_thread_ = threadstack_.top().second;
      threading_interface* next_tinfo = threadstack_.top().first;

      to_delete_.push_back(thr_todelete);

      //tell the launcher that i'm done
      app::ptr a = boost::dynamic_pointer_cast<app>(thr_todelete);
      if (a)
        {
          launchinfo::ptr linfo = a->get_launch_info();
          if (linfo) //if i have linfo, i was launched.
            {
              complete_message::ptr com = complete_message::construct(linfo,
                  succ);
              com->set_time(timestamp(10e-9));
              amis_[AMI::AMI_COMPUTE]->execute_kernel(AMI::AMI_COMP_SLEEP, com);
            }
          task_to_thread_.erase(a->get_sid().task_);
        }



      //nullify references to current thread
      a = 0;
      thr_todelete = 0;

      current_tinfo->complete_context(next_tinfo);
      // This thread is not permitted to start again.
      throw illformederror(
          "operatingsystem::complete:  Thread switched in again "
            "after calling complete and relinquishing stack.");
    }

    // ---------------------

    nodeaddress::ptr
    operatingsystem::get_my_addr() const
    {
      if (amis_.find(AMI::AMI_QUERY) == amis_.end())
        {
          throw unimplementederror(
              "operatingsystem::get_my_addr can't find query ami");
        }

      ptr_type::ptr p = amis_.find(AMI::AMI_QUERY)->second ->execute_kernel(
          AMI::AMI_QUERY_ADDRESS);
      nodeaddress::ptr casted = boost::dynamic_pointer_cast<nodeaddress>(p);
      if (!p)
        {
          throw nullerror(
              "operatingsystem::get_my_addr - error casting to nodeaddress");
        }

      return casted;
    }

    bool
    operatingsystem::is_task_here(const taskid &id) const
    {
      std::map<taskid, long>::const_iterator it = task_to_thread_.find(id);
      return it != task_to_thread_.end();

    }

    std::string
    operatingsystem::get_current_hostname()
    {
      return get_current_env()->get_hostname(current_tid_);
    }

    long
    operatingsystem::get_task_threadid(const taskid &id) const
    {
      if (id.id == taskid::unbound_task || task_to_thread_.size() == 0)
        return thread::main_thread;

      std::map<taskid, long>::const_iterator it = task_to_thread_.find(id);
      if (it == task_to_thread_.end())
        {
          std::map<taskid, long>::const_iterator it, end =
              task_to_thread_.end();
          for (it = task_to_thread_.begin(); it != end; it++)
            {
              std::cerr << "we have task " << it->first.id << "\n";
            }
          sst_throw(valueerror, "invalid task id %d passed to operating system on node %s",
              id.id, get_my_addr()->to_string().c_str());
        }
      return it->second;
    }

    nodeaddress::ptr
    operatingsystem::get_task_addr(softwareid sid) const
    {
      if (env_.find(sid.app_) == env_.end())
        {
          sst_throw(oserror,
              "operatingsystem::get_task_addr(%s) - no environment entry for this app",
              sid.to_string().c_str()
          );
        }

      return env_.find(sid.app_)->second->get_node_address(sid.task_);
    }

    void
    operatingsystem::register_lib(const library::ptr &lib)
    {
      if (libs_.find(lib->get_lib_name()) != libs_.end())
        {
          sst_throw(oserror,
              "operatingsystem: trying to register a lib %s that is already registered on node %s",
              lib->get_lib_name().c_str(), get_my_addr()->to_string().c_str());
        }
      else if (lib->get_lib_name() == "")
        {
          throw oserror(
              "operatingsystem: trying to register a lib with no name");
        }
      SSTMAC_DEBUG << "OS: trying to register lib " << lib->get_lib_name()
            << "\n";
      libs_[lib->get_lib_name()] = lib;
      lib->init_os(ptr(this));
      lib->register_child_libs();
    }

    void
    operatingsystem::unregister_lib(const library::ptr &lib)
    {
      SSTMAC_DEBUG << "OS: trying to unregister lib " << lib->get_lib_name()
            << "\n";
      if (libs_.find(lib->get_lib_name()) == libs_.end())
        {
          /* std::cerr << "lib name: " << lib->get_lib_name() << std::endl;
           std::cerr << "but we do have: \n";
           std::map::iterator<std::string, library::ptr>:: it, end = libs_.end();
           for(it = libs_.begin(); it != end; it++){
           std::cerr << it->first << ", ";
           }
           throw oserror(
           "operatingsystem: trying to unregister a lib that is not registered");*/
        }
      libs_.erase(lib->get_lib_name());
    }

    void
    operatingsystem::register_ser(const service::ptr &lib)
    {
      register_lib(lib);
      lib->init_full_os(ptr(this));
    }

    library::ptr
    operatingsystem::get_lib(const std::string& name) const
    {
      std::map<std::string, library::ptr>::const_iterator it = libs_.find(name);
      if (it == libs_.end())
        return library::ptr();
      else
        return it->second;
    }

    environment::const_ptr
    operatingsystem::get_env(appid aid) const
    {
      if (env_.find(aid) == env_.end())
        {
          std::cout << "appid - " << aid.to_string() << std::endl;
          throw nullerror("os: environment not found");
        }
      return env_.find(aid)->second;
    }

    long
    operatingsystem::get_current_threadid() const
    {
      if (threadstack_.size() > 0)
        {
          thread_data_t td = get_current_context();
          if (td.first != des_context_)
            return td.second->get_id();
          return -1;
        }
      else
        {
          return -1;
        }
    }

    void
    operatingsystem::stack_check()
    {
      thread::ptr thr = OS_API::get_current_thread();
      if (!thr)
        return; //main thread

      if (thr->get_id() == thread::main_thread)
        return;

      void* stack = thr->stack();

      int x;
      void* testptr = &x;

      //stack is usually allocated backwards - we haven't overrun yet
      if (testptr > stack)
        return;

      sst_throw(illformederror,
          "stack pointer on thread %d has gone past end of stack",
          thr->get_id());
    }

    timestamp
    operatingsystem::now() const
    {
      return eventman_->now();
    }

    void
    operatingsystem::set_eventmanager(const eventmanager::ptr &man)
    {
      OS_API::set_eventmanager(man);
      if (!statics_done)
        {
          man->register_stat(call_graph_);
          man->register_stat(ftq_trace_);
          statics_done = true;
        }
    }

    void
    operatingsystem::add_thread(const thread::ptr &t)
    {
      t->init_thread(des_context_, stackalloc_.alloc(),
          stackalloc_.stacksize(), operatingsystem::ptr(this), NULL);

      threads_.push_back(t);
    }

    void
    operatingsystem::add_application(const app::ptr& a, launchinfo::ptr &linfo)
    {
      ftq_trace_->register_app(a->aid().id, sstprintf("app%d", a->aid().id));

      a->consume_params(linfo->params_);
      a->set_launch_info(linfo);

      add_thread(a);

      task_to_thread_[a->get_sid().task_] = a->get_id();
      //thread_to_task_[a->get_id()] = a->get_sid().task_;
    }

    void
    operatingsystem::add_task(const taskid& id)
    {
      task_to_thread_[id] = get_current_threadid();
      // thread_to_task_[get_current_threadid()] = id;
    }

    void
    operatingsystem::start_apps()
    {
      std::list<thread::ptr>::iterator it;

      for (it = threads_.begin(); it != threads_.end(); it++)
        {
          thread::ptr thr = *it;
          if (!thr->is_started())
            {
              // current_ = &((*it)->context_);
              //swap_context(caller_context_, (*it)->context_);
#if ENABLE_DEBUG_SWAP
              int aid = thr->aid().id;
              int tid = thr->tid().id;
              context_map_[aid][tid] = thr->context_;
#endif
#if HAVE_GRAPHVIZ
            {
                void** stack = call_graph_->allocate_trace();
                thr->set_backtrace(stack);
            }
#endif
              switch_to_thread(thread_data_t(thr->context_, thr));
            }
        }

      //check pending messages
      int psize = pending_messages_.size();
      for (int i = 0; i < psize; i++)
        {
          sst_message::ptr msg = pending_messages_.front();
          pending_messages_.pop_front();
          handle_message(msg);
        }
    }

    void
    operatingsystem::handle_message(const sst_message::ptr &msg)
    {
      if (msg->has_callback())
        {
          eventhandler::ptr callback = msg->pop_callback();
          callback->handle(msg);
          return;
        }

      std::string libn = msg->get_lib_name();
      std::map<std::string, library::ptr>::const_iterator it = libs_.find(libn);
      if (it == libs_.end())
        {
          SSTMAC_DEBUG << "OS[" << get_my_addr()->to_string()
                << "]: didn't find lib named " << libn << "\n";
          /* if (pending_messages_.size() > 10)
           std::cerr << "WARNING: OS storing message for lib name: " << libn
           << ", size of pending_messages queue is "
           << (pending_messages_.size() + 1) << std::endl; */

          pending_messages_.push_back(msg);
          std::stringstream sstr;

          hw::node* mynode = dynamic_cast<hw::node*>(this->getAMI(AMI::AMI_COMPUTE));

          std::cerr << "Valid libraries on " << this << ":\n";
          std::map<std::string, library::ptr>::const_iterator it, end = libs_.end();
          for  (it = libs_.begin(); it != end; ++it){
            std::cerr << it->first << std::endl;
          }

          sst_throw(oserror, "operatingsystem::handle_message: can't find library %s"
                    " for message %s on %s",
                    msg->get_lib_name().c_str(), msg->to_string().c_str(),
                    mynode ? mynode->to_string().c_str() : "node");

          if (!restarting_)
            {
              SSTMAC_DEBUG << "OS[" << get_my_addr()->to_string()
                    << "]: queueing up \n";
              pending_messages_.push_back(msg);
            }
        }
      else
        {
          SSTMAC_DEBUG << "OS[" << get_my_addr()->to_string()
                << "]: delivering to lib " << libn << "\n";
          it->second->incoming_message(msg);
        }
    }

    void
    operatingsystem::init_scheduler(const hw::computescheduler::ptr& sched)
    {
      thread_scheduler_ = sched;
    }

    void
    operatingsystem::apps_kill()
    {
      std::list<thread::ptr>::iterator it;

      if (dbg_.is_active())
        dbg_ << "OS: killing threads \n";
      for (it = threads_.begin(); it != threads_.end(); it++)
        {
          (*it)->unregister_all_libs();

          to_delete_.push_back(*it);
        }

    }

    void
    operatingsystem::kill_app(app::ptr app)
    {
      app->unregister_all_libs();
      to_delete_.push_back(app);
    }

    std::list<app::ptr>
    operatingsystem::get_app_ptrs(appid aid)
    {
      cleanup(); //just in case

      std::list<app::ptr> ret;
      std::list<thread::ptr>::iterator iter, end = threads_.end();
      for (iter = threads_.begin(); iter != end; iter++)
        {
          app::ptr a = boost::dynamic_pointer_cast<app>(*iter);
          if (a)
            {
              if (a->get_sid().app_ == aid)
                {
                  ret.push_back(a);
                }
            }
        }

      return ret;
    }

    app::ptr
    operatingsystem::get_app_ptr(softwareid sid)
    {
      cleanup(); //just in case

      app::ptr ret;
      std::list<thread::ptr>::iterator iter, end = threads_.end();
      for (iter = threads_.begin(); iter != end; iter++)
        {
          app::ptr a = boost::dynamic_pointer_cast<app>(*iter);
          if (a)
            {
              if (a->get_sid() == sid)
                {
                  return a;
                }
            }
        }

      return ret;
    }

  }
} //end of namespace sstmac
