/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/process/thread.h>
#include <sstmac/software/process/os_api.h>
#include <sstmac/software/process/key.h>
#include <sstmac/software/libraries/library.h>
#include <sstmac/common/errors.h>
#include <iostream>
#include <exception>
#include <unistd.h>  // getpagesize()
#include <sys/mman.h>
#include <memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <sstmac/software/process/operatingsystem.h>

#include <sstmac/software/api/mpi/mpiapi.h>

//#include <signal.h>
//#include <dlfcn.h> // dladdr (not a POSIX call)
//#include <cxxabi.h> // Demangle C++ names (not a POSIX call)

namespace sstmac
{
  namespace sw
  {

    long thread::THREAD_ID_CNT = 0;
    const appid thread::main_thread_aid(-1);
    const taskid thread::main_thread_tid(-1);

    //
    // Private method that gets called by the scheduler.
    //
    void
    thread::init_thread(threading_interface* threadcopy, void *stack,
        int stacksize, const OS_API::ptr& os, threading_interface *yield_to)
    {
      stack_ = stack;
      stacksize_ = stacksize;
      os_ = os;

      threadid_ = THREAD_ID_CNT++;

      p_txt_ = process_context::construct();

      isInit = true;

      //there were libs that wanted to be registered
      if (!pending_libs_.empty())
        {
          std::list<library::ptr>::iterator it;
          for (it = pending_libs_.begin(); it != pending_libs_.end(); it++)
            {
              os_->register_lib(*it);
            }
        }

      pending_libs_.clear();

      init_os(os);

      context_ = threadcopy->copy();

      threadinfo* info = new threadinfo();
      info->thethread = ptr(this);

      context_->start_context(stack, stacksize, run_routine, info, yield_to);

    }

    ptr_type::ptr
    thread::build_api(int aid, const std::string &name)
    {
        if (parent_app_){
            return parent_app_->build_api(aid, name);
        }
        else {
            sst_throw(illformederror, "thread::build_api: thread has no parent thread");
            return ptr_type::ptr();
        }
    }

    void
    thread::init_os(const OS_APIPtr &os)
    {
        //nothing by default
    }

    app::ptr
    thread::get_parent_app() const {
        return parent_app_;
    }

    /**
     * This can get called by anyone to have a thread exit, including during normal app termination
     */
    void
    thread::kill()
    {
      // We are done, ask the scheduler to remove this task from the
      // list of active tasks.
      state_ = DONE;
      // self->get_scheduler()->unblock(self->key_);
      // self->get_scheduler()->complete(self->stack_);
      unregister_all_libs();

      //smart ptrs really suck here
      if (os_->nreferences() == 1)
        {
          sst_throw(valueerror, "nreferences of OS reached a value that is too low before completing thread");
        }

      os_->complete_thread(true);
    }

    boost::intrusive_ptr<OS_API>
    thread::get_os() const
    {
      return os_;
    }

    /**
     * This is the function that gets called by the OS during thread cleanup
     */
    void
    thread::destroy()
    {
      context_->destroy_context();

      delete context_;
      operatingsystem::stackalloc_.free(stack_);

    }

    void*
    thread::stack() const
    {
      return stack_;
    }

    size_t
    thread::stacksize() const
    {
      return stacksize_;
    }

    //
    // Run routine that defines the initial context for this task.
    //
    void
    thread::run_routine(void* threadptr)
    {
      threadinfo* info = (threadinfo*) threadptr;
      assert(info);
      ptr self = info->thethread;
      delete info;

      if (self == 0)
        {
          // You silly person -- you probably forgot to call thread::set_self.
          throw nullerror(
              "thread::init:  thread::get_self returns a null pointer."
                "  You need to call thread::set_self when you construct "
                "a new (derived) thread object.");
        }
      // Go.
      if (self->initialized())
        {
          self->state_ = ACTIVE;
          bool success = false;
          try
            {
              self->run();
              success = true;
            }
          catch (const std::exception &ex)
            {
              std::cerr << "thread terminated with exception: " << ex.what()
                  << "\n";
              // should forward the exception to the main thread,
              // but for now will abort
              std::cerr << "aborting" << std::endl;
              abort();
            }
          catch (...)
            {
              std::cerr << "thread terminated with exception: aborting"
                  << std::endl;
              abort();
            }

          self->kill();
          self = 0;
        }
      else
        {
          throw oserror("thread::run_routine:  Task has not "
            "been initialized.");
        }
    }

    //
    // Constuctor.
    //
    thread::thread(const softwareid& sid) :
      state_(PENDING), isInit(false), backtrace_(0), bt_nfxn_(0),
          last_bt_collect_nfxn_(0), aid_(sid.app_), tid_(sid.task_),
          threadid_(thread::main_thread), key_(key::construct()),
          core_affinity_(thread::no_core_affinity)
    {
    }

    thread::thread() :
      state_(PENDING), isInit(false), backtrace_(0), bt_nfxn_(0),
          last_bt_collect_nfxn_(0), aid_(operatingsystem::get_current_aid()),
          tid_(operatingsystem::get_current_tid()),
          threadid_(thread::main_thread), key_(key::construct()),
          core_affinity_(thread::no_core_affinity)
    {
    }

    void
    thread::append_backtrace(void* fxn)
    {
      backtrace_[bt_nfxn_] = fxn;
      bt_nfxn_++;
    }

    void
    thread::pop_backtrace()
    {
      --bt_nfxn_;
      last_bt_collect_nfxn_ = std::min(last_bt_collect_nfxn_, bt_nfxn_);
    }

    void
    thread::collect_backtrace(int nfxn)
    {
      last_bt_collect_nfxn_ = nfxn;
    }

    void
    thread::register_lib(const boost::intrusive_ptr<library> &lib)
    {
      if (isInit)
        os_->register_lib(lib);
      else
        pending_libs_.push_back(lib);

      libs_.push_back(lib);
    }

    void
    thread::spawn(const thread::ptr &thr)
    {
        thr->parent_app_ = get_parent_app();
        os_->start_thread(thr);
    }

    void
    thread::unregister_lib(const boost::intrusive_ptr<library> &lib)
    {
      os_->unregister_lib(lib);
    }

    void
    thread::unregister_all_libs()
    {
      std::list<library::ptr>::iterator it, end = libs_.end();
      for (it = libs_.begin(); it != end; it++)
        {
          library::ptr lib = *it;
          lib->unregister_all_libs();
          unregister_lib(lib);
        }
      libs_.clear();
    }

    timestamp
    thread::now()
    {
      return os_->now();
    }

    nodeaddress::ptr
    thread::get_physical_address()
    {
      return os_->get_my_addr();
    }

    //
    // Goodbye.
    //
    thread::~thread()
    {
    }

    //
    // Start a new thread.
    //
    void
    thread::start_thread(const thread::ptr& thr)
    {
      thr->p_txt_ = p_txt_;
      os_->start_thread(thr);
    }

    void
    thread::set_affinity(int affinity)
    {
      core_affinity_ = affinity;
    }

    int
    thread::get_affinity() const
    {
      return core_affinity_;
    }

    //
    // Join threads.
    //
    void
    thread::join()
    {
      if (!this->initialized()){
          // We can't context switch the caller out without a scheduler.
          sst_throw(illformederror,
            "thread::join: target thread has not been initialized.");
      }
      os_->join_thread(ptr(this));
    }

    boost::intrusive_ptr<key>
    thread::get_join_key()
    {
      return key_;
    }

    bool
    thread::initialized() const
    {
      return isInit;
    }

  }
} // end of namespace sstmac
