/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_PROCESS_THREAD_H_INCLUDED
#define SSTMAC_SOFTWARE_PROCESS_THREAD_H_INCLUDED

#include <sstmac/software/process/softwareid.h>
#include <sstmac/common/nocopy.h>
#include <sstmac/common/debug.h>
#include <sstmac/common/nodeaddress.h>
#include <sstmac/common/ptr_type.h>
#include <sstmac/software/process/process_context.h>
#include <unistd.h>
#include <utility>
#include <sstmac/software/process/api.h>

namespace sstmac
{
  namespace sw
  {

    // operatingsystem and key need to include this header to get thread_data_t
    // Forward declare these rather than including them to avoid circular
    // dependency issues.
    class threading_interface;
    class key;
    class app;
    class library;
    class OS_API;

    // We need to know the thread associated with each threading_interface
    // so current thread pointer can be maintained
    // (used for threadstack_ in operatingsystem and block_ in key )
    class thread;
    typedef std::pair<threading_interface*, boost::intrusive_ptr<thread> >
        thread_data_t;



    /**
     * Base class for tasks that can be cooperative scheduled using threadcontext.
     */

    struct threadinfo
    {
      boost::intrusive_ptr<thread> thethread;
    };

    class thread :
        virtual public nocopy,
        public virtual ptr_type
    {
    protected:
        typedef boost::intrusive_ptr<key> keyPtr;
        typedef boost::intrusive_ptr<app> appPtr;
        typedef boost::intrusive_ptr<library> libraryPtr;
        typedef boost::intrusive_ptr<OS_API> OS_APIPtr;

    public:
      typedef boost::intrusive_ptr<thread> ptr;

      friend class operatingsystem;

      /// Help resolve deadlock situations.
      enum state
      {
        PENDING, ACTIVE, SUSPENDED, BLOCKED, DONE
      };

      virtual std::string
      to_string() const
      {
        return "thread";
      }

      void set_api(thread* thr)
      {
        thr->apis_ = apis_;
      }

      virtual ptr_type::ptr
      build_api(int aid, const std::string& name);

      virtual void
      init_os(const OS_APIPtr& os);

      template <class T>
      void
      add_api(const typename T::ptr& api)
      {
        if (API<T>::id == API<T>::null_id){
            sst_throw(illformederror, "API %s is not initialized."
                      "Make sure API is statically initialized correctly"
                      " using ImplementAPI macro.",
                      API<T>::name);
        }
        apis_[API<T>::id] = api;
      }

      template <class T>
      typename T::ptr
      get_api(){
        if (API<T>::id == API<T>::null_id){
            sst_throw(illformederror, "API %s is not initialized."
                      "Make sure API is statically initialized correctly"
                      " using ImplementAPI macro.",
                      API<T>::name);
        }

        ptr_type::ptr api = apis_[API<T>::id];
        if (!api){
            api = build_api(API<T>::id, API<T>::name);
        }

        typename T::ptr casted = boost::dynamic_pointer_cast<T>(api);
        if (!casted){
            sst_throw(valueerror, "Failed to cast API of type %s",
                      API<T>::name);
        }
        return casted;
      }

      virtual appPtr
      get_parent_app() const;

      static const int no_core_affinity = -1;
      static const int no_socket_affinity = -1;
      static const int main_thread = -1;
      static const int nic_thread = -2;
      static const int rdma_thread = -3;
      static const appid main_thread_aid;
      static const taskid main_thread_tid;

    protected:
      std::map<int, ptr_type::ptr> apis_;

      /// Monitor state for deadlock detection.
      state state_;

      /// Each thread can only run under one OS/scheduler.
      OS_APIPtr os_;

      std::queue<boost::intrusive_ptr<key> > joiners_;

      appPtr parent_app_; // who created this one. null if launch/os.

      process_context::ptr p_txt_;

    private:

      bool isInit;

      void** backtrace_;

      int bt_nfxn_;

      int last_bt_collect_nfxn_;

      /// The stack given to this thread.
      void *stack_;
      /// The stacksize.
      size_t stacksize_;

      appid aid_;

      taskid tid_;

      static long THREAD_ID_CNT;

      long threadid_;

      threading_interface* context_;

      /// This key gets used when other threads ask to wait (join) this thread.
      keyPtr key_;

      std::list<libraryPtr> pending_libs_;

      std::list<libraryPtr> libs_;

      int core_affinity_;

    private:

      void
      destroy();

      /// Run routine that defines the initial context for this task.
      /// This routine calls the virtual thread::run method.
      static void
      run_routine(void* threadptr);

    protected:
      /// Initialize private data.
      thread(const softwareid& sid);

      thread();

      void
      register_lib(const boost::intrusive_ptr<library> &lib);
      void
      unregister_lib(const boost::intrusive_ptr<library> &lib);
      void
      unregister_all_libs();

      timestamp
      now();

      nodeaddress::ptr
      get_physical_address();

    public:
      virtual
      ~thread();

      process_context::ptr
      get_process_context()
      {
        return p_txt_;
      }

      /// Get current thread state.
      state
      get_state() const
      {
        return state_;
      }

      appid aid() const
      {
          return aid_;
      }

      taskid tid() const
      {
          return tid_;
      }

      void
      spawn(const thread::ptr& thr);

      bool
      is_started()
      {
        return (state_ != PENDING);
      }

      long
      get_id() const
      {
        return threadid_;
      }

      virtual void
      kill();

      OS_APIPtr
      get_os() const;

      void*
      stack() const;

      size_t
      stacksize() const;

      void**
      get_backtrace() const
      {
        return backtrace_;
      }

      int
      get_last_backtrace_nfxn() const
      {
          return last_bt_collect_nfxn_;
      }

      int
      get_backtrace_nfxn() const
      {
          return bt_nfxn_;
      }

      virtual

      void
      append_backtrace(void* fxn);

      void
      pop_backtrace();

      void set_backtrace(void** bt){
          backtrace_ = bt;
      }

      void collect_backtrace(int nfxn);

      int
      get_affinity() const;

      void
      set_affinity(int affinity);

      /// method that gets called by the operating system.
      void
      init_thread(threading_interface* tocopy, void *stack, int stacksize,
          const OS_APIPtr& os, threading_interface *yield_to);

      /// Derived types need to override this method.
      virtual void
      run() = 0;

      /// A convenience request to start a new thread.
      /// The current thread has to be initialized for this to work.
      void
      start_thread(const thread::ptr& thr);

      void
      join();

      keyPtr
      get_join_key();

      /// Test whether the current task has been initialized (activated)
      /// by a scheduler.
      bool
      initialized() const;

    };

  }
} // end of namespace sstmac
#endif
