/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_BACKENDS_NATIVE_LAUNCH_ALLOCATIONSTRATEGY_H_INCLUDED
#define SSTMAC_BACKENDS_NATIVE_LAUNCH_ALLOCATIONSTRATEGY_H_INCLUDED

#include <sstmac/common/rng.h>
#include <sstmac/common/factories/factory.h>
#include <sstmac/common/debug.h>
#include <sstmac/common/ptr_type.h>
#include <sstmac/common/nodeaddress.h>
#include <sstmac/hardware/network/topology/topology.h>
#include <sstmac/common/sim_parameters.h>

namespace sstmac
{
  namespace sw
  {

    /**
     * Strategy type for assigning processes to nodes in a parallel run.
     *
     */
    class allocationstrategy : public factory_type
    {
    public:
      typedef std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp,
          std::allocator<nodeaddress::ptr> > node_set;

    protected:
      /// Check the arguments for calls to allocate and throw if they
      /// are bad.
      void
      check_request(int nnode, node_set &allocation);

      static debug dbg_;

      static hw::topology::ptr topology_;

      allocationstrategy();

      static node_set allocated_;
      static node_set available_;
      static bool node_map_built_;

    public:

      static void
      clear_statics();

      void
      init(const sstmac::hw::topology::ptr& top)
      {
        set_topology(top);

        long num_avail = top->num_nodes();

        if (!node_map_built_)
        {
          for (int i = 0; i < num_avail; ++i)
          {
            nodeaddress::ptr addr = nodeaddress::convert_from_uniqueid(i);
            available_.insert(addr);
          }
          node_map_built_ = true;
        }
      }

      virtual void
      set_topology(const sstmac::hw::topology::ptr& top)
      {
        topology_ = top;
      }

      static sstmac::hw::topology::ptr
      get_topology()
      {
        return topology_;
      }

      virtual std::string
      to_string() const
      {
        return "allocationstrategy";
      }
      typedef boost::intrusive_ptr<allocationstrategy> ptr;

      virtual
      ~allocationstrategy() throw ();

      /// Get nodes.
      /// @param nnode number of nodes requested
      /// @param allocation returns the nodes that have been allocated
      /// @param available the available set of nodes
      /// @param allocated the allocated set of nodes
      /// Nodes are removed from available and put into allocated.
      /// If not enough nodes are available, then this will throw.
      virtual void
      allocate(int nnode, node_set& allocation) = 0;

      virtual void
      release(node_set &alloc);

      virtual void
      remove_from_available(nodeaddress::ptr addr);
    };

    class firstavailableallocation : public allocationstrategy
    {
    protected:
      firstavailableallocation();
    public:
      typedef boost::intrusive_ptr<firstavailableallocation> ptr;

      static firstavailableallocation::ptr
      construct();

      static ptr
      construct(sim_parameters::ptr params)
      {
        return construct();
      }

      virtual
      ~firstavailableallocation() throw ();
      void
      allocate(int nnode_requested, node_set& allocation);

    };

    class randomallocation : public allocationstrategy
    {
      RNG::UniformInteger::ptr rng_;
    protected:

      randomallocation(unsigned seed);
      randomallocation(const RNG::UniformInteger::ptr &rng);
    public:
      typedef boost::intrusive_ptr<randomallocation> ptr;

      static randomallocation::ptr
      construct(unsigned seed = 1337);
      static randomallocation::ptr
      construct(const RNG::UniformInteger::ptr &rng);

      static ptr
      construct(sim_parameters::ptr params)
      {
        return construct();
      }

      virtual
      ~randomallocation() throw ();
      void
      allocate(int nnode_requested, node_set& allocation);

    };
    DeclareFactory1InitParam(allocationstrategy, hw::topology::ptr)
    ;

  }
} // end of namespace sstmac

#endif
