/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/services/launch/indexing/blockindexing.h>
#include <sstmac/common/errors.h>
#include <math.h>
#include <sstream>
#include <sstmac/software/services/launch/launcher.h>

namespace sstmac  { namespace sw
  {

SSTRegister("block", indexstrategy, blockindexing,
  "Tries to group consecutive MPI ranks on the same node (i.e. in a block). Otherwise, indexes in the same order as the allocation list.");


  //
  // Default constructor.
  //
  blockindexing::blockindexing() {
  }
  
  blockindexing::~blockindexing() throw() {
  }

  //
  // Static method to allocate object.
  //
  blockindexing::ptr blockindexing::construct(sim_parameters::ptr params) {
    return blockindexing::ptr(new blockindexing);
  }

  void blockindexing::allocate(const appid& aid,
                               const node_set &nodes,
                               int ppn,
                               std::vector<nodeaddress::ptr> &result,
                               int nproc)
  {
	  if(nproc < 0){
		  nproc = nodes.size() * ppn;
	  }
    if(nodes.empty()) {
      throw valueerror("blockindexing::allocate:  Empty node vector.");
    }
    if(ppn <= 0) {
      throw valueerror("blockindexing::allocate:  ppn must be > 0.");
    }
    if( ((int)(ppn*nodes.size())) <nproc) {
      throw valueerror("blockindexing::allocate:  require ppn*nodes.size()>=nproc");
    }
    result.clear();
    node_set::const_iterator it, end = nodes.end();
    long i = 0;

    char hostname[64];
    environment::ptr env = launcher::get_env(aid);
    std::stringstream sstr;
    for(it = nodes.begin(); it != end && i < ceil((double)nproc / (double)ppn); ++it) {
      nodeaddress::ptr addr = *it;
      SSTMAC_DEBUG << "blockindexing: considering nodeaddress " << addr->to_string() << "\n";
      int64_t this_ppn;
      if (((int)result.size() + ppn) > nproc) this_ppn = nproc - result.size();
      else this_ppn = ppn;

      //just give it a fake hostname for now
      sprintf(hostname, "nid%ld", addr->unique_id());
      for (int tid=result.size(); tid < this_ppn; ++tid){
          env->add_hostname(tid, hostname, addr);
          sstr << hostname << "\n";
      }

      result.insert(result.end(), this_ppn, addr);
      i++;
    }
    env->setenv("SSTMAC_NODELIST", sstr.str());
  }

} } // end of namespace sstmac
