/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/services/launch/indexing/randomindexing.h>
#include <sstmac/common/errors.h>
#include <algorithm>
#include <stdlib.h>

namespace sstmac
{
  namespace sw
  {
    SSTRegister("random", indexstrategy, randomindexing,
                "randomly assigns tasks to nodes");

    randomindexing::randomindexing()
    {
      rng_ = RNG::SimpleCombo::construct();
    }

    // Hi.
    randomindexing::randomindexing(unsigned seed)
    {
      rng_ = RNG::SimpleCombo::construct(seed);
    }

    randomindexing::randomindexing(const RNG::UniformInteger::ptr &rng)
    {
      rng_ = rng;
    }

    randomindexing::ptr
    randomindexing::construct(sim_parameters::ptr params)
    {
      if(params->has_param("random_indexer_seed")){
        long seed = params->get_long_param("random_indexer_seed");
        return randomindexing::ptr(new randomindexing(seed));
      }else
      return randomindexing::ptr(new randomindexing);
    }

    randomindexing::ptr
    randomindexing::construct(unsigned seed)
    {
      return randomindexing::ptr(new randomindexing(seed));
    }

    randomindexing::ptr
    randomindexing::construct(const RNG::UniformInteger::ptr &rng)
    {
      return randomindexing::ptr(new randomindexing(rng));
    }

    randomindexing::~randomindexing() throw ()
    {
    }

    void
    randomindexing::allocate(
        const appid& aid,
        const node_set &nodes,
        int ppn, std::vector<nodeaddress::ptr> &result, int nproc)
    {

      if (nproc < 0)
        {
          nproc = nodes.size() * ppn;
        }

      if (nodes.empty())
        {
          throw valueerror("randomindexing::allocate:  Empty node vector.");
        }
      if (ppn <= 0)
        {
          throw valueerror("randomindexing::allocate:  ppn must be > 0.");
        }
      if (int(ppn * nodes.size()) < nproc)
        {
          throw valueerror(
              "randomindexing::allocate: require ppn*nodes.size()>=nproc");
        }

      node_set::iterator iter = nodes.begin();
      for(long i = 0; i < nproc / ppn; i++){
          iter++;
      }

      // result = nodes;
      for (int i = 0; i < ppn; ++i)
        {
          result.insert(result.end(), nodes.begin(), iter);
        }

      RNG::UniformInteger_functor rngf(rng_);
      std::random_shuffle(result.begin(), result.end(), rngf);
      //std::random_shuffle(result.begin(), result.end());

      if (dbg_.is_active())
        {
          for (uint i = 0; i < result.size(); i++)
            {
              dbg_ << "randomindexing: pos " << i << " is "
                  << result[i]->to_string() << "\n";
            }
        }

      result.resize(nproc);
    }

  }
} // end of namespace sstmac
