/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/software/services/launch/launcher.h>
#include <unistd.h>
#include <getopt.h>

ImplementFactory(sstmac::sw::launcher);

namespace sstmac
{
  namespace sw
  {

    debug launcher::dbg_;

    vis::vis_topology::vismaptype launcher::vismap_;
    vis::vis_display::ptr launcher::vis_;

    launcher::app_to_node_to_task_map_t launcher::app_to_task_map_;
    launcher::app_to_task_to_node_map_t launcher::app_to_node_map_;
    std::map<appid, environment::ptr> launcher::envmap_;

    nodeaddress::ptr
    launcher::sid_to_node(const appid &aid, const taskid &tid)
    {
        app_to_task_to_node_map_t::const_iterator ait = app_to_node_map_.find(aid.id);
        if (ait == app_to_node_map_.end()){
            sst_throw(valueerror, "cannot find app id %d in node lookup", aid.id);
        }

        task_to_node_map_t::const_iterator tit = ait->second.find(tid.id);
        if (tit == ait->second.end()){
            sst_throw(valueerror, "cannot find task id %d in node lookup for app %d", tid.id, aid.id);
        }

        return tit->second;
    }

    void
    launcher::init(const hw::topology::ptr& top)
    {
        /**
          sstkeyword {
            gui=firstavailable;
            docstring=The scheme for allocating a block of nodes from the machine to run the application.
            This does NOT index MPI ranks from 0 to N. It just allocates the set of nodes on which the
            application will run.;
          }
        */
        allocationstrategy::ptr alloc
            = allocationstrategy_factory::get_param("launch_allocation", params_, top);
        /**
          sstkeyword {
            gui=block;
            docstring=The scheme for assigning and indexing MPI ranks from 0 to N on a given node allocation.;
          }
        */
        indexstrategy::ptr ind = indexstrategy_factory::get_param("launch_indexing", params_, top);
        std::list<launchinfo::ptr> linfos; launcher_factory::construct_launchers(params_, linfos);
        launcher::init(linfos, alloc, ind);
    }

    void
    launcher::init(std::list<launchinfo::ptr> &tolaunch, allocationstrategy::ptr &a, indexstrategy::ptr &ind)
    {
      alloc_ = a;
      indexing_ = ind;
      tolaunch_ = tolaunch;
    }

    environment::ptr
    launcher::get_env(const appid &aid)
    {
        std::map<appid, environment::ptr>::const_iterator it = envmap_.find(aid);
        if (it == envmap_.end()){
            sst_throw(valueerror, "no environment for id %d registered with launcher", aid.id);
        }
        return it->second;
    }

    void
    launcher::finish()
    {
        std::list<launchinfo::ptr>::iterator it = tolaunch_.begin(), end = tolaunch_.end();
        for ( ; it != end; ++it){
            (*it)->finish();
        }
        alloc_ = 0;
        indexing_ = 0;
        tolaunch_.clear();
    }

    launcher::~launcher() throw()
    {
    }

    nodeaddress::ptr
    launcher::get_node(const appid& aid, const taskid& tid)
    {
        return sid_to_node(aid, tid);
    }

    void
    launcher_factory::parse_aprun(const std::string& cmd, long& nproc, long& nproc_per_node,
        std::vector<int>& core_affinities, std::string& appname)
    {
      int aprun_numa_containment = 0;
      option aprun_gopt[] =
        {
          { "n", required_argument, NULL, 'n' },
          { "N", required_argument, NULL, 'N' },
          { "ss", no_argument, &aprun_numa_containment, 1 },
          { "cc", required_argument, NULL, 'C' },
          { "S", required_argument, NULL, 'S' },
          { "d", required_argument, NULL, 'd' },
          { NULL, 0, NULL, '\0' } };

      char cmdline_str[200];
      ::strcpy(cmdline_str, cmd.c_str());
      char *argv[50];
      int argc = 0;
      char* pch = strtok(cmdline_str, " ");
      while (pch != 0)
        {
          argv[argc] = pch;
          ++argc;
          pch = strtok(0, " ");
        }

      int _nproc = -1;
      int _nproc_per_node = -1;
      int ch;
      int option_index = 1;
      std::string core_aff_str;
      optind = 1;
      while ((ch = getopt_long_only(argc, argv, "n:N:S:d:", aprun_gopt,
          &option_index)) != -1)
        {
          switch (ch)
            {
          case 0: //this set an input flag
            break;
          case 'n':
            _nproc = atoi(optarg);
            break;
          case 'N':
            _nproc_per_node = atoi(optarg);
            break;
          case 'S':
            break;
          case 'C':
            core_aff_str = optarg;
            break;
          default:
            break;
            }
        }
      if (_nproc <= 0)
        throw inputerror(
            "aprun allocator did not receive a valid -n specification");

      if (_nproc_per_node == 0 || _nproc_per_node > _nproc)
        throw inputerror(
            "aprun allocator did not receive a valid -N specification");

      if (core_aff_str.size() > 0) //we got a core affinity spec
        {
          std::string tosep = core_aff_str;
          std::deque<std::string> tok;
          std::string space = ",";
          pst::BasicStringTokenizer::tokenize(tosep, tok, space);
          std::deque<std::string>::const_iterator it, end = tok.end();
          for (it = tok.begin(); it != end; ++it)
            {
              std::string core = *it;
              core_affinities.push_back(atoi(core.c_str()));
            }
        }

      nproc = _nproc;
      nproc_per_node = _nproc_per_node;
    }

    static debug dbg_;

    void
    launcher_factory::construct_launchers(const sim_parameters::ptr& params, std::list<launchinfo::ptr>& ret)
    {

      int appnum = 1;
      std::stringstream ss;

      if (!dbg_.is_real()){
          dbg_ = debug("launch");
      }

      while (true)
        {
          ss << "launch_app" << appnum;
          if (!params->has_param(ss.str()))
            {
              break;
            }

          std::string launch_prefix = ss.str();
          std::string appname = params->get_param(launch_prefix);
          sstmac::sw::appid aid(appnum);
          timestamp start(0);





          long tasks = 0;
          long ntask_per_node = 0;
          std::vector<int> core_affinities;
          std::string launch_cmd_key = launch_prefix + "_cmd";
          if (params->has_param(launch_cmd_key))
          {
              /** Check for an aprun launch */
              std::string launch_cmd = params->get_param(launch_cmd_key);
              size_t pos = launch_cmd.find_first_of(' ');
              std::string launcher;
              if (pos != std::string::npos)
                  launcher = launch_cmd.substr(0, pos);
              else
                  launcher = launch_cmd;

              if (launcher == "aprun"){
                  parse_aprun(launch_cmd, tasks, ntask_per_node, core_affinities, appname);
                  if (ntask_per_node == -1) //nothing given
                  {
                      int ncores = params->get_optional_int_param("node_cores", 1);
                      ntask_per_node = ncores > tasks ? tasks : ncores;
                  }
              }
              else {
                  sst_throw(valueerror, "invalid launcher %s given", launcher.c_str());
              }
          }
          else //standard launch
            {

              tasks = params->get_long_param(launch_prefix + "_size");
              ntask_per_node = params->get_optional_long_param(
                  launch_prefix + "_ntask_per_node", 1);

              if (params->has_param(launch_prefix + "_core_affinities"))
                {
                  params->get_vector_param(launch_prefix + "_core_affinities",
                      core_affinities);
                }
            }

          start = params->get_optional_time_param(launch_prefix + "_start", timestamp(0));
          long extra = params->get_optional_long_param("launch_extra_nodes", 0);
          launchinfo::ptr li = launchinfo::construct(appname, aid, params, start,
              tasks, ntask_per_node, extra, core_affinities);
          ret.push_back(li);

          appnum++;
          ss.str("");
        }
    }


    }
  }
