/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_BACKENDS_NATIVE_LAUNCH_LAUNCHER_H_INCLUDED
#define SSTMAC_BACKENDS_NATIVE_LAUNCH_LAUNCHER_H_INCLUDED

#include <sstmac/common/timestamp.h>
#include <sstmac/common/ptr_type.h>
#include <sstmac/common/debug.h>
#include <sstmac/common/factories/factory.h>
#include <sstmac/software/services/service.h>
#include <sstmac/common/vis/vis.h>
#include <sstmac/software/services/launch/allocation/allocationstrategy.h>
#include <sstmac/software/services/launch/indexing/indexstrategy.h>
#include <sstmac/software/services/launch/launchinfo.h>

namespace sstmac
{
  namespace sw
  {

    /**
     * A launcher that can be cooperatively scheduled by a very naive scheduler.
     */
    class launcher :
            public service,
            public factory_type,
            public vis::vis_comp
    {

    public:
      virtual std::string
      to_string() const
      {
        return "launcher";
      }

      typedef boost::intrusive_ptr<launcher> ptr;

      /// Hasta la vista.
      virtual ~launcher() throw ();

      virtual void
      apps_failed() = 0;

      virtual void
      app_finished(appid a)
      {
      }

      virtual void finish();

      virtual void
      complete()
      {
        is_completed_ = true;
      }

      /// Test whether this launcher has completed.
      virtual bool
      completed() const
      {
        return is_completed_;
      }

      virtual void
      init(std::list<launchinfo::ptr> &tolaunch, allocationstrategy::ptr &a, indexstrategy::ptr &ind);

      void init(const hw::topology::ptr& top);

      static nodeaddress::ptr
      get_node(const appid& aid, const taskid& tid);

      static environment::ptr
      get_env(const appid& aid);

    protected:
        typedef std::multimap<nodeaddress::ptr, taskid, nodeaddress::nodeaddresscomp> node_to_task_map_t;
        typedef boost::unordered_map<int, nodeaddress::ptr> task_to_node_map_t;
        typedef std::map<int, node_to_task_map_t> app_to_node_to_task_map_t;
        typedef std::map<int, task_to_node_map_t> app_to_task_to_node_map_t;

        static app_to_node_to_task_map_t app_to_task_map_;
        static app_to_task_to_node_map_t app_to_node_map_;
        static std::map<appid, environment::ptr> envmap_;

        typedef std::map<std::string,std::string> envmap;

      static nodeaddress::ptr
      sid_to_node(const appid& aid, const taskid& tid);

      sim_parameters::ptr params_;

      launcher(const sim_parameters::ptr& params) :
          is_completed_(false), params_(params)
      {
        if (!dbg_.is_real())
          {
            dbg_ = debug("<debug> launcher | launch");
            logger::register_logger(&dbg_);
          }

        libname_ = "launcher";
      }

      allocationstrategy::ptr alloc_;
      indexstrategy::ptr indexing_;
      std::list<launchinfo::ptr> tolaunch_;

      static vis::vis_topology::vismaptype vismap_;
      static vis::vis_display::ptr vis_;

      static debug dbg_;

      bool is_completed_;
    };

    class launcher_factory : public SSTFactory1InitParam<launcher, hw::topology::ptr>
    {
        public:
            static void
            construct_launchers(const sstmac::sim_parameters::ptr& params,
                                std::list<boost::intrusive_ptr<sstmac::sw::launchinfo> >& linfos);

            static void
            parse_aprun(const std::string& cmd, long& nproc, long& nproc_per_node,
                std::vector<int>& core_affinities, std::string& appname);
    };

  }
} // end of namespace sstmac

#endif
