/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/software/skeletons/test/mpiprotocol_test.h>
#include <sstmac/software/api/mpi/mpi_implementation/rdmampi.h>
#include <sstmac/software/api/mpi/mpiapi.h>
#include <sstmac/software/libraries/compute/lib_sleep.h>
#include <sstmac/sstmacro.h>

#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>
#include <cstring>
#include <stdio.h>

namespace sstmac
{
  namespace sw
  {
    SSTRegister("mpiprotocol | MPI_protocol | mpi_protocol", app, mpiprotocol_test);

// Pointer types.

//
// Hello.
//
    mpiprotocol_test::ptr
    mpiprotocol_test::construct(softwareid id)
    {
      mpiprotocol_test::ptr obj(new mpiprotocol_test(id));
      //  obj->set_self(obj);
      return obj;
    }

//
// Goodbye.
//
    mpiprotocol_test::~mpiprotocol_test() throw ()
    {
    }

//
// Get a copy.
//
    app::ptr
    mpiprotocol_test::clone(softwareid id)
    {
      return mpiprotocol_test::construct(id);
    }

    void
    mpiprotocol_test::consume_params(const sim_parameters::ptr &params)
    {
    }

//
// Go.
//
    void
    mpiprotocol_test::skeleton_main()
    {
      if (mpi() == 0)
        {
          throw nullerror("mpiprotocol::run:  mpiapi pointer is null.");
        }
      SSTMAC_DEBUG << "mpiprotocol[" << id_.to_string()
            << "]: mpi initializing... \n";

      MPI_Init(NULL, NULL);

      int rank, size;
      MPI_Comm_rank(MPI_COMM_WORLD, &rank);
      MPI_Comm_size(MPI_COMM_WORLD, &size);

      mpi_implementation::ptr mpi_impl = mpi()->queue()->get_implementation();
      rdmampi::ptr rdma_impl = boost::dynamic_pointer_cast<rdmampi>(mpi_impl);
      if (!rdma_impl)
          sst_throw(illformederror, "mpiprotocol test could not find rdma implementation");
      int eager_count = rdma_impl->max_eager_msg_size();
      int rendezvous_count = 2 * rdma_impl->max_eager_msg_size();

      SSTMAC_DEBUG << "mpiprotocol[" << rank
            << "]: mpi initialized\n";

      if (size != 2)
          throw illformederror("mpiprotocol should only be run with 2 processes");

      //make sure the recv is posted first
      if (rank == 0)
      {
          MPI_Barrier(MPI_COMM_WORLD);
          MPI_Send(MPI_PAYLOAD_IGNORE, eager_count, MPI_BYTE, 1, 100, MPI_COMM_WORLD);
      }
      else
      {
          MPI_Request req;
          MPI_Irecv(MPI_PAYLOAD_IGNORE, eager_count, MPI_BYTE, 0, 100, MPI_COMM_WORLD, &req);
          MPI_Barrier(MPI_COMM_WORLD);
          MPI_Wait(&req, MPI_STATUS_IGNORE);
      }

      SSTMAC_DEBUG << "mpiprotocol[" << rank
            << "]: finished first test\n";

      if (rank == 0){
          //post the send first
        MPI_Send(MPI_PAYLOAD_IGNORE, eager_count, MPI_BYTE, 1, 100, MPI_COMM_WORLD);
      }
      else
      {
          app::sleep(timestamp(0.005));
          MPI_Recv(MPI_PAYLOAD_IGNORE, eager_count, MPI_BYTE, 0, 100, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
      }

      SSTMAC_DEBUG << "mpiprotocol[" << rank
            << "]: finished second test\n";

      if (rank == 0){
          //post the send first
        MPI_Send(MPI_PAYLOAD_IGNORE, rendezvous_count, MPI_BYTE, 1, 100, MPI_COMM_WORLD);
      }
      else
      {
          MPI_Recv(MPI_PAYLOAD_IGNORE, rendezvous_count, MPI_BYTE, 0, 100, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
      }

      SSTMAC_DEBUG << "mpiprotocol[" << rank
            << "]: finished third test\n";

      //now send a really big message
      if (rank == 0){

      }

      MPI_Finalize();

      SSTMAC_DEBUG << "mpiprotocol[" << rank
            << "]: finalized\n";

    }

  }
} // end of namespace sstmac.

