#include <tests/unittests/util/util.h>
#include <sstmac/hardware/network/topology/routing/fattreerouter.h>

using namespace sstmac;
using namespace sstmac::hw;

void
test_fattree4(UnitTest& unit)
{
    topology::clear_rng();

    unittest_params params;
    sstmac_env::params = params;
    params["topology_geometry"] = "3 4";
    topology::ptr top = topology_factory::get_value("fattree", params);
    regular_topology::ptr ftree = boost::dynamic_pointer_cast<regular_topology>(top);
    assertTrue(unit, "fat tree cast topology", ftree);

    /**
        Bottom Up
        Level 0: 0-15
        Level 1: 16-31
        Level 2: 32-39
        OR
        Top Down
        Level 0: 32-39
        Level 1: 16-31
        Level 2: 0-15
    */
    topology::maptype switches;
    init_switches(switches, params, top);

    {
        coordinates coords = get_vector(0,1);
        long nid = ftree->get_switch_id(coords);
        networkswitch::ptr sw = switches[naddr(nid)];
        router::ptr router = sw->get_router();

        long dst = ftree->get_node_id(get_vector(2,1,0,0,0));

        router::outports ports;
        router->get_productive_ports_to_node(naddr(dst), ports);
        assertEqual(unit, "num productive ports", ports.size(), 2);
        assertEqual(unit, "productive port", ports[0], router->convert_to_outport(fattree::down_dimension, 1));
        assertEqual(unit, "productive port", ports[1], router->convert_to_outport(fattree::down_dimension, 5));
    }

    {
        coordinates coords = get_vector(2,0,2,1);
        long nid = ftree->get_switch_id(coords);
        networkswitch::ptr sw = switches[naddr(nid)];
        router::ptr router = sw->get_router();

        long dst = ftree->get_node_id(get_vector(2,1,1,0,1));

        router::outports ports;
        router->get_productive_ports_to_node(naddr(dst), ports);
        assertEqual(unit, "num productive ports", ports.size(), 4);
        for (int k=0; k < ports.size(); ++k)
            assertEqual(unit, "productive port", ports[k], router->convert_to_outport(fattree::up_dimension, k));
    }

    {
        coordinates coords = get_vector(1,1,2);
        long nid = ftree->get_switch_id(coords);
        networkswitch::ptr sw = switches[naddr(nid)];
        router::ptr router = sw->get_router();

        long dst = ftree->get_node_id(get_vector(2,1,2,0,0));

        router::outports ports;
        router->get_productive_ports_to_node(naddr(dst), ports);
        assertEqual(unit, "num productive ports", ports.size(), 1);
        assertEqual(unit, "productive port", ports[0], router->convert_to_outport(fattree::down_dimension, 2));
    }

    {
        coordinates c1 = get_vector(2,0,1,0);
        coordinates c2 = get_vector(2,1,0,1);
        int dist = ftree->minimal_distance(c1,c2);
        //I need to go up two steps
        int correct = 2 * 2;
        assertEqual(unit, "min distance", dist, correct);
    }

    {
        coordinates c1 = get_vector(2,0,1,0);
        coordinates c2 = get_vector(1,0,1);
        int dist = ftree->minimal_distance(c1,c2);
        //I need to go up one step
        int correct = 1;
        assertEqual(unit, "min distance", dist, correct);
    }

}


void
test_fattree2(UnitTest& unit)
{
    topology::clear_rng();

    unittest_params params;
    sstmac_env::params = params;
    params["topology_geometry"] = "5 2";
    params["network_nodes_per_switch"] = "2";
    topology::ptr top = topology_factory::get_value("fattree", params);
    regular_topology::ptr ftree = boost::dynamic_pointer_cast<regular_topology>(top);
    assertTrue(unit, "fat tree cast topology", ftree);

    topology::maptype switches;
    init_switches(switches, params, top);

    /**
        Bottom Up
        Level 0: 0-15
        Level 1: 16-31
        Level 2: 32-47
        Level 3: 48-63
        Level 4: 64-71
        OR
        Top Down
        Level 0: 64-71
        Level 1: 48-63
        Level 2: 32-47
        Level 3: 16-31
        Level 4: 0-15
    */

    {
        coordinates coords = get_vector(0,1);
        long nid = ftree->get_switch_id(coords);
        networkswitch::ptr sw = switches[naddr(nid)];
        router::ptr router = sw->get_router();

        long dst = ftree->get_node_id(get_vector(4,1,0,1,0,0));

        router::outports ports;
        router->get_productive_ports_to_node(naddr(dst), ports);
        assertEqual(unit, "num productive ports", ports.size(), 2);
        assertEqual(unit, "productive port", ports[0], router->convert_to_outport(fattree::down_dimension, 0));
        assertEqual(unit, "productive port", ports[1], router->convert_to_outport(fattree::down_dimension, 2));
    }

    {
        coordinates coords = get_vector(4,0,1,0,1,0);
        long nid = ftree->get_switch_id(coords);
        networkswitch::ptr sw = switches[naddr(nid)];
        router::ptr router = sw->get_router();

        long dst = ftree->get_node_id(get_vector(4,1,0,1,0,0,0));

        router::outports ports;
        router->get_productive_ports_to_node(naddr(dst), ports);
        assertEqual(unit, "num productive ports", ports.size(), 2);
        for (int k=0; k < ports.size(); ++k)
            assertEqual(unit, "productive port", ports[k], router->convert_to_outport(fattree::up_dimension, k));
    }

    {
        coordinates c1 = get_vector(4,0,1,0,1,0);
        coordinates c2 = get_vector(4,0,1,1,1,0);
        int dist = ftree->minimal_distance(c1,c2);
        //I need to go up two steps
        int correct = 2 * 2;
        assertEqual(unit, "min distance", dist, correct);
    }

    {
        coordinates c1 = get_vector(2,0,1,1);
        coordinates c2 = get_vector(4,0,1,1,1,0);
        int dist = ftree->minimal_distance(c1,c2);
        //I only need to go down two
        int correct = 2;
        assertEqual(unit, "min distance", dist, correct);
    }

    {
        coordinates c1 = get_vector(4,0,1,1,1,0);
        coordinates c2 = get_vector(2,0,1,1);
        int dist = ftree->minimal_distance(c1,c2);
        //I only need to go up two
        int correct = 2;
        assertEqual(unit, "min distance", dist, correct);
    }

    {
        coordinates c1 = get_vector(2,0,0,1);
        coordinates c2 = get_vector(4,0,1,1,1,0);
        int dist = ftree->minimal_distance(c1,c2);
        //I only need to go up one then down three
        int correct = 1 + 1 + 2;
        assertEqual(unit, "min distance", dist, correct);
    }

    {
        coordinates c1 = get_vector(4,0,1,1,1,0);
        coordinates c2 = get_vector(2,0,0,1);
        int dist = ftree->minimal_distance(c1,c2);
        //I only need to go up three then down one
        int correct = 2 + 1 + 1;
        assertEqual(unit, "min distance", dist, correct);
    }

    {
        coordinates c1 = get_vector(4,0,1,1,1,0);
        coordinates c2 = get_vector(4,1,1,0,0,1);
        int dist = ftree->minimal_distance(c1,c2);
        //I only need to go all the way up and back down
        int correct = 8;
        assertEqual(unit, "min distance", dist, correct);
    }

    {
        coordinates c1 = get_vector(3,0,1,0,0);
        coordinates c2 = get_vector(3,0,1,0,1);
        int dist = ftree->minimal_distance(c1,c2);
        //I'm on the right branch, but wrong switch
        int correct = 2;
        assertEqual(unit, "min distance", dist, correct);
    }

}
