#include "test_topology.h"
#include <sstmac/hardware/network/topology/dragonfly.h>


using namespace sstmac;
using namespace sstmac::hw;

void
test_dragonfly_traffic(UnitTest& unit)
{
    topology::clear_rng();

    unittest_params params;
    sstmac_env::params = params;
    params["topology_geometry"] = "4 6 8";
    params["network_nodes_per_switch"] = "1";
    params["optical_link_weight"] = "2.0";
    params["topology_group_connections"] = "4";
    params["topology_seed"] = "14";
    topology::ptr top = topology_factory::get_value("dragonfly", params);
    regular_topology::ptr dfly = boost::dynamic_pointer_cast<regular_topology>(top);
    assertTrue(unit, "dragonfly cast topology", dfly);

    topology::maptype switches;
    init_switches(switches, params, top);

    long dst;
    std::vector<nodeaddress::ptr> partners;

    coordinates src_coords = get_vector(1, 2, 3);
    long nid = dfly->get_node_id(src_coords);
    coordinates test_coords = dfly->get_node_coords(nid);
    assertEqual(unit, "coord transform", test_coords, 1, 2, 3);
    nodeaddress::ptr src = nodeaddress::convert_from_uniqueid(nid);

    partners.clear();
    dfly->get_send_partners(traffic_pattern::tornado, src, partners);
    assertEqual(unit, "num tornado partners", partners.size(), 1);
    dst = dfly->get_node_id(get_vector(2,4,6));
    assertEqual(unit, "tornado partner", partners[0]->unique_id(), dst);

    partners.clear();
    dfly->get_send_partners(traffic_pattern::bit_complement, src, partners);
    assertEqual(unit, "num BC partners", partners.size(), 1);
    dst = dfly->get_node_id(get_vector(2,3,4));
    assertEqual(unit, "BC partner", partners[0]->unique_id(), dst);

    partners.clear();
    dfly->get_send_partners(traffic_pattern::nearest_neighbor, src, partners);
    //would be 8, but one of the group connections is a "self" connection
    assertEqual(unit, "num NN partners", partners.size(), 7);

    dst = dfly->get_node_id(get_vector(2,2,3));
    if (partners.size() >= 1)
        assertEqual(unit, "NN partner", partners[0]->unique_id(), dst);

    dst = dfly->get_node_id(get_vector(0,2,3));
    if (partners.size() >= 2)
        assertEqual(unit, "NN partner", partners[1]->unique_id(), dst);

    dst = dfly->get_node_id(get_vector(1,3,3));
    if (partners.size() >= 3)
        assertEqual(unit, "NN partner", partners[2]->unique_id(), dst);

    dst = dfly->get_node_id(get_vector(1,1,3));
    if (partners.size() >= 4)
        assertEqual(unit, "NN partner", partners[3]->unique_id(), dst);

    dst = dfly->get_node_id(get_vector(1,2,1));
    if (partners.size() >= 5)
        assertEqual(unit, "NN partner", partners[4]->unique_id(), dst);

    dst = dfly->get_node_id(get_vector(1,2,5));
    if (partners.size() >= 6)
        assertEqual(unit, "NN partner", partners[5]->unique_id(), dst);

    dst = dfly->get_node_id(get_vector(1,2,7));
    if (partners.size() >= 7)
        assertEqual(unit, "NN partner", partners[6]->unique_id(), dst);
}

void
test_dragonfly_v1(UnitTest& unit)
{
    topology::clear_rng();

    unittest_params params;
    sstmac_env::params = params;
    params["topology_geometry"] = "4 4 4";
    params["network_nodes_per_switch"] = "2";
    params["optical_link_weight"] = "2.0";
    params["topology_group_connections"] = "4";
    params["topology_seed"] = "14";
    topology::ptr top = topology_factory::get_value("dragonfly", params);
    regular_topology::ptr dfly = boost::dynamic_pointer_cast<regular_topology>(top);
    assertTrue(unit, "dragonfly cast topology", dfly);

    topology::maptype switches;
    init_switches(switches, params, top);

    {
        long nids[] = {0, 17, 53, 60};
        for (int i=0; i < 4; ++i){
            long nid = nids[i];
            //make sure the functions work back and forth
            coordinates coords = dfly->get_switch_coords(nid);
            long test_id = dfly->get_switch_id(coords);
            coordinates test_coords = dfly->get_switch_coords(test_id);
            assertEqual(unit, "dfly switch id", test_id, nid);
            assertEqual(unit, "dfly coords", test_coords, coords);
        }
    }

    {
    networkswitch::ptr sw = switches[naddr(7)];
    int dim,dir; long port;
    router::ptr router = sw->get_router();
    port = router->route(msg(4)); router->convert_to_dimdir(port, dim, dir);
    assertEqual(unit, "route dragonfly dim", dim, dragonfly::x_dimension);
    assertEqual(unit, "route dragonfly dir", dir, 2);
    assert_dim_dir(unit, "dragonfly dim/dir",
                   sw, dim, dir, 6);
    }

    {
    networkswitch::ptr sw = switches[naddr(6)];
    int dim,dir; long port;
    router::ptr router = sw->get_router();
    port = router->route(msg(4)); router->convert_to_dimdir(port, dim, dir);
    assertEqual(unit, "route dragonfly dim", dim, dragonfly::y_dimension);
    assertEqual(unit, "route dragonfly dir", dir, 0);
    assert_dim_dir(unit, "dragonfly dim/dir",
                   sw, dim, dir, 2);
    }

    {
    networkswitch::ptr sw = switches[naddr(29)];
    int dim,dir; long port;
    router::ptr router = sw->get_router();
    port = router->route(msg(4)); router->convert_to_dimdir(port, dim, dir);
    assertEqual(unit, "route dragonfly dim", dim, dragonfly::g_dimension);
    assertEqual(unit, "route dragonfly dir", dir, 0);
    assert_dim_dir(unit, "dragonfly dim/dir",
                  sw, dim, dir, 13);
    }

}


void
test_dragonfly_v2(UnitTest& unit)
{
    topology::clear_rng();

    unittest_params params;
    sstmac_env::params = params;
    params["topology_geometry"] = "4 4 8";
    params["network_nodes_per_switch"] = "2";
    params["optical_link_weight"] = "2.0";
    params["topology_group_connections"] = "3";
    params["topology_seed"] = "14";
    topology::ptr top = topology_factory::get_value("dragonfly", params);
    regular_topology::ptr dfly = boost::dynamic_pointer_cast<regular_topology>(top);
    assertTrue(unit, "dragonfly cast topology", dfly);

    topology::maptype switches;
    init_switches(switches, params, top);

    {
        long nids[] = {0, 17, 71, 108};
        for (int i=0; i < 4; ++i){
            long nid = nids[i];
            //make sure the functions work back and forth
            coordinates coords = dfly->get_switch_coords(nid);
            long test_id = dfly->get_switch_id(coords);
            coordinates test_coords = dfly->get_switch_coords(test_id);
            assertEqual(unit, "dfly switch id", test_id, nid);
            assertEqual(unit, "dfly coords", test_coords, coords);
        }
    }


    {
    networkswitch::ptr sw = switches[naddr(7)];
    long port; int dim,dir;
    dfly->minimal_route_to_node(naddr(7), naddr(4), dim, dir);
    assertEqual(unit, "route dragonfly dim", dim, dragonfly::x_dimension);
    assertEqual(unit, "route dragonfly dir", dir, 2);
    assert_dim_dir(unit, "dragonfly dim/dir",
                   sw, dim, dir, 6);
    }


    {
    networkswitch::ptr sw = switches[naddr(6)];
    int dim,dir;
    dfly->minimal_route_to_node(naddr(6), naddr(4), dim, dir);
    assertEqual(unit, "route dragonfly dim", dim, dragonfly::y_dimension);
    assertEqual(unit, "route dragonfly dir", dir, 0);
    assert_dim_dir(unit, "dragonfly dim/dir",
                   sw, dim, dir, 2);
    }

    {
    networkswitch::ptr sw = switches[naddr(29)];
    int dim,dir;
    dfly->minimal_route_to_node(naddr(29), naddr(4), dim, dir);
    assertEqual(unit, "route dragonfly dim", dim, dragonfly::x_dimension);
    assertEqual(unit, "route dragonfly dir", dir, 2);
    assert_dim_dir(unit, "dragonfly dim/dir",
                  sw, dim, dir, 30);
    }

    {
    networkswitch::ptr sw = switches[naddr(30)];
    int dim,dir;
    dfly->minimal_route_to_node(naddr(30), naddr(4), dim, dir);
    assertEqual(unit, "route dragonfly dim", dim, dragonfly::g_dimension);
    assertEqual(unit, "route dragonfly dir", dir, 0);
    assert_dim_dir(unit, "dragonfly dim/dir",
                  sw, dim, dir, 14);
    }

}
