#include "test_topology.h"
#include <sstmac/hardware/network/topology/fattree.h>
#include <sstmac/hardware/network/topology/routing/fattreerouter.h>

using namespace sstmac;
using namespace sstmac::hw;

void
test_fattree4(UnitTest& unit)
{
    topology::clear_rng();

    unittest_params params;
    sstmac_env::params = params;
    params["topology_geometry"] = "3 4";
    topology::ptr top = topology_factory::get_value("fattree", params);
    regular_topology::ptr ftree = boost::dynamic_pointer_cast<regular_topology>(top);
    assertTrue(unit, "fat tree cast topology", ftree);

    /**
        Bottom Up
        Level 0: 0-15
        Level 1: 16-31
        Level 2: 32-39
        OR
        Top Down
        Level 0: 32-39
        Level 1: 16-31
        Level 2: 0-15
    */

    {
        long nids[] = {0, 13, 19, 37};
        for (int i=0; i < 4; ++i){
            long nid = nids[i];
            //make sure the functions work back and forth
            coordinates coords = ftree->get_switch_coords(nid);
            long test_id = ftree->get_switch_id(coords);
            coordinates test_coords = ftree->get_switch_coords(test_id);
            assertEqual(unit, "ftree switch id", test_id, nid);
            assertEqual(unit, "ftree coords", test_coords, coords);
        }
    }

    //lets test some coordinates for switches
    {
        coordinates coords = ftree->get_switch_coords(36);
        assertEqual(unit, "fat tree coord check", coords, 0, 4);
    }

    {
        coordinates coords = ftree->get_switch_coords(21);
        assertEqual(unit, "fat tree coord check", coords, 1, 1, 1);
    }

    {
        coordinates coords = ftree->get_switch_coords(13);
        assertEqual(unit, "fat tree coord check", coords, 2, 3, 1, 0);
    }
}


void
test_fattree2(UnitTest& unit)
{
    topology::clear_rng();

    unittest_params params;
    sstmac_env::params = params;
    params["topology_geometry"] = "5 2";
    params["network_nodes_per_switch"] = "2";
    topology::ptr top = topology_factory::get_value("fattree", params);
    regular_topology::ptr ftree = boost::dynamic_pointer_cast<regular_topology>(top);
    assertTrue(unit, "fat tree cast topology", ftree);

    topology::maptype switches;
    init_switches(switches, params, top);

    {
        long nids[] = {0, 13, 19, 37, 51, 66};
        for (int i=0; i < 6; ++i){
            long nid = nids[i];
            //make sure the functions work back and forth
            coordinates coords = ftree->get_switch_coords(nid);
            long test_id = ftree->get_switch_id(coords);
            coordinates test_coords = ftree->get_switch_coords(test_id);
            assertEqual(unit, "ftree switch id", test_id, nid);
            assertEqual(unit, "ftree coords", test_coords, coords);
        }
    }

    /**
        Bottom Up
        Level 0: 0-15
        Level 1: 16-31
        Level 2: 32-47
        Level 3: 48-63
        Level 4: 64-71
        OR
        Top Down
        Level 0: 64-71
        Level 1: 48-63
        Level 2: 32-47
        Level 3: 16-31
        Level 4: 0-15
    */

    //lets test some coordinates for switches
    {
        coordinates coords = ftree->get_switch_coords(69);
        assertEqual(unit, "fat tree coord check", coords, 0, 5);
    }

    //lets test some coordinates for switches
    {
        coordinates coords = ftree->get_switch_coords(58);
        assertEqual(unit, "fat tree coord check", coords, 1, 1, 2);
    }

    //lets test some coordinates for switches
    {
        coordinates coords = ftree->get_switch_coords(37);
        assertEqual(unit, "fat tree coord check", coords, 2, 0, 1, 1);
    }

    //lets test some coordinates for switches
    {
        coordinates coords = ftree->get_switch_coords(30);
        assertEqual(unit, "fat tree coord check", coords, 3, 1, 1, 1, 0);
    }

    //lets test some coordinates for switches
    {
        coordinates coords = ftree->get_switch_coords(13);
        assertEqual(unit, "fat tree coord check", coords, 4, 1, 1, 0, 1, 0);
    }

    {
    networkswitch::ptr sw = switches[nodeaddress::convert_from_uniqueid(7)];
    router::ptr router = sw->get_router();

    int dim,dir; long port;
    port = router->route(msg(22)); router->convert_to_dimdir(port, dim, dir);
    assertEqual(unit, "route fat tree", dim, fattree::up_dimension);
    assert_dim_dir(unit, "fat tree dim/dir",
                   sw, dim, dir, 22);

    port = router->route(msg(23)); router->convert_to_dimdir(port, dim, dir);
    // this should rotate the paths taken
    assertEqual(unit, "route fat tree", dim, fattree::up_dimension);
    assert_dim_dir(unit, "fat tree dim/dir",
                   sw, dim, dir, 23);

    port = router->route(msg(15)); router->convert_to_dimdir(port, dim, dir);
    assertEqual(unit, "route fat tree", dim, (int) topology::EJECT);
    }


    {
    networkswitch::ptr sw = switches[nodeaddress::convert_from_uniqueid(26)];
    router::ptr router = sw->get_router();

    int dim,dir; long port;
    port = router->route(msg(4)); router->convert_to_dimdir(port, dim, dir);
    assertEqual(unit, "route fat tree", dim, fattree::up_dimension);
    assert_dim_dir(unit, "fat tree dim/dir",
                    sw, dim, dir, 40);

    // this should rotate the paths taken
    port = router->route(msg(4)); router->convert_to_dimdir(port, dim, dir);
    assert_dim_dir(unit, "fat tree dim/dir",
                    sw, dim, dir, 42);

    port = router->route(msg(15)); router->convert_to_dimdir(port, dim, dir);
    assertEqual(unit, "route fat tree", dim, fattree::up_dimension);

    port = router->route(msg(23)); router->convert_to_dimdir(port, dim, dir);
    assertEqual(unit, "route fat tree", dim, fattree::down_dimension);
    assert_dim_dir(unit, "fat tree dim/dir",
                    sw, dim, dir, 11);
    }


    {
    networkswitch::ptr sw = switches[nodeaddress::convert_from_uniqueid(65)];
    router::ptr router = sw->get_router();
    int dim,dir; long port;
    port = router->route(msg(4)); router->convert_to_dimdir(port, dim, dir);
    assertEqual(unit, "route fat tree from top", dim, fattree::down_dimension);
    assert_dim_dir(unit, "fat tree dim/dir at top",
                    sw, dim, dir, 50);
    port = router->route(msg(15)); router->convert_to_dimdir(port, dim, dir);
    assertEqual(unit, "route fat tree", dim, fattree::down_dimension);
    assert_dim_dir(unit, "fat tree dim/dir at top",
                    sw, dim, dir, 51);
    port = router->route(msg(23)); router->convert_to_dimdir(port, dim, dir);
    assertEqual(unit, "route fat tree", dim, fattree::down_dimension);
    assert_dim_dir(unit, "fat tree dim/dir at top",
                    sw, dim, dir, 58);
    }


}
