#include "test_topology.h"
#include <sstmac/hardware/network/topology/hdtorus.h>
#include <sstmac/hardware/network/topology/routing/valiant_routing.h>

using namespace sstmac;
using namespace sstmac::hw;

void test_torus_traffic(UnitTest& unit)
{
    unittest_params params;
    params["topology_geometry"] = "5 5 5";
    params["topology_redundant"] = "1 1 1";
    params["network_nodes_per_switch"] = "2";
    topology::ptr top = topology_factory::get_value("hdtorus", params);
    regular_topology::ptr torus = boost::dynamic_pointer_cast<regular_topology>(top);

    topology::maptype switches;
    init_switches(switches, params, top);

    assertTrue(unit, "torus cast topology", torus);

    long dst;
    std::vector<nodeaddress::ptr> partners;

    coordinates src_coords = get_vector(1, 3, 4, 1);
    long nid = torus->get_node_id(src_coords);
    coordinates test_coords = torus->get_node_coords(nid);
    assertEqual(unit, "node id", nid, 233);
    assertEqual(unit, "coord transform", test_coords, 1, 3, 4, 1);
    nodeaddress::ptr src = nodeaddress::convert_from_uniqueid(nid);

    partners.clear();
    torus->get_send_partners(traffic_pattern::tornado, src, partners);
    assertEqual(unit, "num tornado partners", partners.size(), 1);
    dst = torus->get_node_id(get_vector(3,0,1,1));
    assertEqual(unit, "tornado partner", partners[0]->unique_id(), dst);

    partners.clear();
    torus->get_send_partners(traffic_pattern::nearest_neighbor, src, partners);
    assertEqual(unit, "num NN partners", partners.size(), 6);
    dst = torus->get_node_id(get_vector(2,3,4,1));
    assertEqual(unit, "NN partner", partners[0]->unique_id(), dst);
    dst = torus->get_node_id(get_vector(0,3,4,1));
    assertEqual(unit, "NN partner", partners[1]->unique_id(), dst);
    dst = torus->get_node_id(get_vector(1,4,4,1));
    assertEqual(unit, "NN partner", partners[2]->unique_id(), dst);
    dst = torus->get_node_id(get_vector(1,2,4,1));
    assertEqual(unit, "NN partner", partners[3]->unique_id(), dst);
    dst = torus->get_node_id(get_vector(1,3,0,1));
    assertEqual(unit, "NN partner", partners[4]->unique_id(), dst);
    dst = torus->get_node_id(get_vector(1,3,3,1));
    assertEqual(unit, "NN partner", partners[5]->unique_id(), dst);

    partners.clear();
    torus->get_send_partners(traffic_pattern::bit_complement, src, partners);
    assertEqual(unit, "num tornado partners", partners.size(), 1);
    dst = torus->get_node_id(get_vector(3,1,0,1));
    assertEqual(unit, "tornado partner", partners[0]->unique_id(), dst);
}

void test_torus(UnitTest& unit)
{
    unittest_params params;
    params["topology_geometry"] = "5 5 5";
    params["topology_redundant"] = "1 1 1";
    params["network_nodes_per_switch"] = "2";
    topology::ptr top = topology_factory::get_value("hdtorus", params);
    regular_topology::ptr torus = boost::dynamic_pointer_cast<regular_topology>(top);

    topology::maptype switches;
    init_switches(switches, params, top);

    assertTrue(unit, "torus cast topology", torus);

    {
        long nids[] = {0, 13, 48, 97};
        for (int i=0; i < 4; ++i){
            long nid = nids[i];
            //make sure the functions work back and forth
            coordinates coords = torus->get_switch_coords(nid);
            long test_id = torus->get_switch_id(coords);
            coordinates test_coords = torus->get_switch_coords(test_id);
            assertEqual(unit, "torus switch id", test_id, nid);
            assertEqual(unit, "torus coords", test_coords, coords);
        }
    }

    {
    long nid = 25;
    coordinates coords = torus->get_node_coords(nid);
    long nid_test = torus->get_node_id(coords);
    assertEqual(unit, "torus node coords", coords, 2, 2, 0, 1);
    assertEqual(unit, "torus node id mapping", nid_test, nid);
    }

    {
    long sid = 37;
    coordinates coords = torus->get_switch_coords(sid);
    long sid_test = torus->get_switch_id(coords);
    assertEqual(unit, "torus switch coords", coords, 2, 2, 1);
    assertEqual(unit, "torus switch id mapping", sid_test, sid);
    }

    {
    coordinates coords = get_vector(1,0,2);
    long sid = torus->get_switch_id(coords);
    coordinates test_coords = torus->get_switch_coords(sid);
    assertEqual(unit, "torus switch coords mapping", test_coords, coords);
    assertEqual(unit, "torus switch id", sid, 51);
    }

    {
    coordinates coords = get_vector(1,0,2,1);
    long nid = torus->get_node_id(coords);
    coordinates test_coords = torus->get_node_coords(nid);
    assertEqual(unit, "torus node coords mapping", test_coords, coords);
    assertEqual(unit, "torus node id", nid, 103);
    }

    nodeaddress::ptr swid = nodeaddress::convert_from_uniqueid(21);
    router::ptr router = router_factory::get_value("regular", params, swid, torus);
    router->set_switch(switches[swid]);

    {
    int dim = 1;
    int dir = topology::POS;
    int outport = router->convert_to_outport(dim, dir);
    int testdim, testdir;
    router->convert_to_dimdir(outport, testdim, testdir);
    assertEqual(unit, "torus outport to dim", testdim, dim);
    assertEqual(unit, "torus outport to dir", testdir, dir);
    }

    {
        int dim, dir;
        torus->minimal_route_to_coords(
            get_vector(0,0,0),
            get_vector(0,0,4),
            dim, dir);
        assertEqual(unit, "torus dir", dir, (int) topology::NEG);

        torus->minimal_route_to_coords(
            get_vector(0,0,4),
            get_vector(0,0,0),
            dim, dir);
        assertEqual(unit, "torus dir", dir, (int) topology::POS);

        torus->minimal_route_to_coords(
            get_vector(0,0,1),
            get_vector(0,0,4),
            dim, dir);
        assertEqual(unit, "torus dir", dir, (int) topology::NEG);

        torus->minimal_route_to_coords(
            get_vector(0,0,2),
            get_vector(0,0,4),
            dim, dir);
        assertEqual(unit, "torus dir", dir, (int) topology::POS);

        torus->minimal_route_to_coords(
            get_vector(0,0,3),
            get_vector(0,0,4),
            dim, dir);
        assertEqual(unit, "torus dir", dir, (int) topology::POS);
    }

    {
        int dim, dir;
        torus->minimal_route_to_coords(
            get_vector(0,0,0),
            get_vector(0,4,0),
            dim, dir);
        assertEqual(unit, "torus dir", dir, (int) topology::NEG);

        torus->minimal_route_to_coords(
            get_vector(0,4,0),
            get_vector(0,0,0),
            dim, dir);
        assertEqual(unit, "torus dir", dir, (int) topology::POS);

        torus->minimal_route_to_coords(
            get_vector(0,1,0),
            get_vector(0,4,0),
            dim, dir);
        assertEqual(unit, "torus dir", dir, (int) topology::NEG);

        torus->minimal_route_to_coords(
            get_vector(0,2,0),
            get_vector(0,4,0),
            dim, dir);
        assertEqual(unit, "torus dir", dir, (int) topology::POS);

        torus->minimal_route_to_coords(
            get_vector(0,3,0),
            get_vector(0,4,0),
            dim, dir);
        assertEqual(unit, "torus dir", dir, (int) topology::POS);
    }

    {
    int dim = 0;
    int dir = topology::NEG;
    int outport = router->convert_to_outport(dim, dir);
    int testdim, testdir;
    router->convert_to_dimdir(outport, testdim, testdir);
    assertEqual(unit, "torus outport to dim", testdim, dim);
    assertEqual(unit, "torus outport to dir", testdir, dir);
    }
}

#define PRINT_VALIANT 0

void
test_torus_valiant(UnitTest& unit)
{
    unittest_params params;
    params["topology_geometry"] = "5 5 5";
    params["topology_redundant"] = "1 1 1";
    params["network_nodes_per_switch"] = "2";
    params["topology_seed"] = "479";
    params["default_routing"] = "valiant";
    topology::ptr top = topology_factory::get_value("hdtorus", params);
    regular_topology::ptr regtop = boost::dynamic_pointer_cast<regular_topology>(top);
    hdtorus::ptr torus = boost::dynamic_pointer_cast<hdtorus>(top);

    topology::maptype switches;
    init_switches(switches, params, top);

    assertTrue(unit, "torus cast topology", torus);
    if (!torus)
        return;


    {
        // switch <2,0,0>
        interconnect_message::ptr my_msg = msg(4);
        // switch <2,4,3>
        networkswitch::ptr src_sw = switches[naddr(97)];
        coordinates src_coords = regtop->get_switch_coords(naddr(97));
        assertEqual(unit, "src sw coords", src_coords, 2, 4, 3);
        int dim, dir; long port;
        port = src_sw->get_router()->route(my_msg); src_sw->get_router()->convert_to_dimdir(port, dim, dir);

        valiant_routing_info::ptr rinfo
            = safe_cast(my_msg->rinfo(), valiant_routing_info);
        //overwrite whatever happened before
        rinfo->set_dest_switch(naddr(102));
        port = src_sw->get_router()->route(my_msg); src_sw->get_router()->convert_to_dimdir(port, dim, dir);

        networkswitch::ptr next_sw = get_switch_at_dim_dir(src_sw, dim, dir);

        assertEqual(unit, "valiant middle switch",
            rinfo->get_dest_switch()->unique_id(), 102);
        assertEqual(unit, "valiant first switch",
            next_sw->get_addr()->unique_id(), 77);

        assertEqual(unit, "torus outport to dim", dim, 1);
        assertEqual(unit, "torus outport to dir", dir, (int) topology::POS);

        //figure out the number of hops to the middle switch
        coordinates mid_coords = regtop->get_switch_coords(rinfo->get_dest_switch());
        coordinates next_coords = regtop->get_switch_coords(next_sw->get_addr());
        int dist = 0;
        for (int i=0; i < next_coords.size(); ++i){
            #if PRINT_VALIANT
            std::cout << src_coords[i]
                      << "->"
                      << next_coords[i]
                      << "->"
                      << mid_coords[i]
                      << std::endl;
            #endif
            int dir = torus->get_shortest_direction(i, src_coords[i], mid_coords[i]);
            dist += torus->get_distance(i, dir, src_coords[i], mid_coords[i]);
        }

        for (int s=1; s < dist; ++s){

            src_sw = next_sw;
            src_coords = next_coords;
            port = src_sw->get_router()->route(my_msg); src_sw->get_router()->convert_to_dimdir(port, dim, dir);

            // should be a minimal route
            int mindim, mindir;
            torus->minimal_route_to_coords(
                src_coords,
                mid_coords,
                mindim, mindir
            );

            next_sw = get_switch_at_dim_dir(src_sw, dim, dir);
            assertEqual(unit, "next valiant step dim", dim, mindim);
            assertEqual(unit, "next valiant step dir", dir, (int) mindir);

            coordinates mid_coords = regtop->get_switch_coords(rinfo->get_dest_switch());
            coordinates next_coords = regtop->get_switch_coords(next_sw->get_addr());

            for (int i=0; i < next_coords.size(); ++i){
                #if PRINT_VALIANT
                std::cout << src_coords[i]
                          << "->"
                          << next_coords[i]
                          << "->"
                          << mid_coords[i]
                          << std::endl;
                #endif
            }
        }

        //now we should be routing to the final node
        nodeaddress::ptr ej_addr = top->node_to_ejector_addr(my_msg->toaddr());
        coordinates dst_coords = regtop->get_switch_coords(ej_addr);
        assertEqual(unit, "final dst coords", dst_coords, 2, 0, 0);

        src_sw = next_sw;
        src_coords = next_coords;
        port = src_sw->get_router()->route(my_msg); src_sw->get_router()->convert_to_dimdir(port, dim, dir);
        next_sw = get_switch_at_dim_dir(src_sw, dim, dir);

        //this should match minimal routing to final destination
        int mindim, mindir;
        torus->minimal_route_to_coords(
            src_coords,
            dst_coords,
            mindim, mindir
        );

        assertEqual(unit, "next valiant step dim", dim, mindim);
        assertEqual(unit, "next valiant step dir", dir, mindir);

    }
}

