#include <sstmac/main/sstmac.h>

#include <sstmac/common/fileio.h>

#include <sstmac/backends/native/eventmap.h>
#include <sstmac/backends/native/manager.h>

#include <sstmac/common/parameters/file_parameters.h>
#include <sstmac/common/errors.h>
#include <sstmac/keywords/keyword_registration.h>

#include <sstmac/hardware/network/switches/networkswitch.h>
#include <sstmac/hardware/network/interconnect/switchinterconnect.h>

#include <sstmac/common/basicstringtokenizer.h>
#include <boost/lexical_cast.hpp>
#include <sys/types.h>
#include <sys/shm.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <getopt.h>

using namespace sstmac;
using namespace sstmac::hw;
using namespace sstmac::native;


using boost::lexical_cast;

static std::ostream&
operator<<(std::ostream &os, const opts &oo)
{
  if (oo.help)
    os << "  --help \\\n";
  if (oo.color)
    os << "  --color \\\n";
  os << "  --debug=\"" << oo.debug << "\" \\\n";
  os << "  --configfile=\"" << oo.configfile << "\" \\\n";
  os << "  --mode=\"" << oo.mode << "\" \\\n";
  os << "  --stoptime=" << oo.stoptime.sec();
  return os;
}

int
parse_opts(int argc, char **argv, opts &oo)
{
  int dompitest = 0;
  int printnodes = 0;
  int no_wall_time = 0;
  int print_params = 0;
  option gopt[] =
  {
  { "help", no_argument, NULL, 'h' },
  { "include", required_argument, NULL, 'i' },
  { "debug", required_argument, NULL, 'd' },
  { "stoptime", required_argument, NULL, 't' },
  { "configfile", required_argument, NULL, 'f' },
  { "runnumber", required_argument, NULL, 'r' },
  { "param", required_argument, NULL, 'p' },
  { "srun", required_argument, NULL, 's' },
  { "mode", required_argument, NULL, 'm' },
  { "color", no_argument, NULL, 'c' },
  { "print-apps", no_argument, NULL, 'P' },
  { "parse", required_argument, NULL, 'u' }, //u = undumpi
      { "network", required_argument, NULL, 'n' },
      { "mpitest", no_argument, &dompitest, 1 },
      { "print-nodes", no_argument, &printnodes, 1 },
      { "print-params", no_argument, &print_params, 1 },
      { "no-wall-time", no_argument, &no_wall_time, 1 },
      { NULL, 0, NULL, '\0' } };
  int ch;
  bool errorflag = false;
  std::list<std::pair<std::string, std::string> > paramlist;
  oo.params = command_parameters::construct();
  while ((ch = getopt_long(argc, argv, "Phcd:f:t:p:m:r:n:u:i:", gopt, NULL))
      != -1)
  {
    switch (ch)
    {
    case 0:
      //this set an input flag
      break;
    case 'h':
      oo.help = 1;
      break;
    case 'd':
      oo.debug = optarg;
      break;
    case 'r':
      oo.run_num = atoi(optarg);
      break;
    case 'f':
      oo.configfile = optarg;
      break;
    case 'u':
      oo.dumpimetafile = optarg;
      break;
    case 't':
      oo.stoptime = timestamp(atof(optarg));
      break;
    case 'n':
      oo.network = optarg;
      break;
    case 'P':
      oo.print_apps = true;
      break;
    case 'i':
    {
      file_parameters::ptr params = file_parameters::construct(optarg);
      params->combine_into(oo.params);
    }
      break;
    case 'p':
    {
      std::pair<std::string, std::string> keyval;
      sim_parameters::ptr scope = oo.params->parse_param(optarg, keyval);
      paramlist.push_back(keyval);
      scope->add_param(keyval.first, keyval.second);
      break;
    }
    case 'm':
      oo.mode = optarg;
      break;

    case 'c':
      oo.color = true;
      break;

    default:
      std::cerr << "Unhandled input flag.\n";
      errorflag = true;
      std::cout << "optarg: " << optarg << std::endl;
      break;
    }
    if (oo.help)
    {

      return PARSE_OPT_EXIT_SUCCESS;
    }
    if (errorflag)
    {
      std::cerr << "Error return\n";
      return PARSE_OPT_EXIT_FAIL;
    }
  }

  if (print_params)
  {
    oo.print_params = true;
  }

  /** check to see if we should do an mpitestall */
  if (dompitest)
  {
    if (oo.configfile != "")
    {
      std::cerr
          << "Cannot set config file with --configfile and also request --mpitest"
          << std::endl;
      return PARSE_OPT_EXIT_FAIL;
    }
    oo.configfile = "mpitestall.ini";

    if (oo.network == "")
    {
      std::cerr
          << "For --mpitest, you must specify a network and node model with --network"
          << std::endl;
      return PARSE_OPT_EXIT_FAIL;
    }
  }
  if (printnodes)
  {
    oo.params->add_param_override("launch_app1", "mpi_printnodes");
  }

  /** double negative, sorry about that */
  oo.print_walltime = !no_wall_time;

  /** Set sensible defaults for some of the paramters if not set*/
  if (oo.configfile == "")
    oo.configfile = "parameters.ini";

  std::cerr << argv[0] << " \\\n" << oo;
  if (paramlist.size())
  {
    for (std::list<std::pair<std::string, std::string> >::iterator i =
        paramlist.begin(); i != paramlist.end(); i++)
    {
      std::cerr << " \\\n  --param \"" << i->first << "=" << i->second << "\"";
    }
  }
  std::cerr << "\n";

  return PARSE_OPT_SUCCESS;
}

int
try_main(int argc, char **argv)
{
  opts oo;
  int parse_status = parse_opts(argc, argv, oo);
  if (parse_status == PARSE_OPT_SUCCESS)
  {
    //set up the search path
    SSTFileIO::init();

    if (oo.debug != "")
      logger::set_user_param(oo.debug);
    if (oo.color)
      logger::set_debug_coloring(true);

    //use the config file to set up file search paths
    size_t pos = oo.configfile.find_last_of('/');
    if (pos != std::string::npos)
    {
      std::string dir = oo.configfile.substr(0, pos + 1);
      SSTFileIO::add_path(dir);
    }

    KeywordRegistration::do_validation_ = false;

    file_parameters::ptr params = file_parameters::construct(oo.configfile);

    if (oo.params)
    { // there were command-line overrides
      oo.params->combine_into(params);
    }

    sstmac_env::params = params;

    /** initialize static variables */
    nodeid::init_statics();

    eventmap::ptr ev_map = eventmap::construct();

    ev_map->params_ = params;

    manager::ptr mgr = manager::construct(ev_map);

    manager::global = mgr;

    mgr->build_interconnect(params);

    mgr->setup();

    interconnect::ptr interconn = mgr->get_interconnect();
    switchinterconnect::ptr fab = boost::dynamic_pointer_cast<switchinterconnect>(interconn);
    if (!fab){
        sst_throw(illformederror, "MPI parallel partitioner only compatible with switch interconnect");
    }

    topology::maptype sw = fab->get_switches();
    topology::ptr top = fab->get_topology();

    int r = params->get_int_param("myrank");
    std::stringstream ss;
    ss << "partition" << r << ".metis";
    std::ofstream output;
    output.open(ss.str().c_str());

    long numsw = sw.size();
    long edges = top->radix() * numsw / 2; //divide by two to make it unidirectional

    output << numsw << " " << edges << std::endl;

    for (long a = 0; a < sw.size(); a++)
    {
      nodeaddress::ptr addr = nodeaddress::convert_from_uniqueid(a);
      networkswitch::ptr net = sw[addr];
      std::vector<nodeaddress::ptr> con = net->get_connected_switches();
      for (int i = 0; i < con.size(); i++)
      {
        output << (con[i]->unique_id() + 1) << " ";
      }
      output << std::endl;
    }
    output.close();

    mgr->done();

  }
  else if (parse_status == PARSE_OPT_EXIT_SUCCESS)
  {
    return 0;
  }
  else if (parse_status == PARSE_OPT_EXIT_FAIL)
  {
    return 1;
  }
  return 0;
}

int
main(int argc, char** argv)
{
    try{
        return try_main(argc, argv);
    }
  catch (const std::exception &e)
  {
    std::cout << argv[0] << ": caught exception while setting up simulation:\n"
        << e.what() << "\n";
    return 1;
  }
  catch (...)
  {
    std::cout << argv[0]
        << ": caught unknown exception while setting up simulation\n";
    return 1;
  }
  return 0;
}

