//------------------------------------------------------------------------------
// 
//     SST/Macro tools set
//
//------------------------------------------------------------------------------

#include "dumpi/libundumpi/libundumpi.h"
#include "analyzer/trace/node/callback/callback.h"
#include "analyzer/trace/node/record/callrecord.h"

namespace sstmac { namespace tools { namespace analyzer { 

#define START_TIME(dumpiTime) ((double)dumpiTime->start.sec + (double)dumpiTime->start.nsec * 1e-9)
#define END_TIME(dumpiTime) ((double)dumpiTime->stop.sec + (double)dumpiTime->stop.nsec * 1e-9)

// Create a CallRecord when receiving a callback from dumpi
template< typename ParamType, dumpi_function function>
int OnDumpiCallback(const ParamType* prm, uint16_t thread, const dumpi_time *cpu, const dumpi_time *wall, const dumpi_perfinfo *perf, void *uarg)
{
    // Recovery user argument
    MPIContext* context = static_cast<DumpiCallbackArg*>(uarg)->m_context;
    RecordCallback& callback = static_cast<DumpiCallbackArg*>(uarg)->m_callback;    

    // Construct call record
    CallRecord<ParamType> callRecord(
            context, 
            function,
            TimeRange(START_TIME(cpu), END_TIME(cpu)),
            TimeRange(START_TIME(wall), END_TIME(wall)),
            prm);

    // Recover context
    callRecord.RecoverContext();    

    // Notify
    callback(callRecord);
        
    return true;
}

void SetDumpiCallbacks(libundumpi_callbacks& cb)
{
    cb.on_send                      = OnDumpiCallback<dumpi_send, DUMPI_Send>;
    cb.on_recv                      = OnDumpiCallback<dumpi_recv, DUMPI_Recv>;
    cb.on_get_count                 = OnDumpiCallback<dumpi_get_count, DUMPI_Get_count>;
    cb.on_bsend                     = OnDumpiCallback<dumpi_bsend, DUMPI_Bsend>;
    cb.on_ssend                     = OnDumpiCallback<dumpi_ssend, DUMPI_Ssend>;
    cb.on_rsend                     = OnDumpiCallback<dumpi_rsend, DUMPI_Rsend>;
    cb.on_buffer_attach             = OnDumpiCallback<dumpi_buffer_attach, DUMPI_Buffer_attach>;
    cb.on_buffer_detach             = OnDumpiCallback<dumpi_buffer_detach, DUMPI_Buffer_detach>;
    cb.on_isend                     = OnDumpiCallback<dumpi_isend, DUMPI_Isend>;
    cb.on_ibsend                    = OnDumpiCallback<dumpi_ibsend, DUMPI_Ibsend>;
    cb.on_issend                    = OnDumpiCallback<dumpi_issend, DUMPI_Issend>;
    cb.on_irsend                    = OnDumpiCallback<dumpi_irsend, DUMPI_Irsend>;
    cb.on_irecv                     = OnDumpiCallback<dumpi_irecv, DUMPI_Irecv>;
    cb.on_wait                      = OnDumpiCallback<dumpi_wait, DUMPI_Wait>;
    cb.on_test                      = OnDumpiCallback<dumpi_test, DUMPI_Test>;
    cb.on_request_free              = OnDumpiCallback<dumpi_request_free, DUMPI_Request_free>;
    cb.on_waitany                   = OnDumpiCallback<dumpi_waitany, DUMPI_Waitany>;
    cb.on_testany                   = OnDumpiCallback<dumpi_testany, DUMPI_Testany>;
    cb.on_waitall                   = OnDumpiCallback<dumpi_waitall, DUMPI_Waitall>;
    cb.on_testall                   = OnDumpiCallback<dumpi_testall, DUMPI_Testall>;
    cb.on_waitsome                  = OnDumpiCallback<dumpi_waitsome, DUMPI_Waitsome>;
    cb.on_testsome                  = OnDumpiCallback<dumpi_testsome, DUMPI_Testsome>;
    cb.on_iprobe                    = OnDumpiCallback<dumpi_iprobe, DUMPI_Iprobe>;
    cb.on_probe                     = OnDumpiCallback<dumpi_probe, DUMPI_Probe>;
    cb.on_cancel                    = OnDumpiCallback<dumpi_cancel, DUMPI_Cancel>;
    cb.on_test_cancelled            = OnDumpiCallback<dumpi_test_cancelled, DUMPI_Test_cancelled>;
    cb.on_send_init                 = OnDumpiCallback<dumpi_send_init, DUMPI_Send_init>;
    cb.on_bsend_init                = OnDumpiCallback<dumpi_bsend_init, DUMPI_Bsend_init>;
    cb.on_ssend_init                = OnDumpiCallback<dumpi_ssend_init, DUMPI_Ssend_init>;
    cb.on_rsend_init                = OnDumpiCallback<dumpi_rsend_init, DUMPI_Rsend_init>;
    cb.on_recv_init                 = OnDumpiCallback<dumpi_recv_init, DUMPI_Recv_init>;
    cb.on_start                     = OnDumpiCallback<dumpi_start, DUMPI_Start>;
    cb.on_startall                  = OnDumpiCallback<dumpi_startall, DUMPI_Startall>;
    cb.on_sendrecv                  = OnDumpiCallback<dumpi_sendrecv, DUMPI_Sendrecv>;
    cb.on_sendrecv_replace          = OnDumpiCallback<dumpi_sendrecv_replace, DUMPI_Sendrecv_replace>;
    cb.on_type_contiguous           = OnDumpiCallback<dumpi_type_contiguous, DUMPI_Type_contiguous>;
    cb.on_type_vector               = OnDumpiCallback<dumpi_type_vector, DUMPI_Type_vector>;
    cb.on_type_hvector              = OnDumpiCallback<dumpi_type_hvector, DUMPI_Type_hvector>;
    cb.on_type_indexed              = OnDumpiCallback<dumpi_type_indexed, DUMPI_Type_indexed>;
    cb.on_type_hindexed             = OnDumpiCallback<dumpi_type_hindexed, DUMPI_Type_hindexed>;
    cb.on_type_struct               = OnDumpiCallback<dumpi_type_struct, DUMPI_Type_struct>;
    cb.on_address                   = OnDumpiCallback<dumpi_address, DUMPI_Address>;
    cb.on_type_extent               = OnDumpiCallback<dumpi_type_extent, DUMPI_Type_extent>;
    cb.on_type_size                 = OnDumpiCallback<dumpi_type_size, DUMPI_Type_size>;
    cb.on_type_lb                   = OnDumpiCallback<dumpi_type_lb, DUMPI_Type_lb>;
    cb.on_type_ub                   = OnDumpiCallback<dumpi_type_ub, DUMPI_Type_ub>;
    cb.on_type_commit               = OnDumpiCallback<dumpi_type_commit, DUMPI_Type_commit>;
    cb.on_type_free                 = OnDumpiCallback<dumpi_type_free, DUMPI_Type_free>;
    cb.on_get_elements              = OnDumpiCallback<dumpi_get_elements, DUMPI_Get_elements>;
    cb.on_pack                      = OnDumpiCallback<dumpi_pack, DUMPI_Pack>;
    cb.on_unpack                    = OnDumpiCallback<dumpi_unpack, DUMPI_Unpack>;
    cb.on_pack_size                 = OnDumpiCallback<dumpi_pack_size, DUMPI_Pack_size>;
    cb.on_barrier                   = OnDumpiCallback<dumpi_barrier, DUMPI_Barrier>;
    cb.on_bcast                     = OnDumpiCallback<dumpi_bcast, DUMPI_Bcast>;
    cb.on_gather                    = OnDumpiCallback<dumpi_gather, DUMPI_Gather>;
    cb.on_gatherv                   = OnDumpiCallback<dumpi_gatherv, DUMPI_Gatherv>;
    cb.on_scatter                   = OnDumpiCallback<dumpi_scatter, DUMPI_Scatter>;
    cb.on_scatterv                  = OnDumpiCallback<dumpi_scatterv, DUMPI_Scatterv>;
    cb.on_allgather                 = OnDumpiCallback<dumpi_allgather, DUMPI_Allgather>;
    cb.on_allgatherv                = OnDumpiCallback<dumpi_allgatherv, DUMPI_Allgatherv>;
    cb.on_alltoall                  = OnDumpiCallback<dumpi_alltoall, DUMPI_Alltoall>;
    cb.on_alltoallv                 = OnDumpiCallback<dumpi_alltoallv, DUMPI_Alltoallv>;
    cb.on_reduce                    = OnDumpiCallback<dumpi_reduce, DUMPI_Reduce>;
    cb.on_op_create                 = OnDumpiCallback<dumpi_op_create, DUMPI_Op_create>;
    cb.on_op_free                   = OnDumpiCallback<dumpi_op_free, DUMPI_Op_free>;
    cb.on_allreduce                 = OnDumpiCallback<dumpi_allreduce, DUMPI_Allreduce>;
    cb.on_reduce_scatter            = OnDumpiCallback<dumpi_reduce_scatter, DUMPI_Reduce_scatter>;
    cb.on_scan                      = OnDumpiCallback<dumpi_scan, DUMPI_Scan>;
    cb.on_group_size                = OnDumpiCallback<dumpi_group_size, DUMPI_Group_size>;
    cb.on_group_rank                = OnDumpiCallback<dumpi_group_rank, DUMPI_Group_rank>;
    cb.on_group_translate_ranks     = OnDumpiCallback<dumpi_group_translate_ranks, DUMPI_Group_translate_ranks>;
    cb.on_group_compare             = OnDumpiCallback<dumpi_group_compare, DUMPI_Group_compare>;
    cb.on_comm_group                = OnDumpiCallback<dumpi_comm_group, DUMPI_Comm_group>;
    cb.on_group_union               = OnDumpiCallback<dumpi_group_union, DUMPI_Group_union>;
    cb.on_group_intersection        = OnDumpiCallback<dumpi_group_intersection, DUMPI_Group_intersection>;
    cb.on_group_difference          = OnDumpiCallback<dumpi_group_difference, DUMPI_Group_difference>;
    cb.on_group_incl                = OnDumpiCallback<dumpi_group_incl, DUMPI_Group_incl>;
    cb.on_group_excl                = OnDumpiCallback<dumpi_group_excl, DUMPI_Group_excl>;
    cb.on_group_range_incl          = OnDumpiCallback<dumpi_group_range_incl, DUMPI_Group_range_incl>;
    cb.on_group_range_excl          = OnDumpiCallback<dumpi_group_range_excl, DUMPI_Group_range_excl>;
    cb.on_group_free                = OnDumpiCallback<dumpi_group_free, DUMPI_Group_free>;
    cb.on_comm_size                 = OnDumpiCallback<dumpi_comm_size, DUMPI_Comm_size>;
    cb.on_comm_rank                 = OnDumpiCallback<dumpi_comm_rank, DUMPI_Comm_rank>;
    cb.on_comm_compare              = OnDumpiCallback<dumpi_comm_compare, DUMPI_Comm_compare>;
    cb.on_comm_dup                  = OnDumpiCallback<dumpi_comm_dup, DUMPI_Comm_dup>;
    cb.on_comm_create               = OnDumpiCallback<dumpi_comm_create, DUMPI_Comm_create>;
    cb.on_comm_split                = OnDumpiCallback<dumpi_comm_split, DUMPI_Comm_split>;
    cb.on_comm_free                 = OnDumpiCallback<dumpi_comm_free, DUMPI_Comm_free>;
    cb.on_comm_test_inter           = OnDumpiCallback<dumpi_comm_test_inter, DUMPI_Comm_test_inter>;
    cb.on_comm_remote_size          = OnDumpiCallback<dumpi_comm_remote_size, DUMPI_Comm_remote_size>;
    cb.on_comm_remote_group         = OnDumpiCallback<dumpi_comm_remote_group, DUMPI_Comm_remote_group>;
    cb.on_intercomm_create          = OnDumpiCallback<dumpi_intercomm_create, DUMPI_Intercomm_create>;
    cb.on_intercomm_merge           = OnDumpiCallback<dumpi_intercomm_merge, DUMPI_Intercomm_merge>;
    cb.on_keyval_create             = OnDumpiCallback<dumpi_keyval_create, DUMPI_Keyval_create>;
    cb.on_keyval_free               = OnDumpiCallback<dumpi_keyval_free, DUMPI_Keyval_free>;
    cb.on_attr_put                  = OnDumpiCallback<dumpi_attr_put, DUMPI_Attr_put>;
    cb.on_attr_get                  = OnDumpiCallback<dumpi_attr_get, DUMPI_Attr_get>;
    cb.on_attr_delete               = OnDumpiCallback<dumpi_attr_delete, DUMPI_Attr_delete>;
    cb.on_topo_test                 = OnDumpiCallback<dumpi_topo_test, DUMPI_Topo_test>;
    cb.on_cart_create               = OnDumpiCallback<dumpi_cart_create, DUMPI_Cart_create>;
    cb.on_dims_create               = OnDumpiCallback<dumpi_dims_create, DUMPI_Dims_create>;
    cb.on_graph_create              = OnDumpiCallback<dumpi_graph_create, DUMPI_Graph_create>;
    cb.on_graphdims_get             = OnDumpiCallback<dumpi_graphdims_get, DUMPI_Graphdims_get>;
    cb.on_graph_get                 = OnDumpiCallback<dumpi_graph_get, DUMPI_Graph_get>;
    cb.on_cartdim_get               = OnDumpiCallback<dumpi_cartdim_get, DUMPI_Cartdim_get>;
    cb.on_cart_get                  = OnDumpiCallback<dumpi_cart_get, DUMPI_Cart_get>;
    cb.on_cart_rank                 = OnDumpiCallback<dumpi_cart_rank, DUMPI_Cart_rank>;
    cb.on_cart_coords               = OnDumpiCallback<dumpi_cart_coords, DUMPI_Cart_coords>;
    cb.on_graph_neighbors_count     = OnDumpiCallback<dumpi_graph_neighbors_count, DUMPI_Graph_neighbors_count>;
    cb.on_graph_neighbors           = OnDumpiCallback<dumpi_graph_neighbors, DUMPI_Graph_neighbors>;
    cb.on_cart_shift                = OnDumpiCallback<dumpi_cart_shift, DUMPI_Cart_shift>;
    cb.on_cart_sub                  = OnDumpiCallback<dumpi_cart_sub, DUMPI_Cart_sub>;
    cb.on_cart_map                  = OnDumpiCallback<dumpi_cart_map, DUMPI_Cart_map>;
    cb.on_graph_map                 = OnDumpiCallback<dumpi_graph_map, DUMPI_Graph_map>;
    cb.on_get_processor_name        = OnDumpiCallback<dumpi_get_processor_name, DUMPI_Get_processor_name>;
    cb.on_get_version               = OnDumpiCallback<dumpi_get_version, DUMPI_Get_version>;
    cb.on_errhandler_create         = OnDumpiCallback<dumpi_errhandler_create, DUMPI_Errhandler_create>;
    cb.on_errhandler_set            = OnDumpiCallback<dumpi_errhandler_set, DUMPI_Errhandler_set>;
    cb.on_errhandler_get            = OnDumpiCallback<dumpi_errhandler_get, DUMPI_Errhandler_get>;
    cb.on_errhandler_free           = OnDumpiCallback<dumpi_errhandler_free, DUMPI_Errhandler_free>;
    cb.on_error_string              = OnDumpiCallback<dumpi_error_string, DUMPI_Error_string>;
    cb.on_error_class               = OnDumpiCallback<dumpi_error_class, DUMPI_Error_class>;
    cb.on_wtime                     = OnDumpiCallback<dumpi_wtime, DUMPI_Wtime>;
    cb.on_wtick                     = OnDumpiCallback<dumpi_wtick, DUMPI_Wtick>;
    cb.on_init                      = OnDumpiCallback<dumpi_init, DUMPI_Init>;
    cb.on_finalize                  = OnDumpiCallback<dumpi_finalize, DUMPI_Finalize>;
    cb.on_initialized               = OnDumpiCallback<dumpi_initialized, DUMPI_Initialized>;
    cb.on_abort                     = OnDumpiCallback<dumpi_abort, DUMPI_Abort>;
    cb.on_close_port                = OnDumpiCallback<dumpi_close_port, DUMPI_Close_port>;
    cb.on_comm_accept               = OnDumpiCallback<dumpi_comm_accept, DUMPI_Comm_accept>;
    cb.on_comm_connect              = OnDumpiCallback<dumpi_comm_connect, DUMPI_Comm_connect>;
    cb.on_comm_disconnect           = OnDumpiCallback<dumpi_comm_disconnect, DUMPI_Comm_disconnect>;
    cb.on_comm_get_parent           = OnDumpiCallback<dumpi_comm_get_parent, DUMPI_Comm_get_parent>;
    cb.on_comm_join                 = OnDumpiCallback<dumpi_comm_join, DUMPI_Comm_join>;
    cb.on_comm_spawn                = OnDumpiCallback<dumpi_comm_spawn, DUMPI_Comm_spawn>;
    cb.on_comm_spawn_multiple       = OnDumpiCallback<dumpi_comm_spawn_multiple, DUMPI_Comm_spawn_multiple>;
    cb.on_lookup_name               = OnDumpiCallback<dumpi_lookup_name, DUMPI_Lookup_name>;
    cb.on_open_port                 = OnDumpiCallback<dumpi_open_port, DUMPI_Open_port>;
    cb.on_publish_name              = OnDumpiCallback<dumpi_publish_name, DUMPI_Publish_name>;
    cb.on_unpublish_name            = OnDumpiCallback<dumpi_unpublish_name, DUMPI_Unpublish_name>;
    cb.on_accumulate                = OnDumpiCallback<dumpi_accumulate, DUMPI_Accumulate>;
    cb.on_get                       = OnDumpiCallback<dumpi_get, DUMPI_Get>;
    cb.on_put                       = OnDumpiCallback<dumpi_put, DUMPI_Put>;
    cb.on_win_complete              = OnDumpiCallback<dumpi_win_complete, DUMPI_Win_complete>;
    cb.on_win_create                = OnDumpiCallback<dumpi_win_create, DUMPI_Win_create>;
    cb.on_win_fence                 = OnDumpiCallback<dumpi_win_fence, DUMPI_Win_fence>;
    cb.on_win_free                  = OnDumpiCallback<dumpi_win_free, DUMPI_Win_free>;
    cb.on_win_get_group             = OnDumpiCallback<dumpi_win_get_group, DUMPI_Win_get_group>;
    cb.on_win_lock                  = OnDumpiCallback<dumpi_win_lock, DUMPI_Win_lock>;
    cb.on_win_post                  = OnDumpiCallback<dumpi_win_post, DUMPI_Win_post>;
    cb.on_win_start                 = OnDumpiCallback<dumpi_win_start, DUMPI_Win_start>;
    cb.on_win_test                  = OnDumpiCallback<dumpi_win_test, DUMPI_Win_test>;
    cb.on_win_unlock                = OnDumpiCallback<dumpi_win_unlock, DUMPI_Win_unlock>;
    cb.on_win_wait                  = OnDumpiCallback<dumpi_win_wait, DUMPI_Win_wait>;
    cb.on_alltoallw                 = OnDumpiCallback<dumpi_alltoallw, DUMPI_Alltoallw>;
    cb.on_exscan                    = OnDumpiCallback<dumpi_exscan, DUMPI_Exscan>;
    cb.on_add_error_class           = OnDumpiCallback<dumpi_add_error_class, DUMPI_Add_error_class>;
    cb.on_add_error_code            = OnDumpiCallback<dumpi_add_error_code, DUMPI_Add_error_code>;
    cb.on_add_error_string          = OnDumpiCallback<dumpi_add_error_string, DUMPI_Add_error_string>;
    cb.on_comm_call_errhandler      = OnDumpiCallback<dumpi_comm_call_errhandler, DUMPI_Comm_call_errhandler>;
    cb.on_comm_create_keyval        = OnDumpiCallback<dumpi_comm_create_keyval, DUMPI_Comm_create_keyval>;
    cb.on_comm_delete_attr          = OnDumpiCallback<dumpi_comm_delete_attr, DUMPI_Comm_delete_attr>;
    cb.on_comm_free_keyval          = OnDumpiCallback<dumpi_comm_free_keyval, DUMPI_Comm_free_keyval>;
    cb.on_comm_get_attr             = OnDumpiCallback<dumpi_comm_get_attr, DUMPI_Comm_get_attr>;
    cb.on_comm_get_name             = OnDumpiCallback<dumpi_comm_get_name, DUMPI_Comm_get_name>;
    cb.on_comm_set_attr             = OnDumpiCallback<dumpi_comm_set_attr, DUMPI_Comm_set_attr>;
    cb.on_comm_set_name             = OnDumpiCallback<dumpi_comm_set_name, DUMPI_Comm_set_name>;
    cb.on_file_call_errhandler      = OnDumpiCallback<dumpi_file_call_errhandler, DUMPI_File_call_errhandler>;
    cb.on_grequest_complete         = OnDumpiCallback<dumpi_grequest_complete, DUMPI_Grequest_complete>;
    cb.on_grequest_start            = OnDumpiCallback<dumpi_grequest_start, DUMPI_Grequest_start>;
    cb.on_init_thread               = OnDumpiCallback<dumpi_init_thread, DUMPI_Init_thread>;
    cb.on_is_thread_main            = OnDumpiCallback<dumpi_is_thread_main, DUMPI_Is_thread_main>;
    cb.on_query_thread              = OnDumpiCallback<dumpi_query_thread, DUMPI_Query_thread>;
    cb.on_status_set_cancelled      = OnDumpiCallback<dumpi_status_set_cancelled, DUMPI_Status_set_cancelled>;
    cb.on_status_set_elements       = OnDumpiCallback<dumpi_status_set_elements, DUMPI_Status_set_elements>;
    cb.on_type_create_keyval        = OnDumpiCallback<dumpi_type_create_keyval, DUMPI_Type_create_keyval>;
    cb.on_type_delete_attr          = OnDumpiCallback<dumpi_type_delete_attr, DUMPI_Type_delete_attr>;
    cb.on_type_dup                  = OnDumpiCallback<dumpi_type_dup, DUMPI_Type_dup>;
    cb.on_type_free_keyval          = OnDumpiCallback<dumpi_type_free_keyval, DUMPI_Type_free_keyval>;
    cb.on_type_get_attr             = OnDumpiCallback<dumpi_type_get_attr, DUMPI_Type_get_attr>;
    cb.on_type_get_contents         = OnDumpiCallback<dumpi_type_get_contents, DUMPI_Type_get_contents>;
    cb.on_type_get_envelope         = OnDumpiCallback<dumpi_type_get_envelope, DUMPI_Type_get_envelope>;
    cb.on_type_get_name             = OnDumpiCallback<dumpi_type_get_name, DUMPI_Type_get_name>;
    cb.on_type_set_attr             = OnDumpiCallback<dumpi_type_set_attr, DUMPI_Type_set_attr>;
    cb.on_type_set_name             = OnDumpiCallback<dumpi_type_set_name, DUMPI_Type_set_name>;
    cb.on_type_match_size           = OnDumpiCallback<dumpi_type_match_size, DUMPI_Type_match_size>;
    cb.on_win_call_errhandler       = OnDumpiCallback<dumpi_win_call_errhandler, DUMPI_Win_call_errhandler>;
    cb.on_win_create_keyval         = OnDumpiCallback<dumpi_win_create_keyval, DUMPI_Win_create_keyval>;
    cb.on_win_delete_attr           = OnDumpiCallback<dumpi_win_delete_attr, DUMPI_Win_delete_attr>;
    cb.on_win_free_keyval           = OnDumpiCallback<dumpi_win_free_keyval, DUMPI_Win_free_keyval>;
    cb.on_win_get_attr              = OnDumpiCallback<dumpi_win_get_attr, DUMPI_Win_get_attr>;
    cb.on_win_get_name              = OnDumpiCallback<dumpi_win_get_name, DUMPI_Win_get_name>;
    cb.on_win_set_attr              = OnDumpiCallback<dumpi_win_set_attr, DUMPI_Win_set_attr>;
    cb.on_win_set_name              = OnDumpiCallback<dumpi_win_set_name, DUMPI_Win_set_name>;
    cb.on_alloc_mem                 = OnDumpiCallback<dumpi_alloc_mem, DUMPI_Alloc_mem>;
    cb.on_comm_create_errhandler    = OnDumpiCallback<dumpi_comm_create_errhandler, DUMPI_Comm_create_errhandler>;
    cb.on_comm_get_errhandler       = OnDumpiCallback<dumpi_comm_get_errhandler, DUMPI_Comm_get_errhandler>;
    cb.on_comm_set_errhandler       = OnDumpiCallback<dumpi_comm_set_errhandler, DUMPI_Comm_set_errhandler>;
    cb.on_file_create_errhandler    = OnDumpiCallback<dumpi_file_create_errhandler, DUMPI_File_create_errhandler>;
    cb.on_file_get_errhandler       = OnDumpiCallback<dumpi_file_get_errhandler, DUMPI_File_get_errhandler>;
    cb.on_file_set_errhandler       = OnDumpiCallback<dumpi_file_set_errhandler, DUMPI_File_set_errhandler>;
    cb.on_finalized                 = OnDumpiCallback<dumpi_finalized, DUMPI_Finalized>;
    cb.on_free_mem                  = OnDumpiCallback<dumpi_free_mem, DUMPI_Free_mem>;
    cb.on_get_address               = OnDumpiCallback<dumpi_get_address, DUMPI_Get_address>;
    cb.on_info_create               = OnDumpiCallback<dumpi_info_create, DUMPI_Info_create>;
    cb.on_info_delete               = OnDumpiCallback<dumpi_info_delete, DUMPI_Info_delete>;
    cb.on_info_dup                  = OnDumpiCallback<dumpi_info_dup, DUMPI_Info_dup>;
    cb.on_info_free                 = OnDumpiCallback<dumpi_info_free, DUMPI_Info_free>;
    cb.on_info_get                  = OnDumpiCallback<dumpi_info_get, DUMPI_Info_get>;
    cb.on_info_get_nkeys            = OnDumpiCallback<dumpi_info_get_nkeys, DUMPI_Info_get_nkeys>;
    cb.on_info_get_nthkey           = OnDumpiCallback<dumpi_info_get_nthkey, DUMPI_Info_get_nthkey>;
    cb.on_info_get_valuelen         = OnDumpiCallback<dumpi_info_get_valuelen, DUMPI_Info_get_valuelen>;
    cb.on_info_set                  = OnDumpiCallback<dumpi_info_set, DUMPI_Info_set>;
    cb.on_pack_external             = OnDumpiCallback<dumpi_pack_external, DUMPI_Pack_external>;
    cb.on_pack_external_size        = OnDumpiCallback<dumpi_pack_external_size, DUMPI_Pack_external_size>;
    cb.on_request_get_status        = OnDumpiCallback<dumpi_request_get_status, DUMPI_Request_get_status>;
    cb.on_type_create_darray        = OnDumpiCallback<dumpi_type_create_darray, DUMPI_Type_create_darray>;
    cb.on_type_create_hindexed      = OnDumpiCallback<dumpi_type_create_hindexed, DUMPI_Type_create_hindexed>;
    cb.on_type_create_hvector       = OnDumpiCallback<dumpi_type_create_hvector, DUMPI_Type_create_hvector>;
    cb.on_type_create_indexed_block = OnDumpiCallback<dumpi_type_create_indexed_block, DUMPI_Type_create_indexed_block>;
    cb.on_type_create_resized       = OnDumpiCallback<dumpi_type_create_resized, DUMPI_Type_create_resized>;
    cb.on_type_create_struct        = OnDumpiCallback<dumpi_type_create_struct, DUMPI_Type_create_struct>;
    cb.on_type_create_subarray      = OnDumpiCallback<dumpi_type_create_subarray, DUMPI_Type_create_subarray>;
    cb.on_type_get_extent           = OnDumpiCallback<dumpi_type_get_extent, DUMPI_Type_get_extent>;
    cb.on_type_get_true_extent      = OnDumpiCallback<dumpi_type_get_true_extent, DUMPI_Type_get_true_extent>;
    cb.on_unpack_external           = OnDumpiCallback<dumpi_unpack_external, DUMPI_Unpack_external>;
    cb.on_win_create_errhandler     = OnDumpiCallback<dumpi_win_create_errhandler, DUMPI_Win_create_errhandler>;
    cb.on_win_get_errhandler        = OnDumpiCallback<dumpi_win_get_errhandler, DUMPI_Win_get_errhandler>;
    cb.on_win_set_errhandler        = OnDumpiCallback<dumpi_win_set_errhandler, DUMPI_Win_set_errhandler>;
    cb.on_file_open                 = OnDumpiCallback<dumpi_file_open, DUMPI_File_open>;
    cb.on_file_close                = OnDumpiCallback<dumpi_file_close, DUMPI_File_close>;
    cb.on_file_delete               = OnDumpiCallback<dumpi_file_delete, DUMPI_File_delete>;
    cb.on_file_set_size             = OnDumpiCallback<dumpi_file_set_size, DUMPI_File_set_size>;
    cb.on_file_preallocate          = OnDumpiCallback<dumpi_file_preallocate, DUMPI_File_preallocate>;
    cb.on_file_get_size             = OnDumpiCallback<dumpi_file_get_size, DUMPI_File_get_size>;
    cb.on_file_get_group            = OnDumpiCallback<dumpi_file_get_group, DUMPI_File_get_group>;
    cb.on_file_get_amode            = OnDumpiCallback<dumpi_file_get_amode, DUMPI_File_get_amode>;
    cb.on_file_set_info             = OnDumpiCallback<dumpi_file_set_info, DUMPI_File_set_info>;
    cb.on_file_get_info             = OnDumpiCallback<dumpi_file_get_info, DUMPI_File_get_info>;
    cb.on_file_set_view             = OnDumpiCallback<dumpi_file_set_view, DUMPI_File_set_view>;
    cb.on_file_get_view             = OnDumpiCallback<dumpi_file_get_view, DUMPI_File_get_view>;
    cb.on_file_read_at              = OnDumpiCallback<dumpi_file_read_at, DUMPI_File_read_at>;
    cb.on_file_read_at_all          = OnDumpiCallback<dumpi_file_read_at_all, DUMPI_File_read_at_all>;
    cb.on_file_write_at             = OnDumpiCallback<dumpi_file_write_at, DUMPI_File_write_at>;
    cb.on_file_write_at_all         = OnDumpiCallback<dumpi_file_write_at_all, DUMPI_File_write_at_all>;
    cb.on_file_iread_at             = OnDumpiCallback<dumpi_file_iread_at, DUMPI_File_iread_at>;
    cb.on_file_iwrite_at            = OnDumpiCallback<dumpi_file_iwrite_at, DUMPI_File_iwrite_at>;
    cb.on_file_read                 = OnDumpiCallback<dumpi_file_read, DUMPI_File_read>;
    cb.on_file_read_all             = OnDumpiCallback<dumpi_file_read_all, DUMPI_File_read_all>;
    cb.on_file_write                = OnDumpiCallback<dumpi_file_write, DUMPI_File_write>;
    cb.on_file_write_all            = OnDumpiCallback<dumpi_file_write_all, DUMPI_File_write_all>;
    cb.on_file_iread                = OnDumpiCallback<dumpi_file_iread, DUMPI_File_iread>;
    cb.on_file_iwrite               = OnDumpiCallback<dumpi_file_iwrite, DUMPI_File_iwrite>;
    cb.on_file_seek                 = OnDumpiCallback<dumpi_file_seek, DUMPI_File_seek>;
    cb.on_file_get_position         = OnDumpiCallback<dumpi_file_get_position, DUMPI_File_get_position>;
    cb.on_file_get_byte_offset      = OnDumpiCallback<dumpi_file_get_byte_offset, DUMPI_File_get_byte_offset>;
    cb.on_file_read_shared          = OnDumpiCallback<dumpi_file_read_shared, DUMPI_File_read_shared>;
    cb.on_file_write_shared         = OnDumpiCallback<dumpi_file_write_shared, DUMPI_File_write_shared>;
    cb.on_file_iread_shared         = OnDumpiCallback<dumpi_file_iread_shared, DUMPI_File_iread_shared>;
    cb.on_file_iwrite_shared        = OnDumpiCallback<dumpi_file_iwrite_shared, DUMPI_File_iwrite_shared>;
    cb.on_file_read_ordered         = OnDumpiCallback<dumpi_file_read_ordered, DUMPI_File_read_ordered>;
    cb.on_file_write_ordered        = OnDumpiCallback<dumpi_file_write_ordered, DUMPI_File_write_ordered>;
    cb.on_file_seek_shared          = OnDumpiCallback<dumpi_file_seek_shared, DUMPI_File_seek_shared>;
    cb.on_file_get_position_shared  = OnDumpiCallback<dumpi_file_get_position_shared, DUMPI_File_get_position_shared>;
    cb.on_file_read_at_all_begin    = OnDumpiCallback<dumpi_file_read_at_all_begin, DUMPI_File_read_at_all_begin>;
    cb.on_file_read_at_all_end      = OnDumpiCallback<dumpi_file_read_at_all_end, DUMPI_File_read_at_all_end>;
    cb.on_file_write_at_all_begin   = OnDumpiCallback<dumpi_file_write_at_all_begin, DUMPI_File_write_at_all_begin>;
    cb.on_file_write_at_all_end     = OnDumpiCallback<dumpi_file_write_at_all_end, DUMPI_File_write_at_all_end>;
    cb.on_file_read_all_begin       = OnDumpiCallback<dumpi_file_read_all_begin, DUMPI_File_read_all_begin>;
    cb.on_file_read_all_end         = OnDumpiCallback<dumpi_file_read_all_end, DUMPI_File_read_all_end>;
    cb.on_file_write_all_begin      = OnDumpiCallback<dumpi_file_write_all_begin, DUMPI_File_write_all_begin>;
    cb.on_file_write_all_end        = OnDumpiCallback<dumpi_file_write_all_end, DUMPI_File_write_all_end>;
    cb.on_file_read_ordered_begin   = OnDumpiCallback<dumpi_file_read_ordered_begin, DUMPI_File_read_ordered_begin>;
    cb.on_file_read_ordered_end     = OnDumpiCallback<dumpi_file_read_ordered_end, DUMPI_File_read_ordered_end>;
    cb.on_file_write_ordered_begin  = OnDumpiCallback<dumpi_file_write_ordered_begin, DUMPI_File_write_ordered_begin>;
    cb.on_file_write_ordered_end    = OnDumpiCallback<dumpi_file_write_ordered_end, DUMPI_File_write_ordered_end>;
    cb.on_file_get_type_extent      = OnDumpiCallback<dumpi_file_get_type_extent, DUMPI_File_get_type_extent>;
    cb.on_register_datarep          = OnDumpiCallback<dumpi_register_datarep, DUMPI_Register_datarep>;
    cb.on_file_set_atomicity        = OnDumpiCallback<dumpi_file_set_atomicity, DUMPI_File_set_atomicity>;
    cb.on_file_get_atomicity        = OnDumpiCallback<dumpi_file_get_atomicity, DUMPI_File_get_atomicity>;
    cb.on_file_sync                 = OnDumpiCallback<dumpi_file_sync, DUMPI_File_sync>;
    cb.on_iotest                    = OnDumpiCallback<dumpio_test, DUMPIO_Test>;
    cb.on_iowait                    = OnDumpiCallback<dumpio_wait, DUMPIO_Wait>;
    cb.on_iotestall                 = OnDumpiCallback<dumpio_testall, DUMPIO_Testall>;
    cb.on_iowaitall                 = OnDumpiCallback<dumpio_waitall, DUMPIO_Waitall>;
    cb.on_iotestany                 = OnDumpiCallback<dumpio_testany, DUMPIO_Testany>;
    cb.on_iowaitany                 = OnDumpiCallback<dumpio_waitany, DUMPIO_Waitany>;
    cb.on_iowaitsome                = OnDumpiCallback<dumpio_waitsome, DUMPIO_Waitsome>;
    cb.on_iotestsome                = OnDumpiCallback<dumpio_testsome, DUMPIO_Testsome>;
    cb.on_function_enter            = OnDumpiCallback<dumpi_func_call, DUMPI_Function_enter>;
    cb.on_function_exit             = OnDumpiCallback<dumpi_func_call, DUMPI_Function_exit>;
}

} } } //namespace
