//------------------------------------------------------------------------------
// 
//     SST/Macro tools set
//
//------------------------------------------------------------------------------

#include "analyzer/trace/node/context/mpicontext.h"
#include "common/utilities/assert.h"

namespace sstmac { namespace tools { namespace analyzer {

template< typename ParamType >
CallRecord<ParamType>::CallRecord(MPIContext* context, int function, TimeRange cpuTime, TimeRange wallTime, const ParamType* param)
    : CallRecordBase(context, function, cpuTime, wallTime)
    , m_param(param)
{}

template< typename ParamType >
CallRecord<ParamType>::~CallRecord()
{}

template< typename ParamType >
bool CallRecord<ParamType>::RecoverContext()const
{
    //Do nothing in default implementation
    return true;
}

template< typename ParamType >
int CallRecord<ParamType>::GetBytesSend()const
{
    return 0;
}

template< typename ParamType >
int CallRecord<ParamType>::GetBytesSend(int /*nodeRank*/) const
{
    return 0;
}

template< typename ParamType >
void CallRecord<ParamType>::GetDest(std::vector<int>& nodeVec, bool /*local*/)const
{
    nodeVec.clear();
}

template< typename ParamType >
void CallRecord<ParamType>::GetSource(std::vector<int>& nodeVec, bool /*local*/)const
{
    nodeVec.clear();
}

template< typename ParamType >
bool CallRecord<ParamType>::IsCollective() const
{
    return false;
}

template< typename ParamType >
bool CallRecord<ParamType>::IsInit() const
{
    return false;
}

template< typename ParamType >
bool CallRecord<ParamType>::IsFinalize() const
{
    return false;
}

////////////////////////////////////////////////////////////////////////////////
// Recover context specialization
#define IMPLEMENT_RECOVERCONTEXT(FUNC)                  \
template<>                                              \
bool CallRecord<FUNC>::RecoverContext() const           \
{                                                       \
    return m_context->context_##FUNC(m_param);          \
}

IMPLEMENT_RECOVERCONTEXT(dumpi_comm_dup);
IMPLEMENT_RECOVERCONTEXT(dumpi_comm_create);
IMPLEMENT_RECOVERCONTEXT(dumpi_comm_split);
IMPLEMENT_RECOVERCONTEXT(dumpi_comm_free);
IMPLEMENT_RECOVERCONTEXT(dumpi_group_union);
IMPLEMENT_RECOVERCONTEXT(dumpi_group_intersection);
IMPLEMENT_RECOVERCONTEXT(dumpi_group_difference);
IMPLEMENT_RECOVERCONTEXT(dumpi_group_incl);
IMPLEMENT_RECOVERCONTEXT(dumpi_group_excl);
IMPLEMENT_RECOVERCONTEXT(dumpi_group_range_incl);
IMPLEMENT_RECOVERCONTEXT(dumpi_group_range_excl);
IMPLEMENT_RECOVERCONTEXT(dumpi_group_free);

////////////////////////////////////////////////////////////////////////////////
// GetDest specialization
#define IMPLEMENT_GETNODES_SEND(FUNC)                                                                       \
template<> void CallRecord<FUNC>::GetDest(std::vector<int>& nodeVec, bool local) const                      \
{                                                                                                           \
    nodeVec.resize(1);                                                                                      \
    nodeVec[0] = local ? m_param->dest : m_context->CommRankToWorldRank(m_param->dest, m_param->comm);      \
}                                                                                                           \

#define IMPLEMENT_GETNODES_RECV(FUNC)                                                                       \
template<> void CallRecord<FUNC>::GetSource(std::vector<int>& nodeVec, bool local) const                    \
{                                                                                                           \
    nodeVec.resize(1);                                                                                      \
    nodeVec[0] = local ? m_param->source : m_context->CommRankToWorldRank(m_param->source, m_param->comm);  \
}

#define IMPLEMENT_GETNODES_SCATTER(FUNC)                                                                \
template<> void CallRecord<FUNC>::GetDest(std::vector<int>& nodeVec, bool local) const                  \
{                                                                                                       \
    if(m_param->root != m_context->GetCommRank(m_param->comm))                                          \
    {                                                                                                   \
        nodeVec.clear();                                                                                \
        return;                                                                                         \
    }                                                                                                   \
    int commSize = m_context->GetCommSize(m_param->comm);                                               \
    nodeVec.resize(commSize);                                                                           \
    for (int i = 0; i < commSize; i++)                                                                  \
    {                                                                                                   \
        nodeVec[i] = local ? i : m_context->CommRankToWorldRank(i, m_param->comm);                      \
    }                                                                                                   \
}                                                                                                       \
template<> void CallRecord<FUNC>::GetSource(std::vector<int>& nodeVec, bool local) const                \
{                                                                                                       \
    nodeVec.resize(1);                                                                                  \
    nodeVec[0] = local ? m_param->root : m_context->CommRankToWorldRank(m_param->root, m_param->comm);  \
}

#define IMPLEMENT_GETNODES_GATHER(FUNC)                                                                 \
template<> void CallRecord<FUNC>::GetDest(std::vector<int>& nodeVec, bool local) const                  \
{                                                                                                       \
    if (m_param->comm != DUMPI_COMM_WORLD) return;                                                      \
    int rank = m_context->GetCommRank(m_param->comm);                                                   \
    if (rank == m_param->root) return;                                                                  \
    unsigned mask = 1;                                                                                  \
    for (int i = 0; i < m_context->GetLogSize(); i++, mask = mask << 1)                                 \
    {                                                                                                   \
        if (rank & mask) break;                                                                         \
    }                                                                                                   \
    nodeVec.push_back(local ? rank - mask: m_context->CommRankToWorldRank(rank - mask, m_param->comm)); \
}                                                                                                       \
template<> void CallRecord<FUNC>::GetSource(std::vector<int>& nodeVec, bool local) const                \
{                                                                                                       \
    if (m_param->comm != DUMPI_COMM_WORLD) return;                                                      \
    int rank = m_context->GetCommRank(m_param->comm);                                                   \
    unsigned mod = 2;                                                                                   \
    for (int i = 0; i < m_context->GetLogSize(); i++)                                                   \
    {                                                                                                   \
        int dest = rank + mod;                                                                          \
        if (rank % mod == 0 && dest < m_context->GetCommSize(m_param->comm))                            \
            nodeVec.push_back(local ? dest : m_context->CommRankToWorldRank(dest, m_param->comm));      \
    }                                                                                                   \
}

#define IMPLEMENT_GETNODES_ALLTOALL(FUNC)                                               \
template<> void CallRecord<FUNC>::GetDest(std::vector<int>& nodeVec, bool local) const  \
{                                                                                       \
    int commSize = m_context->GetCommSize(m_param->comm);                               \
    nodeVec.resize(commSize);                                                           \
    for (int i = 0; i < commSize; i++)                                                  \
    {                                                                                   \
        nodeVec[i] = local ? i : m_context->CommRankToWorldRank(i, m_param->comm);      \
    }                                                                                   \
}                                                                                       \
template<> void CallRecord<FUNC>::GetSource(std::vector<int>& nodeVec, bool local) const  \
{                                                                                       \
    int commSize = m_context->GetCommSize(m_param->comm);                               \
    nodeVec.resize(commSize);                                                           \
    for (int i = 0; i < commSize; i++)                                                  \
    {                                                                                   \
        nodeVec[i] = local ? i : m_context->CommRankToWorldRank(i, m_param->comm);      \
    }                                                                                   \
}

IMPLEMENT_GETNODES_SEND(dumpi_send);
IMPLEMENT_GETNODES_SEND(dumpi_bsend);
IMPLEMENT_GETNODES_SEND(dumpi_ssend);
IMPLEMENT_GETNODES_SEND(dumpi_rsend);
IMPLEMENT_GETNODES_SEND(dumpi_isend);
IMPLEMENT_GETNODES_SEND(dumpi_ibsend);
IMPLEMENT_GETNODES_SEND(dumpi_issend);
IMPLEMENT_GETNODES_SEND(dumpi_irsend);
IMPLEMENT_GETNODES_SEND(dumpi_sendrecv);
IMPLEMENT_GETNODES_SEND(dumpi_sendrecv_replace);

IMPLEMENT_GETNODES_RECV(dumpi_recv);
IMPLEMENT_GETNODES_RECV(dumpi_irecv);
IMPLEMENT_GETNODES_RECV(dumpi_sendrecv);
IMPLEMENT_GETNODES_RECV(dumpi_sendrecv_replace);

IMPLEMENT_GETNODES_SCATTER(dumpi_bcast);
IMPLEMENT_GETNODES_SCATTER(dumpi_scatter);
IMPLEMENT_GETNODES_SCATTER(dumpi_scatterv);
IMPLEMENT_GETNODES_GATHER(dumpi_gather);
IMPLEMENT_GETNODES_GATHER(dumpi_gatherv);
IMPLEMENT_GETNODES_GATHER(dumpi_reduce);
IMPLEMENT_GETNODES_ALLTOALL(dumpi_allgather);
IMPLEMENT_GETNODES_ALLTOALL(dumpi_allgatherv);
IMPLEMENT_GETNODES_ALLTOALL(dumpi_allreduce);
IMPLEMENT_GETNODES_ALLTOALL(dumpi_scan);
IMPLEMENT_GETNODES_ALLTOALL(dumpi_alltoall);
IMPLEMENT_GETNODES_ALLTOALL(dumpi_alltoallv);
IMPLEMENT_GETNODES_ALLTOALL(dumpi_alltoallw);
IMPLEMENT_GETNODES_ALLTOALL(dumpi_reduce_scatter);

//IMPLEMENT_GETDEST_SINGLE(dumpi_put, m_param->win); //TODO: translate window into destination rank
//IMPLEMENT_GETDEST_SINGLE(dumpi_accumulate, m_param->win);

////////////////////////////////////////////////////////////////////////////////
// Send bytes specialization
#define IMPLEMENT_BYTESSEND(FUNC, ...)                              \
template<> int CallRecord<FUNC>::GetBytesSend(int nodeRank) const   \
{                                                                   \
    ASSERT(nodeRank < m_context->GetCommSize(m_param->comm),        \
            "Node rank out of range when getting bytes send");      \
    return SendBytes(nodeRank, m_context, __VA_ARGS__);             \
}                                                                   \
template<> int CallRecord<FUNC>::GetBytesSend() const               \
{                                                                   \
    std::vector<int> nodeVec;                                       \
    GetDest(nodeVec, true);                                         \
    int total = 0;                                                  \
    for (unsigned int i = 0; i < nodeVec.size(); i++)               \
    {                                                               \
        total += GetBytesSend(nodeVec[i]);                          \
    }                                                               \
    return total;                                                   \
} 

inline int SendBytes(int /*nodeRank*/, const MPIContext* _context, int _count, dumpi_datatype _type)
{
    return _count * _context->GetTypeSize(_type);
}

inline int SendBytes(int nodeRank, const MPIContext* _context, int* _count, dumpi_datatype _type)
{
    return _count[nodeRank] * _context->GetTypeSize(_type);
}

inline int SendBytes(int nodeRank, const MPIContext* _context, int* _count, dumpi_datatype* _type)
{
    return _count[nodeRank] * _context->GetTypeSize(_type[nodeRank]);
}

IMPLEMENT_BYTESSEND(dumpi_send, m_param->count, m_param->datatype);
IMPLEMENT_BYTESSEND(dumpi_bsend, m_param->count, m_param->datatype);
IMPLEMENT_BYTESSEND(dumpi_ssend, m_param->count, m_param->datatype);
IMPLEMENT_BYTESSEND(dumpi_rsend, m_param->count, m_param->datatype);
IMPLEMENT_BYTESSEND(dumpi_isend, m_param->count, m_param->datatype);
IMPLEMENT_BYTESSEND(dumpi_ibsend, m_param->count, m_param->datatype);
IMPLEMENT_BYTESSEND(dumpi_issend, m_param->count, m_param->datatype);
IMPLEMENT_BYTESSEND(dumpi_irsend, m_param->count, m_param->datatype);
IMPLEMENT_BYTESSEND(dumpi_sendrecv, m_param->sendcount, m_param->sendtype);
IMPLEMENT_BYTESSEND(dumpi_sendrecv_replace, m_param->count, m_param->datatype);

IMPLEMENT_BYTESSEND(dumpi_bcast, m_param->count, m_param->datatype);
IMPLEMENT_BYTESSEND(dumpi_scatter, m_param->sendcount, m_param->sendtype);
IMPLEMENT_BYTESSEND(dumpi_scatterv, m_param->sendcounts, m_param->sendtype);
IMPLEMENT_BYTESSEND(dumpi_gather, m_param->sendcount, m_param->sendtype);
IMPLEMENT_BYTESSEND(dumpi_gatherv, m_param->sendcount, m_param->sendtype);
IMPLEMENT_BYTESSEND(dumpi_allgather, m_param->sendcount, m_param->sendtype);
IMPLEMENT_BYTESSEND(dumpi_allgatherv, m_param->sendcount, m_param->sendtype);
IMPLEMENT_BYTESSEND(dumpi_reduce, m_param->count, m_param->datatype);
IMPLEMENT_BYTESSEND(dumpi_allreduce, m_param->count, m_param->datatype);
IMPLEMENT_BYTESSEND(dumpi_scan, m_param->count, m_param->datatype);
IMPLEMENT_BYTESSEND(dumpi_alltoall, m_param->sendcount, m_param->sendtype);
IMPLEMENT_BYTESSEND(dumpi_alltoallv, m_param->sendcounts, m_param->sendtype);
IMPLEMENT_BYTESSEND(dumpi_alltoallw, m_param->sendcounts, m_param->sendtypes);
IMPLEMENT_BYTESSEND(dumpi_reduce_scatter, m_param->recvcounts, m_param->datatype);

//IMPLEMENT_BYTESSEND(dumpi_put, m_param->origincount, m_param->origintype); //TODO: Translate window into destination rank
//IMPLEMENT_BYTESSEND(dumpi_accumulate, m_param->origincount, m_param->origintype);

////////////////////////////////////////////////////////////////////////////////
// IsCollective specialization
#define IMPLEMENT_ISOFTYPE(FUNC, ISTYPE)                \
template<>                                              \
bool CallRecord<FUNC>::ISTYPE() const                   \
{                                                       \
    return true;                                        \
}

IMPLEMENT_ISOFTYPE(dumpi_barrier, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_bcast, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_gather, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_gatherv, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_scatter, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_scatterv, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_allgather, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_allgatherv, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_alltoall, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_alltoallv, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_reduce, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_allreduce, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_reduce_scatter, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_scan, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_comm_dup, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_comm_create, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_comm_split, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_comm_free, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_intercomm_create, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_init, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_finalize, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_comm_spawn, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_comm_spawn_multiple, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_alltoallw, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_exscan, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_init_thread, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_file_open, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_file_close, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_file_set_size, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_file_preallocate, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_file_set_info, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_file_set_view, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_file_read_at_all, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_file_write_at_all, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_file_read_all, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_file_write_all, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_file_read_ordered, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_file_seek_shared, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_file_read_at_all_begin, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_file_read_at_all_end, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_file_write_at_all_begin, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_file_write_at_all_end, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_file_read_all_begin, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_file_read_all_end, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_file_write_all_begin, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_file_write_all_end, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_file_read_ordered_begin, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_file_read_ordered_end, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_file_write_ordered_begin, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_file_write_ordered_end, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_file_set_atomicity, IsCollective);
IMPLEMENT_ISOFTYPE(dumpi_file_sync, IsCollective);

IMPLEMENT_ISOFTYPE(dumpi_init, IsInit);
IMPLEMENT_ISOFTYPE(dumpi_finalize, IsFinalize);

} } } //namespace
