#include <sstream>
#include <iomanip>
#include "common/log/logger.h"
#include "common/utilities/containerhelper.h"
#include "analyzer/trace/trace.h"
#include "analyzer/trace/meta/metadata.h"
#include "analyzer/trace/node/nodedata.h"

namespace sstmac { namespace tools { namespace analyzer {

bool Trace::Init(const std::string& metaPath)
{
    // Load metafile into memory
    if (!m_metaData.Read(metaPath))    
    {
        OUTPUT_ERROR("Failed to load metafile\n");
        return false; 
    }

    int numProcs = m_metaData.GetValue<int>("numprocs"); 
    std::string pathPrefix = GetNodeFilePathPrefix(metaPath, numProcs);

    if (numProcs <= 0 || pathPrefix.empty())
    {
        OUTPUT_ERROR("Invalid metafile!\n");
        return false;
    }

    // Load all nodedata files
    bool success = true;

    for (int i = 0; i < numProcs; i++) 
    {
        m_nodeDataVec.push_back(NodeData(i, numProcs));
        success &= m_nodeDataVec.back().SetFilePath(GetNodeFilePath(pathPrefix, i));
    }

    OUTPUT_ERROR_IF(!success, "Failed to open all trace files, aborting!\n");

    return success;
}

bool Trace::Parse(const RecordCallback& callback)
{
    bool success = true;

    for (unsigned int i = 0; i < m_nodeDataVec.size(); ++i)
    {
        success &= m_nodeDataVec[i].Parse(callback);
    }

    return success;
}

int Trace::GetNodeCount() const
{
    return m_nodeDataVec.size();
}

const StatsMeta& Trace::GetMetaData() const
{
    return m_metaData;
}

std::string Trace::GetNodeFilePathPrefix(const std::string& metaPath, int numProcs) const
{
    // Read info from metafile
    std::string filePrefix = m_metaData.GetValue<std::string>("fileprefix");
    
    // Remove exsiting dir prefix first
    std::string::size_type slashPos = filePrefix.find_last_of('/');
    if (slashPos != std::string::npos) 
    {
        filePrefix = filePrefix.substr(slashPos + 1);
    }

    // Append metafile path prefix to filePrefix
    slashPos = metaPath.find_last_of('/');
    if (slashPos != std::string::npos) 
    {
        filePrefix = metaPath.substr(0, slashPos + 1) + filePrefix;
    }
    
    return filePrefix;
}

std::string Trace::GetNodeFilePath(const std::string& pathPrefix, int nodeRank) const
{
    std::stringstream filePath;

    //TODO: the padding zeros may be more than 4
    filePath << pathPrefix << "-" << std::setfill('0') << std::setw(4) << nodeRank << ".bin";

    return filePath.str();
}

} } } //namespace
