//------------------------------------------------------------------------------
// 
//     SST/Macro tools set
//
//------------------------------------------------------------------------------

#include <sstream>
#include "common/report/reportstats.h"
#include "common/log/logger.h"
#include "common/stats/stats.h"
#include "common/stats/entry/entrydataimpl.h"
#include "common/factory/factory.h"
#include "common/utilities/serializeadapter.h"

namespace sstmac { namespace tools {

const char* ReportStats::ROOT_TAG = "DUMPI_Trace_Report";
const char* ReportStats::META_TAG = "Meta";
const char* ReportStats::STATS_TAG = "Statistics";
const char* ReportStats::STATS_ATTR_COUNT = "count";
const char* ReportStats::ENTRY_TAG = "Entry";
const char* ReportStats::ENTRY_ATTR_NAME = "name";

ReportStats::ReportStats(Stats& stats, EntryCreator entryCreator)
    : m_stats(stats)
    , m_entryCreator(entryCreator)
{}

bool ReportStats::ReadImpl(std::istream& is)
{
    is >> XmlHeader();

    StartTag rootStart;
    is >> rootStart; 

    if(rootStart.m_tag.compare(ROOT_TAG))
    {
        return false;
    }

    //Meta data
    StartTag metaStart;
    is >> metaStart;

    if(metaStart.m_tag.compare(META_TAG))
    {
        return false;
    }

    m_stats.GetMetaData() = metaStart.m_attr;

    EndTag metaEnd;
    is >> metaEnd;

    //Statistics
    StartTag statsStart;
    is >> statsStart;
    std::string* statsCountStr = statsStart.m_attr.Get(STATS_ATTR_COUNT);

    if(!statsCountStr || statsStart.m_tag.compare(STATS_TAG))
    {
        return false;
    }

    stringstream ss(*statsCountStr);
    int statsCount;
    ss >> statsCount;
    for (int i = 0; i < statsCount; i++) 
    {
        StartTag entryStart; 
        is >> entryStart;

        std::string* entryName = entryStart.m_attr.Get(ENTRY_ATTR_NAME);

        if(!entryName)
        {
            OUTPUT_WARNING("Failed to read entry\n");
            return false;
        }

        EntryDataBase::Ptr newData((*m_entryCreator)(*entryName));

        if(!newData)
        {
            OUTPUT_WARNING("Failed to read entry\n");
            return false;
        }

        is >> StartData() >> (*newData) >> EndData(); 
        m_stats.Register(newData->GetClassId(), newData);

        //std::cout << *newData << std::endl;

        EndTag entryEnd;
        is >> entryEnd;
    }

    EndTag statsEnd;
    is >> statsEnd;

    EndTag rootEnd;
    is >> rootEnd;

    return true;
}

bool ReportStats::WriteImpl(std::ostream& os) const
{
    os << XmlHeader();

    os << StartTag(ROOT_TAG);

    os << StartTag(META_TAG, m_stats.GetMetaData());
    os << EndTag(META_TAG);

    StartTag::AttributeHolder attr;
    stringstream ss;
    ss << m_stats.Size();
    attr.Register(STATS_ATTR_COUNT, ss.str());
    os << StartTag(STATS_TAG, attr);

    for (Stats::Iterator iter = m_stats.Begin(); iter != m_stats.End(); ++iter) 
    {
        StartTag::AttributeHolder attr;
        attr.Register(ENTRY_ATTR_NAME, iter->second->GetClassId());

        os << StartTag(ENTRY_TAG, attr); 
        os << StartData() << (*iter->second) << EndData();
        os << EndTag(ENTRY_TAG) << std::flush;
    }

    os << EndTag(STATS_TAG);
    os << EndTag(ROOT_TAG);

    return true;
}

EntryDataBase* ReportStats::DefaultEntryCreator(const std::string& classId)
{
    return Factory<EntryDataBase>::Singleton().Create(classId);
}

} } //namespace
