//------------------------------------------------------------------------------
// 
//     SST/Macro tools set
//
//------------------------------------------------------------------------------

#include "common/numeric/numericdefine.h"

namespace sstmac { namespace tools { namespace comparator {

////////////////////////////////////////////////////////////////////////////////
// General implementation
template<typename T>
bool CriterionGenerate<T>::Compare(const Stats& refStats, const Stats& simStats)
{
    const EntryDataBase::Ptr* refEntryBase = refStats.Get(T::EntryDataType::StaticClassId());
    const EntryDataBase::Ptr* simEntryBase = simStats.Get(T::EntryDataType::StaticClassId());

    if(!refEntryBase || !simEntryBase)
    {
        return false;
    }

    const typename T::EntryDataPtr refEntry = boost::static_pointer_cast<typename T::EntryDataType>(*refEntryBase);
    const typename T::EntryDataPtr simEntry = boost::static_pointer_cast<typename T::EntryDataType>(*simEntryBase);

    CompareImpl(refEntry, simEntry);

    return true;
}

template<typename T>
void CompareImpl(typename T::EntryDataPtr refEntry, typename T::EntryDataPtr simEntry)
{}

////////////////////////////////////////////////////////////////////////////////
// Overall
template<>
void CriterionGenerate<criterion::Overall>::CompareImpl(criterion::Overall::EntryDataPtr refEntry, criterion::Overall::EntryDataPtr simEntry)
{
    m_numericSimilarity = Gaussian(refEntry->m_bytes, simEntry->m_bytes);
    m_numericSimilarity *= Gaussian(refEntry->m_waitTime, simEntry->m_waitTime);
    m_numericSimilarity *= Gaussian(refEntry->m_computeTime, simEntry->m_computeTime);

    m_logicSimilarity = 1.0f;
}

////////////////////////////////////////////////////////////////////////////////
// ByNode
template<>
void CriterionGenerate<criterion::ByNode>::CompareImpl(criterion::ByNode::EntryDataPtr refEntry, criterion::ByNode::EntryDataPtr simEntry)
{
    unsigned int nodeCount = refEntry->m_bytes.size();
    m_logicSimilarity = Gaussian(nodeCount, simEntry->m_bytes.size());
    m_numericSimilarity = 1.0f;

    if (IsEqual(m_logicSimilarity, 1.0f))
    {
        m_bytesDiff.resize(nodeCount);
        m_waitTimeDiff.resize(nodeCount);
        m_computeTimeDiff.resize(nodeCount);

        for (unsigned int i = 0; i < nodeCount; i++) 
        {
            m_bytesDiff[i].Set(refEntry->m_bytes[i], simEntry->m_bytes[i]);
            m_waitTimeDiff[i].Set(refEntry->m_waitTime[i], simEntry->m_waitTime[i]);
            m_computeTimeDiff[i].Set(refEntry->m_computeTime[i], simEntry->m_computeTime[i]);

            m_numericSimilarity *= Gaussian(refEntry->m_bytes[i], simEntry->m_bytes[i]);
            m_numericSimilarity *= Gaussian(refEntry->m_waitTime[i], simEntry->m_waitTime[i]);
            m_numericSimilarity *= Gaussian(refEntry->m_computeTime[i], simEntry->m_computeTime[i]);
        }
    }
    else
    {
        m_nodeCount.first = refEntry->m_bytes.size(); 
        m_nodeCount.second = simEntry->m_bytes.size();
    }
}

////////////////////////////////////////////////////////////////////////////////
// MpiCall
template<>
void CriterionGenerate<criterion::MpiCall>::CompareImpl(criterion::MpiCall::EntryDataPtr refEntry, criterion::MpiCall::EntryDataPtr simEntry)
{
    unsigned int funcCount = refEntry->m_count.size();
    ASSERT(funcCount == simEntry->m_count.size(), "Expect same number of profiled calls");

    m_logicSimilarity = 1.0f;
    m_numericSimilarity = 1.0f;

    for (unsigned int i = 0; i < funcCount; i++) 
    {
        if(refEntry->m_count[i] != 0 || simEntry->m_count[i] != 0)
        {
            m_func.push_back(i);

            m_callCountDiff.push_back(Difference(refEntry->m_count[i], simEntry->m_count[i]));
            m_avgBlockTimeDiff.push_back(Difference(refEntry->m_avgBlockTime[i], simEntry->m_avgBlockTime[i]));
            m_sdBlockTimeDiff.push_back(Difference(refEntry->m_sdBlockTime[i], simEntry->m_sdBlockTime[i]));

            m_numericSimilarity *= Gaussian(refEntry->m_avgBlockTime[i], simEntry->m_avgBlockTime[i]);
            m_numericSimilarity *= Gaussian(refEntry->m_sdBlockTime[i], simEntry->m_sdBlockTime[i]);

            if(refEntry->m_count[i] != simEntry->m_count[i])
            {
                m_logicSimilarity *= Gaussian(refEntry->m_count[i], simEntry->m_count[i]);
            }
        }
    }
}

////////////////////////////////////////////////////////////////////////////////
// Collective
template<>
void CriterionGenerate<criterion::Collective>::CompareImpl(criterion::Collective::EntryDataPtr refEntry, criterion::Collective::EntryDataPtr simEntry)
{
    unsigned int collCount = refEntry->m_collective.size();
    m_logicSimilarity = Gaussian(collCount, simEntry->m_collective.size());
    m_numericSimilarity = 1.0f;

    if (IsEqual(m_logicSimilarity, 1.0f))
    {
        unsigned int logicDiff = 0;

        for (unsigned int i = 0; i < collCount; i++) 
        {
            if(refEntry->m_collective[i] == simEntry->m_collective[i])
            {
                m_collective.push_back(refEntry->m_collective[i]);
                m_avgTimeDiff.push_back(Difference(refEntry->m_avgTime[i], simEntry->m_avgTime[i]));
                m_sdTimeDiff.push_back(Difference(refEntry->m_sdTime[i], simEntry->m_sdTime[i]));

                m_numericSimilarity *= Gaussian(refEntry->m_avgTime[i], simEntry->m_avgTime[i]);
                m_numericSimilarity *= Gaussian(refEntry->m_sdTime[i], simEntry->m_sdTime[i]);
            }
            else
            {
                logicDiff++;
            }
        }

        m_logicSimilarity = Gaussian((float)(collCount - logicDiff), (float)collCount);
    }
    else
    {
        m_collCount.first = refEntry->m_collective.size();
        m_collCount.second = simEntry->m_collective.size();
    }
}

////////////////////////////////////////////////////////////////////////////////
// Transmission
template<>
void CriterionGenerate<criterion::TransmissionBytes>::CompareImpl(criterion::TransmissionBytes::EntryDataPtr refEntry, criterion::TransmissionBytes::EntryDataPtr simEntry)
{
    //unsigned int nodeCount = refEntry->m_spyPlot.size();
    //m_logicSimilarity = Gaussian(nodeCount, simEntry->m_spyPlot.size());
    //m_numericSimilarity = 1.0f;

    //if (IsEqual(m_logicSimilarity, 1.0f))
    //{
        //m_spyPlotDiff.resize(nodeCount);

        //for (unsigned int src = 0; src < nodeCount; src++) 
        //{
            //m_spyPlotDiff[src].resize(nodeCount);

            //for (unsigned int dest = 0; dest < nodeCount; dest++) 
            //{
                //double refBytes = refEntry->m_spyPlot[src][dest];
                //double simBytes = simEntry->m_spyPlot[src][dest];

                //m_spyPlotDiff[src][dest].Set(refBytes, simBytes);
                //m_numericSimilarity *= Gaussian(refBytes, simBytes);
            //}
        //}
    //}
    //else
    //{
        //m_nodeCount.first = refEntry->m_spyPlot.size();
        //m_nodeCount.second = simEntry->m_spyPlot.size();
    //}
}

} } } //namespace
