//------------------------------------------------------------------------------
// 
//     SST/Macro tools set
//
//------------------------------------------------------------------------------

#include <cstdio>
#include <cmath>
#include <plstream.h>
#include "common/factory/factory.h"
#include "common/utilities/dumpihelper.h"
#include "visualizer/plot/entry/entryplotimpl.h"

FACTORY_REGISTER_CLASS(sstmac::tools::visualizer, EntryPlotBase, sstmac::tools::visualizer, EntryPlotOverall);
FACTORY_REGISTER_CLASS(sstmac::tools::visualizer, EntryPlotBase, sstmac::tools::visualizer, EntryPlotByNode);
FACTORY_REGISTER_CLASS(sstmac::tools::visualizer, EntryPlotBase, sstmac::tools::visualizer, EntryPlotMpiCall);
FACTORY_REGISTER_CLASS(sstmac::tools::visualizer, EntryPlotBase, sstmac::tools::visualizer, EntryPlotCollective);
FACTORY_REGISTER_CLASS(sstmac::tools::visualizer, EntryPlotBase, sstmac::tools::visualizer, EntryPlotTransmission);

namespace sstmac { namespace tools { namespace visualizer {

static const int WHITE = 0;
static const int RED  = 1;
static const int YELLOW = 2;
static const int GREEN = 3;
static const int LGREEN = 4;
static const int PEACH = 5;
static const int BROWN = 6;
static const int GRAY = 7;
static const int DRED = 8;
static const int BLUE = 9;
static const int PURPLE = 10;
static const int CYAN = 11;
static const int LBLUE = 12;
static const int PINK = 13;
static const int ORANGE = 14;
static const int BLACK = 15;

////////////////////////////////////////////////////////////////////////////////
// Helper
void plfbox( plstream* pls, int color, PLFLT x0, PLFLT y0 )
{
    PLFLT x[4];
    PLFLT y[4];

    x[0] = x0;
    y[0] = 0.;
    x[1] = x0;
    y[1] = y0;
    x[2] = x0 + 1.;
    y[2] = y0;
    x[3] = x0 + 1.;
    y[3] = 0.;
    pls->fill( 4, x, y );
    pls->col0( color );
    pls->lsty( 1 );
    pls->line( 4, x, y );
}

////////////////////////////////////////////////////////////////////////////////
// Overall
void EntryPlotOverall::Plot(plstream* pls)
{
    double totalTime = m_waitTime + m_computeTime;
    double percent[2] = {m_waitTime / totalTime, m_computeTime / totalTime};
    int color[2] = {ORANGE, LGREEN};
    char label[2][20];

    sprintf(label[0], "%.0f%% Wait time", percent[0] * 100.0);
    sprintf(label[1], "%.0f%% Compute time", percent[1] * 100.0);

    static const PLFLT PLOT_SIZE = 10.0;
    static const PLFLT PLOT_CENTRE = 5.0;
    static const PLFLT PIE_RADIUS = 3.0;
    static const int POLY_COUNT = 400;
    static const PLFLT THETA_FACTOR = 2.0 * M_PI / POLY_COUNT;

    pls->adv(0);
    pls->vpas(0.1, 0.9, 0.1, 0.9, 1.0);
    pls->wind(0.0, PLOT_SIZE, 0.0, PLOT_SIZE);

    pls->col0(GRAY);
    pls->box("bc", 0, 0, "bc", 0, 0);

    pls->col0(BLACK);
    pls->lab("", "", "Timing distribution");

    int theta0 = 0;
    int dthet = 1;

    for (int i = 0; i < 2; i++) 
    {
        PLFLT x[POLY_COUNT + 2];
        PLFLT y[POLY_COUNT + 2];

        int j = 0;
        x[j] = PLOT_CENTRE;
        y[j++] = PLOT_CENTRE;

        int theta1 = theta0 + POLY_COUNT * percent[i];
        if (i == 1)
        {
            theta1 = POLY_COUNT;
        }

        int theta = theta0;
        for (; theta <= theta1; theta += dthet) 
        {
            x[j] = PLOT_CENTRE + PIE_RADIUS * cos(THETA_FACTOR * theta);
            y[j++] = PLOT_CENTRE + PIE_RADIUS * sin(THETA_FACTOR * theta);
        }

        pls->col0(color[i]);
        pls->psty(0);
        pls->fill(j, x, y);

        // Compute lable just and posision
        PLFLT just = THETA_FACTOR * (theta0 + theta1) / 2.0;
        PLFLT dx = 0.25 * cos(just);
        PLFLT dy = 0.25 * sin(just);

        if ( theta0 + theta1 < POLY_COUNT / 2 || theta0 + theta1 > POLY_COUNT + POLY_COUNT / 2 )
        {
            just = 0.0;
        }
        else
        {
            just = 1.0;
        }

        pls->col0(GRAY);
        pls->ptex(x[j / 2] + dx, y[j / 2] + dy, 1.0, 0.0, just, label[i]);

        theta0 = theta - dthet;
    }
}

////////////////////////////////////////////////////////////////////////////////
// ByNode
void EntryPlotByNode::Plot(plstream* pls)
{
    int nodeCount = m_waitTime.size();

    pls->adv(0);
    pls->vsta();
    pls->wind(0, nodeCount, 0.0, (m_avgWaitTime + m_sdWaitTime) * 1.2);

    pls->col0(GRAY);
    pls->box("bcint", 0, 0, "bcint", 0, 0);

    pls->col0(BLACK);
    pls->lab("Rank", "Time (seconds)", "Wait time by nodes");

    for (int i = 0; i < nodeCount; i++) 
    {
        pls->col0(CYAN);
        pls->psty(0);
        plfbox(pls, BLUE, i, m_waitTime[i]);
    }
}

////////////////////////////////////////////////////////////////////////////////
// MpiCall
void EntryPlotMpiCall::Plot(plstream* pls)
{
    int maxOccr = 0;

    std::vector<int> occuredFunc;

    for (unsigned int i = 0; i < m_count.size(); i++) 
    {
        if(m_count[i] > 0)
        {
            occuredFunc.push_back(i); 

            if(m_count[i] > maxOccr)
            {
                maxOccr = m_count[i];
            }
        }
    }

    unsigned int funcCount = occuredFunc.size();

    pls->adv(0);
    pls->vsta();
    pls->wind(0, funcCount, 0.0, maxOccr * 1.2);

    pls->col0(GRAY);
    pls->box("bcit", 1, 0, "bcint", 0, 0);

    pls->col0(BLACK);
    pls->lab("MPI function", "Count", "MPI Call Histogram");

    for (unsigned int i = 0; i < funcCount; i++) 
    {
        int func = occuredFunc[i];

        if (m_count[func] > 0)
        {
            pls->col0(CYAN);
            pls->psty(0);
            plfbox(pls, BLUE, i, m_count[func]);

            pls->col0(GRAY);
            pls->mtex("tv", -1.0, (i + 0.5) / funcCount, 1.0, Dumpi::GetName(func).c_str());
        }
    }
}

////////////////////////////////////////////////////////////////////////////////
// Collective
void EntryPlotCollective::Plot(plstream* pls)
{
    double maxPeriod = 0.0;

    for (unsigned int i = 0; i < m_avgTime.size(); i++) 
    {
        if(m_avgTime[i] > maxPeriod)
        {
            maxPeriod = m_avgTime[i];
        }
    }

    unsigned int regionCount = m_avgTime.size();

    pls->adv(0);
    pls->vsta();
    pls->wind(-1.0, regionCount + 1.0, 0.0, maxPeriod * 1.2);

    pls->col0(GRAY);
    pls->box("bcit", 1, 0, "bcint", 0, 0);

    pls->col0(BLACK);
    pls->lab("Collective Region", "Time (seconds)", "Collective Region Timing");

    for (unsigned int i = 0; i < regionCount; i++) 
    {
        pls->col0(CYAN);
        pls->psty(0);
        plfbox(pls, BLUE, i, m_avgTime[i]);

        //pls->col0(GRAY);
        //pls->mtex("tv", -1.0, (PLFLT)i / regionCount, 1.0, Dumpi::GetName(m_collective[i].m_beginFunc).c_str());
    }

    //pls->col0(GRAY);
    //pls->mtex("tv", -1.0, 1.0, 1.0, Dumpi::GetName(m_collective.back().m_endFunc).c_str());
}

////////////////////////////////////////////////////////////////////////////////
// Transmission
void EntryPlotTransmission::Plot(plstream* pls)
{
    int nodeCount = m_traffic.size();
    PLFLT** z;
    pls->Alloc2dGrid(&z, nodeCount, nodeCount);

    for (int i = 0; i < nodeCount; i++) 
    {
        for (int j = 0; j < nodeCount; j++) 
        {
            z[i][j] = m_traffic[i][j];
        }
    }

    pls->env(0.0, nodeCount, 0, nodeCount, 1, -1);

    pls->col0(GRAY);
    pls->box("bcint", 0, 0, "bcint", 0, 0);

    pls->col0(BLACK);
    pls->lab("", "", "Transmission Matrix");
    pls->image(z, nodeCount, nodeCount, 0.0, nodeCount, 0.0, nodeCount, 0.0, 0.0, 0.0, nodeCount, 0.0, nodeCount);
    pls->Free2dGrid(z, nodeCount, nodeCount);
}

} } } //namespace
